Package: ggpp
Type: Package
Title: Grammar Extensions to 'ggplot2'
Version: 0.6.1
Date: 2026-07-01
Authors@R: 
    c(
    person("Pedro J.", "Aphalo", email = "pedro.aphalo@helsinki.fi", role = c("aut", "cre"), comment = c(ORCID = "0000-0003-3385-972X")),
    person("Kamil", "Slowikowski", role = "ctb", comment = c(ORCID = "0000-0002-2843-6370")),
    person("Michał", "Krassowski", role = "ctb", comment = c(ORCID = "0000-0002-9638-7785")),
    person("Daniel", "Sabanés Bové", role = "ctb", comment = c(ORCID = "0000-0002-0176-9239")),
    person("Stella", "Banjo", role = "ctb")
    )
Maintainer: Pedro J. Aphalo <pedro.aphalo@helsinki.fi>
Description: Extensions to 'ggplot2' respecting the grammar of graphics 
    paradigm. Geometries: geom_table(), geom_plot() and geom_grob() add insets to 
    plots using native data coordinates, while geom_table_npc(), geom_plot_npc()
    and geom_grob_npc() do the same using "npc" coordinates through new 
    aesthetics "npcx" and "npcy". Statistics: select observations based on 2D 
    density. Positions: radial nudging away from a center point and nudging away
    from a line or curve; combined stacking and nudging; combined dodging and
    nudging.
License: GPL (>= 2)
LazyData: true
LazyLoad: true
ByteCompile: true
Depends: R (>= 4.1.0), ggplot2 (>= 3.5.0)
Imports: stats, grid, grDevices, rlang (>= 1.0.6), vctrs (>= 0.6.0),
        glue (>= 1.6.0), gridExtra (>= 2.3), scales (>= 1.3.0), tibble
        (>= 3.1.8), dplyr (>= 1.1.0), xts (>= 0.13.0), zoo (>= 1.8-11),
        MASS (>= 7.3-58), polynom (>= 1.4-0), lubridate (>= 1.9.0),
        stringr (>= 1.4.0)
Suggests: knitr (>= 1.40), rmarkdown (>= 2.20), ggrepel (>= 0.9.2),
        gginnards(>= 0.2.0), magick (>= 2.7.3), testthat (>= 3.1.5),
        vdiffr (>= 1.0.5)
URL: https://docs.r4photobiology.info/ggpp/,
        https://github.com/aphalo/ggpp
BugReports: https://github.com/aphalo/ggpp/issues
Encoding: UTF-8
VignetteBuilder: knitr
Collate: 'annotate.r' 'compute-npc.r' 'dark-or-light.R'
        'example-data.R' 'geom-grob.R' 'ggpp-legend-draw.R'
        'utilities.R' 'ggp2-margins.R' 'geom-label-linked.r'
        'geom-label-npc.r' 'geom-label-pairwise.r'
        'geom-margin-arrow.r' 'geom-margin-grob.r'
        'geom-margin-point.r' 'geom-plot.R' 'geom-point-linked.r'
        'geom-quadrant-lines.R' 'geom-table.R' 'geom-text-linked.r'
        'geom-text-npc.r' 'geom-text-pairwise.R' 'ggpp.R' 'onload.R'
        'position-dodge-nudge-to.R' 'position-dodge-nudge.R'
        'position-dodge2-nudge.R' 'position-dodge2nudge-to.R'
        'position-jitter-nudge.R' 'position-nudge-center.R'
        'position-nudge-line.R' 'position-stack-nudge.R'
        'position-stacknudge-to.R' 'scale-continuous-npc.r'
        'stat-apply.R' 'stat-dens1d-filter.r' 'stat-dens1d-labels.r'
        'stat-dens2d-filter.r' 'stat-dens2d-labels.r'
        'stat-format-table.R' 'stat-functions.R' 'stat-panel-counts.R'
        'stat-quadrant-counts.R' 'try-data-frame.R' 'weather-data.R'
        'wrap-labels.R'
Config/Needs/website: rmarkdown
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-01 22:30:55 UTC; aphalo
Author: Pedro J. Aphalo [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-3385-972X>),
  Kamil Slowikowski [ctb] (ORCID:
    <https://orcid.org/0000-0002-2843-6370>),
  Michał Krassowski [ctb] (ORCID:
    <https://orcid.org/0000-0002-9638-7785>),
  Daniel Sabanés Bové [ctb] (ORCID:
    <https://orcid.org/0000-0002-0176-9239>),
  Stella Banjo [ctb]
Repository: CRAN
Date/Publication: 2026-07-02 22:10:10 UTC
Built: R 4.6.1; ; 2026-07-02 22:42:22 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(ggplot,ts)
S3method(ggplot,xts)
S3method(heightDetails,titleGrob)
S3method(widthDetails,titleGrob)
export(GeomGrob)
export(GeomGrobNpc)
export(GeomLabelNpc)
export(GeomLabelPairwise)
export(GeomLabelS)
export(GeomPlot)
export(GeomPlotNpc)
export(GeomPointS)
export(GeomQuadrantLines)
export(GeomTable)
export(GeomTableNpc)
export(GeomTextNpc)
export(GeomTextPairwise)
export(GeomTextS)
export(GeomVHLines)
export(GeomXMarginArrow)
export(GeomXMarginGrob)
export(GeomXMarginPoint)
export(GeomYMarginArrow)
export(GeomYMarginGrob)
export(GeomYMarginPoint)
export(PositionDodge2AndNudgeTo)
export(PositionDodgeNudgeTo)
export(PositionFillAndNudge)
export(PositionFillNudgeTo)
export(PositionNudgeCenter)
export(PositionNudgeLine)
export(PositionStackNudgeTo)
export(StatApplyGroup)
export(StatDens1dFilter)
export(StatDens1dFilterG)
export(StatDens1dLabels)
export(StatDens2dFilter)
export(StatDens2dFilterG)
export(StatDens2dLabels)
export(StatFmtTb)
export(StatFunctions)
export(StatGroupCounts)
export(StatPanelCounts)
export(StatQuadrantCounts)
export(annotate)
export(as_npc)
export(as_npcx)
export(as_npcy)
export(compute_npc)
export(compute_npcx)
export(compute_npcy)
export(dark_or_light)
export(draw_key_label_s)
export(draw_key_text_s)
export(geom_grob)
export(geom_grob_npc)
export(geom_label_npc)
export(geom_label_pairwise)
export(geom_label_s)
export(geom_plot)
export(geom_plot_npc)
export(geom_point_s)
export(geom_quadrant_lines)
export(geom_table)
export(geom_table_npc)
export(geom_text_npc)
export(geom_text_pairwise)
export(geom_text_s)
export(geom_vhlines)
export(geom_x_margin_arrow)
export(geom_x_margin_grob)
export(geom_x_margin_point)
export(geom_y_margin_arrow)
export(geom_y_margin_grob)
export(geom_y_margin_point)
export(position_dodge2_keep)
export(position_dodge2nudge)
export(position_dodge2nudge_to)
export(position_dodge_keep)
export(position_dodgenudge)
export(position_dodgenudge_to)
export(position_fill_keep)
export(position_fillnudge)
export(position_fillnudge_to)
export(position_jitter_keep)
export(position_jitternudge)
export(position_nudge_center)
export(position_nudge_centre)
export(position_nudge_keep)
export(position_nudge_line)
export(position_nudge_to)
export(position_stack_keep)
export(position_stack_minmax)
export(position_stacknudge)
export(position_stacknudge_to)
export(scale_npcx_continuous)
export(scale_npcy_continuous)
export(stat_apply_group)
export(stat_centroid)
export(stat_dens1d_filter)
export(stat_dens1d_filter_g)
export(stat_dens1d_labels)
export(stat_dens2d_filter)
export(stat_dens2d_filter_g)
export(stat_dens2d_labels)
export(stat_fmt_tb)
export(stat_functions)
export(stat_group_counts)
export(stat_panel_counts)
export(stat_quadrant_counts)
export(stat_summary_xy)
export(try_data_frame)
export(try_tibble)
export(ttheme_gtbw)
export(ttheme_gtdark)
export(ttheme_gtdefault)
export(ttheme_gtlight)
export(ttheme_gtminimal)
export(ttheme_gtplain)
export(ttheme_gtsimple)
export(ttheme_gtstripes)
export(ttheme_set)
export(wrap_labels)
import(ggplot2)
import(grid)
import(scales)
importFrom(MASS,kde2d)
importFrom(dplyr,mutate)
importFrom(ggplot2,ggplot)
importFrom(polynom,polynomial)
importFrom(rlang,.data)

# ggpp - An Useful g++ Interface

## Table of Contents

- [About](#about)
- [Getting Started](#getting_started)
- [Usage](#usage)

## About <a name = "about"></a>

"ggpp" is a simple g++ interface made with python3, it is useful to run C/C++ code in only one command.<br>
So, instead of:<br>
`g++ example.cpp -o example`<br>
`./example`<br>
And sometimes you don't want the compiled file. So, you also do:<br>
`rm -rf example`<br>
All theses steps, can be reduced with ggpp to:<br>

```
ggpp example.cpp
```

And you will run your C/C++ code without (optionally) leaving a compiled file on the folder.<br>

## Getting Started <a name = "getting_started"></a>

Install ggpp is very simple.

### Prerequisites

You will need **python3** installed on your PC.<br>
For Python 3 installation, download on the [official page](https://www.python.org/downloads/).<br>

- Python 3 is already installed on the most linux distributions.<br>
  Try run "`python3`" on your shell.

As the ggpp is a **g++** interface, you will need it installed too.

As a plus, if you want to copy the content of you C/C++ code automatically after ran it, you will need **xclip** installed on linux. <br>
For install it, you can follow this [useful guide](https://linoxide.com/linux-how-to/copy-paste-commands-output-xclip-linux/). <br>
On Windows, to copy the content, you don't need any prerequisites.

### Installing

Download or clone this repository

```
git clone https://github.com/davigsousa/ggpp.git
```

Then, open the repository folder

```
cd ggpp
```

And run pip3 to install the package.

```
pip3 install .
```

Now, restart your shell and you will be able to use ggpp, try:

```
ggpp --help
```

## Usage <a name = "usage"></a>

You can run your C/C++ code,<br>
just passing the filename as first argument. On Windows, you need file path, like: .\targetfile.cpp <br>

```
ggpp targetfile.cpp
```

You can also pass arguments to the g++ just passing in the end.

```
ggpp targetfile.cpp -Wall -H
```

Currently, there are two arguments to configure ggpp,<br>
they are "-c" or "--clip" to copy file content to the clipboard (xclip is required on linux), <br>
And "-nr" or "--noremove" to make ggpp do not remove the compiled file.

Usage:

```
ggpp example.cpp -c -nr
```

utils::globalVariables(c(
  "from_theme",
  "colour",
  "ink",
  "family",
  "fill",
  "paper",
  "fontsize",
  "borderwidth",
  "bordertype",
  "linewidth",
  "linetype",
  "pointshape",
  "pointsize"
))

.onLoad <- function(libname, pkgname) {

  # if 'ggplot2' >= 4.0.0 we update the Geom "definitions" to retrieve defaults
  # from the theme, as the geom element is available
  if ("element_geom" %in% getNamespaceExports("ggplot2")) {
    ggplot2::update_geom_defaults(
      GeomGrob,
      ggplot2::aes(colour = from_theme(colour %||% ink),
                   family = from_theme(family)))
    ggplot2::update_geom_defaults(
      GeomGrobNpc,
      ggplot2::aes(colour = from_theme(colour %||% ink),
                   family = from_theme(family)))
    ggplot2::update_geom_defaults(
        GeomLabelS,
        ggplot2::aes(colour = from_theme(colour %||% ink),
                     fill = from_theme(fill %||% paper),
                     family = from_theme(family),
                     size = from_theme(fontsize),
                     linewidth = from_theme(borderwidth * 0.5),
                     linetype  = from_theme(bordertype)))
      ggplot2::update_geom_defaults(
        GeomLabelNpc,
        ggplot2::aes(colour = from_theme(colour %||% ink),
                     fill = from_theme(fill %||% paper),
                     family = from_theme(family),
                     size = from_theme(fontsize),
                     linewidth = from_theme(borderwidth * 0.5),
                     linetype  = from_theme(bordertype)))
      ggplot2::update_geom_defaults(
        GeomLabelPairwise,
        ggplot2::aes(colour = from_theme(colour %||% ink),
                     fill = from_theme(fill %||% paper),
                     family = from_theme(family),
                     size = from_theme(fontsize),
                     linewidth = from_theme(borderwidth * 0.5),
                     linetype  = from_theme(bordertype)))
      ggplot2::update_geom_defaults(
        GeomTextS,
        ggplot2::aes(colour = from_theme(colour %||% ink),
                     family = from_theme(family),
                     size = from_theme(fontsize)))
      ggplot2::update_geom_defaults(
        GeomTextNpc,
        ggplot2::aes(colour = from_theme(colour %||% ink),
                     family = from_theme(family),
                     size = from_theme(fontsize)))
      ggplot2::update_geom_defaults(
        GeomTextPairwise,
        ggplot2::aes(colour = from_theme(colour %||% ink),
                     family = from_theme(family),
                     size = from_theme(fontsize)))
      ggplot2::update_geom_defaults(
        GeomPointS,
        ggplot2::aes(shape = from_theme(pointshape),
                     colour = from_theme(colour %||% ink),
                     fill = from_theme(fill %||% NA),
                     size = from_theme(pointsize),
                     stroke = from_theme(borderwidth)))
      ggplot2::update_geom_defaults(
        GeomXMarginPoint,
        ggplot2::aes(shape = from_theme(pointshape),
                     colour = from_theme(colour %||% ink),
                     fill = from_theme(fill %||% NA),
                     size = from_theme(pointsize),
                     stroke = from_theme(borderwidth)))
      ggplot2::update_geom_defaults(
        GeomYMarginPoint,
        ggplot2::aes(shape = from_theme(pointshape),
                     colour = from_theme(colour %||% ink),
                     fill = from_theme(fill %||% NA),
                     size = from_theme(pointsize),
                     stroke = from_theme(borderwidth)))
      ggplot2::update_geom_defaults(
        GeomXMarginArrow,
        ggplot2::aes(colour = from_theme(colour %||% ink),
                     fill = from_theme(fill %||% paper),
                     linewidth = from_theme(linewidth)))
      ggplot2::update_geom_defaults(
        GeomYMarginArrow,
        ggplot2::aes(colour = from_theme(colour %||% ink),
                     fill = from_theme(fill %||% paper),
                     linewidth = from_theme(linewidth)))
      ggplot2::update_geom_defaults(
        GeomXMarginGrob,
        ggplot2::aes(colour = from_theme(colour %||% ink),
                     family = from_theme(family)))
      ggplot2::update_geom_defaults(
        GeomYMarginGrob,
        ggplot2::aes(colour = from_theme(colour %||% ink),
                     family = from_theme(family)))
      ggplot2::update_geom_defaults(
        GeomPlot,
        ggplot2::aes(colour = from_theme(colour %||% ink),
                     family = from_theme(family)))
      ggplot2::update_geom_defaults(
        GeomPlotNpc,
        ggplot2::aes(colour = from_theme(colour %||% ink),
                     family = from_theme(family)))
      ggplot2::update_geom_defaults(
        GeomTable,
        ggplot2::aes(family = from_theme(family),
                     size = from_theme(fontsize)))
      ggplot2::update_geom_defaults(
        GeomTableNpc,
        ggplot2::aes(family = from_theme(family),
                     size = from_theme(fontsize)))
      ggplot2::update_geom_defaults(
        GeomQuadrantLines,
        ggplot2::aes(colour = from_theme(colour %||% ink),
                     linewidth = from_theme(linewidth)))
      ggplot2::update_geom_defaults(
        GeomVHLines,
        ggplot2::aes(colour = from_theme(colour %||% ink),
                     linewidth = from_theme(linewidth)))
    }

}

#' @details Package 'ggpp' provides functions that extend the grammar of
#' graphics as implemented in 'ggplot2'. It attempts to stay true to the
#' original grammar and to respect the naming conventions used in 'ggplot2'.
#'
#' Extensions provided:
#' \itemize{
#' \item Geoms adding support for plot, table and grob insets within the
#' gramamr. Geoms using a parallel pseudo-scale based on native plot coordinates
#' (npc) to allow annotations consistent with the grammar and so supporting
#' facets and grouping. Geoms for annotations on the edges of the plotting
#' area. Geom for easily drawing lines separating the quadrants of a plot.
#' \item Stats for filtering-out/filtering-in observations in regions of a
#' panel or group where the density of observations is high. Statistics
#' simultaneously computing summaries, optionally using different functions,
#' along x and y. Stat computing quadrant counts.
#' \item Position functions implementing multi-directional nudging based on the
#' data.
#' \item Scales. Pseudo-scales supporting npc coordinates for x and y.
#' \item Specializations of the \code{ggplot()} generic accepting time series
#' objects of classes \code{ts} and \code{xts} as data argument.
#' }
#'
#' @section Acknowledgements: We thank Kamil Slowikowski not only for
#'   contributing ideas and code examples to this package but also for adding
#'   new features to his package 'ggrepel' that allow new use cases for
#'   \code{stat_dens2d_labels()}, \code{position_nudge_center()},
#'   \code{position_nudge_line()} and \code{position_nudge_to()} from this
#'   package. This package includes code copied and/or modified from
#'   that in package 'ggplot2'.
#'
#' @references
#' Package 'ggplot2' documentation is available at
#' \url{https://ggplot2.tidyverse.org/}\cr Package 'ggplot2' source code at
#' \url{https://github.com/tidyverse/ggplot2}
#'
#' @import scales grid ggplot2
#' @importFrom ggplot2 ggplot
#' @importFrom rlang .data
#'
#' @importFrom dplyr mutate
#' @importFrom polynom polynomial
#' @importFrom MASS kde2d
#'
"_PACKAGE"

## copied from ggplot2 GitHub development version on 2024-03-02
## needed for new labelGrob() definition

#' Create a text grob with the proper location and margins
#'
#' `titleGrob()` is called when creating titles and labels for axes, legends,
#' and facet strips.
#'
#' @param label Text to place on the plot. These maybe axis titles, axis labels,
#'   facet strip titles, etc.
#' @param x,y x and y locations where the text is to be placed.
#' @param hjust,vjust Horizontal and vertical justification of the text.
#' @param angle Angle of rotation of the text.
#' @param gp Additional graphical parameters in a call to `gpar()`.
#' @param margin Margins around the text. See [ggplot2::margin()] for more
#'   details.
#' @param margin_x,margin_y Whether or not to add margins in the x/y direction.
#' @param debug If `TRUE`, aids visual debugging by drawing a solid
#'   rectangle behind the complete text area, and a point where each label
#'   is anchored.
#'
#' @noRd
titleGrob <- function(label, x, y, hjust, vjust, angle = 0, gp = gpar(),
                      margin = NULL, margin_x = FALSE, margin_y = FALSE,
                      debug = FALSE, check.overlap = FALSE) {
  if (is.null(label)) {
    return(ggplot2::zeroGrob())
  }

  # We rotate the justification values to obtain the correct x and y reference point,
  # since hjust and vjust are applied relative to the rotated text frame in textGrob
  just <- rotate_just(angle, hjust, vjust)

  n <- max(length(x), length(y), 1)
  x <- x %||% grid::unit(rep(just$hjust, n), "npc")
  y <- y %||% grid::unit(rep(just$vjust, n), "npc")
  if (!grid::is.unit(x)) {
    x <- grid::unit(x, "npc")
  }
  if (!grid::is.unit(y)) {
    y <- grid::unit(y, "npc")
  }

  grob <- grid::textGrob(
    label, x, y,
    hjust = hjust, vjust = vjust,
    rot = angle, gp = gp, check.overlap = check.overlap
  )

  # The grob dimensions don't include the text descenders, so these need to be added
  # manually. Because descentDetails calculates the actual descenders of the specific
  # text label, which depends on the label content, we replace the label with one that
  # has the common letters with descenders. This guarantees that the grob always has
  # the same height regardless of whether the text actually contains letters with
  # descenders or not. The same happens automatically with ascenders already.
  descent <- font_descent(gp$fontfamily, gp$fontface, gp$fontsize, gp$cex)

  # Use trigonometry to calculate grobheight and width for rotated grobs. This is only
  # exactly correct when vjust = 1. We need to take the absolute value so we don't make
  # the grob smaller when it's flipped over.
  rad <- (angle[1] %% 360) / 180 * pi
  x_descent <- abs(sin(rad)) * descent
  y_descent <- abs(cos(rad)) * descent

  # Set text size to actual size including descenders
  width  <- grid::unit(1, "grobwidth",  grob) + x_descent
  height <- grid::unit(1, "grobheight", grob) + y_descent

  # Resolve margin
  if (is.null(margin)) {
    margin <- ggplot2::margin(0, 0, 0, 0)
  }
  margin_x <- isTRUE(margin_x)
  margin_y <- isTRUE(margin_y)

  # Initialise new values for position and dimensions
  new_x <- NULL
  new_y <- NULL
  new_width  <- NULL
  new_height <- NULL

  # Calculate new x/width
  if (margin_x) {
    new_width <- grid::unit.c(margin[4], width, margin[2])
    new_x <- x - margin[2] * just$hjust + margin[4] * (1 - just$hjust)
  }

  # Calculate new y/height
  if (margin_y) {
    new_height <- grid::unit.c(margin[1], height, margin[3])
    new_y <- y - margin[1] * just$vjust + margin[3] * (1 - just$vjust)
  }

  # If only one margin is set, the other dimension is a null unit
  if (xor(margin_x, margin_y)) {
    new_width  <- new_width  %||% grid::unit(1, "null")
    new_height <- new_height %||% grid::unit(1, "null")
  }

  # If we haven't touched the new positions/dimensions, use the previous ones
  new_width  <- new_width  %||% width
  new_height <- new_height %||% height
  x <- new_x %||% x
  y <- new_y %||% y

  # Adjust the grob
  grob$x <- x
  grob$y <- y

  # Add debug rectangles/points if necessary
  if (isTRUE(debug)) {
    children <- grid::gList(
      grid::rectGrob(
        x = x, y = y, width = width, height = height,
        hjust = just$hjust, vjust = just$vjust,
        gp = grid::gpar(fill = "cornsilk", col = NA)
      ),
      grid::pointsGrob(x, y, pch = 20, gp = grid::gpar(col = "gold")),
      grob
    )
  } else {
    children <- grid::gList(grob)
  }

  grid::gTree(
    children = children,
    widths   = new_width,
    heights  = new_height,
    cl = "titleGrob"
  )
}

#' @export
widthDetails.titleGrob <- function(x) {
  sum(x$widths)
}

#' @export
heightDetails.titleGrob <- function(x) {
  sum(x$heights)
}

#' Justifies a grob within a larger drawing area
#'
#' `justify_grobs()` can be used to take one or more grobs and draw them justified inside a larger
#' drawing area, such as the cell in a gtable. It is needed to correctly place [`titleGrob`]s
#' with margins.
#'
#' @param grobs The single grob or list of grobs to justify.
#' @param x,y x and y location of the reference point relative to which justification
#'   should be performed. If `NULL`, justification will be done relative to the
#'   enclosing drawing area (i.e., `x = hjust` and `y = vjust`).
#' @param hjust,vjust Horizontal and vertical justification of the grob relative to `x` and `y`.
#' @param int_angle Internal angle of the grob to be justified. When justifying a text
#'   grob with rotated text, this argument can be used to make `hjust` and `vjust` operate
#'   relative to the direction of the text.
#' @param debug If `TRUE`, aids visual debugging by drawing a solid
#'   rectangle behind the complete grob area.
#'
#' @noRd
justify_grobs <- function(grobs, x = NULL, y = NULL, hjust = 0.5, vjust = 0.5,
                          int_angle = 0, debug = FALSE) {
  if (!inherits(grobs, "grob")) {
    if (is.list(grobs)) {
      return(lapply(grobs, justify_grobs, x, y, hjust, vjust, int_angle, debug))
    }
    else {
      stop("Expected an individual 'grob' or list of 'grob' objects, but 'grobs' is a '", class(grobs)[1], "'")
    }
  }

  if (inherits(grobs, "zeroGrob")) {
    return(grobs)
  }

  # adjust hjust and vjust according to internal angle
  just <- rotate_just(int_angle, hjust, vjust)

  x <- x %||% grid::unit(just$hjust, "npc")
  y <- y %||% grid::unit(just$vjust, "npc")


  if (isTRUE(debug)) {
    children <- grid::gList(
      grid::rectGrob(gp = grid::gpar(fill = "lightcyan", col = NA)),
      grobs
    )
  }
  else {
    children = grid::gList(grobs)
  }


  result_grob <- grid::gTree(
    children = children,
    vp = viewport(
      x = x,
      y = y,
      width = grid::grobWidth(grobs),
      height = grid::grobHeight(grobs),
      just = unlist(just)
    )
  )


  if (isTRUE(debug)) {
    #cat("x, y:", c(x, y), "\n")
    #cat("E - hjust, vjust:", c(hjust, vjust), "\n")
    grid::grobTree(
      result_grob,
      grid::pointsGrob(x, y, pch = 20, gp = grid::gpar(col = "mediumturquoise"))
    )
  } else {
    result_grob
  }
}

#' Rotate justification parameters counter-clockwise
#' from 'ggplot2' 3.5.1
#'
#' @param angle angle of rotation, in degrees
#' @param hjust horizontal justification
#' @param vjust vertical justification
#' @return A list with two components, `hjust` and `vjust`, containing the rotated hjust and vjust values
#'
#' @noRd
rotate_just <- function(angle, hjust, vjust) {
  ## Ideally we would like to do something like the following commented-out lines,
  ## but it currently yields unexpected results for angles other than 0, 90, 180, 270.
  ## Problems arise in particular in cases where the horizontal and the vertical
  ## alignment model differ, for example, where horizontal alignment is relative to a
  ## point but vertical alignment is relative to an interval. This case arises for
  ## x and y axis tick labels.
  ##
  ## For more details, see: https://github.com/tidyverse/ggplot2/issues/2653

  # # convert angle to radians
  #rad <- (angle %||% 0) * pi / 180
  #
  #hnew <- cos(rad) * hjust - sin(rad) * vjust + (1 - cos(rad) + sin(rad)) / 2
  #vnew <- sin(rad) * hjust + cos(rad) * vjust + (1 - cos(rad) - sin(rad)) / 2

  angle <- (angle %||% 0) %% 360

  if (is.character(hjust)) {
    hjust <- match(hjust, c("left", "right")) - 1
    hjust[is.na(hjust)] <- 0.5
  }
  if (is.character(vjust)) {
    vjust <- match(vjust, c("bottom", "top")) - 1
    vjust[is.na(vjust)] <- 0.5
  }

  # Apply recycle rules
  size  <- vctrs::vec_size_common(angle, hjust, vjust)
  angle <- vctrs::vec_recycle(angle, size)
  hjust <- vctrs::vec_recycle(hjust, size)
  vjust <- vctrs::vec_recycle(vjust, size)

  # Find quadrant on circle
  case <- findInterval(angle, c(0, 90, 180, 270, 360))

  hnew <- hjust
  vnew <- vjust

  is_case <- which(case == 2) # 90 <= x < 180
  hnew[is_case] <- 1 - vjust[is_case]
  vnew[is_case] <- hjust[is_case]

  is_case <- which(case == 3) # 180 <= x < 270
  hnew[is_case] <- 1 - hjust[is_case]
  vnew[is_case] <- 1 - vjust[is_case]

  is_case <- which(case == 4) # 270 <= x < 360
  hnew[is_case] <- vjust[is_case]
  vnew[is_case] <- 1 - hjust[is_case]

  list(hjust = hnew, vjust = vnew)
}


descent_cache <- new.env(parent = emptyenv())
# Important: This function is not vectorized. Do not use to look up multiple
# font descents at once.
font_descent <- function(family = "", face = "plain", size = 12, cex = 1) {
  cur_dev <- names(grDevices::dev.cur())
  if (cur_dev == "null device") {
    cache <- FALSE   # don't cache if no device open
  } else {
    cache <- TRUE
  }
  key <- paste0(cur_dev, ':', family, ':', face, ":", size, ":", cex)
  # we only look up the first result; this function is not vectorized
  key <- key[1]

  descent <- descent_cache[[key]]

  if (is.null(descent)) {
    descent <- grid::convertHeight(grid::grobDescent(grid::textGrob(
      label = "gjpqyQ",
      gp = grid::gpar(
        fontsize = size,
        cex = cex,
        fontfamily = family,
        fontface = face
      )
    )), 'inches')

    if (cache) {
      descent_cache[[key]] <- descent
    }
  }

  descent
}

test_that("justify_grobs() checks input", {
  expect_error(justify_grobs(1))
})

test_that("justify_grobs() passes input", {
  expect_no_error(justify_grobs(grid::nullGrob()))
  expect_no_error(justify_grobs(grid::nullGrob(), debug = TRUE))
})

test_that("rotate_just() accepts NULL", {
  expect_no_error(rotate_just(NULL, 0.5, 0.5))
})

test_that("rotate_just() works as expected", {
  expect_equal(rotate_just(NULL, 1, 1), list(hjust = 1, vjust = 1))
  expect_equal(rotate_just(45, 1, 1), list(hjust = 1, vjust = 1))
  expect_equal(rotate_just(135, 1, 1), list(hjust = 0, vjust = 1))
  expect_equal(rotate_just(225, 1, 1), list(hjust = 0, vjust = 0))
  expect_equal(rotate_just(315, 1, 1), list(hjust = 1, vjust = 0))
})

#' Position scales for continuous data (npcx & npcy)
#'
#' \code{scale_npcx_continuous()} and \code{scale_npcy_continuous()} are
#' scales for continuous npcx and npcy aesthetics expressed in "npc" units.
#' There are no variants. Obviously limits are always the full range of "npc"
#' units and transformations meaningless. These scales are used by the newly
#' defined aesthetics \code{npcx} and \code{npcy}.
#'
#' @param ... Other arguments passed on to \code{continuous_scale()}
#' @name scale_continuous_npc
#'
#' @return A \code{"Scale"} object.
#'
#' @export
#'
scale_npcx_continuous <- function(...) {
  ggplot2::continuous_scale(
    aesthetics = "npcx",
    palette = identity, name = NULL, breaks = NULL,
    minor_breaks = NULL, labels = NULL, limits = c(NA_real_, NA_real_),
    expand = c(0, 0, 0, 0), oob = scales::censor, na.value = NA_real_, transform = "identity",
    guide = "none", position = "bottom", super = ggplot2::ScaleContinuousPosition,
    ...
  )
}

#' @rdname scale_continuous_npc
#' @export
scale_npcy_continuous <- function(...) {
  ggplot2::continuous_scale(
    aesthetics = "npcy",
    palette = identity, name = NULL, breaks = NULL,
    minor_breaks = NULL, labels = NULL, limits = c(NA_real_, NA_real_),
    expand = c(0, 0, 0, 0), oob = scales::censor, na.value = NA_real_, transform = "identity",
    guide = "none", position = "bottom", super = ggplot2::ScaleContinuousPosition,
    ...
  )
}

#' @rdname geom_text_s
#' @include ggp2-margins.R utilities.R ggpp-legend-draw.R
#'
#' @param label.padding Amount of padding around label. Defaults to 0.25 lines.
#' @param label.r Radius of rounded corners. Defaults to 0.15 lines.
#'
#' @export
#'
geom_label_s <-
  function(mapping = NULL,
           data = NULL,
           stat = "identity",
           position = "identity",
           ...,
           parse = FALSE,
           nudge_x = 0,
           nudge_y = 0,
           default.colour = NULL,
           default.color = default.colour,
           colour.target = c("text", "box"),
           color.target = colour.target,
           default.alpha = NA,
           alpha.target = "all",
           label.padding = grid::unit(0.25, "lines"),
           label.r = grid::unit(0.15, "lines"),
           segment.linewidth = 0.5,
           add.segments = TRUE,
           box.padding = 1e-06,
           point.padding = 1e-06,
           min.segment.length = 0,
           arrow = NULL,
           size.unit = "mm",
           na.rm = FALSE,
           show.legend = NA,
           inherit.aes = TRUE) {

    colour.target <-
      rlang::arg_match(color.target,
                       values = c("all", "text", "box", "box.line",
                                  "segment", "none"),
                       multiple = TRUE)
    alpha.target <-
      rlang::arg_match(alpha.target,
                       values = c("all", "text", "box", "box.line", "box.fill",
                                  "segment", "none"),
                       multiple = TRUE)

    if (!missing(nudge_x) || !missing(nudge_y)) {
      if (!missing(position) && position != "identity") {
        rlang::abort("You must specify either `position` or `nudge_x`/`nudge_y`.")
      }
      # by default we keep the original positions
      position <- position_nudge_keep(nudge_x, nudge_y)
    }

    ggplot2::layer(
    data = data,
    mapping = mapping,
    stat = stat,
    geom = GeomLabelS,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      parse = parse,
      default.colour = default.color,
      colour.target = colour.target,
      default.alpha = default.alpha,
      alpha.target = alpha.target,
      label.padding = label.padding,
      label.r = label.r,
      segment.linewidth = segment.linewidth,
      add.segments = add.segments,
      box.padding = box.padding,
      point.padding = point.padding,
      min.segment.length = min.segment.length,
      arrow = arrow,
      size.unit = size.unit,
      na.rm = na.rm,
      ...
    )
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
GeomLabelS <-
  ggplot2::ggproto("GeomLabelS", ggplot2::Geom,
                   required_aes = c("x", "y", "label"),

                   non_missing_aes = "angle",

                   default_aes = ggplot2::aes(
                     colour = "black",
                     fill = rgb(1, 1, 1, alpha = 0.75), # "white", but occluded data are visible
                     family = "",
                     size = 3.87,
                     angle = 0,
                     hjust = "position",
                     vjust = "position",
                     alpha = NA,
                     fontface = 1,
                     lineheight = 1.2,
                     linewidth = 0.5,
                     linetype  = "solid"
                   ),

                   draw_panel = function(data,
                                         panel_params,
                                         coord, #panel_scales,
                                         parse = FALSE,
                                         na.rm = FALSE,
                                         size.unit = "mm",
                                         add.segments = TRUE,
                                         default.colour = NULL,
                                         colour.target = "all",
                                         default.alpha = NA,
                                         alpha.target = "fill",
                                         box.padding = 0.25,
                                         point.padding = 1e-06,
                                         min.segment.length = 0,
                                         segment.linewidth = 0.5,
                                         arrow = NULL,
                                         label.padding = unit(0.25, "lines"),
                                         label.r = unit(0.15, "lines")) {

                     add.segments <-
                       add.segments &&
                       all(c("x_orig", "y_orig") %in% colnames(data))

                     default.colour <- check_default_colour(default.colour)

                     # ensure compatibility with 'ggplot2'
                     if (exists("label.size", data)) {
                       data$line.width <-
                         data$label.size * .pt / ggplot2::.stroke
                       data$label.size <- NULL
                     }
                     data$label <- as.character(data$label)
                     data <- subset(data, !is.na(label) & label != "")
                     if (nrow(data) == 0L) {
                       return(nullGrob())
                     }

                     lab <- data$label
                     if (parse) {
                       lab <- parse_safe(lab)
                     }

                     data <- coord$transform(data, panel_params)
                     if (all(c("x_orig", "y_orig") %in% colnames(data))) {
                       data_orig <- data.frame(x = data$x_orig, y = data$y_orig)
                       data_orig <- coord$transform(data_orig, panel_params)
                       data$x_orig <- data_orig$x
                       data$y_orig <- data_orig$y
                     }

                     if (is.character(data$vjust)) {
                       data$vjust <-
                         compute_just2d(data = data,
                                        coord = coord,
                                        panel_params = panel_params,
                                        just = data$vjust,
                                        a = "y", b = "x")
                     }
                     if (is.character(data$hjust)) {
                       data$hjust <-
                         compute_just2d(data = data,
                                        coord = coord,
                                        panel_params = panel_params,
                                        just = data$hjust,
                                        a = "x", b = "y")
                     }

                     size.unit <- resolve_text_unit(size.unit)

                     if (!inherits(label.padding, "margin")) {
                       label.padding <- rep(label.padding, length.out = 4)
                     }

                     if (add.segments) {
                       segments.data <-
                         shrink_segments(data,
                                         point.padding = point.padding,
                                         box.padding = box.padding,
                                         min.segment.length = min.segment.length)
                     }
                     # loop needed as gpar is not vectorized
                     all.grobs <- grid::gList()

                     for (row.idx in 1:nrow(data)) {
                       row <- data[row.idx, , drop = FALSE]
                       text.alpha <-
                         ifelse(any(alpha.target %in% c("all", "text")),
                                row$alpha, default.alpha)
                       segment.alpha <-
                         ifelse(any(alpha.target %in% c("all", "segment")),
                                row$alpha, default.alpha)
                       box.colour.alpha <-
                         ifelse(any(alpha.target %in% c("all", "box", "box.line")),
                                row$alpha, default.alpha)
                       box.fill.alpha <-
                         ifelse(any(alpha.target %in% c("all", "box", "box.fill")),
                                row$alpha, default.alpha)
                       user.grob <- labelGrob(lab[row.idx],
                                              x = unit(row$x, "native"),
                                              y = unit(row$y, "native"),
                                              just = c(row$hjust, row$vjust),
                                              padding = label.padding,
                                              r = label.r,
                                              angle = row$angle,
                                              text.gp = gpar(
                                                col = ifelse(any(colour.target %in% c("all", "text")),
                                                             ggplot2::alpha(row$colour, text.alpha),
                                                             ggplot2::alpha(default.colour, text.alpha)),
                                                fontsize = row$size * size.unit,
                                                fontfamily = row$family,
                                                fontface = row$fontface,
                                                lineheight = row$lineheight
                                              ),
                                              rect.gp = gpar(
                                                col = if (row$linewidth == 0) NA else # lwd = 0 is invalid in 'grid'
                                                  ifelse(any(colour.target %in% c("all", "box", "box.line")),
                                                         ggplot2::alpha(row$colour, box.colour.alpha),
                                                         ggplot2::alpha(default.colour, box.colour.alpha)),
                                                fill = alpha(row$fill, box.fill.alpha),
                                                # lwd = (if (row$linewidth == 0) 1 else row$linewidth) * .pt, # mm -> points (as in 'ggplot2')
                                                lwd = (if (row$linewidth == 0) 0.5 else row$linewidth) * ggplot2::.stroke, # mm -> stroke (correct)
                                                lty = row$linetype
                                              )
                       )

                       # give unique name to each grob
                       user.grob$name <- paste("text.s.grob", row$group, row.idx, sep = ".")

                       if (add.segments) {
                         segment.row <- segments.data[row.idx, , drop = FALSE]
                         if (segment.row$too.short) {
                           segment.grob <- grid::nullGrob()
                         } else {
                           segment.grob <-
                             grid::segmentsGrob(x0 = segment.row$x,
                                                y0 = segment.row$y,
                                                x1 = segment.row$x_orig,
                                                y1 = segment.row$y_orig,
                                                arrow = arrow,
                                                gp = grid::gpar(
                                                  col = if (segment.linewidth == 0) NA else # lwd = 0 is invalid in 'grid'
                                                    ifelse(any(colour.target %in% c("all", "segment")),
                                                           ggplot2::alpha(row$colour, segment.alpha),
                                                           ggplot2::alpha(default.colour, segment.alpha)),
                                                  lwd = (if (segment.linewidth == 0) 0.5 else segment.linewidth) * ggplot2::.stroke),
                                                name = paste("text.s.segment", row$group, row.idx, sep = "."))
                         }
                         all.grobs <- grid::gList(all.grobs, segment.grob, user.grob)
                       } else {
                         all.grobs <- grid::gList(all.grobs, user.grob)
                       }
                     }

                     # name needs to be unique within plot, so we would to know layer
                     #                     grid::grobTree(children = all.grobs, name = "geom.text.s.panel")
                     grid::grobTree(children = all.grobs)

                   },

                   draw_key = draw_key_label_s
  )

labelGrob <- function(label, x = grid::unit(0.5, "npc"), y = grid::unit(0.5, "npc"),
                      just = "center", padding = grid::unit(0.25, "lines"), r = grid::unit(0.1, "snpc"),
                      angle = NULL, default.units = "npc", name = NULL,
                      text.gp = grid::gpar(), rect.gp = grid::gpar(fill = "white"), vp = NULL) {

  if (length(label) != 1) {
    rlang::abort("label must be of length 1")
  }

  if (!grid::is.unit(x))
    x <- grid::unit(x, default.units)
  if (!grid::is.unit(y))
    y <- grid::unit(y, default.units)

  if (!is.null(angle) & is.null(vp)) {
    vp <- grid::viewport(
      angle = angle, x = x, y = y,
      width = grid::unit(0, "cm"), height = grid::unit(0, "cm"),
      gp = grid::gpar(fontsize = text.gp$fontsize)
    )
    x <- grid::unit(rep(0.5, length(x)), "npc")
    y <- grid::unit(rep(0.5, length(y)), "npc")
  }

  descent <- font_descent(
    text.gp$fontfamily, text.gp$fontface, text.gp$fontsize, text.gp$cex
  )
  # To balance labels, we ensure the top includes at least the descent height
  # and subtract the descent height from the bottom padding
  padding[1] <- grid::unit.pmax(padding[1], descent)
  padding[3] <- grid::unit.pmax(padding[3] - descent, unit(0, "pt"))

  hjust <- grid::resolveHJust(just, NULL)
  vjust <- grid::resolveVJust(just, NULL)

  text <- titleGrob(
    label = label, hjust = hjust, vjust = vjust, x = x,
    y = y + (1 - vjust) * descent,
    margin = padding, margin_x = TRUE, margin_y = TRUE,
    gp = text.gp
  )

  height <- grid::heightDetails(text)
  box <- grid::roundrectGrob(
    x = x, y = y + (0.5 - vjust) * height,
    width  = grid::widthDetails(text),
    height = height,
    just   = c(hjust, 0.5),
    r = r, gp = rect.gp, name = "box"
  )

  grid::gTree(children = grid::gList(box, text), name = name, vp = vp)
}

context("geom_label_s")

my.cars <- mtcars[c(TRUE, FALSE, FALSE, FALSE), ]
my.cars$name <- rownames(my.cars)

p <- ggplot(my.cars, aes(wt, mpg, label = name)) +
geom_point(color = "red")

test_that("geom_label_s runs successfully", {
  expect_no_error(
    p +
      geom_label_s(nudge_x = 0.12)
  )
})

test_that("geom_label_s pos_or_nudge", {
  expect_error(
    p +
      geom_label_s(nudge_x = 0.12, position = ""),
    "You must specify either `position` or `nudge_x`/`nudge_y`."
  )
})

#' @title Gene expression data
#'
#' @description A dataset containing reshaped and simplified output from an
#'   analysis of data from RNAseq done with package edgeR. Original data from
#'   gene expression in the plant species \emph{Arabidopsis thaliana}.
#'
#' @docType data
#' @keywords datasets
#' @format A \code{data.frame} object with 1218 rows and 5 variables
#' @family Transcriptomics data
#' @references Rai, Neha; O'Hara, Andrew; Farkas, Daniel; Safronov, Omid;
#' Ratanasopa, Khuanpiroon; Wang, Fang; Lindfors, Anders V.; Jenkins,
#' Gareth I.; Lehto, Tarja; Salojärvi, Jarkko; Brosché, Mikael; Strid. Åke;
#'  Aphalo, Pedro José; Morales, Luis Orlando (2020) The photoreceptor UVR8
#'  mediates the perception of both UV-B and UV-A wavelengths up to 350 nm of
#'  sunlight with responsivity moderated by cryptochromes. \emph{Plant, Cell &
#'  Environment}, 43:1513-1527.
#'
#' @examples
#' colnames(volcano_example.df)
#' head(volcano_example.df)
#'
"volcano_example.df"

#' @title Gene expression data
#'
#' @description A dataset containing reshaped and simplified output from an
#'   analysis of data from RNAseq done with package edgeR. Original data from
#'   gene expression in the plant species \emph{Arabidopsis thaliana}.
#'
#' @docType data
#' @keywords datasets
#' @format A \code{data.frame} object with 6088 rows and 6 variables
#' @family Transcriptomics data
#' @references Rai, Neha; O'Hara, Andrew; Farkas, Daniel; Safronov, Omid;
#' Ratanasopa, Khuanpiroon; Wang, Fang; Lindfors, Anders V.; Jenkins,
#' Gareth I.; Lehto, Tarja; Salojärvi, Jarkko; Brosché, Mikael; Strid. Åke;
#'  Aphalo, Pedro José; Morales, Luis Orlando (2020) The photoreceptor UVR8
#'  mediates the perception of both UV-B and UV-A wavelengths up to 350 nm of
#'  sunlight with responsivity moderated by cryptochromes. \emph{Plant, Cell &
#'  Environment}, 43:1513-1527.
#'
#' @examples
#' colnames(quadrant_example.df)
#' head(quadrant_example.df)
#'
"quadrant_example.df"

#' @title Birch seedlings' size
#'
#' @description A dataset containing the measurements on 350 birch seedlings.
#'
#' @details The data are for seedlings grown in trays with cells or containers
#' of two different volumes. For each of these types of trays, all cells, 1/2 of
#' the cells or 1/4 of the cells contained seedlings. Root-collar diameter (mm),
#' height (cm), dry mass (mg) of stems and roots. Measurements done at the end
#' of the first growing season, after leaf fall.
#'
#' @references Aphalo, P. J. and Rikala, R. (2003) Field performance of
#' silver-birch planting-stock grown at different spacing and in containers of
#' different volume. \emph{New Forests}, 25:93-108. \doi{10.1023/A:1022618810937}.
#'
#' @docType data
#' @keywords datasets
#' @format A \code{data.frame} object with 350 rows and 8 variables.
#' @family Plant growth and morphology data
#'
#' @examples
#' colnames(birch.df)
#' head(birch.df)
#'
#' colnames(birch_dw.df)
#' head(birch_dw.df)
#'
"birch.df"

#' @rdname birch.df
#' @format A \code{data.frame} object with 700 rows and 5 variables.
#'
"birch_dw.df"

#' @title Ivy photosynthesis light response
#'
#' @description A dataset containing photosynthesis measurements on four ivy plants.
#'
#' @details For each plant a light response curve of photosynthesis was measured
#' using a custom-built system and software that allowed controlling the
#' concentrations of water vapour and carbon dioxide at the surface of the
#' leaves, i.e., inside the air boundary layer.
#'
#' @references Aphalo, P. J. (1991) Interactions in Stomatal Function. PhD
#' thesis, University of Edinburgh. \url{http://hdl.handle.net/1842/14758}.
#'
#' @docType data
#' @keywords datasets
#' @format A \code{data.frame} object with 36 rows and 6 variables.
#' @family Plant growth and morphology data
#'
#' @examples
#' colnames(ivy.df)
#' head(ivy.df)
#'
"ivy.df"

#' Add Grobs on the margins
#'
#' Margin grobs can supplement a 2d display with annotations. Margin grobs such
#' as icons or symbols can highlight individual values along a margin. The
#' geometries \code{geom_x_margin_grob()} and \code{geom_y_margin_grob()} behave
#' similarly \code{geom_vline()} and \code{geom_hline()} and share their "double
#' personality" as both annotations and geometries.
#'
#' @param mapping The aesthetic mapping, usually constructed with
#'   \code{\link[ggplot2]{aes}}. Only needs to be set at the layer level if you
#'   are overriding the plot defaults.
#' @param data A layer specific dataset - only needed if you want to override
#'   the plot defaults.
#' @param stat The statistical transformation to use on the data for this layer,
#'   as a string.
#' @param na.rm If \code{FALSE} (the default), removes missing values with a
#'   warning.  If \code{TRUE} silently removes missing values.
#' @param position Position adjustment, either as a string, or the result of a
#'   call to a position adjustment function.
#' @param ... other arguments passed on to \code{\link[ggplot2]{layer}}. This
#'   can include aesthetics whose values you want to set, not map. See
#'   \code{\link[ggplot2]{layer}} for more details.
#' @param show.legend logical. Should this layer be included in the legends?
#'   \code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
#'   never includes, and \code{TRUE} always includes.
#' @param inherit.aes If \code{FALSE}, overrides the default aesthetics, rather
#'   than combining with them. This is most useful for helper functions that
#'   define both data and aesthetics and shouldn't inherit behaviour from the
#'   default plot specification, e.g. \code{\link[ggplot2]{borders}}.
#' @param sides A character string of length one that controls on which side of
#'   the plot the grob annotations appear on. It can be set to a string
#'   containing one of \code{"t"}, \code{"r"}, \code{"b"} or \code{"l"}, for
#'   top, right, bottom, and left.
#' @param grob.shift numeric value expressed in npc units for the shift of the
#'   marginal grob inwards from the edge of the plotting area.
#' @param xintercept,yintercept numeric Parameters that control the position of
#'   the marginal points. If these are set, data, mapping and show.legend are
#'   overridden.
#'
#' @inheritSection geom_text_s Alignment
#'
#' @inheritSection geom_text_s Position functions
#'
#' @family Geometries for marginal annotations in ggplots
#'
#' @return A plot layer instance.
#'
#' @seealso \code{\link[grid]{grid-package}}, \code{\link[ggplot2]{geom_rug}},
#'   and other documentation of package 'ggplot2'.
#'
#' @export
#'
#' @examples
#' # We can add icons to the margin of a plot to signal events
#'
#'
#'
#'
geom_x_margin_grob <- function(mapping = NULL, data = NULL,
                               stat = "identity", position = "identity",
                               ...,
                               xintercept,
                               sides = "b",
                               grob.shift = 0,
                               na.rm = FALSE,
                               show.legend = FALSE,
                               inherit.aes = FALSE) {

  # Act like an annotation
  if (!missing(xintercept)) {
    data <- as.data.frame(list(xintercept = xintercept))
    mapping <- aes(xintercept = xintercept)
    show.legend <- FALSE
  }

  layer(
    data = data,
    mapping = mapping,
    stat = stat,
    geom = GeomXMarginGrob,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      sides = stringr::str_sub(sides, 1L, 1L),
      grob.shift = grob.shift,
      na.rm = na.rm,
      ...
    )
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
GeomXMarginGrob <-
  ggproto("GeomXMarginGrob", Geom,
          required_aes = c("xintercept"),
          default_aes = ggplot2::aes(
            colour = "black",
            family = "",
            fontface = 1,
            angle = 0,
            hjust = 0.5,
            vjust = 0.5,
            alpha = NA,
            vp.width = 1/5, vp.height = 1/5
          ),

          draw_panel = function(data, panel_params, coord, sides = "b",
                                grob.shift = 0.01, na.rm = FALSE) {
            if (nrow(data) == 0) {
              return(grid::nullGrob())
            }

            if (!grid::is.grob(data$label[[1]])) {
              warning("Skipping as object mapped to 'label' is not a list of \"grob\" objects.")
              return(grid::nullGrob())
            }

            data <- coord$transform(data, panel_params)
            if (is.character(data$vjust)) {
              data$vjust <-
                compute_just2d(data = data,
                               coord = coord,
                               panel_params = panel_params,
                               just = data$vjust,
                               a = "y", b = "x")
            }
            if (is.character(data$hjust)) {
              data$hjust <-
                compute_just2d(data = data,
                               coord = coord,
                               panel_params = panel_params,
                               just = data$hjust,
                               a = "x", b = "y")
            }

            user.grobs <- grid::gList()

            # For coord_flip, coord$tranform does not flip the sides where to
            # draw the rugs. We have to flip them.
            flipped <- inherits(coord, 'CoordFlip')
            if (flipped) {
              sides <- chartr('tblr', 'rlbt', sides)
            }

            if (!flipped && !is.null(data$xintercept) &&
                (grepl("b", sides) || grepl("t", sides) ) ) {
              if (grepl("b", sides)) {
                y.pos <- unit(grob.shift, "npc")
              } else if (grepl("t", sides)) {
                y.pos <- unit(1 - grob.shift, "npc")
              }
              for (row.idx in seq_len(nrow(data))) {
                userGrob <- data$label[[row.idx]]

                userGrob$vp <-
                  grid::viewport(x = unit(data$xintercept[row.idx], "native"),
                                 y = y.pos,
                                 width = unit(data$vp.width[row.idx], "npc"),
                                 height = unit(data$vp.height[row.idx], "npc"),
                                 just = c(data$hjust[row.idx], data$vjust[row.idx]),
                                 angle = data$angle[row.idx],
                                 name = paste("geom_x_margin_grob.panel", data$PANEL[row.idx],
                                              "row", row.idx, sep = "."))

                # give unique name to each grob
                userGrob$name <- paste(sides, "margin.grob", row.idx, sep = ".")

                user.grobs[[row.idx]] <- userGrob
              }
            }

            # needed for handling flipped coords
            if (flipped && !is.null(data$yintercept) &&
                (grepl("l", sides) || grepl("r", sides) ) ) {
              if (grepl("l", sides)) {
                x.pos <- unit(grob.shift, "npc")
              } else if (grepl("r", sides)) {
                x.pos <- unit(1 - grob.shift, "npc")
              }
              for (row.idx in seq_len(nrow(data))) {
                userGrob <- data$label[[row.idx]]
                userGrob$vp <-
                  grid::viewport(x = x.pos,
                                 y = unit(data$yintercept[row.idx], "native"),
                                 width = unit(data$vp.width[row.idx], "npc"),
                                 height = unit(data$vp.height[row.idx], "npc"),
                                 just = c(data$hjust[row.idx], data$vjust[row.idx]),
                                 angle = data$angle[row.idx],
                                 name = paste("geom_x_margin_grob.panel", data$PANEL[row.idx],
                                              "row", row.idx, sep = "."))

                # give unique name to each grob
                userGrob$name <- paste(sides, "margin.grob", row.idx, sep = ".")

                user.grobs[[row.idx]] <- userGrob
              }
            }

            # grid.name <- paste("geom_x_margin_grob.panel",
            #                    data$PANEL[row.idx], sep = ".")
            #
            # grid::gTree(children = user.grobs, name = grid.name)
            grid::gTree(children = user.grobs)
          },

          draw_key = function(...) {
            grid::nullGrob()
          }
  )

## y axis marging
#' @rdname geom_x_margin_grob
#' @export
#'
geom_y_margin_grob <- function(mapping = NULL, data = NULL,
                               stat = "identity", position = "identity",
                               ...,
                               yintercept,
                               sides = "l",
                               grob.shift = 0,
                               na.rm = FALSE,
                               show.legend = FALSE,
                               inherit.aes = FALSE) {

  # Act like an annotation
  if (!missing(yintercept)) {
    data <- as.data.frame(list(yintercept = yintercept))
    mapping <- aes(yintercept = yintercept)
    show.legend <- FALSE
  }

  layer(
    data = data,
    mapping = mapping,
    stat = stat,
    geom = GeomYMarginGrob,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      sides = stringr::str_sub(sides, 1L, 1L),
      grob.shift = grob.shift,
      na.rm = na.rm,
      ...
    )
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
GeomYMarginGrob <-
  ggproto("GeomYMarginGrob", Geom,
          required_aes = c("yintercept"),
          default_aes = ggplot2::aes(
            colour = "black",
            family = "",
            fontface = 1,
            angle = 0,
            hjust = 0.5,
            vjust = 0.5,
            alpha = NA,
            vp.width = 1/5, vp.height = 1/5
          ),

          draw_panel = function(data, panel_params, coord, sides = "l",
                                grob.shift = 0, na.rm = FALSE) {
            if (nrow(data) == 0) {
              return(grid::nullGrob())
            }

            if (!grid::is.grob(data$label[[1]])) {
              warning("Skipping as object mapped to 'label' is not a list of \"grob\" objects.")
              return(grid::nullGrob())
            }

            data <- coord$transform(data, panel_params)
            if (is.character(data$vjust)) {
              data$vjust <- compute_just(data$vjust, data$y)
            }
            if (is.character(data$hjust)) {
              data$hjust <- compute_just(data$hjust, data$x)
            }

            user.grobs <- grid::gList()

            # For coord_flip, coord$transform does not flip the sides where to
            # draw the grobs. We have to flip them.
            flipped <- inherits(coord, 'CoordFlip')
            if (flipped) {
              sides <- chartr('tblr', 'rlbt', sides)
            }

            if (flipped && !is.null(data$xintercept) &&
                (grepl("b", sides) || grepl("t", sides) ) ) {
              if (grepl("b", sides)) {
                y.pos <- unit(grob.shift, "npc")
              } else if (grepl("t", sides)) {
                y.pos <- unit(1 - grob.shift, "npc")
              }
              for (row.idx in seq_len(nrow(data))) {
                userGrob <- data$label[[row.idx]]

                userGrob$vp <-
                  grid::viewport(x = unit(data$xintercept[row.idx], "native"),
                                 y = y.pos,
                                 width = unit(data$vp.width[row.idx], "npc"),
                                 height = unit(data$vp.height[row.idx], "npc"),
                                 just = c(data$hjust[row.idx], data$vjust[row.idx]),
                                 angle = data$angle[row.idx],
                                 name = paste("geom_grob.panel", data$PANEL[row.idx],
                                              "row", row.idx, sep = "."))

                # give unique name to each grob
                userGrob$name <- paste(sides, "margin.grob", row.idx, sep = ".")

                user.grobs[[row.idx]] <- userGrob
              }
            }

            # needed for handling flipped coords
            if (!flipped && !is.null(data$yintercept) &&
                (grepl("l", sides) || grepl("r", sides) ) ) {
              if (grepl("l", sides)) {
                x.pos <- unit(grob.shift, "npc")
              } else if (grepl("r", sides)) {
                x.pos <- unit(1 - grob.shift, "npc")
              }
              for (row.idx in seq_len(nrow(data))) {
                userGrob <- data$label[[row.idx]]
                userGrob$vp <-
                  grid::viewport(x = x.pos,
                                 y = unit(data$yintercept[row.idx], "native"),
                                 width = unit(data$vp.width[row.idx], "npc"),
                                 height = unit(data$vp.height[row.idx], "npc"),
                                 just = c(data$hjust[row.idx], data$vjust[row.idx]),
                                 angle = data$angle[row.idx],
                                 name = paste("geom_y_margin_grob.panel", data$PANEL[row.idx],
                                              "row", row.idx, sep = "."))

                # give unique name to each grob
                userGrob$name <- paste("inset.grob", row.idx, sep = ".")

                user.grobs[[row.idx]] <- userGrob
              }
            }

            grid.name <- paste("geom_y_margin_grob.panel",
                               data$PANEL[row.idx], sep = ".")

            grid::gTree(children = user.grobs, name = grid.name)
          },

          draw_key = function(...) {
            grid::nullGrob()
          }
  )

context("geom_x_margin_grob")


p <- ggplot(mtcars, aes(wt, mpg)) +
  geom_point()

test_that("geom_x_margin_grob, missing xintercept", {
  p1 <- p + geom_x_margin_grob()
  expect_error(layer_data(p1))
})

test_that("geom_x_margin_grob, not missing xintercept", {
  p1 <- p + geom_x_margin_grob(xintercept = 1)
  expect_no_error(layer_data(p1))
})

context("geom_y_margin_grob")

p <- ggplot(mtcars, aes(wt, mpg)) +
  geom_point()

test_that("geom_y_margin_grob, missing yintercept", {
  p1 <- p + geom_y_margin_grob()
  result <- expect_error(layer_data(p1))
})

test_that("geom_y_margin_grob, not missing yintercept", {
  p1 <- p + geom_y_margin_grob(yintercept = 1)
  expect_no_error(layer_data(p1))
})

#' Draw functions as curves
#'
#' \code{stat_functions()} computes values from functions and returns new data
#' containing numeric vectors for \code{x} and \code{y}. As function definitions
#' are passed through \code{data} this statistic follows the grammar of graphics in
#' its behaviour.
#'
#' @param mapping The aesthetic mapping, usually constructed with
#'   \code{\link[ggplot2]{aes}} or \code{\link[ggplot2]{aes_}}. Only needs to be
#'   set at the layer level if you are overriding the plot defaults.
#' @param data A layer specific dataset. Useful if the function curve is to be
#'   overlaid on other layers.
#' @param n integer Number of points to interpolate along the x axis.
#' @param geom The geometric object to use display the data
#' @param position The position adjustment to use on this layer
#' @param show.legend logical. Should this layer be included in the legends?
#'   \code{NA}, the default, includes it if any aesthetics are mapped.
#'   \code{FALSE} never includes, and \code{TRUE} always includes.
#' @param inherit.aes If \code{FALSE}, overrides the default aesthetics, rather
#'   than combining with them.
#' @param ... other arguments passed on to \code{\link[ggplot2]{layer}}. This
#'   can include aesthetics whose values you want to set, not map. See
#'   \code{\link[ggplot2]{layer}} for more details.
#' @param na.rm	a logical indicating whether \code{NA} values should be stripped
#'   before the computation proceeds.
#'
#' @details This statistic can be used to plot values computed by functions. As
#'   it follows the grammar of graphics, grouping and facets are supported. In
#'   this it differs from \code{\link[ggplot2]{geom_function}} which behaves
#'   like a plot annotation.
#'
#'   Aesthetics \code{xmin} and \code{xmax} should be mapped to numeric values
#'   defining the range of the vector to be created and passed as argument to
#'   the function to compute the \code{y} values, and returned as \code{x} in
#'   data. \code{n} is the length of this \code{x} vector.
#'
#' @section Computed variables: Data frame with \code{n} rows or a multiple of
#'   this, one for each
#'   row in \code{data}. \describe{
#'   \item{x}{numeric vector}
#'   \item{y}{numeric vactor}
#'   \item{idx}{integer vector, with values corresponding to rows in the input
#'       \code{data}, i.e., for each function}
#'   }
#'   As shown in one example below \code{\link[gginnards]{geom_debug}} can be
#'   used to print the computed values returned by any statistic. The output
#'   shown includes also values mapped to aesthetics.
#'
#' @return A plot layer instance.
#'
#' @export
#'
#' @examples
#'
#' # one function
#'
#' df1 <- data.frame(min = 0, max = pi, fun = I(list(sin)))
#'
#' ggplot(df1, aes(xmin = min, xmax = max, y = fun)) +
#'   stat_functions()
#'
#' ggplot(df1, aes(xmin = min, xmax = max, y = fun)) +
#'   stat_functions(geom = "point", n = 20)
#'
#' # two functions
#'
#' df2 <- data.frame(min = -pi, max = pi,
#'                   fun = I(list(sin, cos)), name = c("sin", "cos"))
#'
#' # each function must be in a separate group for correct plotting of lines
#'
#' ggplot(df2, aes(xmin = min, xmax = max, y = fun, group = after_stat(idx))) +
#'   stat_functions()
#'
#' ggplot(df2, aes(xmin = min, xmax = max, y = fun, colour = name)) +
#'   stat_functions()
#'
#' ggplot(df2, aes(xmin = min, xmax = max, y = fun)) +
#'   stat_functions() +
#'   facet_grid(~ name)
#'
#' # two curves with same function
#'
#' df3 <- data.frame(min = c(-pi, 0),
#'                   max = c(0,pi),
#'                   fun = I(list(sin, sin)),
#'                   name = c("negative", "positive"))
#'
#' ggplot(df3, aes(xmin = min, xmax = max, y = fun, colour = name)) +
#'   stat_functions()
#'
#' # We use geom_debug_group() to see the computed values
#'
#' gginnards.installed <- requireNamespace("gginnards", quietly = TRUE)
#' if (gginnards.installed) {
#'   library(gginnards)
#'
#'   ggplot(df1, aes(xmin = min, xmax = max, y = fun)) +
#'     stat_functions(geom = "debug_group")
#' }
#'
stat_functions <- function(mapping = NULL,
                           data = NULL,
                           n = 101,
                           geom = "line",
                           position = "identity",
                           na.rm = FALSE,
                           show.legend = NA,
                           inherit.aes = TRUE,
                           ...) {

  ggplot2::layer(
    stat = StatFunctions,
    data = data,
    mapping = mapping,
    geom = geom,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(na.rm = na.rm,
                  n = n,
                  ...)
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
StatFunctions <-
  ggplot2::ggproto("StatFunctions", ggplot2::Stat,

                   compute_group = function(data,
                                            scales,
                                            n = 101) {
                     z <- data.frame()

                     for (i in seq_along(data$xmin)) {
                       temp <- data.frame(x = seq(from = data$xmin[i],
                                              to = data$xmax[i],
                                              length.out = n),
                                          idx = i)

                       if (is.list(data[["y"]]) && is.function(data[["y"]][[i]])) {
                         temp$y <- data[["y"]][[i]](temp$x)
                       } else {
                         temp$y <- NA_real_
                       }

                       cols2copy <- setdiff(colnames(data),
                                            c(colnames(temp), "xmin", "xmax"))

                       temp <- cbind(temp, data[rep(i, n) , cols2copy])

                       z <- rbind(z, temp)
                     }

                     z
                   },

                   default_aes = ggplot2::aes(x = ggplot2::after_stat(x)),
                   required_aes = c("xmin", "xmax", "y")
  )

#' Inset graphical objects
#'
#' \code{geom_grob} and \code{geom_grob_npc} add Grobs as insets to the ggplot
#' using syntax similar to that of \code{\link[ggplot2]{geom_text}},
#' \code{\link{geom_text_s}} and \code{\link{geom_text_npc}}.
#' In most respects they behave as any other ggplot geometry: they add a layer
#' containing one or more grobs and grouping and faceting works as usual. The
#' most common use of \code{geom_grob} is to add data labels that are graphical
#' objects rather than text. \code{\link{geom_grob_npc}} is used to add grobs
#' as annotations to plots, but contrary to layer function \code{annotate()},
#' \code{\link{geom_grob_npc}} is data driven and respects grouping and facets,
#' thus plot insets can differ among panels. Of these two geoms only
#' \code{\link{geom_grob}} supports the plotting of segments, as
#' \code{\link{geom_grob_npc}} uses a coordinate system that is unrelated
#' to data units and data.
#'
#' @details You can modify the size of insets with the \code{vp.width} and
#'   \code{vp.height} aesthetics. These can take a number between 0 (smallest
#'   possible inset) and 1 (whole plotting area width or height). The default
#'   value for for both of these aesthetics is 1/5. Thus, in contrast to
#'   \code{\link[ggplot2]{geom_text}}, \code{\link[ggplot2]{geom_label}},
#'   \code{\link{geom_text_s}} and \code{\link{geom_label_s}} the size of the
#'   insets remains the same relative to the size of the plotting area
#'   irrespective of the size the plot is rendered at. The aspect ratio of
#'   insets is preserved and size is adjusted until the whole inset fits within
#'   the viewport.
#'
#'   By default \code{geom_grob} uses \code{\link{position_nudge_center}} and
#'   justification \code{"position"}, while \code{geom_grob_npc} uses
#'   \code{\link[ggplot2]{position_nudge}} and justification \code{"inward"}. In
#'   contrast to \code{\link[ggplot2]{position_nudge}},
#'   \code{\link{position_nudge_center}} and all other position functions
#'   defined in packages 'ggpp' keep the original coordinates thus allowing the
#'   plotting of connecting segments and arrows.
#'
#'   \code{geom_grob} and \code{geom_grob_npc} expect a list of graphic objects
#'   ("grob") to be mapped to the \code{label} aesthetic. These geoms work with
#'   tibbles or data frames as \code{data} as they both support \code{list}
#'   objects as member variables.
#'
#'   The \code{x} and \code{y} aesthetics determine the position of the whole
#'   inset grob, similarly to that of a text label, justification is interpreted
#'   as indicating the position of the grob with respect to its \emph{x} and
#'   \emph{y} coordinates in the data, and \code{angle} is used to rotate the
#'   grob as a whole.
#'
#' @section Plot boundaries and clipping: The "width" and "height" of an inset
#'   as for a text element are 0, so stacking and dodging inset plots will not
#'   work by default, and axis limits are not automatically expanded to include
#'   all inset plots. Obviously, insets do have height and width, but they are
#'   physical units, not data units. The amount of space they occupy on the main
#'   plot is not constant in data units of the base plot: when you modify scale
#'   limits, inset plots stay the same size relative to the physical size of the
#'   base plot.
#'
#' @inheritSection geom_text_s Alignment
#'
#' @inheritSection geom_text_s Position functions
#'
#' @note The insets are stored nested within the main ggplot object and
#'   contain their own copy of the data, and are rendered as grid grobs as normal
#'   ggplots at the time the main ggplot is rendered. They can have different
#'   themes.
#'
#'   Use \code{\link{annotate}} as redefined in 'ggpp' when adding insets
#'   as annotations (automatically available unless 'ggpp' is not attached).
#'   \code{\link[ggplot2]{annotate}} cannot be used with the \code{npcx} and
#'   \code{npcy} pseudo-aesthetics.
#'
#' @seealso \code{\link[grid]{grid-package}}, \code{\link[ggplot2]{geom_text}},
#'   and other documentation of package 'ggplot2'.
#'
#' @param mapping The aesthetic mapping, usually constructed with
#'   \code{\link[ggplot2]{aes}}. Only needs to be set at the layer level if you
#'   are overriding the plot defaults.
#' @param data A layer specific dataset - only needed if you want to override
#'   the plot defaults.
#' @param stat The statistical transformation to use on the data for this layer,
#'   as a string.
#' @param na.rm If \code{FALSE} (the default), removes missing values with a
#'   warning.  If \code{TRUE} silently removes missing values.
#' @param position Position adjustment, either as a string, or the result of a
#'   call to a position adjustment function.
#' @param ... other arguments passed on to \code{\link[ggplot2]{layer}}. This
#'   can include aesthetics whose values you want to set, not map. See
#'   \code{\link[ggplot2]{layer}} for more details.
#' @param show.legend logical. Should this layer be included in the legends?
#'   \code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
#'   never includes, and \code{TRUE} always includes.
#' @param inherit.aes If \code{FALSE}, overrides the default aesthetics, rather
#'   than combining with them. This is most useful for helper functions that
#'   define both data and aesthetics and shouldn't inherit behaviour from the
#'   default plot specification, e.g. \code{\link[ggplot2]{borders}}.
#' @param nudge_x,nudge_y Horizontal and vertical adjustments to nudge the
#'   starting position of each "label". The units for \code{nudge_x} and
#'   \code{nudge_y} are the same as for the data units on the x-axis and y-axis.
#' @param default.colour,default.color A colour definition to use for elements
#'   not targeted by the colour aesthetic.
#' @param colour.target,color.target A vector of character strings; \code{"all"},
#'   \code{"text"}, \code{"box"} and \code{"segment"} or \code{"none"}.
#' @param default.alpha numeric in [0..1] A transparency value to use for
#'   elements not targeted by the alpha aesthetic.
#' @param alpha.target A vector of character strings; \code{"all"},
#'   \code{"text"}, \code{"segment"}, \code{"box"}, \code{"box.line"}, and
#'   \code{"box.fill"} or \code{"none"}.
#' @param add.segments logical Display connecting segments or arrows between
#'   original positions and displaced ones if both are available.
#' @param box.padding,point.padding numeric By how much each end of the segments
#'   should shortened in mm.
#' @param segment.linewidth numeric Width of the segments or arrows in mm.
#' @param min.segment.length numeric Segments shorter that the minimum length
#'   are not rendered, in mm.
#' @param arrow specification for arrow heads, as created by
#'   \code{\link[grid]{arrow}}
#'
#' @return A plot layer instance.
#'
#' @references The idea of implementing a \code{geom_custom()} for grobs has
#'   been discussed as an issue at
#'   \url{https://github.com/tidyverse/ggplot2/issues/1399}.
#'
#' @family geometries adding layers with insets.
#'
#' @export
#'
#' @examples
#' library(tibble)
#' df <- tibble(x = 2, y = 15, grob = list(grid::circleGrob(r = 0.2)))
#'
#' # without nudging no segments are drawn
#' ggplot(data = mtcars,
#'        aes(wt, mpg)) +
#'   geom_point(aes(colour = factor(cyl))) +
#'   geom_grob(data = df,
#'             aes(x, y, label = grob))
#'
#' # with nudging segments are drawn
#' ggplot(data = mtcars,
#'        aes(wt, mpg)) +
#'   geom_point(aes(colour = factor(cyl))) +
#'   geom_grob(data = df,
#'             aes(x, y, label = grob),
#'             nudge_x = 0.5,
#'             colour = "red",
#'             hjust = 0.5,
#'             vjust = 0.5)
#'
#' ggplot(data = mtcars,
#'        aes(wt, mpg)) +
#'   geom_point(aes(colour = factor(cyl))) +
#'   geom_grob(data = df,
#'             aes(x, y, label = grob),
#'             nudge_x = 0.5,
#'             colour = "red",
#'             colour.target = "none",
#'             hjust = 0.5,
#'             vjust = 0.5)
#'
#' # with nudging plotting of segments can be disabled
#' ggplot(data = mtcars,
#'        aes(wt, mpg)) +
#'   geom_point(aes(colour = factor(cyl))) +
#'   geom_grob(data = df,
#'             aes(x, y, label = grob),
#'             add.segments = FALSE,
#'             nudge_x = 0.5,
#'             hjust = 0.5,
#'             vjust = 0.5)
#'
geom_grob <-
  function(mapping = NULL,
           data = NULL,
           stat = "identity",
           position = "identity",
           ...,
           nudge_x = 0,
           nudge_y = 0,
           default.colour = NULL,
           default.color = default.colour,
           colour.target = "segment",
           color.target = colour.target,
           default.alpha = 1,
           alpha.target = "segment",
           add.segments = TRUE,
           box.padding = 0.25,
           point.padding = 1e-06,
           segment.linewidth = 0.5,
           min.segment.length = 0,
           arrow = NULL,
           na.rm = FALSE,
           show.legend = FALSE,
           inherit.aes = FALSE) {

    colour.target <-
      rlang::arg_match(color.target,
                       values = c("segment", "all", "grob", "box", "none"),
                       multiple = TRUE)
    alpha.target <-
      rlang::arg_match(alpha.target,
                       values = c("segment", "all", "grob", "box",
                                  "box.line", "box.fill", "none"),
                       multiple = TRUE)

    if (!missing(nudge_x) || !missing(nudge_y)) {
      if (!missing(position) && position != "identity") {
        rlang::abort("You must specify either `position` or `nudge_x`/`nudge_y`.")
      }
      # original position needed for "position" justification
      position <-
        position_nudge_center(nudge_x, nudge_y, kept.origin = "original")
    }

    ggplot2::layer(
      data = data,
      mapping = mapping,
      stat = stat,
      geom = GeomGrob,
      position = position,
      show.legend = show.legend,
      inherit.aes = inherit.aes,
      params = list(
        default.colour = default.color,
        colour.target = colour.target,
        default.alpha = default.alpha,
        alpha.target = alpha.target,
        add.segments = add.segments,
        box.padding = box.padding,
        point.padding = point.padding,
        segment.linewidth = segment.linewidth,
        min.segment.length = min.segment.length,
        arrow = arrow,
        na.rm = na.rm,
        ...
      )
    )
  }

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
GeomGrob <-
  ggplot2::ggproto("GeomGrob", ggplot2::Geom,
                   required_aes = c("x", "y", "label"),

                   default_aes = ggplot2::aes(
                     colour = "black",
                     family = "",
                     fontface = 1,
                     angle = 0,
                     hjust = 0.5,
                     vjust = 0.5,
                     alpha = NA,
                     vp.width = 1/5, vp.height = 1/5
                   ),

                   draw_panel = function(data,
                                         panel_params,
                                         coord,
                                         default.colour = NULL,
                                         colour.target = "all",
                                         default.alpha = 1,
                                         alpha.target = "all",
                                         na.rm = FALSE,
                                         add.segments = TRUE,
                                         box.padding = 0.25,
                                         point.padding = 1e-06,
                                         segment.linewidth = 0.5,
                                         min.segment.length = 0,
                                         arrow = NULL) {

                     if (nrow(data) == 0) {
                       return(grid::nullGrob())
                     }

                     if (!grid::is.grob(data$label[[1]])) {
                       warning("Skipping as object mapped to 'label' is not a list of \"grob\" objects.")
                       return(grid::nullGrob())
                     }

                     add.segments <- add.segments && all(c("x_orig", "y_orig") %in% colnames(data))

                     default.colour <- check_default_colour(default.colour)

                     # should be called only once!
                     data <- coord$transform(data, panel_params)
                     if (add.segments) {
                       data_orig <- data.frame(x = data$x_orig, y = data$y_orig)
                       data_orig <- coord$transform(data_orig, panel_params)
                       data$x_orig <- data_orig$x
                       data$y_orig <- data_orig$y
                     }

                     if (is.character(data$vjust)) {
                       data$vjust <-
                         compute_just2d(data = data,
                                        coord = coord,
                                        panel_params = panel_params,
                                        just = data$vjust,
                                        a = "y", b = "x")
                     }
                     if (is.character(data$hjust)) {
                       data$hjust <-
                         compute_just2d(data = data,
                                        coord = coord,
                                        panel_params = panel_params,
                                        just = data$hjust,
                                        a = "x", b = "y")
                     }
                     if (add.segments) {
                       segments.data <-
                         shrink_segments(data,
                                         point.padding = point.padding,
                                         box.padding = box.padding,
                                         min.segment.length = min.segment.length)
                     }

                     # loop needed as gpar is not vectorized
                     all.grobs <- grid::gList()
                     user.grobs <- data[["label"]]
                     for (row.idx in 1:nrow(data)) {
                       row <- data[row.idx, , drop = FALSE]
                       grob.alpha <-
                         ifelse(any(alpha.target %in% c("all", "grob")),
                                row$alpha, default.alpha)
                       segment.alpha <-
                         ifelse(any(alpha.target %in% c("all", "segment")),
                                row$alpha, default.alpha)
                       user.grob <- user.grobs[[row.idx]]

                       user.grob$vp <-
                         grid::viewport(x = grid::unit(row$x, "native"),
                                        y = grid::unit(row$y, "native"),
                                        width = grid::unit(row$vp.width, "npc"),
                                        height = grid::unit(row$vp.height, "npc"),
                                        just = c(row$hjust, row$vjust),
                                        angle = row$angle,
                                        name = paste("inset.grob.vp", row$PANEL,
                                                     "row", row.idx, sep = "."))

                       # give unique name to each grob
                       user.grob$name <- paste("inset.grob", row.idx, sep = ".")

                       if (add.segments) {
                         segment.row <- segments.data[row.idx, , drop = FALSE]
                         if (segment.row$too.short) {
                           segment.grob <- grid::nullGrob()
                         } else {
                           segment.grob <-
                             grid::segmentsGrob(x0 = segment.row$x,
                                                y0 = segment.row$y,
                                                x1 = segment.row$x_orig,
                                                y1 = segment.row$y_orig,
                                                arrow = arrow,
                                                gp = grid::gpar(
                                                  col = if (segment.linewidth == 0) NA else # lwd = 0 is invalid in 'grid'
                                                    ifelse(any(colour.target %in% c("all", "segment")),
                                                           ggplot2::alpha(row$colour, segment.alpha),
                                                           ggplot2::alpha(default.colour, segment.alpha)),
                                                  lwd = (if (segment.linewidth == 0) 0.5 else segment.linewidth) * ggplot2::.stroke),
                                                name = paste("grob.s.segment", row$group, row.idx, sep = "."))
                         }
                         all.grobs <- grid::gList(all.grobs, segment.grob, user.grob)
                       } else {
                         all.grobs <- grid::gList(all.grobs, user.grob)
                       }
                     }
                     #    grid::grobTree(children = all.grobs, name = "geom.grob.panel")
                     grid::grobTree(children = all.grobs)

                   }
                   ,

                   draw_key = function(...) {
                     grid::nullGrob()
                   }
  )

#' @rdname geom_grob
#' @export
#'
geom_grob_npc <- function(mapping = NULL,
                          data = NULL,
                          stat = "identity",
                          position = "identity",
                          ...,
                          na.rm = FALSE,
                          show.legend = FALSE,
                          inherit.aes = FALSE) {
  layer(
    data = data,
    mapping = mapping,
    stat = stat,
    geom = GeomGrobNpc,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      na.rm = na.rm,
      ...
    )
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
GeomGrobNpc <-
  ggplot2::ggproto("GeomGrobNpc", ggplot2::Geom,
                   required_aes = c("npcx", "npcy", "label"),

                   default_aes = ggplot2::aes(
                     colour = "black",
                     family = "",
                     fontface = 1,
                     angle = 0,
                     hjust = "inward",
                     vjust = "inward",
                     alpha = NA,
                     vp.width = 1/5, vp.height = 1/5
                   ),

                   draw_panel = function(data,
                                         panel_params,
                                         coord,
                                         na.rm = FALSE) {

                     if (nrow(data) == 0) {
                       return(grid::nullGrob())
                     }

                     if (!grid::is.grob(data$label[[1]])) {
                       warning("Skipping as object mapped to 'label' is not a list of \"grob\".")
                       return(grid::nullGrob())
                     }

                     data$npcx <- compute_npcx(data$npcx)
                     data$npcy <- compute_npcy(data$npcy)

                     if (is.character(data$vjust)) {
                       data$vjust <- compute_just(data$vjust, data$npcy)
                     }
                     if (is.character(data$hjust)) {
                       data$hjust <- compute_just(data$hjust, data$npcx)
                     }

                     user.grobs <- grid::gList()

                     for (row.idx in 1:nrow(data)) {
                       userGrob <- data$label[[row.idx]]

                       userGrob$vp <-
                         grid::viewport(x = grid::unit(data$npcx[row.idx], "npc"),
                                        y = grid::unit(data$npcy[row.idx], "npc"),
                                        width = grid::unit(data$vp.width[row.idx], "npc"),
                                        height = grid::unit(data$vp.height[row.idx], "npc"),
                                        just = c(data$hjust[row.idx], data$vjust[row.idx]),
                                        angle = data$angle[row.idx],
                                        name = paste("geom_grob.panel", data$PANEL[row.idx],
                                                     "row", row.idx, sep = "."))

                       # give unique name to each grob
                       userGrob$name <- paste("inset.grob", row.idx, sep = ".")

                       user.grobs[[row.idx]] <- userGrob
                     }

                     # grid.name <- paste("geom_grob.panel",
                     #                    data$PANEL[row.idx], sep = ".")
                     #
                     # grid::gTree(children = user.grobs, name = grid.name)
                     grid::gTree(children = user.grobs)
                   },

                   draw_key = function(...) {
                     grid::nullGrob()
                   }
  )

context("geom_grob")

library(ggplot2)
library(tibble)

test_that("character label gives error in geom_grob", {
  my.df <- data.frame(x = 1:10, y = 1:10, tb = letters[1:10])
  expect_error(print(ggplot(my.df, aes(x, y, label = tb)) +
                       geom_grob()))
})

test_that("numeric label gives error in geom_grob", {
  my.df <- data.frame(x = 1:10, y = 1:10, tb = 1:10)
  expect_error(print(ggplot(my.df, aes(x, y, label = tb)) +
                       geom_grob()))
})

test_that("geom_grob pos_or_nudge", {
  tb <- tibble(a = 2:4, b = 4:2)
  my.tb <- tibble(x = 0, y = 0, tb = list(tb))
  expect_error(geom_grob(data = my.tb,
                          mapping = aes(x, y, label = tb),
                          position = "stack",
                          nudge_x = 0.5,
                          nudge_y = 0.5),
               "You must specify either `position` or `nudge_x`/`nudge_y`.")
})

test_that("geom_grob no segment", {
  tb <- tibble(a = 2:4, b = 4:2, c = 1:3)
  df <- tibble(x = 2, y = 3, grob = list(grid::circleGrob(r = 0.2)))
  p <- ggplot(data = tb, aes(a,b)) +
    geom_point(aes(colour = factor(c))) +
    geom_grob(data = df,
              aes(x, y, label = grob))

  result <- layer_data(p, 2)[, c("x", "y", "label", "colour")]
  expect_identical(result$colour, "black")
  expect_equal("x_orig" %in% colnames(result), FALSE)
  expect_identical(df$x, result$x)
  expect_identical(df$y, result$y)
})

test_that("geom_grob segment drawn", {
  tb <- tibble(a = 2:4, b = 4:2, c = 1:3)
  df <- tibble(x = 2, y = 3, grob = list(grid::circleGrob(r = 0.2)))
  nudge_x <- 0.8
  colour <- "blue"

  p1 <- ggplot(data = tb, aes(a,b)) +
    geom_point(aes(colour = factor(c))) +
    geom_grob(data = df,
              aes(x, y, label = grob),
              nudge_x = nudge_x,
              colour = colour)

  result1 <- layer_data(p1, 2)[, c("x", "y", "label", "x_orig", "y_orig", "colour")]
  expect_identical(df$x + nudge_x, result1$x)
  expect_identical(df$x, result1$x_orig)
  expect_identical(colour, "blue")
})

test_that("geom_grob segment disabled", {
  tb <- tibble(a = 2:4, b = 4:2, c = 1:3)
  df <- tibble(x = 2, y = 3, grob = list(grid::circleGrob(r = 0.2)))
  nudge_x = 0.8

  p2 <- ggplot(data = tb, aes(a,b)) +
    geom_point(aes(colour = factor(c))) +
    geom_grob(data = df,
              aes(x, y, label = grob),
              add.segments = FALSE,
              nudge_x = nudge_x)

  result2 <- layer_data(p2, 2)[, c("x", "y", "label", "x_orig", "y_orig", "colour")]
  expect_identical(result2$colour, "black")
  expect_identical(df$x + nudge_x, result2$x)
  expect_identical(df$y, result2$y_orig)
})

test_that("geom_grob_npc", {
  tb <- tibble(a = 2:4, b = 4:2, c = 1:3)
  df <- tibble(x = 2, y = 3, grob = list(grid::circleGrob(r = 0.2)))

  p_npc <- ggplot(data = tb, aes(a,b)) +
    geom_point(aes(colour = factor(c))) +
    geom_grob_npc(data = df,
                  aes(npcx = x, npcy = y, label = grob))

  result_p_npc <- layer_data(p_npc, 2)[, c("npcx", "npcy", "label", "colour")]
  expect_identical(result_p_npc$colour, "black")
  expect_identical(result_p_npc$npcx, df$x)
  expect_identical(result_p_npc$npcy, df$y)
})

#' @export
#' @rdname geom_text_npc
#' @param label.padding Amount of padding around label. Defaults to 0.25 lines.
#' @param label.r Radius of rounded corners. Defaults to 0.15 lines.
#' @param label.size Size of label border, in mm.
#'
geom_label_npc <- function(mapping = NULL,
                           data = NULL,
                           stat = "identity",
                           position = "identity",
                           ...,
                           parse = FALSE,
                           nudge_x = 0,
                           nudge_y = 0,
                           label.padding = grid::unit(0.25, "lines"),
                           label.r = grid::unit(0.15, "lines"),
                           label.size = 0.25,
                           size.unit = "mm",
                           na.rm = FALSE,
                           show.legend = FALSE,
                           inherit.aes = FALSE) {

  if (!missing(nudge_x) || !missing(nudge_y)) {
    if (!missing(position) && position != "identity") {
      stop("You must specify either `position` or `nudge_x`/`nudge_y`.",
           call. = FALSE)
    }
    position <- ggplot2::position_nudge(nudge_x, nudge_y)
  }

  ggplot2::layer(
    data = data,
    mapping = mapping,
    stat = stat,
    geom = GeomLabelNpc,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      parse = parse,
      label.padding = label.padding,
      label.r = label.r,
      label.size = label.size,
      size.unit = size.unit,
      na.rm = na.rm,
      ...
    )
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
GeomLabelNpc <- ggplot2::ggproto("GeomLabelNpc", ggplot2::Geom,
  required_aes = c("npcx", "npcy", "label"),

  default_aes = ggplot2::aes(
    colour = "black",
    fill = rgb(1, 1, 1, alpha = 0.75), # "white", but occluded data are visible
    family = "",
    size = 3.87,
    angle = 0,
    hjust = "inward",
    vjust = "inward",
    alpha = NA,
    fontface = 1,
    lineheight = 1.2,
    linewidth = 0.5,
    linetype  = "solid"
  ),

  draw_panel = function(self, data, panel_params, coord, parse = FALSE,
                        na.rm = FALSE,
                        label.padding = unit(0.25, "lines"),
                        label.r = unit(0.15, "lines"),
                        label.size = 0.25,
                        size.unit = "mm") {

    data$npcx <- compute_npcx(data$npcx)
    data$npcy <- compute_npcy(data$npcy)

    ranges <- coord$backtransform_range(panel_params)

    data$x <- ranges$x[1] + data$npcx * (ranges$x[2] - ranges$x[1])
    data$y <- ranges$y[1] + data$npcy * (ranges$y[2] - ranges$y[1])

    lab <- data$label

    if (parse) {
      lab <- parse_safe(as.character(lab))
    }

    data <- coord$transform(data, panel_params)
    if (is.character(data$vjust)) {
      data$vjust <- compute_just(data$vjust, data$y, data$x, data$angle)
    }
    if (is.character(data$hjust)) {
      data$hjust <- compute_just(data$hjust, data$x, data$y, data$angle)
    }
    if (!inherits(label.padding, "margin")) {
      label.padding <- rep(label.padding, length.out = 4)
    }

    size.unit <- resolve_text_unit(size.unit)

    grobs <- lapply(1:nrow(data), function(i) {
      row <- data[i, , drop = FALSE]
      labelGrob(lab[i],
                x = grid::unit(row$x, "native"),
                y = grid::unit(row$y, "native"),
                just = c(row$hjust, row$vjust),
                padding = label.padding,
                r = label.r,
                angle = row$angle,
                text.gp = grid::gpar(
                  col = row$colour,
                  fontsize = row$size * size.unit,
                  fontfamily = row$family,
                  fontface = row$fontface,
                  lineheight = row$lineheight
                ),
                rect.gp = grid::gpar(
                  col = if (isTRUE(all.equal(label.size, 0))) NA else row$colour,
                  fill = ggplot2::alpha(row$fill, row$alpha),
                  lwd = label.size * .pt,
                  lty = row$linetype
                )
      )
    })
    class(grobs) <- "gList"

    ggname("geom_label", grid::grobTree(children = grobs))
  },

  draw_key =  function(...) {
    grid::nullGrob()  # geom meant to be used for annotations
  }
)

context("geom_label_npc")

df <- data.frame(
x = c(0, 0, 1, 1, 0.5),
x.chr = c("left", "left", "right", "right", "center"),
y = c(0, 1, 0, 1, 0.5),
y.chr = c("bottom", "top", "bottom", "top", "middle"),
text = c("bottom-left", "top-left", "bottom-right", "top-right", "center-middle")
)

test_that("geom_label_npc, runs successfully", {
  expect_no_error(
    ggplot(data = mtcars) +
    geom_point(mapping = aes(wt, mpg)) +
    geom_label_npc(data = df, aes(npcx = x, npcy = y, label = text))
  )
})

test_that("geom_label_npc pos_or_nudge", {
  expect_error(
    geom_label_npc(position = "", nudge_x = 0.5),
    "You must specify either `position` or `nudge_x`/`nudge_y`."
  )
})

#' Combined positions jitter and nudge
#'
#' \code{position_jitternudge()} combines into one function the action of
#' \code{\link[ggplot2]{position_jitter}} and
#' \code{\link[ggplot2]{position_nudge}}. It is useful when labels to jittered
#' plots and when adding jitter to text labels linked to points plotted without
#' jitter. It can replace other position functions as it is backwards
#' compatible. Like all other position functions in 'ggpp' and 'ggrepel' it
#' preserves the initial position to allow drawing of segments or arrow linking
#' the original position to the displaced one.
#'
#' @details Jitter with \code{position_jitternudge()} is identical to that with
#'   \code{\link[ggplot2]{position_jitter}} while nudging is enhanced compared
#'   to \code{\link[ggplot2]{position_nudge}} by taking into use cases specific
#'   to the combination of jitter and nudge.
#'
#'   There are two posible uses for this function. First it can be used to label
#'   jittered points in a plot. In this case, it is mandatory to use the same
#'   arguments to \code{width}, \code{height} and \code{seed} when passing
#'   \code{position_jitter()} to \code{geom_point()} and
#'   \code{position_jitternudge()} to \code{geom_text()} or to
#'   \code{geom_label()} or their repulsive equivalents. Otherwise the arrows or
#'   segments will fail to connect to the labels. In other words jittering is
#'   computed twice. Jitter should be identical with the same arguments as
#'   \code{position_jitternudge()} as this last function calls the same code
#'   imported from package 'ggplot2'.
#'
#'   The second use is to jitter labels to be connected to points that have not
#'   been jittered. The return of original positions instead of the jittered
#'   ones is achieved by passing \code{origin = "original"} to override the
#'   default \code{origin = "jittered"}.
#'
#' @family position adjustments
#'
#' @param width,height Amount of vertical and horizontal jitter. The jitter is
#'   added in both positive and negative directions, so the total spread is
#'   twice the value specified here. If omitted, defaults to 40% of the
#'   resolution of the data: this means the jitter values will occupy 80% of the
#'   implied bins. Categorical data is aligned on the integers, so a width or
#'   height of 0.5 will spread the data so it's not possible to see the
#'   distinction between the categories.
#' @param seed A random seed to make the jitter reproducible. Useful if you need
#'   to apply the same jitter twice, e.g., for a point and a corresponding
#'   label. The random seed is reset after jittering. If \code{NA} (the default
#'   value), the seed is initialised with a random value; this makes sure that
#'   two subsequent calls start with a different seed. Use \code{NULL} to use
#'   the current random seed and also avoid resetting (the behaviour of ggplot
#'   2.2.1 and earlier).
#' @param x,y Amount of vertical and horizontal distance to move. A numeric
#'   vector of length 1, or of the same length as rows there are in \code{data},
#'   with nudge values in data rows order.
#' @param direction One of \code{"as.is"}, \code{"alternate"}, \code{"split"},
#'   \code{"split.x"} or \code{"split.y"}. A value of \code{"none"} replicates
#'   the behavior of \code{\link[ggplot2]{position_nudge}}. \code{"split"}
#'   changes the sign of the nudge depending on the direction of the random
#'   jitter applied to each individual observation, which is suitable for
#'   nudging labels outward of the jittered data.
#' @param nudge.from One of \code{"original"}, \code{"jittered"},
#'   \code{"original.y"} (or \code{"jittered.x"}), \code{"original.x"} (or
#'   \code{"jittered.y"}). A value of \code{"original"} applies the nudge before
#'   jittering the observations, while \code{"jittered"} applies the nudging
#'   after jittering.
#' @param kept.origin One of \code{"original"}, \code{"jittered"} or
#'   \code{"none"}.
#'
#' @note When \code{direction = "split"} is used together with no jitter, the
#'   split to left and right, or up and down is done at random.
#'
#' @seealso \code{\link[ggplot2]{position_jitter}},
#' \code{\link[ggplot2]{position_nudge}},
#'  \code{\link[ggrepel]{position_nudge_repel}}.
#'
#' @return A \code{"Position"} object. The layer function within it returns a
#'   data frame, with the jittered + nudged values in columns \code{x} and
#'   \code{y} and by default the jittered values with no nudging as
#'   \code{x_orig} and \code{y_orig}. With \code{nudge.from = "original"} the
#'   original values with no jitter and no nudge applied are returned as
#'   \code{x_orig} and \code{y_orig}.
#'
#' @export
#'
#' @author Michał Krassowski, edited by Pedro J. Aphalo.
#'
#' @source \url{https://github.com/slowkow/ggrepel/issues/161}.
#'
#' @examples
#'
#' jitter <- position_jitter(width = 0.2, height = 2, seed = 123)
#'
#' jitter_nudge <- position_jitternudge(width = 0.2, height = 2,
#'                                      seed = 123, x = 0.1,
#'                                      direction = "split",
#'                                      nudge.from = "jittered")
#' ggplot(mpg[1:20, ],
#'        aes(cyl, hwy, label = drv)) +
#'   geom_point(position = jitter) +
#'   geom_text_s(position = jitter_nudge)
#'
#' jitter_nudge <- position_jitternudge(width = 0.2, height = 2,
#'                                      seed = 123, x = 0.35,
#'                                      direction = "split",
#'                                      nudge.from = "original.x")
#' ggplot(mpg[1:20, ],
#'        aes(cyl, hwy, label = drv)) +
#'   geom_point(position = jitter) +
#'   geom_text_s(position = jitter_nudge)
#'
#' jitter <- position_jitter(width = 0, height = 2, seed = 123)
#'
#' jitter_nudge <- position_jitternudge(width = 0, height = 2,
#'                                      seed = 123, x = 0.4,
#'                                      direction = "split",
#'                                      nudge.from = "original.x")
#' ggplot(mpg[1:20, ],
#'        aes(cyl, hwy, label = drv)) +
#'   geom_point(position = jitter) +
#'   geom_text_s(position = jitter_nudge)
#'
#' jitter_nudge <- position_jitternudge(width = 0, height = 2,
#'                                      seed = 123, x = 0.4,
#'                                      direction = "alternate",
#'                                      nudge.from = "original.x")
#' ggplot(mpg[1:20, ],
#'        aes(cyl, hwy, label = drv)) +
#'   geom_point(position = jitter) +
#'   geom_text_s(position = jitter_nudge)
#'
#' # No nudge, show how points have moved with jitter
#'
#' ggplot(mpg[1:20, ],
#'        aes(cyl, hwy, label = drv)) +
#'   geom_point() +
#'   geom_point_s(position =
#'                position_jitter_keep(width = 0.3, height = 2, seed = 123),
#'                color = "red",
#'                arrow = grid::arrow(length = unit(0.4, "lines")))
#'
position_jitternudge <-
  function(width = NULL,
           height = NULL,
           seed = NA,
           x = 0,
           y = 0,
           direction = c("as.is", "alternate", "split"),
           nudge.from = c("original", "original.x", "original.y",
                          "jittered", "jittered.y", "jittered.x"),
           kept.origin = c("jittered", "original", "none")) {

    direction <- rlang::arg_match(direction)
    nudge.from <- rlang::arg_match(nudge.from)
    kept.origin <- rlang::arg_match(kept.origin)

    fixed.direction <-
    function(x) {1}
  conditional.direction <-
    function(x) {
      # positive or negative jitter
      s <- sign(x)
      if (any(s == 0)) {
        # split with no jitter is at random
        r <- sample(c(-1, 1), size = length(s), replace = TRUE)
        ifelse(s == 0, r, s)
      } else {
        s
      }
    }

  alternate.direction <-
    function(x) {
      rep_len(c(1, -1), length.out = length(x))
    }

  ggplot2::ggproto(NULL, PositionJitterAndNudge,
                   x = x,
                   y = y,
                   .fun_x = switch(direction,
                                   as.is = fixed.direction,
                                   none = fixed.direction,
                                   split = conditional.direction,
                                   split.x = conditional.direction,
                                   split.y = fixed.direction,
                                   alternate = alternate.direction,
                                   alternate.x = alternate.direction,
                                   alternate.y = fixed.direction,
                                   fixed.direction),
                   .fun_y = switch(direction,
                                   as.is = fixed.direction,
                                   none = fixed.direction,
                                   split = conditional.direction,
                                   split.x = fixed.direction,
                                   split.y = conditional.direction,
                                   alternate = alternate.direction,
                                   alternate.x = fixed.direction,
                                   alternate.y = alternate.direction,
                                   fixed.direction),
                   nudge.from = nudge.from,
                   kept.origin = kept.origin,
                   width = width,
                   height = height,
                   seed = seed
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @noRd
PositionJitterAndNudge <-
  ggplot2::ggproto("PositionDodgeAndNudge", ggplot2::PositionJitter,
                   x = 0,
                   y = 0,

                   setup_params = function(self, data) {
                     c(
                       list(nudge_x = self$x,
                            nudge_y = self$y,
                            .fun_x = self$.fun_x,
                            .fun_y = self$.fun_y,
                            x.reorder = !is.null(self$x) && length(self$x) > 1 && length(self$x) < nrow(data),
                            y.reorder = !is.null(self$y) && length(self$y) > 1 && length(self$y) < nrow(data),
                            nudge.from = self$nudge.from,
                            kept.origin = self$kept.origin),
                       ggplot2::ggproto_parent(ggplot2::PositionJitter, self)$setup_params(data)
                     )
                   },

                   compute_layer = function(self, data, params, layout) {

                     if (length(params$nudge_x) > nrow(data)) {
                       warning("Argument 'x' longer than data: some values dropped!")
                     }
                     if (length(params$nudge_y) > nrow(data)) {
                       warning("Argument 'y' longer than data: some values dropped!")
                     }
                     x_orig <- data$x
                     y_orig <- data$y

                     # operate on the jittered positions
                     data = ggplot2::ggproto_parent(ggplot2::PositionJitter, self)$compute_layer(data, params, layout)

                     x_jittered <- data$x
                     y_jittered <- data$y
                     if (params$nudge.from %in% c("original", "original.x", "jittered.y")) {
                       data$x <- x_orig
                     }
                     if (params$nudge.from %in% c("original", "original.y", "jittered.x")) {
                       data$y <- y_orig
                     }

                     if (params$x.reorder) {
                       params$nudge_x <- rep_len(params$nudge_x, nrow(data))[order(order(data$x))]
                     } else {
                       params$nudge_x <- rep_len(params$nudge_x, nrow(data))
                     }
                     if (params$y.reorder) {
                       params$nudge_y <- rep_len(params$nudge_y, nrow(data))[order(order(data$y))]
                     } else {
                       params$nudge_y <- rep_len(params$nudge_y, nrow(data))
                     }

                     # transform only the dimensions for which non-zero nudging is requested
                     if (any(params$nudge_x != 0)) {
                       if (any(params$nudge_y != 0)) {
                         data <- ggplot2::transform_position(data,
                                                             function(x) x + params$nudge_x * params$.fun_x(x_jittered - x_orig),
                                                             function(y) y + params$nudge_y * params$.fun_y(y_jittered - y_orig))
                       } else {
                         data <- ggplot2::transform_position(data,
                                                             function(x) x + params$nudge_x * params$.fun_x(x_jittered - x_orig),
                                                             NULL)
                       }
                     } else if (any(params$nudge_y != 0)) {
                       data <- ggplot2::transform_position(data,
                                                           function(x) x,
                                                           function(y) y + params$nudge_y * params$.fun_y(y_jittered - y_orig))
                     }
                     # add origin position (for connecting arrow or segment)
                     if (params$kept.origin == "jittered") {
                       data$x_orig <- x_jittered
                       data$y_orig <- y_jittered
                     } else if (params$kept.origin == "original") {
                       data$x_orig <- x_orig
                       data$y_orig <- y_orig
                     }

                     data
                   },

                   compute_panel = function(self, data, params, scales) {
                     ggplot2::ggproto_parent(PositionJitter, self)$compute_panel(data, params, scales)
                   }
  )

#' @rdname position_jitternudge
#'
#' @export
#'
position_jitter_keep <- function(width = NULL,
                                 height = NULL,
                                 seed = NA,
                                 kept.origin = "original") {
  position_jitternudge(width = width,
                       height = height,
                       seed = seed,
                       x = 0,
                       y = 0,
                       direction = "as.is",
                       nudge.from = "jittered",
                       kept.origin = kept.origin)
}

context("position_jitter_keep")

test_that("test expected arguments", {
  position <- position_jitter_keep()
  expect_identical(position$kept.origin, "original")
  expect_null(position$width)
  expect_null(position$height)
  expect_identical(position$x, 0)
  expect_identical(position$y, 0)
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
  skip_if(condition = utils::packageVersion("ggplot2") < "4.0.0",
          message = "'ggplot2' < 4.0.0")
  expect_identical(position$required_aes, c("x", "y"))
  expect_type(position$aesthetics, "closure")
  expect_type(position$use_defaults, "closure")
})

context("position_jitternudge")

test_that("incorrect kept.origin used", {
  expect_error(
    position_jitternudge(kept.origin = "wrong")
    #,  "`kept.origin` must be one of \"jittered\", \"original\", or \"none\", not\n\"wrong\"."
  )
})

test_that("correct kept.origin used", {
  expect_no_error(position <- position_jitternudge())
  expect_identical(position$kept.origin, "jittered")
})

test_that("correct kept.origin used", {
  expect_no_error(position <- position_jitternudge(kept.origin = "jittered"))
  expect_identical(position$kept.origin, "jittered")
})

test_that("correct kept.origin used", {
  expect_no_error(position <- position_jitternudge(kept.origin = "original"))
  expect_identical(position$kept.origin, "original")
})

test_that("correct kept.origin used", {
  expect_no_error(position <- position_jitternudge(kept.origin = "none"))
  expect_identical(position$kept.origin, "none")
})

test_that("incorrect nudge.from used", {
  expect_error(
    position_jitternudge(nudge.from = "original.z")
    #, "`nudge.from` must be one of \"original\", \"original.x\", \"original.y\",\n\"jittered\", \"jittered.y\", or \"jittered.x\", not \"original.z\".\nℹ Did you mean \"original.x\"?"
  )
})

test_that("correct nudge.from used", {
  expect_no_error(position <- position_jitternudge())
  expect_identical(position$nudge.from, "original")
})

test_that("correct nudge.from used", {
  expect_no_error(position <- position_jitternudge(nudge.from = "original"))
  expect_identical(position$nudge.from, "original")
})

test_that("correct nudge.from used", {
  expect_no_error(position <- position_jitternudge(nudge.from = "original.x"))
  expect_identical(position$nudge.from, "original.x")
})

test_that("correct nudge.from used", {
  expect_no_error(position <- position_jitternudge(nudge.from = "original.y"))
  expect_identical(position$nudge.from, "original.y")
})

test_that("correct nudge.from used", {
  expect_no_error(position <- position_jitternudge(nudge.from = "jittered"))
  expect_identical(position$nudge.from, "jittered")
})

test_that("correct nudge.from used", {
  expect_no_error(position <- position_jitternudge(nudge.from = "jittered.x"))
  expect_identical(position$nudge.from, "jittered.x")
})

test_that("correct nudge.from used", {
  expect_no_error(position <- position_jitternudge(nudge.from = "jittered.y"))
  expect_identical(position$nudge.from, "jittered.y")
})

test_that("test if correct arguments are assigned", {
  position <- position_jitternudge(width = 0.2, height = 2,
                                   seed = 123, x = 0.35,
                                   direction = "split",
                                   nudge.from = "original.x")

  expect_identical(position$kept.origin, "jittered")
  expect_identical(position$nudge.from, "original.x")
  expect_identical(position$width, 0.2)
  expect_identical(position$height, 2)
  expect_identical(position$seed, 123)
  expect_identical(position$x, 0.35)
  expect_identical(position$y, 0)
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
  skip_if(condition = utils::packageVersion("ggplot2") < "4.0.0",
          message = "'ggplot2' < 4.0.0")
  expect_identical(position$required_aes, c("x", "y"))
  expect_type(position$aesthetics, "closure")
  expect_type(position$use_defaults, "closure")
})

test_that("test if correct arguments are assigned with x < 0", {
  position <- position_jitternudge(width = 0.2, height = 2,
                                   seed = 123, x = -0.35, y = -1,
                                   direction = "alternate",
                                   nudge.from = "original.x")

  expect_identical(position$kept.origin, "jittered")
  expect_identical(position$nudge.from, "original.x")
  expect_identical(position$width, 0.2)
  expect_identical(position$height, 2)
  expect_identical(position$seed, 123)
  expect_identical(position$x, -0.35)
  expect_identical(position$y, -1)
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
  skip_if(condition = utils::packageVersion("ggplot2") < "4.0.0",
          message = "'ggplot2' < 4.0.0")
  expect_identical(position$required_aes, c("x", "y"))
  expect_type(position$aesthetics, "closure")
  expect_type(position$use_defaults, "closure")
})

test_that("test if correct arguments are assigned with x < 0", {
  position <- position_jitternudge(width = 0.2, height = 2,
                                   seed = 123, x = -0.35, y = -1,
                                   direction = "as.is",
                                   nudge.from = "original.y",
                                   kept.origin = "original")

  expect_identical(position$kept.origin, "original")
  expect_identical(position$nudge.from, "original.y")
  expect_identical(position$width, 0.2)
  expect_identical(position$height, 2)
  expect_identical(position$seed, 123)
  expect_identical(position$x, -0.35)
  expect_identical(position$y, -1)
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
  skip_if(condition = utils::packageVersion("ggplot2") < "4.0.0",
          message = "'ggplot2' < 4.0.0")
  expect_identical(position$required_aes, c("x", "y"))
  expect_type(position$aesthetics, "closure")
  expect_type(position$use_defaults, "closure")
})

test_that("warn on x or y longer than data", {
  expect_no_warning(
    ggplot(mtcars, aes(x = am, y = mpg, label = mpg)) +
      geom_text_s(
        position = position_jitternudge(x = 0, y = rep(0.5, nrow(mtcars))),
        size = 2.5
      )
  )

  expect_no_warning(
    ggplot(mtcars, aes(x = am, y = mpg, label = mpg)) +
      geom_text_s(
        position = position_jitternudge(y = 0, x = rep(0.05, nrow(mtcars))),
        size = 2.5
      )
  )
})

#' Wrap character strings in a vector
#'
#' Wrap the members of a character vector to a given maximum width by inserting
#' new line characters at word boundaries.
#'
#' @param x character vector, or an object which can be converted to a character
#'   vector by \code{as.character}.
#' @param width a positive integer giving the target column for wrapping lines
#'   in the output.
#' @param indent a positive or negative integer giving the indentation of the
#'   first line in a member character string.
#' @param new.line character sting; use \code{"<br>"} for HTML encoded strings.
#'
#' @details Function \code{wrap_labels()} is a wrapper on \code{link{strwrap}}
#' that returns a vector of character strings instead of a list of vectors. In
#' addition to wrapping, indentation is supported. Wrapping is always at white
#' space, so \code{width = 0} wraps word by word.
#'
#' Because the returned value is a character vector of the same length as the
#' input, this function can be used within a call to \code{aes()} when mapping a
#' character vector to the \code{label} aesthetic, as long as the character
#' strings will not be parsed into R expressions. It can be also used to wrap
#' the strings in a variable stored in a data frame.
#'
#' @return A character vector of the same length as \code{x}, with new line
#'   characters inserted to wrap text lines longer than \code{width}. Names in
#'   \code{x} are preserved in the returned value, no names are added if none
#'   are present in \code{x}.
#'
#' @examples
#' my.text <- c(A = "This is the first string",
#'              B = "This is the second string, which is longer")
#'
#' wrap_labels(my.text, width = 20)
#' wrap_labels(unname(my.text), width = 20)
#'
#' cat(wrap_labels(my.text, width = 20), sep = "\n--\n")
#' cat(wrap_labels(my.text, width = 20, indent = 2), sep = "\n--\n")
#' cat(wrap_labels(my.text, width = 20, indent = -2), sep = "\n--\n")
#'
#' @export
#'
wrap_labels <- function(x,
                        width = 20,
                        indent = 0,
                        new.line = "\n") {
  if (length(x) == 0L) {
    # avoid returning list() as returned by sapply()
    return(character(0L))
  }
  if (indent < 0) {
    exdent = - indent
    indent = 0
  } else {
    exdent = 0
  }
  sapply(X = x,
         FUN = function(x) {
           paste(
             strwrap(x,
                     width = width,
                     prefix = new.line,
                     initial = "",
                     indent = indent,
                     exdent = exdent),
             collapse = "")
         },
         USE.NAMES = FALSE)
}

context("wrap_labels")

test_that("returns always a character vector", {

  expect_is(wrap_labels(" abcd fghij", width = 5), "character")
  expect_is(wrap_labels(NA, width = 5), "character")
  expect_is(wrap_labels(TRUE, width = 5), "character")
  expect_is(wrap_labels(c(123, 1e6), width = 5), "character")
  expect_is(wrap_labels(c(123, NA), width = 5), "character")
  expect_is(wrap_labels(c(123, 1e6), width = 5), "character")
  expect_is(wrap_labels(character(), width = 5), "character")

})

test_that("white space is consumed or replaced", {

expect_equal(wrap_labels("", width = 20), "")
expect_equal(wrap_labels("\n", width = 20), "")
expect_equal(wrap_labels("\t", width = 20), "")
expect_equal(wrap_labels("   ", width = 20), "")
expect_equal(wrap_labels(" abc ", width = 20), "abc")
expect_equal(wrap_labels(" ab cd ", width = 20), "ab cd")
expect_equal(wrap_labels(" ab\ncd ", width = 20), "ab cd")
expect_equal(wrap_labels(" ab\tcd ", width = 20), "ab cd")

})

test_that("text is wrapped and indented", {
  my.text <- c(A = "This is the first string",
               B = "This is the second string, which is longer")

  expect_length(wrap_labels(my.text, width = 10), 2)
  expect_equal(wrap_labels(my.text[["A"]], width = 20),
               "This is the first\nstring")
  expect_named(wrap_labels(my.text, width = 20), c("A", "B"))
  expect_named(wrap_labels(unname(my.text), width = 20), NULL)
  expect_equal(wrap_labels(my.text[["A"]], width = 20, indent = 2),
               "  This is the first\nstring")
  expect_equal(wrap_labels(my.text[["A"]], width = 20, indent = -2),
               "This is the first\n  string")
})

test_that("test width", {
  my.text.line <- "This is a rather long string that needs wrapping"

  expect_equal(wrap_labels(my.text.line, width = 50), my.text.line)
  expect_equal(wrap_labels(my.text.line, width = 30),
               "This is a rather long string\nthat needs wrapping")
  expect_equal(wrap_labels(my.text.line, width = 20),
               "This is a rather\nlong string that\nneeds wrapping")
  expect_equal(wrap_labels(my.text.line, width = 10),
               "This is a\nrather\nlong\nstring\nthat\nneeds\nwrapping")
  expect_equal(wrap_labels(my.text.line, width = 5),
               "This\nis a\nrather\nlong\nstring\nthat\nneeds\nwrapping")
  # split into words at white space!
  expect_equal(wrap_labels(my.text.line, width = 0),
               "This\nis\na\nrather\nlong\nstring\nthat\nneeds\nwrapping")
  # surprising behaviour of strwrap() retained
  expect_equal(wrap_labels(my.text.line, width = -10),
               wrap_labels(my.text.line, width = 0),)

})

#' Text with Normalised Parent Coordinates
#'
#' \code{geom_text_npc()} adds text directly to the plot.
#' \code{geom_label_npc()} draws a rectangle behind the text, making it easier
#' to read. The difference is that \code{x} and \code{y} mappings are expected
#' to be given in \code{npc} graphic units, using pseudo-aesthetics. Their
#' intended use is to add annotations to a plot.
#'
#' @details These geoms are identical to 'ggplot2'
#'   \code{\link[ggplot2]{geom_text}} and \code{\link[ggplot2]{geom_label}}
#'   except that they interpret \code{npcx} and \code{npcy} positions in
#'   \code{npc} units. They translate \code{npcx} and \code{npcy} coordinates
#'   using a pseudo-aesthetic with a fixed scale, the translation is done
#'   separately for each plot panel. All aesthetics other than \emph{x} and
#'   \emph{y} and grouping work as in normal geoms. These include
#'   \code{linetype} and \code{angle} in \code{geom_label_npc()}.
#'
#' @inheritSection geom_text_s Plot boundaries and clipping
#'
#' @inheritSection geom_text_s Alignment
#'
#' @section Alignment: With textual positions and groups a shift is added to
#'   successive labels to avoid overlaps. The shift is based on grouping,
#'   however unused levels are not dropped. In plots with faceting, if not
#'   all groups appear in each panel, there will be blank spaces in between
#'   labels. To solve this pass numeric values for the npc coordinates of each
#'   label instead of character strings.
#'
#' @param mapping The aesthetic mapping, usually constructed with
#'   \code{\link[ggplot2]{aes}}. Only needs
#'   to be set at the layer level if you are overriding the plot defaults.
#' @param data A layer specific data set - only needed if you want to override
#'   the plot defaults.
#' @param stat The statistical transformation to use on the data for this layer,
#'   as a string.
#' @param na.rm If \code{FALSE} (the default), removes missing values with a
#'   warning.  If \code{TRUE} silently removes missing values.
#' @param position Position adjustment, either as a string, or the result of a
#'   call to a position adjustment function.
#' @param ... other arguments passed on to \code{\link[ggplot2]{layer}}. This
#'   can include aesthetics whose values you want to set, not map. See
#'   \code{\link[ggplot2]{layer}} for more details.
#' @param parse If TRUE, the labels will be parsed into expressions and
#'   displayed as described in \code{?plotmath}.
#' @param show.legend logical. Should this layer be included in the legends?
#'   \code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
#'   never includes, and \code{TRUE} always includes.
#' @param inherit.aes If \code{FALSE}, overrides the default aesthetics, rather
#'   than combining with them. This is most useful for helper functions that
#'   define both data and aesthetics and shouldn't inherit behaviour from the
#'   default plot specification, e.g. \code{\link[ggplot2]{borders}}.
#' @param nudge_x,nudge_y Horizontal and vertical adjustment to nudge labels by.
#'   Useful for offsetting text from points, particularly on discrete scales.
#' @param check_overlap If `TRUE`, text that overlaps previous text in the same
#'   layer will not be plotted.
#' @param size.unit How the `size` aesthetic is interpreted: as millimetres
#'   (`"mm"`, default), points (`"pt"`), centimetres (`"cm"`), inches (`"in"`),
#'   or picas (`"pc"`).
#'
#' @seealso \code{\link[ggplot2]{geom_text}} and
#'    \code{\link[ggplot2]{geom_label}} for additional details.
#'
#' @rdname geom_text_npc
#'
#' @export
#' @examples
#'
#' df <- data.frame(
#'   x = c(0, 0, 1, 1, 0.5),
#'   x.chr = c("left", "left", "right", "right", "center"),
#'   y = c(0, 1, 0, 1, 0.5),
#'   y.chr = c("bottom", "top", "bottom", "top", "middle"),
#'   text = c("bottom-left", "top-left", "bottom-right", "top-right", "center-middle")
#' )
#'
#' ggplot(df) +
#'   geom_text_npc(aes(npcx = x, npcy = y, label = text))
#'
#' ggplot(df) +
#'   geom_text_npc(aes(npcx = x.chr, npcy = y.chr, label = text))
#'
#' ggplot(df) +
#'   geom_text_npc(aes(npcx = x.chr, npcy = y.chr, label = text),
#'                 angle = 90)
#'
#' ggplot(data = mtcars, mapping = aes(wt, mpg)) +
#'   geom_point() +
#'   geom_text_npc(data = df, aes(npcx = x, npcy = y, label = text))
#'
#' ggplot(data = mtcars, mapping = aes(wt, mpg)) +
#'   geom_point() +
#'   geom_text_npc(data = df, aes(npcx = x, npcy = y, label = text)) +
#'   expand_limits(y = 40, x = 6)
#'
#' ggplot(data = mtcars) +
#'   geom_point(mapping = aes(wt, mpg)) +
#'   geom_label_npc(data = df, aes(npcx = x, npcy = y, label = text))
#'
#' ggplot(data = mtcars) +
#'   geom_point(mapping = aes(wt, mpg)) +
#'   geom_label_npc(data = df, aes(npcx = x.chr, npcy = y.chr, label = text),
#'                  angle = 90) # ignored by ggplot2 < 3.5.0
#'
geom_text_npc <- function(mapping = NULL, data = NULL,
                      stat = "identity", position = "identity",
                      ...,
                      parse = FALSE,
                      nudge_x = 0,
                      nudge_y = 0,
                      check_overlap = FALSE,
                      na.rm = FALSE,
                      show.legend = FALSE,
                      inherit.aes = FALSE)
{
  if (!missing(nudge_x) || !missing(nudge_y)) {
    if (!missing(position) && position != "identity") {
      stop("You must specify either `position` or `nudge_x`/`nudge_y`.", call. = FALSE)
    }

    position <- ggplot2::position_nudge(nudge_x, nudge_y)
  }

  ggplot2::layer(
    data = data,
    mapping = mapping,
    stat = stat,
    geom = GeomTextNpc,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      parse = parse,
      check_overlap = check_overlap,
      na.rm = na.rm,
      ...
    )
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
GeomTextNpc <- ggplot2::ggproto("GeomTextNpc", ggplot2::Geom,
  required_aes = c("npcx", "npcy", "label"),

  default_aes = ggplot2::aes(
    colour = "black",
    family = "",
    size = 3.87,
    angle = 0,
    hjust = "inward",
    vjust = "inward",
    alpha = NA,
    fontface = 1,
    lineheight = 1.2
  ),

  draw_panel = function(data, panel_params, coord, parse = FALSE,
                        na.rm = FALSE, check_overlap = FALSE) {

    data$npcx <- compute_npcx(data$npcx)
    data$npcy <- compute_npcy(data$npcy)

    ranges <- coord$backtransform_range(panel_params)

    data$x <- ranges$x[1] + data$npcx * (ranges$x[2] - ranges$x[1])
    data$y <- ranges$y[1] + data$npcy * (ranges$y[2] - ranges$y[1])

    ggplot2::GeomText$draw_panel(data = data,
                                 panel_params = panel_params,
                                 coord = coord,
                                 parse = parse,
                                 na.rm = na.rm,
                                 check_overlap = check_overlap)
  },

  draw_key = function(...) {
    grid::nullGrob()
  }
)

context("geom_text_npc")

test_that("multiple_rows_tb", {
  df <- data.frame(
    x = c(0, 0, 1, 1, 0.5),
    y = c(0, 1, 0, 1, 0.5),
    text = c("bottom-left", "top-left", "bottom-right", "top-right", "center"),
    to.parse = c("alpha^2", "beta[2]", "delta[a]", "gamma^{2345}", "sqrt(a, b)=infinity")
  )
  vdiffr::expect_doppelganger("geom_text_npc1",
                              ggplot(df) +
                                geom_text_npc(aes(npcx = x, npcy = y, label = text)))
  vdiffr::expect_doppelganger("geom_text_npc2",
                              ggplot(data = mtcars, mapping = aes(wt, mpg)) +
                                geom_point() +
                                geom_text_npc(data = df, aes(npcx = x, npcy = y, label = text)))
  vdiffr::expect_doppelganger("geom_text_npc3",
                              ggplot(data = mtcars, mapping = aes(wt, mpg)) +
                                geom_point() +
                                geom_text_npc(data = df, aes(npcx = x, npcy = y, label = text)) +
                                expand_limits(y = 40, x = 6))
  vdiffr::expect_doppelganger("geom_text_npc4",
                              ggplot(data = mtcars) +
                                geom_point(mapping = aes(wt, mpg)) +
                                geom_text_npc(data = df,
                                              mapping = aes(npcx = x, npcy = y, label = to.parse),
                                              parse = TRUE))
  expect_error(
    ggplot(data = mtcars) +
      geom_point(mapping = aes(wt, mpg)) +
      geom_text_npc(data = df,
                    mapping = aes(npcx = x, npcy = y, label = to.parse),
                    parse = TRUE,
                    nudge_y = 0.1,
                    position = "stack")
  )

  expect_no_error(
    ggplot(data = mtcars) +
      geom_point(mapping = aes(wt, mpg)) +
      geom_text_npc(data = df,
                    mapping = aes(npcx = x, npcy = y, label = to.parse),
                    parse = TRUE,
                    nudge_x = 0.1,
                    position = "identity")
  )
})

#' Points linked by a segment
#'
#' The geometry \code{"geom_point_s"} provides a super set of the capabilities of
#' geom \code{\link[ggplot2]{geom_point}} from package 'ggplot2' by allowing
#' plotting of arrows or segments joining the original position of displaced observations
#' to their current position rendered as points or graphic symbols. The most
#' common use is to demonstrate the action of different position functions. It
#' can be also used to highlight observations.
#'
#' @details The plotting of segments is similar in idea to that implemented in
#'   \code{\link[ggrepel]{geom_text_repel}} and relies on position functions
#'   that rename instead of only replacing the original \code{x} and \code{y}
#'   coordinates from the \code{data} object.
#'
#'   By default this geom uses \code{\link{position_nudge_center}} which is backwards
#'   compatible with \code{\link[ggplot2]{position_nudge}} but provides additional control
#'   on the direction of the nudging.
#'
#' @inheritSection geom_text_s Position functions
#'
#' @inherit geom_grob note return
#
#' @seealso \code{\link[ggplot2]{geom_point}}.
#'
#' @param mapping Set of aesthetic mappings created by
#'   \code{\link[ggplot2]{aes}}. If specified and \code{inherit.aes = TRUE} (the
#'   default), is combined with the default mapping at the top level of the
#'   plot. You only need to supply \code{mapping} if there isn't a mapping
#'   defined for the plot.
#' @param data A data frame. If specified, overrides the default data frame
#'   defined at the top level of the plot.
#' @param stat The statistical transformation to use on the data for this layer,
#'   as a string.
#' @param position Position adjustment, either as a string, or the result of a
#'   call to a position adjustment function.
#' @param na.rm If \code{FALSE} (the default), removes missing values with a
#'   warning.  If \code{TRUE} silently removes missing values.
#' @param show.legend logical. Should this layer be included in the legends?
#'   \code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
#'   never includes, and \code{TRUE} always includes.
#' @param inherit.aes If \code{FALSE}, overrides the default aesthetics, rather
#'   than combining with them. This is most useful for helper functions that
#'   define both data and aesthetics and shouldn't inherit behaviour from the
#'   default plot specification, e.g. \code{\link[ggplot2]{borders}}.
#' @param ... other arguments passed on to \code{\link[ggplot2]{layer}}. There
#'   are three types of arguments you can use here:
#'   \itemize{ \item Aesthetics: to set an aesthetic to a fixed value, like
#'   \code{colour = "red"} or \code{size = 3}. \item Other arguments to the
#'   layer, for example you override the default \code{stat} associated with the
#'   layer. \item Other arguments passed on to the stat. }
#' @param nudge_x,nudge_y Horizontal and vertical adjustments to nudge the
#'   starting position of each text label. The units for \code{nudge_x} and
#'   \code{nudge_y} are the same as for the data units on the x-axis and y-axis.
#' @param move.point logical If \code{TRUE} the point is drawn at the nudged
#'   position while if \code{FALSE} the point is drawn at the original position.
#' @param default.colour,default.color A colour definition to use for elements not targeted by
#'   the colour aesthetic.
#' @param colour.target,color.target A character string, one of \code{"all"},
#'   \code{"point"} and \code{"segment"} or \code{"none"}.
#' @param default.alpha numeric in [0..1] A transparency value to use for
#'   elements not targeted by the alpha aesthetic.
#' @param alpha.target A character string, one of \code{"all"},
#'   \code{"segment"}, \code{"point"}, or \code{"none"}.
#' @param add.segments logical Display connecting segments or arrows between
#'   original positions and displaced ones if both are available.
#' @param box.padding,point.padding numeric By how much each end of the segments
#'   should shortened in mm.
#' @param segment.linewidth numeric Width of the segments or arrows in mm.
#' @param min.segment.length numeric Segments shorter that the minimum length
#'   are not rendered, in mm.
#' @param arrow specification for arrow heads, as created by
#'   \code{\link[grid]{arrow}}
#'
#' @export
#'
#' @examples
#'
#' # Same output as with geom_point()
#' ggplot(mpg[1:20, ],
#'        aes(cyl, hwy)) +
#'   geom_point_s(colour = "blue")
#'
#' # with segment drawn after nudging
#' ggplot(mpg[1:20, ],
#'        aes(cyl, hwy, label = drv)) +
#'   geom_point_s(position = position_nudge_keep(x = 0.2),
#'                colour = "red") +
#'   geom_point_s(colour = "blue") +
#'   expand_limits(x = c(3.5, 8.5))
#'
#' ggplot(mpg[1:20, ],
#'        aes(cyl, hwy, label = drv)) +
#'   geom_point_s(position = position_nudge_keep(x = 0.2),
#'                colour = "blue",
#'                move.point = FALSE) +
#'   expand_limits(x = c(3.5, 8.5))
#'
#' # with segment drawn after nudging
#' ggplot(mpg[1:20, ],
#'        aes(cyl, hwy, label = drv)) +
#'   geom_point_s(position = position_nudge_keep(x = 0.2),
#'                colour = "red",
#'                colour.target = "all") +
#'   geom_point_s(colour = "blue")
#'
#' ggplot(mpg[1:20, ],
#'        aes(cyl, hwy, label = drv)) +
#'   geom_point_s(position = position_nudge_keep(x = 0.2),
#'                colour = "red",
#'                colour.target = "segment") +
#'   geom_point_s(colour = "blue")
#'
#' ggplot(mpg[1:20, ],
#'        aes(cyl, hwy, label = drv)) +
#'   geom_point_s(position = position_nudge_keep(x = 0.2),
#'                colour = "red",
#'                colour.target = "point") +
#'   geom_point_s(colour = "blue")
#'
#' ggplot(mpg[1:50, ],
#'        aes(cyl, hwy, label = drv)) +
#'   geom_point_s(position = position_jitternudge(width = 0.66, height = 2,
#'                                                seed = 456,
#'                                                nudge.from = "jittered",
#'                                                kept.origin = "original"),
#'                colour = "red",
#'                alpha = 0.3, alpha.target = "segment",
#'                arrow = grid::arrow(length = grid::unit(0.4, "lines"),
#'                                    ends = "first")) +
#'   geom_point_s(colour = "blue")
#'
geom_point_s <- function(mapping = NULL, data = NULL,
                         stat = "identity", position = "identity",
                         ...,
                         nudge_x = 0,
                         nudge_y = 0,
                         move.point = TRUE,
                         arrow = grid::arrow(length = unit(1/3, "lines"),
                                             ends = "first"),
                         default.colour = NULL,
                         default.color = default.colour,
                         colour.target = "point",
                         color.target = colour.target,
                         default.alpha = NA,
                         alpha.target = "all",
                         add.segments = TRUE,
                         box.padding = 0.25,
                         point.padding = 1e-06,
                         segment.linewidth = 0.5,
                         min.segment.length = 0,
                         na.rm = FALSE,
                         show.legend = NA,
                         inherit.aes = TRUE) {

  colour.target <-
    rlang::arg_match(color.target,
                     values = c("point", "all", "segment", "none"),
                     multiple = TRUE)
  alpha.target <-
    rlang::arg_match(alpha.target,
                     values = c("point", "all", "segment", "none"),
                     multiple = TRUE)

  if (!missing(nudge_x) || !missing(nudge_y)) {
    if (!missing(position) && position != "identity") {
      rlang::abort("You must specify either `position` or `nudge_x`/`nudge_y`.")
    }
    # We drop original positions if they are not needed for segments
    # (Justification is not supported)
    position <-
      position_nudge_center(nudge_x, nudge_y,
                            kept.origin = ifelse(add.segments,
                                                 "original", "none"))
  }

  layer(
    data = data,
    mapping = mapping,
    stat = stat,
    geom = GeomPointS,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      move.point = move.point,
      default.colour = default.color,
      colour.target = colour.target,
      default.alpha = default.alpha,
      alpha.target = alpha.target,
      add.segments = add.segments,
      box.padding = box.padding,
      point.padding = point.padding,
      segment.linewidth = segment.linewidth,
      min.segment.length = min.segment.length,
      arrow = arrow,
      na.rm = na.rm,
      ...
    )
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
GeomPointS <-
  ggplot2::ggproto("GeomPointS", Geom,
                   required_aes = c("x", "y"),
                   non_missing_aes = c("size", "shape", "colour"),
                   default_aes = ggplot2::aes(
                     shape = 19,
                     colour = "black",
                     size = 1.5,
                     fill = NA,
                     alpha = NA,
                     stroke = 0.5
                   ),

                   draw_panel = function(data,
                                         panel_params,
                                         coord,
                                         move.point = TRUE,
                                         default.colour = NULL,
                                         colour.target = "point",
                                         default.alpha = NA,
                                         alpha.target = "segment",
                                         na.rm = FALSE,
                                         arrow = NULL,
                                         box.padding = 0.25,
                                         point.padding = 1e-06,
                                         segment.linewidth = 0.5,
                                         min.segment.length = 0,
                                         add.segments = FALSE) {

                     if (is.character(data$shape)) {
                       data$shape <- translate_shape_string(data$shape)
                     }

                     if (nrow(data) == 0L) {
                       return(nullGrob())
                     }

                     default.colour <- check_default_colour(default.colour)

                     add.segments <- add.segments && all(c("x_orig", "y_orig") %in% colnames(data))

                     data <- coord$transform(data, panel_params)
                     if (all(c("x_orig", "y_orig") %in% colnames(data))) {
                       data_orig <- data.frame(x = data$x_orig, y = data$y_orig)
                       data_orig <- coord$transform(data_orig, panel_params)
                       data$x_orig <- data_orig$x
                       data$y_orig <- data_orig$y
                     }

                     if (add.segments) {
                       segments.data <-
                         shrink_segments(data,
                                         point.padding = point.padding,
                                         box.padding = box.padding,
                                         min.segment.length = min.segment.length)
                     }
                     # loop needed as gpar is not vectorized
                     all.grobs <- grid::gList()

                     if (move.point) {
                       x0.name <- "x"
                       y0.name <- "y"
                       x1.name <- "x_orig"
                       y1.name <- "y_orig"
                     } else {
                       x0.name <- "x_orig"
                       y0.name <- "y_orig"
                       x1.name <- "x"
                       y1.name <- "y"
                     }
                     for (row.idx in 1:nrow(data)) {
                       row <- data[row.idx, , drop = FALSE]
                       point.alpha <-
                         ifelse(any(alpha.target %in% c("all", "point")),
                                row$alpha, default.alpha)
                       segment.alpha <-
                         ifelse(any(alpha.target %in% c("all", "segment")),
                                row$alpha, default.alpha)
                       user.grob <- grid::pointsGrob(
                         row[[x0.name]], row[[y0.name]], default.units = "native",
                         pch = row$shape,
                         gp = gpar(
                           col = ifelse(any(colour.target %in% c("all", "point")),
                                        ggplot2::alpha(row$colour, point.alpha),
                                        ggplot2::alpha(default.colour, point.alpha)),
                           fill = alpha(row$fill, point.alpha),
                           # Stroke is added around the outside of the point
                           fontsize = row$size * .pt + row$stroke * ggplot2::.stroke / 2,
                           lwd = row$stroke * ggplot2::.stroke / 2
                         )
                       )

                       # give unique name to each grob
                       user.grob$name <- paste("point.s.grob", row$group, row.idx, sep = ".")

                       if (add.segments) {
                         segment.row <- segments.data[row.idx, , drop = FALSE]
                         if (segment.row$too.short) {
                           segment.grob <- grid::nullGrob()
                         } else {
                           segment.grob <-
                             grid::segmentsGrob(x0 = segment.row[[x0.name]],
                                                y0 = segment.row[[y0.name]],
                                                x1 = segment.row[[x1.name]],
                                                y1 = segment.row[[y1.name]],
                                                arrow = arrow,
                                                gp = grid::gpar(
                                                  col = if (segment.linewidth == 0) NA else # lwd = 0 is invalid in 'grid'
                                                    ifelse(any(colour.target %in% c("all", "segment")),
                                                           ggplot2::alpha(row$colour, segment.alpha),
                                                           ggplot2::alpha(default.colour, segment.alpha)),
                                                  lwd = (if (segment.linewidth == 0) 0.5 else segment.linewidth) * ggplot2::.stroke),
                                                name = paste("text.s.segment", row$group, row.idx, sep = "."))
                         }
                         all.grobs <- grid::gList(all.grobs, segment.grob, user.grob)
                       } else {
                         all.grobs <- grid::gList(all.grobs, user.grob)
                       }
                     }

                     # name needs to be unique within plot, so we would to know other layers
                     #                     grid::grobTree(children = all.grobs, name = "geom.point.s.panel")
                     grid::grobTree(children = all.grobs)

                   },

                   draw_key = ggplot2::draw_key_point
  )

# adopts from 'ggplot2' rejected pull-request #5143 by Teun van den Brand
# addresing issue #5088.
# 'blank' and 'dot' deviate from the grammar of graphics but
# are consistent with underlying R and 'grid' pch codes
#
translate_shape_string <- function(shape_string) {
  # strings of length 0 or 1 are interpreted as symbols by grid
  if (nchar(shape_string[1]) <= 1) {
    return(shape_string)
  }

  pch_table <- c(
    "square open"           = 0,
    "circle open"           = 1,
    "triangle open"         = 2,
    "plus"                  = 3,
    "cross"                 = 4,
    "diamond open"          = 5,
    "triangle down open"    = 6,
    "square cross"          = 7,
    "asterisk"              = 8,
    "diamond plus"          = 9,
    "circle plus"           = 10,
    "star"                  = 11,
    "square plus"           = 12,
    "circle cross"          = 13,
    "square triangle"       = 14,
    "triangle square"       = 14,
    "square"                = 15,
    "circle small"          = 16,
    "triangle"              = 17,
    "diamond"               = 18,
    "circle"                = 19,
    "bullet"                = 20,
    "circle filled"         = 21,
    "square filled"         = 22,
    "diamond filled"        = 23,
    "triangle filled"       = 24,
    "triangle down filled"  = 25,
    "dot"                   = 46,
    "blank"                 = NA
  )

  shape_match <- charmatch(shape_string, names(pch_table), nomatch = 99)

  invisible_shape <- anyNA(shape_match)
  invalid_strings <- shape_match == 99
  nonunique_strings <- shape_match == 0

  if (any(invalid_strings)) {
    bad_string <- unique(shape_string[invalid_strings])
    n_bad <- length(bad_string)

    collapsed_names <- sprintf("\n* '%s'", bad_string[1:min(5, n_bad)])

    more_problems <- if (n_bad > 5) {
      sprintf("\n* ... and %d more problem%s", n_bad - 5, ifelse(n_bad > 6, "s", ""))
    } else {
      ""
    }

    stop(paste("Can't find shape name:", collapsed_names, more_problems))
  }

  if (any(nonunique_strings)) {
    bad_string <- unique(shape_string[nonunique_strings])
    n_bad <- length(bad_string)

    n_matches <- vapply(
      bad_string[1:min(5, n_bad)],
      function(shape_string) sum(grepl(paste0("^", shape_string), names(pch_table))),
      integer(1)
    )

    collapsed_names <- sprintf(
      "\n* '%s' partially matches %d shape names",
      bad_string[1:min(5, n_bad)], n_matches
    )

    more_problems <- if (n_bad > 5) {
      sprintf("\n* ... and %d more problem%s", n_bad - 5, ifelse(n_bad > 6, "s", ""))
    } else {
      ""
    }

    stop(paste("Shape names must be unambiguous:", collapsed_names, more_problems))
  }

  if (invisible_shape) {
    message("Invisible shape in use: not all data points plotted!!")
  }
  unname(pch_table[shape_match])
}

context("geom_point_s")

test_that("geom_point_s runs successfully", {
  expect_no_error(
    ggplot(mpg[1:20, ],
           aes(cyl, hwy, label = drv)) +
      geom_point_s(position = position_nudge_keep(x = 0.2),
                   color = "red") +
      geom_point_s()
  )
})


test_that("geom_point_s pos_or_nudge", {
  expect_error(
    ggplot(mpg[1:20, ],
           aes(cyl, hwy, label = drv)) +
      geom_point_s(position = "",
                   nudge_x = 0.5),
    "You must specify either `position` or `nudge_x`/`nudge_y`."
  )

  expect_silent(
    ggplot(mpg[1:20, ],
           aes(cyl, hwy, label = drv)) +
      geom_point_s(nudge_x = 0.2, nudge_y = -0.2)
    )

})

test_that("translate_shape_string maps correctly", {
  expected <- c(0:14, 14:25)
  shape.strings <- c("square open",
  "circle open",
  "triangle open",
  "plus",
  "cross",
  "diamond open",
  "triangle down open",
  "square cross",
  "asterisk",
  "diamond plus",
  "circle plus",
  "star",
  "square plus",
  "circle cross",
  "square triangle",
  "triangle square",
  "square",
  "circle small",
  "triangle",
  "diamond",
  "circle",
  "bullet",
  "circle filled",
  "square filled",
  "diamond filled",
  "triangle filled",
  "triangle down filled")
  result <- translate_shape_string(shape.strings)
  expect_equal(expected, result)

  expect_error(translate_shape_string("wrong shape")) # no match
  expect_error(translate_shape_string("circ")) # multiple matches

  # chars are passed through
  expect_equal(translate_shape_string("c"), "c")
  expect_equal(translate_shape_string(""), "")
  expect_equal(translate_shape_string("1"), "1")
  expect_error(translate_shape_string("11"))
})

#' @title Replace labels in data based on 2D density
#'
#' @description \code{stat_dens2d_labels()} Sets values mapped to the
#'   \code{label} aesthetic to \code{""} or a user provided character string
#'   based on the local density in regions of a plot panel. Its main use is
#'   together with repulsive geoms from package \code{\link[ggrepel]{ggrepel}}.
#'   If there is no mapping to \code{label} in \code{data}, the mapping is set
#'   to \code{rownames(data)}, with a message.
#'
#' @details \code{stat_dens2d_labels()} is designed to work together with
#'   geometries from package 'ggrepel'. To avoid text labels being plotted over
#'   unlabelled points all the rows in data need to be retained but
#'   labels replaced with the empty character string, \code{""}. Function
#'   \code{\link{stat_dens2d_filter}} cannot be used with the repulsive geoms
#'   from 'ggrepel' because it drops observations.
#'
#'   \code{stat_dens2d_labels()} can be useful also in other situations, as the
#'   substitution character string can be set by the user by passing an argument
#'   to \code{label.fill}. If this argument is \code{NULL} the unselected rows
#'   are filtered out identically as by \code{stat_dens2d_filter}.
#'
#'   The local density of observations in 2D (\emph{x} and \emph{y}) is computed
#'   with function \code{\link[MASS]{kde2d}} and used to select observations,
#'   passing to the geom all the rows in its \code{data} input but with with the
#'   text of labels replaced in those "not kept". The default is to select
#'   observations in sparse regions of the plot, but the selection can be
#'   inverted so that only observations in the densest regions are returned.
#'   Specific observations can be protected from having the label replaced by
#'   passing a suitable argument to \code{keep.these}. Logical and integer
#'   vectors function as indexes to rows in \code{data}, while a character
#'   vector is compared to values in the variable mapped to the \code{label}
#'   aesthetic. A function passed as argument to \code{keep.these} will receive
#'   as its first argument the values in the variable mapped to \code{label} and
#'   should return a character, logical or numeric vector as described above.
#'
#'   How many labels are retained intact in addition to those in
#'   \code{keep.these} is controlled with arguments passed to \code{keep.number}
#'   and \code{keep.fraction}. \code{keep.number} sets the maximum number of
#'   observations selected, whenever \code{keep.fraction} results in fewer
#'   observations selected, it is obeyed.
#'
#'   Computation of density and of the default bandwidth require at least
#'   two observations with different values. If data do not fulfill this
#'   condition, they are kept only if \code{keep.fraction = 1}. This is correct
#'   behavior for a single observation, but can be surprising in the case of
#'   multiple observations.
#'
#'   Parameters \code{keep.these} and \code{exclude.these} make it possible to
#'   force inclusion or exclusion of observations after the density is computed.
#'   In case of conflict, \code{exclude.these} overrides \code{keep.these}.
#'
#' @note Which points are kept and which not depends on how dense a grid is used
#'   and how flexible the density surface estimate is. This depends on the
#'   values passed as arguments to parameters \code{n}, \code{bw} and
#'   \code{kernel}. It is also important to be aware that both
#'   \code{geom_text()} and \code{geom_text_repel()} can avoid overplotting by
#'   discarding labels at the plot rendering stage, i.e., what is plotted may
#'   differ from what is returned by this statistic.
#'
#' @param mapping The aesthetic mapping, usually constructed with
#'   \code{\link[ggplot2]{aes}} or \code{\link[ggplot2]{aes_}}. Only needs
#'   to be set at the layer level if you are overriding the plot defaults.
#' @param data A layer specific dataset - only needed if you want to override
#'   the plot defaults.
#' @param geom The geometric object to use display the data.
#' @param keep.fraction numeric [0..1]. The fraction of the observations (or
#'   rows) in \code{data} to be retained.
#' @param keep.number integer Set the maximum number of observations to retain,
#'   effective only if obeying \code{keep.fraction} would result in a larger
#'   number.
#' @param keep.sparse logical If \code{TRUE}, the default, observations from the
#'   more sparse regions are retained, if \code{FALSE} those from the densest
#'   regions.
#' @param keep.these,exclude.these character vector, integer vector, logical
#'   vector or function that takes one or more variables in data selected by
#'   \code{these.target}. Negative integers behave as in R's extraction methods.
#'   The rows from \code{data} indicated by \code{keep.these} and
#'   \code{exclude.these} are kept or excluded irrespective of the local
#'   density.
#' @param these.target character, numeric or logical selecting one or more
#'   column(s) of \code{data}. If \code{TRUE} the whole \code{data} object is
#'   passed.
#' @param pool.along character, one of \code{"none"} or \code{"x"},
#'   indicating if selection should be done pooling the observations along the
#'   \emph{x} aesthetic, or separately on either side of \code{xintercept}.
#' @param xintercept,yintercept numeric The split points for the data filtering.
#' @param invert.selection logical If \code{TRUE}, the complement of the
#'   selected rows are returned.
#' @param h vector of bandwidths for x and y directions. Defaults to normal
#'   reference bandwidth (see bandwidth.nrd). A scalar value will be taken to
#'   apply to both directions.
#' @param n Number of grid points in each direction. Can be scalar or a length-2
#'   integer vector
#' @param label.fill character vector of length 1, a function or \code{NULL}.
#' @param return.density logical vector of lenght 1. If \code{TRUE} add columns
#'   \code{"density"} and \code{"keep.obs"} to the returned data frame.
#' @param position The position adjustment to use for overlapping points on this
#'   layer
#' @param show.legend logical. Should this layer be included in the legends?
#'   \code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
#'   never includes, and \code{TRUE} always includes.
#' @param inherit.aes If \code{FALSE}, overrides the default aesthetics, rather
#'   than combining with them. This is most useful for helper functions that
#'   define both data and aesthetics and shouldn't inherit behaviour from the
#'   default plot specification, e.g. \code{\link[ggplot2]{borders}}.
#' @param ... other arguments passed on to \code{\link[ggplot2]{layer}}. This
#'   can include aesthetics whose values you want to set, not map. See
#'   \code{\link[ggplot2]{layer}} for more details.
#' @param na.rm	a logical value indicating whether NA values should be stripped
#'   before the computation proceeds.
#'
#' @return A plot layer instance. Using as output \code{data} the input
#'   \code{data} after value substitution based on a 2D the filtering criterion.
#'
#' @seealso \code{\link{stat_dens2d_filter}} and \code{\link[MASS]{kde2d}} used
#'   internally. Parameters \code{n}, \code{h} in this statistic correspond to
#'   the parameters with the same name in this imported function. Limits are set
#'   to the limits of the plot scales.
#'
#' @family statistics returning a subset of data
#'
#' @export
#'
#' @examples
#'
#' random_string <-
#'   function(len = 6) {
#'     paste(sample(letters, len, replace = TRUE), collapse = "")
#'   }
#'
#' # Make random data.
#' set.seed(1001)
#' d <- tibble::tibble(
#'   x = rnorm(100),
#'   y = rnorm(100),
#'   group = rep(c("A", "B"), c(50, 50)),
#'   lab = replicate(100, { random_string() })
#' )
#'
#' # using defaults
#' ggplot(data = d, aes(x, y, label = lab)) +
#'   geom_point() +
#'   stat_dens2d_labels()
#'
#' ggplot(data = d, aes(x, y, label = lab)) +
#'   geom_point() +
#'   stat_dens2d_labels(keep.these = "zoujdg")
#'
#' ggplot(data = d, aes(x, y, label = lab)) +
#'   geom_point() +
#'   stat_dens2d_labels(keep.these = function(x) {grepl("^z", x)})
#'
#' ggplot(data = d, aes(x, y, label = lab)) +
#'   geom_point() +
#'   stat_dens2d_labels(geom = "text_s",
#'                      position = position_nudge_center(x = 0.1, y = 0.1,
#'                                                       center_x = mean,
#'                                                       center_y = mean),
#'                      vjust = "outward_mean", hjust = "outward_mean") +
#'   expand_limits(x = c(-4, 4.5))
#'
#' ggrepel.installed <- requireNamespace("ggrepel", quietly = TRUE)
#' if (ggrepel.installed) {
#'   library(ggrepel)
#'
#'   ggplot(data = d, aes(x, y, label = lab, colour = group)) +
#'     geom_point() +
#'     stat_dens2d_labels(geom = "text_repel")
#'
#'   ggplot(data = d, aes(x, y, label = lab, colour = group)) +
#'     geom_point() +
#'     stat_dens2d_labels(geom = "text_repel", label.fill = NA)
#'
#' # we keep labels starting with "a" across the whole plot, but all in sparse
#' # regions. To achieve this we pass as argument to label.fill a function
#' # instead of a character string.
#'   label.fun <- function(x) {ifelse(grepl("^a", x), x, "")}
#'   ggplot(data = d, aes(x, y, label = lab, colour = group)) +
#'     geom_point() +
#'     stat_dens2d_labels(geom = "text_repel", label.fill = label.fun)
#' }
#'
#' # Using geom_debug_group() we can see that all 100 rows in \code{d} are
#' # returned.
#' gginnards.installed <- requireNamespace("gginnards", quietly = TRUE)
#' if (gginnards.installed) {
#'   library(gginnards)
#'
#'   ggplot(data = d, aes(x, y, label = lab)) +
#'     geom_point() +
#'     stat_dens2d_labels(geom = "debug_group")
#' }
#' if (gginnards.installed) {
#'   ggplot(data = d, aes(x, y, label = lab)) +
#'     geom_point() +
#'     stat_dens2d_labels(geom = "debug_group", return.density = TRUE)
#' }
#' if (gginnards.installed) {
#'   ggplot(data = d, aes(x, y)) +
#'     geom_point() +
#'     stat_dens2d_labels(geom = "debug_group")
#' }
#'
stat_dens2d_labels <-
  function(mapping = NULL,
           data = NULL,
           geom = "text",
           position = "identity",
           ...,
           keep.fraction = 0.10,
           keep.number = Inf,
           keep.sparse = TRUE,
           keep.these = FALSE,
           exclude.these = FALSE,
           these.target = "label",
           pool.along = c("xy", "x", "y", "none"),
           xintercept = 0,
           yintercept = 0,
           invert.selection = FALSE,
           h = NULL,
           n = NULL,
           label.fill = "",
           return.density = FALSE,
           na.rm = TRUE,
           show.legend = FALSE,
           inherit.aes = TRUE) {

    pool.along <- rlang::arg_match(pool.along)

    if (length(label.fill) > 1L) {
      stop("Length for 'label.fill' is not 1: ", label.fill)
    }
    if (is.numeric(label.fill)) {
      stop("'label.fill' should not be a 'numeric' value: ", label.fill)
    }
    if (any(is.na(keep.fraction) | keep.fraction < 0 | keep.fraction > 1)) {
      stop("Out of range or missing value for 'keep.fraction': ", keep.fraction)
    }
    if (any(is.na(keep.number) | keep.number < 0)) {
      stop("Out of range or missing value for 'keep.number': ", keep.number)
    }
    max.expected.length <- c(none = 4L, x = 2L, y = 2L, xy = 1L)[pool.along]
    if (length(keep.fraction) > max.expected.length) {
      if (max.expected.length == 4L) {
        stop("Length of 'keep.fraction' should not exceed 4")
      } else {
        warning("'keep.fraction' is too long, did you forget to set 'pool.along'?")
      }
    }
    if (length(keep.number) > max.expected.length) {
      if (max.expected.length == 4L) {
        stop("Length of 'keep.number' should not exceed 4")
      } else {
        warning("'keep.number' is too long, did you forget to set 'pool.along'?")
      }
    }

    ggplot2::layer(
      stat = StatDens2dLabels, data = data, mapping = mapping, geom = geom,
      position = position, show.legend = show.legend, inherit.aes = inherit.aes,
      params = list(na.rm = na.rm,
                    keep.fraction = keep.fraction,
                    keep.number = keep.number,
                    keep.sparse = keep.sparse,
                    keep.these = keep.these,
                    exclude.these = exclude.these,
                    these.target = these.target,
                    pool.along = pool.along,
                    xintercept = xintercept,
                    yintercept = yintercept,
                    invert.selection = invert.selection,
                    h = h,
                    n = n,
                    label.fill = label.fill,
                    return.density = return.density,
                    ...)
    )
  }


#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
StatDens2dLabels <-
  ggplot2::ggproto(
    "StatDens2dLabels",
    ggplot2::Stat,
    compute_panel =
      function(data,
               scales,
               keep.fraction,
               keep.number,
               keep.sparse,
               keep.these,
               exclude.these,
               these.target,
               pool.along,
               xintercept,
               yintercept,
               invert.selection,
               h,
               n,
               label.fill,
               return.density) {

        force(data)
        if (!exists("label", data) && !is.null(label.fill)) {
          data[["label"]] <- rownames(data)
        }

        keep.these <- these2logical(these = keep.these,
                                    data = data,
                                    these.target = these.target)

        exclude.these <- these2logical(these = exclude.these,
                                       data = data,
                                       these.target = these.target)

        # discard redundant splits
        if (pool.along != "xy") {
          if (pool.along == "y" &&
              !(xintercept < max(data[["x"]]) &&
                xintercept > min(data[["x"]]))) {
            pool.along <- "xy"
          } else if (pool.along == "x" &&
                     !(yintercept < max(data[["y"]]) &&
                       yintercept > min(data[["y"]]))) {
            pool.along <- "xy"
          } else if (pool.along == "none") {
            if (!(xintercept < max(data[["x"]]) &&
                  xintercept > min(data[["x"]])) &&
                !(yintercept < max(data[["y"]]) &&
                  yintercept > min(data[["y"]]))) {
              pool.along <- "xy"
            } else if (!(xintercept < max(data[["x"]]) &&
                         xintercept > min(data[["x"]]))) {
              pool.along <- "x"
            } else if (!(yintercept < max(data[["y"]]) &&
                         yintercept > min(data[["y"]]))) {
              pool.along <- "y"
            }
          }
        }

        # make list of logical vectors
        if (pool.along == "y") {
          selectors <-list(q12 = data[["x"]] <= xintercept,
                           q34 = data[["x"]] > xintercept)
          if (length(keep.fraction) != 2L) {
            keep.fraction <- rep_len(keep.fraction, length.out = 2)
          }
          if (length(keep.number) != 2L) {
            if (length(keep.number) == 1L) {
              keep.number <- keep.number %/% 2
            }
            keep.number <- rep_len(keep.number, length.out = 2)
          }
          num.rows <- sapply(selectors, sum) # selectors are logical
        } else if (pool.along == "x") {
          selectors <-list(q23 = data[["y"]] <= yintercept,
                           q41 = data[["y"]] > yintercept)
          if (length(keep.fraction) != 2L) {
            keep.fraction <- rep_len(keep.fraction, length.out = 2)
          }
          if (length(keep.number) != 2L) {
            if (length(keep.number) == 1L) {
              keep.number <- keep.number %/% 2
            }
            keep.number <- rep_len(keep.number, length.out = 2)
          }
          num.rows <- sapply(selectors, sum) # selectors are logical
        } else if (pool.along == "none") {
          selectors <-list(q1 = data[["y"]] >= yintercept & data[["x"]] >= xintercept,
                           q2 = data[["y"]] < yintercept & data[["x"]] >= xintercept,
                           q3 = data[["y"]] < yintercept & data[["x"]] < xintercept,
                           q4 = data[["y"]] > yintercept & data[["x"]] < xintercept)
          if (length(keep.fraction) != 4L) {
            keep.fraction <- rep_len(keep.fraction, length.out = 4)
          }
          if (length(keep.number) != 4L) {
            if (length(keep.number) == 1L) {
              keep.number <- keep.number %/% 4
            }
            keep.number <- rep_len(keep.number, length.out = 4)
          }
          num.rows <- sapply(selectors, sum) # selectors are logical
        } else {
          keep.fraction <- keep.fraction[[1]] # can be a vector or a list
          keep.number <- keep.number[[1]]
          num.rows <- nrow(data)
          selectors <- list(all = rep.int(TRUE, times = num.rows))
        }

        # vectorized
        too.large.frac <- num.rows * keep.fraction > keep.number
        keep.fraction[too.large.frac] <-
          keep.number[too.large.frac] / num.rows[too.large.frac]

        # estimate 2D density
        # data with fewer than 2 distinct values preventgs density() estimation
        if (length(unique(data$x)) >= 2L &&
            length(unique(data$y)) >= 2L) {
          if (is.null(h)) {
            h <- c(MASS::bandwidth.nrd(data$x), MASS::bandwidth.nrd(data$y))
          }
          if (is.null(n)) {
            n <- trunc(sqrt(nrow(data))) * 8L
          }
          kk <-  MASS::kde2d(
            data[["x"]], data[["y"]], h = h, n = n,
            lims = c(scales$x$dimension(), scales$y$dimension()))
          dimnames(kk[["z"]]) <- list(kk[["x"]], kk[["y"]])

          # compute 2D density at each observation's coordinates
          kx <- cut(data$x, kk$x, labels = FALSE, include.lowest = TRUE)
          ky <- cut(data$y, kk$y, labels = FALSE, include.lowest = TRUE)
          kz <- sapply(seq_along(kx), function(i) kk$z[kx[i], ky[i]])
        } else {
          kz <- rep_len(1, nrow(data))
        }

        # we construct one logical vector by adding observations/label to be kept
        # we may have a list of 1, 2, or 4 logical vectors
        keep <- logical(nrow(data))
        for (i in seq_along(selectors)) {
          if (keep.fraction[i] == 1) {
            keep[ selectors[[i]] ] <- TRUE
          } else if (keep.fraction[i] != 0 && length(selectors[[i]]) >= 2L) {
            if (keep.sparse) {
              keep[ selectors[[i]] ] <-
                kz[ selectors[[i]] ] < stats::quantile(kz[ selectors[[i]] ],
                                                       keep.fraction[i],
                                                       names = FALSE,
                                                       type = 8)
            } else {
              keep[ selectors[[i]] ] <-
                kz[ selectors[[i]] ] >= stats::quantile(kz[ selectors[[i]] ],
                                                        1 - keep.fraction[i],
                                                        names = FALSE,
                                                        type = 8)
            }
          }
        }
        keep <- (keep | keep.these) & !exclude.these

        if (invert.selection){
          keep <- !keep
        }

        if (return.density) {
          data[["keep.obs"]] <- keep
          data[["density"]] <- kz
        }

        if (is.null(label.fill)) {
          data <- data[keep, ]
        } else if (is.function(label.fill)) {
          data[["label"]][!keep] <- label.fill(data[["label"]][!keep])
        } else if (is.na(label.fill)) {
          # NA_logical_, the default NA, cannot always be assigned to character
          label.fill <- NA_character_
          data[["label"]][!keep] <- label.fill
        } else if (is.character(label.fill)) {
          data[["label"]][!keep] <- label.fill
        } else if (is.logical(label.fill)) {
          if (label.fill) {
            data[["label"]][!keep] <- ""
          } # if FALSE data is not modified
        } else {
          stop("'label.fill' is : ", mode(label.fill),
               " instead of 'character' or 'function'.")
        }
        data
      },

    required_aes = c("x", "y")
  )


# Utils for stats that subset data

#' Convert keep.these argument into logical vector
#'
#' @param these character vector, integer vector, logical vector or
#'   function that takes the variable mapped to the \code{label} aesthetic as
#'   first argument and returns a character vector or a logical vector. These
#'   rows from \code{data} are selected irrespective of the local density.
#' @param data data.frame The plot layer's data set.
#' @param these.target character The name of the variable (or aesthetic) in
#'   which to select the target passed through \code{these}.
#'
#' @keywords internal
#'
these2logical <- function(these,
                          data,
                          these.target = "label") {
  if (length(these)) {
    if (is.character(these) || is.function(these)) {
      if (is.character(these.target) && any(these.target == "label") &&
          !exists("label", where = data, mode = "character", inherits = FALSE)) {
        data$label <- rownames(data)
      }
      if (is.character(these.target)) {
        orig.num.targets <- length(unique(these.target))
        these.target <- intersect(these.target, colnames(data))
        if (length(these.target) == 0L) {
          stop("Variables in 'these.target' not in 'data'")
        } else if (orig.num.targets > length(these.target)) {
          warning("Some variables in 'these.target' not in 'data'")
        }
      }
    }
    if (is.function(these)) {
      these <- these(data[ , these.target, drop = TRUE]) # any vector
    }
    if (is.character(these)) {
      stopifnot(is.character(data[[these.target[1]]]))
      these <- data[[these.target[1]]] %in% these # logical vector
    }
    if (is.numeric(these)) { # positional indices
      temp <- rep_len(FALSE, length.out = nrow(data))
      temp[these] <- TRUE
      these <- temp
    }
    if (is.logical(these)) { # logical indices, if short recycle
      if (length(these) >= 1L && length(these) < nrow(data)) {
        these <- rep_len(these, length.out = nrow(data))
      } else if (length(these) > nrow(data)) {
        stop("Logical vector 'keep.these' or 'exclude.these' longer than data")
      }
    }
    if (anyNA(these)) {
      warning("Discarding 'NA's in 'keep.these' or 'exclude-these'")
      these[is.na(these)] <- FALSE
    }
  } else { # replace NULL and vectors with length zero with FALSE
    these <- rep_len(FALSE, length.out = nrow(data))
  }
  these
}

context("stat_dens2d_filter")

library(ggplot2)
library(tibble)

make_data_tbl <- function(nrow = 100, rfun = rnorm, ...) {
  if (nrow %% 2) {
    nrow <- nrow + 1
  }

  set.seed(1001)

  tibble::tibble(
    x = rfun(nrow, ...),
    y = rfun(nrow, ...),
    group = rep(c("A", "B"), c(nrow / 2, nrow / 2))
  )
}

test_that("filter_params", {
  # Here I have included expect_error() and expect_no_error() tests
  # Many of the "no error" tests complement the vdiffr tests further down

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter(keep.fraction = NA)
  )

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter(keep.fraction = 5)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter(keep.fraction = 1)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter(keep.fraction = 0)
  )

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter(keep.fraction = -1)
  )

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter(keep.number = NA)
  )

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter(keep.number = -1)
  )

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(keep.fraction = NA)
  )

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(keep.fraction = 5)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(keep.fraction = 1)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(keep.fraction = 0)
  )

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(keep.fraction = -1)
  )

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(keep.number = NA)
  )

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(keep.number = -1)
  )

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(keep.fraction = rep(1/3, 5),
                           pool.along = "none")
  )

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(keep.number = rep(10, 5),
                           pool.along = "none")
  )

  testthat::expect_warning(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(keep.fraction = rep(1/3, 4))
  )

  testthat::expect_warning(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(keep.number = rep(10, 4))
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(keep.these = 1:3)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(keep.these = TRUE)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(keep.these = FALSE)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(keep.these = rep(c(TRUE, FALSE), 10L))
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(keep.these = function(x) {grepl("^a", x)})
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(keep.these = integer())
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(keep.these = double())
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(keep.these = character())
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(pool.along = "none")
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(pool.along = "none",
                           xintercept = Inf,
                           yintercept = Inf)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(pool.along = "none",
                           xintercept = -Inf,
                           yintercept = -Inf)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(pool.along = "none",
                           xintercept = -Inf,
                           yintercept = Inf)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(pool.along = "none",
                           xintercept = Inf,
                           yintercept = -Inf)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(pool.along = "none",
                           xintercept = Inf)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(pool.along = "none",
                           xintercept = -Inf)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(pool.along = "none",
                           yintercept = Inf)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(pool.along = "none",
                           yintercept = -Inf)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(pool.along = "none",
                           xintercept = -Inf,
                           yintercept = Inf)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(pool.along = "none",
                           xintercept = Inf,
                           yintercept = -Inf)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(pool.along = "none",
                           xintercept = Inf)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(pool.along = "none",
                           xintercept = -Inf)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(pool.along = "none",
                           yintercept = Inf)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(pool.along = "none",
                           yintercept = -Inf)
  )


  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(pool.along = "none",
                           xintercept = -Inf,
                           yintercept = Inf)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(pool.along = "none",
                           xintercept = Inf,
                           yintercept = -Inf)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(pool.along = "y",
                           xintercept = Inf)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(pool.along = "y",
                           xintercept = -Inf)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(pool.along = "x",
                           yintercept = Inf)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens2d_filter_g(pool.along = "x",
                           yintercept = -Inf)
  )

})

test_that("numbers_tb", {
  vdiffr::expect_doppelganger("stat_d2d_fltg_01",
                              ggplot(data = make_data_tbl(6), aes(x, y)) +
                                geom_point() +
                                stat_dens2d_filter_g(colour = "red")
  )
  vdiffr::expect_doppelganger("stat_d2d_flt_01",
                              ggplot(data = make_data_tbl(6), aes(x, y)) +
                                geom_point() +
                                stat_dens2d_filter(colour = "red")
  )
  vdiffr::expect_doppelganger("stat_d2d_flt_02",
                              ggplot(data = make_data_tbl(6), aes(x, y)) +
                                geom_point() +
                                stat_dens2d_filter(colour = "red",
                                                   keep.fraction = 1/2)
  )
  vdiffr::expect_doppelganger("stat_d2d_flt_03",
                              ggplot(data = make_data_tbl(20), aes(x, y)) +
                                geom_point() +
                                stat_dens2d_filter(colour = "red")
  )
  vdiffr::expect_doppelganger("stat_d2d_flt_04",
                              ggplot(data = make_data_tbl(100), aes(x, y)) +
                                geom_point() +
                                stat_dens2d_filter(colour = "red")
  )
  vdiffr::expect_doppelganger("stat_d2d_flt_05",
                              ggplot(data = make_data_tbl(500), aes(x, y)) +
                                geom_point() +
                                stat_dens2d_filter(colour = "red")
  )
  vdiffr::expect_doppelganger("stat_d2d_flt_06",
                              ggplot(data = make_data_tbl(2000), aes(x, y)) +
                                geom_point() +
                                stat_dens2d_filter(colour = "red")
  )
  vdiffr::expect_doppelganger("stat_d2d_flt_07",
                              ggplot(data = make_data_tbl(2000), aes(x, y)) +
                                geom_point() +
                                stat_dens2d_filter(colour = "red",
                                                   keep.fraction = 0.01)
  )
  vdiffr::expect_doppelganger("stat_d2d_flt_08",
                              ggplot(data = make_data_tbl(2000), aes(x, y)) +
                                geom_point() +
                                stat_dens2d_filter(colour = "red",
                                                   keep.sparse = FALSE)
  )
  vdiffr::expect_doppelganger("stat_d2d_flt_09",
                              ggplot(data = make_data_tbl(2000), aes(x, y)) +
                                geom_point() +
                                stat_dens2d_filter(colour = "red",
                                                   keep.sparse = FALSE)+
                                stat_dens2d_filter(colour = "blue")
  )
  vdiffr::expect_doppelganger("stat_d2d_flt_08s",
                              ggplot(data = make_data_tbl(2000), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red",
                                                   invert.selection = TRUE)
  )
  vdiffr::expect_doppelganger("stat_d2d_flt_09s",
                              ggplot(data = make_data_tbl(2000), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red",
                                                   invert.selection = TRUE)+
                                stat_dens1d_filter(colour = "blue")
  )
  vdiffr::expect_doppelganger("stat_d2d_flt_10",
                              ggplot(data = make_data_tbl(2000), aes(x, y)) +
                                geom_point() +
                                stat_dens2d_filter(colour = "red",
                                                   keep.fraction = 0.01,
                                                   keep.sparse = FALSE)
  )
  # vdiffr::expect_doppelganger("stat_d2d_flt_11",
  #                             ggplot(data = make_data_tbl(10000), aes(x, y)) +
  #                               geom_point() +
  #                               stat_dens2d_filter(colour = "red")
  # )
  # vdiffr::expect_doppelganger("stat_d2d_flt_12",
  #                             ggplot(data = make_data_tbl(10000), aes(x, y)) +
  #                               geom_point() +
  #                               stat_dens2d_filter(colour = "red", keep.fraction = 0.01)
  # )
  vdiffr::expect_doppelganger("stat_d2d_flt_13",
                              ggplot(data = make_data_tbl(1000, rfun = runif), aes(x, y)) +
                                geom_point() +
                                stat_dens2d_filter(colour = "red", keep.fraction = 0.1)
  )
  vdiffr::expect_doppelganger("stat_d2d_flt_14",
                              ggplot(data = make_data_tbl(1000, rfun = rgamma, shape = 2), aes(x, y)) +
                                geom_point() +
                                stat_dens2d_filter(colour = "red", keep.fraction = 0.1)
  )
  vdiffr::expect_doppelganger("stat_d2d_flt_15",
                              ggplot(data = make_data_tbl(1000, rfun = rgamma, shape = 6), aes(x, y)) +
                                geom_point() +
                                stat_dens2d_filter(colour = "red", keep.fraction = 0.1)
  )
  vdiffr::expect_doppelganger("stat_d2d_flt_16",
                              ggplot(data = make_data_tbl(1000, rfun = rbeta, shape1 = 3, shape2 = 12), aes(x, y)) +
                                geom_point() +
                                stat_dens2d_filter(colour = "red", keep.fraction = 0.1)
  )
  vdiffr::expect_doppelganger("stat_d2d_flt_17",
                              ggplot(data = make_data_tbl(1000, rfun = rbeta, shape1 = 3, shape2 = 12), aes(x, y)) +
                                geom_point() +
                                stat_dens2d_filter(colour = "red", keep.fraction = 0.1) +
                                scale_y_log10()
  )
  vdiffr::expect_doppelganger("stat_d2d_flt_18",
                              ggplot(data = make_data_tbl(1000, rfun = rbeta, shape1 = 3, shape2 = 12), aes(x, y)) +
                                geom_point() +
                                stat_dens2d_filter(colour = "red", keep.fraction = 0.1, return.density = TRUE)
  )
})

make_labs_tbl <- function(nrow = 100, rfun = rnorm, ...) {
  if (nrow %% 2) {
    nrow <- nrow + 1
  }

  random_string <-
    function(len = 6) {
      paste(sample(letters, len, replace = TRUE), collapse = "")
    }

  set.seed(1001)

  tibble::tibble(
    x = rfun(nrow, ...),
    y = rfun(nrow, ...),
    group = rep(c("A", "B"), c(nrow / 2, nrow / 2)),
    lab = replicate(nrow, { random_string() })
  )
}

test_that("labels_params", {

  testthat::expect_error(
    ggplot( data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens2d_labels(label.fill = rep("", 2))
  )

  testthat::expect_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens2d_labels(keep.fraction = NA)
  )

  testthat::expect_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens2d_labels(keep.fraction = 5)
  )

  testthat::expect_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens2d_labels(keep.fraction = -1)
  )

  testthat::expect_no_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens2d_labels(keep.fraction = 0)
  )

  testthat::expect_no_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens2d_labels(keep.fraction = 1)
  )

  testthat::expect_warning(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens2d_labels(keep.fraction = rep(0.1, 2))
  )

  testthat::expect_no_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens2d_labels(keep.fraction = rep(0.1, 2),
                         pool.along = "none")
  )

  testthat::expect_no_error(
    ggplot(data = make_labs_tbl(160), aes(x, y, label = lab)) +
      stat_dens2d_labels(keep.fraction = rep(0.1, 4),
                         pool.along = "none")
  )

  testthat::expect_error(
    ggplot(data = make_labs_tbl(160), aes(x, y, label = lab)) +
      stat_dens2d_labels(keep.fraction = rep(0.1, 6),
                         pool.along = "none")
  )

  testthat::expect_no_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens2d_labels(keep.fraction = rep(0.1, 2),
                         pool.along = "x")
  )

  testthat::expect_no_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens2d_labels(keep.fraction = rep(0.1, 2),
                         pool.along = "y")
  )

  testthat::expect_no_error(
    ggplot(data = make_labs_tbl(100), aes(x, y, label = lab)) +
      stat_dens2d_labels(keep.fraction = 0.1,
                         pool.along = "xy")
  )

  testthat::expect_warning(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens2d_labels(keep.fraction = rep(0.1, 5))
  )

  testthat::expect_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens2d_labels(keep.number = NA)
  )

  testthat::expect_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens2d_labels(keep.number = -1)
  )

  testthat::expect_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens2d_labels(label.fill = 123)
  )

  testthat::expect_no_error(
    ggplot(data = make_labs_tbl(20), aes(x, y)) +
      stat_dens2d_labels()
  )

  vdiffr::expect_doppelganger("stat_d2d_lbl_fill_x01",
                              ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red", label.fill = "")
  )

  vdiffr::expect_doppelganger("stat_d2d_lbl_fill_x02",
                              ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red", label.fill = NA)
  )

  vdiffr::expect_doppelganger("stat_d2d_lbl_fill_x03",
                              ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red", label.fill = "z")
  )

  vdiffr::expect_doppelganger("stat_d2d_lbl_fill_x04",
                              ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red", label.fill = toupper)
  )

  vdiffr::expect_doppelganger("stat_d2d_lbl_fill_x05",
                              ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(label.fill = FALSE)
  )

  vdiffr::expect_doppelganger("stat_d2d_lbl_keep_frac_x01",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red", keep.fraction = 1/3)
  )

  vdiffr::expect_doppelganger("stat_d2d_lbl_keep_frac_x02",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red",
                                                   keep.fraction = 1/3,
                                                   pool.along = "none")
  )

  vdiffr::expect_doppelganger("stat_d2d_lbl_keep_frac_x03",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red",
                                                   keep.fraction = 1/3,
                                                   pool.along = "none",
                                                   xintercept = -1)
  )

  vdiffr::expect_doppelganger("stat_d2d_lbl_keep_frac_x04",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red",
                                                   keep.fraction = 1/3,
                                                   pool.along = "none",
                                                   xintercept = 1)
  )

  vdiffr::expect_doppelganger("stat_d2d_lbl_keep_frac_x05",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red",
                                                   keep.fraction = 1/3,
                                                   pool.along = "none",
                                                   xintercept = -3)
  )

  vdiffr::expect_doppelganger("stat_d2d_lbl_keep_frac_x06",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red",
                                                   keep.fraction = 1/3,
                                                   pool.along = "none",
                                                   xintercept = 3)
  )

  vdiffr::expect_doppelganger("stat_d2d_lbl_keep_frac_x07",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red",
                                                   keep.fraction = 1,
                                                   keep.number = 6)
  )

  vdiffr::expect_doppelganger("stat_d2d_lbl_keep_frac_x08",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red",
                                                   keep.fraction = 1,
                                                   keep.number = 12,
                                                   pool.along = "none")
  )

  vdiffr::expect_doppelganger("stat_d2d_lbl_keep_frac_x09",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red",
                                                   keep.fraction = 1,
                                                   keep.number = c(1, 5),
                                                   pool.along = "none")
  )

  vdiffr::expect_doppelganger("stat_d2d_lbl_keep_frac_x10",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red",
                                                   keep.fraction = 1,
                                                   keep.number = c(5, 1),
                                                   pool.along = "none")
  )

  vdiffr::expect_doppelganger("stat_d2d_lbl_keep_frac_x11",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red",
                                                   keep.fraction = c(1/3, 0),
                                                   pool.along = "none")
  )

  vdiffr::expect_doppelganger("stat_d2d_lbl_keep_frac_x12",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red",
                                                   keep.fraction = c(0, 1/3),
                                                   pool.along = "none")
  )
})

test_that("labels_tb", {
  vdiffr::expect_doppelganger("stat_d2d_lbl_x01",
                              ggplot(data = make_labs_tbl(6), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red", keep.fraction = 0)
  )
  vdiffr::expect_doppelganger("stat_d2d_lbl_x02",
                              ggplot(data = make_labs_tbl(6), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red",
                                                   keep.fraction = 1/2)
  )
  vdiffr::expect_doppelganger("stat_d2d_lbl_x03",
                              ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red")
  )
  vdiffr::expect_doppelganger("stat_d2d_lbl_x04",
                              ggplot(data = make_labs_tbl(100), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red")
  )
  vdiffr::expect_doppelganger("stat_d2d_lbl_x05",
                              ggplot(data = make_labs_tbl(500), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red")
  )
  vdiffr::expect_doppelganger("stat_d2d_lbl_x06",
                              ggplot(data = make_labs_tbl(2000), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red")
  )
  vdiffr::expect_doppelganger("stat_d2d_lbl_x07",
                              ggplot(data = make_labs_tbl(2000), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red",
                                                   keep.fraction = 0.01)
  )
  vdiffr::expect_doppelganger("stat_d2d_lbl_x08",
                              ggplot(data = make_labs_tbl(2000), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red",
                                                   keep.sparse = FALSE)
  )
  vdiffr::expect_doppelganger("stat_d2d_lbl_x09",
                              ggplot(data = make_labs_tbl(2000), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red",
                                                   keep.sparse = FALSE) +
                                stat_dens2d_labels(colour = "blue")
  )
  vdiffr::expect_doppelganger("stat_d2d_lbl_x10",
                              ggplot(data = make_labs_tbl(2000), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red",
                                                   keep.fraction = 0.01,
                                                   keep.sparse = FALSE)
  )
  # vdiffr::expect_doppelganger("stat_d2d_lbl_x11",
  #                             ggplot(data = make_labs_tbl(10000), aes(x, y, label = lab)) +
  #                               geom_point() +
  #                               stat_dens2d_labels(colour = "red")
  # )
  # vdiffr::expect_doppelganger("stat_d2d_lbl_x12",
  #                             ggplot(data = make_labs_tbl(10000), aes(x, y, label = lab)) +
  #                               geom_point() +
  #                               stat_dens2d_labels(colour = "red",
  #                                                  keep.fraction = 0.01)
  # )
  vdiffr::expect_doppelganger("stat_d2d_lbl_x13",
                              ggplot(data = make_labs_tbl(1000, rfun = runif), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red", keep.fraction = 0.1)
  )
  vdiffr::expect_doppelganger("stat_d2d_lbl_x14",
                              ggplot(data = make_labs_tbl(1000, rfun = rgamma, shape = 2), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red", keep.fraction = 0.1)
  )
  vdiffr::expect_doppelganger("stat_d2d_lbl_x15",
                              ggplot(data = make_labs_tbl(1000, rfun = rgamma, shape = 6), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red", keep.fraction = 0.1)
  )
  vdiffr::expect_doppelganger("stat_d2d_lbl_x16",
                              ggplot(data = make_labs_tbl(1000, rfun = rbeta, shape1 = 3, shape2 = 12), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red", keep.fraction = 0.1)
  )
  vdiffr::expect_doppelganger("stat_d2d_lbl_x17",
                              ggplot(data = make_labs_tbl(1000, rfun = rbeta, shape1 = 3, shape2 = 12), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens2d_labels(colour = "red", keep.fraction = 0.1) +
                                scale_x_log10()
  )
})

test_that("these2logicalarguments with label", {
  df <- data.frame(x = 1:10, y = 1:10, label = letters[1:10])
  expect_equal(these2logical(FALSE, df), rep_len(FALSE, nrow(df)))
  expect_warning(these2logical(NA, df))
  expect_equal(these2logical(NULL, df), rep_len(FALSE, nrow(df)))
  expect_equal(these2logical(numeric(), df), rep_len(FALSE, nrow(df)))
  expect_error(these2logical(nrow(df) + 1, df))
  expect_equal(these2logical(logical(), df), rep_len(FALSE, nrow(df)))
  expect_equal(these2logical(3:4, df), c(F, F, T, T, F, F, F, F, F, F))
  expect_equal(these2logical(-c(1:2, 5:10), df), c(F, F, T, T, F, F, F, F, F, F))
  expect_equal(these2logical(c("c", "d"), df), c(F, F, T, T, F, F, F, F, F, F))
  expect_equal(these2logical(function(x) {x %in% c("c", "d")}, df), c(F, F, T, T, F, F, F, F, F, F))
  expect_equal(these2logical(c("3", "4"), df), rep_len(FALSE, nrow(df)))
  expect_equal(these2logical(function(x) {x %in% c("3", "4")}, df), rep_len(FALSE, nrow(df)))
  expect_equal(these2logical(function(x) {x %in% c(3, 4)}, df,
                                  these.target = "x"), c(F, F, T, T, F, F, F, F, F, F))
  expect_equal(these2logical(function(d) {d$x %in% c(3, 4)}, df,
                                  these.target = c("x", "y")), c(F, F, T, T, F, F, F, F, F, F))
  expect_equal(these2logical(function(d) {d$label %in% c("c", "d")}, df,
                                  these.target = c("x", "y", "label")), c(F, F, T, T, F, F, F, F, F, F))
  expect_equal(these2logical(function(d) {d$x %in% c(3, 4)}, df,
                                  these.target = TRUE), c(F, F, T, T, F, F, F, F, F, F))
  expect_error(these2logical(function(d) {d$x %in% c(3, 4)}, df,
                                  these.target = "zz"))
  expect_warning(these2logical(function(x) {x %in% c(3, 4)}, df,
                                  these.target = c("zz", "x")))
})

test_that("these2logicalarguments without label", {
  df <- data.frame(x = 1:10, y = 1:10)
  expect_equal(these2logical(FALSE, df), rep_len(FALSE, nrow(df)))
  expect_warning(these2logical(NA, df))
  expect_equal(these2logical(NULL, df), rep_len(FALSE, nrow(df)))
  expect_equal(these2logical(numeric(), df), rep_len(FALSE, nrow(df)))
  expect_error(these2logical(nrow(df) + 1, df))
  expect_equal(these2logical(logical(), df), rep_len(FALSE, nrow(df)))
  expect_equal(these2logical(3:4, df), c(F, F, T, T, F, F, F, F, F, F))
  expect_equal(these2logical(-c(1:2, 5:10), df), c(F, F, T, T, F, F, F, F, F, F))
  expect_equal(these2logical(c("c", "d"), df), rep_len(FALSE, nrow(df)))
  expect_equal(these2logical(function(x) {x %in% c("c", "d")}, df), rep_len(FALSE, nrow(df)))
  expect_equal(these2logical(c("3", "4"), df), c(F, F, T, T, F, F, F, F, F, F))
  expect_equal(these2logical(function(x) {x %in% c("3", "4")}, df), c(F, F, T, T, F, F, F, F, F, F))
  expect_equal(these2logical(function(x) {x %in% c(3, 4)}, df,
                                  these.target = "x"), c(F, F, T, T, F, F, F, F, F, F))
  expect_equal(these2logical(function(d) {d$x %in% c(3, 4)}, df,
                                  these.target = c("x", "y")), c(F, F, T, T, F, F, F, F, F, F))
  expect_equal(these2logical(function(d) {d$label %in% c("3", "4")}, df,
                                  these.target = c("x", "y", "label")), c(F, F, T, T, F, F, F, F, F, F))
  expect_equal(these2logical(function(d) {d$x %in% c(3, 4)}, df,
                                  these.target = TRUE), c(F, F, T, T, F, F, F, F, F, F))
})

#' Compute NPC coordinates
#'
#' Translate and/or compute NPC (Normalised Parent Coordinates) for use with
#' aesthetics \code{x} and \code{y}.
#'
#' @details Functions \code{compute_npcx} and \code{compute_npcy} convert
#'   character-encoded positions to npc units and shift positions to avoid
#'   overlaps when grouping is active. If numeric, they validate the npc values.
#'   Function \code{compute_npcx} does the translation either for both \code{x}
#'   and \code{y} aesthetics, but does not implement a shift for grpups.
#'   Functions \code{as_npcx()},  \code{as_npcy()} and \code{as_npc()} are
#'   wrappers on these functions that return the value as objects of class
#'   \code{"AsIs"} so that in 'ggplot2' >= 3.5.0 they can be used with any layer
#'   function.
#'
#'   These functions use NPC (Normalised Parent Coordinates) instead of data
#'   coordinates. They translate named positions into numeric values in [0..1]
#'   and they can also shift the position according to the group, e.g., for each
#'   increase in the group number displace the position inwards or outwards, by
#'   a user-supplied distance. They make it possible to set automatically set
#'   default positions for grouped text labels.
#'
#'   Out of bounds numeric values are constrained to [0..1]. Unrecognized
#'   character values are silently converted into \code{NA_integer_}.
#'
#' @note The \emph{as_npc()} functions make it easier the use of NPC coordinates
#'   with 'ggplot2' >= 3.5.0. The _compute_ functions are used by several layer
#'   functions in packages 'ggpp' and 'ggpmisc', are compatible with 'ggplot2'
#'   <= 3.4.4 and can be useful to developers of other 'ggplot2' extensions.
#'
#' @param x numeric or if character, one of "right", "left", "maximum",
#'   "minimum", "centre", "center" or "middle".
#' @param y numeric or if character, one of "top", "bottom", "maximum",
#'   "minimum", "centre", "center" or "middle".
#' @param a numeric or if character, one of "right", "left", "top", "maximum",
#'   "minimum", "bottom", "centre", "center" or "middle".
#' @param group integer vector, ggplot's group id. Used to shift coordinates to
#'   avoid overlaps.
#' @param h.step,v.step numeric [0..1] The step size for shifting coordinates
#'   in npc units. Usually << 1.
#' @param margin.npc numeric [0..1] The margin added towards the nearest
#'   plotting area edge when converting character coordinates into npc. Usually
#'   << 1.
#' @param each.len integer The number of steps per group.
#' @param ... named arguments passed to \code{compute_npcx()} or
#'   \code{compute_npcy()}.
#'
#' @return A numeric vector with values in the range [0..1] representing
#'   npc coordinates.
#'
#' @examples
#' compute_npcx("right")
#' compute_npcx(c("left", "right"))
#' compute_npcx(c("minimum", "maximum"))
#' compute_npcx(c("left", "right"), margin.npc = 0)
#' compute_npcy("bottom")
#' compute_npcy("bottom", group = 1L:3L)
#' compute_npcy("bottom", group = 1L:3L, v.step = 0.2)
#' compute_npcy("bottom", group = 2L)
#' compute_npcx(0.5)
#' compute_npcx(1)
#' compute_npcx(-2)
#'
#' as_npc("right")
#' class(as_npc("right"))
#' class(compute_npcx("right"))
#'
#' @export
#'
compute_npcx <- function(x,
                         group = 1L,
                         h.step = 0.1,
                         margin.npc = 0.05,
                         each.len = 1) {
  group <- abs(group)
  if (is.factor(x)) {
    x <- as.character(x)
  }
  if (is.character(x)) {
    # we must handle character vectors with length > 1
    map <- c(right = 1 - margin.npc,
             left = 0 + margin.npc,
             maximum = 1,
             minimum = 0,
             centre = 0.5,
             center = 0.5,
             middle = 0.5,
             NA_real_)
    x <- unname(map[x])
  }
  expanded.group <- integer()
  for (i in seq_along(group)) {
    temp <- seq(from = 1, by = 1, length.out = each.len) +
      (group[i] - 1) * each.len
    expanded.group <- c(expanded.group, temp)
  }
  if (any(expanded.group > 0L) && h.step != 0) {
    x <- x + (expanded.group - 1) * h.step * ifelse(x < 0.5, 1, -1)
  }
  x <- ifelse(x > 1, 1, x)
  x <- ifelse(x < 0, 0, x)
  x
}

#' @rdname compute_npcx
#'
#' @export
#'
compute_npcy <- function(y,
                         group = 1L,
                         v.step = 0.1,
                         margin.npc = 0.05,
                         each.len = 1) {
  group <- abs(group)
  if (is.factor(y)) {
    y <- as.character(y)
  }
  if (is.character(y)) {
    # we must handle character vectors with length > 1
    map <- c(top = 1 - margin.npc,
             bottom = 0 + margin.npc,
             maximum = 1,
             minimum = 0,
             centre = 0.5,
             center = 0.5,
             middle = 0.5,
             NA_real_)
    y <- unname(map[y])
  }
  expanded.group <- integer()
  for (i in seq_along(group)) {
    temp <- seq(from = 1, by = 1, length.out = each.len) +
      (group[i] - 1) * each.len
    expanded.group <- c(expanded.group, temp)
  }
  if (any(expanded.group > 1L) && v.step != 0) {
    y <- y + (expanded.group - 1L) * v.step * ifelse(y < 0.5, 1, -1)
  }
  y <- ifelse(y > 1, 1, y)
  y <- ifelse(y < 0, 0, y)
  y
}

#' @rdname compute_npcx
#'
#' @export
#'
as_npcx <- function(x, ...) {
  I(compute_npcx(x = x, ...))
}

#' @rdname compute_npcx
#'
#' @export
#'
as_npcy <- function(y, ...) {
  I(compute_npcy(y = y, ...))
}

#' @rdname compute_npcx
#'
#' @export
#'
compute_npc <- function(a,
                        margin.npc = 0.05) {
  if (is.factor(a)) {
    a <- as.character(a)
  }
  if (is.character(a)) {
    # we must handle character vectors with length > 1
    map <- c(top = 1 - margin.npc,
             bottom = 0 + margin.npc,
             right = 1 - margin.npc,
             left = 0 + margin.npc,
             maximum = 1,
             minimum = 0,
             centre = 0.5,
             center = 0.5,
             middle = 0.5,
             NA_real_)
    a <- unname(map[a])
  } else if (is.numeric(a)) {
    a <- ifelse(a > 1, 1, a)
    a <- ifelse(a < 0, 0, a)
    a
  }
  a
}

#' @rdname compute_npcx
#'
#' @export
#'
as_npc <- function(a,
                   margin.npc = 0.05) {
  I(compute_npc(a = a, margin.npc = margin.npc))
}

context("compute_npcy")

test_that("character input value returns correct numeric value", {
  expect_equal(compute_npcy("top"), 0.95)
  expect_equal(compute_npcy("bottom"), 0.05)
  expect_equal(compute_npcy("bottom", group = 1L:3L), c(0.05, 0.15, 0.25))
  expect_equal(compute_npcy("bottom", group = 2L), 0.15)
  expect_equal(compute_npcy("center"), 0.5)
  expect_equal(compute_npcy("middle"), 0.5)
  expect_equal(compute_npcy("left"), NA)
})

test_that("factor input values returns correct numeric value", {
  y <- factor(c("top", "bottom"))
  expect_equal(compute_npcy(y), c(0.95, 0.05))
})


test_that("numeric input values returns the same value if
          between 0 and 1", {
            y <- c(0.5, 1)
            expect_equal(compute_npcy(y), y)
          })

test_that("numeric input values
          returns 0 if less than 0,
          returns 1 if greater than 1", {
            y <- c(-0.5, 2)
            expect_equal(compute_npcy(y), c(0,1))
          })

test_that("Value is returned 'AsIs'", {
  expect_is(as_npcy("top"), "AsIs")
  expect_equal(as_npcy("top"), I(0.95))
  expect_equal(as_npcy("bottom"), I(0.05))
  expect_equal(as_npcy("middle"), I(0.5))
  expect_equal(as_npcx("center"), I(0.5))
  expect_equal(as_npcx("centre"), I(0.5))
})

context("compute_npc")

test_that("character input value returns correct numeric value", {
  expect_equal(compute_npc("right"), 0.95)
  expect_equal(compute_npc("left"), 0.05)
  expect_equal(compute_npc(c("left", "right")), c(0.05, 0.95))
  expect_equal(compute_npc("centre"), 0.5)
  expect_equal(compute_npc("center"), 0.5)
  expect_equal(compute_npc("middle"), 0.5)
  expect_equal(compute_npc("bottom"), 0.05)
  expect_equal(compute_npc("top"), 0.95)
})

test_that("factor input values returns correct numeric value", {
  x <- factor(c("right", "left"))
  expect_equal(compute_npc(x), c(0.95, 0.05))

})

test_that("numeric input value returns the same value if
          between 0 and 1", {
  x <- c(0.5, 1)
  expect_equal(compute_npc(x), x)
})

test_that("numeric input value
          returns 0 if less than 0,
          returns 1 if greater than 1", {
            x <- c(-0.5, 2)
            expect_equal(compute_npc(x), c(0,1))
          })

test_that("Value is returned 'AsIs'", {
  expect_is(as_npc("right"), "AsIs")
  expect_equal(as_npc("right"), I(0.95))
  expect_equal(as_npc("left"), I(0.05))
  expect_equal(as_npc("centre"), I(0.5))
  expect_equal(as_npc("center"), I(0.5))
})

#' Combined positions stack and nudge
#'
#' \code{position_stacknudge()} is useful when labelling plots such as stacked
#' bars, stacked columns, stacked lines, etc. In contrast to
#' \code{\link[ggplot2]{position_nudge}}, \code{position_stacknudge()} returns
#' in \code{data} both the original coordinates and the nudged coordinates.
#'
#' \code{position_fillnudge()} is useful when labelling plots such as filled
#' bars, filled columns, filled lines, etc. In contrast to
#' \code{\link[ggplot2]{position_nudge}}, \code{position_fillnudge()} returns
#' in \code{data} both the original coordinates and the nudged coordinates.
#'
#' The wrapper \code{position_nudge_keep()} has the same signature and
#' behaviour as \code{\link[ggplot2]{position_nudge}} nad provides an easier to
#' remember name when the need is only to have access to both the original and
#' nudged coordinates.
#'
#' These position functions are backwards compatible with
#' \code{\link[ggplot2]{position_nudge}} but extends it by adding support for
#' stacking and for geometries that make use of the original position to draw
#' connecting segments or arrows.
#'
#' The wrapper \code{position_stack_keep()} has the same signature and
#' behaviour as \code{\link[ggplot2]{position_stack}} and provides an easier to
#' remember name when the need is only to have access to both the original and
#' nudged coordinates.
#'
#' The wrapper \code{position_fill_keep()} has the same signature and
#' behaviour as \code{\link[ggplot2]{position_fill}} and provides an easier to
#' remember name when the need is only to have access to both the original and
#' nudged coordinates.
#'
#' The wrapper \code{position_stack_minmax()} has the same signature and
#' behaviour as \code{\link{position_stacknudge}} but stacks y, ymin and
#' ymax in parallel, making it possible to stack summaries with error bars,
#' works correctly with \code{geom_pointrange()}, \code{geom_linerange()} and
#' \code{geom_errorbar()}.
#'
#' @family position adjustments
#'
#' @param vjust Vertical adjustment for geoms that have a position (like points
#'   or lines), not a dimension (like bars or areas). Set to 0 to align with the
#'   bottom, 0.5 for the middle, and 1 (the default) for the top.
#' @param reverse If TRUE, will reverse the default stacking order. This is
#'   useful if you're rotating both the plot and legend.
#' @param x,y Amount of vertical and horizontal distance to move. A numeric
#'   vector of length 1, or of the same length as rows there are in \code{data},
#'   with nudge values in data rows order.
#' @param direction One of \code{"none"}, \code{"split"}, \code{"split.x"} or
#'   \code{"split.y"}. A value of \code{"none"} replicates the behavior of
#'   \code{\link[ggplot2]{position_nudge}}. At the moment \code{"split"} changes
#'   the sign of the nudge at zero, which is suitable for column plots with
#'   negative slices.
#' @param kept.origin One of \code{"original"}, \code{"stacked"} or
#'   \code{"none"}.
#'
#' @seealso \code{\link[ggplot2]{position_nudge}},
#'   \code{\link[ggplot2]{position_stack}},
#'   \code{\link[ggrepel]{position_nudge_repel}}.
#'
#' @return A \code{"Position"} object.
#'
#' @export
#'
#' @author Michał Krassowski, edited by Pedro J. Aphalo.
#'
#' @source \url{https://github.com/slowkow/ggrepel/issues/161}.
#'
#' @examples
#'
#' df <- data.frame(x1 = c("a", "a", "b", "b", "b"),
#'                  x2 = c(1, 2, 1, 3, -1),
#'                  grp = c("some long name", "other name", "some name",
#'                          "another name", "some long name"))
#'
#' # Add labels to a horizontal column plot (stacked by default)
#' ggplot(data = df, aes(x1, x2, group = grp)) +
#'   geom_col(aes(fill = grp), width=0.5) +
#'   geom_vline(xintercept = 0) +
#'   geom_text(
#'     aes(label = grp),
#'     position = position_stacknudge(vjust = 0.5, y = 0.3)) +
#'   theme(legend.position = "none")
#'
#' # Add labels to a vertical column plot (stacked by default)
#' ggplot(data = df, aes(x2, x1, group = grp)) +
#'   geom_col(aes(fill = grp), width=0.5) +
#'   geom_vline(xintercept = 0) +
#'   geom_text(
#'     aes(label = grp),
#'     position = position_stacknudge(vjust = 0.5, x = -0.3),
#'     angle = 90) +
#'   theme(legend.position = "none")
#'
#' # Add labels to a vertical column plot (stacked by default)
#' ggplot(data = subset(df, x1 >= 0), aes(x1, x2, group = grp)) +
#'   geom_col(aes(fill = grp), width=0.5, position = position_fill()) +
#'   geom_vline(xintercept = 0) +
#'   geom_text(
#'     aes(label = grp),
#'     position = position_fillnudge(vjust = 0.5, x = -0.3),
#'     angle = 90) +
#'   theme(legend.position = "none")
#'
#' # Add label at a fixed distance from the top of each column slice
#' ggplot(data = df, aes(x1, x2, group = grp)) +
#'   geom_col(aes(fill = grp), width=0.5) +
#'   geom_vline(xintercept = 0) +
#'   geom_text(
#'     aes(label = grp),
#'     position = position_stacknudge(vjust = 1, y = -0.2)) +
#'   theme(legend.position = "none")
#'
#' # Use geom_text_s(), geom_text_repel() or geom_label_repel() to link
#' # label to labelled segment or object with an arrow
#' ggplot(data = df, aes(x2, x1, group = grp)) +
#'   geom_col(aes(fill = grp), width=0.5) +
#'   geom_vline(xintercept = 0) +
#'   geom_text_s(
#'     aes(label = grp),
#'     position = position_stacknudge(vjust = 0.5, y = 0.35),
#'     vjust = "bottom") +
#'   theme(legend.position = "none")
#'
#' ggplot(birch_dw.df,
#'        aes(y = dry.weight * 1e-3, x = Density, fill = Part)) +
#'    stat_summary(geom = "col", fun = mean,
#'                 position = "stack", alpha = 0.7, width = 0.67) +
#'    stat_summary(geom = "linerange", fun.data = mean_cl_normal,
#'                 position = position_stack_minmax()) +
#'   labs(y = "Seedling dry mass (g)") +
#'   scale_fill_grey(start = 0.7, end = 0.3) +
#'   facet_wrap(facets = vars(Container))
#'

position_stacknudge <-
  function(vjust = 1,
           reverse = FALSE,
           x = 0,
           y = 0,
           direction = c("none", "split", "split.x", "split.y"),
           kept.origin = c("stacked", "original", "none")) {

    direction <- rlang::arg_match(direction)
    kept.origin <- rlang::arg_match(kept.origin)

    ggplot2::ggproto(NULL, PositionStackAndNudge,
                     x = x,
                     y = y,
                     .fun_x = switch(direction,
                                     none = function(x) {1},
                                     split = sign,
                                     split.y = function(x) {1},
                                     split.x = sign),
                     .fun_y = switch(direction,
                                     none = function(x) {1},
                                     split = sign,
                                     split.x = function(x) {1},
                                     split.y = sign),
                     kept.origin = kept.origin,
                     vjust = vjust,
                     reverse = reverse
    )
  }

#' @export
#' @rdname position_stacknudge
position_fillnudge <-
  function(vjust = 1,
           reverse = FALSE,
           x = 0,
           y = 0,
           direction = c("none", "split", "split.x", "split.y"),
           kept.origin = c("stacked", "original", "none")) {

  direction <- rlang::arg_match(direction)
  kept.origin <- rlang::arg_match(kept.origin)

  ggplot2::ggproto(NULL, PositionFillAndNudge,
                   x = x,
                   y = y,
                   .fun_x = switch(direction,
                                   none = function(x) {1},
                                   split = sign,
                                   split.y = function(x) {1},
                                   split.x = sign),
                   .fun_y = switch(direction,
                                   none = function(x) {1},
                                   split = sign,
                                   split.x = function(x) {1},
                                   split.y = sign),
                   kept.origin = kept.origin,
                   vjust = vjust,
                   reverse = reverse
  )
}


#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @noRd
PositionStackAndNudge <-
  ggplot2::ggproto("PositionStackAndNudge", ggplot2::PositionStack,
                   x = 0,
                   y = 0,

                   setup_params = function(self, data) {
                     c(
                       list(nudge_x = self$x,
                            nudge_y = self$y,
                            .fun_x = self$.fun_x,
                            .fun_y = self$.fun_y,
                            x.reorder = !is.null(self$x) && length(self$x) > 1 && length(self$x) < nrow(data),
                            y.reorder = !is.null(self$y) && length(self$y) > 1 && length(self$y) < nrow(data),
                            kept.origin = self$kept.origin),
                       ggplot2::ggproto_parent(ggplot2::PositionStack, self)$setup_params(data)
                     )
                   },

                   setup_data = function(self, data, params) {
                     data <- flip_data(data, params$flipped_aes)
                     if (is.null(params$var)) {
                       return(data)
                     }

                     data$ymax <- switch(params$var,
                                         y = data$y,
                                         ymax = as.numeric(ifelse(data$ymax == 0, data$ymin, data$ymax))
                     )

                     data <- remove_missing(
                       data,
                       vars = c("x", "xmin", "xmax", "y"),
                       name = "position_stack"
                     )
                     flip_data(data, params$flipped_aes)
                   },

                   compute_layer = function(self, data, params, layout) {
                     if (length(params$nudge_x) > nrow(data)) {
                       warning("Argument 'x' longer than data: some values dropped!")
                     }
                     if (length(params$nudge_y) > nrow(data)) {
                       warning("Argument 'y' longer than data: some values dropped!")
                     }
                     x_orig <- data$x
                     y_orig <- data$y

                     # operate on the stacked positions (updated in August 2020)
                     data = ggplot2::ggproto_parent(ggplot2::PositionStack, self)$compute_layer(data, params, layout)
                     x_stacked <- data$x
                     y_stacked <- data$y

                     if (params$x.reorder) {
                       params$nudge_x <- rep_len(params$nudge_x, nrow(data))[order(order(data$x))]
                     } else {
                       params$nudge_x <- rep_len(params$nudge_x, nrow(data))
                     }
                     if (params$y.reorder) {
                       params$nudge_y <- rep_len(params$nudge_y, nrow(data))[order(order(data$y))]
                     } else {
                       params$nudge_y <- rep_len(params$nudge_y, nrow(data))
                     }

                     # transform only the dimensions for which non-zero nudging is requested
                     if (any(params$nudge_x != 0)) {
                       if (any(params$nudge_y != 0)) {
                         data <- ggplot2::transform_position(data,
                                                             function(x) x + params$nudge_x * params$.fun_x(x),
                                                             function(y) y + params$nudge_y * params$.fun_y(y))
                       } else {
                         data <- ggplot2::transform_position(data,
                                                             function(x) x + params$nudge_x * params$.fun_x(x),
                                                             NULL)
                       }
                     } else if (any(params$nudge_y != 0)) {
                       data <- ggplot2::transform_position(data,
                                                           function(x) x,
                                                           function(y) y + params$nudge_y * params$.fun_y(y))
                     }

                     if (params$kept.origin == "stacked") {
                       data$x_orig <- x_stacked
                       data$y_orig <- y_stacked
                     } else if (params$kept.origin == "original") {
                       data$x_orig <- x_orig
                       data$y_orig <- y_orig
                     }

                     data
                   },

                   compute_panel = function(self, data, params, scales) {
                     ggplot2::ggproto_parent(PositionStack, self)$compute_panel(data, params, scales)
                   }
  )

#' @rdname position_stacknudge
#'
#' @export
#'
position_stack_keep <-
  function(vjust = 1,
           reverse = FALSE,
           kept.origin = "original") {
    position_stacknudge(vjust = vjust,
                        reverse = reverse,
                        x = 0,
                        y = 0,
                        direction = "none",
                        kept.origin = kept.origin)
  }

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
PositionFillAndNudge <-
  ggplot2::ggproto("PositionFillAndNudge", PositionStackAndNudge,
                   fill = TRUE
  )

#' @rdname position_stacknudge
#'
#' @export
#'
position_fill_keep <-
  function(vjust = 1,
           reverse = FALSE,
           kept.origin = "original") {
    position_fillnudge(vjust = vjust,
                       reverse = reverse,
                       x = 0,
                       y = 0,
                       direction = "none",
                       kept.origin = kept.origin)
  }

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @noRd
PositionStackMinMax <-
  ggplot2::ggproto("PositionStackMinMax", ggplot2::PositionStack,
                   x = 0,
                   y = 0,
                   var = "y",

                   setup_params = function(self, data) {
                     c(
                       list(nudge_x = self$x,
                            nudge_y = self$y,
                            .fun_x = self$.fun_x,
                            .fun_y = self$.fun_y,
                            x.reorder = !is.null(self$x) && length(self$x) > 1 && length(self$x) < nrow(data),
                            y.reorder = !is.null(self$y) && length(self$y) > 1 && length(self$y) < nrow(data),
                            kept.origin = self$kept.origin,
                            var = self$var,
                            fill = FALSE),
                       ggplot2::ggproto_parent(ggplot2::PositionStack, self)$setup_params(data)
                     )
                   },

                   setup_data = function(self, data, params) {
                     data <- flip_data(data, params$flipped_aes)
                     if (!(exists("ymax", data) && exists("ymin", data) && exists("y", data))) {
                       stop("position_stack_minmax() requires y, ymin and ymax mappings in data")
                     }
                     flip_data(data, params$flipped_aes)
                   },

                   compute_layer = function(self, data, params, layout) {

                     if (length(params$nudge_x) > nrow(data)) {
                       warning("Argument 'x' longer than data: some values dropped!")
                     }
                     if (length(params$nudge_y) > nrow(data)) {
                       warning("Argument 'y' longer than data: some values dropped!")
                     }
                     x_orig <- data$x
                     if (exists("xmin", data)) {
                       xmin_delta <-  data$xmin - data$x
                     } else {
                       xmin_delta <-  0
                     }
                     if (exists("xmax", data)) {
                       xmax_delta <-  data$xmax - data$x
                     } else {
                       xmax_delta <-  0
                     }

                     y_orig <- data$y
                     if (exists("ymin", data)) {
                       ymin_delta <-  data$ymin - data$y
                     } else {
                       ymin_delta <-  0
                     }
                     if (exists("ymax", data)) {
                       ymax_delta <-  data$ymax - data$y
                     } else {
                       ymax_delta <-  0
                     }

                     # operate on the stacked positions (updated in August 2020)
                     data = ggplot2::ggproto_parent(ggplot2::PositionStack, self)$compute_layer(data, params, layout)
                     x_stacked <- data$x
                     y_stacked <- data$y
                     if (exists("xmin", data)) data$xmin <- data$x + xmin_delta
                     if (exists("xmax", data)) data$xmax <- data$x + xmax_delta
                     if (exists("ymin", data)) data$ymin <- data$y + ymin_delta
                     if (exists("ymax", data)) data$ymax <- data$y + ymax_delta

                     if (params$x.reorder) {
                       params$nudge_x <- rep_len(params$nudge_x, nrow(data))[order(order(data$x))]
                     } else {
                       params$nudge_x <- rep_len(params$nudge_x, nrow(data))
                     }
                     if (params$y.reorder) {
                       params$nudge_y <- rep_len(params$nudge_y, nrow(data))[order(order(data$y))]
                     } else {
                       params$nudge_y <- rep_len(params$nudge_y, nrow(data))
                     }

                     # transform only the dimensions for which non-zero nudging is requested
                     if (any(params$nudge_x != 0)) {
                       if (any(params$nudge_y != 0)) {
                         data <- ggplot2::transform_position(data,
                                                             function(x) x + params$nudge_x * params$.fun_x(x),
                                                             function(y) y + params$nudge_y * params$.fun_y(y))
                       } else {
                         data <- ggplot2::transform_position(data,
                                                             function(x) x + params$nudge_x * params$.fun_x(x),
                                                             NULL)
                       }
                     } else if (any(params$nudge_y != 0)) {
                       data <- ggplot2::transform_position(data,
                                                           function(x) x,
                                                           function(y) y + params$nudge_y * params$.fun_y(y))
                     }

                     if (params$kept.origin == "stacked") {
                       data$x_orig <- x_stacked
                       data$y_orig <- y_stacked
                     } else if (params$kept.origin == "original") {
                       data$x_orig <- x_orig
                       data$y_orig <- y_orig
                     }

                     data
                   },

                   compute_panel = function(self, data, params, scales) {
                     ggplot2::ggproto_parent(PositionStack, self)$compute_panel(data, params, scales)
                   }
  )

#' @rdname position_stacknudge
#'
#' @export
#'
position_stack_minmax <-
  function(vjust = 1,
           reverse = FALSE,
           x = 0,
           y = 0,
           direction = c("none", "split", "split.x", "split.y"),
           kept.origin = c("stacked", "original", "none")) {

    direction <- rlang::arg_match(direction)
    kept.origin <- rlang::arg_match(kept.origin)

    ggplot2::ggproto(NULL, PositionStackMinMax,
                     x = x,
                     y = y,
                     .fun_x = switch(direction,
                                     none = function(x) {1},
                                     split = sign,
                                     split.y = function(x) {1},
                                     split.x = sign),
                     .fun_y = switch(direction,
                                     none = function(x) {1},
                                     split = sign,
                                     split.x = function(x) {1},
                                     split.y = sign),
                     kept.origin = kept.origin,
                     vjust = vjust,
                     reverse = reverse
    )
  }

context("position_fill_keep")

test_that("test vjust, reverse, x, y and kept.origin default arguments", {
  position <- position_fill_keep()
  expect_identical(position$vjust, 1)
  expect_identical(position$reverse, FALSE)
  expect_identical(position$x, 0)
  expect_identical(position$y, 0)
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
  expect_identical(position$kept.origin, "original")
  expect_identical(position$fill, TRUE)
  skip_if(condition = utils::packageVersion("ggplot2") < "4.0.0",
          message = "'ggplot2' < 4.0.0")
  expect_false(position$reverse)
  expect_identical(position$required_aes, character(0))
  expect_type(position$aesthetics, "closure")
  expect_type(position$use_defaults, "closure")
})

context("position_fillnudge")

test_that("incorrect kept.origin used", {
  expect_error(
    position_fillnudge(kept.origin = "") #,
#    "`kept.origin` must be one of \"stacked\", \"original\", or \"none\", not \"\"."
# rlang arg_match() adds to the error message characters that break expect_error()
  )
})

test_that("correct kept.origin used", {
  position <- position_fillnudge(kept.origin = "none")
  expect_no_error(position)
})

test_that("correct kept.origin used", {
  position <- position_fillnudge(kept.origin = "original")
  expect_no_error(position)
})

test_that("incorrect direction used", {
  expect_error(
    position_stack_minmax(direction = "") #,
    #    "`kept.origin` must be one of \"stacked\", \"original\", or \"none\", not \"\"."
    # rlang arg_match() adds to the error message characters that break expect_error()
  )
})

test_that("correct direction used", {
  position <- position_stack_minmax(direction = "none")
  expect_no_error(position)
})

test_that("correct direction used", {
  position <- position_stack_minmax(direction = "split")
  expect_no_error(position)
})

test_that("correct direction used", {
  position <- position_stack_minmax(direction = "split.x")
  expect_no_error(position)
})

test_that("correct direction used", {
  position <- position_stack_minmax(direction = "split.y")
  expect_no_error(position)
})

test_that("correct reverse, vjust and x assigned", {
  position <- position_fillnudge(kept.origin = "none", vjust = 0.5, x = -0.3)
  expect_identical(position$kept.origin, "none")
  expect_false(position$reverse)
  expect_identical(position$vjust, 0.5)
  expect_identical(position$y, 0)
  expect_identical(position$x, -0.3)
  expect_identical(position$fill, TRUE)
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
  skip_if(condition = utils::packageVersion("ggplot2") < "4.0.0",
          message = "'ggplot2' < 4.0.0")
  expect_false(position$reverse)
  expect_identical(position$required_aes, character(0))
  expect_type(position$aesthetics, "closure")
  expect_type(position$use_defaults, "closure")
})

test_that("warn on x or y longer than data", {
  expect_no_warning(
  ggplot(mtcars, aes(x = am, y = mpg, label = mpg)) +
    geom_text_s(
      position = position_stacknudge(x = 0.05, y = rep(5, nrow(mtcars))),
      size = 2.5
    )
)

expect_no_warning(
  ggplot(mtcars, aes(x = am, y = mpg, label = mpg)) +
    geom_text_s(
      position = position_stacknudge(y = 0, x = rep(0.05, nrow(mtcars))),
      size = 2.5
    )
)
})

context("position_stack_minmax")

test_that("incorrect kept.origin used", {
  expect_error(
    position_stack_minmax(kept.origin = "") #,
    #    "`kept.origin` must be one of \"stacked\", \"original\", or \"none\", not \"\"."
    # rlang arg_match() adds to the error message characters that break expect_error()
  )
})

test_that("correct kept.origin used", {
  expect_no_error(position <- position_stack_minmax(kept.origin = "none"))
  expect_identical(position$kept.origin, "none")
})

test_that("correct kept.origin used", {
  expect_no_error(position <- position_stack_minmax(kept.origin = "original"))
  expect_identical(position$kept.origin, "original")
})

test_that("incorrect direction used", {
  expect_error(
    position_stack_minmax(direction = "") #,
    #    "`kept.origin` must be one of \"stacked\", \"original\", or \"none\", not \"\"."
    # rlang arg_match() adds to the error message characters that break expect_error()
  )
})

test_that("correct direction used 'none'", {
  expect_no_error(position_stack_minmax(direction = "none"))
})

test_that("correct direction used 'split'", {
  expect_no_error(position_stack_minmax(direction = "split"))
})

test_that("correct direction used 'split.x'", {
  expect_no_error(position_stack_minmax(direction = "split.x"))
})

test_that("correct direction used 'split.y'", {
  expect_no_error(position_stack_minmax(direction = "split.y"))
})

test_that("test vjust, reverse, x, y and kept.origin arguments", {
  position <- position_stack_minmax()
  expect_identical(position$vjust, 1)
  expect_identical(position$reverse, FALSE)
  expect_identical(position$x, 0)
  expect_identical(position$y, 0)
  expect_identical(position$kept.origin, "stacked")
  expect_identical(position$var, "y")
  expect_identical(position$fill, FALSE)
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
  skip_if(condition = utils::packageVersion("ggplot2") < "4.0.0",
          message = "'ggplot2' < 4.0.0")
  expect_false(position$reverse)
  expect_identical(position$required_aes, character(0))
  expect_type(position$aesthetics, "closure")
  expect_type(position$use_defaults, "closure")
})

context("positions")

test_that("stacknudge", {
  df <- data.frame(x1 = c(1, 2, 1, 3, -1),
                   x2 = c("a", "a", "b", "b", "b"),
                   grp = c("some long name", "other name", "some name",
                           "another name", "some long name"))

  p <-
    ggplot(data = df, aes(x1, x2, group = grp)) +
    geom_col(aes(fill = grp), width = 0.5) +
    theme(legend.position = "none")

  vdiffr::expect_doppelganger("stack_nudge1",
                              p + geom_text(
                                aes(label = grp),
                                position = position_stacknudge(vjust = 0.5, y = 0.3))
  )
  vdiffr::expect_doppelganger("stack_nudge2",
                              p + geom_text_s(
                                aes(label = grp),
                                vjust = 0,
                                position = position_stacknudge(vjust = 0.5, y = 0.3))
  )
  vdiffr::expect_doppelganger("stack_nudge3",
                              p + geom_text(
                                aes(label = grp),
                                hjust = 0,
                                position = position_stacknudge(vjust = 0, x = 0.1))
  )
  vdiffr::expect_doppelganger("stack_nudge4",
                              p + geom_text(
                                aes(label = grp),
                                hjust = 0, angle = 90,
                                position = position_stacknudge(vjust = 0, x = 0.1, y = -0.21))
  )
  vdiffr::expect_doppelganger("stack_nudge5",
                              p + geom_text_s(
                                aes(label = grp),
                                vjust = 0.5,
                                box.padding = 0.99,
                                position = position_stacknudge(vjust = 0.5, y = c(0.3, -0.3)))
  )
  vdiffr::expect_doppelganger("stack_nudge6",
                              p + geom_text_s(
                                aes(label = grp),
                                vjust = 0.5,
                                box.padding = 0.99,
                                position = position_stacknudge(vjust = 0.5,
                                                               y = c(rep(0.3, 2), rep(-0.3, 3))))
  )
  vdiffr::expect_doppelganger("stack_nudge7",
                              ggplot(data = df, aes(x1, x2, group = grp)) +
                                geom_col(aes(fill = grp),
                                         width = 0.5,
                                         position = position_stack(reverse = TRUE)) +
                                theme(legend.position = "none") +
                                geom_text(
                                  aes(label = grp),
                                  position = position_stacknudge(vjust = 0.5,
                                                                 y = 0.3,
                                                                 reverse = TRUE))
  )
})

test_that("fillnudge", {
  df <- data.frame(x1 = c(1, 2, 1, 3, -1),
                   x2 = c("a", "a", "b", "b", "b"),
                   grp = c("some long name", "other name", "some name",
                           "another name", "some long name"))

  p <-
    ggplot(data = df, aes(x1, x2, group = grp)) +
    geom_col(aes(fill = grp), width = 0.5, position = "fill") +
    theme(legend.position = "none")

  vdiffr::expect_doppelganger("fill_nudge1",
                              p + geom_text(aes(label = grp),
                                            position = position_fillnudge(vjust = 0.5, y = 0.3))
  )
  vdiffr::expect_doppelganger("fill_nudge2",
                              p + geom_text_s(aes(label = grp),
                                              vjust = 0,
                                              position = position_fillnudge(vjust = 0.5, y = 0.3))
  )
  vdiffr::expect_doppelganger("fill_nudge3",
                              p + geom_text(aes(label = grp),
                                            angle = 90,
                                            position = position_fillnudge(vjust = 0, x = 0.1))
  )
  vdiffr::expect_doppelganger("fill_nudge4",
                              p + geom_text(aes(label = grp),
                                            hjust = 0, angle = 90,
                                            position = position_fillnudge(vjust = 0, x = 0.1, y = -0.21))
  )
  vdiffr::expect_doppelganger("fill_nudge5",
                              p + geom_text_s(aes(label = grp),
                                              vjust = 0.5,
                                              box.padding = 0.99,
                                              position = position_fillnudge(vjust = 0.5, y = c(0.3, -0.3)))
  )
  vdiffr::expect_doppelganger("fill_nudge6",
                              p + geom_text_s(aes(label = grp),
                                              vjust = 0.5,
                                              box.padding = 0.99,
                                              position = position_fillnudge(vjust = 0.5,
                                                                             y = c(rep(0.3, 2), rep(-0.3, 3))))
  )
  vdiffr::expect_doppelganger("fill_nudge7",
                              ggplot(data = df, aes(x1, x2, group = grp)) +
                                geom_col(aes(fill = grp),
                                         width = 0.5,
                                         position = position_fill(reverse = TRUE)) +
                                theme(legend.position = "none") +
                                geom_text(aes(label = grp),
                                          position = position_fillnudge(vjust = 0.5,
                                                                        y = 0.3,
                                                                        reverse = TRUE))
  )
})

test_that("dodgenudge", {
  df <- data.frame(x1 = c(1, 2, 1, 3, -1),
                   x2 = c("a", "a", "b", "b", "b"),
                   grp = c("some long name", "other name", "some name",
                           "another name", "some long name"))

  p <-
    ggplot(data = df, aes(x2, x1, group = grp)) +
    geom_col(aes(fill = grp),
             width=0.5,
             position = position_dodge(width = 0.5)) +
    theme(legend.position = "none")

  vdiffr::expect_doppelganger("dodge_nudge1",
                              p + geom_text(
                                aes(label = grp),
                                position = position_dodgenudge(width = 0.5, y = 0.1))
  )
  vdiffr::expect_doppelganger("dodge_nudge2",
                              p + geom_text(aes(label = grp),
                                            position = position_dodgenudge(width = 0.5, y = 0.1,
                                                                           direction = "split"))
  )
  vdiffr::expect_doppelganger("dodge_nudge3",
                              p + geom_text_s(aes(label = grp),
                                              hjust = 0,
                                              position = position_dodgenudge(width = 0.5, y = 0.2,
                                                                             direction = "split"))
  )
  vdiffr::expect_doppelganger("dodge_nudge4",
                              p + geom_text(
                                aes(label = grp),
                                hjust = "center",
                                position = position_dodgenudge(width = 0.5, y = -0.075,
                                                               direction = "split"))
  )
  # assumes the default order of factor levels
  vdiffr::expect_doppelganger("dodge_nudge5",
                              p + geom_text(
                                aes(label = grp),
                                position = position_dodgenudge(width = 0.5,
                                                               y = c(rep(0.1, 4), -0.1)))
  )
  vdiffr::expect_doppelganger("dodge_nudge6",
                              p + geom_text(aes(label = grp),
                                            position = position_dodgenudge(width = 0.5,
                                                                           y = 0.1))
  )
  skip_if(utils::packageVersion("ggplot2") < "4.0.0", "'ggplot2' < 4.0.0")
  vdiffr::expect_doppelganger("dodge_nudge7",
                              ggplot(data = df, aes(x2, x1, group = grp)) +
                                geom_col(
                                  aes(fill = grp),
                                  width=0.5,
                                  position = position_dodge(width = 0.5,
                                                            reverse = TRUE)) +
                                theme(legend.position = "none") +
                                geom_text(
                                  aes(label = grp),
                                  position = position_dodgenudge(width = 0.5,
                                                                 y = 0.1,
                                                                 reverse = TRUE))
  )
})

test_that("nudge_keep", {
  df <- data.frame(x1 = c(1, 2, 1, 3, -1),
                   x2 = c("a", "b", "c", "d", "e"),
                   grp = c("some long name", "other name", "some name",
                           "another name", "some long name"))

  p <-
    ggplot(data = df, aes(x2, x1, group = grp)) +
    geom_col(aes(fill = grp)) +
    theme(legend.position = "none")

  vdiffr::expect_doppelganger("nudge_keep1",
                              p + geom_text_s(
                                aes(label = grp),
                                vjust = 0, hjust = 0.5,
                                position = position_nudge_keep(y = 0.12))
  )
  vdiffr::expect_doppelganger("nudge_keep2",
                              p + geom_text_s(
                                aes(label = grp),
                                vjust = 0, hjust = "left",
                                position = position_nudge_keep(y = 0.12))
  )
  vdiffr::expect_doppelganger("nudge_keep3",
                              p + geom_text_s(
                                aes(label = grp),
                                vjust = 0, hjust = "inward",
                                position = position_nudge_keep(y = 0.12))
  )
  vdiffr::expect_doppelganger("nudge_keep4",
                              p + geom_text_s(
                                aes(label = grp),
                                position = position_nudge_keep(y = 0.12))
  )
  vdiffr::expect_doppelganger("nudge_keep5",
                              p + geom_text_s(
                                aes(label = grp),
                                position = position_nudge_keep(y = c(rep(0.12, 4), -0.12)))
  )
  vdiffr::expect_doppelganger("nudge_keep6",
                              p + geom_text_s(
                                aes(label = grp),
                                position = position_nudge_keep(y = c(rep(0.12, 4), -0.12),
                                                               obey_grouping = TRUE))
  )
  vdiffr::expect_doppelganger("nudge_keep7",
                              p + geom_text_s(
                                aes(label = grp),
                                position = position_nudge_keep(y = c(0.12, -0.12)))
  )
})

test_that("nudge_keep works with duration", {
  df <- data.frame(x1 = lubridate::as.duration(c(1, 2, 1, 3, -1)),
                   x2 = c("a", "b", "c", "d", "e"),
                   grp = c("some long name", "other name", "some name",
                           "another name", "some long name"))

  p <-
    ggplot(data = df, aes(x2, x1, group = grp)) +
    geom_col(aes(fill = grp)) +
    theme(legend.position = "none")

  vdiffr::expect_doppelganger("nudge_keep_duration1",
                              p + geom_text_s(
                                aes(label = grp),
                                vjust = 0, hjust = 0.5,
                                position = position_nudge_keep(y = 0.12))
  )
  vdiffr::expect_doppelganger("nudge_keep_duration2",
                              p + geom_text_s(
                                aes(label = grp),
                                vjust = 0, hjust = "left",
                                position = position_nudge_keep(y = 0.12))
  )
})

test_that("nudge_center", {
  df <- data.frame(
    x = c(1,3,2,5,4,2.5),
    y = c("abc","cd","d","c","bcd","a")
  )

  vdiffr::expect_doppelganger("nudge_center0",
                              ggplot(df, aes(x, y, label = y)) +
                                geom_point() +
                                geom_text(
                                  position = position_nudge_center(center_x = 3, center_y = 4,
                                                                   x = 0.1, y = 0.2)
                                )
  )
  vdiffr::expect_doppelganger("nudge_center1",
                              ggplot(df, aes(x, y, label = y)) +
                                geom_point() +
                                geom_text(hjust = 0, vjust = 0,
                                          position = position_nudge_center(x = 0.05,
                                                                           y = 0.07)
                                )
  )
  vdiffr::expect_doppelganger("nudge_center1a",
                              ggplot(df, aes(x, y, label = y)) +
                                geom_point() +
                                geom_text(hjust = 0, vjust = 0,
                                          position = position_nudge_center(x = 0.05,
                                                                           y = 0.07,
                                                                           direction = "none")
                                )
  )
  vdiffr::expect_doppelganger("nudge_center1b",
                              ggplot(df, aes(x, y, label = y)) +
                                geom_point() +
                                geom_text(hjust = 0, vjust = 0,
                                          position = position_nudge_center(x = c(0.05, -0.05),
                                                                           y = 0.07)
                                )
  )
  vdiffr::expect_doppelganger("nudge_center1c",
                              ggplot(df, aes(x, y, label = y)) +
                                geom_point() +
                                geom_text(hjust = 0, vjust = 0,
                                          position = position_nudge_center(x = 0.05,
                                                                           y = c(0.1, -0.1))
                                )
  )
  vdiffr::expect_doppelganger("nudge_center2",
                              ggplot(df, aes(x, y)) +
                                geom_point() +
                                geom_text(aes(label = y),
                                          hjust = "outward", vjust = "outward",
                                          position = position_nudge_center(x = 0.05,
                                                                           y = 0.07,
                                                                           direction = "split"))
  )
  vdiffr::expect_doppelganger("nudge_center3",
                              ggplot(df, aes(x, y)) +
                                geom_point() +
                                geom_text(aes(label = y),
                                          hjust = "outward", vjust = "outward",
                                          position = position_nudge_center(x = 0.05,
                                                                           y = 0.07,
                                                                           direction = "radial"))
  )
  vdiffr::expect_doppelganger("nudge_center4",
                              ggplot(df, aes(x, y)) +
                                geom_point() +
                                geom_text(aes(label = y),
                                          vjust = "outward", hjust = "outward",
                                          position = position_nudge_center(x = 0.06,
                                                                           y = 0.08,
                                                                           center_y = 2,
                                                                           center_x = 1.5,
                                                                           direction = "split"))
  )
  vdiffr::expect_doppelganger("nudge_center5",
                              ggplot(df, aes(x, y)) +
                                geom_point() +
                                geom_text(aes(label = y),
                                          vjust = "outward", hjust = "outward",
                                          position = position_nudge_center(x = 0.06,
                                                                           y = 0.08,
                                                                           center_y = 2,
                                                                           center_x = 1.5,
                                                                           direction = "radial"))
  )
  vdiffr::expect_doppelganger("nudge_center6",
                              ggplot(df, aes(x, y)) +
                                geom_point() +
                                geom_text(aes(label = y),
                                          vjust = "outward", hjust = "outward",
                                          position = position_nudge_center(x = 0.06,
                                                                           y = 0.08,
                                                                           center_y = 2))
  )
  vdiffr::expect_doppelganger("nudge_center7",
                              ggplot(df, aes(x, y)) +
                                geom_point() +
                                geom_text(aes(label = y),
                                          vjust = "outward", hjust = "outward",
                                          position = position_nudge_center(x = 0.06,
                                                                           y = 0.08,
                                                                           center_x = 1.5))
                                )
})

test_that("nudge_center works with time", {
  df <- data.frame(
    x = lubridate::dmy("01-01-1070") + lubridate::hours(c(1,3,2,5,4,6)),
    y = c("abc","cd","d","c","bcd","a")
  )

  vdiffr::expect_doppelganger("nudge_center_time0",
                              ggplot(df, aes(x, y, label = y)) +
                                geom_point() +
                                geom_text(
                                  position = position_nudge_center(center_y = 4,
                                                                   x = 600, # numeric = seconds
                                                                   y = 0.2)
                                )
  )
  vdiffr::expect_doppelganger("nudge_center_time1",
                              ggplot(df, aes(x, y, label = y)) +
                                geom_point() +
                                geom_text(hjust = 0, vjust = 0,
                                          position = position_nudge_center(x = 300, # numeric = seconds
                                                                           y = 0.07)
                                )
  )
})

test_that("nudge_center works with duration", {
  df <- data.frame(
    x = lubridate::as.duration(lubridate::hours(c(1,3,2,5,4,6))),
    y = c("abc","cd","d","c","bcd","a")
  )

  vdiffr::expect_doppelganger("nudge_center_duration0",
                              ggplot(df, aes(x, y, label = y)) +
                                geom_point() +
                                geom_text(
                                  position = position_nudge_center(x = 300, # numeric = seconds
                                                                   y = 0.2)
                                ) # +
#                               scale_x_time() # depends on package 'hms'
  )
  vdiffr::expect_doppelganger("nudge_center_duration1",
                              ggplot(df, aes(x, y, label = y)) +
                                geom_point() +
                                geom_text(hjust = 0, vjust = 0,
                                          position = position_nudge_center(x = 300, # numeric = seconds
                                                                           y = 0.07)
                                ) # +
#                               scale_x_time() # depends on package 'hms'
  )
})

test_that("nudge_line", {
  set.seed(16532)
  df <- data.frame(
    x = -10:10,
    y = (-10:10)^2,
    yy = (-10:10)^2 + rnorm(21, 0, 4),
    yyy = (-10:10) + rnorm(21, 0, 4),
    l = letters[1:21]
  )

  expect_error(ggplot(df, aes(x, y, label = l)) +
                  geom_text(position = position_nudge_line(kept.origin = "")))
  expect_error(ggplot(df, aes(x, y, label = l)) +
                 geom_text(position = position_nudge_line(kept.origin = "zz")))

  { # output under linux does not match snapshots created under Windows
    skip_on_os(c("mac", "linux", "solaris"))
    vdiffr::expect_doppelganger("nudge_line1",
                                ggplot(df, aes(x, y, label = l)) +
                                  geom_line() +
                                  geom_point() +
                                  geom_text(position = position_nudge_line())
    )
    vdiffr::expect_doppelganger("nudge_line2",
                                ggplot(df, aes(x, y, label = l)) +
                                  geom_line() +
                                  geom_point() +
                                  geom_text(position = position_nudge_line(xy_relative = -0.03))
    )
  }

  vdiffr::expect_doppelganger("nudge_line3",
                              ggplot(df, aes(x, x * 2 + 5, label = l)) +
                                geom_abline(intercept = 5, slope = 2) +
                                geom_point() +
                                geom_text(position = position_nudge_line(abline = c(5, 2)))
  )
  vdiffr::expect_doppelganger("nudge_line3a",
                              ggplot(df, aes(x, x * 2 + 5, label = l)) +
                                geom_abline(intercept = 5, slope = 2) +
                                geom_point() +
                                geom_text(position = position_nudge_line(abline = c(5, 2),
                                                                         x = c(-0.5, 0.5),
                                                                         y = c(-0.5, 0.5)))
  )
  vdiffr::expect_doppelganger("nudge_line4",
                              ggplot(subset(df, x >= 0), aes(x, yyy)) +
                                geom_point() +
                                geom_text(aes(label = l),
                                          position = position_nudge_line(direction = "split"))
  )
  vdiffr::expect_doppelganger("nudge_line5",
                              ggplot(subset(df, x >= 0), aes(y, yy)) +
                                stat_smooth(method = "lm", formula = y ~ x) +
                                geom_point() +
                                geom_text(aes(label = l),
                                          position = position_nudge_line(method = "lm",
                                                                         direction = "split"))
  )
  vdiffr::expect_doppelganger("nudge_line6",
                              ggplot(subset(df, x >= 0), aes(y, yy)) +
                                stat_smooth(method = "lm", formula = y ~ x) +
                                geom_point() +
                                geom_text(aes(label = l),
                                          position = position_nudge_line(method = "linear",
                                                                         direction = "split"))
  )
  vdiffr::expect_doppelganger("nudge_line7",
                              ggplot(subset(df, x >= 0), aes(y, yy)) +
                                stat_smooth(method = "lm", formula = y ~ x) +
                                geom_point() +
                                geom_text(aes(label = l),
                                          position = position_nudge_line(line_nudge = 1.5))
  )
})

test_that("nudge_to", {
  df <- data.frame(
    x = c(1,3,2,5,4,2.5),
    y = c(2, 1, 2.5, 1.8, 2.8, 1.5),
    label = c("abc","cd","d","c","bcd","a")
  )

  # errors ------------------------------------------------------------------

  expect_error(
    ggplot(df, aes(x, y, label = label)) +
      geom_text_s(position =
                    position_nudge_to(y = 3,
                                      kept.origin = "wrong value"))
  )
  expect_error(
    ggplot(df, aes(x, y, label = label)) +
      geom_text_s(position =
                    position_nudge_to(y = 3,
                                      x.action = "wrong value"))
  )
  expect_error(
    ggplot(df, aes(x, y, label = label)) +
      geom_text_s(position =
                    position_nudge_to(y = 3,
                                      y.action = "wrong value"))
  )
  expect_error(
    ggplot(df, aes(x, y, label = label)) +
      geom_point() +
      geom_text_s(position =
                    position_nudge_to(x = NA_real_,
                                      y.action = "spread"))
  )
  expect_error(
    ggplot(df, aes(x, y, label = label)) +
      geom_point() +
      geom_text_s(position =
                    position_nudge_to(x = NA,
                                      y.action = "spread"))
  )

  expect_no_error(
    ggplot(df, aes(x, y, label = label)) +
      geom_text_s(position =
                    position_nudge_to(y = 3,
                                      kept.origin = "none"))
  )
  expect_no_error(
    ggplot(df, aes(x, y, label = label)) +
      geom_text_s(position =
                    position_nudge_to(y = 3,
                                      x.action = "none"))
  )
  expect_no_error(
    ggplot(df, aes(x, y, label = label)) +
      geom_text_s(position =
                    position_nudge_to(y = 3,
                                      y.action = "none"))
  )


  expect_no_error(
    ggplot(df, aes(x, y, label = label)) +
                    geom_text_s(position =
                                  position_nudge_to(y = 3,
                                                    kept.origin = "original"))
    )
  expect_no_error(
    ggplot(df, aes(x, y, label = label)) +
                    geom_point() +
                    geom_text_s(position =
                                  position_nudge_to(y = 3,
                                                    x.action = "spread"))
    )
  expect_no_error(
    ggplot(df, aes(x, y, label = label)) +
      geom_point() +
      geom_text(position =
                    position_nudge_to(x = 4,
                                      y.action = "spread"))
  )
  expect_no_error(
    ggplot(df, aes(x, y, label = label)) +
      geom_point() +
      geom_text_s(position =
                  position_nudge_to(x = 0,
                                    y.action = "spread"))
  )


# warnings ----------------------------------------------------------------

  ## works outside testthat, but warning seems to be muted by testthat
  # expect_warning(
  #   ggplot(df, aes(x, y, label = label)) +
  #     geom_point() +
  #     geom_text_s(position =
  #                   position_nudge_to(x = rep(0, 7),
  #                                     y.action = "spread"))
  # )

  expect_no_warning(
    ggplot(df, aes(x, y, label = label)) +
      geom_point() +
      geom_text_s(position =
                    position_nudge_to(x = rep(0, 6),
                                      y.action = "spread"))
  )

  expect_no_warning(
    ggplot(df, aes(x, y, label = label)) +
      geom_point() +
      geom_text_s(position =
                    position_nudge_to(x = NULL,
                                      y.action = "spread"))
  )

  expect_no_warning(
    ggplot(df, aes(x, y, label = label)) +
      geom_point() +
      geom_text_s(position =
                    position_nudge_to(x = numeric(0),
                                      y.action = "spread"))
  )

  # x.action = none, y.action = none -----------------------------------------

  vdiffr::expect_doppelganger("nudge_to1",
                              ggplot(df, aes(x, y, label = label)) +
                                geom_point() +
                                geom_text_s(position = position_nudge_to(y = 3),
                                            vjust = 0)
  )
  vdiffr::expect_doppelganger("nudge_to1a",
                              ggplot(df, aes(x, y, label = label)) +
                                geom_point() +
                                geom_text_s(position = position_nudge_to(y = 3 + (1:6) / 6),
                                            vjust = 0)
  )
  vdiffr::expect_doppelganger("nudge_to2",
                              ggplot(df, aes(x, y, label = label)) +
                                geom_point() +
                                geom_text_s(position = position_nudge_to(x = 6),
                                            hjust = 0)
  )

  # x.action = spread, y.action = none --------------------------------------

  vdiffr::expect_doppelganger("nudge_to_spread_x1",
                              ggplot(df, aes(x, y, label = label)) +
                                geom_point() +
                                geom_text_s(position = position_nudge_to(x.action = "spread", y = 3),
                                            hjust = 0)
  )
  vdiffr::expect_doppelganger("nudge_to_expand_x1",
                              ggplot(df, aes(x, y, label = label)) +
                                geom_point() +
                                geom_text_s(position = position_nudge_to(x.action = "spread",
                                                                         y = 3,
                                                                         x.expansion = 0.1),
                                            hjust = 0)
  )
  vdiffr::expect_doppelganger("nudge_to_contract_x1",
                              ggplot(df, aes(x, y, label = label)) +
                                geom_point() +
                                geom_text_s(position = position_nudge_to(x.action = "spread",
                                                                         y = 3,
                                                                         x.expansion = -0.1),
                                            hjust = 0)
  )
  vdiffr::expect_doppelganger("nudge_to_contract_expand_x1",
                              ggplot(df, aes(x, y, label = label)) +
                                geom_point() +
                                geom_text_s(position = position_nudge_to(x.action = "spread",
                                                                         y = 3,
                                                                         x.expansion = c(-0.1, 0.1)),
                                            hjust = 0)
  )
  vdiffr::expect_doppelganger("nudge_to_spread_x2",
                              ggplot(df, aes(x, y, label = label)) +
                                geom_point() +
                                geom_text_s(position = position_nudge_to(x.action = "spread",
                                                                         x = c(2, 4), y = 3),
                                            hjust = 0)
  )
  vdiffr::expect_doppelganger("nudge_to_expand_x2",
                              ggplot(df, aes(x, y, label = label)) +
                                geom_point() +
                                geom_text_s(position = position_nudge_to(x.action = "spread",
                                                                         x = c(2, 4), y = 3,
                                                                         x.expansion = 0.1),
                                            hjust = 0)
  )

  # x.action = none, y.action = spread --------------------------------------

  vdiffr::expect_doppelganger("nudge_to_spread_y1",
                              ggplot(df, aes(x, y, label = label)) +
                                geom_point() +
                                geom_text_s(position = position_nudge_to(x = 6, y.action = "spread"),
                                            hjust = 0)
  )
  vdiffr::expect_doppelganger("nudge_to_expand_y1",
                              ggplot(df, aes(x, y, label = label)) +
                                geom_point() +
                                geom_text_s(position = position_nudge_to(x = 6,
                                                                         y.action = "spread",
                                                                         y.expansion = 0.1),
                                            hjust = 0)
  )
  vdiffr::expect_doppelganger("nudge_to_contract_y1",
                              ggplot(df, aes(x, y, label = label)) +
                                geom_point() +
                                geom_text_s(position = position_nudge_to(x = 6,
                                                                         y.action = "spread",
                                                                         y.expansion = -0.1),
                                            hjust = 0)
  )
  vdiffr::expect_doppelganger("nudge_to_contract_expand_y1",
                              ggplot(df, aes(x, y, label = label)) +
                                geom_point() +
                                geom_text_s(position = position_nudge_to(x = 6,
                                                                         y.action = "spread",
                                                                         y.expansion = c(-0.1, 0.1)),
                                            hjust = 0)
  )
  vdiffr::expect_doppelganger("nudge_to_spread_y2",
                              ggplot(df, aes(x, y, label = label)) +
                                geom_point() +
                                geom_text_s(position = position_nudge_to(x = 6, y = c(1, 3),
                                                                         y.action = "spread"),
                                            hjust = 0)
  )
  vdiffr::expect_doppelganger("nudge_to_expand_y2",
                              ggplot(df, aes(x, y, label = label)) +
                                geom_point() +
                                geom_text_s(position = position_nudge_to(x = 6, y = c(1, 3),
                                                                         y.action = "spread",
                                                                         y.expansion = 0.1),
                                            hjust = 0)
  )

  # x.action = spread, y.action = spread ------------------------------------

  vdiffr::expect_doppelganger("nudge_to_spread_x1y1",
                              ggplot(df, aes(x, y, label = label)) +
                                geom_point() +
                                geom_text_s(position = position_nudge_to(x.action = "spread",
                                                                         y.action = "spread"))
  )
  vdiffr::expect_doppelganger("nudge_to_expand_x1y1",
                              ggplot(df, aes(x, y, label = label)) +
                                geom_point() +
                                geom_text_s(position = position_nudge_to(x.action = "spread",
                                                                         y.action = "spread",
                                                                         x.expansion = 0.05,
                                                                         y.expansion = 0.05))
  )
})

#' Weather data
#'
#' A data set containing weather data measured in Viikki, Helsinki,
#' Finland. Values for all variables are means of 12 readings at 5 seconds
#' intervals. Sun angles were computed with R package 'photobiology'.
#'
#' The variables are as follows:
#' \itemize{
#'   \item time (yyyy-mm-dd hh:mm:ss)
#'   \item PAR_umol (umol m-2 s-1)
#'   \item PAR_diff_fr (/1)
#'   \item global_watt (W m-2)
#'   \item day_of_year
#'   \item month_of_year
#'   \item month_name
#'   \item calendar_year
#'   \item solar_time (h)
#'   \item sun_elevation (degrees above horizon)
#'   \item sun_azimuth (degrees)
#'   \item was_sunny (T/F)
#'   \item wind_speed (m s-1)
#'   \item wind_direction (degrees)
#'   \item air_temperature_C (C)
#'   \item air_RH (%)
#'   \item air_DP (C)
#'   \item air_pressure
#' }
#'
#' @docType data
#' @keywords datasets
#' @format A tibble with 18 columns and 1440 rows.
#'
#' @examples
#' names(weather_18_june_2019.df)
#' head(weather_18_june_2019.df)
#' nrow(weather_18_june_2019.df)
#'
#' @references P. J. Aphalo, unpublished data.
#'
"weather_18_june_2019.df"

#' Stack plus nudge labels to new positions
#'
#' Functions \code{position_stacknudge_to()} and \code{position_fillnudge_to()}
#' are meant to complement \code{position_stack()} and \code{position_fill()}
#' from 'ggplot2', adding as a second action that of \code{position_nudge_to()}.
#' These positions are generally useful for adjusting the position of labels or
#' text. As with other position functions in this package, the original
#' positions are preserved to allow the text or labels to be linked back to
#' their original position with a segment or arrow.
#'
#' @family position adjustments
#'
#' @param vjust Vertical adjustment for geoms that have a position (like points
#'   or lines), not a dimension (like bars or areas). Set to 0 to align with the
#'   bottom, 0.5 for the middle, and 1 (the default) for the top.
#' @param reverse If TRUE, will reverse the default stacking order. This is
#'   useful if you're rotating both the plot and legend.
#' @param x,y Coordinates of the destination position. A vector of mode
#'   \code{numeric}, that is extended if needed, to the same length as rows
#'   there are in \code{data}. The default, \code{NULL}, leaves the original
#'   coordinates unchanged after dodging.
#' @param x.action,y.action character string, one of \code{"none"}, or
#'   \code{"spread"}. With \code{"spread"} distributing the positions
#'   within the range of argument \code{x} or \code{y}, if non-null, or the
#'   range the variable mapped to \emph{x} or \code{y}, otherwise.
#' @param x.distance,y.distance character or numeric Currently only
#'   \code{"equal"} is implemented.
#' @param x.expansion,y.expansion numeric vectors of length 1 or 2, as a
#'   fraction of width of the range.
#' @param kept.origin One of \code{"original"}, \code{"stacked"} or
#'   \code{"none"}.
#'
#' @details These positions apply sequentially two actions, in the order they
#'   appear in their names. The applied stacking is similar to that by
#'   \code{\link[ggplot2]{position_stack}} and
#'   \code{\link[ggplot2]{position_fill}} while nudging is different to that
#'   by \code{\link[ggplot2]{position_nudge}} and equal to that applied by
#'   \code{\link{position_nudge_to}}.
#'
#'   The nudged to \code{x} and/or \code{y} values replace the original
#'   ones in \code{data}, while the original or the stacked coordinates are
#'   returned in \code{x_orig} and \code{y_orig}. Nudge values supported are
#'   those of \emph{mode} numeric, thus including dates and times when they
#'   match the mapped data.
#'
#'   If the length of \code{x} and/or \code{y} is more than one but less than
#'   rows are present in the data, the vector is both recycled and reordered so
#'   that the nudges are applied sequentially based on the data values. If their
#'   length matches the number of rows in data, they are assumed to be already
#'   in data order.
#'
#'   When applying stacking, the return of original positions instead of the
#'   stacked ones is achieved by passing \code{origin = "original"} instead of
#'   the default of \code{origin = "stacked"}.
#'
#' @note Irrespective of the action, the ordering of rows in \code{data} is
#'   preserved.
#'
#' @return A \code{"Position"} object.
#'
#' @seealso \code{\link{position_nudge_to}},
#'   \code{\link[ggplot2]{position_stack}}.
#'
#' @examples
#'
#' cols.df <- data.frame(x1 = c(1, 2, 1, 3),
#'                        x2 = c("a", "a", "b", "b"),
#'                        grp = c("A", "B", "C", "D"))
#'
#'  ggplot(data = cols.df, aes(x2, x1, group = grp)) +
#'  geom_col(aes(fill = grp), width=0.5, position = "stack") +
#'    geom_vline(xintercept = 0) +
#'    geom_text_s(
#'      aes(label = x1),
#'      position = position_stacknudge_to(x = c(1.4, 1.4, 1.6, 1.6))) +
#'    theme(legend.position = "none")
#'
#'  ggplot(data = cols.df, aes(x2, x1, group = grp)) +
#'  geom_col(aes(fill = grp), width=0.5, position = "fill") +
#'    geom_vline(xintercept = 0) +
#'    geom_text_s(
#'      aes(label = grp),
#'      position = position_fillnudge_to(x = c(1.4, 1.4, 1.6, 1.6))) +
#'    theme(legend.position = "none")
#'
#'  ggplot(data = cols.df, aes(x2, x1, group = grp)) +
#'  geom_col(aes(fill = grp), width=0.5, position = "fill") +
#'    geom_vline(xintercept = 0) +
#'    geom_text_s(
#'      aes(label = x1),
#'      position = position_fillnudge_to(vjust = 0.5,
#'                                       x = c(1.4, 1.4, 1.6, 1.6))) +
#'    theme(legend.position = "none")
#'
#'  ggplot(data = cols.df, aes(x2, x1, group = grp)) +
#'  geom_col(aes(fill = grp), width=0.5, position = "fill") +
#'    geom_vline(xintercept = 0) +
#'    geom_text_s(
#'      aes(label = x1), vjust = 0.5,
#'      position = position_fillnudge_to(vjust = 0.5,
#'                                       x = c(1.4, 1.4, 1.6, 1.6),
#'                                       y = c(0.85, 0.35, 0.85, 0.35))) +
#'    theme(legend.position = "none")
#'
#' @export
#'
position_stacknudge_to <-
  function(vjust = 1,
           reverse = FALSE,
           x = NULL,
           y = NULL,
           x.action = c("none", "spread"),
           y.action = c("none", "spread"),
           x.distance = "equal",
           y.distance = "equal",
           x.expansion = 0,
           y.expansion = 0,
           kept.origin = c("stacked", "original", "none")) {

    stopifnot("'x' must be NULL or of mode numeric" = length(x) == 0 ||
                (!anyNA(x) && mode(x) == "numeric"))
    stopifnot("'y' must be NULL or of mode numeric" = length(y) == 0 ||
                (!anyNA(y) && mode(y) == "numeric"))

    # this works as long as nudge and mapped variable are of the same class
    # ggplot2's behaviour has been in the past and seems to be again to expect
    # numeric seconds for POSIXct and numeric days for Date time shifts
    if (lubridate::is.instant(x)) {
      x <- as.numeric(x)
    }
    if (lubridate::is.instant(y)) {
      y <- as.numeric(y)
    }

    ggplot2::ggproto(NULL, PositionStackNudgeTo,
                     x = x,
                     y = y,
                     x.action = rlang::arg_match(x.action),
                     y.action = rlang::arg_match(y.action),
                     x.distance = x.distance,
                     y.distance = y.distance,
                     x.expansion = rep_len(x.expansion, 2),
                     y.expansion = rep_len(y.expansion, 2),
                     kept.origin = rlang::arg_match(kept.origin),
                     vjust = vjust,
                     reverse = reverse
    )
  }

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
PositionStackNudgeTo <-
  ggplot2::ggproto(
    "PositionStackNudgeTo",
    ggplot2::PositionStack,
    fill = FALSE,
    x = NULL,
    y = NULL,

    setup_params = function(self, data) {
      c(list(x = self$x,
             y = self$y,
             x.action = self$x.action,
             y.action = self$y.action,
             x.distance = self$x.distance,
             y.distance = self$y.distance,
             x.expansion = self$x.expansion,
             y.expansion = self$y.expansion,
             x.reorder = !is.null(self$x) && length(self$x) > 1 && length(self$x) < nrow(data),
             y.reorder = !is.null(self$y) && length(self$y) > 1 && length(self$y) < nrow(data),
             kept.origin = self$kept.origin,
             vjust = self$vjust,
             reverse = self$reverse),
        ggplot2::ggproto_parent(ggplot2::PositionStack, self)$setup_params(data))
    },

    setup_data = function(self, data, params) {
      data <- flip_data(data, params$flipped_aes)
      if (is.null(params$var)) {
        return(data)
      }

      data$ymax <- switch(params$var,
                          y = data$y,
                          ymax = as.numeric(ifelse(data$ymax == 0, data$ymin, data$ymax))
      )

      data <- remove_missing(
        data,
        vars = c("x", "xmin", "xmax", "y"),
        name = "position_stack"
      )
      flip_data(data, params$flipped_aes)
    },

    compute_layer = function(self, data, params, layout) {
      x_orig <- x_stacked <- data$x
      y_orig <- data$y

      # operate on the stacked positions (updated in August 2020)
      data = ggplot2::ggproto_parent(ggplot2::PositionStack, self)$compute_layer(data, params, layout)
      y_stacked <- data$y

      # compute/convert x nudges
      if (!length(params$x)) {
        # set default x
        if (params$x.action == "none") {
          params$x <- rep_len(0, nrow(data))
        } else if (params$x.action == "spread") {
          params$x <- range(x_stacked)
        }
      } else if (is.numeric(params$x)) {
        # check user supplied x
        if (length(params$x) > nrow(data)) {
          warning("Argument 'x' longer than data: some values dropped!")
        }
        if (params$x.action == "none") {
          # recycle or trim x as needed
          if (params$x.reorder) {
            params$x <- rep_len(params$x, nrow(data))[order(order(data$x))] - x_stacked
          } else {
            params$x <- rep_len(params$x, nrow(data)) - x_stacked
          }
        } else if (params$x.action == "spread") {
          params$x <- range(params$x)
        }
      }

      if (params$x.action == "spread") {
        # apply x.expansion to x
        x.spread <- diff(params$x)
        params$x[1] <- params$x[1] - params$x.expansion[1] * x.spread
        params$x[2] <- params$x[2] + params$x.expansion[2] * x.spread
        if (params$x.distance == "equal") {
          # evenly spaced sequence of positions ordered as in data
          params$x <- seq(from = params$x[1],
                          to = params$x[2],
                          length.out = nrow(data))[order(order(data$x))] - x_stacked
        }
        # other strategies to distribute positions could be added here
      }

      # compute/convert y nudges
      if (!length(params$y)) {
        # set default y
        if (params$y.action == "none") {
          params$y <- rep_len(0, nrow(data))
        } else if (params$y.action == "spread") {
          params$y <- range(y_stacked)
        }
      } else if (is.numeric(params$y)) {
        # check user supplied y
        if (length(params$y) > nrow(data)) {
          warning("Argument 'y' longer than data: some values dropped!")
        }
        if (params$y.action == "none") {
          # recycle or trim y as needed
          if (params$y.reorder) {
            params$y <- rep_len(params$y, nrow(data))[order(order(data$y))] - y_stacked
          } else {
            params$y <- rep_len(params$y, nrow(data)) - y_stacked
          }
        } else if (params$y.action == "spread") {
          params$y <- range(params$y)
        }
      }

      if (params$y.action == "spread") {
        y.spread <- diff(params$y)
        params$y[1] <- params$y[1] - params$y.expansion[1] * y.spread
        params$y[2] <- params$y[2] + params$y.expansion[2] * y.spread
        if (params$y.distance == "equal") {
          # evenly spaced sequence ordered as in data
          params$y <- seq(from = params$y[1],
                          to = params$y[2],
                          length.out = nrow(data))[order(order(data$y))] - y_stacked
        }
        # other strategies could be added here
      }

      # As in 'ggplot2' we apply the nudge to xmin, xmax, xend, ymin, ymax, and yend.
      # Transform the dimensions for which not all nudges are zero
      if (any(params$x != 0)) {
        if (any(params$y != 0)) {
          data <- transform_position(data, function(x) x + params$x, function(y) y + params$y)
        } else {
          data <- transform_position(data, function(x) x + params$x, NULL)
        }
      } else if (any(params$y != 0)) {
        data <- transform_position(data, NULL, function(y) y + params$y)
      }
      # add original position
      if (params$kept.origin == "stacked") {
        data$x_orig <- x_stacked
        data$y_orig <- y_stacked
      } else if (params$kept.origin == "original") {
        data$x_orig <- x_orig
        data$y_orig <- y_orig
      }
      data
    },

    compute_panel = function(self, data, params, scales) {
      ggplot2::ggproto_parent(PositionStack, self)$compute_panel(data, params, scales)
    }
  )

#' @rdname position_stacknudge_to
#'
#' @export
#'
position_fillnudge_to <-
  function(vjust = 1,
           reverse = FALSE,
           x = NULL,
           y = NULL,
           x.action = c("none", "spread"),
           y.action = c("none", "spread"),
           x.distance = "equal",
           y.distance = "equal",
           x.expansion = 0,
           y.expansion = 0,
           kept.origin = c("stacked", "original", "none")) {

    stopifnot("'x' must be NULL or of mode numeric" = length(x) == 0 ||
                (!anyNA(x) && mode(x) == "numeric"))
    stopifnot("'y' must be NULL or of mode numeric" = length(y) == 0 ||
                (!anyNA(y) && mode(y) == "numeric"))

    # this works as long as nudge and mapped variable are of the same class
    # ggplot2's behaviour has been in the past and seems to be again to expect
    # numeric seconds for POSIXct and numeric days for Date time shifts
    if (lubridate::is.instant(x)) {
      x <- as.numeric(x)
    }
    if (lubridate::is.instant(y)) {
      y <- as.numeric(y)
    }

    ggplot2::ggproto(NULL, PositionFillNudgeTo,
                     x = x,
                     y = y,
                     x.action = rlang::arg_match(x.action),
                     y.action = rlang::arg_match(y.action),
                     x.distance = x.distance,
                     y.distance = y.distance,
                     x.expansion = rep_len(x.expansion, 2),
                     y.expansion = rep_len(y.expansion, 2),
                     kept.origin = rlang::arg_match(kept.origin),
                     vjust = vjust,
                     reverse = reverse
    )
  }

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
PositionFillNudgeTo <-
  ggplot2::ggproto("PositionFillNudgeTo", PositionStackNudgeTo,
                   fill = TRUE
  )

context("position_stacknudge_to")

test_that("incorrect kept.origin used", {
  expect_error(
    position_stacknudge_to(kept.origin = "") #,
#    "`kept.origin` must be one of \"stacked\", \"original\", or \"none\", not \"\"."
# rlang arg_match() adds to the error message characters that break expect_error()
  )
})

test_that("correct kept.origin used", {
  expect_no_error(
    position <- position_stacknudge_to()
  )
  expect_identical(position$kept.origin, "stacked")
})

test_that("correct kept.origin used", {
  expect_no_error(
    position <- position_stacknudge_to(kept.origin = "none")
  )
  expect_identical(position$kept.origin, "none")
})

test_that("correct kept.origin used", {
  expect_no_error(
    position <- position_stacknudge_to(kept.origin = "original")
  )
  expect_identical(position$kept.origin, "original")
})

test_that("correct kept.origin used", {
  expect_no_error(
    position <- position_stacknudge_to(kept.origin = "stacked")
  )
  expect_identical(position$kept.origin, "stacked")
})

test_that("non-numeric y used", {
  expect_error(
    position_stacknudge_to(y = "bad")
    #, "'y' must be NULL or of mode numeric"
  )
})

test_that("NA y used", {
  expect_error(
    position_stacknudge_to(y = NA)
    #, "'y' must be NULL or of mode numeric"
  )
})

test_that("non-numeric x used", {
  expect_error(
    position_stacknudge_to(x = "bad")
    #, "'x' must be NULL or of mode numeric"
  )
})

test_that("NA x used", {
  expect_error(
    position_stacknudge_to(x = "NA")
    #, "'x' must be NULL or of mode numeric"
  )
})

test_that("test if correct arguments are assigned", {
  position <- position_stacknudge_to()
  expect_false(position$reverse)
  expect_identical(position$vjust, 1)
  expect_null(position$x)
  expect_null(position$y)
  expect_identical(position$x.action, "none")
  expect_identical(position$y.action, "none")
  expect_identical(position$x.distance, "equal")
  expect_identical(position$y.distance, "equal")
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
})

test_that("test if correct arguments are assigned with 'y.action'", {
  position <- position_stacknudge_to(y.action = "spread")
  expect_false(position$reverse)
  expect_identical(position$vjust, 1)
  expect_null(position$x)
  expect_null(position$y)
  expect_identical(position$x.action, "none")
  expect_identical(position$y.action, "spread")
  expect_identical(position$x.distance, "equal")
  expect_identical(position$y.distance, "equal")
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
})

test_that("test if correct arguments are assigned with 'x.action'", {
  position <- position_stacknudge_to(x.action = "spread")
  expect_false(position$reverse)
  expect_identical(position$vjust, 1)
  expect_null(position$x)
  expect_null(position$y)
  expect_identical(position$x.action, "spread")
  expect_identical(position$y.action, "none")
  expect_identical(position$x.distance, "equal")
  expect_identical(position$y.distance, "equal")
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
})


test_that("warn on x or y longer than data", {
  expect_no_warning(
  ggplot(mtcars, aes(x = am, y = mpg, label = mpg)) +
    geom_point(position = "fill") +
    geom_text_s(
      position = position_stacknudge_to(x = (1:nrow(mtcars)) * 0.9 / nrow(mtcars)),
      size = 2.5
    )
)

expect_no_warning(
  ggplot(mtcars, aes(x = am, y = mpg, label = mpg)) +
    geom_point(position = "fill") +
    geom_text_s(
      position = position_stacknudge_to(x = c(0.3, 0.7)),
      size = 2.5
    )
)

# warning is issue but not "seen" by 'testthat'
# expect_warning(
#   ggplot(mtcars, aes(x = am, y = mpg, label = mpg)) +
#     geom_point(position = "fill") +
#     geom_text_s(
#       position = position_stacknudge_to(x = rep(c(0.3, 0.7), nrow(mtcars))),
#       size = 2.5
#     )
# )

})

#' Apply a function to x or y values
#'
#' \code{stat_summary_xy()} and \code{stat_centroid()} are similar to
#' \code{ggplot2::stat_summary()} but summarize both \code{x} and \code{y}
#' values in the same plot layer. Differently to \code{stat_summary()} no
#' grouping based on data \code{values} is done; the grouping respected is that
#' already present based on mappings to aesthetics. This makes it possible to
#' highlight the actual location of the centroid with \code{geom_point()},
#' \code{geom_text()}, and similar geometries. Instead, if we use
#' \code{geom_rug()} they are only a convenience avoiding the need to add two
#' separate layers and flipping one of them using \code{orientation = "y"}.
#'
#' \code{stat_apply_group} applies functions to data.
#' When possible it is preferable to use transformations through scales or
#' summary functions such as \code{ggplot2::stat_summary()},
#' \code{stat_summary_xy()} or \code{stat_centroid()}. There are some
#' computations that are not scale transformations but are not usual summaries
#' either, as the number of data values does not decrease all the way to one row
#' per group. A typical case for a summary is the computation of quantiles. For
#' transformations are cumulative ones, e.g., using \code{cumsum()},
#' \code{runmed()} and similar functions. Obviously, it is always possible to
#' apply such functions to the data before plotting and passing them to a single
#' layer function. However, it can be useful to apply such functions on-the-fly
#' to ensure that grouping is consistent between computations and aesthetics.
#' One particularity of these statistics is that they can apply simultaneously
#' different functions to \code{x} values and to \code{y} values when needed. In
#' contrast to these statistics, \code{\link[ggplot2]{geom_smooth}} applies a
#' function that takes both \code{x} and \code{y} values as arguments.
#'
#' @param mapping The aesthetic mapping, usually constructed with
#'   \code{\link[ggplot2]{aes}}. Only needs to be set at the layer level if you
#'   are overriding the plot defaults.
#' @param data A layer specific dataset - only needed if you want to override
#'   the plot defaults.
#' @param geom The geometric object to use display the data
#' @param position The position adjustment to use for overlapping points on this
#'   layer
#' @param show.legend logical. Should this layer be included in the legends?
#'   \code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
#'   never includes, and \code{TRUE} always includes.
#' @param inherit.aes If \code{FALSE}, overrides the default aesthetics, rather
#'   than combining with them. This is most useful for helper functions that
#'   define both data and aesthetics and shouldn't inherit behaviour from the
#'   default plot specification, e.g. \code{\link[ggplot2]{borders}}.
#' @param ... other arguments passed on to \code{\link[ggplot2]{layer}}. This
#'   can include aesthetics whose values you want to set, not map. See
#'   \code{\link[ggplot2]{layer}} for more details.
#' @param na.rm	a logical value indicating whether NA values should be stripped
#'   before the computation proceeds.
#' @param .fun.x,.fun.y,.fun function to be applied or the name of the function
#'   to be applied as a character string.
#' @param .fun.x.args,.fun.y.args,.fun.args additional arguments to be passed to
#'   the function as a named list.
#'
#' @details These four statistics are similar. They differ on whether they
#'   return a single or multiple rows of data per group.
#'
#' @return A data frame with the same variables as the data input, with either a
#'   single or multiple rows, with the values of \code{x} and \code{y} variables
#'   replaced by the values returned by the applied functions, or possibly
#'   filled with \code{NA} if no function was supplied or available by default.
#'   If the applied function returns a named vector, the names are copied into
#'   columns \code{x.names} and/or \code{y.names}. If the summary function
#'   applied returns a one row data frame, it will be column bound keeping
#'   the column names, but overwritting columns x and/or y with y from the
#'   summary data frame. In the names returned by \code{.fun.x} the letter
#'   "y" is replaced by "x". These allows the use of the same functions as in
#'   \code{ggplot2::stat_summary()}.
#'
#'   \describe{
#'   \item{x}{x-value as returned by \code{.fun.x}, with names removed}
#'   \item{y}{y-value as returned by \code{.fun.y}, with names removed}
#'   \item{x.names}{if the x-value returned by \code{.fun.x} is named, these names}
#'   \item{y.names}{if the y-value returned by \code{.fun.y} is named, these names}
#'   \item{xmin, xmax}{values returned by \code{.fun.x} under these names, if present}
#'   \item{ymin, ymax}{values returned by \code{.fun.y} under these names, if present}
#'   \item{<other>}{additional values as returned by \code{.fun.y} under other names}
#'   }
#'
#' @note The applied function(s) must accept as first argument a vector that
#'   matches the variables mapped to \code{x} or \code{y} aesthetics. For
#'   \code{stat_summary_xy()} and \code{stat_centroid()} the function(s) to be
#'   applied is(are) expected to return a vector of length 1 or a data frame
#'   with only one row, as \code{mean_se()}, \code{mean_cl_normal()}
#'   \code{mean_cl_boot()}, \code{mean_sdl()} and \code{median_hilow()} from
#'   'ggplot2' do.
#'
#'   For \code{stat_apply_group} the vectors returned by the
#'   the functions applied to \code{x} and \code{y} must be of exactly the same
#'   length. When only one of \code{.fun.x} or \code{.fun.y} are passed a
#'   function as argument, the other variable in the returned data is filled
#'   with \code{NA_real_}. If other values are desired, they can be set by means
#'   of a user-defined function.
#'
#' @references
#'
#' Answers to question "R ggplot on-the-fly calculation by grouping variable" at
#' \url{https://stackoverflow.com/questions/51412522}.
#'
#' @examples
#' set.seed(123456)
#' my.df <- data.frame(X = rep(1:20,2),
#'                     Y = runif(40),
#'                     category = rep(c("A","B"), each = 20))
#'
#' # make sure rows are ordered for X as we will use functions that rely on this
#' my.df <- my.df[order(my.df[["X"]]), ]
#'
#' # Centroid
#' ggplot(my.df, aes(x = X, y = Y, colour = category)) +
#'   stat_centroid(shape = "cross", size = 6) +
#'   geom_point()
#'
#' ggplot(my.df, aes(x = X, y = Y, colour = category)) +
#'   stat_centroid(geom = "rug", linewidth = 1.5, .fun = median) +
#'   geom_point()
#'
#' ggplot(my.df, aes(x = X, y = Y, colour = category)) +
#'   stat_centroid(geom = "text", aes(label = category)) +
#'   geom_point()
#'
#' ggplot(my.df, aes(x = X, y = Y, colour = category)) +
#'   stat_summary_xy(geom = "pointrange",
#'                   .fun.x = mean, .fun.y = mean_se) +
#'   geom_point()
#'
#' # quantiles
#' ggplot(my.df, aes(x = X, y = Y, colour = category)) +
#'   geom_point() +
#'   stat_apply_group(geom = "rug", .fun.y = quantile, .fun.x = quantile)
#'
#' ggplot(my.df, aes(x = X, y = Y)) +
#'   geom_point() +
#'   stat_apply_group(geom = "rug", sides = "lr", color = "darkred",
#'                    .fun.y = quantile) +
#'   stat_apply_group(geom = "text", hjust = "right", color = "darkred",
#'                    .fun.y = quantile,
#'                    .fun.x = function(x) {rep(22, 5)}, # set x to 22
#'                    mapping = aes(label = after_stat(y.names))) +
#'                    expand_limits(x = 21)
#'
#' my.probs <- c(0.25, 0.5, 0.75)
#' ggplot(my.df, aes(x = X, y = Y, colour = category)) +
#'   geom_point() +
#'   stat_apply_group(geom = "hline",
#'                    aes(yintercept = after_stat(y)),
#'                    .fun.y = quantile,
#'                    .fun.y.args = list(probs = my.probs))
#'
#' # cummulative summaries
#' ggplot(my.df, aes(x = X, y = Y, colour = category)) +
#'   stat_apply_group(.fun.x = function(x) {x},
#'                    .fun.y = cummax)
#'
#' ggplot(my.df, aes(x = X, y = Y, colour = category)) +
#'   stat_apply_group(.fun.x = cumsum, .fun.y = cumsum)
#'
#' # diff returns a shorter vector by 1 for each group
#' ggplot(my.df, aes(x = X, y = Y, colour = category)) +
#'   stat_apply_group(.fun.x = function(x) {x[-1L]},
#'                    .fun.y = diff, na.rm = TRUE)
#'
#' # Running summaries
#' ggplot(my.df, aes(x = X, y = Y, colour = category)) +
#'   geom_point() +
#'   stat_apply_group(.fun.x = function(x) {x},
#'                    .fun.y = runmed, .fun.y.args = list(k = 5))
#'
#' # Rescaling per group
#' ggplot(my.df, aes(x = X, y = Y, colour = category)) +
#'   stat_apply_group(.fun.x = function(x) {x},
#'                    .fun.y = function(x) {(x - min(x)) / (max(x) - min(x))})
#'
#' # inspecting the returned data
#' if (requireNamespace("gginnards", quietly = TRUE)) {
#'   library(gginnards)
#'
#'   ggplot(my.df, aes(x = X, y = Y, colour = category)) +
#'     stat_centroid(.fun = mean_se, geom = "debug_group")
#'
#'   ggplot(my.df, aes(x = X, y = Y, colour = category)) +
#'     stat_summary_xy(.fun.y = mean_se, geom = "debug_group")
#'
#'   ggplot(my.df, aes(x = X, y = Y, colour = category)) +
#'     stat_apply_group(.fun.y = cumsum, geom = "debug_group")
#'
#'   ggplot(my.df, aes(x = X, y = Y, colour = category)) +
#'     geom_point() +
#'     stat_apply_group(geom = "debug_group",
#'                     .fun.x = quantile,
#'                     .fun.x.args = list(probs = my.probs),
#'                     .fun.y = quantile,
#'                    .fun.y.args = list(probs = my.probs))
#' }
#'
#' @rdname stat_apply
#'
#' @export
#' @family summary stats
#'
stat_apply_group <- function(mapping = NULL,
                             data = NULL,
                             geom = "line",
                             .fun.x = NULL,
                             .fun.x.args = list(),
                             .fun.y = NULL,
                             .fun.y.args = list(),
                             position = "identity",
                             na.rm = FALSE,
                             show.legend = FALSE,
                             inherit.aes = TRUE,
                             ...) {
  .fun.x.null <- is.null(.fun.x)
  if (.fun.x.null) {
    .fun.x <- function(x) {x}
  }
  .fun.y.null <- is.null(.fun.y)
  if (.fun.y.null) {
    .fun.y <- function(x) {x}
  }
  ggplot2::layer(
    stat = StatApplyGroup,
    data = data,
    mapping = mapping,
    geom = geom,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(.fun.x = .fun.x,
                  .fun.x.args = .fun.x.args,
                  .fun.x.null = .fun.x.null,
                  .fun.y = .fun.y,
                  .fun.y.args = .fun.y.args,
                  .fun.y.null = .fun.y.null,
                  na.rm = na.rm,
                  single.row = FALSE,
                  ...)
  )
}

#' @rdname stat_apply
#'
#' @export
#'
stat_summary_xy <- function(mapping = NULL,
                            data = NULL,
                            geom = "point",
                            .fun.x = NULL,
                            .fun.x.args = list(),
                            .fun.y = NULL,
                            .fun.y.args = list(),
                            position = "identity",
                            na.rm = FALSE,
                            show.legend = FALSE,
                            inherit.aes = TRUE,
                            ...) {
  .fun.x.null <- is.null(.fun.x)
  if (.fun.x.null) {
    .fun.x <- function(x) {x}
  }
  .fun.y.null <- is.null(.fun.y)
  if (.fun.y.null) {
    .fun.y <- function(x) {x}
  }
  ggplot2::layer(
    stat = StatApplyGroup, data = data, mapping = mapping, geom = geom,
    position = position, show.legend = show.legend, inherit.aes = inherit.aes,
    params = list(.fun.x = .fun.x,
                  .fun.x.args = .fun.x.args,
                  .fun.x.null = .fun.x.null,
                  .fun.y = .fun.y,
                  .fun.y.args = .fun.y.args,
                  .fun.y.null = .fun.y.null,
                  na.rm = na.rm,
                  single.row = TRUE,
                  ...)
  )
}

#' @rdname stat_apply
#'
#' @export
#'
stat_centroid <- function(mapping = NULL,
                          data = NULL,
                          geom = "point",
                          .fun = NULL,
                          .fun.args = list(),
                          position = "identity",
                          na.rm = FALSE,
                          show.legend = FALSE,
                          inherit.aes = TRUE,
                          ...) {
  if (is.null(.fun)) {
    .fun <- ggplot2::mean_se
  }
  ggplot2::layer(
    stat = StatApplyGroup,
    data = data,
    mapping = mapping,
    geom = geom,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(.fun.x = .fun,
                  .fun.x.args = .fun.args,
                  .fun.x.null = FALSE,
                  .fun.y = .fun,
                  .fun.y.args = .fun.args,
                  .fun.y.null = FALSE,
                  na.rm = na.rm,
                  single.row = TRUE,
                  ...)
  )
}

#' \code{Stat*} Objects
#'
#' All \code{stat_*} functions (like \code{stat_bin}) return a layer that
#' contains a \code{Stat*} object (like \code{StatBin}). The \code{Stat*}
#' object is responsible for rendering the data in the plot.
#'
#' Each of the \code{Stat*} objects is a \code{\link[ggplot2]{ggproto}} object, descended
#' from the top-level \code{Stat}, and each implements various methods and
#' fields. To create a new type of Stat object, you typically will want to
#' implement one or more of the following:
#'
#' @name Stats
#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
#' @seealso \code{\link[ggplot2]{ggplot2-ggproto}}
#' @keywords internal
StatApplyGroup <-
  ggplot2::ggproto("StatApplyGroup", ggplot2::Stat,
                   compute_group = function(data,
                                            scales,
                                            .fun.x, .fun.x.args, .fun.x.null,
                                            .fun.y, .fun.y.args, .fun.y.null,
                                            single.row,
                                            na.rm = FALSE) {

                     new.data <- data[1, ]
                     unique_value_cols <-
                       sapply(data, function(x) {length(unique(x)) == 1L})

                     if (sum(!unique_value_cols) > 2L) {
                       bad.vars <-
                         setdiff(names(data)[!unique_value_cols], c("x", "y"))
                       warning("Non-unique values in columns: ", bad.vars)
                       new.data[ , bad.vars] <- NA
                     }

                     if (single.row) {
                       # stat_apply_fun_rw(data,
                       #                   new.data,
                       #                   scales,
                       #                   .fun.x, .fun.x.args, .fun.x.null,
                       #                   .fun.y, .fun.y.args, .fun.y.null)
                       if (!.fun.x.null) {
                         args <- c(unname(data["x"]), .fun.x.args)
                         new.x <- do.call(.fun.x, args = args)

                         stopifnot("'new.x' is NULL." = !is.null(new.x))

                         if (is.vector(new.x) && (length(new.x) == 1L)) {
                           if(!is.null(names(new.x))) {
                             new.data[["x.names"]] <- names(new.x)
                             new.x <- unname(new.x)
                           }
                           new.data[["x"]] <- new.x
                         } else if (is.data.frame(new.x) && (nrow(new.x) == 1L)) {
                           # functions like mean_se() return columns y, ymin and ymax.
                           names(new.x) <- gsub("y", "x", names(new.x))
                           if ("x" %in% names(new.x)) {
                             new.data <- cbind(new.data[ , setdiff(names(new.data), "x")],
                                               new.x)
                           } else {
                             new.data <- cbind(new.data, new.x)
                           }
                         } else {
                           warning("Object of class '", class(new.x),
                                   "' with names '", names(new.x),
                                   "' from '.fun.x', is incompatible.")
                         }
                       }

                       if (!.fun.y.null) {
                         args <- c(unname(data["y"]), .fun.y.args)
                         new.y <- do.call(.fun.y, args = args)

                         stopifnot("'new.y' is NULL." = !is.null(new.y))

                         if (is.vector(new.y) && (length(new.y) == 1L)) {
                           if(!is.null(names(new.y))) {
                             new.data[["y.names"]] <- names(new.y)
                             new.y <- unname(new.y)
                           }
                           new.data[["y"]] <- new.y
                         } else if (is.data.frame(new.y) && (nrow(new.y) == 1L)) {
                           if ("y" %in% names(new.y)) {
                             new.data <- cbind(new.data[ , setdiff(names(new.data), "y")],
                                               new.y)
                           } else {
                             new.data <- cbind(new.data, new.y)
                           }
                         } else {
                           warning("Object of class '", class(new.y),
                                   "' with names '", names(new.y),
                                   "' from '.fun.y', is incompatible.")
                         }
                       }

                       new.data
                     } else {
                       # stat_apply_fun_vc(data,
                       #                   new.data,
                       #                   scales,
                       #                   .fun.x, .fun.x.args, .fun.x.null,
                       #                   .fun.y, .fun.y.args, .fun.y.null)
                       if (!.fun.x.null) {
                         args <- c(unname(data["x"]), .fun.x.args)
                         new.x <- do.call(.fun.x, args = args)
                         stopifnot("'new.x' must be 'numeric'" =
                                     is.numeric(new.x))
                         x.names <- names(new.x)
                       }

                       if (!.fun.y.null) {
                         args <- c(unname(data["y"]), .fun.y.args)
                         new.y <- do.call(.fun.y, args = args)
                         stopifnot("'new.y' must be 'numeric'" =
                                     is.numeric(new.y))
                         y.names <- names(new.y)
                       }

                       if (.fun.x.null && .fun.y.null) {
                         new.data[["y"]] <- NA_real_
                         new.data[["y"]] <- NA_real_
                       } else if (.fun.y.null) {
                         new.data <- new.data[rep(1L, length(new.x)), ]
                         new.data[["x"]] <- new.x
                         new.data[["y"]] <- NA_real_
                         if (!is.null(x.names)) {
                           new.data[["x.names"]] <- x.names
                         }
                       } else if (.fun.x.null) {
                         new.data <- new.data[rep(1L, length(new.y)), ]
                         new.data[["x"]] <- NA_real_
                         new.data[["y"]] <- new.y
                         if (!is.null(y.names)) {
                           new.data[["y.names"]] <- y.names
                         }
                       } else if (length(new.x) == length(new.y)) {
                         new.data <- new.data[rep(1L, length(new.y)), ]
                         new.data[["x"]] <- new.x
                         new.data[["y"]] <- new.y
                         if (!is.null(x.names)) {
                           new.data[["x.names"]] <- x.names
                         }
                         if (!is.null(y.names)) {
                           new.data[["y.names"]] <- y.names
                         }
                       } else {
                         warning("x and y summaries of different length!")
                         new.data[["y"]] <- NA_real_
                         new.data[["y"]] <- NA_real_
                       }

                       new.data
                     }
                   },

                   required_aes = c("x", "y"),
                   extra_params = c("na.rm")
  )

context("stat_centroid")

set.seed(123456)
my.df <- data.frame(X = rep(1:5,2),
                    Y = runif(10),
                    category = rep(c("A","B"), each = 5))

# make sure rows are ordered for X as we will use functions that rely on this
my.df <- my.df[order(my.df[["X"]]), ]


test_that("stat_centroid, using default shape and size", {
  p1 <- ggplot(my.df, aes(x = X, y = Y, colour = category)) +
    stat_centroid() +
    geom_point()

  result <- layer_data(p1)[, c("xmin", "xmax", "ymin", "ymax", "shape", "size")]
  expected <- data.frame(
    xmin = c(2.292893, 2.292893),
    xmax = c(3.707107, 3.707107),
    ymin = c(0.4278716, 0.2311237),
    ymax = c(0.6303108, 0.5629344),
    shape = c(19, 19),
    size = c(1.5, 1.5)
  )
  expect_identical(format(result, digits = 7), format(expected, digits = 7))
})

test_that("stat_centroid, changed default shape and size", {
  p1 <- ggplot(my.df, aes(x = X, y = Y, colour = category)) +
    stat_centroid(shape = "cross", size = 6) +
    geom_point()

  result <- layer_data(p1)[, c("xmin", "xmax", "ymin", "ymax", "shape", "size")]

  expected <- data.frame(
    xmin = c(2.292893, 2.292893),
    xmax = c(3.707107, 3.707107),
    ymin = c(0.4278716, 0.2311237),
    ymax = c(0.6303108, 0.5629344),
    shape = c("cross", "cross"),
    size = c(6, 6)
  )
  expect_identical(format(result, digits = 7), format(expected, digits = 7))
})

test_that("stat_centroid, geom 'rug' function median applied", {
  p1 <- ggplot(my.df, aes(x = X, y = Y, colour = category)) +
    stat_centroid(geom = "rug", linewidth = 1.5, .fun = median) +
    geom_point()

  result <- layer_data(p1)[, c("x", "y", "linewidth", "linetype")]

  expected <- data.frame(
    x = c(3,3),
    y = c(0.3912557, 0.1983447),
    linewidth = c(1.5, 1.5),
    linetype = c(1, 1)
  )
  expect_identical(format(result, digits = 7), format(expected, digits = 7))
})

context("stat_summary_xy")

library(ggplot2)

test_that("stat_summary_xy", {
  tst.df <- data.frame(
    x = c(2.2, -0.2, -0.2, -2.5, -0.6, -0.1),
    y = c(1.1, -0.6, -0.9, -1.6, 0.3, 1.6),
    group = c("A", "A", "A", "B", "B", "B")
  )
  tst.df.na <- data.frame(
    x = c(2.2, -0.2, -0.2, -2.5, -0.6, 1.1, 0.1),
    y = c(1.1, -0.6, -0.9, -1.6, 0.3, 1.6, 1.6),
    group = c("A", "A", "A", "B", "B", "B", "B")
  )
  result <- ggplot(tst.df, aes(x, y)) +
    stat_summary_xy(.fun.x = mean)
  result <- ggplot2::layer_data(result)
  result <- result$x
  expected <- mean(tst.df$x)
  expect_identical(result, expected)

  result <- ggplot(tst.df, aes(x, y)) +
    stat_summary_xy(.fun.y = mean_se)
  result <- ggplot2::layer_data(result)
  result <- result[, c("y", "ymin", "ymax")]
  expected <- mean_se(tst.df$y)
  expect_identical(result, expected)

  result <- ggplot(tst.df, aes(x, y)) +
    stat_summary_xy(.fun.x = min)
  result <- ggplot2::layer_data(result)
  result <- result$x
  expected <- min(tst.df$x)
  expect_identical(result, expected)

  result <- ggplot(tst.df.na, aes(x, y)) +
    stat_summary_xy(
      .fun.y = mean,
      .fun.x = mean,
      .fun.y.args = list(na.rm = TRUE),
      .fun.x.args = list(na.rm = TRUE)
    )
  result <- ggplot2::layer_data(result)
  result <- result[, c("x", "y")]
  expected <- as.data.frame(lapply(na.omit(tst.df.na[, c("x", "y")]), mean, na.rm = TRUE))
  expect_identical(result, expected)

  function_passed <- "mean_se"
  result <- ggplot(tst.df, aes(x, y)) +
    stat_summary_xy(.fun.y = function_passed)
  result <- ggplot2::layer_data(result)
  result <- result[, c("y", "ymin", "ymax")]
  expected <- mean_se(tst.df$y)
  expect_identical(result, expected)

  result <- ggplot(tst.df, aes(x, y)) +
    stat_summary_xy(.fun.y = function(x) mean(x))
  result <- ggplot2::layer_data(result)
  result <- result$y
  expected <- mean(tst.df$y)
  expect_identical(result, expected)
})

# could do with additional tests
test_that("stat_apply_group", {
  tst.df <- data.frame(
    x = c(2.2, -0.2, -0.2, -2.5, -0.6, -0.1),
    y = c(1.1, -0.6, -0.9, -1.6, 0.3, 1.6),
    group = c("A", "A", "A", "B", "B", "B")
  )
  result <- ggplot(tst.df, aes(x, y, group = group)) +
    stat_apply_group(.fun.x = mean)
  result <- ggplot2::layer_data(result)
  result <- result$x
  expected <- c(mean(tst.df$x[1:3]), mean(tst.df$x[4:6]))
  expect_identical(result, expected)

  result <- ggplot(tst.df, aes(x, y, group = group)) +
    stat_apply_group(.fun.x = min)
  result <- ggplot2::layer_data(result)
  result <- result$x
  expected <- c(min(tst.df$x[1:3]), min(tst.df$x[4:6]))
  expect_identical(result, expected)

})

library(ggplot2)
library(ggpp)

context("stat_apply_group")

test_that("stat_apply_group", {
  my.df.unsorted <- data.frame(X = rep(1:20,2),
                      Y = runif(40)*10,
                      category = rep(c("A","B"), each = 20))

  my.df <- my.df.unsorted[order(my.df.unsorted[["X"]]), ]

  result <- ggplot(my.df, aes(x = X, y = Y, colour = category)) +
    geom_point() +
    stat_apply_group(geom = "rug", .fun.y = quantile, .fun.x = quantile)
  result <- ggplot2::layer_data(result,2)
  result <- result[result$group == 1,]$x |> as.data.frame()
  expected <- quantile(my.df[my.df$category == "A",]$X)[1:5] |>
    as.data.frame()
  rownames(result) <- NULL
  rownames(expected) <- NULL
  colnames(result) <- NULL
  colnames(expected) <- NULL
  expect_identical(result, expected)

  result <- ggplot(my.df, aes(x = X, y = Y, colour = category)) +
    geom_point() +
    stat_apply_group(geom = "rug", .fun.y = quantile, .fun.x = quantile)
  result <- ggplot2::layer_data(result,2)
  result <- result[result$group == 1,]$y |> as.data.frame()
  expected <- quantile(my.df[my.df$category == "A",]$Y)[1:5] |>
    as.data.frame()
  rownames(result) <- NULL
  rownames(expected) <- NULL
  colnames(result) <- NULL
  colnames(expected) <- NULL
  expect_identical(result, expected)

  expect_silent(result <- ggplot(my.df, aes(x = X, y = Y)) +
    geom_point() +
    stat_apply_group(geom = "rug", sides = "lr", color = "darkred",
                     .fun.y = quantile) +
    stat_apply_group(geom = "text", hjust = "right", color = "darkred",
                     .fun.y = quantile,
                     .fun.x = function(x) {rep(22, 5)}, # set x to 22
                     mapping = aes(label = after_stat(y.names))) +
                     expand_limits(x = 21))
  expect_s3_class(result, "ggplot")

  my.probs <- c(0.25, 0.5, 0.75)
  result <- ggplot(my.df, aes(x = X, y = Y, colour = category)) +
    geom_point() +
    stat_apply_group(geom = "hline",
                     aes(yintercept = after_stat(y)),
                     .fun.y = quantile,
                     .fun.y.args = list(probs = my.probs))
  result <- ggplot2::layer_data(result,2)
  result <- result[result$group == 1,]$y |> as.data.frame()
  expected <- quantile(my.df[my.df$category == "A",]$Y, probs = my.probs) |>
    as.data.frame()
  rownames(result) <- NULL
  rownames(expected) <- NULL
  colnames(result) <- NULL
  colnames(expected) <- NULL
  expect_identical(result, expected)

  # cummulative summaries
  result <- ggplot(my.df.unsorted, aes(x = X, y = Y, colour = category)) +
    stat_apply_group(.fun.x = function(x) {x},
                     .fun.y = cummax)
  result <- ggplot2::layer_data(result)
  result <- result[result$group == 1,]$y
  expected <- cummax(my.df.unsorted[my.df.unsorted$category == "A",]$Y)
  expect_identical(result, expected)

  result <- ggplot(my.df, aes(x = X, y = Y, colour = category)) +
    stat_apply_group(.fun.x = cumsum, .fun.y = cumsum)
  result <- ggplot2::layer_data(result)
  result <- result[result$group == 1,]$x |> as.numeric()
  expected <- cumsum(my.df[my.df$category == "A",]$X) |> as.numeric()
  expect_identical(result, expected)
})

#' @title Label pairwise comparisons
#' @include ggp2-margins.R utilities.R ggpp-legend-draw.R
#'
#' @description Add a plot layer with a text label and a segment connecting two
#'   values along the \code{x} aesthetic. These are usually two levels of a
#'   factor mapped to the \code{x} aesthetic when used to report significance or
#'   highlighting pairwise comparisons.
#'
#' @section Under development!: This geometry is still under development
#'   and its user interface subject to change.
#'
#' @details Geometries \code{geom_text_pairwise()} and
#'   \code{geom_label_pairwise()} have an interface similar to that of
#'   \code{\link[ggplot2]{geom_text}} and \code{\link[ggplot2]{geom_label}}, but
#'   add a segment connecting two values along \code{x}. In the most
#'   frequent use case they add a segment connecting pairs of levels from a
#'   grouping factor mapped to the \emph{x} or \emph{y} aesthetic. They can
#'   also be used to label ranges of values.
#'
#'   The segment extends from \code{xmin} to \code{xmax}, and the text label is
#'   located at \code{x} with a default that positions the label at the centre
#'   of the bar. The ends of the bar can be terminated with arrow heads given
#'   by parameter \code{arrow}, with a default of a plain segment without
#'   arrow tips. The text label is located slightly above the segment by the
#'   default value of \code{vjust} in \code{geom_text_pairwise()} and on top
#'   of the segment in \code{geom_label_pairwise()}.
#'
#'   Layer functions \code{geom_text_pairwise()} and
#'   \code{geom_label_pairwise()} use by default
#'   \code{\link[ggplot2]{position_nudge}}. Nudging affects both text label and
#'   bar, and its default of no displacement will very rarely need to be
#'   changed.
#'
#'   Differently to \code{geom_text_repel()} and \code{geom_label_repel()},
#'   \code{geom_text_pairwise()} and \code{geom_label_pairwise()} do not make
#'   use of additional aesthetics for the segments or boxes, but instead allow
#'   the choice of which elements are targeted by the usual 'ggplot2' aesthetics
#'   and which are rendered using a default constant value. In the grammar of
#'   graphics using the same aesthetic with multiple meanings is not allowed,
#'   thus, the approach used in package 'ggpp' attempts to enforce this.
#'
#' @section Plot boundaries and clipping: Note that when you change the scale
#'   limits for \emph{x} and/or \emph{y} of a plot, text labels stay the same
#'   size, as determined by the \code{size} aesthetic, given in millimetres. The
#'   actual size as seen in the plotted output is decided during the rendering
#'   of the plot to a graphics device. Limits are expanded only to include the
#'   anchor point of the labels because the "width" and "height" of a text
#'   element are 0 (as seen by ggplot2). Text labels do have height and width,
#'   but in grid units, not data units. Either function
#'   \code{\link[ggplot2]{expand_limits}} or the scale expansion can be used to
#'   ensure text labels remain within the plotting area.
#'
#' @section Alignment: You can modify text alignment with the \code{vjust} and
#'   \code{hjust} aesthetics. These can either be a number between 0
#'   (right/bottom) and 1 (top/left) or a character (\code{"left"},
#'   \code{"middle"}, \code{"right"}, \code{"bottom"}, \code{"center"},
#'   \code{"top"}). Values outside the range 0..1 displace the text label so
#'   that the anchor point is outside the text label. In addition, you can use
#'   special alignments for justification including \code{"position"},
#'   \code{"inward"} and \code{"outward"}. Inward always aligns text towards the
#'   center of the plotting area, and outward aligns it away from the center of
#'   the plotting area. If tagged with \code{_mean} or \code{_median} (e.g.,
#'   \code{"outward_mean"}) the mean or median of the data in the panel along
#'   the corresponding axis is used as center. If the characters following the
#'   underscore represent a number (e.g., \code{"outward_10.5"}) the reference
#'   point will be this value in data units. Position justification is computed
#'   based on the direction of the displacement of the position of the label so
#'   that each individual text or label is justified outwards from its original
#'   position. The default justification is \code{"identity"}.
#'
#' @param mapping Set of aesthetic mappings created by
#'   \code{\link[ggplot2]{aes}}. With \code{inherit.aes = FALSE}
#'   (the default) it is not combined with the default mapping at the top level of
#'   the plot. You always need to supply a \code{mapping} unless you set
#'   \code{inherit.aes = TRUE}.
#' @param data A data frame. If specified, overrides the default data frame
#'   defined at the top level of the plot.
#' @param stat The statistical transformation to use on the data for this layer,
#'   as a string.
#' @param position Position adjustment, either as a string, or the result of a
#'   call to a position adjustment function.
#' @param orientation character The orientation of the layer. The default
#'   (\code{NA}) automatically determines the orientation from the aesthetic
#'   mapping. In the rare event that this fails it can be given explicitly by
#'   setting orientation to either "x" or "y".
#' @param parse If \code{TRUE}, the labels will be parsed into expressions and
#'   displayed as described in \code{?plotmath}.
#' @param na.rm If \code{FALSE} (the default), removes missing values with a
#'   warning.  If \code{TRUE} silently removes missing values.
#' @param show.legend logical. Should this layer be included in the legends?
#'   \code{NA} includes a legend if any aesthetics are mapped.
#'   \code{FALSE}, the default, never includes it, and \code{TRUE} always includes it.
#' @param inherit.aes If \code{FALSE}, overrides the default aesthetics, rather
#'   than combining them.
#' @param check_overlap If \code{TRUE}, text that overlaps previous text in the
#'   same layer will not be plotted. \code{check_overlap} takes place at draw
#'   time and in the order of the data, thus its action depends of the size at
#'   which the plot is drawn.
#' @param size.unit How the `size` aesthetic is interpreted: as millimetres
#'   (`"mm"`, default), points (`"pt"`), centimetres (`"cm"`), inches (`"in"`),
#'   or picas (`"pc"`).
#' @param ... other arguments passed on to \code{\link[ggplot2]{layer}}. There
#'   are three types of arguments you can use here:
#'
#'   \itemize{ \item Aesthetics: to set an aesthetic to a fixed value, like
#'   \code{colour = "red"} or \code{size = 3}. \item Other arguments to the
#'   layer, for example you override the default \code{stat} associated with the
#'   layer. \item Other arguments passed on to the stat. }
#' @param nudge_x,nudge_y Horizontal and vertical adjustments to nudge the
#'   starting position of each text label. The units for \code{nudge_x} and
#'   \code{nudge_y} are the same as for the data units on the x-axis and y-axis.
#' @param default.colour,default.color A colour definition to use for elements
#'   not targeted by the colour aesthetic.
#' @param colour.target,color.target A vector of character strings; \code{"all"},
#'   \code{"text"}, \code{"segment"}, \code{"box"}, \code{"box.line"}, and
#'   \code{"box.fill"} or \code{"none"}.
#' @param default.alpha numeric in [0..1] A transparency value to use for
#'   elements not targeted by the alpha aesthetic.
#' @param alpha.target A vector of character strings; \code{"all"},
#'   \code{"text"}, \code{"segment"}, \code{"box"}, \code{"box.line"}, and
#'   \code{"box.fill"} or \code{"none"}.
#' @param segment.linewidth numeric Width of the segments or arrows in mm.
#' @param arrow specification for arrow heads, as created by
#'   \code{\link[grid]{arrow}}
#'
#' @section Aesthetics: Layer functions \code{geom_text_pairwise()} and
#'   \code{geom_label_pairwise()} require aesthetics \code{xmin}, \code{xmax},
#'   \code{x}, \code{y} and
#'   \code{label} and support aesthetics: \code{alpha}, \code{colour},
#'   \code{group}, \code{size} (of text), \code{family}, \code{fontface},
#'   \code{linewidth}, \code{linetype}, \code{hjust} and \code{vjust}. In addition,
#'   \code{geom_text_pairwise} supports \code{angle} and \code{geom_label_pairwise} supports
#'   \code{fill}. See
#'   \code{\link[ggplot2]{aes_colour_fill_alpha}},
#'   \code{\link[ggplot2]{aes_linetype_size_shape}},
#'   \code{\link[ggplot2]{aes_position}}, and
#'   \code{\link[ggplot2]{aes_group_order}}.
#'
#'   In 'ggplot2' \code{linewidth} when applied to the border of the box drawn
#'   by \code{geom_label()} is given in points rather than in mm because of a
#'   historical error in the code. In other geometries such as
#'   \code{geom_segment()} \code{linewidth} is given in mm. As in
#'   \code{geom_label_pairwise()} it is important to remain consistent among
#'   different \code{linewidth} specifications, mm are used both for the box
#'   border and linking segment. To imitate the behaviour of \code{geom_label()}
#'   a correction factor of 0.75 (more exactly 1 pt = 0.7528 mm) can be used for
#'   the line width of the border of the box.
#'
#' @seealso \code{\link{geom_text_s}}, \code{\link{geom_label_s}},
#'   \code{\link[ggplot2]{geom_text}}, \code{\link[ggplot2]{geom_label}} and
#'   other documentation of package 'ggplot2'.
#'
#' @return A plot layer instance.
#'
#' @export
#'
#' @examples
#'
#' my.cars <- mtcars
#' my.cars$name <- rownames(my.cars)
#' p1 <- ggplot(my.cars, aes(x = factor(cyl), y = mpg)) +
#'        geom_boxplot(width = 0.33)
#'
#' # With a factor mapped to x, highlight pairs
#'
#' my.pairs <-
#'   data.frame(A = 1:2, B = 2:3, bar.height = c(12, 30),
#'              p.value = c(0.01, 0.05678))
#' p1 +
#'   geom_text_pairwise(data = my.pairs,
#'                      aes(xmin = A, xmax = B,
#'                          y = bar.height,
#'                          label = p.value),
#'                      parse = TRUE)
#'
#' p1 +
#'   geom_text_pairwise(data = my.pairs,
#'                      aes(xmin = A, xmax = B,
#'                          y = bar.height,
#'                          label = sprintf("italic(P)~`=`~%.2f", p.value)),
#'                      arrow = grid::arrow(angle = 90,
#'                                          length = unit(1, "mm"),
#'                                          ends = "both"),
#'                      parse = TRUE)
#'
#' p1 +
#'   geom_text_pairwise(data = my.pairs,
#'                      aes(xmin = A, xmax = B,
#'                          y = bar.height,
#'                          label = sprintf("italic(P)~`=`~%.2f", p.value)),
#'                      colour = "red",
#'                      arrow = grid::arrow(angle = 90,
#'                                          length = unit(1, "mm"),
#'                                          ends = "both"),
#'                      parse = TRUE)
#'
#' p1 +
#'   geom_label_pairwise(data = my.pairs,
#'                       aes(xmin = A, xmax = B,
#'                           y = bar.height,
#'                           label = sprintf("italic(P)~`=`~%.2f", p.value)),
#'                       colour = "red", size = 2.75,
#'                       arrow = grid::arrow(angle = 30,
#'                                           length = unit(1.5, "mm"),
#'                                           ends = "both"),
#'                       parse = TRUE)
#'
#' p1 +
#'   geom_text_pairwise(data = my.pairs,
#'                      aes(xmin = A, xmax = B,
#'                          y = bar.height,
#'                          label = sprintf("italic(P)~`=`~%.2f", p.value)),
#'                      colour = "red", colour.target = "segment",
#'                      arrow = grid::arrow(angle = 90,
#'                                          length = unit(1, "mm"),
#'                                          ends = "both"),
#'                      parse = TRUE)
#'
#' p1 +
#'   geom_text_pairwise(data = my.pairs,
#'                      aes(xmin = A, xmax = B,
#'                          y = bar.height,
#'                          label = sprintf("italic(P)~`=`~%.2f", p.value)),
#'                      colour = "red", colour.target = "text",
#'                      arrow = grid::arrow(angle = 90,
#'                                          length = unit(1, "mm"),
#'                                          ends = "both"),
#'                      parse = TRUE)
#'
#' # with a numeric vector mapped to x, indicate range
#'
#' p2 <-
#'   ggplot(my.cars, aes(disp, mpg)) +
#'     geom_point()
#'
#' my.ranges <-
#'   data.frame(A = c(50, 400),
#'              B = c(200, 500),
#'              bar.height = 5,
#'              text = c("small", "large"))
#'
#' p2 +
#'   geom_text_pairwise(data = my.ranges,
#'                      aes(xmin = A, xmax = B,
#'                      y = bar.height, label = text))
#'
#' p2 +
#'   geom_text_pairwise(data = my.ranges,
#'                      aes(xmin = A, xmax = B,
#'                          y = bar.height, label = text),
#'                      angle = 90, hjust = -0.1)
#'
#' p2 +
#'   geom_label_pairwise(data = my.ranges,
#'                       aes(xmin = A, xmax = B,
#'                           y = bar.height, label = text),
#'                      angle = 90, hjust = -0.1)
#'
#' p2 +
#'   geom_label_pairwise(data = my.ranges,
#'                       aes(xmin = A, xmax = B,
#'                           y = bar.height, label = text))
#'
#' p2 +
#'   geom_text_pairwise(data = my.ranges,
#'                      aes(xmin = A, xmax = B,
#'                          y = bar.height, label = text),
#'                      arrow = grid::arrow(ends = "both", length = unit(2, "mm")))
#'
#' # implicit orientation = "y"
#' p3 <- ggplot(my.cars, aes(x = mpg, y = factor(cyl))) +
#'        geom_boxplot(width = 0.33)
#'
#' p3 +
#'   geom_text_pairwise(data = my.pairs,
#'                      aes(ymin = A, ymax = B,
#'                          x = bar.height,
#'                          label = p.value))
#'
#' p3 +
#'   geom_label_pairwise(data = my.pairs,
#'                       aes(ymin = A, ymax = B,
#'                           x = bar.height,
#'                           label = p.value))
#'
geom_text_pairwise <- function(mapping = NULL,
                               data = NULL,
                               stat = "identity",
                               position = "identity",
                               ...,
                               orientation = NA,
                               parse = FALSE,
                               nudge_x = 0,
                               nudge_y = 0,
                               default.colour = "black",
                               default.color = default.colour,
                               colour.target = "all",
                               color.target = colour.target,
                               default.alpha = NA,
                               alpha.target = "all",
                               segment.linewidth = 0.5,
                               arrow = NULL,
                               check_overlap = FALSE,
                               size.unit = "mm",
                               na.rm = FALSE,
                               show.legend = NA,
                               inherit.aes = FALSE) {

  colour.target <-
    rlang::arg_match(color.target,
                     values = c("all", "text", "segment", "none"),
                     multiple = TRUE)
  alpha.target <-
    rlang::arg_match(alpha.target,
                     values = c("all", "text", "segment", "none"),
                     multiple = TRUE)


  if (!missing(nudge_x) || !missing(nudge_y)) {
    if (!missing(position) && position != "identity") {
      rlang::abort("You must specify either `position` or `nudge_x`/`nudge_y`.")
    }
    # original position needed for "position" justification
    position <-
      ggplot2::position_nudge(nudge_x, nudge_y)
  }

  ggplot2::layer(
    data = data,
    mapping = mapping,
    stat = stat,
    geom = GeomTextPairwise,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      orientation = orientation,
      parse = parse,
      default.colour = default.color,
      colour.target = colour.target,
      default.alpha = default.alpha,
      alpha.target = alpha.target,
      segment.linewidth = segment.linewidth,
      arrow = arrow,
      check_overlap = check_overlap,
      size.unit = size.unit,
      na.rm = na.rm,
      ...
    )
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
GeomTextPairwise <-
  ggplot2::ggproto("GeomTextPairwise", ggplot2::Geom,
                   required_aes = c("xmin|ymin", "xmax|ymax", "y|x", "label"),

                   default_aes = ggplot2::aes(
                     x = NA_real_,
                     y = NA_real_,
                     colour = "black",
                     size = 3.88,
                     angle = 0,
                     hjust = 0.5,
                     vjust = -0.5,
                     alpha = NA,
                     family = "",
                     fontface = 1,
                     lineheight = 1.2
                   ),

                   draw_panel = function(data,
                                         panel_params,
                                         coord, #panel_scales,
                                         parse = FALSE,
                                         size.unit = "mm",
                                         default.colour = "black",
                                         colour.target = "all",
                                         default.alpha = NA,
                                         alpha.target = "all",
                                         na.rm = FALSE,
                                         check_overlap = FALSE,
                                         segment.linewidth = 0.5,
                                         arrow = NULL,
                                         orientation = NA) {

                     if (is.null(orientation) || is.na(orientation)) {
                       if (all(c("xmin", "xmax") %in% colnames(data)) &&
                           !(all(is.na(data[["xmin"]])) || all(is.na(data[["xmax"]])))) {
                         orientation <- "x"
                       } else if (all(c("ymin", "ymax") %in% colnames(data)) &&
                                  !(all(is.na(data[["ymin"]])) || all(is.na(data[["ymax"]])))) {
                         orientation <- "y"
                       } else {
                         stop("Unable to guess 'orientation' from mapped aesthetics!")
                       }
                     }
                     data$label <- as.character(data$label)
                     data <- subset(data, !is.na(label) & label != "")
                     if (nrow(data) == 0L) {
                       return(nullGrob())
                     }

                     lab <- data$label
                     if (parse) {
                       lab <- parse_safe(lab)
                     }

                     data <- coord$transform(data, panel_params)

                     if (is.character(data$vjust)) {
                       data$vjust <-
                         compute_just2d(data = data,
                                        coord = coord,
                                        panel_params = panel_params,
                                        just = data$vjust,
                                        a = "y", b = "x")
                     }
                     if (is.character(data$hjust)) {
                       data$hjust <-
                         compute_just2d(data = data,
                                        coord = coord,
                                        panel_params = panel_params,
                                        just = data$hjust,
                                        a = "x", b = "y")
                     }

                     size.unit <- resolve_text_unit(size.unit)

                     if (orientation == "x") {
                       # loop needed as gpar is not vectorized
                       all.grobs <- grid::gList()

                       for (row.idx in 1:nrow(data)) {
                         row <- data[row.idx, , drop = FALSE]
                         if (is.na(row$x)) {
                           row$x <- (row$xmin + row$xmax) / 2
                         }
                         text.alpha <-
                           ifelse(any(alpha.target %in% c("all", "text")),
                                  row$alpha, default.alpha)
                         segment.alpha <-
                           ifelse(any(alpha.target %in% c("all", "segment")),
                                  row$alpha, default.alpha)
                         user.grob <- grid::textGrob(
                           lab[row.idx],
                           x = row$x, y = row$y, default.units = "native",
                           hjust = row$hjust, vjust = row$vjust,
                           rot = row$angle,
                           gp = gpar(
                             col = ifelse(any(colour.target %in% c("all", "text")),
                                          ggplot2::alpha(row$colour, text.alpha),
                                          ggplot2::alpha(default.colour, text.alpha)),
                             fontsize = row$size * size.unit,
                             fontfamily = row$family,
                             fontface = row$fontface,
                             lineheight = row$lineheight
                           ),
                           check.overlap = check_overlap
                         )

                         # give unique name to each grob
                         user.grob$name <-
                           paste("text.pairwise.grob", row$group, row.idx, sep = ".")

                         segment.grob <-
                           grid::segmentsGrob(x0 = row$xmin,
                                              y0 = row$y,
                                              x1 = row$xmax,
                                              y1 = row$y,
                                              arrow = arrow,
                                              gp = grid::gpar(
                                                col = if (segment.linewidth == 0) NA else # lwd = 0 is invalid in 'grid'
                                                  ifelse(any(colour.target %in% c("all", "segment")),
                                                         ggplot2::alpha(row$colour, segment.alpha),
                                                         ggplot2::alpha(default.colour, segment.alpha)),
                                                lwd = (if (segment.linewidth == 0) 0.5 else segment.linewidth) * ggplot2::.stroke),
                                              name = paste("text.pr.segment", row$group, row.idx, sep = ".")
                           )
                         all.grobs <- grid::gList(all.grobs, segment.grob, user.grob)
                       }
                     } else {
                       # loop needed as gpar is not vectorized
                       all.grobs <- grid::gList()

                       for (row.idx in 1:nrow(data)) {
                         row <- data[row.idx, , drop = FALSE]
                         if (is.na(row$y)) {
                           row$y <- (row$ymin + row$ymax) / 2
                         }
                         text.alpha <-
                           ifelse(any(alpha.target %in% c("all", "text")),
                                  row$alpha, default.alpha)
                         segment.alpha <-
                           ifelse(any(alpha.target %in% c("all", "segment")),
                                  row$alpha, default.alpha)
                         user.grob <- grid::textGrob(
                           lab[row.idx],
                           y = row$y, x = row$x, default.units = "native",
                           hjust = row$hjust, vjust = row$vjust,
                           rot = row$angle + 90,
                           gp = gpar(
                             col = ifelse(any(colour.target %in% c("all", "text")),
                                          ggplot2::alpha(row$colour, text.alpha),
                                          ggplot2::alpha(default.colour, text.alpha)),
                             fontsize = row$size * size.unit,
                             fontfamily = row$family,
                             fontface = row$fontface,
                             lineheight = row$lineheight
                           ),
                           check.overlap = check_overlap
                         )

                         # give unique name to each grob
                         user.grob$name <-
                           paste("text.pairwise.grob", row$group, row.idx, sep = ".")

                         segment.grob <-
                           grid::segmentsGrob(x0 = row$x,
                                              y0 = row$ymin,
                                              x1 = row$x,
                                              y1 = row$ymax,
                                              arrow = arrow,
                                              gp = grid::gpar(
                                                col = if (segment.linewidth == 0) NA else # lwd = 0 is invalid in 'grid'
                                                  ifelse(any(colour.target %in% c("all", "segment")),
                                                         ggplot2::alpha(row$colour, segment.alpha),
                                                         ggplot2::alpha(default.colour, segment.alpha)),
                                                lwd = (if (segment.linewidth == 0) 0.5 else segment.linewidth) * ggplot2::.stroke),
                                              name = paste("text.pr.segment", row$group, row.idx, sep = ".")
                           )
                         all.grobs <- grid::gList(all.grobs, segment.grob, user.grob)
                       }
                     }

                     # name needs to be unique within plot, so we would need to know other layers
                     #                     grid::grobTree(children = all.grobs, name = "geom.text.pairwise.panel")
                     grid::grobTree(children = all.grobs)

                   },

                   draw_key = draw_key_text_s
  )

context("geom_text_pairwise")

test_that("text_pairwise_default", {
  df <- data.frame(
    x.min = c(1, 3, 5),
    x.max = c(2, 4, 7),
    y = c(2, 4, 3),
    text = c("1-2", "3-4", "5-7")
  )

  vdiffr::expect_doppelganger("geom_text_pairwise_inherit.aes",
                              ggplot() +
                                geom_text_pairwise(data = df,
                                                   mapping = aes(xmin = x.min,
                                                                 xmax = x.max,
                                                                 y = y,
                                                                 label = text))
  )

  vdiffr::expect_doppelganger("geom_text_pairwise_default",
                              ggplot(df,
                                     aes(xmin = x.min,
                                         xmax = x.max,
                                         y = y,
                                         label = text)) +
                                geom_text_pairwise(inherit.aes = TRUE)
  )

  vdiffr::expect_doppelganger("geom_text_pairwise_hjust",
                              ggplot(df,
                                     aes(xmin = x.min,
                                         xmax = x.max,
                                         x = x.min,
                                         y = y,
                                         label = text)) +
                                geom_text_pairwise(inherit.aes = TRUE,
                                                   hjust = 0)
  )

  vdiffr::expect_doppelganger("geom_text_pairwise_below",
                              ggplot(df,
                                     aes(xmin = x.min,
                                         xmax = x.max,
                                         y = y,
                                         label = text)) +
                                geom_text_pairwise(inherit.aes = TRUE,
                                                   vjust = 1.1)
  )

  vdiffr::expect_doppelganger("geom_text_pairwise_arrow",
                              ggplot(df,
                                     aes(xmin = x.min,
                                         xmax = x.max,
                                         y = y,
                                         label = text)) +
                                geom_text_pairwise(inherit.aes = TRUE,
                                                   arrow = grid::arrow(ends = "both"))
  )

    vdiffr::expect_doppelganger("geom_text_pairwise_angle",
                              ggplot(df,
                                     aes(xmin = x.min,
                                         xmax = x.max,
                                         y = y,
                                         label = text)) +
                                geom_text_pairwise(inherit.aes = TRUE,
                                                   hjust = -0.2,
                                                   vjust = 0.5,
                                                   angle = 90)
  )

})

test_that("text_pairwise_segment", {
# segment.linewidth is a parameter, not an aesthetic
  df <- data.frame(
    x.min = c(1, 3, 5),
    x.max = c(2, 4, 7),
    y = c(2, 4, 3),
    text = c("1-2", "3-4", "5-7")
  )

  vdiffr::expect_doppelganger("geom_text_pairwise_segment.linewidth",
                              ggplot(df,
                                     aes(xmin = x.min,
                                         xmax = x.max,
                                         y = y,
                                         label = text)) +
                                geom_text_pairwise(inherit.aes = TRUE,
                                                   segment.linewidth = 1)
  )

  vdiffr::expect_doppelganger("geom_text_pairwise_colour",
                              ggplot(df,
                                     aes(xmin = x.min,
                                         xmax = x.max,
                                         y = y,
                                         label = text)) +
                                geom_text_pairwise(inherit.aes = TRUE,
                                                   colour = "red")
  )

  vdiffr::expect_doppelganger("geom_text_pairwise_color",
                              ggplot(df,
                                     aes(xmin = x.min,
                                         xmax = x.max,
                                         y = y,
                                         label = text)) +
                                geom_text_pairwise(inherit.aes = TRUE,
                                                   color = "red")
  )

  vdiffr::expect_doppelganger("geom_text_pairwise_color.target.all",
                              ggplot(df,
                                     aes(xmin = x.min,
                                         xmax = x.max,
                                         y = y,
                                         label = text)) +
                                geom_text_pairwise(inherit.aes = TRUE,
                                                   color.target = "all",
                                                   color = "red")
  )

  vdiffr::expect_doppelganger("geom_text_pairwise_color.target.text",
                              ggplot(df,
                                     aes(xmin = x.min,
                                         xmax = x.max,
                                         y = y,
                                         label = text)) +
                                geom_text_pairwise(inherit.aes = TRUE,
                                                   color.target = "text",
                                                   color = "red")
  )

  vdiffr::expect_doppelganger("geom_text_pairwise_color.target.segment",
                              ggplot(df,
                                     aes(xmin = x.min,
                                         xmax = x.max,
                                         y = y,
                                         label = text)) +
                                geom_text_pairwise(inherit.aes = TRUE,
                                                   color.target = "segment",
                                                   color = "red")
  )

}
)

test_that("text_pairwise_position", {
  df <- data.frame(
    x.min = c(1, 3, 5),
    x.max = c(2, 4, 7),
    y = c(2, 4, 3),
    text = c("1-2", "3-4", "5-7")
  )

  vdiffr::expect_doppelganger("geom_text_pairwise_nudge_x",
                              ggplot(df,
                                     aes(xmin = x.min,
                                         xmax = x.max,
                                         y = y,
                                         label = text)) +
                                geom_text_pairwise(inherit.aes = TRUE,
                                                   nudge_x = 0.5)
  )

  vdiffr::expect_doppelganger("geom_text_pairwise_nudge_y",
                              ggplot(df,
                                     aes(xmin = x.min,
                                         xmax = x.max,
                                         y = y,
                                         label = text)) +
                                geom_text_pairwise(inherit.aes = TRUE,
                                                   nudge_y = 0.5)
  )

  expect_error(ggplot(df,
                      aes(xmin = x.min,
                          xmax = x.max,
                          y = y,
                          label = text)) +
                 geom_text_pairwise(inherit.aes = TRUE,
                                    nudge_x = 0.5,
                                    position = position_nudge(x = 0.5, y = 0))
  )

})

#' @title Filter observations by local 1D density
#'
#' @description \code{stat_dens1d_filter} Filters-out/filters-in observations in
#'   regions of a plot panel with high density of observations, based on the
#'   values mapped to one of \code{x} and \code{y} aesthetics.
#'   \code{stat_dens1d_filter_g} does the same filtering by group instead of by
#'   panel. This second stat is useful for highlighting observations, while the
#'   first one tends to be most useful when the aim is to prevent clashes among
#'   text labels. By default the data are handled all together, but it is also
#'   possible to control labeling separately in each tail.
#'
#' @details The 1D density of observations of \emph{x} or \emph{y} is computed
#'   with function \code{\link[stats]{density}} and used to select observations,
#'   passing to the geom a subset of the rows in its \code{data} input. The
#'   default is to select observations in sparse regions of the plot, but the
#'   selection can be inverted so that only observations in the densest regions
#'   are returned. Specific observations can be protected from being deselected
#'   and "kept" by passing a suitable argument to \code{keep.these}. Logical and
#'   integer vectors work as indexes to rows in \code{data}, while a values in a
#'   character vector are compared to the character values mapped to the
#'   \code{label} aesthetic. A function passed as argument to keep.these will
#'   receive as argument the values in the variable mapped to \code{label} and
#'   should return a character, logical or numeric vector as described above. If
#'   no variable has been mapped to \code{label}, row names are used in its
#'   place.
#'
#'   How many rows are retained in addition to those in \code{keep.these} is
#'   controlled with arguments passed to \code{keep.number} and
#'   \code{keep.fraction}. \code{keep.number} sets the maximum number of
#'   observations selected, whenever \code{keep.fraction} results in fewer
#'   observations selected, it is obeyed. If `xintercept` is a finite value
#'   within the \emph{x} range of the data and \code{pool.along}
#'   is passed \code{"none"} the data as are split into two groups
#'   and \code{keep.number} and \code{keep.fraction} are applied separately to
#'   each tail with density still computed jointly from all observations. If the
#'   length of \code{keep.number} and \code{keep.fraction} is one, this value
#'   is used for both tails, if their length is two, the first value is use
#'   for the left tail and the second value for the right tail.
#'
#'   Computation of density and of the default bandwidth require at least
#'   two observations with different values. If data do not fulfill this
#'   condition, they are kept only if \code{keep.fraction = 1}. This is correct
#'   behavior for a single observation, but can be surprising in the case of
#'   multiple observations.
#'
#'   Parameters \code{keep.these} and \code{exclude.these} make it possible to
#'   force inclusion or exclusion of observations after the density is computed.
#'   In case of conflict, \code{exclude.these} overrides \code{keep.these}.
#'
#' @note Which points are kept and which not depends on how dense and flexible
#'   is the density curve estimate. This depends on the values passed as
#'   arguments to parameters \code{n}, \code{bw} and \code{kernel}. It is
#'   also important to be aware that both \code{geom_text()} and
#'   \code{geom_text_repel()} can avoid over plotting by discarding labels at
#'   the plot rendering stage, i.e., what is plotted may differ from what is
#'   returned by this statistic.
#'
#' @param mapping The aesthetic mapping, usually constructed with
#'   \code{\link[ggplot2]{aes}} or \code{\link[ggplot2]{aes_}}. Only needs to be
#'   set at the layer level if you are overriding the plot defaults.
#' @param data A layer specific dataset - only needed if you want to override
#'   the plot defaults.
#' @param geom The geometric object to use display the data.
#' @param keep.fraction numeric vector of length 1 or 2 [0..1]. The fraction of
#'   the observations (or rows) in \code{data} to be retained.
#' @param keep.number integer vector of length 1 or 2. Set the maximum number of
#'   observations to retain, effective only if obeying \code{keep.fraction}
#'   would result in a larger number.
#' @param keep.sparse logical If \code{TRUE}, the default, observations from the
#'   more sparse regions are retained, if \code{FALSE} those from the densest
#'   regions.
#' @param keep.these,exclude.these character vector, integer vector, logical
#'   vector or function that takes one or more variables in data selected by
#'   \code{these.target}. Negative integers behave as in R's extraction methods.
#'   The rows from \code{data} indicated by \code{keep.these} and
#'   \code{exclude.these} are kept or excluded irrespective of the local
#'   density.
#' @param these.target character, numeric or logical selecting one or more
#'   column(s) of \code{data}. If \code{TRUE} the whole \code{data} object is
#'   passed.
#' @param pool.along character, one of \code{"none"} or \code{"x"},
#'   indicating if selection should be done pooling the observations along the
#'   \emph{x} aesthetic, or separately on either side of \code{xintercept}.
#' @param xintercept numeric The split point for the data filtering. If
#'   \code{NA} the data are not split.
#' @param invert.selection logical If \code{TRUE}, the complement of the
#'   selected rows are returned.
#' @param bw numeric or character The smoothing bandwidth to be used. If
#'   numeric, the standard deviation of the smoothing kernel. If character, a
#'   rule to choose the bandwidth, as listed in \code{\link[stats]{bw.nrd}}.
#' @param adjust numeric A multiplicative bandwidth adjustment. This makes it
#'   possible to adjust the bandwidth while still using the a bandwidth
#'   estimator through an argument passed to \code{bw}. The larger the value
#'   passed to \code{adjust} the stronger the smoothing, hence decreasing
#'   sensitivity to local changes in density.
#' @param kernel character See \code{\link{density}} for details.
#' @param n numeric Number of equally spaced points at which the density is to
#'   be estimated for applying the cut point. See \code{\link{density}} for
#'   details.
#' @param return.density logical vector of lenght 1. If \code{TRUE} add columns
#'   \code{"density"} and \code{"keep.obs"} to the returned data frame.
#' @param orientation	character The aesthetic along which density is computed.
#'   Given explicitly by setting orientation to either \code{"x"} or \code{"y"}.
#' @param position The position adjustment to use for overlapping points on this
#'   layer
#' @param show.legend logical. Should this layer be included in the legends?
#'   \code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
#'   never includes, and \code{TRUE} always includes.
#' @param inherit.aes If \code{FALSE}, overrides the default aesthetics, rather
#'   than combining with them. This is most useful for helper functions that
#'   define both data and aesthetics and shouldn't inherit behaviour from the
#'   default plot specification, e.g. \code{\link[ggplot2]{borders}}.
#' @param ... other arguments passed on to \code{\link[ggplot2]{layer}}. This
#'   can include aesthetics whose values you want to set, not map. See
#'   \code{\link[ggplot2]{layer}} for more details.
#' @param na.rm	a logical value indicating whether \code{NA} values should be
#'   stripped before the computation proceeds.
#'
#' @return A plot layer instance. Using as output \code{data} a subset of the
#'   rows in input \code{data} retained based on a 1D filtering criterion.
#'
#' @seealso \code{\link[stats]{density}} used internally.
#'
#' @family statistics returning a subset of data
#'
#' @examples
#'
#' random_string <-
#'   function(len = 6) {
#'     paste(sample(letters, len, replace = TRUE), collapse = "")
#'   }
#'
#' # Make random data.
#' set.seed(1001)
#' d <- tibble::tibble(
#'   x = rnorm(100),
#'   y = rnorm(100),
#'   group = rep(c("A", "B"), c(50, 50)),
#'   lab = replicate(100, { random_string() })
#' )
#' d$xg <- d$x
#' d$xg[51:100] <- d$xg[51:100] + 1
#'
#' # highlight the 1/10 of observations in sparsest regions of the plot
#' ggplot(data = d, aes(x, y)) +
#'   geom_point() +
#'   geom_rug(sides = "b") +
#'   stat_dens1d_filter(colour = "red") +
#'   stat_dens1d_filter(geom = "rug", colour = "red", sides = "b")
#'
#' # highlight the 1/4 of observations in densest regions of the plot
#' ggplot(data = d, aes(x, y)) +
#'   geom_point() +
#'   geom_rug(sides = "b") +
#'   stat_dens1d_filter(colour = "blue",
#'                      keep.fraction = 1/4, keep.sparse = FALSE) +
#'   stat_dens1d_filter(geom = "rug", colour = "blue",
#'                      keep.fraction = 1/4, keep.sparse = FALSE,
#'                      sides = "b")
#'
#' # switching axes
#' ggplot(data = d, aes(x, y)) +
#'   geom_point() +
#'   geom_rug(sides = "l") +
#'   stat_dens1d_filter(colour = "red", orientation = "y") +
#'   stat_dens1d_filter(geom = "rug", colour = "red", orientation = "y",
#'                      sides = "l")
#'
#' # highlight 1/10 plus 1/10 observations in high and low density regions
#' ggplot(data = d, aes(x, y)) +
#'   geom_point() +
#'   geom_rug(sides = "b") +
#'   stat_dens1d_filter(colour = "red") +
#'   stat_dens1d_filter(geom = "rug", colour = "red", sides = "b") +
#'   stat_dens1d_filter(colour = "blue", keep.sparse = FALSE) +
#'   stat_dens1d_filter(geom = "rug",
#'                      colour = "blue", keep.sparse = FALSE, sides = "b")
#'
#' # selecting the 1/10 observations in sparsest regions and their complement
#' ggplot(data = d, aes(x, y)) +
#'   stat_dens1d_filter(colour = "red") +
#'   stat_dens1d_filter(geom = "rug", colour = "red", sides = "b") +
#'   stat_dens1d_filter(colour = "blue", invert.selection = TRUE) +
#'   stat_dens1d_filter(geom = "rug",
#'                      colour = "blue", invert.selection = TRUE, sides = "b")
#'
#' # density filtering done jointly across groups
#' ggplot(data = d, aes(xg, y, colour = group)) +
#'   geom_point() +
#'   geom_rug(sides = "b", colour = "black") +
#'   stat_dens1d_filter(shape = 1, size = 3, keep.fraction = 1/4, adjust = 2)
#'
#' # density filtering done independently for each group
#' ggplot(data = d, aes(xg, y, colour = group)) +
#'   geom_point() +
#'   geom_rug(sides = "b") +
#'   stat_dens1d_filter_g(shape = 1, size = 3, keep.fraction = 1/4, adjust = 2)
#'
#' # density filtering done jointly across groups by overriding grouping
#' ggplot(data = d, aes(xg, y, colour = group)) +
#'   geom_point() +
#'   geom_rug(sides = "b") +
#'   stat_dens1d_filter_g(colour = "black",
#'                        shape = 1, size = 3, keep.fraction = 1/4, adjust = 2)
#'
#' # label observations
#' ggplot(data = d, aes(x, y, label = lab, colour = group)) +
#'   geom_point() +
#'   stat_dens1d_filter(geom = "text", hjust = "outward")
#'
#' # looking under the hood with gginnards::geom_debug_group()
#' gginnards.installed <- requireNamespace("gginnards", quietly = TRUE)
#' if (gginnards.installed) {
#'   library(gginnards)
#'
#'   ggplot(data = d, aes(x, y, label = lab, colour = group)) +
#'     stat_dens1d_filter(geom = "debug_group")
#' }
#'
#' if (gginnards.installed) {
#'   ggplot(data = d, aes(x, y, label = lab, colour = group)) +
#'     stat_dens1d_filter(geom = "debug_group", return.density = TRUE)
#' }
#' @export
#'
stat_dens1d_filter <-
  function(mapping = NULL,
           data = NULL,
           geom = "point",
           position = "identity",
           ...,
           keep.fraction = 0.10,
           keep.number = Inf,
           keep.sparse = TRUE,
           keep.these = FALSE,
           exclude.these = FALSE,
           these.target = "label",
           pool.along = c("x", "none"),
           xintercept = 0,
           invert.selection = FALSE,
           bw = "SJ",
           kernel = "gaussian",
           adjust = 1,
           n = 512,
           return.density = FALSE,
           orientation = c("x", "y"),
           na.rm = TRUE,
           show.legend = FALSE,
           inherit.aes = TRUE) {

    pool.along <- rlang::arg_match(pool.along)
    orientation <- rlang::arg_match(orientation)

    if (any(is.na(keep.fraction) | keep.fraction < 0 | keep.fraction > 1)) {
      stop("Out of range or missing value for 'keep.fraction': ", keep.fraction)
    }
    if (any(is.na(keep.number) | keep.number < 0)) {
      stop("Out of range or missing value for 'keep.number': ", keep.number)
    }

    ggplot2::layer(
      stat = StatDens1dFilter, data = data, mapping = mapping, geom = geom,
      position = position, show.legend = show.legend, inherit.aes = inherit.aes,
      params = list(na.rm = na.rm,
                    keep.fraction = keep.fraction,
                    keep.number = keep.number,
                    keep.sparse = keep.sparse,
                    keep.these = keep.these,
                    exclude.these = exclude.these,
                    these.target = these.target,
                    pool.along = pool.along,
                    xintercept = xintercept,
                    invert.selection = invert.selection,
                    bw = bw,
                    adjust = adjust,
                    kernel = kernel,
                    n = n,
                    return.density = return.density,
                    orientation = orientation,
                    ...)
    )
  }

#' @rdname stat_dens1d_filter
#'
#' @export
#'
stat_dens1d_filter_g <-
  function(mapping = NULL, data = NULL,
           geom = "point", position = "identity",
           keep.fraction = 0.10,
           keep.number = Inf,
           keep.sparse = TRUE,
           keep.these = FALSE,
           exclude.these = FALSE,
           these.target = "label",
           pool.along = c("x", "none"),
           xintercept = 0,
           invert.selection = FALSE,
           na.rm = TRUE, show.legend = FALSE,
           inherit.aes = TRUE,
           bw = "SJ",
           adjust = 1,
           kernel = "gaussian",
           n = 512,
           return.density = FALSE,
           orientation = c("x", "y"),
           ...) {

    pool.along <- rlang::arg_match(pool.along)
    orientation <- rlang::arg_match(orientation)

    if (any(is.na(keep.fraction) | keep.fraction < 0 | keep.fraction > 1)) {
      stop("Out of range or missing value for 'keep.fraction': ", keep.fraction)
    }
    if (any(is.na(keep.number) | keep.number < 0)) {
      stop("Out of range or missing value for 'keep.number': ", keep.number)
    }

    ggplot2::layer(
      stat = StatDens1dFilterG, data = data, mapping = mapping, geom = geom,
      position = position, show.legend = show.legend, inherit.aes = inherit.aes,
      params = list(na.rm = na.rm,
                    keep.fraction = keep.fraction,
                    keep.number = keep.number,
                    keep.sparse = keep.sparse,
                    keep.these = keep.these,
                    exclude.these = exclude.these,
                    these.target = these.target,
                    pool.along = pool.along,
                    xintercept = xintercept,
                    invert.selection = invert.selection,
                    bw = bw,
                    kernel = kernel,
                    adjust = adjust,
                    n = n,
                    return.density = return.density,
                    orientation = orientation,
                    ...)
    )
  }

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
StatDens1dFilter <-
  ggplot2::ggproto(
    "StatDens1dFilter",
    ggplot2::Stat,
    compute_panel =
#      dens1d_flt_compute_fun,
# code duplicated below to ensure it is all seen by 'covr'
  function(data,
           scales,
           keep.fraction,
           keep.number,
           keep.sparse,
           keep.these,
           exclude.these,
           these.target,
           pool.along,
           xintercept,
           invert.selection,
           bw,
           kernel,
           adjust,
           n,
           orientation,
           return.density) {

    force(data)

    keep.these <- these2logical(these = keep.these,
                                data = data,
                                these.target = these.target)

    exclude.these <- these2logical(these = exclude.these,
                                   data = data,
                                   these.target = these.target)

    # discard redundant splits and make list of logical vectors
    if (pool.along != "x" &&
        xintercept < max(data[[orientation]]) &&
        xintercept > min(data[[orientation]])) {

      selectors <-list(low.tail = data[[orientation]] <= xintercept,
                       high.tail = data[[orientation]] > xintercept)
      if (length(keep.fraction) != 2L) {
        keep.fraction <- rep_len(keep.fraction, length.out = 2)
      }
      if (length(keep.number) != 2L) {
        if (length(keep.number) == 1L) {
          keep.number <- keep.number %/% 2
        }
        keep.number <- rep_len(keep.number, length.out = 2)
      }
      num.rows <- sapply(selectors, sum) # selectors are logical
    } else {
      keep.fraction <- keep.fraction[[1]] # can be a vector or a list
      keep.number <- keep.number[[1]]
      num.rows <- nrow(data)
      selectors <- list(all = rep.int(TRUE, times = num.rows))
    }

    # vectorized
    too.large.frac <- num.rows * keep.fraction > keep.number
    keep.fraction[too.large.frac] <-
      keep.number[too.large.frac] / num.rows[too.large.frac]

    # density on a grid
    # data with fewer than 2 rows is as a special case as density() fails
    if (length(unique(data[[orientation]])) >= 2L) {
      dens <-
        stats::density(data[[orientation]],
                       bw = bw, kernel = kernel, adjust = adjust, n = n,
                       from = scales[[orientation]]$dimension()[1],
                       to = scales[[orientation]]$dimension()[2])

      # estimate density at each observations coordinates
      fdens <- stats::splinefun(dens$x, dens$y) # y contains estimate of density
      dens <- fdens(data[[orientation]])
    } else {
      if (nrow(data) > 1L) {
        message("Density not computed, too few distinct values in '", orientation, "'")
      }
      dens <- rep_len(1, nrow(data))
    }

    # we construct one logical vector by adding observations/label to be kept
    # we may have a list of 1 or 2 logical vectors
    keep <- logical(nrow(data))
    for (i in seq_along(selectors)) {
      if (keep.fraction[i] == 1 ||
          (length(selectors[[i]]) < 2L && keep.fraction[i] >= 0.5)) {
        keep[ selectors[[i]] ] <- TRUE
      } else if (keep.fraction[i] != 0 && length(selectors[[i]]) >= 2L) {
        if (keep.sparse) {
          keep[ selectors[[i]] ] <-
            dens[ selectors[[i]] ] < stats::quantile(dens[ selectors[[i]] ],
                                                     keep.fraction[i],
                                                     names = FALSE,
                                                     type = 8)
        } else {
          keep[ selectors[[i]] ] <-
            dens[ selectors[[i]] ] >= stats::quantile(dens[ selectors[[i]] ],
                                                      1 - keep.fraction[i],
                                                      names = FALSE,
                                                      type = 8)
        }
      }
    }
    keep <- (keep | keep.these) & !exclude.these

    if (invert.selection){
      keep <- !keep
    }

    if (return.density) {
      data[["keep.obs"]] <- keep
      data[["density"]] <- dens
    }

    data[keep, ]
  },

    required_aes = "x|y"
  )

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
StatDens1dFilterG <-
  ggplot2::ggproto(
    "StatDens1dFilterG",
    ggplot2::Stat,
    compute_group =
      #      dens1d_flt_compute_fun,
      # code duplicated above to ensure it is all seen by 'covr'
      function(data,
               scales,
               keep.fraction,
               keep.number,
               keep.sparse,
               keep.these,
               exclude.these,
               these.target,
               pool.along,
               xintercept,
               invert.selection,
               bw,
               kernel,
               adjust,
               n,
               orientation,
               return.density) {

        force(data)

        keep.these <- these2logical(these = keep.these,
                                    data = data,
                                    these.target = these.target)

        exclude.these <- these2logical(these = exclude.these,
                                       data = data,
                                       these.target = these.target)

        # discard redundant splits and make list of logical vectors
        if (pool.along != "x" &&
            xintercept < max(data[[orientation]]) &&
            xintercept > min(data[[orientation]])) {

          selectors <-list(low.tail = data[[orientation]] <= xintercept,
                           high.tail = data[[orientation]] > xintercept)
          if (length(keep.fraction) != 2L) {
            keep.fraction <- rep_len(keep.fraction, length.out = 2)
          }
          if (length(keep.number) != 2L) {
            if (length(keep.number) == 1L) {
              keep.number <- keep.number %/% 2
            }
            keep.number <- rep_len(keep.number, length.out = 2)
          }
          num.rows <- sapply(selectors, sum) # selectors are logical
        } else {
          keep.fraction <- keep.fraction[[1]] # can be a vector or a list
          keep.number <- keep.number[[1]]
          num.rows <- nrow(data)
          selectors <- list(all = rep.int(TRUE, times = num.rows))
        }

        # vectorized
        too.large.frac <- num.rows * keep.fraction > keep.number
        keep.fraction[too.large.frac] <-
          keep.number[too.large.frac] / num.rows[too.large.frac]

        # density on a grid
        # data with fewer than 2 rows is as a special case as density() fails
        if (length(unique(data[[orientation]])) >= 2L) {
          dens <-
            stats::density(data[[orientation]],
                           bw = bw, kernel = kernel, adjust = adjust, n = n,
                           from = scales[[orientation]]$dimension()[1],
                           to = scales[[orientation]]$dimension()[2])

          # estimate density at each observations coordinates
          fdens <- stats::splinefun(dens$x, dens$y) # y contains estimate of density
          dens <- fdens(data[[orientation]])
        } else {
          if (nrow(data) > 1L) {
            message("Density not computed, too few distinct values in '", orientation, "'")
          }
          dens <- rep_len(1, nrow(data))
        }

        # we construct one logical vector by adding observations/label to be kept
        # we may have a list of 1 or 2 logical vectors
        keep <- logical(nrow(data))
        for (i in seq_along(selectors)) {
          if (keep.fraction[i] == 1) {
            keep[ selectors[[i]] ] <- TRUE
          } else if (keep.fraction[i] != 0 && length(selectors[[i]]) >= 2L) {
            if (keep.sparse) {
              keep[ selectors[[i]] ] <-
                 dens[ selectors[[i]] ] < stats::quantile(dens[ selectors[[i]] ],
                                                         keep.fraction[i],
                                                         names = FALSE,
                                                         type = 8)
            } else {
              keep[ selectors[[i]] ] <-
                dens[ selectors[[i]] ] >= stats::quantile(dens[ selectors[[i]] ],
                                                          1 - keep.fraction[i],
                                                          names = FALSE,
                                                          type = 8)
            }
          }
        }
        keep <- (keep | keep.these) & !exclude.these

        if (invert.selection){
          keep <- !keep
        }

        if (return.density) {
          data[["keep.obs"]] <- keep
          data[["density"]] <- dens
        }

        data[keep, ]
      },

    required_aes = "x|y"
  )

context("stat_dens1d_filter_g")

random_string <-
  function(len = 6) {
    paste(sample(letters, len, replace = TRUE), collapse = "")
  }

# Make random data.
set.seed(1001)
d <- tibble::tibble(
  x = rnorm(10),
  y = rnorm(10),
  group = rep(c("A", "B"), c(5, 5)),
  lab = replicate(10, { random_string() })
)

d$xg <- d$x
d$xg[6:10] <- d$xg[6:10] + 1

test_that("stat_dens1d_filter_g, default arguments", {
  sd <- stat_dens1d_filter_g()
  expect_identical(sd$stat_params, stat_dens1d_filter_g()$stat_params)
})

test_that("stat_dens1d_filter_g, change default arguments", {
  sd1 <- stat_dens1d_filter_g(shape = 1, size = 3, keep.fraction = 1/4, adjust = 2)
  expect_true(sd1$stat_params$keep.fraction == 1/4)
  expect_true(sd1$stat_params$adjust == 2)

})

test_that("stat_dens1d_filter_g, incorrect argument", {
  expect_error(ggplot(data = d, aes(xg, y, colour = group)) +
    geom_point() +
    geom_rug(sides = "b") +
    stat_dens1d_filter_g(colour = "black",
                         shape = 1, size = 3, keep.fraction = 2, adjust = 2))
})

context("stat_dens1d_filter")

library(tibble)

make_data_tbl <- function(nrow = 100, rfun = rnorm, ...) {
  if (nrow %% 2) {
    nrow <- nrow + 1
  }

  set.seed(1001)

  random_string <-
    function(len = 6) {
      paste(sample(letters, len, replace = TRUE), collapse = "")
    }

  tibble::tibble(
    x = rfun(nrow, ...),
    y = rfun(nrow, ...),
    group = rep(c("A", "B"), c(nrow / 2, nrow / 2)),
                lab = replicate(nrow, { random_string() })
  )
}

test_that("filter_x_params", {

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens1d_filter(keep.fraction = NA)
  )

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens1d_filter(keep.fraction = 5)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens1d_filter(keep.fraction = 1)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens1d_filter(keep.fraction = 0)
  )

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens1d_filter(keep.fraction = -1)
  )

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens1d_filter(keep.number = NA)
  )

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens1d_filter(keep.number = -1)
  )

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens1d_filter_g(keep.fraction = NA)
  )

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens1d_filter_g(keep.fraction = 5)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens1d_filter_g(keep.fraction = 1)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens1d_filter_g(keep.fraction = 0)
  )

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens1d_filter_g(keep.fraction = -1)
  )

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens1d_filter_g(keep.number = NA)
  )

  testthat::expect_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens1d_filter_g(keep.number = -1)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens1d_filter_g(keep.these = 1:3)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens1d_filter_g(keep.these = TRUE)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens1d_filter_g(keep.these = FALSE)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens1d_filter_g(keep.these = rep(c(TRUE, FALSE), 10L))
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y, label = lab)) +
      stat_dens1d_filter_g(keep.these = function(x) {grepl("^1", x)})
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens1d_filter_g(keep.these = function(x) {grepl("^1", x)})
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens1d_filter_g(keep.these = integer())
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(20), aes(x, y)) +
      stat_dens1d_filter_g(keep.these = double())
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(2), aes(x, y)) +
      stat_dens1d_filter_g(keep.fraction = 0)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(1), aes(x, y)) +
      stat_dens1d_filter_g(keep.fraction = 0)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(2), aes(x, y)) +
      stat_dens1d_filter_g(keep.fraction = 1)
  )

  testthat::expect_no_error(
    ggplot(data = make_data_tbl(1), aes(x, y)) +
      stat_dens1d_filter_g(keep.fraction = 1)
  )

  testthat::expect_no_error(
    ggplot(data = data.frame(x = 1:2, y = 1), aes(x, y)) +
      stat_dens1d_filter_g(keep.fraction = 0)
  )

  testthat::expect_no_error(
    ggplot(data = data.frame(x = 1:2, y = 1), aes(x, y)) +
      stat_dens1d_filter_g(keep.fraction = 1)
  )

  testthat::expect_no_error(
    ggplot(data = data.frame(x = 1:2, y = 1), aes(x, y)) +
      stat_dens1d_filter_g(keep.fraction = 0, orientation = "y")
  )

  testthat::expect_no_error(
    ggplot(data = data.frame(x = 1:2, y = 1), aes(x, y)) +
      stat_dens1d_filter_g(keep.fraction = 1, orientation = "y")
  )

  testthat::expect_no_error(
    ggplot(data = data.frame(x = 1:2, y = 1), aes(x, y)) +
      stat_dens1d_filter_g(keep.fraction = 0.5, orientation = "y")
  )

})

test_that("numbers_x_tb", {
  vdiffr::expect_doppelganger("stat_d1d_fltg_01",
                              ggplot(data = make_data_tbl(10), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter_g(colour = "red")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_01",
                              ggplot(data = make_data_tbl(10), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_02",
                              ggplot(data = make_data_tbl(10), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red",
                                                   keep.fraction = 1/2)
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_03a",
                              ggplot(data = make_data_tbl(20), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_03",
                              ggplot(data = make_data_tbl(20), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_04",
                              ggplot(data = make_data_tbl(100), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_05",
                              ggplot(data = make_data_tbl(500), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_06",
                              ggplot(data = make_data_tbl(2000), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_07",
                              ggplot(data = make_data_tbl(2000), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red",
                                                   keep.fraction = 0.01)
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_08",
                              ggplot(data = make_data_tbl(1000), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red",
                                                   keep.sparse = FALSE)
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_09",
                              ggplot(data = make_data_tbl(1000), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red",
                                                   keep.sparse = FALSE)+
                                stat_dens1d_filter(colour = "blue")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_08s",
                              ggplot(data = make_data_tbl(2000), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red",
                                                   invert.selection = TRUE)
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_09s",
                              ggplot(data = make_data_tbl(2000), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red",
                                                   invert.selection = TRUE)+
                                stat_dens1d_filter(colour = "blue")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_10",
                              ggplot(data = make_data_tbl(2000), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red",
                                                   keep.fraction = 0.01,
                                                   keep.sparse = FALSE)
  )
  # vdiffr::expect_doppelganger("stat_d1d_flt_11",
  #                             ggplot(data = make_data_tbl(10000), aes(x, y)) +
  #                               geom_point() +
  #                               stat_dens1d_filter(colour = "red")
  # )
  # vdiffr::expect_doppelganger("stat_d1d_flt_12",
  #                             ggplot(data = make_data_tbl(10000), aes(x, y)) +
  #                               geom_point() +
  #                               stat_dens1d_filter(colour = "red", keep.fraction = 0.01)
  # )
  vdiffr::expect_doppelganger("stat_d1d_flt_13",
                              ggplot(data = make_data_tbl(1000, rfun = runif), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red", keep.fraction = 0.1)
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_14",
                              ggplot(data = make_data_tbl(1000, rfun = rgamma, shape = 2), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red", keep.fraction = 0.1)
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_15",
                              ggplot(data = make_data_tbl(1000, rfun = rgamma, shape = 6), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red", keep.fraction = 0.1)
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_16",
                              ggplot(data = make_data_tbl(1000, rfun = rbeta, shape1 = 3, shape2 = 12), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red", keep.fraction = 0.1)
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_17",
                              ggplot(data = make_data_tbl(1000, rfun = rbeta, shape1 = 3, shape2 = 12), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red", keep.fraction = 0.1) +
                                scale_y_log10()
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_18",
                              ggplot(data = make_data_tbl(1000, rfun = rbeta, shape1 = 3, shape2 = 12), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red", keep.fraction = 0.1, return.density = TRUE)
  )
})

test_that("numbers_y_tb", {
  vdiffr::expect_doppelganger("stat_d1d_flt_y01",
                              ggplot(data = make_data_tbl(10), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red",
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_y02",
                              ggplot(data = make_data_tbl(10), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red",
                                                   keep.fraction = 1/2,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_y03",
                              ggplot(data = make_data_tbl(20), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red",
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_y04",
                              ggplot(data = make_data_tbl(100), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red",
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_y05",
                              ggplot(data = make_data_tbl(500), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red",
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_y06",
                              ggplot(data = make_data_tbl(2000), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red",
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_y07",
                              ggplot(data = make_data_tbl(2000), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red",
                                                   keep.fraction = 0.01,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_y08",
                              ggplot(data = make_data_tbl(2000), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red",
                                                   keep.sparse = FALSE,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_y09",
                              ggplot(data = make_data_tbl(2000), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red",
                                                   keep.sparse = FALSE,
                                                   orientation = "y")+
                                stat_dens1d_filter(colour = "blue",
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_y10",
                              ggplot(data = make_data_tbl(2000), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red",
                                                   keep.fraction = 0.01,
                                                   keep.sparse = FALSE,
                                                   orientation = "y")
  )
  # vdiffr::expect_doppelganger("stat_d1d_flt_y11",
  #                             ggplot(data = make_data_tbl(10000), aes(x, y)) +
  #                               geom_point() +
  #                               stat_dens1d_filter(colour = "red",
  #                                                  orientation = "y")
  # )
  # vdiffr::expect_doppelganger("stat_d1d_flt_y12",
  #                             ggplot(data = make_data_tbl(10000), aes(x, y)) +
  #                               geom_point() +
  #                               stat_dens1d_filter(colour = "red",
  #                                                  keep.fraction = 0.01,
  #                                                  orientation = "y")
  # )
  vdiffr::expect_doppelganger("stat_d1d_flt_y13",
                              ggplot(data = make_data_tbl(1000, rfun = runif), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red", keep.fraction = 0.1,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_y14",
                              ggplot(data = make_data_tbl(1000, rfun = rgamma, shape = 2), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red", keep.fraction = 0.1,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_y15",
                              ggplot(data = make_data_tbl(1000, rfun = rgamma, shape = 6), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red", keep.fraction = 0.1,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_y16",
                              ggplot(data = make_data_tbl(1000, rfun = rbeta, shape1 = 3, shape2 = 12), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red", keep.fraction = 0.1,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_flt_y17",
                              ggplot(data = make_data_tbl(1000, rfun = rbeta, shape1 = 3, shape2 = 12), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_filter(colour = "red", keep.fraction = 0.1,
                                                   orientation = "y") +
                                scale_y_log10()
  )
})

make_labs_tbl <- function(nrow = 100, rfun = rnorm, ...) {
  if (nrow %% 2) {
    nrow <- nrow + 1
  }

  random_string <-
    function(len = 6) {
      paste(sample(letters, len, replace = TRUE), collapse = "")
    }

  set.seed(1001)

  tibble::tibble(
    x = rfun(nrow, ...),
    y = rfun(nrow, ...),
    group = rep(c("A", "B"), c(nrow / 2, nrow / 2)),
    lab = replicate(nrow, { random_string() })
  )
}

test_that("labels_x_params", {

  testthat::expect_error(
    ggplot( data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens1d_labels(label.fill = rep("", 2))
  )

  testthat::expect_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens1d_labels(keep.fraction = NA)
  )

  testthat::expect_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens1d_labels(keep.fraction = 5)
  )

  testthat::expect_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens1d_labels(keep.fraction = -1)
  )

  testthat::expect_no_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens1d_labels(keep.fraction = 0)
  )

  testthat::expect_no_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens1d_labels(keep.fraction = 1)
  )

  testthat::expect_no_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens1d_labels(keep.fraction = rep(0.1, 2))
  )

  testthat::expect_no_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens1d_labels(keep.fraction = rep(0.1, 2),
                         pool.along = "none")
  )

  testthat::expect_no_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens1d_labels(keep.fraction = rep(0.1, 5))
  )

  testthat::expect_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens1d_labels(keep.number = NA)
  )

  testthat::expect_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens1d_labels(keep.number = -1)
  )

  testthat::expect_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens1d_labels(label.fill = 123)
  )

  testthat::expect_no_error(
    ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
      stat_dens1d_labels()
  )

  testthat::expect_no_error(
    ggplot(data = make_labs_tbl(2), aes(x, y, label = lab)) +
      stat_dens1d_labels(keep.fraction = 0)
  )

  testthat::expect_no_error(
    ggplot(data = make_labs_tbl(1), aes(x, y, label = lab)) +
      stat_dens1d_labels(keep.fraction = 0)
  )

  testthat::expect_no_error(
    ggplot(data = make_labs_tbl(2), aes(x, y, label = lab)) +
      stat_dens1d_labels(keep.fraction = 1)
  )

  testthat::expect_no_error(
    ggplot(data = make_labs_tbl(1), aes(x, y, label = lab)) +
      stat_dens1d_labels(keep.fraction = 1)
  )

  testthat::expect_no_error(
    ggplot(data = data.frame(x = 1:2, y = 1), aes(x, y, label = lab)) +
      stat_dens1d_labels(keep.fraction = 0)
  )

  testthat::expect_no_error(
    ggplot(data = data.frame(x = 1:2, y = 1), aes(x, y, label = lab)) +
      stat_dens1d_labels(keep.fraction = 1)
  )

  testthat::expect_no_error(
    ggplot(data = data.frame(x = 1:2, y = 1), aes(x, y, label = lab)) +
      stat_dens1d_labels(keep.fraction = 0, orientation = "y")
  )

  testthat::expect_no_error(
    ggplot(data = data.frame(x = 1:2, y = 1), aes(x, y, label = lab)) +
      stat_dens1d_filter_g(keep.fraction = 1, orientation = "y")
  )

  testthat::expect_no_error(
    ggplot(data = data.frame(x = 1:2, y = 1), aes(x, y, label = lab)) +
      stat_dens1d_labels(keep.fraction = 0.5, orientation = "y")
  )

  vdiffr::expect_doppelganger("stat_d1d_lbl_fill_x01",
                                      ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
                                        geom_point() +
                                        stat_dens1d_labels(colour = "red", label.fill = "")
          )
          vdiffr::expect_doppelganger("stat_d1d_lbl_fill_x02",
                                      ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
                                        geom_point() +
                                        stat_dens1d_labels(colour = "red", label.fill = NA)
          )
          vdiffr::expect_doppelganger("stat_d1d_lbl_fill_x03",
                                      ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
                                        geom_point() +
                                        stat_dens1d_labels(colour = "red", label.fill = "z")
          )
          vdiffr::expect_doppelganger("stat_d1d_lbl_fill_x04",
                                      ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
                                        geom_point() +
                                        stat_dens1d_labels(colour = "red", label.fill = toupper)
          )
          vdiffr::expect_doppelganger("stat_d1d_lbl_fill_x05",
                                      ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
                                        geom_point() +
                                        stat_dens1d_labels(label.fill = FALSE)
          )
          vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_x01",
                                      ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                        geom_point() +
                                        stat_dens1d_labels(colour = "red", keep.fraction = 1/3)
          )
          vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_x02",
                                      ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                        geom_point() +
                                        stat_dens1d_labels(colour = "red",
                                                           keep.fraction = 1/3,
                                                           pool.along = "none")
          )
          vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_x03",
                                      ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                        geom_point() +
                                        stat_dens1d_labels(colour = "red",
                                                           keep.fraction = 1/3,
                                                           pool.along = "none",
                                                           xintercept = -1)
          )
          vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_x04",
                                      ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                        geom_point() +
                                        stat_dens1d_labels(colour = "red",
                                                           keep.fraction = 1/3,
                                                           pool.along = "none",
                                                           xintercept = 1)
          )
          vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_x05",
                                      ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                        geom_point() +
                                        stat_dens1d_labels(colour = "red",
                                                           keep.fraction = 1/3,
                                                           pool.along = "none",
                                                           xintercept = -3)
          )
          vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_x06",
                                      ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                        geom_point() +
                                        stat_dens1d_labels(colour = "red",
                                                           keep.fraction = 1/3,
                                                           pool.along = "none",
                                                           xintercept = 3)
          )
          vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_x07",
                                      ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                        geom_point() +
                                        stat_dens1d_labels(colour = "red",
                                                           keep.fraction = 1,
                                                           keep.number = 6)
          )
          vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_x08",
                                      ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                        geom_point() +
                                        stat_dens1d_labels(colour = "red",
                                                           keep.fraction = 1,
                                                           keep.number = 6,
                                                           pool.along = "none")
          )
          vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_x09",
                                      ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                        geom_point() +
                                        stat_dens1d_labels(colour = "red",
                                                           keep.fraction = 1,
                                                           keep.number = c(1, 5),
                                                           pool.along = "none")
          )
          vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_x10",
                                      ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                        geom_point() +
                                        stat_dens1d_labels(colour = "red",
                                                           keep.fraction = 1,
                                                           keep.number = c(5, 1),
                                                           pool.along = "none")
          )
          vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_x11",
                                      ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                        geom_point() +
                                        stat_dens1d_labels(colour = "red",
                                                           keep.fraction = c(1/3, 0),
                                                           pool.along = "none")
          )
          vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_x12",
                                      ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                        geom_point() +
                                        stat_dens1d_labels(colour = "red",
                                                           keep.fraction = c(0, 1/3),
                                                           pool.along = "none")
          )
})

test_that("labels_y_params", {
  vdiffr::expect_doppelganger("stat_d1d_lbl_fill_y01",
                              ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red", label.fill = "",
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_fill_y02",
                              ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red", label.fill = NA,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_fill_y03",
                              ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red", label.fill = "z",
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_fill_y04",
                              ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red", label.fill = toupper,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_fill_y05",
                              ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(label.fill = FALSE,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_y01",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red", keep.fraction = 1/3,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_y02",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   keep.fraction = 1/3,
                                                   pool.along = "none",
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_y03",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   keep.fraction = 1/3,
                                                   pool.along = "none",
                                                   xintercept = -1,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_y04",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   keep.fraction = 1/3,
                                                   pool.along = "none",
                                                   xintercept = 1,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_y05",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   keep.fraction = 1/3,
                                                   pool.along = "none",
                                                   xintercept = -3,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_y06",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   keep.fraction = 1/3,
                                                   pool.along = "none",
                                                   xintercept = 3,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_y07",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   keep.fraction = 1,
                                                   keep.number = 6,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_y08",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   keep.fraction = 1,
                                                   keep.number = 6,
                                                   pool.along = "none",
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_y09",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   keep.fraction = 1,
                                                   keep.number = c(1, 5),
                                                   pool.along = "none",
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_y10",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   keep.fraction = 1,
                                                   keep.number = c(5, 1),
                                                   pool.along = "none",
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_y11",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   keep.fraction = c(1/3, 0),
                                                   pool.along = "none",
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_keep_frac_y12",
                              ggplot(data = make_labs_tbl(18), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   keep.fraction = c(0, 1/3),
                                                   pool.along = "none",
                                                   orientation = "y")
  )
})

test_that("labels_x_tb", {
  # test non-mapped label filled with row names
  vdiffr::expect_doppelganger("stat_d1d_nolbl_x01",
                              ggplot(data = make_data_tbl(10), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   orientation = "x")
  )
  vdiffr::expect_doppelganger("stat_d1d_nolbl_x02",
                              ggplot(data = make_labs_tbl(10), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   orientation = "x")
  )
  # test with variable mapped to label aesthetic
  vdiffr::expect_doppelganger("stat_d1d_lbl_x01",
                              ggplot(data = make_labs_tbl(10), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   orientation = "x")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_x02",
                              ggplot(data = make_labs_tbl(10), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   keep.fraction = 1/2,
                                                   orientation = "x")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_x03",
                              ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   orientation = "x")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_x04",
                              ggplot(data = make_labs_tbl(100), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   orientation = "x")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_x05",
                              ggplot(data = make_labs_tbl(500), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   orientation = "x")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_x06",
                              ggplot(data = make_labs_tbl(2000), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   orientation = "x")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_x07",
                              ggplot(data = make_labs_tbl(2000), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   keep.fraction = 0.01,
                                                   orientation = "x")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_x08",
                              ggplot(data = make_labs_tbl(1000), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   keep.sparse = FALSE,
                                                   orientation = "x")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_x09",
                              ggplot(data = make_labs_tbl(1000), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   keep.sparse = FALSE,
                                                   orientation = "x")+
                                stat_dens1d_labels(colour = "blue",
                                                   orientation = "x")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_x10",
                              ggplot(data = make_labs_tbl(2000), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   keep.fraction = 0.01,
                                                   keep.sparse = FALSE,
                                                   orientation = "x")
  )
  # vdiffr::expect_doppelganger("stat_d1d_lbl_x11",
  #                             ggplot(data = make_labs_tbl(10000), aes(x, y, label = lab)) +
  #                               geom_point() +
  #                               stat_dens1d_labels(colour = "red",
  #                                                  orientation = "x")
  # )
  # vdiffr::expect_doppelganger("stat_d1d_lbl_x12",
  #                             ggplot(data = make_labs_tbl(10000), aes(x, y, label = lab)) +
  #                               geom_point() +
  #                               stat_dens1d_labels(colour = "red",
  #                                                  keep.fraction = 0.01,
  #                                                  orientation = "x")
  # )
  vdiffr::expect_doppelganger("stat_d1d_lbl_x13",
                              ggplot(data = make_labs_tbl(1000, rfun = runif), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red", keep.fraction = 0.1,
                                                   orientation = "x")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_x14",
                              ggplot(data = make_labs_tbl(1000, rfun = rgamma, shape = 2), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red", keep.fraction = 0.1,
                                                   orientation = "x")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_x15",
                              ggplot(data = make_labs_tbl(1000, rfun = rgamma, shape = 6), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red", keep.fraction = 0.1,
                                                   orientation = "x")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_x16",
                              ggplot(data = make_labs_tbl(1000, rfun = rbeta, shape1 = 3, shape2 = 12), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red", keep.fraction = 0.1,
                                                   orientation = "x")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_x17",
                              ggplot(data = make_labs_tbl(1000, rfun = rbeta, shape1 = 3, shape2 = 12), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red", keep.fraction = 0.1,
                                                   orientation = "x") +
                                scale_x_log10()
  )
})

test_that("labels_y_tb", {
  # test non-mapped label filled with row names
  vdiffr::expect_doppelganger("stat_d1d_nolbl_y01",
                              ggplot(data = make_data_tbl(10), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_nolbl_y02",
                              ggplot(data = make_labs_tbl(10), aes(x, y)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   orientation = "y")
  )
  # test with variable mapped to label aesthetic
  vdiffr::expect_doppelganger("stat_d1d_lbl_y01",
                              ggplot(data = make_labs_tbl(10), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_y02",
                              ggplot(data = make_labs_tbl(10), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   keep.fraction = 1/2,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_y03",
                              ggplot(data = make_labs_tbl(20), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_y04",
                              ggplot(data = make_labs_tbl(100), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_y05",
                              ggplot(data = make_labs_tbl(500), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_y06",
                              ggplot(data = make_labs_tbl(2000), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_y07",
                              ggplot(data = make_labs_tbl(2000), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   keep.fraction = 0.01,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_y08",
                              ggplot(data = make_labs_tbl(2000), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   keep.sparse = FALSE,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_y09",
                              ggplot(data = make_labs_tbl(2000), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   keep.sparse = FALSE,
                                                   orientation = "y")+
                                stat_dens1d_labels(colour = "blue",
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_y10",
                              ggplot(data = make_labs_tbl(2000), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red",
                                                   keep.fraction = 0.01,
                                                   keep.sparse = FALSE,
                                                   orientation = "y")
  )
  # vdiffr::expect_doppelganger("stat_d1d_lbl_y11",
  #                             ggplot(data = make_labs_tbl(10000), aes(x, y, label = lab)) +
  #                               geom_point() +
  #                               stat_dens1d_labels(colour = "red",
  #                                                  orientation = "y")
  # )
  # vdiffr::expect_doppelganger("stat_d1d_lbl_y12",
  #                             ggplot(data = make_labs_tbl(10000), aes(x, y, label = lab)) +
  #                               geom_point() +
  #                               stat_dens1d_labels(colour = "red",
  #                                                  keep.fraction = 0.01,
  #                                                  orientation = "y")
  # )
  vdiffr::expect_doppelganger("stat_d1d_lbl_y13",
                              ggplot(data = make_labs_tbl(1000, rfun = runif), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red", keep.fraction = 0.1,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_y14",
                              ggplot(data = make_labs_tbl(1000, rfun = rgamma, shape = 2), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red", keep.fraction = 0.1,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_y15",
                              ggplot(data = make_labs_tbl(1000, rfun = rgamma, shape = 6), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red", keep.fraction = 0.1,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_y16",
                              ggplot(data = make_labs_tbl(1000, rfun = rbeta, shape1 = 3, shape2 = 12), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red", keep.fraction = 0.1,
                                                   orientation = "y")
  )
  vdiffr::expect_doppelganger("stat_d1d_lbl_y17",
                              ggplot(data = make_labs_tbl(1000, rfun = rbeta, shape1 = 3, shape2 = 12), aes(x, y, label = lab)) +
                                geom_point() +
                                stat_dens1d_labels(colour = "red", keep.fraction = 0.1,
                                                   orientation = "y") +
                                scale_y_log10()
  )
})

#' Reference arrows on the margins
#'
#' Small arrows on plot margins can supplement a 2d display with annotations.
#' Arrows can be used to highlight specific values along a margin. The geometries
#' \code{geom_x_margin_arrow()} and \code{geom_y_margin_arrow()} behave
#' similarly \code{geom_vline()} and \code{geom_hline()} and share their "double
#' personality" as both annotations and geometries.
#'
#' @param mapping The aesthetic mapping, usually constructed with
#'   \code{\link[ggplot2]{aes}}. Only needs to be set at the layer level if you
#'   are overriding the plot defaults.
#' @param data A layer specific dataset - only needed if you want to override
#'   the plot defaults.
#' @param stat The statistical transformation to use on the data for this layer,
#'   as a string.
#' @param na.rm If \code{FALSE} (the default), removes missing values with a
#'   warning.  If \code{TRUE} silently removes missing values.
#' @param position Position adjustment, either as a string, or the result of a
#'   call to a position adjustment function.
#' @param ... other arguments passed on to \code{\link[ggplot2]{layer}}. This
#'   can include aesthetics whose values you want to set, not map. See
#'   \code{\link[ggplot2]{layer}} for more details.
#' @param show.legend logical. Should this layer be included in the legends?
#'   \code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
#'   never includes, and \code{TRUE} always includes.
#' @param inherit.aes If \code{FALSE}, overrides the default aesthetics, rather
#'   than combining with them. This is most useful for helper functions that
#'   define both data and aesthetics and shouldn't inherit behaviour from the
#'   default plot specification, e.g., \code{\link[ggplot2]{borders}}.
#' @param sides A string that controls which sides of the plot the rugs appear
#'   on. It can be set to a string containing any combination of \code{"trbl"},
#'   for top, right, bottom, and left.
#' @param arrow.length numeric value expressed in npc units for the length of the
#'   arows inwards from the edge of the plotting area.
#' @param xintercept,yintercept numeric Parameters that control the position of
#'   the marginal points. If these are set, data, mapping and show.legend are
#'   overridden.
#'
#' @family Geometries for marginal annotations in ggplots
#'
#' @return A plot layer instance.
#'
#' @export
#' @examples
#' p <- ggplot(mtcars, aes(wt, mpg)) +
#'   geom_point()
#' p
#' p + geom_x_margin_arrow(xintercept = 3.5)
#' p + geom_y_margin_arrow(yintercept = c(18, 28, 15))
#' p + geom_x_margin_arrow(data = data.frame(x = c(2.5, 4.5)),
#'                         mapping = aes(xintercept = x))
#' p + geom_x_margin_arrow(data = data.frame(x = c(2.5, 4.5)),
#'                         mapping = aes(xintercept = x),
#'                         sides="tb")
#'
geom_x_margin_arrow <- function(mapping = NULL, data = NULL,
                                stat = "identity", position = "identity",
                                ...,
                                xintercept,
                                sides = "b",
                                arrow.length = 0.03,
                                na.rm = FALSE,
                                show.legend = FALSE,
                                inherit.aes = FALSE) {

  # Act like an annotation
  if (!missing(xintercept)) {
    data <- as.data.frame(list(xintercept = xintercept))
    mapping <- ggplot2::aes(xintercept = xintercept)
    show.legend <- FALSE
  }

  ggplot2::layer(
    data = data,
    mapping = mapping,
    stat = stat,
    geom = GeomXMarginArrow,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      sides = sides,
      arrow.length = arrow.length,
      na.rm = na.rm,
      ...
    )
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
GeomXMarginArrow <-
  ggplot2::ggproto("GeomXMarginArrow", ggplot2::Geom,
                   required_aes = c("xintercept"),
                   non_missing_aes = c("linewidth", "shape", "colour"),
                   default_aes = ggplot2::aes(
                     colour = "black",
                     fill = "red",
                     linewidth = 1,
                     alpha = NA
                   ),

  draw_panel = function(data, panel_params, coord, sides = "b",
                        arrow.length = 0.03, na.rm = FALSE) {
    if (is.character(data$shape)) {
      data$shape <- translate_shape_string(data$shape)
    }

    rugarrows <- list()
    data <- coord$transform(data, panel_params)

    # For coord_flip, coord$tranform does not flip the sides where to
    # draw the rugs. We have to flip them.
    flipped <- inherits(coord, "CoordFlip")
    if (flipped) {
      sides <- chartr("tblr", "rlbt", sides)
    }

    arrow <- arrow(length = unit(arrow.length / 3, "npc"), ends = "first", type = "open")
    gp <- gpar(col = alpha(data$colour, data$alpha), lty = data$linetype, lwd = data$linewidth * .pt)
    if (!flipped && !is.null(data$xintercept)) {
      if (grepl("b", sides)) {
        rugarrows$x_b <- grid::segmentsGrob(
          x0 = grid::unit(data$xintercept, "native"),
          x1 = grid::unit(data$xintercept, "native"),
          y0 = grid::unit(0, "npc"),
          y1 = grid::unit(arrow.length, "npc"),
          arrow = arrow,
          gp = gp
        )
      }

      if (grepl("t", sides)) {
        rugarrows$x_t <- grid::segmentsGrob(
          x0 = grid::unit(data$xintercept, "native"),
          x1 = grid::unit(data$xintercept, "native"),
          y0 = grid::unit(1, "npc"), y1 = unit(1 - arrow.length, "npc"),
          arrow = arrow,
          gp = gp
        )
      }
    }

    # needed for handling flipped coords
    if (flipped && !is.null(data$yintercept)) {
      if (grepl("l", sides)) {
        rugarrows$y_l <- segmentsGrob(
          y0 = unit(data$yintercept, "native"), y1 = unit(data$yintercept, "native"),
          x0 = unit(0, "npc"), x1 = unit(arrow.length, "npc"),
          arrow = arrow,
          gp = gp
        )
      }

      if (grepl("r", sides)) {
        rugarrows$y_r <- segmentsGrob(
          y0 = unit(data$yintercept, "native"), y1 = unit(data$yintercept, "native"),
          x0 = unit(1, "npc"), x1 = unit(1 - arrow.length, "npc"),
          arrow = arrow,
          gp = gp
        )
      }
    }

    gTree(children = do.call("gList", rugarrows))
  },

  draw_key = ggplot2:::draw_key_path
)

## y axis marging
#' @rdname geom_x_margin_arrow
#' @export
#'
geom_y_margin_arrow <- function(mapping = NULL, data = NULL,
                                stat = "identity", position = "identity",
                                ...,
                                yintercept,
                                sides = "l",
                                arrow.length = 0.03,
                                na.rm = FALSE,
                                show.legend = FALSE,
                                inherit.aes = FALSE) {

  # Act like an annotation
  if (!missing(yintercept)) {
    data <- as.data.frame(list(yintercept = yintercept))
    mapping <- ggplot2::aes(yintercept = yintercept)
    show.legend <- FALSE
  }

  ggplot2::layer(
    data = data,
    mapping = mapping,
    stat = stat,
    geom = GeomYMarginArrow,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      sides = sides,
      arrow.length = arrow.length,
      na.rm = na.rm,
      ...
    )
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
GeomYMarginArrow <-
  ggproto("GeomYMarginArrow", Geom,
          required_aes = c("yintercept"),
          non_missing_aes = c("linewidth", "shape", "colour"),
          default_aes = ggplot2::aes(
            colour = "black",
            fill = "red",
            linewidth = 1,
            alpha = NA
          ),

          draw_panel = function(data, panel_params, coord, sides = "l",
                                arrow.length = 0.03, na.rm = FALSE) {
            if (is.character(data$shape)) {
              data$shape <- translate_shape_string(data$shape)
            }

            rugarrows <- list()
            data <- coord$transform(data, panel_params)

            # For coord_flip, coord$tranform does not flip the sides where to
            # draw the rugs. We have to flip them.
            flipped <- inherits(coord, "CoordFlip")
            if (flipped) {
              sides <- chartr("tblr", "rlbt", sides)
            }

            arrow <- arrow(length = unit(arrow.length / 3, "npc"), ends = "first", type = "open")
            gp <- gpar(col = alpha(data$colour, data$alpha), lty = data$linetype, lwd = data$linewidth * .pt)

            if (!flipped && !is.null(data$yintercept)) {
              if (grepl("l", sides)) {
                rugarrows$y_l <- segmentsGrob(
                  y0 = unit(data$yintercept, "native"), y1 = unit(data$yintercept, "native"),
                  x0 = unit(0, "npc"), x1 = unit(arrow.length, "npc"),
                  arrow = arrow,
                  gp = gp
                )
              }

              if (grepl("r", sides)) {
                rugarrows$y_r <- segmentsGrob(
                  y0 = unit(data$yintercept, "native"), y1 = unit(data$yintercept, "native"),
                  x0 = unit(1, "npc"), x1 = unit(1 - arrow.length, "npc"),
                  arrow = arrow,
                  gp = gp
                )
              }
            }
            # needed for handling flipped coords
            if (flipped && !is.null(data$xintercept)) {
              if (grepl("b", sides)) {
                rugarrows$x_b <- segmentsGrob(
                  x0 = unit(data$xintercept, "native"), x1 = unit(data$xintercept, "native"),
                  y0 = unit(0, "npc"), y1 = unit(arrow.length, "npc"),
                  arrow = arrow,
                  gp = gp
                )
              }

              if (grepl("t", sides)) {
                rugarrows$x_t <- segmentsGrob(
                  x0 = unit(data$xintercept, "native"), x1 = unit(data$xintercept, "native"),
                  y0 = unit(1, "npc"), y1 = unit(1 - arrow.length, "npc"),
                  arrow = arrow,
                  gp = gp
                )
              }
            }

            gTree(children = do.call("gList", rugarrows))
          },


          draw_key = ggplot2:::draw_key_path
  )

context("geom_x_margin_arrow")

test_that("geom_x_margin_arrow, missing xintercept", {
  p <- ggplot(mtcars, aes(wt, mpg)) +
    geom_point()
  p1 <- p + geom_x_margin_arrow()

  result <- expect_error(layer_data(p1))
})

test_that("geom_x_margin_arrow, xintercept single value", {
  p <- ggplot(mtcars, aes(wt, mpg)) +
    geom_point()
  p1 <- p + geom_x_margin_arrow(xintercept = 3.5)

  expect_true(p1$layers[[2]]$data$xintercept == 3.5)
})

test_that("geom_x_margin_arrow, xintercept dataframe", {
  p <- ggplot(mtcars, aes(wt, mpg)) +
    geom_point()
  p1 <- p + geom_x_margin_arrow(data = data.frame(x = c(2.5, 4.5)),
                                mapping = aes(xintercept = x))

  expect_true(p1$layers[[2]]$data$x[1] == 2.5)
  expect_true(p1$layers[[2]]$data$x[2] == 4.5)
})

#' @rdname position_dodgenudge_to
#'
#' @export
#'
position_dodge2nudge_to <-
  function(width = 1,
           preserve = c("total", "single"),
           padding = 0.1,
           reverse = FALSE,
           x = NULL,
           y = NULL,
           x.action = c("none", "spread"),
           y.action = c("none", "spread"),
           x.distance = "equal",
           y.distance = "equal",
           x.expansion = 0,
           y.expansion = 0,
           kept.origin = c("dodged", "original", "none")) {

    stopifnot("'x' must be NULL or of mode numeric" = length(x) == 0 ||
                (!anyNA(x) && mode(x) == "numeric"))
    stopifnot("'y' must be NULL or of mode numeric" = length(y) == 0 ||
                (!anyNA(y) && mode(y) == "numeric"))

    # this works as long as nudge and mapped variable are of the same class
    # ggplot2's behaviour has been in the past and seems to be again to expect
    # numeric seconds for POSIXct and numeric days for Date time shifts
    if (lubridate::is.instant(x)) {
      x <- as.numeric(x)
    }
    if (lubridate::is.instant(y)) {
      y <- as.numeric(y)
    }

    ggplot2::ggproto(NULL, PositionDodge2AndNudgeTo,
                     x = x,
                     y = y,
                     x.action = rlang::arg_match(x.action),
                     y.action = rlang::arg_match(y.action),
                     x.distance = x.distance,
                     y.distance = y.distance,
                     x.expansion = rep_len(x.expansion, 2),
                     y.expansion = rep_len(y.expansion, 2),
                     kept.origin = rlang::arg_match(kept.origin),
                     width = width,
                     preserve = rlang::arg_match(preserve),
                     padding = padding,
                     reverse = reverse
    )
  }

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
PositionDodge2AndNudgeTo <-
  ggplot2::ggproto(
    "PositionDodge2AndNudgeTo",
    Position,
    x = NULL,
    y = NULL,

    setup_params = function(self, data) {
      list(x = self$x,
           y = self$y,
           x.action = self$x.action,
           y.action = self$y.action,
           x.distance = self$x.distance,
           y.distance = self$y.distance,
           x.expansion = self$x.expansion,
           y.expansion = self$y.expansion,
           x.reorder = !is.null(self$x) && length(self$x) > 1 && length(self$x) < nrow(data),
           y.reorder = !is.null(self$y) && length(self$y) > 1 && length(self$y) < nrow(data),
           kept.origin = self$kept.origin,
           width = self$width,
           preserve = self$preserve,
           padding = self$padding,
           reverse = self$reverse
      )
    },

    compute_layer = function(self, data, params, layout) {
      x_orig <- data$x
      y_orig <- data$y
      if (!is.na(params$width)) {
        # operate on the dodged positions
        data = ggplot2::ggproto_parent(ggplot2::PositionDodge2, self)$compute_layer(data, params, layout)
      }
      x_dodged <- data$x
      y_dodged <- data$y

      # compute/convert x nudges
      if (!length(params$x)) {
        # set default x
        if (params$x.action == "none") {
          params$x <- rep_len(0, nrow(data))
        } else if (params$x.action == "spread") {
          params$x <- range(x_dodged)
        }
      } else if (is.numeric(params$x)) {
        # check user supplied x
        if (length(params$x) > nrow(data)) {
          warning("Argument 'x' longer than data: some values dropped!")
        }
        if (params$x.action == "none") {
          # recycle or trim x as needed
          if (params$x.reorder) {
            params$x <- rep_len(params$x, nrow(data))[order(order(data$x))] - x_dodged
          } else {
            params$x <- rep_len(params$x, nrow(data)) - x_dodged
          }
        } else if (params$x.action == "spread") {
          params$x <- range(params$x)
        }
      }

      if (params$x.action == "spread") {
        # apply x.expansion to x
        x.spread <- diff(params$x)
        params$x[1] <- params$x[1] - params$x.expansion[1] * x.spread
        params$x[2] <- params$x[2] + params$x.expansion[2] * x.spread
        if (params$x.distance == "equal") {
          # evenly spaced sequence of positions ordered as in data
          params$x <- seq(from = params$x[1],
                          to = params$x[2],
                          length.out = nrow(data))[order(order(data$x))] - x_dodged
        }
        # other strategies to distribute positions could be added here
      }

      # compute/convert y nudges
      if (!length(params$y)) {
        # set default y
        if (params$y.action == "none") {
          params$y <- rep_len(0, nrow(data))
        } else if (params$y.action == "spread") {
          params$y <- range(y_dodged)
        }
      } else if (is.numeric(params$y)) {
        # check user supplied y
        if (length(params$y) > nrow(data)) {
          warning("Argument 'y' longer than data: some values dropped!")
        }
        if (params$y.action == "none") {
          # recycle or trim y as needed
          if (params$y.reorder) {
            params$y <- rep_len(params$y, nrow(data))[order(order(data$y))] - y_dodged
          } else {
            params$y <- rep_len(params$y, nrow(data)) - y_dodged
          }
        } else if (params$y.action == "spread") {
          params$y <- range(params$y)
        }
      }

      if (params$y.action == "spread") {
        y.spread <- diff(params$y)
        params$y[1] <- params$y[1] - params$y.expansion[1] * y.spread
        params$y[2] <- params$y[2] + params$y.expansion[2] * y.spread
        if (params$y.distance == "equal") {
          # evenly spaced sequence ordered as in data
          params$y <- seq(from = params$y[1],
                          to = params$y[2],
                          length.out = nrow(data))[order(order(data$y))] - y_dodged
        }
        # other strategies could be added here
      }

      # As in 'ggplot2' we apply the nudge to xmin, xmax, xend, ymin, ymax, and yend.
      # Transform the dimensions for which not all nudges are zero
      if (any(params$x != 0)) {
        if (any(params$y != 0)) {
          data <- transform_position(data, function(x) x + params$x, function(y) y + params$y)
        } else {
          data <- transform_position(data, function(x) x + params$x, NULL)
        }
      } else if (any(params$y != 0)) {
        data <- transform_position(data, NULL, function(y) y + params$y)
      }
      # add original position
      if (params$kept.origin == "dodged" && !is.na(params$width)) {
        data$x_orig <- x_dodged
        data$y_orig <- y_dodged
      } else if (params$kept.origin == "original") {
        data$x_orig <- x_orig
        data$y_orig <- y_orig
      }

      data
    },

    compute_panel = function(self, data, params, scales) {
      ggplot2::ggproto_parent(PositionDodge2, self)$compute_panel(data, params, scales)
    }
  )

context("position_dodge2nudge_to")

test_that("incorrect kept.origin used", {
  expect_error(
    position_dodge2nudge_to(kept.origin = "")
    #, "`kept.origin` must be one of \"dodged\", \"original\", or \"none\", not \"\"."
  )
})

test_that("incorrect preserve used", {
  expect_error(
    position_dodge2nudge_to(preserve = "")
    #, "`preserve` must be one of \"total\", or \"single\", not \"\"."
  )
})

test_that("non-numeric y used", {
  expect_error(
    position_dodge2nudge_to(y = "bad")
    #, "'y' must be NULL or of mode numeric"
  )
})

test_that("NA y used", {
  expect_error(
    position_dodge2nudge_to(y = NA)
    #, "'y' must be NULL or of mode numeric"
  )
})

test_that("non-numeric x used", {
  expect_error(
    position_dodge2nudge_to(y = "bad")
    #, "'x' must be NULL or of mode numeric"
  )
})

test_that("NA x used", {
  expect_error(
    position_dodge2nudge_to(x = "NA")
    #, "'x' must be NULL or of mode numeric"
  )
})

test_that("correct kept.origin used", {
  expect_error(
    position_dodge2nudge_to(kept.origin = "stacked")
  )
})

test_that("correct kept.origin used", {
  expect_no_error(
    position_dodge2nudge_to(kept.origin = "original")
  )
})

test_that("correct kept.origin used", {
  expect_no_error(
    position_dodge2nudge_to(kept.origin = "none")
  )
})

test_that("correct kept.origin used", {
  expect_no_error(
    position_dodge2nudge_to(kept.origin = "dodged")
  )
})


test_that("test if correct arguments are assigned", {
  position <- position_dodge2nudge_to(kept.origin = "none")
  expect_false(position$reverse)
  expect_identical(position$kept.origin, "none")
  expect_identical(position$padding, 0.1)
  expect_identical(position$preserve, "total")
  expect_identical(position$width, 1)
  expect_null(position$x)
  expect_null(position$y)
  expect_identical(position$x.action, "none")
  expect_identical(position$y.action, "none")
  expect_identical(position$x.distance, "equal")
  expect_identical(position$y.distance, "equal")
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
  skip_if(condition = utils::packageVersion("ggplot2") < "4.0.0",
          message = "'ggplot2' < 4.0.0")
  expect_false(position$reverse)
  expect_identical(position$required_aes, character(0))
  expect_type(position$aesthetics, "closure")
  expect_type(position$use_defaults, "closure")
  skip("Not yet implemented")
  expect_identical(position$orientation, "x")
})

test_that("test if correct arguments are assigned with 'y.action'", {
  position <- position_dodge2nudge_to(kept.origin = "none", y.action = "spread")
  expect_false(position$reverse)
  expect_identical(position$kept.origin, "none")
  expect_identical(position$padding, 0.1)
  expect_identical(position$preserve, "total")
  expect_identical(position$width, 1)
  expect_null(position$x)
  expect_null(position$y)
  expect_identical(position$x.action, "none")
  expect_identical(position$y.action, "spread")
  expect_identical(position$x.distance, "equal")
  expect_identical(position$y.distance, "equal")
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
  skip_if(condition = utils::packageVersion("ggplot2") < "4.0.0",
          message = "'ggplot2' < 4.0.0")
  expect_false(position$reverse)
  expect_identical(position$required_aes, character(0))
  expect_type(position$aesthetics, "closure")
  expect_type(position$use_defaults, "closure")
  skip("Not yet implemented")
  expect_identical(position$orientation, "x")
})

test_that("test if correct arguments are assigned with 'x.action'", {
  position <- position_dodge2nudge_to(kept.origin = "none", y.action = "spread")
  expect_false(position$reverse)
  expect_identical(position$kept.origin, "none")
  expect_identical(position$padding, 0.1)
  expect_identical(position$preserve, "total")
  expect_identical(position$width, 1)
  expect_null(position$x)
  expect_null(position$y)
  expect_identical(position$x.action, "none")
  expect_identical(position$y.action, "spread")
  expect_identical(position$x.distance, "equal")
  expect_identical(position$y.distance, "equal")
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
  skip_if(condition = utils::packageVersion("ggplot2") < "4.0.0",
          message = "'ggplot2' < 4.0.0")
  expect_false(position$reverse)
  expect_identical(position$required_aes, character(0))
  expect_type(position$aesthetics, "closure")
  expect_type(position$use_defaults, "closure")
  skip("Not yet implemented")
  expect_identical(position$orientation, "x")
})

test_that("dodge2nudge_to plots are correct", {
    expect_silent(
      ggplot(mtcars, aes(x = am, y = mpg, label = mpg)) +
        geom_text_s(
          position = position_dodge2nudge_to(width = 0.85),
          size = 2.5
        )
    )

  expect_silent(
    ggplot(mtcars, aes(x = am, y = mpg, label = mpg, group = am)) +
      geom_point(position = position_dodge2(width = 0.1)) +
      geom_text_s(
        position = position_dodge2nudge_to(width = 0.1,
                                           x = 0.5,
                                           kept.origin = "dodged"),
        size = 2.5
      )
  )

  vdiffr::expect_doppelganger("dodge2-nudge-to-1",
                              ggplot(mtcars, aes(x = am, y = mpg, label = mpg, group = am)) +
                                geom_point(position = position_dodge2(width = 0.1)) +
                                geom_text_s(
                                  position = position_dodge2nudge_to(width = 0.1,
                                                                     x = 0.5,
                                                                     kept.origin = "dodged"),
                                  size = 2.5
                                )
  )

  vdiffr::expect_doppelganger("dodge2-nudge-to-2",
                              ggplot(mtcars, aes(y = am, x = mpg, label = mpg, group = am)) +
                                geom_point(position = position_dodge2(width = 0.1)) +
                                geom_text_s(
                                  position = position_dodge2nudge_to(width = 0.1,
                                                                     y = 0.5,
                                                                     kept.origin = "dodged"),
                                  size = 2.5
                                )
  )

#  skip_on_os(os = "mac")
  vdiffr::expect_doppelganger("dodge3-nudge-to-3",
                              ggplot(mtcars, aes(x = am, y = mpg, label = mpg, group = am)) +
                                geom_point(position = position_dodge2(width = 0.1,
                                                                      padding = 0.5)) +
                                geom_text_s(
                                  position = position_dodge2nudge_to(width = 0.1,
                                                                     padding = 0.5,
                                                                     x = 0.5,
                                                                     y.expansion = 0.1,
                                                                     kept.origin = "dodged"),
                                  size = 2.5
                                )
  )

  vdiffr::expect_doppelganger("dodge3-nudge-to-4",
                              ggplot(mtcars, aes(x = am, y = mpg, label = mpg, group = am)) +
                                geom_point(position = position_dodge2(width = 0.1)) +
                                geom_text_s(
                                  position = position_dodge2nudge_to(width = 0.1,
                                                                     padding = 0.5,
                                                                     y = 2,
                                                                     kept.origin = "dodged"),
                                  size = 2.5
                                )
  )


}
)

#' @rdname geom_text_pairwise
#' @include ggp2-margins.R utilities.R ggpp-legend-draw.R
#'
#' @param label.padding Amount of padding around label. Defaults to 0.25 lines.
#' @param label.r Radius of rounded corners. Defaults to 0.15 lines.
#'
#' @export
#'
geom_label_pairwise <-
  function(mapping = NULL,
           data = NULL,
           stat = "identity",
           position = "identity",
           ...,
           orientation = NA,
           parse = FALSE,
           nudge_x = 0,
           nudge_y = 0,
           default.colour = NULL,
           default.color = default.colour,
           colour.target = "all",
           color.target = colour.target,
           default.alpha = NA,
           alpha.target = "segment",
           label.padding = grid::unit(0.25, "lines"),
           label.r = grid::unit(0.15, "lines"),
           segment.linewidth = 0.5,
           arrow = NULL,
           size.unit = "mm",
           na.rm = FALSE,
           show.legend = FALSE,
           inherit.aes = FALSE) {

    colour.target <-
      rlang::arg_match(color.target,
                       values = c("all", "text", "box", "box.line",
                                  "segment", "none"),
                       multiple = TRUE)
    alpha.target <-
      rlang::arg_match(alpha.target,
                       values = c("all", "text", "box", "box.line", "box.fill",
                                  "segment", "none"),
                       multiple = TRUE)

    if (!missing(nudge_x) || !missing(nudge_y)) {
      if (!missing(position) && position != "identity") {
        rlang::abort("You must specify either `position` or `nudge_x`/`nudge_y`.")
      }
      # by default we do not keep the original positions
      position <- ggplot2::position_nudge(nudge_x, nudge_y)
    }

    ggplot2::layer(
    data = data,
    mapping = mapping,
    stat = stat,
    geom = GeomLabelPairwise,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      orientation = orientation,
      parse = parse,
      default.colour = default.color,
      colour.target = colour.target,
      default.alpha = default.alpha,
      alpha.target = alpha.target,
      label.padding = label.padding,
      label.r = label.r,
      segment.linewidth = segment.linewidth,
      arrow = arrow,
      size.unit = size.unit,
      na.rm = na.rm,
      ...
    )
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
GeomLabelPairwise <-
  ggplot2::ggproto("GeomLabelPairwise", ggplot2::Geom,
                   required_aes = c("xmin|ymin", "xmax|ymax", "y|x", "label"),

                   default_aes = ggplot2::aes(
                     x = NA_real_,
                     y = NA_real_,
                     colour = "black",
                     fill = rgb(1, 1, 1, alpha = 0.75), # "white", but occluded data are visible
                     family = "",
                     size = 3.87,
                     angle = 0,
                     hjust = 0.5,
                     vjust = 0.5,
                     alpha = NA,
                     fontface = 1,
                     lineheight = 1.2,
                     linewidth = 0.5,
                     linetype  = "solid"
                   ),

                   draw_panel = function(data, panel_params, coord, #panel_scales,
                                         parse = FALSE,
                                         na.rm = FALSE,
                                         size.unit = "mm",
                                         default.colour = NULL,
                                         colour.target = "all",
                                         default.alpha = NA,
                                         alpha.target = "fill",
                                         segment.linewidth = 0.5,
                                         arrow = NULL,
                                         label.padding = unit(0.25, "lines"),
                                         label.r = unit(0.15, "lines"),
                                         orientation = NA) {

                     if (is.null(orientation) || is.na(orientation)) {
                       if (all(c("xmin", "xmax") %in% colnames(data)) &&
                               !(all(is.na(data[["xmin"]])) || all(is.na(data[["xmax"]])))) {
                         orientation <- "x"
                       } else if (all(c("ymin", "ymax") %in% colnames(data)) &&
                                  !(all(is.na(data[["ymin"]])) || all(is.na(data[["ymax"]])))) {
                         orientation <- "y"
                       } else {
                         stop("Unable to guess 'orientation' from mapped aesthetics!")
                       }
                     }
                     default.colour <- check_default_colour(default.colour)

                     data$label <- as.character(data$label)
                     data <- subset(data, !is.na(label) & label != "")
                     if (nrow(data) == 0L) {
                       return(nullGrob())
                     }

                     lab <- data$label
                     if (parse) {
                       lab <- parse_safe(lab)
                     }

                     data <- coord$transform(data, panel_params)
                     if (all(c("x_orig", "y_orig") %in% colnames(data))) {
                       data_orig <- data.frame(x = data$x_orig, y = data$y_orig)
                       data_orig <- coord$transform(data_orig, panel_params)
                       data$x_orig <- data_orig$x
                       data$y_orig <- data_orig$y
                     }

                     if (is.character(data$vjust)) {
                       data$vjust <-
                         compute_just2d(data = data,
                                        coord = coord,
                                        panel_params = panel_params,
                                        just = data$vjust,
                                        a = "y", b = "x")
                     }
                     if (is.character(data$hjust)) {
                       data$hjust <-
                         compute_just2d(data = data,
                                        coord = coord,
                                        panel_params = panel_params,
                                        just = data$hjust,
                                        a = "x", b = "y")
                     }

                     size.unit <- resolve_text_unit(size.unit)

                     if (!inherits(label.padding, "margin")) {
                       label.padding <- rep(label.padding, length.out = 4)
                     }

                     if (orientation == "x") {
                       # loop needed as gpar is not vectorized
                       all.grobs <- grid::gList()

                       for (row.idx in 1:nrow(data)) {
                         row <- data[row.idx, , drop = FALSE]
                         if (is.na(row$x)) {
                           row$x <- (row$xmin + row$xmax) / 2
                         }
                         text.alpha <-
                           ifelse(any(alpha.target %in% c("all", "text")),
                                  row$alpha, default.alpha)
                         segment.alpha <-
                           ifelse(any(alpha.target %in% c("all", "segment")),
                                  row$alpha, default.alpha)
                         box.colour.alpha <-
                           ifelse(any(alpha.target %in% c("all", "box", "box.line")),
                                  row$alpha, default.alpha)
                         box.fill.alpha <-
                           ifelse(any(alpha.target %in% c("all", "box", "box.fill")),
                                  row$alpha, default.alpha)
                         user.grob <- labelGrob(lab[row.idx],
                                                x = unit(row$x, "native"),
                                                y = unit(row$y, "native"),
                                                just = c(row$hjust, row$vjust),
                                                padding = label.padding,
                                                r = label.r,
                                                angle = row$angle,
                                                text.gp = gpar(
                                                  col = ifelse(any(colour.target %in% c("all", "text")),
                                                               ggplot2::alpha(row$colour, text.alpha),
                                                               ggplot2::alpha(default.colour, text.alpha)),
                                                  fontsize = row$size * size.unit,
                                                  fontfamily = row$family,
                                                  fontface = row$fontface,
                                                  lineheight = row$lineheight
                                                ),
                                                rect.gp = gpar(
                                                  col = if (row$linewidth == 0) NA else # lwd = 0 is invalid in 'grid'
                                                    ifelse(any(colour.target %in% c("all", "box")),
                                                           ggplot2::alpha(row$colour, box.colour.alpha),
                                                           ggplot2::alpha(default.colour, box.colour.alpha)),
                                                  fill = alpha(row$fill, box.fill.alpha),
                                                  lwd = (if (row$linewidth == 0) 1 else row$linewidth) * .pt, # mm -> points (as in 'ggplot2')
                                                  # lwd = (if (row$linewidth == 0) 0.5 else row$linewidth) * ggplot2::.stroke, # mm -> stroke (correct)
                                                  lty = row$linetype
                                                )
                         )

                         # give unique name to each grob
                         user.grob$name <- paste("label.pw.grob", row$group, row.idx, sep = ".")

                         if (!is.null(arrow) && !inherits(arrow, "arrow")) {
                           shape <- arrow[[1]]
                           arrow <- NULL
                         } else {
                           shape <- NULL
                         }

                         segment.grob <-
                           grid::segmentsGrob(x0 = row$xmin,
                                              y0 = row$y,
                                              x1 = row$xmax,
                                              y1 = row$y,
                                              arrow = arrow,
                                              gp = grid::gpar(
                                                col = if (segment.linewidth == 0) NA else # lwd = 0 is invalid in 'grid'
                                                  ifelse(any(colour.target %in% c("all", "segment")),
                                                         ggplot2::alpha(row$colour, segment.alpha),
                                                         ggplot2::alpha(default.colour, segment.alpha)),
                                                lwd = (if (segment.linewidth == 0) 0.5 else segment.linewidth) * ggplot2::.stroke),
                                              name = paste("text.pw.segment", row$group, row.idx, sep = "."))
                         all.grobs <- grid::gList(all.grobs, segment.grob, user.grob)
                       }
                     } else {
                       # loop needed as gpar is not vectorized
                       all.grobs <- grid::gList()

                       for (row.idx in 1:nrow(data)) {
                         row <- data[row.idx, , drop = FALSE]
                         if (is.na(row$y)) {
                           row$y <- (row$ymin + row$ymax) / 2
                         }
                         text.alpha <-
                           ifelse(any(alpha.target %in% c("all", "text")),
                                  row$alpha, default.alpha)
                         segment.alpha <-
                           ifelse(any(alpha.target %in% c("all", "segment")),
                                  row$alpha, default.alpha)
                         box.colour.alpha <-
                           ifelse(any(alpha.target %in% c("all", "box", "box.line")),
                                  row$alpha, default.alpha)
                         box.fill.alpha <-
                           ifelse(any(alpha.target %in% c("all", "box", "box.fill")),
                                  row$alpha, default.alpha)
                         user.grob <- labelGrob(lab[row.idx],
                                                x = unit(row$x, "native"),
                                                y = unit(row$y, "native"),
                                                just = c(row$hjust, row$vjust),
                                                padding = label.padding,
                                                r = label.r,
                                                angle = row$angle + 90,
                                                text.gp = gpar(
                                                  col = ifelse(any(colour.target %in% c("all", "text")),
                                                               ggplot2::alpha(row$colour, text.alpha),
                                                               ggplot2::alpha(default.colour, text.alpha)),
                                                  fontsize = row$size * size.unit,
                                                  fontfamily = row$family,
                                                  fontface = row$fontface,
                                                  lineheight = row$lineheight
                                                ),
                                                rect.gp = gpar(
                                                  col = if (row$linewidth == 0) NA else # lwd = 0 is invalid in 'grid'
                                                    ifelse(any(colour.target %in% c("all", "box")),
                                                           ggplot2::alpha(row$colour, box.colour.alpha),
                                                           ggplot2::alpha(default.colour, box.colour.alpha)),
                                                  fill = alpha(row$fill, box.fill.alpha),
                                                  lwd = (if (row$linewidth == 0) 1 else row$linewidth) * .pt, # mm -> points (as in 'ggplot2')
                                                  # lwd = (if (row$linewidth == 0) 0.5 else row$linewidth) * ggplot2::.stroke, # mm -> stroke (correct)
                                                  lty = row$linetype
                                                )
                         )

                         # give unique name to each grob
                         user.grob$name <- paste("label.pw.grob", row$group, row.idx, sep = ".")

                         if (!is.null(arrow) && !inherits(arrow, "arrow")) {
                           shape <- arrow[[1]]
                           arrow <- NULL
                         } else {
                           shape <- NULL
                         }

                         segment.grob <-
                           grid::segmentsGrob(x0 = row$x,
                                              y0 = row$ymin,
                                              x1 = row$x,
                                              y1 = row$ymax,
                                              arrow = arrow,
                                              gp = grid::gpar(
                                                col = if (segment.linewidth == 0) NA else # lwd = 0 is invalid in 'grid'
                                                  ifelse(any(colour.target %in% c("all", "segment")),
                                                         ggplot2::alpha(row$colour, segment.alpha),
                                                         ggplot2::alpha(default.colour, segment.alpha)),
                                                lwd = (if (segment.linewidth == 0) 0.5 else segment.linewidth) * ggplot2::.stroke),
                                              name = paste("text.pw.segment", row$group, row.idx, sep = "."))
                         all.grobs <- grid::gList(all.grobs, segment.grob, user.grob)
                       }
                     }

                     # name needs to be unique within plot, so we would have to know layer name to use as basis
                     #                     grid::grobTree(children = all.grobs, name = "geom.text.s.panel")
                     grid::grobTree(children = all.grobs)

                   },

                   draw_key = draw_key_label_s
  )

#'  Check default.colour argument
#'
#' if possible use ink from theme as default colour
#'
#' @param default.colour The value to be checked
#' @param theme The theme from where to fetch a replacement value when
#'   \code{default.colour} is \code{NULL}.
#'
#' @details The replacement value is fetched from the theme only if
#'   'ggplot2' >= 4.0.0 and otherwise \code{"black"} is used. \code{NA}
#'   values are passed through as they can be used to disable rendering
#'   of grob elements.
#'
#' @return A colour definition or NA
#
#' @keywords internal
#'
check_default_colour <-
  function(default.colour, theme = NULL) {
    if (is.null(default.colour)) {
      if (utils::packageVersion("ggplot2") >= "4.0.0") {
        if (is.null(theme)) {
          theme <- ggplot2::get_theme()
        }
        ggplot2::calc_element("geom", theme)@ink
      } else {
        "black"
      }
    } else {
      default.colour[[1L]] # trim vectors or lists
    }
  }

# utility functions not exported by 'ggplot2'
# constant copied from geom-.R
.pt <- 72.27 / 25.4

# function copied from ggplot2's utilities.R

"%||%" <- function(a, b) {
  if (!is.null(a)) a else b
}

#' Parse takes a vector of n lines and returns m expressions.
#' See https://github.com/tidyverse/ggplot2/issues/2864 for discussion.
#'
#' parse(text = c("alpha", "", "gamma"))
#' #> expression(alpha, gamma)
#'
#' parse_safe(text = c("alpha", "", "gamma"))
#' #> expression(alpha, NA, gamma)
#'
#' @noRd
parse_safe <- function(text) {
  stopifnot(is.character(text))
  out <- vector("expression", length(text))
  for (i in seq_along(text)) {
    expr <- parse(text = text[[i]])
    out[[i]] <- if (length(expr) == 0) NA else expr[[1]]
  }
  out
}

# Helpers to convert degrees to radians and vice versa
rad2deg <- function(rad) rad * 180 / pi
deg2rad <- function(deg) deg * pi / 180


# function from ggplot2, needed in annotate() but not exported
compact <- function (x)
{
  null <- vapply(x, is.null, logical(1))
  x[!null]
}

# function from ggplot2 (not current), needed in annotate() but not exported
new_data_frame <- function (x = list(), n = NULL) {
  if (length(x) != 0 && is.null(names(x))) {
    rlang::abort("Elements must be named")
  }
  lengths <- vapply(x, length, integer(1))
  if (is.null(n)) {
    n <- if (length(x) == 0 || min(lengths) == 0)
      0
    else max(lengths)
  }
  for (i in seq_along(x)) {
    if (lengths[i] == n)
      next
    if (lengths[i] != 1) {
      rlang::abort("Elements must equal the number of rows or 1")
    }
    x[[i]] <- rep(x[[i]], n)
  }
  tibble::as_tibble(x)
}

# Used in annotations to ensure printed even when no
# global data
# copied from ggplot2's utilities.r
dummy_data <- function() new_data_frame(list(x = NA), n = 1)

# This function is used to vectorise the following pattern:
#
# obj$name1 <- obj$name1 %||% value
# obj$name2 <- obj$name2 %||% value
#
# and express this pattern as:
#
# replace_null(obj, name1 = value, name2 = value)
replace_null <- function(obj, ..., env = rlang::caller_env()) {
  # Collect dots without evaluating
  dots <- rlang::enexprs(...)
  # Select arguments that are null in `obj`
  nms  <- names(dots)
  nms  <- nms[vapply(obj[nms], is.null, logical(1))]
  # Replace those with the evaluated dots
  obj[nms] <- rlang::inject(list(!!!dots[nms]), env = env)
  obj
}

# From 'ggplot2' 4.0.0
descent_cache <- new.env(parent = emptyenv())
# Important: This function is not vectorized. Do not use to look up multiple
# font descents at once.
font_descent <- function(family = "", face = "plain", size = 12, cex = 1) {
  cur_dev <- names(grDevices::dev.cur())
  if (cur_dev == "null device") {
    cache <- FALSE   # don't cache if no device open
  } else {
    cache <- TRUE
  }
  key <- paste0(cur_dev, ':', family, ':', face, ":", size, ":", cex)
  # we only look up the first result; this function is not vectorized
  key <- key[1]

  descent <- descent_cache[[key]]

  if (is.null(descent)) {
    descent <- grid::convertHeight(grid::grobDescent(textGrob(
      label = "gjpqyQ",
      gp = gg_par(
        fontsize = size,
        cex = cex,
        fontfamily = family,
        fontface = face
      )
    )), 'inches')

    if (cache) {
      descent_cache[[key]] <- descent
    }
  }

  descent
}

test_that("parse_safe works with simple expressions", {
  expect_equal(
    parse_safe(c("", " ", "     ")),
    expression(NA, NA, NA)
  )

  expect_equal(
    parse_safe(c("A", "B", "C")),
    expression(A, B, C)
  )

  expect_equal(
    parse_safe(c("alpha", "", "gamma", " ")),
    expression(alpha, NA, gamma, NA)
  )

  expect_equal(
    parse_safe(c(NA, "a", NA, "alpha")),
    expression(NA, a, NA, alpha)
  )
})

test_that("parse_safe works with multi expressions", {
  expect_equal(
    parse_safe(c(" \n", "\n ", " \n  \n  ")),
    expression(NA, NA, NA)
  )

  expect_equal(
    parse_safe(c("alpha ~ beta", "beta \n gamma", "")),
    expression(alpha ~ beta, beta, NA)
  )

  expect_equal(
    parse_safe(c("alpha ~ beta", " ", "integral(f(x) * dx, a, b)")),
    expression(alpha ~ beta, NA, integral(f(x) * dx, a, b))
  )

  expect_equal(
    parse_safe(c(NA, 1, 2, "a \n b")),
    expression(NA, 1, 2, a)
  )
})

test_that("new_data_frame", {
  expect_is(new_data_frame(), "data.frame")
  df <- new_data_frame(list(A = 1:2, B = letters[1:2]))
  expect_is(df, "data.frame")
  expect_named(df, c("A", "B"))
  expect_equal(nrow(df), 2L)
  expect_error(new_data_frame(list(1:2, letters[1:2])))
  expect_error(new_data_frame(list(A = 1:2, B = letters[1:2]), n = 0))
  expect_error(new_data_frame(list(A = 1:2, B = letters[1:2]), n = 10))
})

#' Dodge plus nudge labels to new positions
#'
#' Functions \code{position_dodgenudge_to()} and
#' \code{position_dodge2nudge_to()} are meant to complement
#' \code{position_dodge()} and \code{position_dodge2()} from 'ggplot2', adding
#' as a second action that of \code{position_nudge_to()}. These positions are
#' generally useful for adjusting the position of labels or text. As with other
#' position functions in this package, the original positions are preserved to
#' allow the text or labels to be linked back to their original position with a
#' segment or arrow.
#'
#' @family position adjustments
#'
#' @param width Dodging width, when different to the width of the individual
#'   elements. This is useful when you want to align narrow geoms with wider
#'   geoms. See the examples.
#' @param preserve Should dodging preserve the total width of all elements at a
#'   position, or the width of a single element?.
#' @param reverse logical If TRUE, will reverse the default dodging order.
#' @param padding Padding between elements at the same position. Elements are
#'   shrunk by this proportion to allow space between them. Defaults to 0.1.
#' @param reverse If TRUE, will reverse the default dodging order.
#' @param x,y Coordinates of the destination position. A vector of mode
#'   \code{numeric}, that is extended if needed, to the same length as rows
#'   there are in \code{data}. The default, \code{NULL}, leaves the original
#'   coordinates unchanged after dodging.
#' @param x.action,y.action character string, one of \code{"none"}, or
#'   \code{"spread"}. With \code{"spread"} distributing the positions
#'   within the range of argument \code{x} or \code{y}, if non-null, or the
#'   range the variable mapped to \emph{x} or \code{y}, otherwise.
#' @param x.distance,y.distance character or numeric Currently only \code{"equal"} is
#'   implemented.
#' @param x.expansion,y.expansion numeric vectors of length 1 or 2, as a
#'   fraction of width of the range.
#' @param kept.origin One of \code{"original"}, \code{"dodged"} or
#'   \code{"none"}.
#'
#' @details These positions apply sequentially two actions, in the order they
#'   appear in their names. The applied dodge is similar to that by
#'   \code{\link[ggplot2]{position_dodge}} and
#'   \code{\link[ggplot2]{position_dodge2}} while nudging is different to that
#'   by \code{\link[ggplot2]{position_nudge}} and equal to that applied by
#'   \code{\link{position_nudge_to}}.
#'
#'   The dodged and nudged to \code{x} and/or \code{y} values replace the
#'   original ones in \code{data}, while the original or the dodged coordinates
#'   are returned in \code{x_orig} and \code{y_orig}. Nudge values supported are
#'   those of \emph{mode} numeric, thus including dates and times when they
#'   match the mapped data.
#'
#'   If the length of \code{x} and/or \code{y} is more than one but less than
#'   rows are present in the data, the vector is both recycled and reordered so
#'   that the nudges are applied sequentially based on the data values. If their
#'   length matches the number of rows in data, they are assumed to be already
#'   in data order.
#'
#'   The intended use is to label dodged bars, boxplots or points with labels
#'   aligned. In this case, it is mandatory to use the same argument to
#'   \code{width} when passing \code{position_dodge()} to \code{geom_col()} and
#'   \code{position_dodgenudge_to()} to \code{geom_text()}, \code{geom_label()},
#'   \code{geom_text_s()}, \code{geom_label_s()} or their repulsive equivalents
#'   from package 'ggrepel'. Otherwise the arrows or segments will fail to
#'   connect to the labels.
#'
#'   When applying dodging, the return of original positions instead of the
#'   dodged ones is achieved by passing \code{origin = "original"} instead of
#'   the default of \code{origin = "dodged"}.
#'
#' @note Irrespective of the action, the ordering of rows in \code{data} is
#'   preserved.
#'
#' @return A \code{"Position"} object.
#'
#' @seealso \code{\link{position_nudge_to}},
#'   \code{\link[ggplot2]{position_dodge}},
#'   \code{\link[ggplot2]{position_dodge2}}.
#'
#' @export
#'
#' @examples
#'
#' df <- data.frame(
#'   x = c(1,3,2,5,4,2.5),
#'   y = c(2, 3, 2.5, 1.8, 2.8, 1.5),
#'   grp = c("A", "A", "A", "B", "B", "B"),
#'   grp.inner = c("a", "b", "c", "a", "b", "c"),
#'   label = c("abc","cd","d","c","bcd","a")
#' )
#'
#' # default is no nudging
#' ggplot(df, aes(grp, y, label = label, fill = label)) +
#'   geom_col(position = position_dodge(width = 0.92)) +
#'   geom_text(position = position_dodgenudge_to(width = 0.92),
#'             vjust = -0.2) +
#'   theme(legend.position = "none")
#'
#' ggplot(df, aes(grp, y, label = label, fill = grp.inner)) +
#'   geom_col(position = position_dodge(width = 0.92)) +
#'   geom_text(position = position_dodgenudge_to(width = 0.92),
#'             vjust = -0.2)
#'
#' ggplot(df, aes(grp, y, label = label, fill = label)) +
#'   geom_col(position = position_dodge2(width = 0.92)) +
#'   geom_text(position = position_dodge2nudge_to(width = 0.92),
#'             vjust = -0.2) +
#'   theme(legend.position = "none")
#'
#' ggplot(df, aes(grp, y, label = label, fill = grp.inner)) +
#'   geom_col(position = position_dodge2(width = 0.92)) +
#'   geom_text(position = position_dodge2nudge_to(width = 0.92),
#'             vjust = -0.2)
#'
#' # nudging all labels to a given y value
#' ggplot(df, aes(grp, y, label = label, fill = grp.inner)) +
#'   geom_col(position = position_dodge(width = 0.92)) +
#'   geom_text(position = position_dodgenudge_to(width = 0.92, y = 0.8))
#'
#' ggplot(df, aes(grp, y, label = label, fill = grp.inner)) +
#'   geom_col(position = position_dodge2(width = 0.92)) +
#'   geom_text(position = position_dodge2nudge_to(width = 0.92, y = 0.8))
#'
#' ggplot(df, aes(grp, y, label = label, fill = grp.inner)) +
#'   geom_col(position = position_dodge(width = 0.92)) +
#'   geom_text(position = position_dodgenudge_to(width = 0.92, y = 0.8))
#'
#' ggplot(df[-1, ], aes(grp, y, label = label, fill = grp.inner)) +
#'   geom_col(position = position_dodge(width = 0.92)) +
#'   geom_text(position = position_dodgenudge_to(width = 0.92, y = 0.8))
#'
#' ggplot(df[-1, ], aes(grp, y, label = label, fill = grp.inner)) +
#'   geom_col(position = position_dodge(width = 0.92, preserve = "total")) +
#'   geom_text(position = position_dodgenudge_to(width = 0.92, y = 0.8,
#'                                               preserve = "total"))
#'
#' ggplot(df[-1, ], aes(grp, y, label = label, fill = grp.inner)) +
#'   geom_col(position = position_dodge(width = 0.92, preserve = "single")) +
#'   geom_text(position = position_dodgenudge_to(width = 0.92, y = 0.8,
#'                                               preserve = "single"))
#'
#' ggplot(df[-1, ], aes(grp, y, label = label, fill = grp.inner)) +
#'   geom_col(position = position_dodge2(width = 0.92, preserve = "total")) +
#'   geom_text(position = position_dodge2nudge_to(width = 0.92, y = 0.8,
#'                                               preserve = "total"))
#'
#' ggplot(df[-1, ], aes(grp, y, label = label, fill = grp.inner)) +
#'   geom_col(position = position_dodge2(width = 0.92, preserve = "single")) +
#'   geom_text(position = position_dodge2nudge_to(width = 0.92, y = 0.8,
#'                                               preserve = "single"))
#'
position_dodgenudge_to <-
  function(width = 1,
           preserve = c("total", "single"),
           reverse = FALSE,
           x = NULL,
           y = NULL,
           x.action = c("none", "spread"),
           y.action = c("none", "spread"),
           x.distance = "equal",
           y.distance = "equal",
           x.expansion = 0,
           y.expansion = 0,
           kept.origin = c("dodged", "original", "none")) {

    stopifnot("'x' must be NULL or of mode numeric" = length(x) == 0 ||
                (!anyNA(x) && mode(x) == "numeric"))
    stopifnot("'y' must be NULL or of mode numeric" = length(y) == 0 ||
                (!anyNA(y) && mode(y) == "numeric"))

    # this works as long as nudge and mapped variable are of the same class
    # ggplot2's behaviour has been in the past and seems to be again to expect
    # numeric seconds for POSIXct and numeric days for Date time shifts
    if (lubridate::is.instant(x)) {
      x <- as.numeric(x)
    }
    if (lubridate::is.instant(y)) {
      y <- as.numeric(y)
    }

    ggplot2::ggproto(NULL, PositionDodgeNudgeTo,
                     x = x,
                     y = y,
                     x.action = rlang::arg_match(x.action),
                     y.action = rlang::arg_match(y.action),
                     x.distance = x.distance,
                     y.distance = y.distance,
                     x.expansion = rep_len(x.expansion, 2),
                     y.expansion = rep_len(y.expansion, 2),
                     kept.origin = rlang::arg_match(kept.origin),
                     width = width,
                     reverse = reverse,
                     preserve = rlang::arg_match(preserve)
    )
  }

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
PositionDodgeNudgeTo <-
  ggplot2::ggproto(
    "PositionDodgeNudgeTo",
    Position,
    x = NULL,
    y = NULL,
    width = NULL,
    preserve = "total",
    orientation = "x",
    reverse = NULL,

    setup_params = function(self, data) {
      list(x = self$x,
           y = self$y,
           x.action = self$x.action,
           y.action = self$y.action,
           x.distance = self$x.distance,
           y.distance = self$y.distance,
           x.expansion = self$x.expansion,
           y.expansion = self$y.expansion,
           x.reorder = !is.null(self$x) && length(self$x) > 1 && length(self$x) < nrow(data),
           y.reorder = !is.null(self$y) && length(self$y) > 1 && length(self$y) < nrow(data),
           kept.origin = self$kept.origin,
           width = self$width,
           reverse = self$reverse %||% TRUE,
           preserve = self$preserve
      )
    },

    compute_layer = function(self, data, params, layout) {
      x_orig <- data$x
      y_orig <- data$y
      if (!is.na(params$width)) {
        # operate on the dodged positions
        data = ggplot2::ggproto_parent(ggplot2::PositionDodge, self)$compute_layer(data, params, layout)
      }
      x_dodged <- data$x
      y_dodged <- data$y

      # compute/convert x nudges
      if (!length(params$x)) {
        # set default x
        if (params$x.action == "none") {
          params$x <- rep_len(0, nrow(data))
        } else if (params$x.action == "spread") {
          params$x <- range(x_dodged)
        }
      } else if (is.numeric(params$x)) {
        # check user supplied x
        if (length(params$x) > nrow(data)) {
          warning("Argument 'x' longer than data: some values dropped!")
        }
        if (params$x.action == "none") {
          # recycle or trim x as needed
          if (params$x.reorder) {
            params$x <- rep_len(params$x, nrow(data))[order(order(data$x))] - x_dodged
          } else {
            params$x <- rep_len(params$x, nrow(data)) - x_dodged
          }
        } else if (params$x.action == "spread") {
          params$x <- range(params$x)
        }
      }

      if (params$x.action == "spread") {
        # apply x.expansion to x
        x.spread <- diff(params$x)
        params$x[1] <- params$x[1] - params$x.expansion[1] * x.spread
        params$x[2] <- params$x[2] + params$x.expansion[2] * x.spread
        if (params$x.distance == "equal") {
          # evenly spaced sequence of positions ordered as in data
          params$x <- seq(from = params$x[1],
                          to = params$x[2],
                          length.out = nrow(data))[order(order(data$x))] - x_dodged
        }
        # other strategies to distribute positions could be added here
      }

      # compute/convert y nudges
      if (!length(params$y)) {
        # set default y
        if (params$y.action == "none") {
          params$y <- rep_len(0, nrow(data))
        } else if (params$y.action == "spread") {
          params$y <- range(y_dodged)
        }
      } else if (is.numeric(params$y)) {
        # check user supplied y
        if (length(params$y) > nrow(data)) {
          warning("Argument 'y' longer than data: some values dropped!")
        }
        if (params$y.action == "none") {
          # recycle or trim y as needed
          if (params$y.reorder) {
            params$y <- rep_len(params$y, nrow(data))[order(order(data$y))] - y_dodged
          } else {
            params$y <- rep_len(params$y, nrow(data)) - y_dodged
          }
        } else if (params$y.action == "spread") {
          params$y <- range(params$y)
        }
      }

      if (params$y.action == "spread") {
        y.spread <- diff(params$y)
        params$y[1] <- params$y[1] - params$y.expansion[1] * y.spread
        params$y[2] <- params$y[2] + params$y.expansion[2] * y.spread
        if (params$y.distance == "equal") {
          # evenly spaced sequence ordered as in data
          params$y <- seq(from = params$y[1],
                          to = params$y[2],
                          length.out = nrow(data))[order(order(data$y))] - y_dodged
        }
        # other strategies could be added here
      }

      # As in 'ggplot2' we apply the nudge to xmin, xmax, xend, ymin, ymax, and yend.
      # Transform the dimensions for which not all nudges are zero
      if (any(params$x != 0)) {
        if (any(params$y != 0)) {
          data <- transform_position(data, function(x) x + params$x, function(y) y + params$y)
        } else {
          data <- transform_position(data, function(x) x + params$x, NULL)
        }
      } else if (any(params$y != 0)) {
        data <- transform_position(data, NULL, function(y) y + params$y)
      }
      # add original position
      if (params$kept.origin == "dodged" && !is.na(params$width)) {
        data$x_orig <- x_dodged
        data$y_orig <- y_dodged
      } else if (params$kept.origin == "original") {
        data$x_orig <- x_orig
        data$y_orig <- y_orig
      }

      data
    },

    compute_panel = function(self, data, params, scales) {
      ggplot2::ggproto_parent(PositionDodge, self)$compute_panel(data, params, scales)
    }
  )

#' Nudge labels to new positions
#'
#' \code{position_nudge_to()} differs from \code{\link[ggplot2]{position_nudge}}
#' in that the coordinates of the new position are given directly, rather than
#' as a displacement from the original location. It optionally sets an even
#' spacing among positions within a range. As with other position functions in
#' this package, the original positions are preserved to allow the text or
#' labels to be linked back to their original position with a segment or arrow.
#'
#' @family position adjustments
#'
#' @param x,y Coordinates of the destination position. A vector of mode
#'   \code{numeric}, that is extended if needed, to the same length as rows
#'   there are in \code{data}. The default, \code{NULL}, leaves the original
#'   coordinates unchanged.
#' @param x.action,y.action character string, one of \code{"none"}, or
#'   \code{"spread"}. With \code{"spread"} distributing the positions
#'   within the range of argument \code{x} or \code{y}, if non-null. Otherwise,
#'   using the range the variable mapped to \emph{x} or \code{y}.
#' @param x.distance,y.distance character or numeric Currently only
#'   \code{"equal"} is implemented, indicating equal spacing between the
#'   spread positions.
#' @param x.expansion,y.expansion numeric vectors of length 1 or 2, as a
#'   fraction of width of the range used to spread positions.
#' @param kept.origin One of \code{"original"} or \code{"none"}.
#'
#' @details The nudged to
#'   \code{x} and/or \code{y} values replace the original ones in
#'   \code{data}, while the original coordinates are returned in
#'   \code{x_orig} and \code{y_orig}. Nudge values supported are those of
#'   \emph{mode} numeric, thus including dates and times when they match the
#'   mapped data.
#'
#'   If the length of \code{x} and/or \code{y} is more than one but less than
#'   the rows present in the \code{data}, the vector is both recycled and
#'   reordered so that the nudges are applied sequentially based on the data
#'   values. If their length matches the number of rows in \code{data}, they are
#'   assumed to be already in \code{data} order.
#'
#'   Irrespective of the action, the ordering of rows in \code{data} is
#'   preserved.
#'
#' @note The current implementation DOES NOT support flipping geoms with the
#' \code{orientation} argument or implicitly by the mapping. It DOES NOT
#' apply scale transformations when spreading the positions.
#'
#' @return A \code{"Position"} object.
#'
#' @seealso \code{\link[ggplot2]{position_nudge}},
#'   \code{\link[ggrepel]{position_nudge_repel}}.
#'
#' @export
#'
#' @examples
#' # The examples below exemplify the features of position_nudge_to().
#' # Please see the vignette for examples of use cases.
#'
#' df <- data.frame(
#'   x = c(1,3,2,5,4,2.5),
#'   y = c(2, 3, 2.5, 1.8, 2.8, 1.5),
#'   grp = c("A", "A", "A", "B", "B", "B"),
#'   grp.inner = c("a", "b", "c", "a", "b", "c"),
#'   label = c("abc","cd","d","c","bcd","a")
#' )
#'
#' # default is no nudging
#' ggplot(df, aes(label, y, label = y)) +
#'   geom_col() +
#'   geom_text(position = position_nudge_to(),
#'             vjust = -0.2)
#'
#' # a single y (or x) value nudges all observations to this data value
#' ggplot(df, aes(label, y, label = y)) +
#'   geom_col() +
#'   geom_label(position = position_nudge_to(y = 1))
#'
#' ggplot(df, aes(x, y, label = label)) +
#'   geom_point() +
#'   geom_text(position = position_nudge_to(y = 3.2))
#'
#' ggplot(df, aes(x, y, label = label)) +
#'   geom_point() +
#'   geom_line() +
#'   geom_text(position = position_nudge_to(y = 0.1))
#'
#' # with a suitable geom, segments or arrows can be added
#' ggplot(df, aes(x, y, label = label)) +
#'   geom_point() +
#'   geom_text_s(position = position_nudge_to(y = 2.25))
#'
#' # alternating in y value order because y has fewer values than rows in data
#' ggplot(df, aes(x, y, label = label)) +
#'   geom_point() +
#'   geom_text_s(position = position_nudge_to(y = c(3, 0.1)))
#'
#' # in data row order with as many nudge y values as rows in data
#' ggplot(df, aes(x, y, label = label)) +
#'   geom_point() +
#'   geom_text_s(position = position_nudge_to(y = c(1.8, 2.3, 1.3, 2.8, 3, 0.1)))
#'
#' # spread the values at equal distance within the available space
#' ggplot(df, aes(x, y, label = label)) +
#'   geom_point() +
#'   geom_text_s(position =
#'                 position_nudge_to(y = 4, x.action = "spread"))
#'
#' ggplot(df, aes(x, y, label = label)) +
#'   geom_point() +
#'   geom_text_s(position =
#'                 position_nudge_to(y = 4, x.action = "spread")) +
#' scale_x_log10()
#'
#' # spread the values at equal distance within the expanded available space
#' ggplot(df, aes(x, y, label = label)) +
#'   geom_point() +
#'   geom_text_s(position =
#'                 position_nudge_to(y = 4, x.action = "spread", x.expansion = 0.1))
#'
#' # spread the values at equal distance within the range given by x
#' ggplot(df, aes(x, y, label = label)) +
#'   geom_point() +
#'   geom_text_s(position =
#'                 position_nudge_to(y = 4, x = c(1.5, 4), x.action = "spread"))
#'
#' # currently if scale transformations are used, the x and/or y arguments must
#' # be transformed. WARNING: This will change in the near future!!
#'
#' ggplot(df, aes(x, y, label = label)) +
#'   geom_point() +
#'   geom_text_s(position =
#'                 position_nudge_to(y = 4, x = log10(c(1.5, 4)), x.action = "spread")) +
#' scale_x_log10()
#'
#' ggplot(df, aes(x, y, label = label)) +
#'   geom_point() +
#'   geom_text_s(position =
#'                 position_nudge_to(y = log10(4), x.action = "spread")) +
#' scale_y_log10()
#'
position_nudge_to <-
  function(x = NULL,
           y = NULL,
           x.action = c("none", "spread"),
           y.action = c("none", "spread"),
           x.distance = "equal",
           y.distance = "equal",
           x.expansion = 0,
           y.expansion = 0,
           kept.origin = c("original", "none")) {

    position_dodgenudge_to(width = NA_real_, # used as flag to disable dodging
                           preserve = "total",
                           x = x,
                           y = y,
                           x.action = rlang::arg_match(x.action),
                           y.action = rlang::arg_match(y.action),
                           x.distance = x.distance,
                           y.distance = y.distance,
                           x.expansion = x.expansion,
                           y.expansion = y.expansion,
                           kept.origin = rlang::arg_match(kept.origin)
    )
  }

#' Inset tables
#'
#' \code{geom_table} and \code{geom_table_npc} add data frames as table insets
#' to the base ggplot, using syntax similar to that of
#' \code{\link[ggplot2]{geom_text}} and \code{\link{geom_text_s}}. In most
#' respects they behave as any other ggplot geometry: they add a layer
#' containing one or more grobs and grouping and faceting works as usual. The
#' most common use of \code{geom_table} is to add data labels that are whole
#' tables rather than text. \code{\link{geom_table_npc}} is used to add tables
#' as annotations to plots, but contrary to layer function \code{annotate},
#' \code{\link{geom_table_npc}} is data driven and respects grouping and facets,
#' thus plot insets can differ among panels.
#'
#' @details \code{geom_table} and \code{geom_table_npc} expect a list of data frames
#'   (\code{"data.frame"} class or derived) to be mapped to the \code{label}
#'   aesthetic. These geoms work with tibbles or data frames as \code{data} as
#'   they both support \code{list} objects as member variables.
#'
#'   A table is built with function \code{gridExtra::gtable} for each
#'   data frame in the list, and formatted according to a \code{ttheme} (table
#'   theme) list object or \code{ttheme} constructor function passed as argument
#'   to parameter \code{table.theme}. If the value passed as argument to
#'   \code{table.theme} is \code{NULL} the table theme used is that set as
#'   default through R option \code{ggpmisc.ttheme.default} at the time the plot
#'   is rendered or the \code{\link{ttheme_gtdefault}} constructor function if
#'   not set.
#'
#'   If the argument passed to \code{table.theme} or set through R option
#'   \code{ggpmisc.ttheme.default} is a constructor function (passing its name
#'   without parenthesis), the values mapped to \code{size}, \code{colour},
#'   \code{fill}, \code{alpha}, and \code{family} aesthetics will the passed to
#'   this theme constructor for each individual table. In contrast, if a ready
#'   constructed \code{ttheme} stored as a list object is passed as argument
#'   (e.g., by calling the constructor, using constructor name followed by
#'   parenthesis), it is used as is, with mappings to aesthetics \code{colour},
#'   \code{fill}, \code{alpha}, and \code{family} ignored if present. By default
#'   the constructor \code{ttheme_gtdefault} is used and \code{colour} and
#'   \code{fill}, are mapped to \code{NA}. Mapping these aesthetics to \code{NA}
#'   triggers the use the values set in the \code{ttheme}. As the table is built
#'   with function \code{gridExtra::gtable()}, for details, please, consult
#'   \code{\link[gridExtra]{tableGrob}} and \code{\link{ttheme_gtdefault}}.
#'
#'   The character strings in the data frame can be parsed into R expressions so
#'   the inset tables can include maths. With \code{parse = TRUE} parsing is
#'   attempted on each table cell, but failure triggers fall-back to rendering
#'   without parsing, on a cell by cell basis. Thus, a table can contain a
#'   mixture cells and/or headings that require parsing or not (see the
#'   documentation in \link[gridExtra]{gridExtra-package} for details).
#'
#'   The \code{x} and \code{y} aesthetics determine the position of the whole
#'   inset table, similarly to that of a text label, justification is
#'   interpreted as indicating the position of the inset table with respect to
#'   its \emph{horizontal} and \emph{vertical} axes (rows and columns in the
#'   data frame), and \code{angle} is used to rotate the inset table as a whole.
#'   The default for the \code{colour} and \code{fill} aesthetics is to retrieve
#'   them from the table theme, while \code{family} and \code{size} are
#'   retrieved from the \code{geom} element of the ggplot theme.
#'
#'   Of these two geoms only \code{\link{geom_table}} supports the plotting of
#'   connecting segments or arrows when its position has been modified by a
#'   \code{position} function. This is because \code{\link{geom_table_npc}} uses
#'   a coordinate system that is unrelated to data units, scales or data in
#'   other plot layers. In the case of \code{geom_table_npc}, \code{npcx} and
#'   \code{npcy} pseudo-aesthetics determine the position of the inset table. By
#'   default, \code{hjust} and \code{vjust} are set to \code{"inward"} to avoid
#'   clipping at the edge of the plot canvas.
#'
#' @inheritSection geom_text_s Alignment
#'
#' @inheritSection geom_text_s Position functions
#'
#' @inheritSection geom_grob Plot boundaries and clipping
#'
#' @note Complex tables with annotations or different colouring of rows or cells
#'   can be constructed with functions in package 'gridExtra' or in any other
#'   way as long as they can be saved as grid graphical objects and then added
#'   to a ggplot as a new layer with \code{\link{geom_grob}}.
#'
#' @inherit geom_grob note return seealso references
#'
#' @seealso Formatting of tables \code{stat_fmt_table},
#'   \code{\link{ttheme_gtdefault}}, \code{\link{ttheme_set}},
#'   \code{\link[gridExtra]{tableGrob}}.
#'
#' @inheritParams geom_grob
#'
#' @param table.theme NULL, list or function A gridExtra ttheme defintion, or
#'   a constructor for a ttheme or NULL for default. If \code{nULL} the theme
#'   is retrieved from R option \code{} at the time the plot is rendered.
#' @param table.rownames,table.colnames logical flag to enable or disable
#'   printing of row names and column names.
#' @param table.hjust numeric Horizontal justification for the core and column
#'   headings of the table.
#' @param parse If TRUE, the labels will be parsed into expressions and
#'   displayed as described in \code{?plotmath}.
#' @param default.colour,default.color A colour definition to use for elements
#'   not targeted by the colour aesthetic. If \code{NA} the colours in the table
#'   theme are not modified.
#' @param colour.target,color.target A vector of character strings with one or
#'   more of \code{"all"}, \code{"table"}, \code{"table.text"},
#'   \code{"table.rules"}, \code{"segment"} or \code{"none"}.
#' @param default.alpha numeric in [0..1] A transparency value to use for
#'   elements not targeted by the alpha aesthetic. If \code{NA} the alpha
#'   channel of the colour definitions is not modified.
#' @param alpha.target A vector of character strings with one or
#'   more of \code{"all"}, \code{"table"}, \code{"table.text"},
#'   \code{"table.rules"} and \code{"table.canvas"}, \code{"segment"} or
#'   \code{"none"}.
#' @param fontsize.scaling A scaling factor to apply to the \emph{size}
#'   aesthetic retrieved from the theme or mapped, applied to table text.
#'
#' @references This geometry is inspired on answers to two questions in
#'   Stackoverflow. In contrast to these earlier examples, the current geom
#'   obeys the grammar of graphics, and attempts to be consistent with the
#'   behaviour of 'ggplot2' geometries.
#'   \url{https://stackoverflow.com/questions/12318120/adding-table-within-the-plotting-region-of-a-ggplot-in-r}
#'   \url{https://stackoverflow.com/questions/25554548/adding-sub-tables-on-each-panel-of-a-facet-ggplot-in-r?}
#'
#' @family geometries adding layers with insets
#'
#' @export
#'
#' @examples
#' library(dplyr)
#' library(tibble)
#'
#' theme_set(theme_bw())
#'
#' mtcars |>
#'   group_by(cyl) |>
#'   summarize(wt = mean(wt), mpg = mean(mpg)) |>
#'   ungroup() |>
#'   mutate(wt = sprintf("%.2f", wt),
#'          mpg = sprintf("%.1f", mpg)) -> tb
#'
#' df <- data.frame(x = 5.45, y = 34, tb = I(list(tb)))
#'
#' # using defaults
#' ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
#'   geom_point() +
#'   geom_table(data = df,
#'              aes(x = x, y = y, label = tb))
#'
#' # settings aesthetics to constants
#' ggplot(mtcars,
#'        aes(wt, mpg, colour = factor(cyl))) +
#'   geom_point() +
#'   geom_table(data = df,
#'              aes(x = x, y = y, label = tb),
#'              color = "red",
#'              fill = "#FFCCCC",
#'              family = "serif", size = 5,
#'              angle = 90, vjust = 0)
#'
#' # passing a theme constructor as argument
#' ggplot(mtcars,
#'        aes(wt, mpg, colour = factor(cyl))) +
#'   geom_point() +
#'   geom_table(data = df,
#'              aes(x = x, y = y, label = tb),
#'              table.theme = ttheme_gtstripes) +
#'   theme_classic()
#'
#' # transparency
#' ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
#'   geom_point() +
#'   geom_table(data = df,
#'              aes(x = x, y = y, label = tb),
#'              alpha = 0.5) +
#'   theme_bw()
#'
#' ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
#'   geom_point() +
#'   geom_table(data = df,
#'              aes(x = x, y = y, label = tb),
#'              alpha = 0.5, alpha.target = "table.canvas")
#'
#' df2 <- tibble(x = 5.45,
#'               y = c(34, 29, 24),
#'               x1 = c(2.29, 3.12, 4.00),
#'               y1 = c(26.6, 19.7, 15.1),
#'               cyl = c(4, 6, 8),
#'               tb = list(tb[1, 1:3], tb[2, 1:3], tb[3, 1:3]))
#'
#' # mapped aesthetics
#' ggplot(mtcars,
#'        aes(wt, mpg, color = factor(cyl))) +
#'   geom_point() +
#'   geom_table(data = df2,
#'              inherit.aes = TRUE,
#'              mapping = aes(x = x, y = y, label = tb))
#'
#' ggplot(mtcars,
#'        aes(wt, mpg, color = factor(cyl))) +
#'   geom_point() +
#'   geom_table(data = df2,
#'              inherit.aes = TRUE,
#'              colour.target = "table.rules",
#'              mapping = aes(x = x, y = y, label = tb))
#'
#' # nudging and segments
#' ggplot(mtcars,
#'        aes(wt, mpg, color = factor(cyl))) +
#'   geom_point(show.legend = FALSE) +
#'   geom_table(data = df2,
#'              inherit.aes = TRUE,
#'              mapping = aes(x = x1, y = y1, label = tb),
#'              nudge_x = 0.7, nudge_y = 3,
#'              vjust = 0.5, hjust = 0.5,
#'              arrow = arrow(length = unit(0.5, "lines"))) +
#'   theme_classic()
#'
#' ggplot(mtcars,
#'        aes(wt, mpg, color = factor(cyl))) +
#'   geom_point(show.legend = FALSE) +
#'   geom_table(data = df2,
#'              inherit.aes = TRUE,
#'              mapping = aes(x = x1, y = y1, label = tb),
#'              nudge_x = 0.7, nudge_y = 3,
#'              vjust = 0.5, hjust = 0.5,
#'              arrow = arrow(length = unit(0.5, "lines")),
#'              colour.target = c("table.rules", "segment")) +
#'   theme_classic()
#'
#' # Using native plot coordinates instead of data coordinates
#' dfnpc <- tibble(x = 0.95, y = 0.95, tb = list(tb))
#'
#' ggplot(mtcars,
#'        aes(wt, mpg, colour = factor(cyl))) +
#'   geom_point() +
#'   geom_table_npc(data = dfnpc,
#'                  aes(npcx = x, npcy = y, label = tb))
#'
geom_table <- function(mapping = NULL,
                       data = NULL,
                       stat = "identity",
                       position = "identity",
                       ...,
                       nudge_x = 0,
                       nudge_y = 0,
                       default.colour = NA,
                       default.color = default.colour,
                       colour.target = "table.text",
                       color.target = colour.target,
                       default.alpha = 1,
                       alpha.target = "all",
                       fontsize.scaling = 0.825,
                       add.segments = TRUE,
                       box.padding = 0.25,
                       point.padding = 1e-06,
                       segment.linewidth = 0.5,
                       min.segment.length = 0,
                       arrow = NULL,
                       table.theme = NULL,
                       table.rownames = FALSE,
                       table.colnames = TRUE,
                       table.hjust = 0.5,
                       parse = FALSE,
                       na.rm = FALSE,
                       show.legend = FALSE,
                       inherit.aes = FALSE) {

  colour.target <-
    rlang::arg_match(color.target,
                     values = c("segment", "all", "box", "text", "table",
                                "table.text", "table.rules", "none"),
                     multiple = TRUE)
  alpha.target <-
    rlang::arg_match(alpha.target,
                     values = c("segment", "all", "box", "text", "table",
                                "table.text", "table.rules", "table.canvas",
                                "none"),
                     multiple = TRUE)

  if (!missing(nudge_x) || !missing(nudge_y)) {
    if (!missing(position) && position != "identity") {
      rlang::abort("You must specify either `position` or `nudge_x`/`nudge_y`.")
    }
    # We do not keep the original positions if they will not be used
    position <-
      position_nudge_center(nudge_x, nudge_y,
                            kept.origin = ifelse(add.segments,
                                                 "original", "none"))
  }

  if (is.character(table.hjust)) {
    table.hjust <- switch(table.hjust,
                          left = 0,
                          middle = 0.5,
                          center = 0.5,
                          right = 1,
                          0.5)
  }

  ggplot2::layer(
    data = data,
    mapping = mapping,
    stat = stat,
    geom = GeomTable,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      default.colour = default.color,
      colour.target = colour.target,
      default.alpha = default.alpha,
      alpha.target = alpha.target,
      fontsize.scaling = fontsize.scaling,
      add.segments = add.segments,
      box.padding = box.padding,
      point.padding = point.padding,
      segment.linewidth = segment.linewidth,
      min.segment.length = min.segment.length,
      arrow = arrow,
      table.theme = table.theme,
      table.rownames = table.rownames,
      table.colnames = table.colnames,
      table.hjust = table.hjust,
      parse = parse,
      na.rm = na.rm,
      ...
    )
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
GeomTable <-
  ggplot2::ggproto("GeomTable", ggplot2::Geom,
                   required_aes = c("x", "y", "label"),
                   default_aes = ggplot2::aes(
                     colour = NA,
                     fill = NA,
                     family = "",
                     size = 3.2,
                     angle = 0,
                     hjust = "inward",
                     vjust = "inward",
                     alpha = NA,
                     fontface = 1,
                     lineheight = 1.2
                   ),

          draw_panel = function(data,
                                panel_params,
                                coord,
                                add.segments = TRUE,
                                box.padding = 0.25,
                                point.padding = 1e-06,
                                segment.linewidth = 0.5,
                                min.segment.length = 0,
                                arrow = NULL,
                                table.theme = NULL,
                                table.rownames = FALSE,
                                table.colnames = TRUE,
                                table.hjust = 0.5,
                                parse = FALSE,
                                default.colour = NA,
                                colour.target = "table.text",
                                default.alpha = NA,
                                alpha.target = "all",
                                fontsize.scaling = 0.825,
                                na.rm = FALSE) {

            if (nrow(data) == 0) {
              return(grid::nullGrob())
            }

            if (!is.data.frame(data$label[[1]])) {
              warning("Skipping as object mapped to 'label' is not a list of ",
                      "\"tibble\" or \"data.frame\" objects.")
              return(grid::nullGrob())
            }

            # replace NULL with default at time of rendering!
            if (is.null(table.theme)) {
              table.theme <-
                getOption("ggpmisc.ttheme.default",
                          default = ggpp::ttheme_gtdefault)
            }

            default.colour <- check_default_colour(default.colour)

            add.segments <-
              add.segments && all(c("x_orig", "y_orig") %in% colnames(data))

            # should be called only once!
            data <- coord$transform(data, panel_params)
            if (add.segments) {
              data_orig <- data.frame(x = data$x_orig, y = data$y_orig)
              data_orig <- coord$transform(data_orig, panel_params)
              data$x_orig <- data_orig$x
              data$y_orig <- data_orig$y
            }

            if (is.character(data$vjust)) {
              data$vjust <-
                compute_just2d(data = data,
                               coord = coord,
                               panel_params = panel_params,
                               just = data$vjust,
                               a = "y", b = "x")
            }
            if (is.character(data$hjust)) {
              data$hjust <-
                compute_just2d(data = data,
                               coord = coord,
                               panel_params = panel_params,
                               just = data$hjust,
                               a = "x", b = "y")
            }
            if (add.segments) {
              segments.data <-
                shrink_segments(data,
                                point.padding = point.padding,
                                box.padding = box.padding,
                                min.segment.length = min.segment.length)
            }
            # for the table itself default.colour NA uses the table theme
            # box and segment use "black" as default as in earlier versions
            if (is.na(default.colour)) {
              off.table.default.colour <- "black"
            } else {
              off.table.default.colour <- default.colour
            }

            # loop needed as gpar is not vectorized
            all.grobs <- grid::gList()
            for (row.idx in seq_len(nrow(data))) {
              row <- data[row.idx, , drop = FALSE]
              # Build the table
              if (is.function(table.theme)) {
                # tableGrob puts all the padding on the same side unless
                # just = this makes it difficult to compute a suitable value
                # for table.x without knowing the width of the column. The code
                # here at least ensures that whatever its length the whole text
                # is always displayed.
                table.x <- table.hjust
                # alpha
                text.alpha <-
                  ifelse(any(alpha.target %in% c("all", "table", "table.text", "text")),
                         row$alpha, default.alpha)
                rules.alpha <-
                  ifelse(any(alpha.target %in% c("all", "table", "table.rules")),
                         row$alpha, default.alpha)
                canvas.alpha <-
                  ifelse(any(alpha.target %in% c("all", "table", "table.canvas")),
                         row$alpha, default.alpha)
                segment.alpha <-
                  ifelse(any(alpha.target %in% c("all", "segment")),
                         row$alpha, default.alpha)
                box.alpha <-
                  ifelse(any(alpha.target %in% c("all", "box")),
                         row$alpha, default.alpha)
                # colour
                text.colour <-
                  ifelse(any(colour.target %in% c("all", "table", "table.text", "text")),
                         row$colour, default.colour)
                # text.alpha applied in ttheme constructor
                rules.colour <-
                  ifelse(any(colour.target %in% c("all", "table", "table.rules")),
                         row$colour, default.colour) |>
                  ggplot2::alpha(rules.alpha)
                segment.colour <-
                  ifelse(any(colour.target %in% c("all", "segment")),
                         row$colour, off.table.default.colour) |>
                  ggplot2::alpha(segment.alpha)
                box.colour <-
                  ifelse(any(colour.target %in% c("all", "box")),
                         row$colour, off.table.default.colour) |>
                  ggplot2::alpha(box.alpha)
                # fill
                canvas.fill <- row$fill |>
                  ggplot2::alpha(canvas.alpha)

                if (is.na(canvas.fill)) {
                  core.params <-
                    list(fg_params = list(hjust = table.hjust, x = table.x))
                  rowhead.params <- list(fg_params = list(hjust = 1, x = 0.9))
                  colhead.params <- list(fg_params = list(hjust = table.hjust,
                                                          x = table.x))
                } else {
                  # override ttheme fill for background canvas
                  core.params <-
                    list(fg_params = list(hjust = table.hjust, x = table.x),
                         bg_params = list(fill = canvas.fill))
                  rowhead.params <-
                    list(fg_params = list(hjust = 1, x = 0.9),
                         bg_params = list(fill = canvas.fill))
                  colhead.params <- list(fg_params = list(hjust = table.hjust,
                                                          x = table.x),
                                         bg_params = list(fill = canvas.fill))
                }
                # override ttheme colour for background rules
                if (!is.na(rules.colour)) {
                  rules.colour <- ggplot2::alpha(rules.colour, rules.alpha)
                  core.params$bg_params$col <- rules.colour
                  rowhead.params$bg_params$col <- rules.colour
                  colhead.params$bg_params$col <- rules.colour
                }
                if (is.na(text.colour)) {
                  # use theme's default base_colour
                  this.table.theme <-
                    table.theme(base_size = row$size * fontsize.scaling * .pt,
                                base_family = row$family,
                                parse = parse,
                                text.alpha = text.alpha,
                                canvas.alpha = canvas.alpha,
                                rules.alpha = rules.alpha,
                                rowhead = rowhead.params,
                                colhead = colhead.params,
                                core = core.params)
                } else {
                  # use colour from data$colour
                  this.table.theme <-
                    table.theme(base_size = row$size * fontsize.scaling * .pt,
                                base_colour = text.colour,
                                base_family = row$family,
                                parse = parse,
                                text.alpha = text.alpha,
                                canvas.alpha = canvas.alpha,
                                rules.alpha = rules.alpha,
                                rowhead = rowhead.params,
                                colhead = colhead.params,
                                core = core.params)
                }
              } else if (is.list(table.theme)) {
                this.table.theme <- table.theme
              }

              table.tb <- data[["label"]][[row.idx]]
              user.grob <-
                gridExtra::tableGrob(
                  d = table.tb,
                  theme = this.table.theme,
                  rows = if (table.rownames) rownames(table.tb) else NULL,
                  cols = if (table.colnames) colnames(table.tb) else NULL
                )

              user.grob$vp <-
                grid::viewport(x = grid::unit(row$x, "native"),
                               y = grid::unit(row$y, "native"),
                               width = sum(user.grob$widths),
                               height = sum(user.grob$heights),
                               just = c(row$hjust, row$vjust),
                               angle = row$angle,
                               name = paste("inset.table.vp", row$PANEL,
                                            "row", row.idx, sep = "."))

              # give unique name to each grob
              user.grob$name <- paste("inset.table", row.idx, sep = ".")

              if (add.segments) {
                segment.row <- segments.data[row.idx, , drop = FALSE]
                if (segment.row$too.short) {
                  segment.grob <- grid::nullGrob()
                } else {
                  segment.grob <-
                    grid::segmentsGrob(x0 = segment.row$x,
                                       y0 = segment.row$y,
                                       x1 = segment.row$x_orig,
                                       y1 = segment.row$y_orig,
                                       arrow = arrow,
                                       gp = grid::gpar(
                                         col = if (segment.linewidth == 0) NA else
                                           # lwd = 0 is invalid in 'grid'
                                           segment.colour,
                                         lwd = if (segment.linewidth == 0) 0.5 else
                                           segment.linewidth * ggplot2::.stroke),
                                       name = paste("table.s.segment",
                                                    row$group,
                                                    row.idx, sep = "."))
                }
                all.grobs <- grid::gList(all.grobs, segment.grob, user.grob)
              } else {
                all.grobs <- grid::gList(all.grobs, user.grob)
              }
            }
            #    grid::grobTree(children = all.grobs, name = "geom.table.panel")
            grid::grobTree(children = all.grobs)

          },

          draw_key = function(...) {
            grid::nullGrob()
          }
  )

#' @rdname geom_table
#' @export
#'
geom_table_npc <- function(mapping = NULL, data = NULL,
                           stat = "identity", position = "identity",
                           ...,
                           default.colour = NA,
                           default.color = default.colour,
                           colour.target = "table.text",
                           color.target = colour.target,
                           default.alpha = 1,
                           alpha.target = "all",
                           fontsize.scaling = 0.825,
                           table.theme = NULL,
                           table.rownames = FALSE,
                           table.colnames = TRUE,
                           table.hjust = 0.5,
                           parse = FALSE,
                           na.rm = FALSE,
                           show.legend = FALSE,
                           inherit.aes = FALSE) {

  colour.target <-
    rlang::arg_match(color.target,
                     values = c("segment", "all", "box", "text", "table",
                                "table.text", "table.rules", "none"),
                     multiple = TRUE)
  alpha.target <-
    rlang::arg_match(alpha.target,
                     values = c("segment", "all", "box", "text", "table",
                                "table.text", "table.rules", "table.canvas",
                                "none"),
                     multiple = TRUE)

  if (is.character(table.hjust)) {
    table.hjust <- switch(table.hjust,
                          left = 0,
                          middle = 0.5,
                          center = 0.5,
                          right = 1,
                          0.5)
  }

  ggplot2::layer(
    data = data,
    mapping = mapping,
    stat = stat,
    geom = GeomTableNpc,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      default.colour = default.color,
      colour.target = colour.target,
      default.alpha = default.alpha,
      alpha.target = alpha.target,
      fontsize.scaling = fontsize.scaling,
      table.theme = table.theme,
      table.rownames = table.rownames,
      table.colnames = table.colnames,
      table.hjust = table.hjust,
      parse = parse,
      na.rm = na.rm,
      ...
    )
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
GeomTableNpc <-
  ggplot2::ggproto("GeomTableNpc", ggplot2::Geom,
          required_aes = c("npcx", "npcy", "label"),

          default_aes = ggplot2::aes(
            colour = NA,
            fill = NA,
            size = 3.2,
            angle = 0,
            hjust = "inward",
            vjust = "inward",
            alpha = NA,
            family = "",
            fontface = 1,
            lineheight = 1.2
          ),

          draw_panel =
            function(data,
                     panel_params,
                     coord,
                     table.theme = NULL,
                     table.rownames = FALSE,
                     table.colnames = TRUE,
                     table.hjust = 0.5,
                     parse = FALSE,
                     default.colour = NA,
                     colour.target = "table.text",
                     default.alpha = NA,
                     alpha.target = "all",
                     fontsize.scaling = 0.825,
                     na.rm = FALSE) {

              if (nrow(data) == 0) {
                return(grid::nullGrob())
              }

              if (!is.data.frame(data$label[[1]])) {
                warning("Skipping as object mapped to 'label' is not a list of ",
                        "\"tibble\" or \"data.frame\" objects.")
                return(grid::nullGrob())
              }

              # replace NULL with default
              if (is.null(table.theme)) {
                table.theme <-
                  getOption("ggpmisc.ttheme.default",
                            default = ggpp::ttheme_gtdefault)
              }

              default.colour <- check_default_colour(default.colour)

              data$npcx <- compute_npcx(data$npcx)
              data$npcy <- compute_npcy(data$npcy)

              if (is.character(data$vjust)) {
                data$vjust <- compute_just(data$vjust, data$npcy)
              }
              if (is.character(data$hjust)) {
                data$hjust <- compute_just(data$hjust, data$npcx)
              }

              # for the table itself default.colour NA uses the table theme
              # box and segment use "black" as default as in earlier versions
              if (is.na(default.colour)) {
                off.table.default.colour <- "black"
              } else {
                off.table.default.colour <- default.colour
              }

              tb.grobs <- grid::gList()
              for (row.idx in seq_len(nrow(data))) {
                row <- data[row.idx, , drop = FALSE]
                # Build the table
                if (is.function(table.theme)) {
                  # tableGrob puts all the padding on the same side unless just = 0.5
                  # this makes it difficult to compute a suitable value for table.x
                  # without knowing the width of the column. The code here at least
                  # ensures that whatever its length the whole text is always displayed.
                  table.x <- table.hjust
                  # text position in cell depends on hjust
#                  table.x <- if(table.hjust == 0.5) 0.5 else table.hjust * 0.8 + 0.1
                  # alpha
                  text.alpha <-
                    ifelse(any(alpha.target %in% c("all", "table", "table.text", "text")),
                           row$alpha, default.alpha)
                  rules.alpha <-
                    ifelse(any(alpha.target %in% c("all", "table", "table.rules")),
                           row$alpha, default.alpha)
                  canvas.alpha <-
                    ifelse(any(alpha.target %in% c("all", "table", "table.canvas")),
                           row$alpha, default.alpha)
                  box.alpha <-
                    ifelse(any(alpha.target %in% c("all", "box")),
                           row$alpha, default.alpha)
                  # colour
                  text.colour <-
                    ifelse(any(colour.target %in% c("all", "table", "table.text", "text")),
                           row$colour, default.colour) |>
                    ggplot2::alpha(text.alpha)
                  rules.colour <-
                    ifelse(any(colour.target %in% c("all", "table", "table.rules")),
                           row$colour, default.colour) |>
                    ggplot2::alpha(rules.alpha)
                  box.colour <-
                    ifelse(any(colour.target %in% c("all", "box")),
                           row$colour, off.table.default.colour) |>
                    ggplot2::alpha(box.alpha)
                  # fill
                  canvas.fill <- row$fill |>
                    ggplot2::alpha(canvas.alpha)

                  if (is.na(canvas.fill)) {
                    core.params <-
                      list(fg_params = list(hjust = table.hjust, x = table.x))
                    rowhead.params <- list(fg_params = list(hjust = 1, x = 0.9))
                    colhead.params <- list(fg_params = list(hjust = table.hjust,
                                                            x = table.x))
                  } else {
                    # override ttheme fill for background canvas
                    core.params <-
                      list(fg_params = list(hjust = table.hjust, x = table.x),
                           bg_params = list(fill = canvas.fill))
                    rowhead.params <-
                      list(fg_params = list(hjust = 1, x = 0.9),
                           bg_params = list(fill = canvas.fill))
                    colhead.params <- list(fg_params = list(hjust = table.hjust,
                                                            x = table.x),
                                           bg_params = list(fill = canvas.fill))
                  }
                  # override ttheme colour for background rules
                  if (!is.na(rules.colour)) {
                    rules.colour <- ggplot2::alpha(rules.colour, rules.alpha)
                    core.params$bg_params$col <- rules.colour
                    rowhead.params$bg_params$col <- rules.colour
                    colhead.params$bg_params$col <- rules.colour
                  }
                  if (is.na(text.colour)) {
                    # use theme's default base_colour
                    this.table.theme <-
                      table.theme(base_size = row$size * fontsize.scaling * .pt,
                                  base_family = row$family,
                                  parse = parse,
                                  text.alpha = text.alpha,
                                  canvas.alpha = canvas.alpha,
                                  rules.alpha = rules.alpha,
                                  rowhead = rowhead.params,
                                  colhead = colhead.params,
                                  core = core.params)
                  } else {
                    # use colour from data$colour
                    this.table.theme <-
                      table.theme(base_size = row$size * fontsize.scaling * .pt,
                                  base_colour = text.colour,
                                  base_family = row$family,
                                  parse = parse,
                                  text.alpha = text.alpha,
                                  canvas.alpha = canvas.alpha,
                                  rules.alpha = rules.alpha,
                                  rowhead = rowhead.params,
                                  colhead = colhead.params,
                                  core = core.params)
                  }
                } else if (is.list(table.theme)) {
                  this.table.theme <- table.theme
                }

                table.tb <- data[["label"]][[row.idx]]
                gtb <-
                  gridExtra::tableGrob(
                    d = table.tb,
                    theme = this.table.theme,
                    rows = if (table.rownames) rownames(table.tb) else NULL,
                    cols = if (table.colnames) colnames(table.tb) else NULL
                  )

                gtb$vp <-
                  grid::viewport(x = grid::unit(data$npcx[row.idx], "native"),
                                 y = grid::unit(data$npcy[row.idx], "native"),
                                 width = sum(gtb$widths),
                                 height = sum(gtb$heights),
                                 just = c(data$hjust[row.idx], data$vjust[row.idx]),
                                 angle = data$angle[row.idx],
                                 name = paste("geom_table.panel", data$PANEL[row.idx],
                                              "row", row.idx, sep = "."))

                # give unique name to each table
                gtb$name <- paste("table", row.idx, sep = ".")

                tb.grobs[[row.idx]] <- gtb
              }

              grid::grobTree(children = tb.grobs)
            },

          draw_key = function(...) {
            grid::nullGrob()
          }
  )

#' Table themes
#'
#' Additional theme constructors for use with \code{\link{geom_table}}.
#'
#' @details These wrapper functions are table theme (\code{ttheme}) constructors
#'   making it easier to change the style of tables created with
#'   \code{\link[gridExtra]{tableGrob}}. When passed as argument to
#'   \code{\link{geom_table}} the table theme's \code{base_colour},
#'   \code{base_family}, \code{base_colour} and \code{base_size} function as
#'   defaults for the text in the body of the table. They are overriden if the
#'   corresponding text-related aesthetics are mapped or set to a constant
#'   through the usual 'ggplot2' mechanisms. On the other hand the properties
#'   of the background fill, rules and column and row headings can be set only
#'   through the theme. The \code{ttheme} constructors defined in
#'   'ggpp' have formal parameters for \code{alpha} transparency of the
#'   text, background and rules. Transparency is useful as plot insets can
#'   accidentally overlap observations hiding them from view depending on the
#'   stacking order of plot layers.
#'
#'   These theme constructors are wrappers on the constructors
#'   \code{gridExtra::ttheme_default()} and \code{gridExtra::ttheme_minimal()}.
#'   They can also be used directly with \code{\link[gridExtra]{grid.table}} if
#'   desired.
#'
#' @param base_size numeric, default font size of text in table.
#' @param base_colour	default font colour for text in table.
#' @param base_family	default font family for text in table.
#' @param parse	logical, behaviour for parsing text as plotmath.
#' @param padding length-2 unit vector specifying the horizontal and vertical
#'   padding of text within each cell.
#' @param text.alpha,canvas.alpha,rules.alpha numeric in [0..1] Transparency
#'   applied to table \code{base_colour}, to table body background
#'   \code{fill} and rules \code{colour}, respectively.
#' @param ... further arguments to control the gtable.
#'
#' @return A \code{list} object that can be used as \code{ttheme} in the
#'   construction of tables with functions from package 'gridExtra'.
#'
#' @export
#'
#' @family geometries for adding insets to ggplots
#'
#' @examples
#' library(dplyr)
#' library(tibble)
#'
#' mtcars |>
#'   group_by(cyl) |>
#'   summarize(wt = mean(wt), mpg = mean(mpg)) |>
#'   ungroup() |>
#'   mutate(wt = sprintf("%.2f", wt),
#'          mpg = sprintf("%.1f", mpg)) -> tb
#'
#' df <- tibble(x = 5.45, y = 34, tb = list(tb))
#'
#' # Same as the default theme constructor
#' ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
#'   geom_point() +
#'   geom_table(data = df, aes(x = x, y = y, label = tb),
#'              table.theme = ttheme_gtdefault) +
#'   theme_classic()
#'
#' # Minimal theme constructor
#' ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
#'   geom_point() +
#'   geom_table(data = df, aes(x = x, y = y, label = tb),
#'              table.theme = ttheme_gtminimal) +
#'   theme_classic()
#'
#' # A theme with white background
#' ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
#'   geom_point() +
#'   geom_table(data = df, aes(x = x, y = y, label = tb),
#'              table.theme = ttheme_gtbw) +
#'   theme_bw()
#'
#' # Base colour of theme overridden by aesthetic constant
#' ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
#'   geom_point() +
#'   geom_table(data = df, aes(x = x, y = y, label = tb),
#'              table.theme = ttheme_gtbw, colour = "darkblue") +
#'   theme_bw()
#'
#' # A theme with dark background
#' ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
#'   geom_point() +
#'   geom_table(data = df, aes(x = x, y = y, label = tb),
#'              table.theme = ttheme_gtdark) +
#'   theme_dark()
#'
#' # Base colour of theme overridden by aesthetic constant
#' ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
#'   geom_point() +
#'   geom_table(data = df, aes(x = x, y = y, label = tb),
#'              table.theme = ttheme_gtdark, colour = "yellow") +
#'   theme_dark()
#'
#' # A theme with light background
#' ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
#'   geom_point() +
#'   geom_table(data = df, aes(x = x, y = y, label = tb),
#'              table.theme = ttheme_gtlight)
#'
#' # Base colour of theme overridden by aesthetic constant
#' ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
#'   geom_point() +
#'   geom_table(data = df, aes(x = x, y = y, label = tb),
#'              table.theme = ttheme_gtlight, colour = "darkred")
#'
#' # Base colour of theme overridden by aesthetic constant
#' ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
#'   geom_point() +
#'   geom_table(data = df, aes(x = x, y = y, label = tb),
#'              table.theme = ttheme_gtsimple)
#'
#' # Default colour of theme superceded by aesthetic constant
#' ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
#'   geom_point() +
#'   geom_table(data = df, aes(x = x, y = y, label = tb),
#'              table.theme = ttheme_gtstripes) +
#'   theme_dark()
#'
#' # Transparency of table background fill and grid lines colour
#' ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
#'   geom_point() +
#'   geom_table(data = df, aes(x = 1.5, y = y, label = tb),
#'              table.theme = ttheme_gtplain(canvas.alpha = 0.5,
#'                                           rules.alpha = 0.2)) +
#'   theme_classic()
#'
#' # Transparency of table background fill and grid lines colour
#' # and table text base colour: black with 50% transparency
#' ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
#'   geom_point() +
#'   geom_table(data = df, aes(x = 1.5, y = y, label = tb),
#'              table.theme = ttheme_gtplain(text.alpha = 0.5)) +
#'   theme_classic()
#'
ttheme_gtdefault <- function(base_size = 10,
                             base_colour = "black",
                             base_family = "",
                             parse = FALSE,
                             padding = grid::unit(c(0.8, 0.6), "char"),
                             text.alpha = NA,
                             rules.alpha = NA,
                             canvas.alpha = NA,
                             ...) {
  core <-
    list(bg_params =
           list(fill = ggplot2::alpha(c("grey95", "grey90"), canvas.alpha),
                col = ggplot2::alpha("white", rules.alpha)),
         fg_params =
           list(col = ggplot2::alpha(base_colour, text.alpha)))
  colhead <-
    list(bg_params = list(fill = ggplot2::alpha("grey80", canvas.alpha),
                          col = ggplot2::alpha("white", rules.alpha)),
         fg_params =
           list(col = ggplot2::alpha(base_colour, text.alpha)))
  rowhead <-
    list(bg_params = list(fill = ggplot2::alpha("grey80", canvas.alpha),
                          col = ggplot2::alpha("white", rules.alpha)),
         fg_params =
           list(col = ggplot2::alpha(base_colour, text.alpha)))

  default <-
    gridExtra::ttheme_default(base_size = base_size,
                              base_colour = base_colour,
                              base_family = base_family,
                              parse = parse,
                              padding = padding,
                              core = core,
                              colhead = colhead,
                              rowhead = rowhead,
                              ...)

  utils::modifyList(default, list(...))
}

#' @rdname ttheme_gtdefault
#'
#' @export
#'
ttheme_gtminimal <- function(base_size = 10,
                             base_colour = "black",
                             base_family = "",
                             parse = FALSE,
                             padding = grid::unit(c(0.5, 0.4), "char"),
                             text.alpha = NA,
                             rules.alpha = NA,
                             canvas.alpha = NA,
                             ...) {
  default <-
    gridExtra::ttheme_minimal(base_size = base_size,
                              base_colour = base_colour,
                              base_family = base_family,
                              parse = parse,
                              padding = padding,
                              ...)

  utils::modifyList(default, list(...))
}

#' @rdname ttheme_gtdefault
#'
#' @export
#'
ttheme_gtbw <- function(base_size = 10,
                        base_colour = "black",
                        base_family = "",
                        parse = FALSE,
                        padding = grid::unit(c(1, 0.6), "char"),
                        text.alpha = NA,
                        rules.alpha = NA,
                        canvas.alpha = NA,
                        ...) {
  core <-
    list(bg_params =
           list(fill = ggplot2::alpha("white", canvas.alpha),
                lwd = 1.5,
                col = ggplot2::alpha("grey90", rules.alpha)),
         fg_params =
           list(col = ggplot2::alpha(base_colour, text.alpha)))
  colhead <- rowhead <-
    list(bg_params = list(fill = ggplot2::alpha("grey80", canvas.alpha),
                          lwd = 1.5,
                          col = ggplot2::alpha("grey90", rules.alpha)),
         fg_params =
           list(col = ggplot2::alpha(base_colour, text.alpha)))

  default <-
    gridExtra::ttheme_default(base_size = base_size,
                              base_colour = base_colour,
                              base_family = base_family,
                              parse = parse,
                              padding = padding,
                              core = core,
                              colhead = colhead,
                              rowhead = rowhead)

  utils::modifyList(default, list(...))
}

#' @rdname ttheme_gtdefault
#'
#' @export
#'
ttheme_gtplain <- function(base_size = 10,
                           base_colour = "black",
                           base_family = "",
                           parse = FALSE,
                           padding = grid::unit(c(0.8, 0.6), "char"),
                           text.alpha = NA,
                           rules.alpha = NA,
                           canvas.alpha = NA,
                           ...) {
  core <-
    list(bg_params =
           list(fill = ggplot2::alpha("white", canvas.alpha)),
         fg_params =
           list(col = ggplot2::alpha(base_colour, text.alpha)))
  colhead <- rowhead <-
    list(bg_params = list(fill = ggplot2::alpha("grey90", canvas.alpha)),
         fg_params =
           list(col = ggplot2::alpha(base_colour, text.alpha)))
  default <-
    gridExtra::ttheme_default(base_size = base_size,
                              base_colour = base_colour,
                              base_family = base_family,
                              parse = parse,
                              padding = padding,
                              core = core,
                              colhead = colhead,
                              rowhead = rowhead)

  utils::modifyList(default, list(...))
}

#' @rdname ttheme_gtdefault
#'
#' @export
#'
ttheme_gtdark <- function(base_size = 10,
                          base_colour = "grey90",
                          base_family = "",
                          parse = FALSE,
                          padding = grid::unit(c(0.8, 0.6), "char"),
                          text.alpha = NA,
                          rules.alpha = NA,
                          canvas.alpha = NA,
                          ...) {
  core <-
    list(bg_params =
           list(fill = ggplot2::alpha("grey30", canvas.alpha),
                lwd = 1.5,
                col = ggplot2::alpha(base_colour, rules.alpha)),
         fg_params =
           list(col = ggplot2::alpha(base_colour, text.alpha)))
  colhead <- rowhead <-
    list(bg_params = list(fill = ggplot2::alpha("black", canvas.alpha),
                          lwd = 1.5,
                          col = ggplot2::alpha(base_colour, rules.alpha)),
         fg_params =
           list(col = ggplot2::alpha(base_colour, text.alpha)))
  default <-
    gridExtra::ttheme_default(base_size = base_size,
                              base_colour = base_colour,
                              base_family = base_family,
                              parse = parse,
                              padding = padding,
                              core = core,
                              colhead = colhead,
                              rowhead = rowhead)

  utils::modifyList(default, list(...))
}

#' @rdname ttheme_gtdefault
#'
#' @export
#'
ttheme_gtlight <- function(base_size = 10,
                           base_colour = "grey10",
                           base_family = "",
                           parse = FALSE,
                           padding = grid::unit(c(0.8, 0.6), "char"),
                           text.alpha = NA,
                           rules.alpha = NA,
                           canvas.alpha = NA,
                           ...) {
  core <-
    list(bg_params =
           list(fill = ggplot2::alpha("white", canvas.alpha),
                lwd = 1.5,
                col = ggplot2::alpha(base_colour, rules.alpha)),
         fg_params =
           list(col = ggplot2::alpha(base_colour, text.alpha)))
  colhead <- rowhead <-
    list(bg_params = list(fill = ggplot2::alpha("grey80", canvas.alpha),
                          lwd = 1.5,
                          col = ggplot2::alpha(base_colour, rules.alpha)),
         fg_params =
           list(col = ggplot2::alpha(base_colour, text.alpha)))

  default <-
    gridExtra::ttheme_default(base_size = base_size,
                              base_colour = base_colour,
                              base_family = base_family,
                              parse = parse,
                              padding = padding,
                              core = core,
                              colhead = colhead,
                              rowhead = rowhead)

  utils::modifyList(default, list(...))
}

#' @rdname ttheme_gtdefault
#'
#' @export
#'
ttheme_gtsimple <- function(base_size = 10,
                            base_colour = "grey10",
                            base_family = "",
                            parse = FALSE,
                            padding = grid::unit(c(0.5, 0.4), "char"),
                            text.alpha = NA,
                            rules.alpha = NA,
                            canvas.alpha = NA,
                            ...) {
  core <-
    list(bg_params =
           list(fill = ggplot2::alpha("white", canvas.alpha),
                lwd = 0,
                col = NA),
         fg_params =
           list(col = ggplot2::alpha(base_colour, text.alpha)))
  colhead <- rowhead <-
    list(bg_params = list(fill = ggplot2::alpha("grey80", canvas.alpha),
                          lwd = 0,
                          col = NA),
         fg_params =
           list(col = ggplot2::alpha(base_colour, text.alpha)))

  default <-
    gridExtra::ttheme_default(base_size = base_size,
                              base_colour =base_colour,
                              base_family = base_family,
                              parse = parse,
                              padding = padding,
                              core = core,
                              colhead = colhead,
                              rowhead = rowhead)

  utils::modifyList(default, list(...))
}

#' @rdname ttheme_gtdefault
#'
#' @export
#'
ttheme_gtstripes <- function(base_size = 10,
                             base_colour = "grey10",
                             base_family = "",
                             parse = FALSE,
                             padding = grid::unit(c(0.8, 0.6), "char"),
                             text.alpha = NA,
                             rules.alpha = NA,
                             canvas.alpha = NA,
                             ...) {
  core <-
    list(bg_params =
           list(fill = ggplot2::alpha(c("white", "grey90"), canvas.alpha),
                lwd = 0,
                col = NA),
         fg_params =
           list(col = ggplot2::alpha(base_colour, text.alpha)))
  rowhead <- colhead <-
    list(bg_params = list(fill = ggplot2::alpha("grey75", canvas.alpha),
                          lwd = 0,
                          col = NA),
         fg_params =
           list(col = ggplot2::alpha(base_colour, text.alpha)))
  # colhead <-
  #   list(bg_params = list(fill = ggplot2::alpha("white", canvas.alpha),
  #                         lwd = 0,
  #                         col = NA))
  default <-
    gridExtra::ttheme_default(base_size = base_size,
                              base_colour = base_colour,
                              base_family = base_family,
                              parse = parse,
                              padding = padding,
                              core = core,
                              colhead = colhead,
                              rowhead = rowhead)

  utils::modifyList(default, list(...))
}

#' Set default table theme
#'
#' Set R option to the theme to use as current default. This function is
#' implemented differently but is used in the same way as
#' \code{ggplot2::theme_set()} but affects the default table-theme instead
#' of the plot theme.
#'
#' @note The ttheme is set when a plot object is constructed, and consequently
#' the option setting does not affect rendering of ready built plot objects.
#'
#' @param table.theme NULL, list or function A gridExtra ttheme defintion, or
#'   a constructor for a ttheme or NULL for default.
#'
#' @return A named list with the previous value of the option.
#'
#' @export
#'
#' @examples
#' library(dplyr)
#' library(tibble)
#'
#' mtcars |>
#'   group_by(cyl) |>
#'   summarize(wt = mean(wt), mpg = mean(mpg)) |>
#'   ungroup() |>
#'   mutate(wt = sprintf("%.2f", wt),
#'          mpg = sprintf("%.1f", mpg)) -> tb
#'
#' df <- tibble(x = 5.45, y = 34, tb = list(tb))
#'
#' # Same as the default theme constructor
#' ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
#'   geom_point() +
#'   geom_table(data = df, aes(x = x, y = y, label = tb))
#'
#' # set a new default
#' old_ttheme <- ttheme_set(ttheme_gtstripes)
#'
#' ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
#'   geom_point() +
#'   geom_table(data = df, aes(x = x, y = y, label = tb))
#'
#' # restore previous setting
#' ttheme_set(old_ttheme)
#'
ttheme_set <- function(table.theme = NULL) {
  stopifnot("Bad argument passed to 'table.theme'" =
              is.null(table.theme) ||
              is.function(table.theme) ||
              is.list(table.theme))
  invisible(options(ggpmisc.ttheme.default = table.theme)[[1]])
}

#' @rdname ttheme_set
#'
set_ttheme <- ttheme_set

context("geom_tb")

library(tibble)
library(dplyr)

test_that("character label gives error in geom_table", {
  my.df <- data.frame(x = 1:10, y = 1:10, tb = letters[1:10])
  expect_error(print(ggplot(my.df, aes(x, y, label = tb)) +
                       geom_table()))
})

test_that("numeric label gives error in geom_table", {
  my.df <- data.frame(x = 1:10, y = 1:10, tb = 1:10)
  expect_error(print(ggplot(my.df, aes(x, y, label = tb)) +
                       geom_table()))
})

get_tb <- function() {
  mtcars |>
    group_by(cyl) |>
    summarize(wt = mean(wt), mpg = mean(mpg)) |>
    ungroup() |>
    mutate(
      wt = sprintf("%.2f", wt),
      mpg = sprintf("%.1f", mpg)
    )
}

test_that("geom_table works as expected", {
  p <- ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
    geom_point()
  tb <- get_tb()

  tbbl <- tibble(x = 5.45, y = 34, tb = list(tb))
  result <- expect_silent(
    p + geom_table(data = tbbl, aes(x = x, y = y, label = tb))
  )
  expect_s3_class(result, "ggplot")

  df <- data.frame(x = 5.45, y = 34, tb = I(list(tb)))
  result <- expect_silent(
    p + geom_table(data = df, aes(x = x, y = y, label = tb))
  )
  expect_s3_class(result, "ggplot")
})

test_that("geom_table_npc works as expected", {
  p <- ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
    geom_point()
  tb <- get_tb()

  tbbl <- tibble(x = 0.5, y = 0.5, tb = list(tb))
  result <- expect_silent(
    p + geom_table_npc(data = tbbl, aes(npcx = x, npcy = y, label = tb))
  )
  expect_s3_class(result, "ggplot")

  df <- data.frame(x = 0.5, y = 0.5, tb = I(list(tb)))
  result <- expect_silent(
    p + geom_table_npc(data = df, aes(npcx = x, npcy = y, label = tb))
  )
  expect_s3_class(result, "ggplot")

  df <- data.frame(x = 0.5, y = 0.5, tb = I(list(tb)))
  result <- expect_silent(
    p + geom_table_npc(data = df, aes(npcx = x, npcy = y, label = tb),
                       table.hjust = 0)
  )
  expect_s3_class(result, "ggplot")

  result <- expect_silent(
    p + geom_table_npc(data = df, aes(npcx = x, npcy = y, label = tb),
                       table.hjust = "center")
  )
  expect_s3_class(result, "ggplot")
})

# test_that("data.frame", {
#   my.df <- data.frame(x = 1:10, y = 1:10, tb = letters[1:10])
#   expect_warning(ggplot(my.df, aes(x, y, label = tb)) +
#                    geom_table())
#   vdiffr::expect_doppelganger("geom_table_text_label",
#                               suppressWarnings(
#                                 ggplot(my.df, aes(x, y, label = tb)) +
#                                   geom_table()
#                               )
#   )
# })

test_that("multiple_rows_tb works in geom_table", {
  tb <- tibble(Z1 = LETTERS[2:4], z1 = letters[4:2])
  tbb <- tibble(Z2 = LETTERS[2:4], z2 = letters[4:2])
  tbbb <- tibble(Z3 = LETTERS[2:4], z3 = letters[4:2])
  my.tb <- tibble(x = 2:4, y = 3:5, tb = list(t1 = tb, t2 = tbb, t3 = tbbb))
  vdiffr::expect_doppelganger("geom_table_multi_row",
                              ggplot() +
                                geom_table(data = my.tb,
                                           mapping = aes(x, y, label = tb)) +
                                lims(x = c(0, 6), y = c(0, 6))
                              )
})

test_that("numbers_tb works in geom_table", {
  my_data.tb <- tibble(x = -5:5, y = -5:5)
  tb <- tibble(a = 2:4, b = 4:2)
  my.tb <- tibble(x = 0, y = 0, tb = list(tb))
  vdiffr::expect_doppelganger("geom_table_geom_data",
                              ggplot() +
                                geom_table(data = my.tb,
                                           mapping = aes(x, y, label = tb))
                              )
  vdiffr::expect_doppelganger("geom_table_plot_data",
                              ggplot(data = my.tb) +
                                geom_table(mapping = aes(x, y, label = tb))
                              )
  vdiffr::expect_doppelganger("geom_table_vjust",
                              ggplot() +
                                geom_table(data = my.tb,
                                           mapping = aes(x, y, label = tb),
                                           vjust = 1)
                              )
  vdiffr::expect_doppelganger("geom_table_vjust_hjust",
                              ggplot(data = my.tb) +
                                geom_table(vjust = 1, hjust = 0,
                                     mapping = aes(x, y, label = tb))
                              )
  vdiffr::expect_doppelganger("geom_table_with_points",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           mapping = aes(x, y, label = tb))
                              )
  vdiffr::expect_doppelganger("geom_table_nudge_x",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           mapping = aes(x, y, label = tb))
  )
  vdiffr::expect_doppelganger("geom_table_nudge_x_nudge_y",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2, nudge_y = 2,
                                           vjust = 0.5, hjust = 1,
                                           mapping = aes(x, y, label = tb))
  )
})

test_that("alpha targets work in geom_table", {
  my_data.tb <- tibble(x = -5:5, y = -5:5)
  tb <- tibble(a = 2:4, b = 4:2)
  my.tb <- tibble(x = 0, y = 0, tb = list(tb))
  vdiffr::expect_doppelganger("geom_table_alpha_aes",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.5,
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_alpha_all",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.5,
                                           alpha.target = "all",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_alpha_segment",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.25,
                                           alpha.target = "segment",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_alpha_table",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.25,
                                           alpha.target = "table",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_alpha_canvas",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.25,
                                           alpha.target = "table.canvas",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_alpha_rules",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.25,
                                           alpha.target = "table.rules",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_alpha_base",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.25,
                                           alpha.target = "table.text",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_alpha_box",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.25,
                                           alpha.target = "box",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_alpha_none",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.25,
                                           alpha.target = "none",
                                           mapping = aes(x, y, label = tb))
  )

})

test_that("colour targets work in geom_table", {
  my_data.tb <- tibble(x = -5:5, y = -5:5)
  tb <- tibble(a = 2:4, b = 4:2)
  my.tb <- tibble(x = 0, y = 0, tb = list(tb))

  vdiffr::expect_doppelganger("geom_table_colour_na",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           colour = NA,
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_colour_aes",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           colour = "red",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_default_colour",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           default.colour = "blue",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_default_colour_none",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           default.colour = "blue",
                                           colour.target = "none",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_colour_all",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           colour = "red",
                                           colour.target = "all",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_colour_segment",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           colour = "red",
                                           colour.target = "segment",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_colour_table",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           colour = "red",
                                           colour.target = "table",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_colour_rules",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           colour = "red",
                                           colour.target = "table.rules",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_colour_base",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           colour = "red",
                                           colour.target = "table.text",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_colour_box",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           colour = "red",
                                           colour.target = "box",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_colour_none",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           colour = "red",
                                           colour.target = "none",
                                           mapping = aes(x, y, label = tb))
  )

})

test_that("colour targets and alpha targets work together in geom_table", {
  my_data.tb <- tibble(x = -5:5, y = -5:5)
  tb <- tibble(a = 2:4, b = 4:2)
  my.tb <- tibble(x = 0, y = 0, tb = list(tb))
  vdiffr::expect_doppelganger("geom_table_colour_alpha_aes",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           colour = "red",
                                           alpha = 0.25,
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_colour_all_alpha",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.25,
                                           alpha.target = "all",
                                           colour = "red",
                                           colour.target = "all",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_colour_segment_alpha_all",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.25,
                                           alpha.target = "all",
                                           colour = "red",
                                           colour.target = "segment",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_colour_segment_alpha_table",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.25,
                                           alpha.target = "table",
                                           colour = "red",
                                           colour.target = "segment",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_colour_table_alpha_table",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.25,
                                           alpha.target = "table",
                                           colour = "red",
                                           colour.target = "table",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_colour_table_alpha_segment",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.25,
                                           alpha.target = "segment",
                                           colour = "red",
                                           colour.target = "table",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_colour_rules_alpha_table",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.25,
                                           alpha.target = "table",
                                           colour = "red",
                                           colour.target = "table.rules",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_colour_rules_alpha_canvas",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.25,
                                           alpha.target = "table.canvas",
                                           colour = "red",
                                           colour.target = "table.rules",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_default_colour_rules_alpha_canvas",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.25,
                                           default.colour = "blue",
                                           alpha.target = "table.canvas",
                                           colour = "red",
                                           colour.target = "table.rules",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_colour_base_alpha_multiple",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.25,
                                           alpha.target = c("table.rules",
                                                            "table.canvas"),
                                           colour = "red",
                                           colour.target = "table.text",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_colour_box2",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.1,
                                           alpha.target = "none",
                                           colour = "red",
                                           colour.target = "box",
                                           mapping = aes(x, y, label = tb))
  )

})

test_that("fill and alpha work together in geom_table", {
  my_data.tb <- tibble(x = -5:5, y = -5:5)
  tb <- tibble(a = 2:4, b = 4:2)
  my.tb <- tibble(x = 0, y = 0, tb = list(tb))
  vdiffr::expect_doppelganger("geom_table_fill_aes",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           fill = "yellow",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_fill_alpha_all",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.25,
                                           alpha.target = "all",
                                           fill = "yellow",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_fill_alpha_table",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.25,
                                           alpha.target = "table",
                                           fill = "yellow",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_fill_alpha_segment",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.25,
                                           alpha.target = "segment",
                                           fill = "yellow",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_fill_alpha_canvas",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.25,
                                           alpha.target = "table.canvas",
                                           fill = "yellow",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_fill_alpha_multiple",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.25,
                                           alpha.target = c("table.rules",
                                                            "table.canvas"),
                                           fill = "yellow",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_fill_alpha_none",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.1,
                                           alpha.target = "none",
                                           fill = "yellow",
                                           mapping = aes(x, y, label = tb))
  )

  vdiffr::expect_doppelganger("geom_table_fill_alpha_na",
                              ggplot(my_data.tb, aes(x, y)) +
                                geom_point() +
                                geom_table(data = my.tb,
                                           nudge_x = -2,
                                           hjust = 1,
                                           alpha = 0.1,
                                           alpha.target = "none",
                                           fill = NA,
                                           mapping = aes(x, y, label = tb))
  )

})

test_that("table themes work in geom_table", {
  my_data.tb <- tibble(x = -5:5, y = -5:5)
  tb <- tibble(a = 2^(2:4), b = 2^(4:2))
  my.tb <- tibble(x = 0, y = 0, tb = list(tb))
  ttheme_set(ttheme_gtdark)
  vdiffr::expect_doppelganger("ttheme_set_dark",
                              ggplot() +
                                geom_table(data = my.tb,
                                           mapping = aes(x, y, label = tb))
                              )
  ttheme_set()
  vdiffr::expect_doppelganger("ttheme_gtdefault_just_1",
                              ggplot() +
                                geom_table(data = my.tb,
                                           table.theme = ttheme_gtdefault,
                                           table.hjust = 1,
                                           mapping = aes(x, y, label = tb))
                              )
  vdiffr::expect_doppelganger("ttheme_gtdefault_just_0",
                              ggplot() +
                                geom_table(data = my.tb,
                                           table.theme = ttheme_gtdefault,
                                           table.hjust = 0,
                                           mapping = aes(x, y, label = tb)))
  vdiffr::expect_doppelganger("ttheme_gtdefault_just_05",
                              ggplot() +
                                geom_table(data = my.tb,
                                           table.theme = ttheme_gtdefault,
                                           table.hjust = 0.5,
                                           mapping = aes(x, y, label = tb)))
  vdiffr::expect_doppelganger("ttheme_gtdefault_just_right",
                              ggplot() +
                                geom_table(data = my.tb,
                                           table.theme = ttheme_gtdefault,
                                           table.hjust = "right",
                                           mapping = aes(x, y, label = tb)))
  vdiffr::expect_doppelganger("ttheme_gtdefault_just_middle",
                              ggplot() +
                                geom_table(data = my.tb,
                                           table.theme = ttheme_gtdefault,
                                           table.hjust = "middle",
                                           mapping = aes(x, y, label = tb)))
  vdiffr::expect_doppelganger("ttheme_gtdefault_just_center",
                              ggplot() +
                                geom_table(data = my.tb,
                                           table.theme = ttheme_gtdefault,
                                           table.hjust = "center",
                                           mapping = aes(x, y, label = tb)))
  vdiffr::expect_doppelganger("ttheme_gtdefault_just_left",
                              ggplot() +
                                geom_table(data = my.tb,
                                           table.theme = ttheme_gtdefault,
                                           table.hjust = "left",
                                           mapping = aes(x, y, label = tb)))
  vdiffr::expect_doppelganger("ttheme_gtdefault_rownames",
                              ggplot() +
                                geom_table(data = my.tb,
                                           table.theme = ttheme_gtdefault,
                                           table.rownames = TRUE,
                                           mapping = aes(x, y, label = tb)))
  vdiffr::expect_doppelganger("ttheme_gtdefault_colnames",
                              ggplot() +
                                geom_table(data = my.tb,
                                           table.theme = ttheme_gtdefault,
                                           table.colnames = FALSE,
                                           mapping = aes(x, y, label = tb)))
  vdiffr::expect_doppelganger("ttheme_gtminimal",
                              ggplot() +
                                geom_table(data = my.tb,
                                           table.theme = ttheme_gtminimal,
                                           mapping = aes(x, y, label = tb)))
  vdiffr::expect_doppelganger("ttheme_gtsimple",
                              ggplot() +
                                geom_table(data = my.tb,
                                           table.theme = ttheme_gtsimple,
                                           mapping = aes(x, y, label = tb)))
  vdiffr::expect_doppelganger("ttheme_gtdark",
                              ggplot() +
                                geom_table(data = my.tb,
                                           table.theme = ttheme_gtdark,
                                           mapping = aes(x, y, label = tb)))
  vdiffr::expect_doppelganger("ttheme_gtlight",
                              ggplot() +
                                geom_table(data = my.tb,
                                           table.theme = ttheme_gtlight,
                                           mapping = aes(x, y, label = tb)))
  vdiffr::expect_doppelganger("ttheme_gtstripes",
                              ggplot() +
                                geom_table(data = my.tb,
                                           table.theme = ttheme_gtstripes,
                                           mapping = aes(x, y, label = tb)))
})

test_that("letters_tb works in geom_table", {
  tb <- tibble(a = LETTERS[2:4], b = letters[4:2])
  my.tb <- tibble(x = 0, y = 0, tb = list(tb))
  vdiffr::expect_doppelganger("geom_table_letters",
                              ggplot() +
                                geom_table(data = my.tb,
                                           mapping = aes(x, y, label = tb))
                              )
})

test_that("parsed_tb works in geom_table", {
  tb <- tibble("alpha" = c("x[2]~\"=\"~a^2", "sqrt(y)"),
               "beta" = c("x[2]~\"=\"~b^2", "sqrt(1/y)"))
  my.tb <- tibble(x = 0, y = 0, tb = list(tb))
  vdiffr::expect_doppelganger("geom_table_parsed_all",
                              ggplot() +
                                geom_table(my.tb, mapping = aes(x, y, label = tb), parse = TRUE)
                              )

  tb <- tibble("alpha" = c("x[2]~\"=\"~a^2", "text"),
               "beta" = c("x[2]~\"=\"~b^2", "1200"))
  my.tb <- tibble(x = 0, y = 0, tb = list(tb))
  vdiffr::expect_doppelganger("geom_table_parsed_partial",
                              ggplot() +
                                geom_table(my.tb, mapping = aes(x, y, label = tb), parse = TRUE)
                              )
  vdiffr::expect_doppelganger("geom_table_parsed_partial_dark",
                              ggplot() +
                                geom_table(my.tb,
                                           table.theme = ttheme_gtdark,
                                           mapping = aes(x, y, label = tb),
                                           parse = TRUE)
  )
})

test_that("pos_or_nudge work in geom_table", {
  tb <- tibble(a = 2:4, b = 4:2)
  my.tb <- tibble(x = 0, y = 0, tb = list(tb))
  expect_error(geom_table(data = my.tb,
                          mapping = aes(x, y, label = tb),
                          position = "stack",
                          nudge_x = 0.5,
                          nudge_y = 0.5),
               "You must specify either `position` or `nudge_x`/`nudge_y`.")
})

test_that("string_left_hjustworks in geom_table", {
  tb <- tibble(a = 2:4, b = 4:2)
  my.tb <- tibble(x = 0, y = 0, tb = list(tb))
x <- geom_table(data = my.tb,
                mapping = aes(x, y, label = tb),
                table.hjust = "left")
expect_equal(x$geom_params$table.hjust, 0)
})

context("ttheme_gtminimal")

test_that("ttheme_gtminimal, default text elements", {
  t1 <- ttheme_gtminimal()

  expect_true(t1$core$fg_params$fontsize == 10)
  expect_true(t1$core$fg_params$col == "#000000FF" ||
                t1$core$fg_params$col == "black")
})

test_that("ttheme_gtminimal, check changes to text elements", {
  t2 <- ttheme_gtminimal(base_size = 15, base_colour = "blue")

  expect_true(t2$core$fg_params$fontsize == 15)
  expect_true(t2$core$fg_params$col == "#0000FFFF" ||
                t2$core$fg_params$col == "blue")
})

context("ttheme_gtplain")

test_that("ttheme_gtplain, default text elements", {
  t1 <- ttheme_gtplain()

  expect_true(t1$core$fg_params$fontsize == 10)
  expect_true(t1$core$fg_params$col == "#000000FF")

  expect_identical(t1$core$bg_params$fill, ttheme_gtplain()$core$bg_params$fill)

  expect_identical(t1$colhead$bg_params$fill, ttheme_gtplain()$colhead$bg_params$fill)

  expect_identical(t1$rowhead$bg_params$fill, ttheme_gtplain()$rowhead$bg_params$fill)
})

test_that("ttheme_gtplain, check changes to text elements", {
  t2 <- ttheme_gtplain(base_size = 15, base_colour = "blue")

  expect_true(t2$core$fg_params$fontsize == 15)
  expect_true(t2$core$fg_params$col == "#0000FFFF")

  expect_identical(t2$core$bg_params$fill, ttheme_gtplain()$core$bg_params$fill)

  expect_identical(t2$colhead$bg_params$fill, ttheme_gtplain()$colhead$bg_params$fill)

  expect_identical(t2$rowhead$bg_params$fill, ttheme_gtplain()$rowhead$bg_params$fill)
})

context("ttheme_gtstripes")

test_that("ttheme_gtstripes, default text elements", {
  t1 <- ttheme_gtstripes()

  expect_true(t1$core$fg_params$fontsize == 10)
  expect_true(t1$core$fg_params$col == "#1A1A1AFF")

  expect_identical(t1$core$bg_params$fill, ttheme_gtstripes()$core$bg_params$fill)
  expect_identical(t1$core$bg_params$col, ttheme_gtstripes()$core$bg_params$col)
  expect_identical(t1$core$bg_params$lwd, ttheme_gtstripes()$core$bg_params$lwd)

  expect_identical(t1$colhead$bg_params$fill, ttheme_gtstripes()$colhead$bg_params$fill)
  expect_identical(t1$colhead$bg_params$col, ttheme_gtstripes()$colhead$bg_params$col)
  expect_identical(t1$colhead$bg_params$lwd, ttheme_gtstripes()$colhead$bg_params$lwd)

  expect_identical(t1$rowhead$bg_params$fill, ttheme_gtstripes()$rowhead$bg_params$fill)
  expect_identical(t1$rowhead$bg_params$col, ttheme_gtstripes()$rowhead$bg_params$col)
  expect_identical(t1$rowhead$bg_params$lwd, ttheme_gtstripes()$rowhead$bg_params$lwd)
})

test_that("ttheme_gtstripes, check changes to text elements", {
  t2 <- ttheme_gtstripes(base_size = 12, base_colour = "darkred")

  expect_true(t2$core$fg_params$fontsize == 12)
  expect_true(t2$core$fg_params$col == "#8B0000FF")

  expect_identical(t2$core$bg_params$fill, ttheme_gtstripes()$core$bg_params$fill)
  expect_identical(t2$core$bg_params$col, ttheme_gtstripes()$core$bg_params$col)
  expect_identical(t2$core$bg_params$lwd, ttheme_gtstripes()$core$bg_params$lwd)

  expect_identical(t2$colhead$bg_params$fill, ttheme_gtstripes()$colhead$bg_params$fill)
  expect_identical(t2$colhead$bg_params$col, ttheme_gtstripes()$colhead$bg_params$col)
  expect_identical(t2$colhead$bg_params$lwd, ttheme_gtstripes()$colhead$bg_params$lwd)

  expect_identical(t2$rowhead$bg_params$fill, ttheme_gtstripes()$rowhead$bg_params$fill)
  expect_identical(t2$rowhead$bg_params$col, ttheme_gtstripes()$rowhead$bg_params$col)
  expect_identical(t2$rowhead$bg_params$lwd, ttheme_gtstripes()$rowhead$bg_params$lwd)
})

#' Inset plots
#'
#' \code{geom_plot} and \code{geom_plot_npc} add ggplot objects as insets to the
#' base ggplot, using syntax similar to that of
#' \code{\link[ggplot2]{geom_label}} and \code{\link{geom_text_s}}.
#' In most respects they behave as any other ggplot geometry: they add a layer
#' containing one or more grobs and grouping and faceting works as usual. The
#' most common use of \code{geom_plot} is to add data labels that are themselves
#' ggplots rather than text. \code{\link{geom_plot_npc}} is used to add ggplots
#' as annotations to plots, but contrary to layer function \code{annotate()},
#' \code{\link{geom_plot_npc}} is data driven and respects grouping and facets,
#' thus plot insets can differ among panels.
#'
#' @details You can modify the size of inset plots with the \code{vp.width} and
#'   \code{vp.height} aesthetics. These can take a number between 0 (smallest
#'   possible inset) and 1 (whole plotting area width or height). The default
#'   value for for both of these aesthetics is 1/5. Thus, in contrast to
#'   \code{\link[ggplot2]{geom_text}} and \code{\link{geom_text_s}} the size of
#'   the insets remains the same relative to the size of the plotting area
#'   irrespective of how the plot is rendered. The aspect ratio of insets is
#'   preserved and size is adjusted until the whole inset fits within the
#'   viewport.
#'
#'   \code{geom_plot} and \code{geom_plot_npc} expect a list of ggplot objects
#'   (\code{"gg"} class) to be mapped to the \code{label} aesthetic. These geoms
#'   work with tibbles or data frames as \code{data} as they both support
#'   \code{list} objects as member variables.
#'
#'   The \code{x} and \code{y} aesthetics determine the position of the whole
#'   inset plot, similarly to that of a text label, justification is interpreted
#'   as indicating the position of the plot with respect to its x and y
#'   coordinates in the data, and \code{angle} is used to rotate the plot as a
#'   whole.
#'
#'   Of these two geoms only \code{\link{geom_plot}} supports the plotting of
#'   segments, as \code{\link{geom_plot_npc}} uses a coordinate system that is
#'   unrelated to data units and data.In the case of \code{geom_plot_npc()},
#'   \code{npcx} and \code{npcy} aesthetics determine the position of the inset
#'   plot.
#'
#' @inheritSection geom_text_s Alignment
#'
#' @inheritSection geom_text_s Position functions
#'
#' @inheritSection geom_grob Plot boundaries and clipping
#'
#' @inherit geom_grob return note seealso references
#'
#' @inheritParams geom_grob
#'
#' @param default.colour,default.color A colour definition to use for elements not targeted by
#'   the colour aesthetic.
#' @param colour.target,color.target A vector of character strings; \code{"all"},
#'   \code{"text"}, \code{"box"} and \code{"segment"}.
#' @param default.alpha numeric in [0..1] A transparency value to use for
#'   elements not targeted by the alpha aesthetic.
#' @param alpha.target A vector of character strings; \code{"all"},
#'   \code{"text"}, \code{"segment"}, \code{"box"}, \code{"box.line"}, and
#'   \code{"box.fill"}.
#'
#' @family geometries adding layers with insets
#'
#' @export
#'
#' @examples
#' # inset plot with enlarged detail from a region of the main plot
#' library(tibble)
#' p <-
#'   ggplot(data = mtcars, mapping = aes(wt, mpg)) +
#'   geom_point()
#'
#' p.crop <- p +
#'           coord_cartesian(xlim = c(3, 4),
#'           ylim = c(13, 16)) +
#'           labs(x = NULL, y = NULL) +
#'           theme_bw(10)
#'
#' df <- data.frame(x = 0.01,
#'                  y = 0.01,
#'                  plot = I(list(p.crop)))
#' p +
#'   expand_limits(x = 0, y = 0) +
#'   geom_plot_npc(data = df,
#'                 aes(npcx = x, npcy = y, label = plot))
#'
#' p +
#'   expand_limits(x = 0, y = 0) +
#'   geom_plot_npc(data = df,
#'                 aes(npcx = x, npcy = y, label = plot,
#'                 vp.width = 1/2, vp.height = 1/4))
#'
#' p +
#'   expand_limits(x = 0, y = 0) +
#'   geom_plot_npc(data = df,
#'                 aes(npcx = x, npcy = y, label = plot),
#'                 vp.width = 1/4, vp.height = 1/4)
#'
#' p +
#'   geom_plot(data = df,
#'             aes(x = x + 3, y = y + 20, label = plot),
#'             nudge_x = -1, nudge_y = - 7,
#'             hjust = 0.5, vjust = 0.5,
#'             arrow = arrow(length = unit(0.5, "lines")),
#'             colour = "red",
#'             vp.width = 1/5, vp.height = 1/5)
#'
geom_plot <- function(mapping = NULL,
                      data = NULL,
                      stat = "identity",
                      position = "identity",
                      ...,
                      nudge_x = 0,
                      nudge_y = 0,
                      default.colour = NULL,
                      default.color = default.colour,
                      colour.target = "box",
                      color.target = colour.target,
                      default.alpha = 1,
                      alpha.target = "all",
                      add.segments = TRUE,
                      box.padding = 0.25,
                      point.padding = 1e-06,
                      segment.linewidth = 0.5,
                      min.segment.length = 0,
                      arrow = NULL,
                      na.rm = FALSE,
                      show.legend = FALSE,
                      inherit.aes = FALSE) {

  colour.target <-
    rlang::arg_match(color.target,
                     values = c("segment", "all", "box", "none"),
                     multiple = TRUE)
  alpha.target <-
    rlang::arg_match(alpha.target,
                     values = c("segment", "all", "box",
                                "box.line", "box.fill", "none"),
                     multiple = TRUE)

  if (!missing(nudge_x) || !missing(nudge_y)) {
    if (!missing(position) && !identical(position, "identity")) {
      rlang::abort("You must specify either `position` or `nudge_x`/`nudge_y`.")
    }
    # original position needed for "position" justification
    position <-
      position_nudge_center(nudge_x, nudge_y, kept.origin = "original")
  }

  ggplot2::layer(
    data = data,
    mapping = mapping,
    stat = stat,
    geom = GeomPlot,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      default.colour = default.color,
      colour.target = colour.target,
      default.alpha = default.alpha,
      alpha.target = alpha.target,
      add.segments = add.segments,
      box.padding = box.padding,
      point.padding = point.padding,
      segment.linewidth = segment.linewidth,
      min.segment.length = min.segment.length,
      arrow = arrow,
      na.rm = na.rm,
      ...
    )
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
GeomPlot <-
  ggplot2::ggproto("GeomPlot", ggplot2::Geom,
          required_aes = c("x", "y", "label"),

          default_aes = ggplot2::aes(
            colour = "black",
            family = "",
            angle = 0,
            hjust = "inward",
            vjust = "inward",
            alpha = NA,
            fontface = 1,
            vp.width = 0.4,
            vp.height = 0.4
          ),

          draw_panel =   function(data,
                                  panel_params,
                                  coord,
                                  add.segments = TRUE,
                                  box.padding = 0.25,
                                  point.padding = 1e-06,
                                  segment.linewidth = 1,
                                  min.segment.length = 0,
                                  arrow = NULL,
                                  default.colour = "black",
                                  colour.target = "all",
                                  default.alpha = 1,
                                  alpha.target = "all",
                                  na.rm = FALSE) {

            if (nrow(data) == 0) {
              return(grid::nullGrob())
            }

            if (!ggplot2::is_ggplot(data$label[[1]])) {
              warning("Skipping as object mapped to 'label' is not",
                      " a list of \"gg\" or \"ggplot\" objects.")
              return(grid::nullGrob())
            }

            default.colour <- check_default_colour(default.colour)

            add.segments <- add.segments && all(c("x_orig", "y_orig") %in% colnames(data))

            # should be called only once!
            data <- coord$transform(data, panel_params)
            if (add.segments) {
              data_orig <- data.frame(x = data$x_orig, y = data$y_orig)
              data_orig <- coord$transform(data_orig, panel_params)
              data$x_orig <- data_orig$x
              data$y_orig <- data_orig$y
            }

            if (is.character(data$vjust)) {
              data$vjust <-
                compute_just2d(data = data,
                               coord = coord,
                               panel_params = panel_params,
                               just = data$vjust,
                               a = "y", b = "x")
            }
            if (is.character(data$hjust)) {
              data$hjust <-
                compute_just2d(data = data,
                               coord = coord,
                               panel_params = panel_params,
                               just = data$hjust,
                               a = "x", b = "y")
            }
            if (add.segments) {
              segments.data <-
                shrink_segments(data,
                                point.padding = point.padding,
                                box.padding = box.padding,
                                min.segment.length = min.segment.length)
            }

            # loop needed as gpar is not vectorized
            all.grobs <- grid::gList()

            for (row.idx in 1:nrow(data)) {
              row <- data[row.idx, , drop = FALSE]
              plot.alpha <-
                ifelse(any(alpha.target %in% c("all", "plot")),
                       row$alpha, default.alpha)
              segment.alpha <-
                ifelse(any(alpha.target %in% c("all", "segment")),
                       row$alpha, default.alpha)
              user.grob <- ggplot2::ggplotGrob(x = data$label[[row.idx]])

              user.grob$vp <-
                grid::viewport(x = grid::unit(row$x, "native"),
                               y = grid::unit(row$y, "native"),
                               width = grid::unit(row$vp.width, "npc"),
                               height = grid::unit(row$vp.height, "npc"),
                               just = c(row$hjust, row$vjust),
                               angle = row$angle,
                               name = paste("inset.plot.vp", row$PANEL,
                                            "row", row.idx, sep = "."))

              # give unique name to each grob
              user.grob$name <- paste("inset.plot", row.idx, sep = ".")

              if (add.segments) {
                segment.row <- segments.data[row.idx, , drop = FALSE]
                if (segment.row$too.short) {
                  segment.grob <- grid::nullGrob()
                } else {
                  segment.grob <-
                    grid::segmentsGrob(x0 = segment.row$x,
                                       y0 = segment.row$y,
                                       x1 = segment.row$x_orig,
                                       y1 = segment.row$y_orig,
                                       arrow = arrow,
                                       gp = grid::gpar(
                                         col = if (segment.linewidth == 0) NA else # lwd = 0 is invalid in 'grid'
                                           ifelse(any(colour.target %in% c("all", "segment")),
                                                  ggplot2::alpha(row$colour, segment.alpha),
                                                  ggplot2::alpha(default.colour, segment.alpha)),
                                         lwd = (if (segment.linewidth == 0) 0.5 else segment.linewidth) * ggplot2::.stroke),
                                       name = paste("plot.s.segment", row$group, row.idx, sep = "."))
                }
                all.grobs <- grid::gList(all.grobs, segment.grob, user.grob)
              } else {
                all.grobs <- grid::gList(all.grobs, user.grob)
              }
            }
            #    grid::grobTree(children = all.grobs, name = "geom.plot.panel")
            grid::grobTree(children = all.grobs)

          },
          draw_key = function(...) {
            grid::nullGrob()
          }
  )

#' @rdname geom_plot
#' @export
#'
geom_plot_npc <- function(mapping = NULL,
                          data = NULL,
                          stat = "identity",
                          position = "identity",
                          ...,
                          na.rm = FALSE,
                          show.legend = FALSE,
                          inherit.aes = FALSE) {
  ggplot2::layer(
    data = data,
    mapping = mapping,
    stat = stat,
    geom = GeomPlotNpc,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      na.rm = na.rm,
      ...
    )
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
GeomPlotNpc <-
  ggplot2::ggproto("GeomPlotNpc", ggplot2::Geom,
          required_aes = c("npcx", "npcy", "label"),

          default_aes = ggplot2::aes(
            colour = "black",
            angle = 0,
            hjust = "inward",
            vjust = "inward",
            alpha = NA,
            family = "",
            fontface = 1,
            vp.width = 0.4, vp.height = 0.4
          ),

          draw_panel = function(data, panel_params, coord,
                                na.rm = FALSE) {

            if (nrow(data) == 0) {
              return(grid::nullGrob())
            }

            if (!ggplot2::is_ggplot(data$label[[1]])) {
              warning("Skipping as object mapped to 'label' is not a list of \"gg\" or \"ggplot\" objects.")
              return(grid::nullGrob())
            }

            data$npcx <- compute_npcx(data$npcx)
            data$npcy <- compute_npcy(data$npcy)

            if (is.character(data$vjust)) {
              data$vjust <- compute_just(data$vjust, data$npcy)
            }
            if (is.character(data$hjust)) {
              data$hjust <- compute_just(data$hjust, data$npcx)
            }

            plot.grobs <- grid::gList()

            for (row.idx in seq_len(nrow(data))) {
              plotGrob <-
                ggplot2::ggplotGrob(x = data$label[[row.idx]])

              plotGrob$vp <- grid::viewport(x = grid::unit(data$npcx[row.idx], "native"),
                                            y = grid::unit(data$npcy[row.idx], "native"),
                                            width = grid::unit(data$vp.width[row.idx], "npc"),
                                            height = grid::unit(data$vp.height[row.idx], "npc"),
                                            just = c(data$hjust[row.idx],
                                                     data$vjust[row.idx]),
                                            angle = data$angle[row.idx],
                                            name = paste("geom_plot.panel",
                                                         data$PANEL[row.idx], "row",
                                                         row.idx, sep = "."))

              # give unique name to each plot
              plotGrob$name <- paste("inset.plot", row.idx, sep = ".")

              plot.grobs[[row.idx]] <- plotGrob
            }

            # grid.name <- paste("geom_plot.panel",
            #                    data$PANEL[row.idx], sep = ".")
            # grid::gTree(children = plot.grobs, name = grid.name)

            grid::gTree(children = plot.grobs)
          },

          draw_key = function(...) {
            grid::nullGrob()
          }
  )

context("geom_plot")

library(tibble)

test_that("character label gives error", {
   my.df <- data.frame(x = 1:10, y = 1:10, tb = letters[1:10])
   expect_error(print(ggplot(my.df, aes(x, y, label = tb)) +
                  geom_plot()))
})


test_that("numeric label gives error", {
  my.df <- data.frame(x = 1:10, y = 1:10, tb = 1:10)
  expect_error(print(ggplot(my.df, aes(x, y, label = tb)) +
                       geom_plot()))
})

test_that("geom_plot works as expected", {
  p1 <- ggplot(mpg, aes(displ, cty)) + geom_point()

  p <- ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
    geom_point()
  # tibble
  tb <- tibble(x = 5.45, y = 34, plot = list(p1))
  result <- expect_silent(
    p + geom_plot(data = tb, aes(x = x, y = y, label = plot))
  )
  expect_s3_class(result, "ggplot")
  # data.frame
  df <- data.frame(x = 5.45, y = 34, plot = I(list(p1)))
  result1 <- expect_silent(
    p + geom_plot(data = df, aes(x = x, y = y, label = plot))
  )
  expect_s3_class(result1, "ggplot")
})

test_that("geom_plot_npc works as expected", {
  p1 <- ggplot(mpg, aes(displ, cty)) + geom_point()

  p <- ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
    geom_point()
  # tibble
  tbnpc <- tibble(x = 0.95, y = 0.95,  plot = list(p1))
  result <- expect_silent(
    p + geom_plot_npc(data = tbnpc, aes(npcx = x, npcy = y, label = plot))
  )
  expect_s3_class(result, "ggplot")

  dfnpc <- data.frame(x = 0.95, y = 0.95,  plot = I(list(p1)))
  result1 <- expect_silent(
    p + geom_plot_npc(data = dfnpc, aes(npcx = x, npcy = y, label = plot))
  )
  expect_s3_class(result1, "ggplot")
})


test_that("multiple_ggplot", {
  p1 <- ggplot(mpg, aes(displ, cty)) + geom_point()
  p2 <- ggplot(mpg, aes(displ, hwy)) + geom_point()
  p3 <- ggplot(mpg, aes(cyl, displ)) + geom_point()
  my.tb <- tibble(x = 0:2, y = 0:2, tb = list(p1, p2, p3))
#  skip("false positive triggered")
  vdiffr::expect_doppelganger("geom_plot_three_plots",
                              ggplot() +
                                geom_plot(data = my.tb,
                                           mapping = aes(x, y, label = tb)) +
                                lims(x = c(0, 2), y = c(0, 2)))
#  skip("false positive triggered")
  vdiffr::expect_doppelganger("geom_plot_three_angles",
                              ggplot() +
                                geom_plot(data = my.tb,
                                          mapping = aes(x, y, label = tb),
                                          angle = 90, vjust = c(1, 0.5, 0)) +
                                lims(x = c(0, 2), y = c(0, 2)))
  # expect_warning(print(ggplot() +
  #                  geom_plot(data = my.tb[2, ],
  #                            mapping = aes(x, y, label = tb),
  #                            angle = 45) +
  #                  lims(x = c(0, 2), y = c(0, 2))))
})

test_that("examples_geom_plot", {
  p <-
    ggplot(data = mtcars, mapping = aes(wt, mpg)) +
    geom_point()

  df <- tibble(x = 0.01, y = 0.01,
               plot = list(p +
                           coord_cartesian(xlim = c(3, 4),
                                           ylim = c(13, 16)) +
                           labs(x = NULL, y = NULL)))

#  skip("false positive triggered")
  vdiffr::expect_doppelganger("geom_plot_npc_1",
                              p +
                                expand_limits(x = 0, y = 0) +
                                geom_plot_npc(data = df,
                                              aes(npcx = x, npcy = y, label = plot))
  )
#  skip("false positive triggered")
  vdiffr::expect_doppelganger("geom_plot_npc_2",
                              p +
                                expand_limits(x = 0, y = 0) +
                                geom_plot_npc(data = df,
                                              aes(npcx = x, npcy = y, label = plot),
                                              angle = 90, vjust = 1)
  )
})

test_that("nudge_x_geom_plot", {
  p1 <- ggplot(mpg, aes(displ, cty)) + geom_point()
  p2 <- ggplot(mpg, aes(displ, hwy)) + geom_point()
  p3 <- ggplot(mpg, aes(cyl, displ)) + geom_point()
  my.tb <- tibble(x = 0:2, y = 0:2, tb = list(p1, p2, p3))
  p <- ggplot() +
    geom_plot(data = my.tb, mapping = aes(x, y, label = tb), nudge_x = 0.1) +
    lims(x = c(0, 2), y = c(0, 2))
  result <- layer_data(p)[, c("x_orig", "y_orig")]
  expected <- data.frame(
    x_orig = c(0, 1, 2),
    y_orig = c(0, 1, 2)
  )
  expect_identical(result, expected)
})

test_that("nudge_x_and_position_geom_plot_fails", {
  p1 <- ggplot(mpg, aes(displ, cty)) + geom_point()
  p2 <- ggplot(mpg, aes(displ, hwy)) + geom_point()
  p3 <- ggplot(mpg, aes(cyl, displ)) + geom_point()
  my.tb <- tibble(x = 0:2, y = 0:2, tb = list(p1, p2, p3))
  expect_error(
    ggplot() +
      geom_plot(data = my.tb, mapping = aes(x, y, label = tb), nudge_x = 0.1, position = position_nudge_keep) +
      lims(x = c(0, 2), y = c(0, 2)),
    "You must specify either `position` or `nudge_x`/`nudge_y`."
  )
})

#' @rdname position_dodgenudge
#'
#' @export
#'
position_dodge2nudge <-
  function(width = 1,
           preserve = c("total", "single"),
           padding = 0.1,
           reverse = FALSE,
           x = 0,
           y = 0,
           direction = c("none", "split", "split.x", "split.y", "center"),
           kept.origin = c("dodged", "original", "none")) {

    preserve <- rlang::arg_match(preserve)
    direction <- rlang::arg_match(direction)
    kept.origin <- rlang::arg_match(kept.origin)

    fun_one <- function(x) {1}

    ggplot2::ggproto(NULL, PositionDodge2AndNudge,
                     x = x,
                     y = y,
                     .fun_x = switch(direction,
                                     none = fun_one,
                                     split = sign,
                                     split.y = fun_one,
                                     split.x = sign,
                                     center = sign),
                     .fun_y = switch(direction,
                                     none = fun_one,
                                     split = sign,
                                     split.x = fun_one,
                                     split.y = sign,
                                     center = sign),
                     kept.origin = kept.origin,
                     width = width,
                     preserve = rlang::arg_match(preserve),
                     padding = padding,
                     reverse = reverse
    )
  }

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @noRd
PositionDodge2AndNudge <-
  ggplot2::ggproto("PositionDodge2AndNudge", ggplot2::PositionDodge2,
                   x = 0,
                   y = 0,

                   setup_params = function(self, data) {
                     c(
                       list(nudge_x = self$x,
                            nudge_y = self$y,
                            .fun_x = self$.fun_x,
                            .fun_y = self$.fun_y,
                            x.reorder = !is.null(self$x) && length(self$x) > 1 && length(self$x) < nrow(data),
                            y.reorder = !is.null(self$y) && length(self$y) > 1 && length(self$y) < nrow(data),
                            kept.origin = self$kept.origin),
                       ggplot2::ggproto_parent(ggplot2::PositionDodge2, self)$setup_params(data)
                     )
                   },

                   compute_layer = function(self, data, params, layout) {

                     if (length(params$nudge_x) > nrow(data)) {
                       warning("Argument 'x' longer than data: some values dropped!")
                     }
                     if (length(params$nudge_y) > nrow(data)) {
                       warning("Argument 'y' longer than data: some values dropped!")
                     }
                     x_orig <- data$x
                     y_orig <- data$y
                     # operate on the dodged positions
                     data = ggplot2::ggproto_parent(ggplot2::PositionDodge2, self)$compute_layer(data, params, layout)

                     x_dodged <- data$x
                     y_dodged <- data$y

                     if (params$x.reorder) {
                       params$nudge_x <- rep_len(params$nudge_x, nrow(data))[order(order(data$x))]
                     } else {
                       params$nudge_x <- rep_len(params$nudge_x, nrow(data))
                     }
                     if (params$y.reorder) {
                       params$nudge_y <- rep_len(params$nudge_y, nrow(data))[order(order(data$y))]
                     } else {
                       params$nudge_y <- rep_len(params$nudge_y, nrow(data))
                     }

                     # transform only the dimensions for which non-zero nudging is requested
                     if (any(params$nudge_x != 0)) {
                       if (any(params$nudge_y != 0)) {
                         data <- ggplot2::transform_position(data,
                                                             function(x) x + params$nudge_x * params$.fun_x(x),
                                                             function(y) y + params$nudge_y * params$.fun_y(y))
                       } else {
                         data <- ggplot2::transform_position(data,
                                                             function(x) x + params$nudge_x * params$.fun_x(x),
                                                             NULL)
                       }
                     } else if (any(params$nudge_y != 0)) {
                       data <- ggplot2::transform_position(data,
                                                           function(x) x,
                                                           function(y) y + params$nudge_y * params$.fun_y(y))
                     }
                     # add original position
                     if (params$kept.origin == "dodged") {
                       data$x_orig <- x_dodged
                       data$y_orig <- y_dodged
                     } else if (params$kept.origin == "original") {
                       data$x_orig <- x_orig
                       data$y_orig <- y_orig
                     }

                     data
                   },

                   compute_panel = function(self, data, params, scales) {
                     ggplot2::ggproto_parent(PositionDodge2, self)$compute_panel(data, params, scales)
                   }
  )

#' @rdname position_dodgenudge
#'
#' @export
#'
position_dodge2_keep <- function(width = 1,
                                 preserve = c("total", "single"),
                                 kept.origin = "original") {
  position_dodge2nudge(width = width,
                       preserve = preserve,
                       x = 0,
                       y = 0,
                       direction = "none",
                       kept.origin = kept.origin)
}

context("position_dodge2nudge")

test_that("incorrect kept.origin used", {
  expect_error(
    position_dodge2nudge(kept.origin = "")
    #, "`kept.origin` must be one of \"dodged\", \"original\", or \"none\", not \"\"."
  )
})

test_that("incorrect direction used", {
  expect_error(
    position_dodge2nudge(direction = "")
    #, "`kept.origin` must be one of \"dodged\", \"original\", or \"none\", not \"\"."
  )
})

test_that("correct kept.origin used", {
  expect_no_error(position <- position_dodge2nudge())
  expect_identical(position$kept.origin, "dodged")
})

test_that("correct kept.origin used", {
  expect_no_error(position <- position_dodge2nudge(kept.origin = "original"))
  expect_identical(position$kept.origin, "original")
})

test_that("correct kept.origin used", {
  expect_no_error(position <- position_dodge2nudge(kept.origin = "dodged"))
  expect_identical(position$kept.origin, "dodged")
})

test_that("correct kept.origin used", {
  expect_no_error(position <- position_dodge2nudge(kept.origin = "none"))
  expect_identical(position$kept.origin, "none")
})

test_that("test if correct arguments are assigned", {
  position <- position_dodge2nudge(kept.origin = "none")
  expect_false(position$reverse)
  expect_identical(position$kept.origin, "none")
  expect_identical(position$padding, 0.1)
  expect_identical(position$preserve, "total")
  expect_identical(position$width, 1)
  expect_identical(position$x, 0)
  expect_identical(position$y, 0)
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
  skip_if(condition = utils::packageVersion("ggplot2") < "4.0.0",
          message = "'ggplot2' < 4.0.0")
  expect_false(position$reverse)
  expect_identical(position$required_aes, character(0))
  expect_type(position$aesthetics, "closure")
  expect_type(position$use_defaults, "closure")
  skip("Not yet implemented")
  expect_identical(position$orientation, "x")
})

test_that("test if correct arguments are assigned with 'split.y'", {
  position <- position_dodge2nudge(kept.origin = "none", direction = "split.y")
  expect_false(position$reverse)
  expect_identical(position$kept.origin, "none")
  expect_identical(position$padding, 0.1)
  expect_identical(position$preserve, "total")
  expect_identical(position$width, 1)
  expect_identical(position$x, 0)
  expect_identical(position$y, 0)
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
  skip_if(condition = utils::packageVersion("ggplot2") < "4.0.0",
          message = "'ggplot2' < 4.0.0")
  expect_false(position$reverse)
  expect_identical(position$required_aes, character(0))
  expect_type(position$aesthetics, "closure")
  expect_type(position$use_defaults, "closure")
  skip("Not yet implemented")
  expect_identical(position$orientation, "x")
})

test_that("test if correct arguments are assigned with 'split.x'", {
  position <- position_dodge2nudge(kept.origin = "none", direction = "split.x")
  expect_false(position$reverse)
  expect_identical(position$kept.origin, "none")
  expect_identical(position$padding, 0.1)
  expect_identical(position$preserve, "total")
  expect_identical(position$width, 1)
  expect_identical(position$x, 0)
  expect_identical(position$y, 0)
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
  skip_if(condition = utils::packageVersion("ggplot2") < "4.0.0",
          message = "'ggplot2' < 4.0.0")
  expect_false(position$reverse)
  expect_identical(position$required_aes, character(0))
  expect_type(position$aesthetics, "closure")
  expect_type(position$use_defaults, "closure")
  skip("Not yet implemented")
  expect_identical(position$orientation, "x")
})

test_that("test if correct arguments are assigned with 'split'", {
  position <- position_dodge2nudge(kept.origin = "none", direction = "split")
  expect_false(position$reverse)
  expect_identical(position$padding, 0.1)
  expect_identical(position$preserve, "total")
  expect_identical(position$width, 1)
  expect_identical(position$x, 0)
  expect_identical(position$y, 0)
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
  skip_if(condition = utils::packageVersion("ggplot2") < "4.0.0",
          message = "'ggplot2' < 4.0.0")
  expect_false(position$reverse)
  expect_identical(position$required_aes, character(0))
  expect_type(position$aesthetics, "closure")
  expect_type(position$use_defaults, "closure")
  skip("Not yet implemented")
  expect_identical(position$orientation, "x")
})

test_that("test if correct arguments are assigned with 'none'", {
  position <- position_dodge2nudge(kept.origin = "none", direction = "none")
  expect_false(position$reverse)
  expect_identical(position$kept.origin, "none")
  expect_identical(position$padding, 0.1)
  expect_identical(position$preserve, "total")
  expect_identical(position$width, 1)
  expect_identical(position$x, 0)
  expect_identical(position$y, 0)
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
  skip_if(condition = utils::packageVersion("ggplot2") < "4.0.0",
          message = "'ggplot2' < 4.0.0")
  expect_false(position$reverse)
  expect_identical(position$required_aes, character(0))
  expect_type(position$aesthetics, "closure")
  expect_type(position$use_defaults, "closure")
  skip("Not yet implemented")
  expect_identical(position$orientation, "x")
})

test_that("test if correct arguments are assigned with 'center'", {
  position <- position_dodge2nudge(kept.origin = "none", direction = "center")
  expect_false(position$reverse)
  expect_identical(position$kept.origin, "none")
  expect_identical(position$padding, 0.1)
  expect_identical(position$preserve, "total")
  expect_identical(position$width, 1)
  expect_identical(position$x, 0)
  expect_identical(position$y, 0)
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
  skip_if(condition = utils::packageVersion("ggplot2") < "4.0.0",
          message = "'ggplot2' < 4.0.0")
  expect_false(position$reverse)
  expect_identical(position$required_aes, character(0))
  expect_type(position$aesthetics, "closure")
  expect_type(position$use_defaults, "closure")
  skip("Not yet implemented")
  expect_identical(position$orientation, "x")
})

test_that("dodge2nudge plots are correct", {
    expect_no_warning(
      ggplot(mtcars, aes(x = am, y = mpg, label = mpg)) +
        geom_text_s(
          position = position_dodge2nudge(width = 0.85,
                                          x = 0, y = rep(0.5, nrow(mtcars))),
          size = 2.5
        )
    )

  expect_no_warning(
    ggplot(mtcars, aes(x = am, y = mpg, label = mpg)) +
      geom_text_s(
        position = position_dodge2nudge(width = 0.85,
                                        y = 0, x = rep(0.05, nrow(mtcars))),
        size = 2.5
      )
  )

  vdiffr::expect_doppelganger("dodge2-nudge1",
                              ggplot(mtcars, aes(x = am, y = mpg, label = mpg)) +
                                geom_text_s(
                                  position = position_dodge2nudge(width = 0.85,
                                                                 y = 0, x = c(0.1)),
                                  size = 2.5
                                )
  )

  vdiffr::expect_doppelganger("dodge2-nudge2",
                              ggplot(mtcars, aes(x = am, y = mpg, label = mpg)) +
                                geom_text_s(
                                  position = position_dodge2nudge(width = 0.85,
                                                                 y = 0, x = c(-0.05, 0.05)),
                                  size = 2.5
                                )
  )

  skip_on_os(os = "mac")
  vdiffr::expect_doppelganger("dodge3-nudge3",
                              ggplot(mtcars, aes(x = am, y = mpg, label = mpg)) +
                                geom_text_s(
                                  position = position_dodge2nudge(width = 0.85,
                                                                  direction = "split",
                                                                  y = 0, x = c(0.1)),
                                  size = 2.5
                                )
  )

}
)

#' Convert an R object into a tibble
#'
#' This functions tries to convert any R object into a data.frame object.
#' If \code{x} is already a data.frame, it is returned as is. If it is
#' a list or a vector it is converted by means of \code{as.data.frame()}.
#' If of any other type, a conversion into an object of class \code{xts} is
#' attempted by means of \code{try.xts()} and if successful the \code{xts}
#' object is converted into a data frame with a variable \code{time}
#' containing times as \code{POSIXct} and the remaining data columns with
#' the time series data. In this conversion row names are stripped.
#'
#' @param x An R object
#' @param time.resolution character The time unit to which the returned time
#'   values will be rounded.
#' @param as.numeric logical If TRUE convert time to numeric, expressed as
#'   fractional calendar years.
#' @param col.names character vector
#'
#' @return A \code{tibble::tibble} object, derived from \code{data.frame}.
#'
#' @note This function can be used to easily convert time series data into a
#'   format that can be easily plotted with package \code{ggplot2}.
#'   \code{try_tibble} is another name for \code{try_data_frame} which tracks
#'   the separation and re-naming of \code{data_frame} into
#'   \code{tibble::tibble} in the imported packages.
#'
#' @section Warning!: The time zone was set to "UTC" by try.xts() in the test
#'   cases I used. Setting TZ to "UTC" can cause some trouble as several
#'   frequently used functions have as default the local or system TZ and will
#'   apply a conversion before printing or plotting time data, which in addition
#'   is affected by summer/winter time transitions. This should be taken into
#'   account as even for yearly data when conversion is to POSIXct a day (1st of
#'   January) will be set, but then shifted some hours if printed on a TZ
#'   different from "UTC". I recommend reading the documentation of package
#'   \code{\link[lubridate]{lubridate-package}} where the irregularities of time
#'   data and the difficulties they cause are very well described. In many cases
#'   when working with time series with yearly observations it is best to work
#'   with numeric values for years.
#'
#' @export
#'
#' @examples
#' class(lynx)
#' try_tibble(lynx)
#' try_tibble(lynx, as.numeric = TRUE)
#' try_tibble(lynx, "year")
#' class(austres)
#' try_tibble(austres)
#' try_tibble(austres, as.numeric = TRUE)
#' try_tibble(austres, "quarter")
#' class(cars)
#' try_tibble(cars)
#'
try_data_frame <- function(x,
                           time.resolution = "month",
                           as.numeric = FALSE,
                           col.names = NULL) {
  if (inherits(x, "data.frame")) {
    return(x)
  }
  if (!xts::xtsible(x) &&
      (is.list(x) || is.factor(x) || is.vector(x) || is.matrix(x))) {
    return(as.data.frame(x))
  }
  if (xts::is.xts(x)) {
    data.xts <- x
  } else {
    stopifnot(xts::xtsible(x))
    data.xts <- xts::as.xts(x)
  }
  times.raw <- zoo::index(data.xts) # because TZ = "UTC"
  # index is never numeric in an xts object
  if (as.numeric) {
    if (inherits(times.raw, "yearmon")) {
      times <- as.numeric(times.raw)
    } else if (inherits(times.raw, "yearqtr")) {
      times <- zoo::as.Date(times.raw)
      times <- lubridate::decimal_date(times)
    } else {
      times <- lubridate::decimal_date(times.raw)
    }
    times <- as.double(times) # remove "ts" class attribute
  } else {
    if (lubridate::is.POSIXct(times.raw) || lubridate::is.Date(times.raw)) {
      times <- times.raw
    } else if (inherits(times.raw, "yearmon")) {
      times <- as.POSIXct(format(times.raw, "%Y-%m-01"),  tz = "UTC", format = "%Y-%m-%d")
    } else if (inherits(times.raw, "yearqtr")) {
      times <- zoo::as.Date(times.raw)
    } else {
      stop("class '", class(times.raw), "' used as index is not supported")
    }
    if (lubridate::tz(times) == "") {
      times <- lubridate::with_tz(times, "UTC")
    }
    if (!is.null(time.resolution) && !is.na(time.resolution)) {
      times <- lubridate::round_date(times, unit = time.resolution)
    }
  }
  if (is.null(names(x))) {
    data.names <- paste(as.character(substitute(x)), sep = "")
  } else {
    data.names <- names(x)
  }

  z <- data.frame(time = times)
  z <- cbind(z, as.data.frame(as.numeric(data.xts)))
  if (length(col.names) == ncol(z)) {
    colnames(z) <- col.names
  } else {
    names(z)[-1] <- data.names
  }
  z
}

#' @export
#'
#' @rdname try_data_frame
try_tibble <- function(x,
                       time.resolution = "month",
                       as.numeric = FALSE,
                       col.names = NULL) {
  z <- try_data_frame(x = x,
                      time.resolution = time.resolution,
                      as.numeric = as.numeric,
                      col.names = col.names)
  rownames(z) <- NULL
  tibble::as_tibble(z)
}

#' Create a new ggplot plot from time series data
#'
#' \code{ggplot()} initializes a ggplot object. It can be used to
#' declare the input spectral object for a graphic and to optionally specify the
#' set of plot aesthetics intended to be common throughout all
#' subsequent layers unless specifically overridden.
#'
#' \code{ggplot()} is typically used to construct a plot
#' incrementally, using the + operator to add layers to the
#' existing ggplot object. This is advantageous in that the
#' code is explicit about which layers are added and the order
#' in which they are added. For complex graphics with multiple
#' layers, initialization with \code{ggplot} is recommended.
#'
#' There are three common ways to invoke \code{ggplot}:
#' \itemize{
#'    \item \code{ggplot(ts, aes(x, y, <other aesthetics>))}
#'    \item \code{ggplot(ts)}
#'   }
#' The first method is recommended if all layers use the same
#' data and the same set of aesthetics, although this method
#' can also be used to add a layer using data from another
#' data frame. See the first example below. The second
#' method specifies the default spectrum object to use for the plot, and the
#' units to be used for y in the plot,
#' but no aesthetics are defined up front. This is useful when
#' one data frame is used predominantly as layers are added,
#' but the aesthetics may vary from one layer to another. The
#' third method specifies the default spectrum object to use for the plot,
#' but no aesthetics are defined up front. This is useful when
#' one spectrum is used predominantly as layers are added,
#' but the aesthetics may vary from one layer to another.
#'
#' @param data Default spectrum dataset to use for plot. If not a spectrum, the
#'   methods used will be those defined in package \code{ggplot2}. See \code{\link[ggplot2]{ggplot}}.
#'   If not specified,
#'   must be suppled in each layer added to the plot.
#' @param mapping Default list of aesthetic mappings to use for plot.
#'   If not specified, in the case of spectral objects, a default mapping will
#'   be used.
#' @param time.resolution character The time unit to which the returned time
#'   values will be rounded.
#' @param as.numeric logical If TRUE convert time to numeric, expressed as
#'   fractional calendar years.
#' @param ... Other arguments passed on to methods. Not currently used.
#' @param environment If an variable defined in the aesthetic mapping is not
#'   found in the data, ggplot will look for it in this environment. It defaults
#'   to using the environment in which \code{ggplot()} is called.
#'
#' @return A \code{"ggplot"} object.
#'
#' @export
#' @examples
#' ggplot(lynx) + geom_line()
#'
#' @note Current implementation does not merge default mapping with user
#' supplied mapping. If user supplies a mapping, it is used as is.
#' To add to the default mapping, aes() can be used by itself to compose
#' the ggplot.
#'
#' @name ggplot
#'
ggplot.ts <-
  function(data, mapping = NULL, ...,
           time.resolution = "day",
           as.numeric = TRUE,
           environment = parent.frame()) {
    data.name <- substitute(data)
    if (!is.name(data.name)) {
      # likely an expression
      data.name <- as.name("y")
    }
    time.name <- as.name("time")
    data.df <- try_tibble(data,
                          time.resolution = time.resolution,
                          as.numeric = as.numeric,
                          col.names = c(time.name, data.name)
                          )
    if (is.null(mapping)) {
      mapping <- ggplot2::aes(x = {{time.name}}, y = {{data.name}})
    }
    ggplot2::ggplot(data = data.df,
                    mapping =  mapping,
                    ... = ...,
                    environment = environment)
  }

#' @rdname ggplot
#'
#' @export
#'
ggplot.xts <- ggplot.ts

context("try_tibble")

lynx

test_that("try_tibble, compare output", {

  tb <- try_tibble(lynx, as.numeric = TRUE)

  expect_equal(length(lynx), nrow(tb))
})

test_that("try_tibble, check class", {

  tb <- try_tibble(lynx, as.numeric = TRUE)

  expect_s3_class(tb, "tbl_df")
})

test_that("try_tibble, check format of time", {

  df <- try_tibble(lynx, "year")
  expect_equal(class(df$time), "Date")

  df <- try_tibble(lynx, as.numeric = TRUE)
  expect_equal(class(df$time), "numeric")
})

test_that("try_tibble, with xts object with hourly data", {

  my.xts <- xts::as.xts(data.frame(a = 1:50,
                                   t = lubridate::now() + lubridate::hours(1:50)))
  df <- try_tibble(my.xts, "year")
  expect_equal(class(df$time)[1], "POSIXct")

  df <- try_tibble(my.xts, as.numeric = TRUE)
  expect_equal(class(df$time), "numeric")
})

test_that("try_tibble, with xts object with daily data", {

  my.xts <- xts::as.xts(data.frame(a = 1:50,
                                   t = lubridate::now() + lubridate::days(1:50)))
  df <- try_tibble(my.xts, "day")
  expect_equal(class(df$time)[1], "POSIXct")

  df <- try_tibble(my.xts, as.numeric = TRUE)
  expect_equal(class(df$time), "numeric")
})

test_that("try_tibble, with xts object with daily data as Dates", {

  my.xts <- xts::as.xts(data.frame(a = 1:50,
                                   t = as.Date(
                                     lubridate::now() + lubridate::days(1:50)
                                   )))
  df <- try_tibble(my.xts, "day")
  expect_equal(class(df$time)[1], "Date")

  df <- try_tibble(my.xts, as.numeric = TRUE)
  expect_equal(class(df$time), "numeric")
})

context("try_data_frame")

# lists and data frames

test_that("try_data_frame, if data frame pass through", {

  arg.df <- data.frame(x = 1:10, y = 10:1)
  df <- try_data_frame(arg.df)

  expect_s3_class(df, "data.frame")
  expect_identical(arg.df, df)
})

test_that("try_data_frame, if list to data frame", {

  arg.lst <- list(x = 1:10, y = 10:1)
  df <- try_data_frame(arg.lst)

  expect_s3_class(df, "data.frame")
  expect_equal(as.data.frame(arg.lst), df)
})

# lynx (year)

test_that("try_data_frame, compare output", {

  df <- try_data_frame(lynx, as.numeric = TRUE)

  expect_equal(length(lynx), nrow(df))
})

test_that("try_data_frame, check if dataframe", {

  df <- try_data_frame(lynx, as.numeric = TRUE)

  expect_s3_class(df, "data.frame")
})

test_that("try_data_frame, check format of time", {

  df <- try_data_frame(lynx, "year")
  expect_equal(class(df$time), "Date")

  df <- try_data_frame(lynx, as.numeric = TRUE)
  expect_equal(class(df$time), "numeric")
})

test_that("try_data_frame, check format of time with xts", {
  arg.xts <- xts::as.xts(lynx)
  df <- try_data_frame(arg.xts, "year")
  expect_equal(class(df$time), "Date")

  df <- try_data_frame(lynx, as.numeric = TRUE)
  expect_equal(class(df$time), "numeric")
})

test_that("ggplot.ts, if list to data frame num", {

  p <- ggplot(lynx)

  expect_s3_class(p$data, "data.frame")
  expect_named(p$data, c("time", "lynx"))
  expect_is(p$data$time, "numeric")
  expect_is(p$data$lynx, "numeric")
})

test_that("ggplot.ts, if list to data frame date", {

  p <- ggplot(lynx, as.numeric = FALSE)

  expect_s3_class(p$data, "data.frame")
  expect_named(p$data, c("time", "lynx"))
  expect_is(p$data$time, "Date")
  expect_is(p$data$lynx, "numeric")
})

# austres (quarterly)

test_that("try_data_frame, compare output", {

  df <- try_data_frame(austres, as.numeric = TRUE)

  expect_equal(length(austres), nrow(df))
})

test_that("try_data_frame, check if dataframe", {

  df <- try_data_frame(austres, as.numeric = TRUE)

  expect_s3_class(df, "data.frame")
})

test_that("try_data_frame, check format of time", {

  df <- try_data_frame(austres, "quarter")
  expect_equal(class(df$time), "Date")

  df <- try_data_frame(austres, "month")
  expect_equal(class(df$time), "Date")

  df <- try_data_frame(austres, "week")
  expect_equal(class(df$time), "Date")

  df <- try_data_frame(austres, "day")
  expect_equal(class(df$time), "Date")

  df <- try_data_frame(austres, as.numeric = TRUE)
  expect_equal(class(df$time), "numeric")
})

context("ggplot.ts")

test_that("ggplot.ts works as expected with the yearly time series.", {
  data <- lynx

  expect_silent(p <- ggplot.ts(data) + geom_line())
  res <- layer_data(p)

  expect_equal(res$x, as.numeric(time(data)))
  expect_equal(res$y, as.numeric(data))

  # argument to data is not an object name
  expect_silent(p <- ggplot.ts(get("data")) + geom_line())
  res <- layer_data(p)

  expect_equal(res$x, as.numeric(time(data)))
  expect_equal(res$y, as.numeric(data))
})

test_that("ggplot.ts works as expected with the monthly time series.", {
  data <- sunspots

  expect_silent(p <- ggplot.ts(data) + geom_line())
  res <- layer_data(p)

  expect_equal(res$x, as.numeric(time(data)))
  expect_equal(res$y, as.numeric(data))
})

test_that("ggplot.ts works as expected with the quaterly time series.", {
  data <- presidents

  expect_silent(p <- ggplot.ts(data) + geom_line())
  res <- layer_data(p)

  expect_equal(res$x, as.numeric(time(data)), tolerance = 0.001)
  expect_equal(res$y, as.numeric(data))
})

test_that("ggplot.ts time resolution works as expected.", {
  data <- sunspot.month
  expected <- zoo::as.Date(time(data))
  expected <- lubridate::round_date(expected, unit = "year")
  expected_pos <- as.numeric(as.POSIXct(expected))

  expect_silent(
    p <- ggplot.ts(data, as.numeric = FALSE, time.resolution = "year") +
      geom_point()
  )
  res <- layer_data(p, 1)

  expect_equal(res$x, expected_pos)
  expect_equal(res$y, as.numeric(data))
})

test_that("ggplot.ts mapping works as expected.", {
  data <- sunspot.month

  data <- sunspot.month
  expected <- zoo::as.Date(time(data))
  expected <- lubridate::round_date(expected, unit = "10 year")

  expected_med <- as.numeric(tapply(as.numeric(data), expected, median))
  expected_q1 <- as.numeric(tapply(as.numeric(data), expected, quantile, probs = 0.25))
  expected_q3 <- as.numeric(tapply(as.numeric(data), expected, quantile, probs = 0.75))

  expect_silent(
    p <- ggplot.ts(
      data,
      mapping = aes(x = data, y = time, group = time),
      as.numeric = FALSE,
      time.resolution = "10 year"
    ) +
      geom_boxplot()
  )

  res <- layer_data(p)

  expect_equal(res$y, unique(as.numeric(as.POSIXct(expected))))
  expect_equal(res$xmiddle, expected_med)
  expect_equal(res$xlower, expected_q1)
  expect_equal(res$xupper, expected_q3)
})

#' @title Select and slice a tibble nested in \code{data}
#'
#' @description \code{stat_fmt_tb} selects, reorders and/or renames columns and
#'   or rows of a tibble nested in \code{data}. It can also apply user supplied
#'   functions to data columns. This stat is intended to be used to pre-process
#'   \code{tibble} objects mapped to the \code{label} aesthetic before adding
#'   them to a plot with \code{geom_table}.
#'
#' @param mapping The aesthetic mapping, usually constructed with
#'   \code{\link[ggplot2]{aes}} or \code{\link[ggplot2]{aes_}}. Only needs to be
#'   set at the layer level if you are overriding the plot defaults.
#' @param data A layer specific dataset - only needed if you want to override
#'   the plot defaults.
#' @param geom The geometric object to use display the data
#' @param position The position adjustment to use for overlapping points on this
#'   layer
#' @param show.legend logical. Should this layer be included in the legends?
#'   \code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
#'   never includes, and \code{TRUE} always includes.
#' @param inherit.aes If \code{FALSE}, overrides the default aesthetics, rather
#'   than combining with them. This is most useful for helper functions that
#'   define both data and aesthetics and shouldn't inherit behaviour from the
#'   default plot specification, e.g. \code{\link[ggplot2]{borders}}.
#' @param ... other arguments passed on to \code{\link[ggplot2]{layer}}. This
#'   can include aesthetics whose values you want to set, not map. See
#'   \code{\link[ggplot2]{layer}} for more details.
#' @param na.rm	a logical indicating whether \code{NA} values should be stripped
#'   before the computation proceeds.
#' @param digits integer indicating the number of significant digits to be
#'   retained in data. Use \code{digits = Inf} to skip.
#' @param tb.vars,tb.rows character or numeric vectors, optionally named, used
#'   to select and/or rename the columns or rows in the table
#'   returned.
#' @param tb.funs named list of functions to be applied to \code{data} columns.
#' @param table.theme NULL, list or function A 'gridExtra' \code{ttheme}
#'   definition, or a constructor for a \code{ttheme} or \code{NULL} for
#'   default.
#' @param table.rownames,table.colnames logical flag to enable or disabling
#'   printing of row names and column names.
#' @param table.hjust numeric Horizontal justification for the core and column
#'   headings of the table.
#' @param parse If \code{TRUE}, the labels will be parsed into expressions and
#'   displayed as described in \code{?plotmath}.
#'
#' @details One or more functions to be applied can be passed in a named list to
#'   parameter `tb.funs`. Functions are matched by name to columns, after column
#'   selection and renaming have been applied.
#'
#' @seealso See \code{\link{geom_table}} for details on how tables respond
#'   to mapped aesthetics and table themes. For details on predefined table
#'   themes see \code{\link{ttheme_gtdefault}}.
#'
#' @section Computed variables: The output of sequentially applying
#'   \code{\link[dplyr]{slice}} with \code{tb.rows} as argument and
#'   \code{\link[dplyr]{select}} with \code{tb.vars} to a list variable
#'   list mapped to \code{label} and containing a single tibble per row
#'   in \code{data}.
#'
#' @return A plot layer instance. Using as output \code{data} a copy of the
#'   input \code{data} in which the data frames mapped to \code{label} have been
#'   modified.
#'
#' @export
#'
#' @examples
#' my.df <-
#'   tibble::tibble(
#'     x = c(1, 2),
#'     y = c(0, 4),
#'     group = c("A", "B"),
#'     tbs = list(a = tibble::tibble(Xa = 1:6, Y = rep(c("x", "y"), 3)),
#'                b = tibble::tibble(Xb = 1:3, Y = "x"))
#'   )
#'
#' ggplot(my.df, aes(x, y, label = tbs)) +
#'   stat_fmt_tb() +
#'   expand_limits(x = c(0,3), y = c(-2, 6))
#'
#' # Hide column names, diplay row names
#' ggplot(my.df, aes(x, y, label = tbs)) +
#'   stat_fmt_tb(table.colnames = FALSE,
#'               table.rownames = TRUE) +
#'   expand_limits(x = c(0,3), y = c(-2, 6))
#'
#' # Use a theme for the table
#' ggplot(my.df, aes(x, y, label = tbs)) +
#'   stat_fmt_tb(table.theme = ttheme_gtlight) +
#'   expand_limits(x = c(0,3), y = c(-2, 6))
#'
#' # selection and renaming by column position
#' ggplot(my.df, aes(x, y, label = tbs)) +
#'   stat_fmt_tb(tb.vars = c(value = 1, group = 2),
#'                tb.rows = 1:3) +
#'   expand_limits(x = c(0,3), y = c(-2, 6))
#'
#' # apply functions to columns
#' ggplot(my.df, aes(x, y, label = tbs)) +
#'   stat_fmt_tb(tb.vars = c(value = 1, group = 2),
#'                tb.rows = 1:3,
#'                tb.funs = list(group = function(x) {sprintf("italic(%s)", x)},
#'                               value = function(x) {ifelse(x > 2, "bold(zz)", x)}),
#'                parse = TRUE) +
#'   expand_limits(x = c(0,3), y = c(-2, 6))
#'
#' # selection, reordering and renaming by column position
#' ggplot(my.df, aes(x, y, label = tbs)) +
#'   stat_fmt_tb(tb.vars = c(group = 2, value = 1),
#'               tb.rows = 1:3) +
#'   expand_limits(x = c(0,3), y = c(-2, 6))
#'
#' # selection and renaming, using partial matching to column name
#' ggplot(my.df, aes(x, y, label = tbs)) +
#'   stat_fmt_tb(tb.vars = c(value = "X", group = "Y"),
#'               tb.rows = 1:3) +
#'   expand_limits(x = c(0,3), y = c(-2, 6))
#'
stat_fmt_tb <- function(mapping = NULL,
                        data = NULL,
                        geom = "table",
                        tb.vars = NULL,
                        tb.rows = NULL,
                        tb.funs = list(),
                        digits = 3,
                        position = "identity",
                        table.theme = NULL,
                        table.rownames = FALSE,
                        table.colnames = TRUE,
                        table.hjust = 0.5,
                        parse = FALSE,
                        na.rm = FALSE,
                        show.legend = FALSE,
                        inherit.aes = TRUE,
                        ...) {

  # replace NULL with default
  if (is.null(table.theme)) {
    table.theme <-
      getOption("ggpmisc.ttheme.default",
                default = ggpp::ttheme_gtdefault)
  }

  ggplot2::layer(
    stat = StatFmtTb, data = data, mapping = mapping, geom = geom,
    position = position, show.legend = show.legend, inherit.aes = inherit.aes,
    params = list(tb.vars = tb.vars,
                  tb.rows = tb.rows,
                  tb.funs = tb.funs,
                  digits = digits,
                  table.theme = table.theme,
                  table.rownames = table.rownames,
                  table.colnames = table.colnames,
                  table.hjust = table.hjust,
                  parse = parse,
                  na.rm = na.rm,
                  ...)
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
StatFmtTb <-
  ggplot2::ggproto("StatFmtTb", ggplot2::Stat,
                   compute_group =   function(data,
                                              scales,
                                              tb.vars = NULL,
                                              tb.rows = NULL,
                                              tb.funs = list(),
                                              digits = 3) {
                     stopifnot(is.list(data$label))

                     for (tb.idx in seq_along(data$label)) {
                       temp_tb <- data$label[tb.idx][[1]]

                       if (!is.data.frame(temp_tb)) {
                         message("Skipping object of class ", class(temp_tb))
                         next()
                       }

                       num.cols <- sapply(temp_tb, is.numeric)
                       temp_tb[num.cols] <-
                         signif(temp_tb[num.cols], digits = digits)

                       if (!is.null(tb.vars)) {
                         if (is.character(tb.vars)) {
                           idxs <- pmatch(tb.vars, colnames(temp_tb))
                           if (length(idxs) < length(tb.vars) || anyNA(idxs)) {
                             warning("Attempt to select nonexistent columns")
                             idxs <- stats::na.omit(idxs)
                             # no renaming possible, as we do not know which name was not matched
                             tb.vars <- unname(tb.vars)
                           }
                         } else {
                           idxs <- unname(tb.vars)
                           if (any(idxs > ncol(temp_tb))) {
                             warning("Attempt to select nonexistent columns")
                             idxs <- idxs[idxs <= ncol(temp_tb)]
                             tb.vars <- tb.vars[idxs]
                           }
                         }
                         # if (length(idxs) < ncol(temp_tb)) {
                         #   message("Dropping column(s) from table.")
                         # }
                         if (length(idxs) < 1L) {
                           message("No matching column(s).")
                           temp_tb <- NULL
                         } else {
                           temp_tb <- temp_tb[ , idxs]
                           if (!is.null(names(tb.vars))) {
                             # support renaming of only some selected columns
                             selector <- names(tb.vars) != ""
                             colnames(temp_tb)[selector] <- names(tb.vars)[selector]
                           }
                         }
                       }

                       if (!is.null(tb.rows) && !is.null(temp_tb)) {
                         if (is.character(tb.rows)) {
                           idxs <- pmatch(tb.rows, rownames(temp_tb))
                           if (length(idxs) < length(tb.rows) || anyNA(idxs)) {
                             warning("Attempt to select nonexistent rows")
                             idxs <- stats::na.omit(idxs)
                             # no renaming possible, as we do not know which name was not matched
                             tb.rows <- unname(tb.rows)
                           }
                         } else {
                           idxs <- unname(tb.rows)
                           if (any(idxs > nrow(temp_tb))) {
                             warning("Attempt to select nonexistent rows")
                             idxs <- idxs[idxs <= nrow(temp_tb)]
                             tb.rows <- tb.rows[idxs]
                           }
                         }
                         # if (length(idxs) < nrow(temp_tb)) {
                         #   message("Dropping row(s) from table.")
                         # }
                         if (length(idxs) < 1L) {
                           warning("No matching row(s).")
                           temp_tb <- NULL
                         } else {
                           temp_tb <- temp_tb[idxs, ]
                           if (!is.null(names(tb.rows))) {
                             # support renaming of only some selected rows
                             selector <- names(tb.rows) != ""
                             colnames(temp_tb)[selector] <- names(tb.rows)[selector]
                           }
                         }
                       }

                       funs2apply <-
                         intersect(names(tb.funs), colnames(temp_tb))

                       for (f in funs2apply) {
                         if (is.character(tb.funs[[f]])) {
                           tb.funs[[f]] <- match.fun(tb.funs[[f]])
                         }
                         if (is.function(tb.funs[[f]])) {
                           temp_tb[[f]] <- tb.funs[[f]](temp_tb[[f]])
                         } else {
                           warning("'tb.funs[[", f, "]]' is not a function or a visible function name", sep = "")
                           temp_tb[[f]] <- NA
                         }
                       }

                       data$label[tb.idx] <- list(temp_tb)

                     }

                     data
                   },

                   required_aes = c("x", "y", "label")
)

context("stat_fmt_tb")

test_that("stat_fmt_tb", {
  my.df <-
    tibble::tibble(
      x = c(1, 2),
      y = c(0, 4),
      group = c("A", "B"),
      tbs = list(a = tibble::tibble(Xa = 1:6, Y = rep(c("x", "y"), 3)),
                 b = tibble::tibble(Xb = 1:3, Y = "x"))
    )

  result <- ggplot(my.df, aes(x, y, label = tbs)) +
    stat_fmt_tb() +
    expand_limits(x = c(0,3), y = c(-2, 6)) +
    geom_point()
  expect_s3_class(result, "ggplot")

  expect_silent(result <- ggplot(my.df, aes(x, y, label = tbs)) +
    stat_fmt_tb() +
    expand_limits(x = c(0,3), y = c(-2, 6)))
  result <- ggplot2::layer_data(result,1)
  result1 <- result$label[[1]]
  expected <- tibble::tibble(Xa = as.numeric(1:6), Y = rep(c("x", "y"), 3))
  expect_identical(result1, expected)

  result2 <- result$label[[2]]
  expected <-tibble::tibble(Xb = as.numeric(1:3), Y = "x")
  expect_identical(result2, expected)

  # Hide column names, diplay row names
  expect_silent(result2 <- ggplot(my.df, aes(x, y, label = tbs)) +
    stat_fmt_tb(table.colnames = FALSE,
                table.rownames = TRUE))

  expect_silent(result3 <- ggplot(my.df, aes(x, y, label = tbs)) +
                  stat_fmt_tb(tb.funs = list(Xa = function(x) {x * 2},
                                             Xb = function(x) {x * 3},
                                             Y = function(x) {toupper(x)})) +
                  expand_limits(x = c(0,3), y = c(-2, 6)))
  result <- ggplot2::layer_data(result3,1)
  expect_equal(result$label[[1]][["Xa"]], my.df$tbs[["a"]][["Xa"]] * 2)
  expect_equal(result$label[[2]][["Xb"]], my.df$tbs[["b"]][["Xb"]] * 3)
  expect_equal(result$label[[1]][["Y"]], toupper(my.df$tbs[["a"]][["Y"]]))
  expect_equal(result$label[[2]][["Y"]], toupper(my.df$tbs[["b"]][["Y"]]))

  # Use a theme for the table
  ggplot(my.df, aes(x, y, label = tbs)) +
    stat_fmt_tb(table.theme = ttheme_gtlight) +
    expand_limits(x = c(0,3), y = c(-2, 6))

  vdiffr::expect_doppelganger("stat_fmt_tb_1",
                              ggplot(my.df, aes(x, y, label = tbs)) +
                                stat_fmt_tb(table.theme = ttheme_gtlight) +
                                expand_limits(x = c(0,4), y = c(-2, 6))
  )

  # selection and renaming by column position
  vdiffr::expect_doppelganger("stat_fmt_tb_2",
                              ggplot(my.df, aes(x, y, label = tbs)) +
                              stat_fmt_tb(tb.vars = c(value = 1, group = 2),
                                           tb.rows = 1:3) +
                              expand_limits(x = c(0,3), y = c(-2, 6)))


  # selection, reordering and renaming by column position
  vdiffr::expect_doppelganger("stat_fmt_tb_3",
                              ggplot(my.df, aes(x, y, label = tbs)) +
                              stat_fmt_tb(tb.vars = c(group = 2, value = 1),
                                          tb.rows = 1:3) +
                              expand_limits(x = c(0,3), y = c(-2, 6)))

  # selection and renaming, using partial matching to column name
  vdiffr::expect_doppelganger("stat_fmt_tb_4",
                              ggplot(my.df, aes(x, y, label = tbs)) +
                              stat_fmt_tb(tb.vars = c(value = "X", group = "Y"),
                                          tb.rows = 1:3) +
                              expand_limits(x = c(0,3), y = c(-2, 6)))
})

#' Reference lines: horizontal plus vertical, and quadrants
#'
#' \code{geom_vhlines()} adds in a single layer both vertical and horizontal
#' guide lines. Can be thought of as a convenience function that helps with
#' producing consistent vertical and horizontal guide lines. It behaves like
#' \code{geom_vline()} and \code{geom_hline()}.
#' \code{geom_quadrant_lines()} displays the boundaries of four quadrants
#' with an arbitrary origin. The quadrants are specified in the same way as
#' in \code{stat_quadrant_counts()} and is intended to be used to add guide
#' lines consistent with the counts by quadrant computed by this stat.
#'
#' @param mapping The aesthetic mapping, usually constructed with
#'   \code{\link[ggplot2]{aes}}. Only needs to be set at the layer level if you
#'   are overriding the plot defaults.
#' @param data A layer specific data set - only needed if you want to override
#'   the plot defaults.
#' @param stat The statistic object to use display the data
#' @param position The position adjustment to use for overlapping points on this
#'   layer
#' @param show.legend logical. Should this layer be included in the legends?
#'   \code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
#'   never includes, and \code{TRUE} always includes.
#' @param inherit.aes If \code{FALSE}, overrides the default aesthetics, rather
#'   than combining with them. This is most useful for helper functions that
#'   define both data and aesthetics and should not inherit behaviour from the
#'   default plot specification, e.g. \code{\link[ggplot2]{borders}}.
#' @param ... other arguments passed on to \code{\link[ggplot2]{layer}}. This
#'   can include aesthetics whose values you want to set, not map. See
#'   \code{\link[ggplot2]{layer}} for more details.
#' @param na.rm	a logical indicating whether NA values should be stripped before
#'   the computation proceeds.
#' @param pool.along character, one of \code{"none"}, \code{"x"}, \code{"y"}, or
#'   \code{"xy"} indicating whether to plot or not lines separating quadrants.
#' @param xintercept,yintercept numeric vectors the coordinates of the origin of the
#'   quadrants.
#'
#' @details While \code{geom_vhlines()} does not provide defaults for the
#'   intercepts and accepts vectors of length > 1, \code{geom_quadrant_lines()}
#'   sets by default the intercepts to zero producing the natural quadrants and
#'   only accepts vectors of length one per panel. That is \code{geom_vhlines()}
#'   can be used to plot a grid while \code{geom_quadrant_lines()} plots at
#'   most one vertical and one horizontal line. In the case of
#'   \code{geom_quadrant_lines()} the pooling along axes can be specified in the
#'   same way as in \code{\link{stat_quadrant_counts}()}.
#'
#' @family Functions for quadrant and volcano plots
#'
#' @seealso \code{\link[ggplot2]{geom_abline}}, the topic where
#'   \code{geom_vline()} and \code{geom_hline()} are described.
#'
#' @return A plot layer instance.
#'
#' @export
#'
#' @examples
#' # generate artificial data
#' set.seed(4321)
#' x <- 1:100
#' y <- rnorm(length(x), mean = 10)
#' my.data <- data.frame(x, y)
#'
#' ggplot(my.data, aes(x, y)) +
#'   geom_quadrant_lines() +
#'   geom_point()
#'
#' ggplot(my.data, aes(x, y)) +
#'   geom_quadrant_lines(linetype = "dotted") +
#'   geom_point()
#'
#' ggplot(my.data, aes(x, y)) +
#'   geom_quadrant_lines(xintercept = 50,
#'                       yintercept = 10,
#'                       colour = "blue") +
#'   geom_point()
#'
#' ggplot(my.data, aes(x, y)) +
#'   geom_quadrant_lines(xintercept = 50,
#'                       pool.along = "y",
#'                       colour = "blue") +
#'   geom_point()
#'
#' ggplot(my.data, aes(x, y)) +
#'   geom_vhlines(xintercept = c(25, 50, 75),
#'                yintercept = 10 ,
#'                linetype = "dotted",
#'                colour = "red") +
#'   geom_point() +
#'   theme_bw()
#'
#' ggplot(my.data, aes(x, y)) +
#'   geom_vhlines(xintercept = c(25, 50, 75),
#'                yintercept = c(10, 8),
#'                linetype = "dotted",
#'                colour = "red") +
#'   geom_point() +
#'   theme_bw()
#'
geom_quadrant_lines <- function(mapping = NULL,
                                data = NULL,
                                stat = "identity",
                                position = "identity",
                                pool.along = c("none", "x", "y", "xy"),
                                xintercept = 0,
                                yintercept = 0,
                                na.rm = FALSE,
                                show.legend = FALSE,
                                inherit.aes = FALSE,
                                ...) {

  pool.along <- rlang::arg_match(pool.along)

  stopifnot("'xintercept' must have length == 1L" = length(xintercept) <= 1,
            "'yintercept' must have length == 1L" = length(yintercept) <= 1)

  # Act like an annotation
  if (!is.null(xintercept) && !is.null(yintercept)) {
    data <- as.data.frame(list(xintercept = xintercept,
                               yintercept = yintercept))
    mapping <- ggplot2::aes(xintercept = xintercept,
                            yintercept = yintercept)
    show.legend <- FALSE
  } else if (missing(xintercept) || missing(yintercept)) {
    stop("Missing 'xintercept' and/or 'yintercept'")
  }

  ggplot2::layer(
    data = data,
    mapping = mapping,
    stat = stat,
    geom = GeomQuadrantLines,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      pool.along = pool.along,
      na.rm = na.rm,
      ...
    )
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
GeomQuadrantLines <-
  ggplot2::ggproto(
    "GeomQuadrantLines", Geom,
    draw_panel = function(data,
                          panel_params,
                          coord,
                          pool.along = "none",
                          lineend = "butt") {
      ranges <- coord$backtransform_range(panel_params)
      data.hline <- data.vline <- data

      if (pool.along == "xy") {
        return(grid::nullGrob())
      }

      if (!grepl("y", x = pool.along, ignore.case = TRUE)) {
        data.hline[["y"]]    <- data[["yintercept"]][1]
        data.hline[["yend"]] <- data[["yintercept"]][1]
        data.hline[["x"]]    <- ranges[["x"]][1]
        data.hline[["xend"]] <- ranges[["x"]][2]
      }

      if (!grepl("x", x = pool.along, ignore.case = TRUE)) {
        data.vline[["x"]]    <- data[["xintercept"]][1]
        data.vline[["xend"]] <- data[["xintercept"]][1]
        data.vline[["y"]]    <- ranges[["y"]][1]
        data.vline[["yend"]] <- ranges[["y"]][2]
      }

      # discard infinite values
      # other off-range values still respect oob
      #
      data.hline <- data.hline[is.finite(data.hline[["y"]])]
      data.vline <- data.vline[is.finite(data.vline[["x"]])]

      data <- switch(tolower(pool.along),
                     none = rbind(data.hline, data.vline),
                     x = data.hline,
                     y = data.vline,
                     data[NULL, ]
      )

      ggplot2::GeomSegment$draw_panel(unique(data), panel_params, coord, lineend = lineend)

    },

    default_aes =
      ggplot2::aes(colour = "black",
                   linewidth = 0.5,
                   linetype = "dashed",
                   alpha = NA),
    required_aes = c("xintercept", "yintercept"),
    non_missing_aes = c("size", "linetype", "colour"),

    draw_key = draw_key_path,

    rename_size = TRUE
  )

#' @rdname geom_quadrant_lines
#'
#' @export
#'
geom_vhlines <- function(mapping = NULL, data = NULL,
                                stat = "identity", position = "identity",
                                xintercept = NULL, yintercept = NULL,
                                na.rm = FALSE, show.legend = FALSE,
                                inherit.aes = FALSE, ...) {
  # Act like an annotation
  if (!is.null(xintercept) && !is.null(yintercept)) {
    num.rows <- max(length(xintercept), length(yintercept))
    data <- as.data.frame(list(xintercept = rep_len(xintercept, num.rows),
                               yintercept = rep_len(yintercept, num.rows)))
    mapping <- ggplot2::aes(xintercept = xintercept,
                            yintercept = yintercept)
    show.legend <- FALSE
  } else if (xor(missing(xintercept), missing(yintercept))) {
    stop("Arguments should be passed either to none or both of 'xintrecept' and 'yintercept'")
  }

  ggplot2::layer(
    data = data,
    mapping = mapping,
    stat = stat,
    geom = GeomVHLines,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      na.rm = na.rm,
      ...
    )
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
GeomVHLines <-
  ggplot2::ggproto(
    "GeomVHLines", Geom,
    draw_panel = function(data, panel_params, coord, lineend = "butt") {
      ranges <- coord$backtransform_range(panel_params)
      data.hline <- data.vline <- data

      # when xintercept and yintercept differ in lengths some lines will
      # be created more than once, which seems harmless.

      data.hline[["y"]]    <- data[["yintercept"]]
      data.hline[["yend"]] <- data[["yintercept"]]
      data.hline[["x"]]    <- ranges[["x"]][1]
      data.hline[["xend"]] <- ranges[["x"]][2]

      data.vline[["x"]]    <- data[["xintercept"]]
      data.vline[["xend"]] <- data[["xintercept"]]
      data.vline[["y"]]    <- ranges[["y"]][1]
      data.vline[["yend"]] <- ranges[["y"]][2]

      data <- rbind(data.hline, data.vline)

      ggplot2::GeomSegment$draw_panel(unique(data), panel_params, coord, lineend = lineend)

    },

    default_aes = ggplot2::aes(
      colour = "black",
      linewidth = 0.5,
      linetype = 1,
      alpha = NA
    ),
    required_aes = c("xintercept", "yintercept"),
    non_missing_aes = c("size", "linetype", "colour"),

    draw_key = draw_key_path,

    rename_size = TRUE
  )

context("geom_quadrant_lines")

library(ggplot2)
  tst.df <- data.frame(
    x = c(5, -0.2, -0.2, -2.5, -0.6, -0.1),
    y = c(1.1, -0.6, -0.9, -1.6, 0.3, 1.6),
    group = c("A", "A", "A", "B", "B", "B")
  )

test_that("geom_quadrant_lines errors on missing intercept", {
  expect_error(
    ggplot(tst.df, aes(x = x, y = y)) +
      geom_point() +
      geom_quadrant_lines(xintercept = NULL)
  )

  expect_error(
    ggplot(tst.df, aes(x = x, y = y)) +
      geom_point() +
      geom_quadrant_lines(yintercept = NULL)
  )

  # errors in 'ggplot2' code!
  # expect_error(
  #   ggplot(tst.df, aes(x = x, y = y)) +
  #     geom_point() +
  #     geom_quadrant_lines(xintercept = NULL,
  #                         yintercept = NULL)
  # )

})

# tests with intercept coords mapped with aes() are missing!!

test_that("geom_quadrant_lines works as expected with default values", {
  res <- ggplot(tst.df, aes(x = x, y = y)) +
    geom_point() +
    geom_quadrant_lines()

  df_res <- layer_data(res, 2)
  gp <- -1L
  attr(gp, "n") <- 1

  expected <- data.frame(
    xintercept = 0,
    yintercept = 0,
    PANEL = factor(1),
    group = gp,
    colour = "black",
    linewidth = 0.5,
    linetype = "dashed",
    alpha = NA
  )

  expect_equal(df_res, expected)
})

test_that("geom_quadrant_lines works as expected with custom parameter values", {
  res <- ggplot(tst.df, aes(x = x, y = y)) +
    geom_point() +
    geom_quadrant_lines(
      xintercept = 0.5,
      yintercept = 0.5,
      colour = "blue",
      linetype = "dashed"
    )

  df_res <- layer_data(res, 2)
  gp <- -1L
  attr(gp, "n") <- 1

  expected <- data.frame(
    xintercept = 0.5,
    yintercept = 0.5,
    PANEL = factor(1),
    group = gp,
    colour = "blue",
    linewidth = 0.5,
    linetype = "dashed",
    alpha = NA
  )

  expect_equal(df_res, expected)

  # pool.along works as expected.
  res <- ggplot(tst.df, aes(x = x, y = y)) +
    geom_point() +
    geom_quadrant_lines(
      pool.along = "y",
      xintercept = 0.5,
      yintercept = 10,
      colour = "blue",
      linetype = "dashed"
    )

  vdiffr::expect_doppelganger("geom_quadrant_lines_pool_y", res)

  res <- ggplot(tst.df, aes(x = x, y = y)) +
    geom_point() +
    geom_quadrant_lines(
      pool.along = "x",
      xintercept = 0.5,
      yintercept = 10,
      colour = "blue",
      linetype = "dashed"
    )

  vdiffr::expect_doppelganger("geom_quadrant_lines_pool_x", res)
})

test_that("geom_vhlines returns an error if neither x nor y intercept is provided", {
  expect_error(
    print(
      ggplot(tst.df, aes(x = x, y = y)) +
        geom_point() +
        geom_vhlines()
    )
  )

  expect_error(
    ggplot(tst.df, aes(x = x, y = y)) +
      geom_point() +
      geom_vhlines(xintercept = 0)
  )

  expect_error(
    ggplot(tst.df, aes(x = x, y = y)) +
      geom_point() +
      geom_vhlines(yintercept = 0)
  )

})

test_that("geom_vhlines works as expected with multiple intercepts", {
  res <- ggplot(tst.df, aes(x = x, y = y)) +
    geom_point() +
    geom_vhlines(
      xintercept = c(-1, 1),
      yintercept = c(0.5, 1)
    )

  df_res <- layer_data(res, 2)
  gp <- c(-1L, -1L)
  attr(gp, "n") <- 1

  expected <- data.frame(
    xintercept = c(-1, 1),
    yintercept = c(0.5, 1),
    PANEL = factor(c(1, 1)),
    group = gp,
    colour = c("black", "black"),
    linewidth = c(0.5, 0.5),
    linetype = c(1, 1),
    alpha = c(NA, NA)
  )

  expect_equal(df_res, expected)
})

#' @title Filter observations by local 2D density
#'
#' @description \code{stat_dens2d_filter} Filters-out/filters-in observations in
#'   regions of a plot panel with high density of observations, based on the
#'   values mapped to both \code{x} and \code{y} aesthetics.
#'   \code{stat_dens2d_filter_g} does the filtering by group instead of by
#'   panel. This second stat is useful for highlighting observations, while the
#'   first one tends to be most useful when the aim is to prevent clashes among
#'   text labels. If there is no mapping to \code{label} in \code{data}, the
#'   mapping is silently set to \code{rownames(data)}.
#'
#' @details The local density of observations in 2D (\emph{x} and \emph{y}) is
#'   computed with function \code{\link[MASS]{kde2d}} and used to select
#'   observations, passing to the geom a subset of the rows in its \code{data}
#'   input. The default is to select observations in sparse regions of the plot,
#'   but the selection can be inverted so that only observations in the densest
#'   regions are returned. Specific observations can be protected from being
#'   deselected and "kept" by passing a suitable argument to \code{keep.these}.
#'   Logical and integer vectors work as indexes to rows in \code{data}, while a
#'   character vector values are compared to the character values mapped to the
#'   \code{label} aesthetic. A function passed as argument to keep.these will
#'   receive as argument the values in the variable mapped to \code{label} and
#'   should return a character, logical or numeric vector as described above. If
#'   no variable has been mapped to \code{label}, row names are used in its
#'   place.
#'
#'   How many rows are retained in addition to those in \code{keep.these} is
#'   controlled with arguments passed to \code{keep.number} and
#'   \code{keep.fraction}. \code{keep.number} sets the maximum number of
#'   observations selected, whenever \code{keep.fraction} results in fewer
#'   observations selected, it is obeyed.
#'
#'   Computation of density and of the default bandwidth require at least
#'   two observations with different values. If data do not fulfill this
#'   condition, they are kept only if \code{keep.fraction = 1}. This is correct
#'   behavior for a single observation, but can be surprising in the case of
#'   multiple observations.
#'
#'   Parameters \code{keep.these} and \code{exclude.these} make it possible to
#'   force inclusion or exclusion of observations after the density is computed.
#'   In case of conflict, \code{exclude.these} overrides \code{keep.these}.
#'
#' @note Which points are kept and which not depends on how dense a grid is used
#'   and how flexible the density surface estimate is. This depends on the
#'   values passed as arguments to parameters \code{n}, \code{bw} and
#'   \code{kernel}. It is also important to be aware that both
#'   \code{geom_text()} and \code{geom_text_repel()} can avoid overplotting by
#'   discarding labels at the plot rendering stage, i.e., what is plotted may
#'   differ from what is returned by this statistic.
#'
#' @param mapping The aesthetic mapping, usually constructed with
#'   \code{\link[ggplot2]{aes}} or \code{\link[ggplot2]{aes_}}. Only needs
#'   to be set at the layer level if you are overriding the plot defaults.
#' @param data A layer specific dataset - only needed if you want to override
#'   the plot defaults.
#' @param geom The geometric object to use display the data.
#' @param keep.fraction numeric [0..1]. The fraction of the observations (or
#'   rows) in \code{data} to be retained.
#' @param keep.number integer Set the maximum number of observations to retain,
#'   effective only if obeying \code{keep.fraction} would result in a larger
#'   number.
#' @param keep.sparse logical If \code{TRUE}, the default, observations from the
#'   more sparse regions are retained, if \code{FALSE} those from the densest
#'   regions.
#' @param keep.these,exclude.these character vector, integer vector, logical
#'   vector or function that takes one or more variables in data selected by
#'   \code{these.target}. Negative integers behave as in R's extraction methods.
#'   The rows from \code{data} indicated by \code{keep.these} and
#'   \code{exclude.these} are kept or excluded irrespective of the local
#'   density.
#' @param these.target character, numeric or logical selecting one or more
#'   column(s) of \code{data}. If \code{TRUE} the whole \code{data} object is
#'   passed.
#' @param pool.along character, one of \code{"none"}, \code{"x"}, \code{"y"}, or
#'   \code{"xy"} indicating if selection should be done pooling the observations
#'   along the \emph{x}, \code{y}, both axes or none based on quadrants given by
#'   \code{xintercept} and \code{yintercept}.
#' @param xintercept,yintercept numeric The center point of the quadrants.
#' @param invert.selection logical If \code{TRUE}, the complement of the
#'   selected rows are returned.
#' @param h vector of bandwidths for x and y directions. Defaults to normal
#'   reference bandwidth (see bandwidth.nrd). A scalar value will be taken to
#'   apply to both directions.
#' @param n Number of grid points in each direction. Can be scalar or a
#'   length-2 integer vector
#' @param return.density logical vector of lenght 1. If \code{TRUE} add columns
#'   \code{"density"} and \code{"keep.obs"} to the returned data frame.
#' @param position The position adjustment to use for overlapping points on this
#'   layer
#' @param show.legend logical. Should this layer be included in the legends?
#'   \code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
#'   never includes, and \code{TRUE} always includes.
#' @param inherit.aes If \code{FALSE}, overrides the default aesthetics, rather
#'   than combining with them. This is most useful for helper functions that
#'   define both data and aesthetics and shouldn't inherit behaviour from the
#'   default plot specification, e.g. \code{\link[ggplot2]{borders}}.
#' @param ... other arguments passed on to \code{\link[ggplot2]{layer}}. This
#'   can include aesthetics whose values you want to set, not map. See
#'   \code{\link[ggplot2]{layer}} for more details.
#' @param na.rm	a logical value indicating whether NA values should be stripped
#'   before the computation proceeds.
#'
#' @return A plot layer instance. Using as output \code{data} a subset of the
#'   rows in input \code{data} retained based on a 2D-density-based filtering
#'   criterion.
#'
#' @seealso \code{\link{stat_dens2d_labels}} and \code{\link[MASS]{kde2d}} used
#'   internally. Parameters \code{n}, \code{h} in these statistics correspond to
#'   the parameters with the same name in this imported function. Limits are set
#'   to the limits of the plot scales.
#'
#' @family statistics returning a subset of data
#'
#' @examples
#'
#' random_string <-
#'   function(len = 6) {
#'     paste(sample(letters, len, replace = TRUE), collapse = "")
#'   }
#'
#' # Make random data.
#' set.seed(1001)
#' d <- tibble::tibble(
#'   x = rnorm(100),
#'   y = rnorm(100),
#'   group = rep(c("A", "B"), c(50, 50)),
#'   lab = replicate(100, { random_string() })
#' )
#'
#' # filter (and here highlight) 1/10 observations in sparsest regions
#' ggplot(data = d, aes(x, y)) +
#'   geom_point() +
#'   stat_dens2d_filter(colour = "red")
#'
#' # filter observations not in the sparsest regions
#' ggplot(data = d, aes(x, y)) +
#'   geom_point() +
#'   stat_dens2d_filter(colour = "blue", invert.selection = TRUE)
#'
#' # filter observations in dense regions of the plot
#' ggplot(data = d, aes(x, y)) +
#'   geom_point() +
#'   stat_dens2d_filter(colour = "blue", keep.sparse = FALSE)
#'
#' # filter 1/2 the observations
#' ggplot(data = d, aes(x, y)) +
#'   geom_point() +
#'   stat_dens2d_filter(colour = "red", keep.fraction = 0.5)
#'
#' # filter 1/2 the observations but cap their number to maximum 12 observations
#' ggplot(data = d, aes(x, y)) +
#'   geom_point() +
#'   stat_dens2d_filter(colour = "red",
#'                      keep.fraction = 0.5,
#'                      keep.number = 12)
#'
#' # density filtering done jointly across groups
#' ggplot(data = d, aes(x, y, colour = group)) +
#'   geom_point() +
#'   stat_dens2d_filter(shape = 1, size = 3, keep.fraction = 1/4)
#'
#' # density filtering done independently for each group
#' ggplot(data = d, aes(x, y, colour = group)) +
#'   geom_point() +
#'   stat_dens2d_filter_g(shape = 1, size = 3, keep.fraction = 1/4)
#'
#' # density filtering done jointly across groups by overriding grouping
#' ggplot(data = d, aes(x, y, colour = group)) +
#'   geom_point() +
#'   stat_dens2d_filter_g(colour = "black",
#'                        shape = 1, size = 3, keep.fraction = 1/4)
#'
#' # label observations
#' ggplot(data = d, aes(x, y, label = lab, colour = group)) +
#'   geom_point() +
#'   stat_dens2d_filter(geom = "text")
#'
#' ggplot(data = d, aes(x, y, label = lab, colour = group)) +
#'   geom_point() +
#'   stat_dens2d_filter(geom = "text",
#'                      keep.these = function(x) {grepl("^u", x)})
#'
#' ggplot(data = d, aes(x, y, label = lab, colour = group)) +
#'   geom_point() +
#'   stat_dens2d_filter(geom = "text",
#'                      keep.these = function(x) {grepl("^u", x)})
#'
#' ggplot(data = d, aes(x, y, label = lab, colour = group)) +
#'   geom_point() +
#'   stat_dens2d_filter(geom = "text",
#'                      keep.these = 1:30)
#'
#' # looking under the hood with gginnards::geom_debug_group()
#' gginnards.installed <- requireNamespace("gginnards", quietly = TRUE)
#' if (gginnards.installed) {
#'   library(gginnards)
#'
#'   ggplot(data = d, aes(x, y, label = lab, colour = group)) +
#'     stat_dens2d_filter(geom = "debug_group")
#' }
#' if (gginnards.installed) {
#'   ggplot(data = d, aes(x, y, label = lab, colour = group)) +
#'     geom_point() +
#'     stat_dens2d_filter(geom = "debug_group", return.density = TRUE)
#' }
#'
#' @export
#'
stat_dens2d_filter <-
  function(mapping = NULL, data = NULL,
           geom = "point", position = "identity",
           ...,
           keep.fraction = 0.10,
           keep.number = Inf,
           keep.sparse = TRUE,
           keep.these = FALSE,
           exclude.these = FALSE,
           these.target = "label",
           pool.along = c("xy", "x", "y", "none"),
           xintercept = 0,
           yintercept = 0,
           invert.selection = FALSE,
           na.rm = TRUE, show.legend = FALSE,
           inherit.aes = TRUE,
           h = NULL,
           n = NULL,
           return.density = FALSE) {

    pool.along <- rlang::arg_match(pool.along)

    if (any(is.na(keep.fraction) | keep.fraction < 0 | keep.fraction > 1)) {
      stop("Out of range or missing value for 'keep.fraction': ", keep.fraction)
    }
    if (any(is.na(keep.number) | keep.number < 0)) {
      stop("Out of range or missing value for 'keep.number': ", keep.number)
    }
    max.expected.length <- unname(c(none = 4L, x = 2L, y = 2L, xy = 1L)[pool.along])
    if (length(keep.fraction) > max.expected.length) {
      if (max.expected.length == 4L) {
        stop("Length of 'keep.fraction' should not exceed 4")
      } else {
        warning("'keep.fraction' is too long, did you forget to set 'pool.along'?")
      }
    }
    if (length(keep.number) > max.expected.length) {
      if (max.expected.length == 4L) {
        stop("Length of 'keep.number' should not exceed 4")
      } else {
        warning("'keep.number' is too long, did you forget to set 'pool.along'?")
      }
    }

    ggplot2::layer(
      stat = StatDens2dFilter, data = data, mapping = mapping, geom = geom,
      position = position, show.legend = show.legend, inherit.aes = inherit.aes,
      params = list(na.rm = na.rm,
                    keep.fraction = keep.fraction,
                    keep.number = keep.number,
                    keep.sparse = keep.sparse,
                    keep.these = keep.these,
                    exclude.these = exclude.these,
                    these.target = these.target,
                    pool.along = pool.along,
                    xintercept = xintercept,
                    yintercept = yintercept,
                    invert.selection = invert.selection,
                    h = h,
                    n = n,
                    return.density = return.density,
                    ...)
    )
  }

#' @rdname stat_dens2d_filter
#'
#' @export
#'
stat_dens2d_filter_g <-
  function(mapping = NULL, data = NULL,
           geom = "point",
           position = "identity",
           ...,
           keep.fraction = 0.10,
           keep.number = Inf,
           keep.sparse = TRUE,
           keep.these = FALSE,
           exclude.these = FALSE,
           these.target = "label",
           pool.along = c("xy", "x", "y", "none"),
           xintercept = 0,
           yintercept = 0,
           invert.selection = FALSE,
           na.rm = TRUE, show.legend = FALSE,
           inherit.aes = TRUE,
           h = NULL,
           n = NULL,
           return.density = FALSE) {

    pool.along <- rlang::arg_match(pool.along)

    if (any(is.na(keep.fraction) | keep.fraction < 0 | keep.fraction > 1)) {
      stop("Out of range or missing value for 'keep.fraction': ", keep.fraction)
    }
    if (any(is.na(keep.number) | keep.number < 0)) {
      stop("Out of range or missing value for 'keep.number': ", keep.number)
    }
    max.expected.length <- unname(c(none = 4L, x = 2L, y = 2L, xy = 1L)[pool.along])
    if (length(keep.fraction) > max.expected.length) {
      if (max.expected.length == 4L) {
        stop("Length of 'keep.fraction' should not exceed 4")
      } else {
        warning("'keep.fraction' is too long, did you forget to set 'pool.along'?")
      }
    }
    if (length(keep.number) > max.expected.length) {
      if (max.expected.length == 4L) {
        stop("Length of 'keep.number' should not exceed 4")
      } else {
        warning("'keep.number' is too long, did you forget to set 'pool.along'?")
      }
    }

    ggplot2::layer(
      stat = StatDens2dFilterG, data = data, mapping = mapping, geom = geom,
      position = position, show.legend = show.legend, inherit.aes = inherit.aes,
      params = list(na.rm = na.rm,
                    keep.fraction = keep.fraction,
                    keep.number = keep.number,
                    keep.sparse = keep.sparse,
                    keep.these = keep.these,
                    exclude.these = exclude.these,
                    these.target = these.target,
                    pool.along = pool.along,
                    xintercept = xintercept,
                    yintercept = yintercept,
                    invert.selection = invert.selection,
                    h = h,
                    n = n,
                    return.density = return.density,
                    ...)
    )
  }


#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
StatDens2dFilter <-
  ggplot2::ggproto(
    "StatDens2dFilter",
    ggplot2::Stat,
    compute_panel =
#      dens2d_flt_compute_fun,
# see below for duplicated code to ensure 'covr' sees it
  function(data,
           scales,
           keep.fraction,
           keep.number,
           keep.sparse,
           keep.these,
           exclude.these,
           these.target,
           pool.along,
           xintercept,
           yintercept,
           invert.selection,
           h,
           n,
           return.density) {

    force(data)

    keep.these <- these2logical(these = keep.these,
                                data = data,
                                these.target = these.target)

    exclude.these <- these2logical(these = exclude.these,
                                   data = data,
                                   these.target = these.target)

    # discard redundant splits
    if (pool.along != "xy") {
      if (pool.along == "y" &&
          !(xintercept < max(data[["x"]]) &&
            xintercept > min(data[["x"]]))) {
        pool.along <- "xy"
      } else if (pool.along == "x" &&
                 !(yintercept < max(data[["y"]]) &&
                   yintercept > min(data[["y"]]))) {
        pool.along <- "xy"
      } else if (pool.along == "none") {
        if (!(xintercept < max(data[["x"]]) &&
              xintercept > min(data[["x"]])) &&
            !(yintercept < max(data[["y"]]) &&
              yintercept > min(data[["y"]]))) {
          pool.along <- "xy"
        } else if (!(xintercept < max(data[["x"]]) &&
                     xintercept > min(data[["x"]]))) {
          pool.along <- "x"
        } else if (!(yintercept < max(data[["y"]]) &&
                     yintercept > min(data[["y"]]))) {
          pool.along <- "y"
        }
      }
    }

    # make list of logical vectors
    if (pool.along == "y") {
      selectors <-list(q12 = data[["x"]] <= xintercept,
                       q34 = data[["x"]] > xintercept)
      if (length(keep.fraction) != 2L) {
        keep.fraction <- rep_len(keep.fraction, length.out = 2)
      }
      if (length(keep.number) != 2L) {
        if (length(keep.number) == 1L) {
          keep.number <- keep.number %/% 2
        }
        keep.number <- rep_len(keep.number, length.out = 2)
      }
      num.rows <- sapply(selectors, sum) # selectors are logical
    } else if (pool.along == "x") {
      selectors <-list(q23 = data[["y"]] <= yintercept,
                       q41 = data[["y"]] > yintercept)
      if (length(keep.fraction) != 2L) {
        keep.fraction <- rep_len(keep.fraction, length.out = 2)
      }
      if (length(keep.number) != 2L) {
        if (length(keep.number) == 1L) {
          keep.number <- keep.number %/% 2
        }
        keep.number <- rep_len(keep.number, length.out = 2)
      }
      num.rows <- sapply(selectors, sum) # selectors are logical
    } else if (pool.along == "none") {
      selectors <-list(q1 = data[["y"]] >= yintercept & data[["x"]] >= xintercept,
                       q2 = data[["y"]] < yintercept & data[["x"]] >= xintercept,
                       q3 = data[["y"]] < yintercept & data[["x"]] < xintercept,
                       q4 = data[["y"]] > yintercept & data[["x"]] < xintercept)
      if (length(keep.fraction) != 4L) {
        keep.fraction <- rep_len(keep.fraction, length.out = 4)
      }
      if (length(keep.number) != 4L) {
        if (length(keep.number) == 1L) {
          keep.number <- keep.number %/% 4
        }
        keep.number <- rep_len(keep.number, length.out = 4)
      }
      num.rows <- sapply(selectors, sum) # selectors are logical
    } else {
      keep.fraction <- keep.fraction[[1]] # can be a vector or a list
      keep.number <- keep.number[[1]]
      num.rows <- nrow(data)
      selectors <- list(all = rep.int(TRUE, times = num.rows))
    }

    # vectorized
    too.large.frac <- num.rows * keep.fraction > keep.number
    keep.fraction[too.large.frac] <-
      keep.number[too.large.frac] / num.rows[too.large.frac]

    # estimate 2D density
    # data with fewer than 2 rows needs to be treated as a special case as
    # density cannot be estimated
    if (length(unique(data$x)) >= 2L &&
        length(unique(data$y)) >= 2L) {
      if (is.null(h)) {
        h <- c(MASS::bandwidth.nrd(data$x), MASS::bandwidth.nrd(data$y))
      }
      if (is.null(n)) {
        n <- trunc(sqrt(nrow(data))) * 8L
      }
      kk <-  MASS::kde2d(
        data[["x"]], data[["y"]], h = h, n = n,
        lims = c(scales$x$dimension(), scales$y$dimension()))
      dimnames(kk[["z"]]) <- list(kk[["x"]], kk[["y"]])

      # compute 2D density at each observation's coordinates
      kx <- cut(data$x, kk$x, labels = FALSE, include.lowest = TRUE)
      ky <- cut(data$y, kk$y, labels = FALSE, include.lowest = TRUE)
      kz <- sapply(seq_along(kx), function(i) kk$z[kx[i], ky[i]])
    } else {
      if (nrow(data) > 1L) {
        message("Density not computed, too few distinct values in 'x' and/or 'y'")
      }
      kz <- rep_len(1, nrow(data))
    }
    # we construct one logical vector by adding observations/label to be kept
    # we may have a list of 1, 2, or 4 logical vectors
    keep <- logical(nrow(data))
    for (i in seq_along(selectors)) {
      if (keep.fraction[i] == 1) {
        keep[ selectors[[i]] ] <- TRUE
      } else if (keep.fraction[i] != 0  && length(selectors[[i]]) >= 2L) {
        if (keep.sparse) {
          keep[ selectors[[i]] ] <-
            kz[ selectors[[i]] ] < stats::quantile(kz[ selectors[[i]] ],
                                                   keep.fraction[i], names = FALSE)
        } else {
          keep[ selectors[[i]] ] <-
            kz[ selectors[[i]] ] >= stats::quantile(kz[ selectors[[i]] ],
                                                    1 - keep.fraction[i], names = FALSE)
        }
      }
    }
    keep <- (keep | keep.these) & !exclude.these

    if (invert.selection) {
      keep <- !keep
    }

    if (return.density) {
      data[["keep.obs"]] <- keep
      data[["density"]] <- kz
    }

    data[keep, ]

  },
    required_aes = c("x", "y")
  )

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
StatDens2dFilterG <-
  ggplot2::ggproto(
    "StatDens2dFilterG",
    ggplot2::Stat,
    compute_group =
      #      dens2d_flt_compute_fun,
      # see above for duplicated code to ensure 'covr' sees it
      function(data,
               scales,
               keep.fraction,
               keep.number,
               keep.sparse,
               keep.these,
               exclude.these,
               these.target,
               pool.along,
               xintercept,
               yintercept,
               invert.selection,
               h,
               n,
               return.density) {

        force(data)

        keep.these <- these2logical(these = keep.these,
                                    data = data,
                                    these.target = these.target)

        exclude.these <- these2logical(these = exclude.these,
                                       data = data,
                                       these.target = these.target)

        # discard redundant splits
        if (pool.along != "xy") {
          if (pool.along == "y" &&
              !(xintercept < max(data[["x"]]) &&
                xintercept > min(data[["x"]]))) {
            pool.along <- "xy"
          } else if (pool.along == "x" &&
                     !(yintercept < max(data[["y"]]) &&
                       yintercept > min(data[["y"]]))) {
            pool.along <- "xy"
          } else if (pool.along == "none") {
            if (!(xintercept < max(data[["x"]]) &&
                  xintercept > min(data[["x"]])) &&
                !(yintercept < max(data[["y"]]) &&
                  yintercept > min(data[["y"]]))) {
              pool.along <- "xy"
            } else if (!(xintercept < max(data[["x"]]) &&
                         xintercept > min(data[["x"]]))) {
              pool.along <- "x"
            } else if (!(yintercept < max(data[["y"]]) &&
                         yintercept > min(data[["y"]]))) {
              pool.along <- "y"
            }
          }
        }

        # make list of logical vectors
        if (pool.along == "y") {
          selectors <-list(q12 = data[["x"]] <= xintercept,
                           q34 = data[["x"]] > xintercept)
          if (length(keep.fraction) != 2L) {
            keep.fraction <- rep_len(keep.fraction, length.out = 2)
          }
          if (length(keep.number) != 2L) {
            if (length(keep.number) == 1L) {
              keep.number <- keep.number %/% 2
            }
            keep.number <- rep_len(keep.number, length.out = 2)
          }
          num.rows <- sapply(selectors, sum) # selectors are logical
        } else if (pool.along == "x") {
          selectors <-list(q23 = data[["y"]] <= yintercept,
                           q41 = data[["y"]] > yintercept)
          if (length(keep.fraction) != 2L) {
            keep.fraction <- rep_len(keep.fraction, length.out = 2)
          }
          if (length(keep.number) != 2L) {
            if (length(keep.number) == 1L) {
              keep.number <- keep.number %/% 2
            }
            keep.number <- rep_len(keep.number, length.out = 2)
          }
          num.rows <- sapply(selectors, sum) # selectors are logical
        } else if (pool.along == "none") {
          selectors <-list(q1 = data[["y"]] >= yintercept & data[["x"]] >= xintercept,
                           q2 = data[["y"]] < yintercept & data[["x"]] >= xintercept,
                           q3 = data[["y"]] < yintercept & data[["x"]] < xintercept,
                           q4 = data[["y"]] > yintercept & data[["x"]] < xintercept)
          if (length(keep.fraction) != 4L) {
            keep.fraction <- rep_len(keep.fraction, length.out = 4)
          }
          if (length(keep.number) != 4L) {
            if (length(keep.number) == 1L) {
              keep.number <- keep.number %/% 4
            }
            keep.number <- rep_len(keep.number, length.out = 4)
          }
          num.rows <- sapply(selectors, sum) # selectors are logical
        } else {
          keep.fraction <- keep.fraction[[1]] # can be a vector or a list
          keep.number <- keep.number[[1]]
          num.rows <- nrow(data)
          selectors <- list(all = rep.int(TRUE, times = num.rows))
        }

        # vectorized
        too.large.frac <- num.rows * keep.fraction > keep.number
        keep.fraction[too.large.frac] <-
          keep.number[too.large.frac] / num.rows[too.large.frac]

        # estimate 2D density
        # data with fewer than 2 rows is as a special case as density() fails
        if (length(unique(data$x)) >= 2L &&
            length(unique(data$y)) >= 2L) {
          if (is.null(h)) {
            h <- c(MASS::bandwidth.nrd(data$x), MASS::bandwidth.nrd(data$y))
          }
          if (is.null(n)) {
            n <- trunc(sqrt(nrow(data))) * 8L
          }
          kk <-  MASS::kde2d(
            data[["x"]], data[["y"]], h = h, n = n,
            lims = c(scales$x$dimension(), scales$y$dimension()))
          dimnames(kk[["z"]]) <- list(kk[["x"]], kk[["y"]])

          # compute 2D density at each observation's coordinates
          kx <- cut(data$x, kk$x, labels = FALSE, include.lowest = TRUE)
          ky <- cut(data$y, kk$y, labels = FALSE, include.lowest = TRUE)
          kz <- sapply(seq_along(kx), function(i) kk$z[kx[i], ky[i]])
        } else {
          if (nrow(data) > 1L) {
            message("Density not computed, too few distinct values in 'x' and/or 'y'")
          }
          kz <- rep_len(1, nrow(data))
        }

        # we construct one logical vector by adding observations/label to be kept
        # we may have a list of 1, 2, or 4 logical vectors
        keep <- logical(nrow(data))
        for (i in seq_along(selectors)) {
          if (keep.fraction[i] == 1) {
            keep[ selectors[[i]] ] <- TRUE
          } else if (keep.fraction[i] != 0 && length(selectors[[i]]) >= 2L) {
            if (keep.sparse) {
              keep[ selectors[[i]] ] <-
                kz[ selectors[[i]] ] < stats::quantile(kz[ selectors[[i]] ],
                                                       keep.fraction[i], names = FALSE)
            } else {
              keep[ selectors[[i]] ] <-
                kz[ selectors[[i]] ] >= stats::quantile(kz[ selectors[[i]] ],
                                                        1 - keep.fraction[i], names = FALSE)
            }
          }
        }
        keep <- (keep | keep.these) & !exclude.these

        if (invert.selection){
          keep <- !keep
        }

        if (return.density) {
          data[["keep.obs"]] <- keep
          data[["density"]] <- kz
        }

        data[keep, ]

      },
    required_aes = c("x", "y")
  )

#' Nudge labels away from a central point
#'
#' \code{position_nudge_center()} is generally useful for adjusting the position of
#' labels or text, both on a discrete or continuous scale. In contrast to
#' \code{\link[ggplot2]{position_nudge}}, \code{position_nudge_center()} returns in \code{data} both
#' the original coordinates and the nudged coordinates.
#'
#' This position function is backwards compatible with \code{\link[ggplot2]{position_nudge}}
#' but extends it by adding support for nudging that varies across the plotting
#' region, either in opposite directions or radially from a virtual \emph{center
#' point}.
#'
#' @family position adjustments
#'
#' @param x,y Amount of vertical and horizontal distance to move. A numeric
#'   vector, that is recycled if shorter than the number of rows in \code{data}.
#' @param center_x,center_y The coordinates of the virtual origin out from which
#'   nudging radiates or splits in opposite directions. A numeric vector of
#'   length 1 or of the same length as rows there are in \code{data}, or a
#'   function returning either of these vectors computed from the variables in
#'   data mapped to \code{x} or \code{y}, respectively.
#' @param direction One of \code{"none"}, \code{"radial"}, or \code{"split"}. A
#'   value of \code{"none"} replicates the behavior of
#'   \code{\link[ggplot2]{position_nudge}}. Which of these three values is the
#'   default depends on the values passed to the other parameters.
#' @param obey_grouping A logical flag indicating whether to obey or not
#'   groupings of the observations. By default, grouping is obeyed when both of
#'   the variables mapped to \emph{x} and \emph{y} are continuous numeric and
#'   ignored otherwise.
#' @param kept.origin One of \code{"original"} or \code{"none"}.
#'
#' @details Positive values as arguments to \code{x} and \code{y} are added to
#'   the original position along either axis. If no arguments are passed to
#'   \code{center_x}, \code{center_y} or \code{direction}, the nudging is
#'   applied as is, as is the case if \code{direction = "none"}. If
#'   non-\code{NULL} arguments are passed to both \code{center_x} and
#'   \code{center_y}, \code{direction = "radial"} is assumed. In this case, if
#'   \code{x} and/or \code{y} positive nudging is applied radially outwards from
#'   the center, while if negative, inwards towards the center. When a
#'   non-\code{NULL} argument is passed only to one of \code{center_x} or
#'   \code{center_y}, \code{direction = "split"} is assumed. In this case when
#'   the initial location of the point is to the left of \code{center_x},
#'   \code{-x} is used instead of \code{x} for nudging, and when the initial
#'   location of the point is to the below of \code{center_y}, \code{-y} is used
#'   instead of \code{y} for nudging. If non-\code{NULL} arguments are passed to
#'   both \code{center_x} and \code{center_y}, and \code{direction} is passed
#'   \code{"split"} as argument, then the split as described above is applied to
#'   both to \emph{x} and \emph{y} coordinates.
#'
#' @note Some situations are handled as special cases. When \code{direction =
#'   "split"} or \code{direction = "radial"}, observations at exactly the _center_
#'   are nudged using \code{x} and \code{y} unchanged. When\code{direction = "split"},
#'   and
#'   both \code{center_x} and \code{center_y} have been supplied, segments are drawn at
#'   eight different possible angles. When segments are exactly horizontal or
#'   vertical they would be shorter than when drawn at the other four angles, in
#'   which case \code{x} or \code{y} are adjusted to ensure these segments are of the same
#'   lengths as those at other angles.
#'
#'   This position is most useful when labelling points forming a cloud or
#'   grouped along vertical or horizontal lines or "divides".
#'
#' @seealso [ggplot2::position_nudge()], [ggrepel::position_nudge_repel()].
#'
#' @return A \code{"Position"} object.
#'
#' @export
#'
#' @examples
#' df <- data.frame(
#'   x = c(1,3,2,5,4,2.5),
#'   y = c("abc","cd","d","c","bcd","a")
#' )
#'
#' # Plain nudging, same as with ggplot2::position_nudge()
#'
#' ggplot(df, aes(x, y, label = y)) +
#'   geom_point() +
#'   geom_text_s(hjust = "left", vjust = "bottom",
#'               position = position_nudge(x = 0.2, y = 0.2))
#'
#' ggplot(df, aes(x, y, label = y)) +
#'   geom_point() +
#'   geom_text_s(add.segments = FALSE,
#'               position = position_nudge_center(x = 0.2, y = 0.2)
#'   )
#'
#' # "split" nudging
#'
#' ggplot(df, aes(x, y)) +
#'   geom_point() +
#'   geom_text_s(aes(label = y),
#'               add.segments = FALSE,
#'               position = position_nudge_center(x = 0.2,
#'                                                y = 0.2,
#'                                                direction = "split"))
#'
#' ggplot(df, aes(x, y)) +
#'   geom_point() +
#'   geom_text_s(aes(label = y),
#'               position = position_nudge_center(x = 0.4,
#'                                                direction = "split"))
#'
#' ggplot(df, aes(x, y)) +
#'   geom_point() +
#'   geom_text_s(aes(label = y),
#'               position = position_nudge_center(y = 0.2,
#'                                                direction = "split"))
#'
#' ggplot(df, aes(x, y)) +
#'   geom_point() +
#'   geom_text_s(aes(label = y),
#'               position = position_nudge_center(x = 0.2,
#'                                                y = 0.3,
#'                                                center_y = 2,
#'                                                center_x = 1.5,
#'                                                direction = "split"))
#'
#' ggplot(df, aes(x, y)) +
#'   geom_point() +
#'   geom_text_s(aes(label = y),
#'               position = position_nudge_center(x = 0.06,
#'                                                y = 0.08,
#'                                                center_y = 2))
#'
#' ggplot(df, aes(x, y)) +
#'   geom_point() +
#'   geom_text_s(aes(label = y),
#'               position = position_nudge_center(x = 0.1,
#'                                                center_x = 2.51))
#'
#' ggplot(df, aes(x, y)) +
#'   geom_point() +
#'   geom_text_s(aes(label = y),
#'               position = position_nudge_center(x = 0.06,
#'                                                y = 0.08,
#'                                                center_x = median,
#'                                                center_y = median,
#'                                                direction = "split"))
#'
#' # "Radial" nudging
#'
#' ggplot(df, aes(x, y)) +
#'   geom_point() +
#'   geom_text_s(aes(label = y),
#'               position = position_nudge_center(x = 0.1,
#'                                                y = 0.2,
#'                                                direction = "radial"))
#'
#' ggplot(df, aes(x, y)) +
#'   geom_point() +
#'   geom_text_s(aes(label = y),
#'               position = position_nudge_center(x = -0.1,
#'                                                y = -0.1,
#'                                                direction = "radial"))
#'
#' df <- data.frame(
#'   x = -10:10,
#'   z = (-10:10)^2,
#'   y = letters[1:21],
#'   group = rep(c("a", "b"), rep(c(11, 10)))
#' )
#'
#' ggplot(df, aes(x, z)) +
#'   geom_point() +
#'   geom_line() +
#'   geom_text_s(aes(label = y),
#'               position = position_nudge_center(x = 0.9,
#'                                                y = 2.7,
#'                                                center_x = mean,
#'                                                center_y = max))
#'
#' ggplot(df, aes(x, z, color = group)) +
#'   geom_point() +
#'   geom_line(color = "black", linetype = "dotted") +
#'   geom_text_s(aes(label = y),
#'               position = position_nudge_center(x = -1.2,
#'                                                y = -3,
#'                                                center_x = 0,
#'                                                center_y = "above_max"))
#'
#' ggplot(df, aes(x, z, color = group)) +
#'   geom_point() +
#'   geom_line(color = "black", linetype = "dotted") +
#'   geom_text(aes(label = y),
#'             vjust = "inward", hjust = "inward",
#'             position = position_nudge_center(x = -0.9,
#'                                              y = -2.7,
#'                                              center_x = mean,
#'                                              center_y = max,
#'                                              obey_grouping = FALSE))
#'
position_nudge_center <-
  function(x = 0,
           y = 0,
           center_x = NULL,
           center_y = NULL,
           direction = NULL,
           obey_grouping = NULL,
           kept.origin = c("original", "none")) {

    kept.origin <- rlang::arg_match(kept.origin)

    if (is.null(direction)) {
      # Set default for 'direction' based on other arguments
      if (is.null(center_x) && is.null(center_y)) {
        direction <- "none"
      } else if (xor(is.null(center_x), is.null(center_y))) {
        direction <- "split"
      } else {
        direction <- "radial"
      }
    }

    if (direction != "none") {
      # Set center if is missing and direction requires it
      if (is.null(center_x)) {
        center_x <- mean
      }
      if (is.null(center_y)) {
        center_y <- mean
      }
    }

    if (is.null(obey_grouping)) {
      # default needs to be set in panel_function when we have access to data
      obey_grouping <- NA
    }

    if (lubridate::is.duration(x)) {
      x <- as.numeric(x)
    }
    if (lubridate::is.duration(y)) {
      y <- as.numeric(y)
    }
    # this works as long as center coords and mapped variable are of the same class
    # ggplot2's behaviour has been in the past and seems to be again to expect
    # numeric seconds for POSIXct and numeric days for Date time shifts
    if (lubridate::is.instant(center_x)) {
      x <- as.numeric(center_x)
    }
    if (lubridate::is.instant(center_y)) {
      y <- as.numeric(center_y)
    }

    ggplot2::ggproto(NULL, PositionNudgeCenter,
                     x = x,
                     y = y,
                     center_x = center_x,
                     center_y = center_y,
                     kept.origin = kept.origin,
                     direction = direction,
                     obey_grouping = obey_grouping
    )
  }

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
PositionNudgeCenter <-
  ggplot2::ggproto(
    "PositionNudgeCenter",
    ggplot2::Position,
    x = 0,
    y = 0,
    center_x = mean,
    center_y = mean,
    direction = "none",
    obey_grouping = NA,

    setup_params = function(self, data) {

      list(x = self$x,
           y = self$y,
           x.reorder = !is.null(self$x) && length(self$x) > 1 && length(self$x) < nrow(data),
           y.reorder = !is.null(self$y) && length(self$y) > 1 && length(self$y) < nrow(data),
           center_x = self$center_x,
           center_y = self$center_y,
           kept.origin = self$kept.origin,
           direction = self$direction,
           obey_grouping = self$obey_grouping)
    },

    compute_panel = function(self, data, params, scales) {

      if (length(params$x) > nrow(data)) {
        warning("Argument 'x' longer than data: some values dropped!")
      }
      if (length(params$y) > nrow(data)) {
        warning("Argument 'y' longer than data: some values dropped!")
      }
      x_orig <- data$x
      y_orig <- data$y
      # we handle grouping by ourselves
      if (is.na(params$obey_grouping)) {
        if (inherits(data$x, "ggplot2_mapped_discrete") ||
            inherits(data$y, "ggplot2_mapped_discrete") ||
            inherits(data$x, "mapped_discrete") ||
            inherits(data$y, "mapped_discrete") ||
            params$direction == "none") {
          # we ignore grouping as position_nudge() does
          params$obey_grouping <- FALSE
        } else {
          if (length(params$x) >= nrow(data)) {
            params$obey_grouping <- FALSE
          } else {
            # we respect groups
            params$obey_grouping <- TRUE
          }
        }
      }

      if (params$obey_grouping) {
        # one group at a time
        groups <- unique(data$group)
      } else {
        # all at once
        groups <- 1
      }
      # Based on the value of 'direction' we adjust the nudge for each point
      x_nudge <- y_nudge <- numeric(nrow(data))
      for (group in groups) {
        if (params$obey_grouping) {
          # selector for rows in current group
          in.grp <- data$group == group
        } else {
          # selector for all rows
          in.grp <- rep(TRUE, nrow(data))
        }
        # compute focal centre by group
        if (is.function(params$center_x)) {
          x_ctr <- params$center_x(as.numeric(data[in.grp, "x"]))
        } else if (is.numeric(params$center_x)) {
          x_ctr <- params$center_x[1]
        } else {
          x_ctr <- -Inf # ensure all observations are to the right
        }
        if (is.function(params$center_y)) {
          y_ctr <- params$center_y(as.numeric(data[in.grp, "y"]))
        } else if (is.numeric(params$center_y)) {
          y_ctr <- params$center_y[1]
        } else {
          y_ctr <- -Inf # ensure all observations are above
        }

        if (params$direction == "radial") {
          # compute x and y nudge for each point
          x_dist <- as.numeric(data[in.grp, "x"]) - x_ctr
          y_dist <- as.numeric(data[in.grp, "y"]) - y_ctr
          # if both x and y position displacements are 0, we force a shift
          overlapping <- (abs(x_dist) < 1e-5) & (abs(y_dist) < 1e-5)
          x_dist <- ifelse(overlapping, 1e-5, x_dist)
          y_dist <- ifelse(overlapping, 1e-5, y_dist)
          angle <- atan2(y_dist, x_dist) + pi / 2
          if (params$x == 0) {
            angle <- ifelse(cos(angle) == 0, 0, angle)
          }
          if (params$y == 0) {
            angle <- ifelse(sin(angle) == 0, pi / 2, angle)
          }

          x_nudge[in.grp] <- params$x * sin(angle)
          y_nudge[in.grp] <- -params$y * cos(angle)
        } else if (params$direction == "split") {
          if (length(params$x) == 1L && length(params$y) == 1L) {
            # ensure horizontal and vertical segments have same length as others
            segment_length <- sqrt(params$x^2 + params$y^2)
            xx <- rep(params$x, nrow(data[in.grp, ]))
            xx <- ifelse(data[in.grp, "y"] == y_ctr,
                         segment_length * sign(xx),
                         xx)
            yy <- rep(params$y, nrow(data[in.grp, ]))
            yy <- ifelse(data[in.grp, "x"] == x_ctr,
                         segment_length * sign(yy),
                         yy)
          }
          x_nudge[in.grp] <- xx * sign(as.numeric(data[in.grp, "x"]) - x_ctr)
          y_nudge[in.grp] <- yy * sign(as.numeric(data[in.grp, "y"]) - y_ctr)
          # if both x and y position displacements are 0, we force a shift
          overlapping <- (x_nudge[in.grp] == 0) & (y_nudge[in.grp] == 0)
          x_nudge[in.grp] <- ifelse(overlapping, xx, x_nudge[in.grp])
          y_nudge[in.grp] <- ifelse(overlapping, yy, y_nudge[in.grp])
        } else {
          if (params$direction != "none") {
            warning("Ignoring unrecognized direction \"",
                    params$direction, "\".")
          }
          # behaviour similar to position_nudge() except for handling of vectors of nudges
          # and possibly respecting groups for alternating nudges
          if (params$x.reorder) {
            x_nudge[in.grp] <- rep_len(params$x, sum(in.grp))[order(order(data$x[in.grp]))]
          } else {
            x_nudge[in.grp] <- rep_len(params$x, sum(in.grp))
          }
          if (params$y.reorder) {
            y_nudge[in.grp] <- rep_len(params$y, sum(in.grp))[order(order(data$y[in.grp]))]
          } else {
            y_nudge[in.grp] <- rep_len(params$y, sum(in.grp))
          }
        }
      }
      # transform only the dimensions for which new coordinates exist
      if (any(x_nudge != 0)) {
        if (any(y_nudge != 0)) {
          data <- transform_position(data, function(x) x + x_nudge, function(y) y + y_nudge)
        } else {
          data <- transform_position(data, function(x) x + x_nudge, NULL)
        }
      } else if (any(y_nudge != 0)) {
        data <- transform_position(data, NULL, function(y) y + y_nudge)
      }
      # add original position
      if (params$kept.origin == "original") {
        data$x_orig <- x_orig
        data$y_orig <- y_orig
      }

      data
    }
  )

#' @rdname position_nudge_center
#'
#' @export
#'
position_nudge_centre <- position_nudge_center

#' Nudge points a fixed distance
#'
#' The function \code{position_nudge_keep()} has an additional parameters
#' compared to \code{\link[ggplot2]{position_nudge}}, \code{obey_grouping} and
#' by default the same behaviour when the values passed as arguments to \code{x}
#' and \code{y} have length one.
#'
#' @details When \code{x} or \code{y} have length > 1, they are treated
#'   specially. If the lengths is the same as there are rows in data, the nudges
#'   are applied in the order of the rows in data. When they are shorter, they
#'   are recycled and applied to the data values after ordering. This makes it
#'   possible to have alternating mudging right and left or up and down. If
#'   \code{obey_grouping = TRUE} is passed in the call, the alternation will
#'   take place within groups.
#'
#'   As other position functions from package 'ggpp', \code{position_nudge_keep()}
#'   by default renames and keeps the original positions of the observations in
#'   \code{data} making it possible to draw connecting segments or conencting
#'   arrows.
#'
#' @family position adjustments
#'
#' @param x,y Amount of vertical and horizontal distance to move. A numeric
#'   vector of length 1, or of the same length as rows there are in \code{data},
#'   with nudge values in data rows order.
#' @param obey_grouping A logical flag indicating whether to obey or not
#'   groupings of the observations. By default, grouping is obeyed when both of
#'   the variables mapped to \emph{x} and \emph{y} are continuous numeric and
#'   ignored otherwise.
#' @param kept.origin One of \code{"original"} or \code{"none"}.
#'
#' @note Irrespective of the action, the ordering of rows in \code{data} is
#'   preserved.
#'
#' @return A \code{"Position"} object.
#'
#' @export
#'
#' @examples
#' df <- data.frame(
#'   x = c(1,3,2,5,4,2.5),
#'   y = c("abc","cd","d","c","bcd","a")
#' )
#'
#' # Plain nudging, same as with ggplot2::position_nudge()
#'
#' ggplot(df, aes(x, y, label = y)) +
#'   geom_point() +
#'   geom_text_s(hjust = "left", vjust = "bottom",
#'               position = position_nudge_keep(x = 0.2, y = 0.2))
#'
#' # alternating nudging
#' ggplot(df, aes(x, y, label = y)) +
#'   geom_point() +
#'   geom_text_s(position = position_nudge_keep(x = c(0.2, -0.2)))
#'
#' # direct nudging
#' ggplot(df, aes(x, y, label = y)) +
#'   geom_point() +
#'   geom_text_s(position = position_nudge_keep(x = rep_len(c(0.2, -0.2), 6)))
#'
position_nudge_keep <-
  function(x = 0,
           y = 0,
           obey_grouping = NULL,
           kept.origin = c("original", "none")) {

    kept.origin <- rlang::arg_match(kept.origin)

    if (is.null(obey_grouping)) {
      # default needs to be set in panel_function when we have access to data
      obey_grouping <- NA
    }

    if (lubridate::is.duration(x)) {
      x <- as.numeric(x)
    }
    if (lubridate::is.duration(y)) {
      y <- as.numeric(y)
    }

    # we reuse the code from the position defined above
    ggplot2::ggproto(NULL, PositionNudgeCenter,
                     x = x,
                     y = y,
                     center_x = NULL,
                     center_y = NULL,
                     kept.origin = kept.origin,
                     direction = "none",
                     obey_grouping = obey_grouping
    )
  }

#' Number of observations in a plot panel
#'
#' \code{stat_panel_counts()} counts the number of observations in each panel.
#' \code{stat_group_counts()} counts the number of observations in each group.
#' By default they add one or more text labels to the top right corner of each
#' panel. Grouping is ignored by \code{stat_panel_counts()}. If no grouping
#' exists, the two statistics behave similarly.
#'
#' @param mapping The aesthetic mapping, usually constructed with
#'   \code{\link[ggplot2]{aes}} or \code{\link[ggplot2]{aes_}}. Only needs to be
#'   set at the layer level if you are overriding the plot defaults.
#' @param data A layer specific dataset. Rarely used, as you will not want to
#'   override the plot defaults.
#' @param geom The geometric object to use display the data
#' @param position The position adjustment to use on this layer
#' @param show.legend logical. Should this layer be included in the legends?
#'   \code{NA}, the default, includes it if any aesthetics are mapped.
#'   \code{FALSE} never includes, and \code{TRUE} always includes.
#' @param inherit.aes If \code{FALSE}, overrides the default aesthetics, rather
#'   than combining with them. This is most useful for helper functions that
#'   define both data and aesthetics and should not inherit behaviour from the
#'   default plot specification, e.g., \code{\link[ggplot2]{borders}}.
#' @param ... other arguments passed on to \code{\link[ggplot2]{layer}}. This
#'   can include aesthetics whose values you want to set, not map. See
#'   \code{\link[ggplot2]{layer}} for more details.
#' @param na.rm	a logical indicating whether \code{NA} values should be stripped
#'   before the computation proceeds.
#' @param label.x,label.y \code{numeric} Coordinates (in npc units) to be used
#'   for absolute positioning of the labels.
#'
#' @details These statistics can be used to automatically count observations in
#'   each panel of a plot, and by default add these counts as text labels. These
#'   statistics, unlike \code{stat_quadrant_counts()} requires only one of
#'   \emph{x} or \emph{y} aesthetics and can be used together with statistics
#'   that have the same requirement, like \code{stat_density()}.
#'
#'   The default position of the label is in the top right corner. When using
#'   facets even with free limits for \emph{x} and \emph{y} axes, the location
#'   of the labels is consistent across panels. This is achieved by use of
#'   \code{geom = "text_npc"} or \code{geom = "label_npc"}. To pass the
#'   positions in native data units to \code{label.x} and \code{label.y}, pass
#'   also explicitly \code{geom = "text"}, \code{geom = "label"} or some other
#'   geometry that use the \emph{x} and/or \emph{y} aesthetics. A vector with
#'   the same length as the number of panels in the figure can be used if
#'   needed.
#'
#' @section Computed variables: Data frame with one or more rows, one for each
#'   group of observations for which counts are counted in \code{data}. \describe{
#'   \item{x,npcx}{x value of label position in data- or npc units, respectively}
#'   \item{y,npcy}{y value of label position in data- or npc units, respectively}
#'   \item{count}{number of  observations as an integer}
#'   \item{count.label}{number of observations as character}}
#'
#'   As shown in one example below \code{\link[gginnards]{geom_debug}} can be
#'   used to print the computed values returned by any statistic. The output
#'   shown includes also values mapped to aesthetics, like \code{label} in the
#'   example. \code{x} and \code{y} are included in the output only if mapped.
#'
#' @note  If a factor is mapped to \code{x} or to \code{y} aesthetics each level
#'   of the factor constitutes a group, in this case the default positioning and
#'   geom using NPC pseudo aesthetics will have to be overriden by passing
#'   \code{geom = "text"} and data coordinates used. The default for factors
#'   may change in the future.
#'
#' @return A plot layer instance. Using as output \code{data} the counts of
#'   observations in each plot panel or per group in each plot panel.
#'
#' @family Functions for quadrant and volcano plots
#'
#' @export
#'
#' @examples
#'
#' # generate artificial data with numeric x and y
#' set.seed(67821)
#' x <- 1:100
#' y <- rnorm(length(x), mean = 10)
#' group <- factor(rep(c("A", "B"), times = 50))
#' my.data <- data.frame(x, y, group)
#'
#' # using automatically generated text labels
#'
#' ggplot(my.data, aes(x, y)) +
#'   geom_point() +
#'   stat_panel_counts()
#'
#' ggplot(my.data, aes(x, y, colour = group)) +
#'   geom_point() +
#'   stat_panel_counts()
#'
#' ggplot(my.data, aes(x, y, colour = group)) +
#'   geom_point() +
#'   stat_group_counts()
#'
#' ggplot(my.data, aes(x, y, colour = group)) +
#'   geom_point() +
#'   stat_group_counts(label.x = "left", hstep = 0.06, vstep = 0)
#'
#' ggplot(my.data, aes(x, y, colour = group)) +
#'   geom_point() +
#'   stat_group_counts(aes(label = after_stat(pc.label)))
#'
#' ggplot(my.data, aes(x, y, colour = group)) +
#'   geom_point() +
#'   stat_group_counts(aes(label = after_stat(pc.label)), digits = 3)
#'
#' ggplot(my.data, aes(x, y, colour = group)) +
#'   geom_point() +
#'   stat_group_counts(aes(label = after_stat(fr.label)))
#'
#' ggplot(my.data, aes(x, y, colour = group)) +
#'   geom_point() +
#'   stat_group_counts(aes(label = after_stat(dec.label)))
#'
#' # one of x or y can be a factor
#' # label.x or label.y along the factor can be set to "factor" together
#' # with the use of geom_text()
#'
#' ggplot(mpg,
#'        aes(factor(cyl), hwy)) +
#'   stat_boxplot() +
#'   stat_group_counts(geom = "text",
#'                     label.y = 10,
#'                     label.x = "factor") +
#'   stat_panel_counts()
#'
#' # Numeric values can be used to build labels with alternative formats
#' # Here with sprintf(), but paste() and format() also work.
#'
#' ggplot(my.data, aes(x, y)) +
#'  geom_point() +
#'  stat_panel_counts(aes(label = sprintf("%i observations",
#'                                        after_stat(count)))) +
#'  scale_y_continuous(expand = expansion(mult = c(0.05, 0.12)))
#'
#' ggplot(mpg,
#'        aes(factor(cyl), hwy)) +
#'   stat_boxplot() +
#'   stat_group_counts(geom = "text",
#'                     aes(label = sprintf("(%i)", after_stat(count))),
#'                     label.y = 10,
#'                     label.x = "factor")
#'
#' ggplot(mpg,
#'        aes(factor(cyl), hwy)) +
#'   stat_boxplot() +
#'   stat_group_counts(aes(label = sprintf("n[%i]~`=`~%i",
#'                                         after_stat(x), after_stat(count))),
#'                     parse = TRUE,
#'                     geom = "text",
#'                     label.y = 10,
#'                     label.x = "factor") +
#'   stat_panel_counts(aes(label = sprintf("sum(n[i])~`=`~%i",
#'                                         after_stat(count))),
#'                     parse = TRUE)
#'
#' # label position
#'
#' ggplot(my.data, aes(y)) +
#'   stat_panel_counts(label.x = "left") +
#'   stat_density(alpha = 0.5)
#'
#' ggplot(my.data, aes(y, colour = group)) +
#'   stat_group_counts(label.y = "top") +
#'   stat_density(aes(fill = group), alpha = 0.3)
#'
#' # The numeric value can be used as a label as is
#'
#' ggplot(mpg,
#'        aes(factor(cyl), hwy)) +
#'   stat_boxplot() +
#'   stat_group_counts(geom = "text",
#'                     aes(label = after_stat(count)),
#'                     label.x = "factor",
#'                     label.y = 10) +
#'   annotate(geom = "text", x = 0.55, y = 10, label = "n[i]~`=`", parse = TRUE)
#'
#' # We use geom_debug_group() to see the computed values
#'
#' gginnards.installed <- requireNamespace("gginnards", quietly = TRUE)
#' if (gginnards.installed) {
#'   library(gginnards)
#'
#'   ggplot(my.data, aes(x, y)) +
#'     geom_point() +
#'     stat_panel_counts(geom = "debug_group")
#' }
#'
#' if (gginnards.installed) {
#'   ggplot(my.data, aes(x, y, colour = group)) +
#'     geom_point() +
#'     stat_group_counts(geom = "debug_group")
#' }
#'
stat_panel_counts <- function(mapping = NULL,
                              data = NULL,
                              geom = "text_npc",
                              position = "identity",
                              label.x = "right",
                              label.y = "top",
                              na.rm = FALSE,
                              show.legend = FALSE,
                              inherit.aes = TRUE,
                              ...) {

  stopifnot(is.null(label.x) || is.numeric(label.x) || is.character(label.x))
  stopifnot(is.null(label.y) || is.numeric(label.y) || is.character(label.y))

  ggplot2::layer(
    stat = StatPanelCounts,
    data = data,
    mapping = mapping,
    geom = geom,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(na.rm = na.rm,
                  label.x = label.x,
                  label.y = label.y,
                  npc.used = grepl("_npc", geom),
                  ...)
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
StatPanelCounts <-
  ggplot2::ggproto("StatPanelCounts", ggplot2::Stat,

                   compute_panel = function(data,
                                            scales,
                                            label.x,
                                            label.y,
                                            npc.used) {

                     force(data)
                     # total count (works because NAs have been already removed)
                     z <- tibble::tibble(count = nrow(data))
                     # label position
                     if (is.character(label.x)) {
                       if (npc.used) {
                         margin.npc <- 0.05
                       } else {
                         # margin set by scale
                         margin.npc <- 0
                       }
                       label.x <- compute_npcx(x = label.x)
                       if (!npc.used) {
                         if ("x" %in% colnames(data)) {
                           x.expanse <- abs(diff(range(data$x)))
                           x.min <- min(data$x)
                           label.x <- label.x * x.expanse + x.min
                         } else {
                           if (data$PANEL[1] == 1L) { # show only once
                             message("No 'x' mapping; 'label.x' requires a numeric argument in data units")
                           }
                           label.x <- NA_real_
                         }
                       }
                     }
                     if (is.character(label.y)) {
                       if (npc.used) {
                         margin.npc <- 0.05
                       } else {
                         # margin set by scale
                         margin.npc <- 0
                       }
                       label.y <- compute_npcy(y = label.y)
                       if (!npc.used) {
                         if ("y" %in% colnames(data)) {
                           y.expanse <- abs(diff(range(data$y)))
                           y.min <- min(data$y)
                           label.y <- label.y * y.expanse + y.min
                         } else {
                           if (data$PANEL[1] == 1L) { # show only once
                             message("No 'y' mapping; 'label.y' requires a numeric argument in data units")
                           }
                           label.y <- NA_real_
                         }
                       }
                     }

                     # returned data frame must contain a "group" column
                     if (! "group" %in% colnames(z)) {
                       z[["group"]] <- -1L
                     }

                     if (npc.used) {
                       z$npcx <- label.x
                       z$x <- NA_real_
                       z$npcy <- label.y
                       z$y <- NA_real_
                     } else {
                       z$x <- label.x
                       z$npcx <- NA_real_
                       z$y <- label.y
                       z$npcy <- NA_real_
                     }
                     z$count.label <- sprintf("n=%i", z$count)
                     z
                   },

                   default_aes =
                     ggplot2::aes(npcx = ggplot2::after_stat(npcx),
                                  npcy = ggplot2::after_stat(npcy),
                                  label = ggplot2::after_stat(count.label),
                                  hjust = "inward",
                                  vjust = "inward"),
                   required_aes = c("x|y")
  )


#' @rdname stat_panel_counts
#'
#' @param hstep,vstep numeric in npc units, the horizontal and vertical step
#'   used between labels for different groups.
#' @param digits integer Number of digits for fraction and percent labels.
#'
#' @export
#'
stat_group_counts <- function(mapping = NULL,
                              data = NULL,
                              geom = "text_npc",
                              position = "identity",
                              label.x = "right",
                              label.y = "top",
                              hstep = 0,
                              vstep = NULL,
                              digits = 2,
                              na.rm = FALSE,
                              show.legend = FALSE,
                              inherit.aes = TRUE,
                              ...) {

  stopifnot(is.null(label.x) || is.numeric(label.x) || is.character(label.x))
  stopifnot(is.null(label.y) || is.numeric(label.y) || is.character(label.y))

  ggplot2::layer(
    stat = StatGroupCounts,
    data = data,
    mapping = mapping,
    geom = geom,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(na.rm = na.rm,
                  label.x = label.x,
                  label.y = label.y,
                  digits = digits,
                  hstep = hstep,
                  vstep = ifelse(is.null(vstep),
                                 ifelse(grepl("label", geom),
                                        0.10,
                                        0.05),
                                 vstep),
                  npc.used = grepl("_npc", geom),
                  ...)
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
StatGroupCounts <-
  ggplot2::ggproto("StatGroupCounts", ggplot2::Stat,

                   compute_panel = function(data,
                                            params,
                                            scales,
                                            label.x,
                                            label.y,
                                            digits = 2,
                                            vstep,
                                            hstep,
                                            npc.used) {

                     force(data)

                     # Build group labels
                     if (exists("grp.label", data)) {
                       if (length(unique(data[["grp.label"]])) > 1L) {
                         warning("Non-unique value in 'data$grp.label' using group index ", data[["group"]][1], " as label.")
                         grp.label <- as.character(unique(data[["group"]]))
                       } else {
                         grp.label <- unique(data[["grp.label"]])
                       }
                     } else {
                       # if nothing mapped to grp.label we use group index as label
                       grp.label <- as.character(unique(data[["group"]]))
                     }

                     # validate label positions
                     group.idx <- abs(unique(data$group))
                     if (length(label.x) >= length(group.idx)) {
                       label.x <- label.x[1:length(group.idx)]
                     } else if (length(label.x) > 0) {
                       label.x <- label.x[1]
                     }
                     if (length(label.y) >= length(group.idx)) {
                       label.y <- label.y[1:length(group.idx)]
                     } else if (length(label.y) > 0) {
                       label.y <- label.y[1]
                     }

                     # Compute number of observations
                     # (works because NAs have been already removed)
                     z <- data.frame()

                     for (g in unique(data$group)) {
                       temp <- cbind(data[data$group == g, ][ 1, ], count = sum(data$group == g))
                       z <- rbind(z, temp)
                     }

                     z$total <- nrow(data)

                     z$count.label <- sprintf("n=%i", z$count)
                     z$pc.label <- sprintf("p=%.*f%%",
                                           digits - 2,
                                           z$count / z$total * 100)
                     z$dec.label <- sprintf("f=%.*f",
                                            digits,
                                            z$count / z$total)
                     z$fr.label <- sprintf("%i / %i",
                                           z$count, z$total)

                     # Compute label positions
                     if (is.character(label.x)) {
                       if (label.x[1] == "factor") {
                         label.x <- z$group
                       } else {
                         if (npc.used) {
                           margin.npc <- 0.05
                         } else {
                           # margin set by scale
                           margin.npc <- 0
                         }
                         label.x <- compute_npcx(x = label.x, group = group.idx, h.step = hstep,
                                                 margin.npc = margin.npc)
                         if (!npc.used) {
                           if ("x" %in% colnames(data)) {
                             x.expanse <- abs(diff(range(data$x)))
                             x.min <- min(data$x)
                             label.x <- label.x * x.expanse + x.min
                           } else {
                             if (data$PANEL[1] == 1L && group.idx == 1L) { # show only once
                               message("No 'x' mapping; 'label.x' requires a numeric argument in data units")
                             }
                             label.x <- NA_real_
                           }
                         }
                       }
                     }

                     if (is.character(label.y)) {
                       if (label.y[1] == "factor") {
                         label.x <- z$group
                       } else {
                         if (npc.used) {
                           margin.npc <- 0.05
                         } else {
                           # margin set by scale
                           margin.npc <- 0
                         }
                         label.y <- compute_npcy(y = label.y, group = group.idx, v.step = vstep,
                                                 margin.npc = margin.npc)
                         if (!npc.used) {
                           if ("y" %in% colnames(data)) {
                             y.expanse <- abs(diff(range(data$y)))
                             y.min <- min(data$y)
                             label.y <- label.y * y.expanse + y.min
                           } else {
                             if (data$PANEL[1] == 1L && group.idx == 1L) { # show only once
                               message("No 'y' mapping; 'label.y' requires a numeric argument in data units")
                             }
                             label.y <- NA_real_
                           }
                         }
                       }
                     }

                     if (npc.used) {
                       z$npcx <- label.x
                       z$x <- NA_real_
                       z$npcy <- label.y
                       z$y <- NA_real_
                     } else {
                       z$x <- label.x
                       z$npcx <- NA_real_
                       z$y <- label.y
                       z$npcy <- NA_real_
                     }
                     z$count.label <- sprintf("n=%i", z$count)
                     z
                   },

                   default_aes =
                     ggplot2::aes(npcx = ggplot2::after_stat(npcx),
                                  npcy = ggplot2::after_stat(npcy),
                                  label = ggplot2::after_stat(count.label),
                                  hjust = "inward",
                                  vjust = "inward"),
                   required_aes = c("x|y")
  )

context("stat_panel_counts")

test_that("stat_panel_counts", {
  tst.df <- data.frame(
    x = c(2.2, -0.2, -0.2, -2.5, -0.6, -0.1),
    y = c(1.1, -0.6, -0.9, -1.6, 0.3, 1.6),
    group = c("A", "A", "A", "B", "B", "B")
  )
  vdiffr::expect_doppelganger("stat_panel_counts_xy",
                              ggplot(tst.df, aes(x, y)) +
                                stat_panel_counts() +
                                geom_point()
                              )

  vdiffr::expect_doppelganger("stat_panel_counts_x",
                              ggplot(tst.df, aes(x)) +
                                stat_panel_counts() +
                                geom_histogram(bins = 3)
                              )

  vdiffr::expect_doppelganger("stat_panel_counts_y",
                              ggplot(tst.df, aes(y)) +
                                stat_panel_counts() +
                                geom_histogram(bins = 3)
                              )

  p <- ggplot(tst.df, aes(x, y)) +
    stat_panel_counts() +
    geom_point()
  result <- layer_data(p)[, c("npcx", "npcy", "label", "count", "hjust", "vjust")]
  expected <- data.frame(
    npcx = 0.95,
    npcy = 0.95,
    label = "n=6",
    count = 6L,
    hjust = "inward",
    vjust = "inward")
  expect_identical(result, expected)

  p <- ggplot(tst.df, aes(x)) +
    stat_panel_counts()
  result <- layer_data(p)[, c("npcx", "npcy", "label", "count", "hjust", "vjust")]
  expect_identical(result, expected)

  p <- ggplot(tst.df, aes(y)) +
    stat_panel_counts()
  result <- layer_data(p)[, c("npcx", "npcy", "label", "count", "hjust", "vjust")]
  expect_identical(result, expected)

  expect_error(ggplot(tst.df, aes(x, y)) +
      stat_panel_counts(label.x = NA))

  expect_error(ggplot(tst.df, aes(x, y)) +
      stat_panel_counts(label.y = NA))

  })


test_that("stat_group_counts", {
  tst.df <- data.frame(
    x = c(2.2, -0.2, -0.2, -2.5, -0.6, -0.1),
    y = c(1.1, -0.6, -0.9, -1.6, 0.3, 1.6),
    group = c("A", "A", "A", "B", "B", "A")
  )
  vdiffr::expect_doppelganger("stat_group_counts_xy",
                              ggplot(tst.df, aes(x, y)) +
                                stat_group_counts() +
                                geom_point()
  )

  vdiffr::expect_doppelganger("stat_group_counts_xy_color",
                              ggplot(tst.df, aes(x, y, color = group)) +
                                stat_group_counts() +
                                geom_point()
  )

  vdiffr::expect_doppelganger("stat_group_counts_x",
                              ggplot(tst.df, aes(x)) +
                                stat_group_counts() +
                                geom_histogram(bins = 3)
  )

  vdiffr::expect_doppelganger("stat_group_counts_y",
                              ggplot(tst.df, aes(y)) +
                                stat_group_counts() +
                                geom_histogram(bins = 3)
  )

  p <- ggplot(tst.df, aes(x, y)) +
    stat_group_counts() +
    geom_point()
  result <- layer_data(p)[, c("npcx", "npcy", "label", "count", "total",
                              "count.label", "pc.label", "dec.label", "fr.label", "hjust", "vjust")]
  expected <- data.frame(
    npcx = 0.95,
    npcy = 0.95,
    label = "n=6",
    count = 6L,
    total = 6L,
    count.label = "n=6",
    pc.label = "p=100%",
    dec.label = "f=1.00",
    fr.label = "6 / 6",
    hjust = "inward",
    vjust = "inward")
  expect_identical(result, expected)

  p <- ggplot(tst.df, aes(x)) +
    stat_group_counts()
  result <- layer_data(p)[, c("npcx", "npcy", "label", "count", "total",
                              "count.label", "pc.label", "dec.label", "fr.label", "hjust", "vjust")]
  expect_identical(result, expected)

  p <- ggplot(tst.df, aes(y)) +
    stat_group_counts()
  result <- layer_data(p)[, c("npcx", "npcy", "label", "count", "total",
                              "count.label", "pc.label", "dec.label", "fr.label", "hjust", "vjust")]
  expect_identical(result, expected)

  p <- ggplot(tst.df, aes(x, y)) +
    stat_group_counts(digits = 3) +
    geom_point()
  result <- layer_data(p)[, c("npcx", "npcy", "label", "count", "total",
                              "count.label", "pc.label", "dec.label", "fr.label", "hjust", "vjust")]
  expected <- data.frame(
    npcx = 0.95,
    npcy = 0.95,
    label = "n=6",
    count = 6L,
    total = 6L,
    count.label = "n=6",
    pc.label = "p=100.0%",
    dec.label = "f=1.000",
    fr.label = "6 / 6",
    hjust = "inward",
    vjust = "inward")
  expect_identical(result, expected)

  p <- ggplot(tst.df, aes(x, y, group = group)) +
    stat_group_counts(digits = 3) +
    geom_point()
  result <- layer_data(p)[, c("npcx", "npcy", "label", "group", "count", "total",
                              "count.label", "pc.label", "dec.label", "fr.label", "hjust", "vjust")]
  expected <- data.frame(
    npcx = 0.95,
    npcy = c(0.95, 0.90),
    label = c("n=4", "n=2"),
    group = 1L:2L,
    count = c(4L, 2L),
    total = 6L,
    count.label =  c("n=4", "n=2"),
    pc.label = c("p=66.7%", "p=33.3%"),
    dec.label = c("f=0.667", "f=0.333"),
    fr.label = c("4 / 6", "2 / 6"),
    hjust = "inward",
    vjust = "inward")
  expect_equal(result, expected)

  expect_error(ggplot(tst.df, aes(x, y)) +
                 stat_group_counts(label.x = NA))

  expect_error(ggplot(tst.df, aes(x, y)) +
                 stat_group_counts(label.y = NA))

  expect_error(ggplot(tst.df, aes(x, y, color = group)) +
                 stat_group_counts(label.x = NA))

  expect_error(ggplot(tst.df, aes(x, y, color = group)) +
                 stat_group_counts(label.y = NA))

})

#' @title Linked Text
#' @include ggp2-margins.R utilities.R ggpp-legend-draw.R
#'
#' @description Linked text geometries are most useful for adding data labels to
#'   plots. `geom_text_s()` and `geom_label_s()` add text to the plot and for
#'   nudged positions link the original location to the nudged text with a
#'   segment or arrow.
#'
#' @details Geometries \code{geom_text_s()} and \code{geom_label_s()} have an
#'   interface similar to that of \code{\link[ggplot2]{geom_text}} and
#'   \code{\link[ggplot2]{geom_label}}, but support additional features.
#'   Similarly to \code{geom_text_repel()} and \code{geom_label_repel()} when
#'   used together with position functions defined in package 'ggpp' they draw a
#'   segment linking the label at a displaced position to the original position,
#'   usually a point corresponding to an observation to which the label refers.
#'   Another difference is that they allow control of to which graphical
#'   elements the mappings to colour and alpha aesthetics are applied.
#'   Differently to \code{geom_label()}, \code{geom_label_s()} obeys aesthetic
#'   mappings to \code{linewidth} and \code{linetype} applied to the line at the
#'   edge of the label box. These features are reflected in the plot key, except
#'   for the segment, assumed not to be used to display information
#'   only in coordination with other graphic elements.
#'
#'   In \code{geom_label_s()} the default \code{fill} is similar to
#'   \code{"white"} but with its \code{alpha} component set to 0.75. This
#'   differs from \code{"white"} used in \code{geom_label()}: the default fill
#'   is semitransparent with the intention that accidental occlusion of
#'   observations is obvious irrespective of the order in which layers are added
#'   to the plot.
#'
#'   Layer functions \code{geom_text_s()} and \code{geom_label_s()} use by
#'   default \code{\link{position_nudge_keep}} which is backwards compatible
#'   with \code{\link[ggplot2]{position_nudge}}. In contrast to
#'   \code{\link[ggplot2]{position_nudge}}, \code{\link{position_nudge_keep}}
#'   and all other position functions defined in packages 'ggpp' and 'ggrepel'
#'   keep the original coordinates, thus allowing the plotting of connecting
#'   segments and arrows.
#'
#'   Differently to \code{geom_text_repel()} and \code{geom_label_repel()},
#'   \code{geom_text_s()} and \code{geom_label_s()} do not make use of
#'   additional aesthetics for the segments or boxes, but instead allow the
#'   choice of which elements are targeted by the aesthetics and which are
#'   rendered in a default colour. In the grammar of graphics using the same
#'   aesthetic with multiple meanings is not allowed, thus, the approach used in
#'   the geometry layer functions from package 'ggpp' attempts to enforce this.
#'
#' @section Plot boundaries and clipping: Note that when you change the scale
#'   limits for \emph{x} and/or \emph{y} of a plot, text labels stay the same
#'   size, as determined by the \code{size} aesthetic, given in millimetres. The
#'   actual size as seen in the plotted output is decided during the rendering
#'   of the plot to a graphics device. Limits are expanded only to include the
#'   anchor point of the labels because the "width" and "height" of a text
#'   element are 0 (as seen by ggplot2). Text labels do have height and width,
#'   but in grid units, not data units.
#'
#' @section Alignment: You can modify text alignment with the \code{vjust} and
#'   \code{hjust} aesthetics. These can either be a number between 0
#'   (right/bottom) and 1 (top/left) or a character (\code{"left"},
#'   \code{"middle"}, \code{"right"}, \code{"bottom"}, \code{"center"},
#'   \code{"top"}). In addition, you can use special alignments for
#'   justification including \code{"position"}, \code{"inward"} and
#'   \code{"outward"}. Inward always aligns text towards the center of the
#'   plotting area, and outward aligns it away from the center of the plotting
#'   area. If tagged with \code{_mean} or \code{_median} (e.g.,
#'   \code{"outward_mean"}) the mean or median of the data in the panel along
#'   the corresponding axis is used as center. If the characters following the
#'   underscore represent a number (e.g., \code{"outward_10.5"}) the reference
#'   point will be this value in data units. Position justification is computed
#'   based on the direction of the displacement of the position of the label so
#'   that each individual text or label is justified outwards from its original
#'   position. The default justification is \code{"position"}.
#'
#'   If no position displacement is applied, or a position function defined in
#'   'ggplot2' is used, these geometries behave similarly to the corresponding
#'   ones from package 'ggplot2' with a default justification of \code{0.5} and
#'   no segment drawn.
#'
#' @section Differences from earlier versions:
#'   The user interface is for the most part stable starting from 'ggpp' (==
#'   0.5.7). In 'ggpp' (== 0.5.0) support for aesthetics related to segments was
#'   removed, and replaced by parameters and a new mechanism for targeting the
#'   usual \code{colour} and \code{alpha} aesthetics to text, border, and
#'   segment.
#'
#' @param mapping Set of aesthetic mappings created by
#'   \code{\link[ggplot2]{aes}}. If specified and with \code{inherit.aes = TRUE}
#'   (the default), it is combined with the default mapping at the top level of
#'   the plot. You only need to supply \code{mapping} if there isn't a mapping
#'   defined for the plot.
#' @param data A data frame. If specified, overrides the default data frame
#'   defined at the top level of the plot.
#' @param stat The statistical transformation to use on the data for this layer,
#'   as a string.
#' @param position Position adjustment, either as a string, or the result of a
#'   call to a position adjustment function.
#' @param parse If \code{TRUE}, the labels will be parsed into expressions and
#'   displayed as described in \code{?plotmath}.
#' @param na.rm If \code{FALSE} (the default), removes missing values with a
#'   warning.  If \code{TRUE} silently removes missing values.
#' @param show.legend logical. Should this layer be included in the legends?
#'   \code{NA}, the default, includes a legend if any aesthetics are mapped.
#'   \code{FALSE} never includes it, and \code{TRUE} always includes it.
#' @param inherit.aes If \code{FALSE}, overrides the default aesthetics, rather
#'   than combining them. This is most useful for helper functions that define
#'   both data and aesthetics and shouldn't inherit behaviour from the default
#'   plot specification, e.g., \code{\link[ggplot2]{borders}}.
#' @param check_overlap If \code{TRUE}, text that overlaps previous text in the
#'   same layer will not be plotted. \code{check_overlap} takes place at draw
#'   time and in the order of the data, thus its action depends of the size at
#'   which the plot is drawn.
#' @param size.unit How the `size` aesthetic is interpreted: as millimetres
#'   (`"mm"`, default), points (`"pt"`), centimetres (`"cm"`), inches (`"in"`),
#'   or picas (`"pc"`).
#' @param ... other arguments passed on to \code{\link[ggplot2]{layer}}. There
#'   are three types of arguments you can use here:
#'
#'   \itemize{ \item Aesthetics: to set an aesthetic to a fixed value, like
#'   \code{colour = "red"} or \code{size = 3}. \item Other arguments to the
#'   layer, for example you override the default \code{stat} associated with the
#'   layer. \item Other arguments passed on to the stat. }
#' @param nudge_x,nudge_y Horizontal and vertical adjustments to nudge the
#'   starting position of each text label. The units for \code{nudge_x} and
#'   \code{nudge_y} are the same as for the data units on the x-axis and y-axis.
#' @param default.colour,default.color A colour definition to use for elements
#'   not targeted by the colour aesthetic.
#' @param colour.target,color.target A vector of character strings; \code{"all"},
#'   \code{"text"}, \code{"segment"}, \code{"box"}, \code{"box.line"}, and
#'   \code{"box.fill"} or \code{"none"}.
#' @param default.alpha numeric in [0..1] A transparency value to use for
#'   elements not targeted by the alpha aesthetic.
#' @param alpha.target A vector of character strings; \code{"all"},
#'   \code{"text"}, \code{"segment"}, \code{"box"}, \code{"box.line"}, and
#'   \code{"box.fill"} or \code{"none"}.
#' @param add.segments logical Display connecting segments or arrows between
#'   original positions and displaced ones if both are available.
#' @param box.padding,point.padding numeric By how much each end of the segments
#'   should shortened in mm.
#' @param segment.linewidth numeric Width of the segments or arrows in mm.
#' @param min.segment.length numeric Segments shorter that the minimum length
#'   are not rendered, in mm.
#' @param arrow specification for arrow heads, as created by
#'   \code{\link[grid]{arrow}}
#'
#' @section Aesthetics: Layer functions \code{geom_text_s()} and
#'   \code{geom_label_s()} require aesthetics \code{x}, \code{y} and
#'   \code{label} and support aesthetics: \code{alpha}, \code{colour},
#'   \code{group}, \code{size} (of text), \code{family}, \code{fontface},
#'   \code{lineheight}, \code{hjust} and \code{vjust}. In addition,
#'   \code{geom_text_s} supports \code{angle} and \code{geom_label_s} supports
#'   \code{fill}, \code{linewidth} and \code{linetype}. See
#'   \code{\link[ggplot2]{aes_colour_fill_alpha}},
#'   \code{\link[ggplot2]{aes_linetype_size_shape}},
#'   \code{\link[ggplot2]{aes_position}}, and
#'   \code{\link[ggplot2]{aes_group_order}}.
#'
#'   In 'ggplot2' \code{linewidth} when applied to the border of the box drawn
#'   by \code{geom_label()} is given in points rather than in mm because of a
#'   historical error in the code. In other geometries such as
#'   \code{geom_segment()} \code{linewidth} is given in mm. As in
#'   \code{geom_label_s()} it is important to remain consistent among
#'   different \code{linewidth} specifications, mm are used both for the box
#'   border and linking segment. To imitate the behaviour of \code{geom_label()}
#'   a correction factor of 0.75 (more exactly 1 pt = 0.7528 mm) can be used for
#'   the line width of the border of the box.
#'
#' @section Position functions: Many layer functions from package 'ggpp' are
#'   designed to work seamlessly with position functions that keep, rather than
#'   discard, the original \code{x} and \code{y} positions in \code{data} when
#'   computing a new displaced position. See \code{\link{position_nudge_keep}},
#'   \code{\link{position_dodge_keep}}, \code{\link{position_jitter_keep}},
#'   \code{\link{position_nudge_center}}, \code{\link{position_nudge_line}},
#'   \code{\link{position_nudge_to}}, \code{\link{position_dodgenudge}},
#'   \code{\link{position_jitternudge}}, and \code{\link{position_stacknudge}}
#'   for examples and details of their use.
#'
#' @seealso \code{\link[ggplot2]{geom_text}}, \code{\link[ggplot2]{geom_label}}
#'   and other documentation of package 'ggplot2'.
#'
#' @return A plot layer instance.
#'
#' @export
#'
#' @examples
#'
#' my.cars <- mtcars[c(TRUE, FALSE, FALSE, FALSE), ]
#' my.cars$name <- rownames(my.cars)
#'
#' # no nudging
#' ggplot(my.cars, aes(wt, mpg, label = name)) +
#'   geom_text_s() +
#'   expand_limits(x = c(2, 6))
#'
#' # base plot
#' p <- ggplot(my.cars, aes(wt, mpg, label = name)) +
#'        geom_point()
#'
#' # Using nudging
#' p +
#'   geom_text_s(nudge_x = 0.12) +
#'   expand_limits(x = 6.2)
#' p +
#'   geom_text_s(nudge_x = -0.12) +
#'   expand_limits(x = 1.5)
#' p +
#'   geom_text_s(nudge_x = 0.12,
#'               arrow = arrow(length = grid::unit(1.5, "mm")),
#'               point.padding = 0.4) +
#'   expand_limits(x = 6.2)
#' p +
#'   geom_text_s(nudge_y = 0.1, nudge_x = 0.07) +
#'   expand_limits(x = 6.2)
#' p +
#'   geom_text_s(nudge_y = 1, angle = 90) +
#'   expand_limits(y = 30)
#' p +
#'   geom_text_s(angle = 90, nudge_y = 1,
#'               arrow = arrow(length = grid::unit(1.5, "mm")),
#'               colour.target = "segment", colour = "red") +
#'   expand_limits(y = 30)
#' p +
#'   geom_text_s(aes(colour = factor(cyl)),
#'               angle = 90, nudge_y = 1,
#'               arrow = arrow(length = grid::unit(1.5, "mm")),
#'               alpha.target = "segment", alpha = 0.3) +
#'   expand_limits(y = 30)
#'
#' p +
#'   geom_label_s(nudge_x = 0.12) +
#'   expand_limits(x = 6.2)
#' p +
#'   geom_label_s(nudge_x = 0.12, linetype = "dotted", linewidth = 0.3) +
#'   expand_limits(x = 6.2)
#' p +
#'   geom_label_s(aes(colour = factor(cyl)),
#'                nudge_x = 0.12,
#'                colour.target = "box",
#'                linewidth = 0.5,
#'                label.r = unit(0, "lines")) +
#'   expand_limits(x = 6.2)
#' p +
#'   geom_label_s(nudge_x = 0.12, linewidth = 0) +
#'   expand_limits(x = 6.2)
#'
#' # No segments
#' p +
#'   geom_label_s(nudge_x = 0.05, segment.linewidth = 0) +
#'   expand_limits(x = 6.2)
#'
#' # Nudging away from arbitrary point
#' p +
#'   geom_label_s(hjust = "outward_1", nudge_x = 0.12) +
#'   expand_limits(x = 6.2)
#' p +
#'   geom_label_s(hjust = "inward_3", nudge_y = 0.4)
#'
#' p +
#'   geom_label_s(nudge_y = 1, angle = 90) +
#'   expand_limits(y = 30)
#'
#' # Add aesthetic mappings and adjust arrows
#' p +
#'   geom_text_s(aes(colour = factor(cyl)),
#'               angle = 90,
#'               nudge_y = 1,
#'               arrow = arrow(angle = 20,
#'                             length = grid::unit(1.5, "mm"),
#'                             ends = "first",
#'                             type = "closed")) +
#'   scale_colour_discrete(l = 40) + # luminance, make colours darker
#'   expand_limits(y = 27)
#'
#' p +
#'   geom_text_s(aes(colour = factor(cyl)),
#'               angle = 90,
#'               nudge_y = 1,
#'               arrow = arrow(angle = 20,
#'                             length = grid::unit(1.5, "mm"),
#'                             ends = "first",
#'                             type = "closed")) +
#'   scale_colour_discrete(l = 40) + # luminance, make colours darker
#'   expand_limits(y = 27)
#'
#' p +
#'   geom_label_s(aes(colour = factor(cyl)),
#'               colour.target = c("box", "text"),
#'               nudge_x = 0.3,
#'               arrow = arrow(angle = 20,
#'                             length = grid::unit(1/3, "lines"))) +
#'   scale_colour_discrete(l = 40) + # luminance, make colours darker
#'   expand_limits(x = 7)
#'
#' p +
#'   geom_label_s(aes(colour = factor(cyl)),
#'               nudge_x = 0.3,
#'               colour.target = c("box", "segment"),
#'               linewidth = 0.5,
#'               arrow = arrow(angle = 20,
#'                             length = grid::unit(1/3, "lines"))) +
#'   scale_colour_discrete(l = 40) + # luminance, make colours darker
#'   expand_limits(x = 7)
#'
#' p +
#'   geom_label_s(aes(colour = factor(cyl), fill = factor(cyl)),
#'               nudge_x = 0.3,
#'               alpha.target = "box",
#'               alpha = 0.1,
#'               linewidth = 0.5,
#'               arrow = arrow(angle = 20,
#'                             length = grid::unit(1/3, "lines"))) +
#'   scale_colour_discrete(l = 40) + # luminance, make colours darker
#'   expand_limits(x = 7)#' # Scale height of text, rather than sqrt(height)
#'
#' p +
#'   geom_text_s(aes(size = wt), nudge_x = -0.1) +
#'   scale_radius(range = c(3,6)) + # override scale_area()
#'     expand_limits(x = c(1.8, 5.5))
#'
geom_text_s <- function(mapping = NULL,
                        data = NULL,
                        stat = "identity",
                        position = "identity",
                        ...,
                        parse = FALSE,
                        nudge_x = 0,
                        nudge_y = 0,
                        default.colour = NULL,
                        default.color = default.colour,
                        colour.target = "text",
                        color.target = colour.target,
                        default.alpha = NA,
                        alpha.target = "all",
                        add.segments = TRUE,
                        box.padding = 0.25,
                        point.padding = 1e-06,
                        segment.linewidth = 0.5,
                        min.segment.length = 0,
                        arrow = NULL,
                        check_overlap = FALSE,
                        size.unit = "mm",
                        na.rm = FALSE,
                        show.legend = NA,
                        inherit.aes = TRUE) {

  colour.target <-
    rlang::arg_match(color.target,
                     values = c("all", "text", "segment", "none"),
                     multiple = TRUE)
  alpha.target <-
    rlang::arg_match(alpha.target,
                     values = c("all", "text", "segment", "none"),
                     multiple = TRUE)


  if (!missing(nudge_x) || !missing(nudge_y)) {
    if (!missing(position) && position != "identity") {
      rlang::abort("You must specify either `position` or `nudge_x`/`nudge_y`.")
    }
    # original position needed for "position" justification
    position <-
      position_nudge_center(nudge_x, nudge_y, kept.origin = "original")
  }

  ggplot2::layer(
    data = data,
    mapping = mapping,
    stat = stat,
    geom = GeomTextS,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      parse = parse,
      default.colour = default.color,
      colour.target = colour.target,
      default.alpha = default.alpha,
      alpha.target = alpha.target,
      add.segments = add.segments,
      box.padding = box.padding,
      point.padding = point.padding,
      segment.linewidth = segment.linewidth,
      min.segment.length = min.segment.length,
      arrow = arrow,
      check_overlap = check_overlap,
      size.unit = size.unit,
      na.rm = na.rm,
      ...
    )
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
GeomTextS <-
  ggplot2::ggproto("GeomTextS", ggplot2::Geom,
                   required_aes = c("x", "y", "label"),

                   non_missing_aes = "angle",

                   default_aes = ggplot2::aes(
                     colour = "black",
                     family = "",
                     size = 3.87,
                     angle = 0,
                     hjust = "position",
                     vjust = "position",
                     alpha = NA,
                     fontface = 1,
                     lineheight = 1.2
                   ),

                   draw_panel = function(data,
                                         panel_params,
                                         coord, #panel_scales,
                                         parse = FALSE,
                                         size.unit = "mm",
                                         default.colour = NULL,
                                         colour.target = "all",
                                         default.alpha = NA,
                                         alpha.target = "all",
                                         na.rm = FALSE,
                                         check_overlap = FALSE,
                                         add.segments = TRUE,
                                         box.padding = 0.25,
                                         point.padding = 1e-06,
                                         segment.linewidth = 0.5,
                                         min.segment.length = 0,
                                         arrow = NULL) {

                     add.segments <- add.segments && any(c("x_orig", "y_orig") %in% colnames(data))

                     default.colour <- check_default_colour(default.colour)

                     data$label <- as.character(data$label)
                     data <- subset(data, !is.na(label) & label != "")
                     if (nrow(data) == 0L) {
                       return(nullGrob())
                     }

                     lab <- data$label
                     if (parse) {
                       lab <- parse_safe(lab)
                     }

                     data <- coord$transform(data, panel_params)

                     if (all(c("x_orig", "y_orig") %in% colnames(data))) {
                       data_orig <- data.frame(x = data$x_orig, y = data$y_orig)
                       data_orig <- coord$transform(data_orig, panel_params)
                       data$x_orig <- data_orig$x
                       data$y_orig <- data_orig$y
                     }

                     if (is.character(data$vjust)) {
                       data$vjust <-
                         compute_just2d(data = data,
                                        coord = coord,
                                        panel_params = panel_params,
                                        just = data$vjust,
                                        a = "y", b = "x")
                     }
                     if (is.character(data$hjust)) {
                       data$hjust <-
                         compute_just2d(data = data,
                                        coord = coord,
                                        panel_params = panel_params,
                                        just = data$hjust,
                                        a = "x", b = "y")
                     }

                     size.unit <- resolve_text_unit(size.unit)

                     if (add.segments) {
                       segments.data <-
                         shrink_segments(data,
                                         point.padding = point.padding,
                                         box.padding = box.padding,
                                         min.segment.length = min.segment.length)
                     }
                     # loop needed as gpar is not vectorized
                     all.grobs <- grid::gList()

                     for (row.idx in 1:nrow(data)) {
                       row <- data[row.idx, , drop = FALSE]
                       text.alpha <-
                         ifelse(any(alpha.target %in% c("all", "text")),
                                row$alpha, default.alpha)
                       segment.alpha <-
                         ifelse(any(alpha.target %in% c("all", "segment")),
                                row$alpha, default.alpha)
                       user.grob <- grid::textGrob(
                         lab[row.idx],
                         row$x, row$y, default.units = "native",
                         hjust = row$hjust, vjust = row$vjust,
                         rot = row$angle,
                         gp = gpar(
                           col = ifelse(any(colour.target %in% c("all", "text")),
                                        ggplot2::alpha(row$colour, text.alpha),
                                        ggplot2::alpha(default.colour, text.alpha)),
                           fontsize = row$size * size.unit,
                           fontfamily = row$family,
                           fontface = row$fontface,
                           lineheight = row$lineheight
                         ),
                         check.overlap = check_overlap
                       )

                       # give unique name to each grob
                       user.grob$name <- paste("text.s.grob", row$group, row.idx, sep = ".")

                       if (add.segments) {
                         segment.row <- segments.data[row.idx, , drop = FALSE]
                         if (segment.row$too.short) {
                           segment.grob <- grid::nullGrob()
                         } else {
                           segment.grob <-
                             grid::segmentsGrob(x0 = segment.row$x,
                                                y0 = segment.row$y,
                                                x1 = segment.row$x_orig,
                                                y1 = segment.row$y_orig,
                                                arrow = arrow,
                                                gp = grid::gpar(
                                                  col = if (segment.linewidth == 0) NA else # lwd = 0 is invalid in 'grid'
                                                    ifelse(any(colour.target %in% c("all", "segment")),
                                                           ggplot2::alpha(row$colour, segment.alpha),
                                                           ggplot2::alpha(default.colour, segment.alpha)),
                                                  lwd = (if (segment.linewidth == 0) 0.5 else segment.linewidth) * ggplot2::.stroke),
                                                name = paste("text.s.segment", row$group, row.idx, sep = "."))
                         }
                         all.grobs <- grid::gList(all.grobs, segment.grob, user.grob)
                       } else {
                         all.grobs <- grid::gList(all.grobs, user.grob)
                       }
                     }

                     # name needs to be unique within plot, so we would need to know other layers
#                     grid::grobTree(children = all.grobs, name = "geom.text.s.panel")
                     grid::grobTree(children = all.grobs)

                   },

                   draw_key = draw_key_text_s
  )

# heavily modified from geom-text.r from 'ggplot2' 3.1.0
# when just is "outward" or "inward" or one of its variants and the geom
# supports the angle aesthetics we need to take into account that justification
# is relative to the text, rather than the plot axes. By using compute_split()
# we add support for definitions of "inward" and "outward" relative to
# arbitrary positions along the axis.
#
# We support "position" (could be called "away") when nudging or other
# displacement has been applied and the original position saved.
#
# This function can handle either hjust or vjust, but only one at a time.
compute_just2d <- function(data,
                           coord,
                           panel_params,
                           just,
                           a = "x",
                           b = a) {
  just <- ifelse(just == "away", "position", just)
  if (a != b) {
    angle <- data$angle
  } else {
    angle <- 0
  }
  if (any(grepl("outward|inward|position", just))) {
    # ensure all angles are in -360...+360
    angle <- angle %% 360
    # ensure correct behaviour for angles in -360...+360
    angle <- ifelse(angle > 180, angle - 360, angle)
    angle <- ifelse(angle < -180, angle + 360, angle)
    rotated_forward <-
      grepl("outward|inward|position", just) & (angle > 45 & angle < 135)
    rotated_backwards <-
      grepl("outward|inward|position", just) & (angle < -45 & angle > -135)

    ab <- ifelse(rotated_forward | rotated_backwards, b, a)
    swap_ab <- rotated_backwards | abs(angle) > 135

    if (any(just == "position")) {
      # compute justification based on saved original position
      # works only with special position functions from 'ggpp' and 'ggrepel'
      ab_orig <- paste(ab, "_orig", sep = "")
      position <- just == "position"
      if (!all(unique(ab_orig) %in% colnames(data))) {
        # this needs to be silent as it is normal behaviour
        just[position] <- "middle"
      } else {
        just[position] <-
          c("left", "middle", "right")[2L + 1L *
                                         sign(data[[ab_orig[1L]]][position] -
                                                data[[ab[1L]]][position])]
      }
    }

    if (any(grepl("outward|inward", just))) {
      # we allow tags for setting center/middle position for just
      just_used <- unique(just)
      just_special <- grep("_mean$|_median$|.*[0-9].*", just_used, value = TRUE)
      middle <- rep(0.5, length(just))
      for (j in just_special) { # skipped if just_special has length zero
        j_selector <- just == j
        if (j %in% c("outward_mean", "inward_mean")) {
          middle[j_selector & !swap_ab] <- mean(data[[a]])
          middle[j_selector & swap_ab] <- mean(data[[b]])
        } else if (j %in% c("outward_median", "inward_median")) {
          middle[j_selector & !swap_ab] <- stats::median(data[[a]])
          middle[j_selector & swap_ab] <- stats::median(data[[b]])
        } else {
          middle[j_selector & swap_ab] <- stats::median(data[[b]])
          middle_a <- as.numeric(gsub("outward_|inward_", "", unique(just)))
          if (a == "x") {
            tmp_data <- tibble::tibble(x = middle_a, y = data[[b]])
            middle[j_selector & !swap_ab] <- coord$transform(tmp_data, panel_params)$x
          } else {
            tmp_data <- tibble::tibble(y = middle_a, x = data[[b]])
            middle[j_selector & !swap_ab] <- coord$transform(tmp_data, panel_params)$y
          }
        }
      }

      just <- gsub("_.*$", "", just)

      # what follows is like in ggplot2 except for split_at
      obs <- data[[a]]
      obs[swap_ab] <- data[[b]][swap_ab]

      inward <- just == "inward"
      just[inward] <- c("left", "middle", "right")[just_dir(obs[inward],
                                                            split_at = middle[inward])]
      outward <- just == "outward"
      just[outward] <- c("right", "middle", "left")[just_dir(obs[outward],
                                                             split_at = middle[outward])]
    }
  }

  unname(c(left = 0, center = 0.5, right = 1,
           bottom = 0, middle = 0.5, top = 1)[just])
}

# modified from geom-text.r from 'ggplot2' 3.1.0 to support arbitrary split
# for "inward" and "outward" justification
just_dir <- function(x, tol = 0.001, split_at = 0.5) {
  out <- rep(2L, length(x))
  out[x < split_at - tol] <- 1L
  out[x > split_at + tol] <- 3L
  out
}

# as in pull request to ggplot2
# can be used by other geoms
compute_just <- function(just, a, b = a, angle = 0) {
  #  As justification direction is relative to the text, not the plotting area
  #  we need to swap x and y if text direction is rotated so that hjust is
  #  applied along y and vjust along x.
  if (any(grepl("outward|inward", just))) {
    # ensure all angles are in -360...+360
    angle <- angle %% 360
    # ensure correct behaviour for angles in -360...+360
    angle <- ifelse(angle > 180, angle - 360, angle)
    angle <- ifelse(angle < -180, angle + 360, angle)
    rotated_forward <-
      grepl("outward|inward", just) & (angle > 45 & angle < 135)
    rotated_backwards <-
      grepl("outward|inward", just) & (angle < -45 & angle > -135)

    ab <- ifelse(rotated_forward | rotated_backwards, b, a)
    just_swap <- rotated_backwards | abs(angle) > 135
    inward <-
      (just == "inward" & !just_swap | just == "outward" & just_swap)
    just[inward] <- c("left", "middle", "right")[just_dir(ab[inward])]
    outward <-
      (just == "outward" & !just_swap) | (just == "inward" & just_swap)
    just[outward] <- c("right", "middle", "left")[just_dir(ab[outward])]

  }

  unname(c(left = 0, center = 0.5, right = 1,
           bottom = 0, middle = 0.5, top = 1)[just])
}

# shorten segments to add padding
# code based on https://stackoverflow.com/questions/22649781/
#
# box.padding and point.pading givern in mm
#
shrink_segments <- function(data,
                            box.padding = 0,
                            point.padding = 0,
                            min.segment.length = 0.5) {
  stopifnot("'box.padding' must be >= 0" = box.padding >= 0,
            "'point.padding' must be >= 0" =  point.padding >= 0)
  segments.data <- data[ , c("x_orig", "y_orig", "x", "y")]
  starting.length <- apply(segments.data, 1,
                           function(x) stats::dist(rbind(x[1:2], x[3:4])))

  zero.len <- starting.length < max(1e-5, (point.padding + box.padding) / 100)

  # padding origin
  if (point.padding != 0) {
    p <- point.padding / starting.length[!zero.len] / 100
    p <- 1 - p
    segments.data[!zero.len, "x_orig"] <- data[!zero.len, "x"] + p * (data[!zero.len, "x_orig"] - data[!zero.len, "x"])
    segments.data[!zero.len, "y_orig"] <- data[!zero.len, "y"] + p * (data[!zero.len, "y_orig"] - data[!zero.len, "y"])
  }
  # padding position
  if (box.padding != 0) {
    p <- box.padding / starting.length[!zero.len] / 100
    p <- 1 - p
    segments.data[!zero.len, "x"] <- data[!zero.len, "x_orig"] + p * (data[!zero.len, "x"] - data[!zero.len, "x_orig"])
    segments.data[!zero.len, "y"] <- data[!zero.len, "y_orig"] + p * (data[!zero.len, "y"] - data[!zero.len, "y_orig"])
  }
  final.length <- apply(segments.data, 1,
                        function(x) stats::dist(rbind(x[1:2], x[3:4])))
  segments.data$too.short <- final.length < min.segment.length
  segments.data
}

# from ggplot2
resolve_text_unit <- function(unit) {
  unit <- rlang::arg_match0(unit, c("mm", "pt", "cm", "in", "pc"))
  switch(
    unit,
    "mm" = .pt,
    "cm" = .pt * 10,
    "in" = 72.27,
    "pc" = 12,
    1
  )
}

# from ggplot2 utilities-grid.R
#
# Name ggplot grid object
# Convenience function to name grid objects
#
# @keyword internal
ggname <- function(prefix, grob) {
  grob$name <- grobName(grob, prefix)
  grob
}

context("geom_text_s")
# Here we not only test geom_text_s() but also function compute_just2d()
# which translates character strings into numeric values of justification
# taking into account angle of rotation and direction splits.

test_that("text_linked_just", {
  df <- data.frame(
    x = c(1, 1, 2, 2, 1.5),
    y = c(1, 2, 1, 2, 1.5),
    text = c("bottom-left", "top-left", "bottom-right", "top-right", "center")
  )

  vdiffr::expect_doppelganger("geom_text_s1",
                              ggplot(df, aes(x, y)) +
                                geom_text_s(aes(label = text))
  )

  vdiffr::expect_doppelganger("geom_text_s2",
                              ggplot(df, aes(x, y)) +
                                geom_text_s(aes(label = text),
                                            vjust = "inward", hjust = "inward")
  )
  vdiffr::expect_doppelganger("geom_text_s3",
                              ggplot(df, aes(x, y)) +
                                geom_text_s(aes(label = text),
                                            hjust = "inward", angle = 33)
  )
  vdiffr::expect_doppelganger("geom_text_s4",
                              ggplot(df, aes(x, y)) +
                                geom_text_s(aes(label = text),
                                            hjust = "inward", angle = 66)
  )
  vdiffr::expect_doppelganger("geom_text_s5",
                              ggplot(df, aes(x, y)) +
                                geom_text_s(aes(label = text),
                                            hjust = "inward", angle = 90)
  )
  vdiffr::expect_doppelganger("geom_text_s7",
                              ggplot(df, aes(x, y)) +
                                geom_text_s(aes(label = text),
                                            vjust = "inward",
                                            hjust = "inward", angle = 33)
  )
  vdiffr::expect_doppelganger("geom_text_s8",
                              ggplot(df, aes(x, y)) +
                                geom_text_s(aes(label = text),
                                            vjust = "inward",
                                            hjust = "inward", angle = 66)
  )
  vdiffr::expect_doppelganger("geom_text_s9",
                              ggplot(df, aes(x, y)) +
                                geom_text_s(aes(label = text),
                                            vjust = "inward",
                                            hjust = "inward", angle = 90)
  )
  vdiffr::expect_doppelganger("geom_text_s10",
                              ggplot(df, aes(x, y)) +
                                geom_text_s(aes(label = text),
                                            vjust = "inward",
                                            hjust = "inward", angle = 90)
  )
  vdiffr::expect_doppelganger("geom_text_s11",
                              ggplot(df, aes(x, y)) +
                                geom_text_s(aes(label = text),
                                            vjust = "inward_1.5",
                                            hjust = "inward_1.5")
  )
  vdiffr::expect_doppelganger("geom_text_s12",
                              ggplot(df, aes(x - 1.5, y - 1.5)) +
                                geom_text_s(aes(label = text),
                                            vjust = "inward_0.0",
                                            hjust = "inward_0.0")
  )
  vdiffr::expect_doppelganger("geom_text_s13",
                              ggplot(df, aes(x, y)) +
                                geom_text_s(aes(label = text),
                                            vjust = "inward_mean",
                                            hjust = "inward_mean")
  )
  vdiffr::expect_doppelganger("geom_text_s14",
                              ggplot(df, aes(x - 1.5, y - 1.5)) +
                                geom_text_s(aes(label = text),
                                            vjust = "inward_mean",
                                            hjust = "inward_mean")
  )
  vdiffr::expect_doppelganger("geom_text_s15",
                              ggplot(df, aes(x, y)) +
                                geom_text_s(aes(label = text),
                                            vjust = "inward_median",
                                            hjust = "inward_median")
  )

  vdiffr::expect_doppelganger("geom_text_s16",
                              ggplot(df, aes(x, y)) +
                                geom_text_s(aes(label = text),
                                            vjust = "outward",
                                            hjust = "outward")
  )
  vdiffr::expect_doppelganger("geom_text_s17",
                              ggplot(df, aes(x, y)) +
                                geom_text_s(aes(label = text),
                                            vjust = "outward_mean",
                                            hjust = "outward_mean")
  )
  vdiffr::expect_doppelganger("geom_text_s18",
                              ggplot(df, aes(x - 1.5, y - 1.5)) +
                                geom_text_s(aes(label = text),
                                            vjust = "outward_mean",
                                            hjust = "outward_mean")
  )
  vdiffr::expect_doppelganger("geom_text_s19",
                              ggplot(df, aes(x, y)) +
                                geom_text_s(aes(label = text),
                                            vjust = "outward_median",
                                            hjust = "outward_median")
  )
  vdiffr::expect_doppelganger("geom_text_s20",
                              ggplot(df, aes(x - 1.5, y - 1.5)) +
                                geom_text_s(aes(label = text),
                                            vjust = "outward_median",
                                            hjust = "outward_median")
  )
  vdiffr::expect_doppelganger("geom_text_s21",
                              ggplot(df, aes(x - 1.5, y - 1.5)) +
                                geom_text_s(aes(label = text),
                                            vjust = "position",
                                            hjust = "position")
  )
  vdiffr::expect_doppelganger("geom_text_s22",
                              ggplot(df, aes(x - 1.5, y - 1.5)) +
                                geom_text_s(aes(label = text),
                                            vjust = "position",
                                            hjust = "position",
                                            nudge_x = 0.01,
                                            nudge_y = 0.01) +
                                expand_limits(x = 0.6)
  )
  vdiffr::expect_doppelganger("geom_text_s23",
                              ggplot(df, aes(x - 1.5, y - 1.5)) +
                                geom_text_s(aes(label = text),
                                            vjust = "position",
                                            hjust = "position",
                                            nudge_x = -0.01,
                                            nudge_y = -0.01) +
                                expand_limits(x = -0.6)
  )
  vdiffr::expect_doppelganger("geom_text_s24",
                              ggplot(df, aes(x - 1.5, y - 1.5)) +
                                geom_text_s(aes(label = text),
                                            vjust = "position",
                                            hjust = "position",
                                            nudge_x = 0.01,
                                            nudge_y = -0.01) +
                                expand_limits(x = 0.6)
  )
  vdiffr::expect_doppelganger("geom_text_s25",
                              ggplot(df, aes(x - 1.5, y - 1.5)) +
                                geom_text_s(aes(label = text),
                                            vjust = "position",
                                            hjust = "position",
                                            nudge_x = -0.01,
                                            nudge_y = 0.01) +
                                expand_limits(x = -0.6)
  )
  vdiffr::expect_doppelganger("geom_text_s26",
                              ggplot(df, aes(x - 1.5, y - 1.5)) +
                                geom_text_s(aes(label = text),
                                            vjust = "middle",
                                            hjust = "center")  +
                                expand_limits(x = c(-0.6, 0.6))
  )
  vdiffr::expect_doppelganger("geom_text_s27",
                              ggplot(df, aes(x - 1.5, y - 1.5)) +
                                geom_text_s(aes(label = text),
                                            vjust = "top",
                                            hjust = "right")  +
                                expand_limits(x = c(-0.6, 0.6))
  )
  vdiffr::expect_doppelganger("geom_text_s28",
                              ggplot(df, aes(x - 1.5, y - 1.5)) +
                                geom_text_s(aes(label = text),
                                            vjust = "bottom",
                                            hjust = "left")  +
                                expand_limits(x = c(-0.6, 0.6))
  )

})

test_that("text_linked_size", {
# size affects only the text
  my.cars <- datasets::mtcars[c(TRUE, FALSE, FALSE, FALSE), ]
  my.cars$name <- rownames(my.cars)

  vdiffr::expect_doppelganger("geom_text_s30",
                              ggplot(my.cars, aes(wt, mpg, label = name)) +
                                geom_point() +
                                geom_text_s(aes(size = wt), nudge_x = -0.1, hjust = "right") +
                                scale_radius(range = c(3,6)) + # override scale_area()
                                expand_limits(x = c(1.8, 5.5))
  )
}
)

test_that("text_linked_linewidth", {
# segment.linewidth is a parameter, not an aesthetic
  my.cars <- datasets::mtcars[c(TRUE, FALSE, FALSE, FALSE), ]
  my.cars$name <- rownames(my.cars)

  vdiffr::expect_doppelganger("geom_text_s31",
                              ggplot(my.cars, aes(wt, mpg, label = name)) +
                                geom_point() +
                                geom_text_s(segment.linewidth = 1, nudge_x = -0.1, hjust = "right") +
                                scale_radius(range = c(3,6)) + # override scale_area()
                                expand_limits(x = c(1.8, 5.5))
  )
  vdiffr::expect_doppelganger("geom_text_s32",
                              ggplot(my.cars, aes(wt, mpg, label = name)) +
                                geom_point() +
                                geom_text_s(segment.linewidth = 0, nudge_x = -0.1, hjust = "right") +
                                scale_radius(range = c(3,6)) + # override scale_area()
                                expand_limits(x = c(1.8, 5.5))
  )
  vdiffr::expect_doppelganger("geom_text_s33",
                              ggplot(my.cars, aes(wt, mpg, label = name)) +
                                geom_point() +
                                geom_text_s(add.segments = FALSE, nudge_x = -0.1, hjust = "right") +
                                scale_radius(range = c(3,6)) + # override scale_area()
                                expand_limits(x = c(1.8, 5.5))
  )
}
)

test_that("text_linked_color", {

  my.cars <- datasets::mtcars[c(TRUE, FALSE, FALSE, FALSE), ]
  my.cars$name <- rownames(my.cars)

  vdiffr::expect_doppelganger("geom_text_s40",
                              ggplot(my.cars, aes(wt, mpg, label = name)) +
                                geom_text_s(aes(colour = wt), nudge_x = -0.1, hjust = "right") +
                                geom_point() +
                                expand_limits(x = c(1.8, 5.5))
  )
  vdiffr::expect_doppelganger("geom_text_s41",
                              ggplot(my.cars, aes(wt, mpg, label = name)) +
                                geom_text_s(aes(colour = wt), colour.target = "segment",
                                            nudge_x = -0.1, hjust = "right") +
                                geom_point() +
                                expand_limits(x = c(1.8, 5.5))
  )
  vdiffr::expect_doppelganger("geom_text_s42",
                              ggplot(my.cars, aes(wt, mpg, label = name)) +
                                geom_text_s(aes(colour = wt), colour.target = "text",
                                            nudge_x = -0.1, hjust = "right") +
                                geom_point() +
                                expand_limits(x = c(1.8, 5.5))
  )
  vdiffr::expect_doppelganger("geom_text_s43",
                              ggplot(my.cars, aes(wt, mpg, label = name)) +
                                geom_text_s(aes(colour = wt), colour.target = "all",
                                            nudge_x = -0.1, hjust = "right") +
                                geom_point() +
                                expand_limits(x = c(1.8, 5.5))
  )
  vdiffr::expect_doppelganger("geom_text_s45",
                              ggplot(my.cars, aes(wt, mpg, label = name)) +
                                geom_text_s(colour = "red", nudge_x = -0.1, hjust = "right") +
                                geom_point() +
                                expand_limits(x = c(1.8, 5.5))
  )
  vdiffr::expect_doppelganger("geom_text_s46",
                              ggplot(my.cars, aes(wt, mpg, label = name)) +
                                geom_text_s(colour = "red", colour.target = "segment",
                                            nudge_x = -0.1, hjust = "right") +
                                geom_point() +
                                expand_limits(x = c(1.8, 5.5))
  )
  vdiffr::expect_doppelganger("geom_text_s47",
                              ggplot(my.cars, aes(wt, mpg, label = name)) +
                                geom_text_s(colour = "red", colour.target = "text",
                                            nudge_x = -0.1, hjust = "right") +
                                geom_point() +
                                expand_limits(x = c(1.8, 5.5))
  )
  vdiffr::expect_doppelganger("geom_text_s48",
                              ggplot(my.cars, aes(wt, mpg, label = name)) +
                                geom_text_s(colour = "red", colour.target = "all",
                                            nudge_x = -0.1, hjust = "right") +
                                geom_point() +
                                expand_limits(x = c(1.8, 5.5))
  )
  vdiffr::expect_doppelganger("geom_text_s49",
                              ggplot(my.cars, aes(wt, mpg, label = name)) +
                                geom_text_s(colour = "red", colour.target = "text",
                                            default.colour = "blue",
                                            nudge_x = -0.1, hjust = "right") +
                                geom_point(colour = "blue") +
                                expand_limits(x = c(1.8, 5.5))
  )
  vdiffr::expect_doppelganger("geom_text_s50",
                              ggplot(my.cars, aes(wt, mpg, label = name)) +
                                geom_text_s(colour = "red", colour.target = "segment",
                                            default.colour = "blue",
                                            nudge_x = -0.1, hjust = "right") +
                                geom_point(colour = "blue") +
                                expand_limits(x = c(1.8, 5.5))
  )
  vdiffr::expect_doppelganger("geom_text_s51",
                              ggplot(my.cars, aes(wt, mpg, label = name)) +
                                geom_text_s(aes(colour = wt), colour.target = "text",
                                            default.colour = "blue",
                                            nudge_x = -0.1, hjust = "right") +
                                geom_point(colour = "blue") +
                                expand_limits(x = c(1.8, 5.5))
  )
  vdiffr::expect_doppelganger("geom_text_s52",
                              ggplot(my.cars, aes(wt, mpg, label = name)) +
                                geom_text_s(aes(colour = wt), colour.target = "segment",
                                            default.colour = "blue",
                                            nudge_x = -0.1, hjust = "right") +
                                geom_point(colour = "blue") +
                                expand_limits(x = c(1.8, 5.5))
  )
}
)

test_that("shrink_segments returns correct values", {
  dfs <- data.frame(x_orig = 1, y_orig = 1, x = 2, y = 2)

  value <- shrink_segments(dfs)
  expect_equal(value[ , -5], dfs)
  expect_false(value$too.short[1])

  value <- ggpp:::shrink_segments(dfs, point.padding = 0.5)
  expect_equal(as.vector(t(value[1:4])),
               c(1.003536, 1.003536, 2.000000, 2.000000),
               tolerance = 1e-5)
  expect_false(value$too.short[1])

  value <- ggpp:::shrink_segments(dfs, box.padding = 0.5)
  expect_equal(as.vector(t(value[1:4])),
               c(1.000000, 1.00000, 1.996464, 1.996464),
               tolerance = 1e-5)
  expect_false(value$too.short[1])

  value <- ggpp:::shrink_segments(dfs,
                                  box.padding = 0.5 - 1e-15,
                                  point.padding = 0.5 - 1e-15)
  expect_equal(as.vector(t(value[1:4])),
               c(1.003536, 1.003536, 1.996464, 1.996464),
               tolerance = 1e-5)
  expect_false(value$too.short[1])

  value <- shrink_segments(data.frame(x_orig = 1, y_orig = 1, x = 1, y = 1))
  expect_equal(as.vector(t(value[1:4])),
               c(1, 1, 1, 1),
               tolerance = 1e-5)
  expect_true(value$too.short[1])

  value <- shrink_segments(data.frame(x_orig = 1, y_orig = 1, x = 1, y = 1),
                           point.padding = 0.5)
  expect_equal(as.vector(t(value[1:4])),
               c(1, 1, 1, 1),
               tolerance = 1e-5)
  expect_true(value$too.short[1])

  value <- shrink_segments(data.frame(x_orig = 1, y_orig = 1, x = 1, y = 1),
                           box.padding = 0.5)
  expect_equal(as.vector(t(value[1:4])),
               c(1, 1, 1, 1),
               tolerance = 1e-5)
  expect_true(value$too.short[1])

  })

#' Number of observations in quadrants
#'
#' \code{stat_quadrant_counts()} counts the number of observations in each
#' quadrant of a plot panel. By default it adds a text label to the far corner
#' of each quadrant. It can also be used to obtain the total number of
#' observations in each of two pairs of quadrants or in the whole panel.
#' Grouping is ignored, so en every case a single count is computed for each
#' quadrant in a plot panel.
#'
#' @param mapping The aesthetic mapping, usually constructed with
#'   \code{\link[ggplot2]{aes}} or \code{\link[ggplot2]{aes_}}. Only needs to be
#'   set at the layer level if you are overriding the plot defaults.
#' @param data A layer specific dataset - only needed if you want to override
#'   the plot defaults.
#' @param geom The geometric object to use display the data
#' @param position The position adjustment to use on this layer
#' @param show.legend logical. Should this layer be included in the legends?
#'   \code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
#'   never includes, and \code{TRUE} always includes.
#' @param inherit.aes If \code{FALSE}, overrides the default aesthetics, rather
#'   than combining with them. This is most useful for helper functions that
#'   define both data and aesthetics and should not inherit behaviour from the
#'   default plot specification, e.g., \code{\link[ggplot2]{borders}}.
#' @param ... other arguments passed on to \code{\link[ggplot2]{layer}}. This
#'   can include aesthetics whose values you want to set, not map. See
#'   \code{\link[ggplot2]{layer}} for more details.
#' @param na.rm	a logical indicating whether \code{NA} values should be stripped
#'   before the computation proceeds.
#' @param quadrants integer vector indicating which quadrants are of interest,
#'   with a \code{OL} indicating the whole plot.
#' @param pool.along character, one of \code{"none"}, \code{"x"} or \code{"y"},
#'   indicating which quadrants to pool to calculate counts by pair of
#'   quadrants.
#' @param xintercept,yintercept numeric the coordinates of the origin of the
#'   quadrants.
#' @param label.x,label.y \code{numeric} Coordinates (in npc units) to be used
#'   for absolute positioning of the labels.
#' @param digits integer Number of digits for fraction and percent labels.
#'
#' @details This statistic can be used to automatically count observations in
#'   each of the four quadrants of a plot, and by default add these counts as
#'   text labels. Values exactly equal to \code{xintercept} or
#'   \code{yintercept} are counted together with those larger than the
#'   intercepts. An argument value of zero, passed to formal parameter
#'   \code{quadrants} is interpreted as a request for the count of all
#'   observations in each plot panel.
#'
#'   The default origin of quadrants is at \code{xintercept = 0},
#'   \code{yintercept = 0}. Also by default, counts are computed for all
#'   quadrants within the \emph{x} and \emph{y} scale limits, but ignoring any
#'   marginal scale expansion. The default positions of the labels is in the
#'   farthest corner or edge of each quadrant using npc coordinates.
#'   Consequently, when using facets even with free limits for \emph{x} and
#'   \emph{y} axes, the location of the labels is consistent across panels. This
#'   is achieved by use of \code{geom = "text_npc"} or \code{geom =
#'   "label_npc"}. To pass the positions in native data units, pass \code{geom =
#'   "text"} explicitly as argument.
#'
#' @section Computed variables: Data frame with one to four rows, one for each
#'   quadrant for which counts are counted in \code{data}. \describe{
#'   \item{quadrant}{integer, one of 0:4} \item{x}{x value of label position in
#'   data units} \item{y}{y value of label position in data units} \item{npcx}{x
#'   value of label position in npc units} \item{npcy}{y value of label position
#'   in npc units} \item{count}{number of  observations in the quadrant(s)}
#'   \item{total}{number of onservations in data}
#'   \item{count.label}{number of observations as character}
#'   \item{pc.label}{percent of observations as character}
#'   \item{fr.label}{fraction of observations as character} }.
#'
#'   As shown in one example below \code{\link[gginnards]{geom_debug}} can be
#'   used to print the computed values returned by any statistic. The output
#'   shown includes also values mapped to aesthetics, like \code{label} in the
#'   example.
#'
#' @return A plot layer instance. Using as output \code{data} the counts of
#'   observations per plot quadrant.
#'
#' @family Functions for quadrant and volcano plots
#'
#' @export
#'
#' @examples
#'
#' # generate artificial data
#' set.seed(4321)
#' x <- -50:50
#' y <- rnorm(length(x), mean = 0)
#' my.data <- data.frame(x, y)
#'
#' # using automatically generated text labels, default origin at (0, 0)
#'
#' ggplot(my.data, aes(x, y)) +
#'   geom_point() +
#'   geom_quadrant_lines() +
#'   stat_quadrant_counts()
#'
#' ggplot(my.data, aes(x, y)) +
#'   geom_point() +
#'   geom_quadrant_lines() +
#'   stat_quadrant_counts(aes(label = after_stat(pc.label)))
#'
#' ggplot(my.data, aes(x, y)) +
#'   geom_point() +
#'   geom_quadrant_lines() +
#'   stat_quadrant_counts(aes(label = after_stat(fr.label)))
#'
#' ggplot(my.data, aes(x, y)) +
#'   geom_point() +
#'   geom_quadrant_lines() +
#'   stat_quadrant_counts(aes(label = after_stat(dec.label)))
#'
#' ggplot(my.data, aes(x, y)) +
#'  geom_point() +
#'   geom_quadrant_lines() +
#'   stat_quadrant_counts(aes(label = sprintf("%i observations", after_stat(count)))) +
#'   scale_y_continuous(expand = expansion(c(0.05, 0.15))) # reserve space
#'
#' # user specified origin
#'
#' ggplot(my.data, aes(x, y)) +
#'   geom_quadrant_lines(colour = "blue", xintercept = 10, yintercept = -1) +
#'   stat_quadrant_counts(colour = "blue", xintercept = 10, yintercept = -1) +
#'   geom_point() +
#'   scale_y_continuous(expand = expansion(mult = 0.15))
#'
#' ggplot(my.data, aes(x, y)) +
#'   geom_quadrant_lines(colour = "blue", xintercept = 10, yintercept = -1) +
#'   stat_quadrant_counts(aes(label = after_stat(pc.label)),
#'                        colour = "blue", xintercept = 10, yintercept = -1) +
#'   geom_point() +
#'   scale_y_continuous(expand = expansion(mult = 0.15))
#'
#' # more digits in labels
#'
#' ggplot(my.data, aes(x, y)) +
#'   geom_quadrant_lines(colour = "blue", xintercept = 10, yintercept = -1) +
#'   stat_quadrant_counts(aes(label = after_stat(pc.label)), digits = 3,
#'                        colour = "blue", xintercept = 10, yintercept = -1) +
#'   geom_point() +
#'   scale_y_continuous(expand = expansion(mult = 0.15))
#'
#' ggplot(my.data, aes(x, y)) +
#'   geom_quadrant_lines(colour = "blue", xintercept = 10, yintercept = -1) +
#'   stat_quadrant_counts(aes(label = after_stat(fr.label)),
#'                        colour = "blue", xintercept = 10, yintercept = -1) +
#'   geom_point() +
#'   scale_y_continuous(expand = expansion(mult = 0.15))
#'
#' # grouped quadrants
#'
#' ggplot(my.data, aes(x, y)) +
#'   geom_quadrant_lines(colour = "blue",
#'                       pool.along = "x") +
#'   stat_quadrant_counts(colour = "blue", label.x = "right",
#'                        pool.along = "x") +
#'   geom_point() +
#'   scale_y_continuous(expand = expansion(mult = 0.15))
#'
#' # whole panel
#'
#' ggplot(my.data, aes(x, y)) +
#'   geom_point() +
#'   stat_quadrant_counts(quadrants = 0, label.x = "left", label.y = "bottom") +
#'   scale_y_continuous(expand = expansion(mult = c(0.15, 0.05)))
#'
#' # use a different geometry
#'
#' ggplot(my.data, aes(x, y)) +
#'   geom_point() +
#'   stat_quadrant_counts(geom = "text") # use geom_text()
#'
#' # Numeric values can be used to build labels with alternative formats
#' # Here with sprintf(), but paste() and format() also work.
#'
#' ggplot(my.data, aes(x, y)) +
#'   geom_quadrant_lines(colour = "blue") +
#'   stat_quadrant_counts(aes(label = sprintf("%i of %i genes",
#'                        after_stat(count), after_stat(total))),
#'                        colour = "blue") +
#'   geom_point() +
#'   scale_y_continuous(expand = expansion(mult = 0.15))
#'
#' # We use geom_debug_group() to see the computed values
#'
#' gginnards.installed <- requireNamespace("gginnards", quietly = TRUE)
#' if (gginnards.installed) {
#'   library(gginnards)
#'
#'   ggplot(my.data, aes(x, y)) +
#'     geom_point() +
#'     stat_quadrant_counts(geom = "debug_group")
#' }
#'
#' if (gginnards.installed) {
#'   ggplot(my.data, aes(x, y)) +
#'     geom_point() +
#'     stat_quadrant_counts(geom = "debug_group", xintercept = 50)
#' }
#'
stat_quadrant_counts <- function(mapping = NULL,
                                 data = NULL,
                                 geom = "text_npc",
                                 position = "identity",
                                 quadrants = NULL,
                                 pool.along = c("none", "x", "y", "xy"),
                                 xintercept = 0,
                                 yintercept = 0,
                                 label.x = NULL,
                                 label.y = NULL,
                                 digits = 2,
                                 na.rm = FALSE,
                                 show.legend = FALSE,
                                 inherit.aes = TRUE,
                                 ...) {

  pool.along <- rlang::arg_match(pool.along)
  if (!is.null(pool.along) && pool.along == "xy") {
    quadrants <- 0L # consistent UI with other stats from 'ggpp'
  }
  stopifnot("'xintercept' should have length == 1L" = length(xintercept) <= 1,
            "'yintercept' should have length == 1L" = length(yintercept) <= 1)
  stopifnot("Length of 'quadrants' shoul be between 0 and 4" = length(quadrants) <= 4)
  stopifnot(is.null(label.x) || is.numeric(label.x) || is.character(label.x))
  stopifnot(is.null(label.y) || is.numeric(label.y) || is.character(label.y))

  ggplot2::layer(
    stat = StatQuadrantCounts,
    data = data,
    mapping = mapping,
    geom = geom,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(na.rm = na.rm,
                  quadrants = quadrants,
                  pool.along = pool.along,
                  xintercept = xintercept,
                  yintercept = yintercept,
                  label.x = label.x,
                  label.y = label.y,
                  digits = digits,
                  ...)
  )
}

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
StatQuadrantCounts <-
  ggplot2::ggproto("StatQuadrantCounts", ggplot2::Stat,

                   compute_panel = function(data,
                                            scales,
                                            quadrants,
                                            pool.along,
                                            xintercept,
                                            yintercept,
                                            label.x,
                                            label.y,
                                            digits = 2) {

                     which_quadrant <- function(x, y) {
                       z <- ifelse(x >= xintercept & y >= yintercept,
                                   1L,
                                   ifelse(x >= xintercept & y < yintercept,
                                          2L,
                                          ifelse(x < xintercept & y < yintercept,
                                                 3L,
                                                 4L)))
                       if (pool.along == "x") {
                         z <- ifelse(z %in% c(1L, 4L), 1L, 2L)
                       } else if(pool.along == "y") {
                         z <- ifelse(z %in% c(1L, 2L), 1L, 4L)
                       }
                       z
                     }

                     force(data)
                     # compute range of whole data
                     range.x <- range(data$x)
                     range.y <- range(data$y)
                     # set position for labels in npc units
                     if (is.null(label.x)) {
                       if (pool.along == "x") {
                         label.x <- rep("centre", 2)
                       } else {
                         label.x <- c("left", "right")
                       }
                     }
                     if (is.null(label.y)) {
                       if (pool.along == "y") {
                         label.y <- rep("centre", 2)
                       } else {
                         label.y <- c("bottom", "top")
                       }
                     }

                     label.x <- compute_npcx(label.x)
                     label.y <- compute_npcy(label.y)

                     label.x <- range(label.x) # ensure length is always 2
                     label.y <- range(label.y) # ensure length is always 2

                     # dynamic default based on data range
                     if (is.null(quadrants)) {
                       if (all(range.x >= xintercept) && all(range.y >= yintercept)) {
                         quadrants <- 1L
                       } else if (all(range.x < xintercept) && all(range.y < yintercept)) {
                         quadrants <- 3L
                       } else if (all(range.x >= xintercept)) {
                         quadrants <- c(1L, 2L)
                       } else if (all(range.y >= yintercept)) {
                         quadrants <- c(1L, 4L)
                       } else {
                         quadrants <- c(1L, 2L, 3L, 4L)
                       }
                     }
                     if (pool.along == "x") {
                       quadrants <- intersect(quadrants, c(1L, 2L))
                     }
                     if (pool.along == "y") {
                       quadrants <- intersect(quadrants, c(1L, 4L))
                     }

                     num.obs <- nrow(data)

                     if (all(is.na(quadrants)) || 0L %in% quadrants) {
                       # total count
                       z <-
                         tibble::tibble(quadrant = 0,
                                        count = num.obs,
                                        total = num.obs,
                                        npcx = label.x[2],
                                        npcy = label.y[2],
                                        x = range.x[2],
                                        y = range.y[2])
                     } else {
                       # counts for the selected quadrants
                       data |>
                         dplyr::mutate(quadrant = which_quadrant(.data$x, .data$y)) |>
                         dplyr::filter(.data$quadrant %in% quadrants) |>
                         dplyr::group_by(.data$quadrant) |>
                         dplyr::summarise(count = length(.data$x)) |> # dplyr::n() triggers error
                         dplyr::ungroup() -> data

                       data$total <- num.obs

                       zero.count.quadrants <- setdiff(quadrants, data$quadrant)

                       if (length(zero.count.quadrants) > 0) {
                         data <-
                           rbind(data,
                                 tibble::tibble(quadrant = zero.count.quadrants,
                                                count = 0L, total = num.obs))
                       }

                       data$count.label <- sprintf("n=%i", data$count)
                       data$pc.label <- sprintf("p=%.*f%%",
                                                digits - 2,
                                                data$count / data$total * 100)
                       data$dec.label <- sprintf("f=%.*f",
                                                 digits,
                                                 data$count / data$total)
                       data$fr.label <- sprintf("%i / %i",
                                                data$count, data$total)

                       z <-
                         data |>
                         dplyr::mutate(npcx = ifelse(.data$quadrant %in% c(1L, 2L),
                                                     label.x[2],
                                                     label.x[1]),
                                       npcy = ifelse(.data$quadrant %in% c(1L, 4L),
                                                     label.y[2],
                                                     label.y[1]),
                                       x = ifelse(.data$quadrant %in% c(1L, 2L),
                                                  range.x[2],
                                                  range.x[1]),
                                       y = ifelse(.data$quadrant %in% c(1L, 4L),
                                                  range.y[2],
                                                  range.y[1]))
                     }

                     z[["count.label"]] <- sprintf("n=%i", z$count)

                     # returned data frame must contain a "group" column
                     if (! "group" %in% colnames(z)) {
                       z[["group"]] <- -1L
                     }

                     z
                   },

                   default_aes =
                     ggplot2::aes(npcx = ggplot2::after_stat(npcx),
                                  npcy = ggplot2::after_stat(npcy),
                                  label = ggplot2::after_stat(count.label),
                                  hjust = "inward",
                                  vjust = "inward"),
                   required_aes = c("x", "y")
  )

context("stat_quadrant_counts")

library(ggplot2)

test_that("stat_quadrant_counts", {
  tst.df <- data.frame(
    x = c(2.2, -0.2, -0.2, -2.5, -0.6, -0.1),
    y = c(1.1, -0.6, -0.9, -1.6, 0.3, 1.6),
    group = c("A", "A", "A", "B", "B", "B")
  )
  vdiffr::expect_doppelganger("stat_quadrant_counts_1",
                              ggplot(tst.df, aes(x, y)) +
                                stat_quadrant_counts())

  p <- ggplot(tst.df, aes(x, y)) +
    stat_quadrant_counts()
  result <- layer_data(p)[, c("npcx", "npcy", "label", "quadrant", "count",
                              "total", "count.label", "pc.label", "dec.label", "fr.label", "hjust", "vjust")]
  expected <- data.frame(
    npcx = c(0.95, 0.05, 0.05, 0.95),
    npcy = c(0.95, 0.05, 0.95, 0.05),
    label = c("n=1", "n=3", "n=2", "n=0"),
    quadrant = c(1L, 3L, 4L, 2L),
    count = c(1L, 3L, 2L, 0L),
    total = 6L,
    count.label = c("n=1", "n=3", "n=2", "n=0"),
    pc.label = c("p=17%", "p=50%", "p=33%", "p=0%"),
    dec.label = c("f=0.17", "f=0.50", "f=0.33", "f=0.00"),
    fr.label = c("1 / 6", "3 / 6", "2 / 6", "0 / 6"),
    hjust = "inward",
    vjust = "inward"
  )
  expect_identical(result, expected)

  p <- ggplot(tst.df, aes(x, y)) +
    stat_quadrant_counts(digits = 3)
  result <- layer_data(p)[, c("npcx", "npcy", "label", "quadrant", "count",
                              "total", "count.label", "pc.label", "dec.label", "fr.label", "hjust", "vjust")]
  expected <- data.frame(
    npcx = c(0.95, 0.05, 0.05, 0.95),
    npcy = c(0.95, 0.05, 0.95, 0.05),
    label = c("n=1", "n=3", "n=2", "n=0"),
    quadrant = c(1L, 3L, 4L, 2L),
    count = c(1L, 3L, 2L, 0L),
    total = 6L,
    count.label = c("n=1", "n=3", "n=2", "n=0"),
    pc.label = c("p=16.7%", "p=50.0%", "p=33.3%", "p=0.0%"),
    dec.label = c("f=0.167", "f=0.500", "f=0.333", "f=0.000"),
    fr.label = c("1 / 6", "3 / 6", "2 / 6", "0 / 6"),
    hjust = "inward",
    vjust = "inward"
  )
  expect_identical(result, expected)

  p <- ggplot(tst.df, aes(x, y)) +
    stat_quadrant_counts(quadrant = 0)
  result <- layer_data(p)[, c("npcx", "npcy", "label", "quadrant", "count", "total", "hjust", "vjust")]
  expected <- data.frame(
    npcx = 0.95,
    npcy = 0.95,
    label = "n=6",
    quadrant = 0L,
    count = 6L,
    total = 6L,
    hjust = "inward",
    vjust = "inward")
  expect_equal(result, expected) # not identical!

  p <- ggplot(tst.df, aes(x, y)) +
    stat_quadrant_counts(pool.along = "x")
  result <- layer_data(p)[, c("npcx", "npcy", "label", "quadrant", "count", "total", "hjust", "vjust")]
  expected <- data.frame(
    npcx = 0.5,
    npcy = c(0.95, 0.05),
    label = "n=3",
    quadrant = c(1L, 2L),
    count = 3L,
    total = 6L,
    hjust = "inward",
    vjust = "inward")
  expect_identical(result, expected)

  p <- ggplot(tst.df, aes(x, y)) +
    stat_quadrant_counts(pool.along = "y")
  result <- layer_data(p)[, c("npcx", "npcy", "label", "quadrant", "count", "total", "hjust", "vjust")]
  expected <- data.frame(
    npcx = c(0.95, 0.05),
    npcy = 0.5,
    label = c("n=1", "n=5"),
    quadrant = c(1L, 4L),
    count = c(1L, 5L),
    total = 6L,
    hjust = "inward",
    vjust = "inward")
  expect_identical(result, expected)

  p <- ggplot(tst.df, aes(x, y)) +
    stat_quadrant_counts(geom = "text")
  result <- layer_data(p)[, c("label", "x", "y", "quadrant", "count", "total", "hjust", "vjust")]
  expected <- data.frame(
    label = c("n=1", "n=3", "n=2", "n=0"),
    x = c(2.2, -2.5, -2.5, 2.2),
    y = c(1.6, -1.6, 1.6, -1.6),
    quadrant = c(1L, 3L, 4L, 2L),
    count = c(1L, 3L, 2L, 0L),
    total = 6L,
    hjust = 0.5,
    vjust = 0.5
  )
  expect_identical(result, expected)

  expect_error(ggplot(tst.df, aes(x, y)) +
                 stat_quadrant_counts(label.x = NA))

  expect_error(ggplot(tst.df, aes(x, y)) +
                 stat_quadrant_counts(label.y = NA))

  expect_error(ggplot(tst.df, aes(x, y)) +
                 stat_quadrant_counts(pool.along = "z"))

  expect_error(ggplot(tst.df, aes(x, y)) +
                 stat_quadrant_counts(quadrants = numeric(5)))

})

#' @title Replace labels in data based on 1D density
#'
#' @description \code{stat_dens1d_labels()} Sets values mapped to the
#'   \code{label} aesthetic to \code{""} or a user provided character string
#'   based on the local density in regions of a plot panel. Its main use is
#'   together with repulsive geoms from package \code{\link[ggrepel]{ggrepel}}
#'   to restrict labeling to the low density tails of a distribution. By default
#'   the data are handled all together, but it is also possible to control
#'   labeling separately in each tail.
#'
#'   If there is no mapping to \code{label} in \code{data}, the mapping is set
#'   to \code{rownames(data)}, with a message.
#'
#' @details \code{stat_dens1d_labels()} is designed to work together with
#'   geometries from package 'ggrepel'. To avoid text labels being plotted over
#'   unlabelled points the corresponding rows in data need to be retained but
#'   labels replaced with the empty character string, \code{""}. Function
#'   \code{\link{stat_dens1d_filter}} cannot be used with the repulsive geoms
#'   from 'ggrepel' because it drops the observations.
#'
#'   \code{stat_dens1d_labels()} can be useful also in other situations, as the
#'   substitution character string can be set by the user by passing an argument
#'   to \code{label.fill}. If this argument is \code{NULL} the unselected rows
#'   are filtered out.
#'
#'   The local density of observations along \emph{x} or \emph{y} is computed
#'   with function \code{\link[stats]{density}} and used to select observations,
#'   passing to the geom all the rows in its \code{data} input but with with the
#'   text of labels replaced in those "not kept". The default is to select
#'   observations in sparse regions of the plot, but the selection can be
#'   inverted so that only observations in the densest regions are returned.
#'   Specific observations can be protected from having the label replaced by
#'   passing a suitable argument to \code{keep.these}. Logical and integer
#'   vectors function as indexes to rows in \code{data}, while a character
#'   vector is compared to values in the variable mapped to the \code{label}
#'   aesthetic. A function passed as argument to keep.these will receive as
#'   argument the values in the variable mapped to \code{label} and should
#'   return a character, logical or numeric vector as described above.
#'
#'   How many labels are retained intact in addition to those in
#'   \code{keep.these} is controlled with arguments passed to \code{keep.number}
#'   and \code{keep.fraction}. \code{keep.number} sets the maximum number of
#'   observations selected, whenever \code{keep.fraction} results in fewer
#'   observations selected, it is obeyed. If \code{xintercept} is a finite value
#'   within the \emph{x} range of the data and \code{pool.along} is passed
#'   \code{"none"} the data are split into two groups and \code{keep.number} and
#'   \code{keep.fraction} are applied separately to each tail with density still
#'   computed jointly from all observations. If the length of \code{keep.number}
#'   and \code{keep.fraction} is one, half this value is used each tail, if
#'   their length is two, the first value is use for the left tail and the
#'   second value for the right tail (or if using \code{orientation = "y"} the
#'   lower and upper tails, respectively).
#'
#'   Computation of density and of the default bandwidth require at least
#'   two observations with different values. If data do not fulfill this
#'   condition, they are kept only if \code{keep.fraction = 1}. This is correct
#'   behavior for a single observation, but can be surprising in the case of
#'   multiple observations.
#'
#'   Parameters \code{keep.these} and \code{exclude.these} make it possible to
#'   force inclusion or exclusion of labels after the density is computed.
#'   In case of conflict, \code{exclude.these} overrides \code{keep.these}.
#'
#' @note Which points are kept and which not depends on how dense and flexible
#'   is the density curve estimate. This depends on the values passed as
#'   arguments to parameters \code{n}, \code{bw} and \code{kernel}. It is
#'   also important to be aware that both \code{geom_text()} and
#'   \code{geom_text_repel()} can avoid overplotting by discarding labels at
#'   the plot rendering stage, i.e., what is plotted may differ from what is
#'   returned by this statistic.
#'
#' @param mapping The aesthetic mapping, usually constructed with
#'   \code{\link[ggplot2]{aes}} or \code{\link[ggplot2]{aes_}}. Only needs to be
#'   set at the layer level if you are overriding the plot defaults.
#' @param data A layer specific dataset - only needed if you want to override
#'   the plot defaults.
#' @param geom The geometric object to use display the data.
#' @param keep.fraction numeric vector of length 1 or 2 [0..1]. The fraction of
#'   the observations (or rows) in \code{data} to be retained.
#' @param keep.number integer vector of length 1 or 2. Set the maximum number of
#'   observations to retain, effective only if obeying \code{keep.fraction}
#'   would result in a larger number.
#' @param keep.sparse logical If \code{TRUE}, the default, observations from the
#'   more sparse regions are retained, if \code{FALSE} those from the densest
#'   regions.
#' @param keep.these,exclude.these character vector, integer vector, logical
#'   vector or function that takes one or more variables in data selected by
#'   \code{these.target}. Negative integers behave as in R's extraction methods.
#'   The rows from \code{data} indicated by \code{keep.these} and
#'   \code{exclude.these} are kept or excluded irrespective of the local
#'   density.
#' @param these.target character, numeric or logical selecting one or more
#'   column(s) of \code{data}. If \code{TRUE} the whole \code{data} object is
#'   passed.
#' @param pool.along character, one of \code{"none"} or \code{"x"},
#'   indicating if selection should be done pooling the observations along the
#'   \emph{x} aesthetic, or separately on either side of \code{xintercept}.
#' @param xintercept numeric The split point for the data filtering.
#' @param invert.selection logical If \code{TRUE}, the complement of the
#'   selected rows are returned.
#' @param bw numeric or character The smoothing bandwidth to be used. If
#'   numeric, the standard deviation of the smoothing kernel. If character, a
#'   rule to choose the bandwidth, as listed in \code{\link[stats]{bw.nrd}}.
#' @param adjust numeric A multiplicative bandwidth adjustment. This makes it
#'   possible to adjust the bandwidth while still using the a bandwidth
#'   estimator through an argument passed to \code{bw}. The larger the value
#'   passed to \code{adjust} the stronger the smoothing, hence decreasing
#'   sensitivity to local changes in density.
#' @param kernel character See \code{\link{density}} for details.
#' @param n numeric Number of equally spaced points at which the density is to
#'   be estimated for applying the cut point. See \code{\link{density}} for
#'   details.
#' @param return.density logical vector of lenght 1. If \code{TRUE} add columns
#'   \code{"density"} and \code{"keep.obs"} to the returned data frame.
#' @param orientation	character The aesthetic along which density is computed.
#'   Given explicitly by setting orientation to either "x" or "y".
#' @param label.fill character vector of length 1 or a function.
#' @param position The position adjustment to use for overlapping points on this
#'   layer
#' @param show.legend logical. Should this layer be included in the legends?
#'   \code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
#'   never includes, and \code{TRUE} always includes.
#' @param inherit.aes If \code{FALSE}, overrides the default aesthetics, rather
#'   than combining with them. This is most useful for helper functions that
#'   define both data and aesthetics and shouldn't inherit behaviour from the
#'   default plot specification, e.g. \code{\link[ggplot2]{borders}}.
#' @param ... other arguments passed on to \code{\link[ggplot2]{layer}}. This
#'   can include aesthetics whose values you want to set, not map. See
#'   \code{\link[ggplot2]{layer}} for more details.
#' @param na.rm	a logical value indicating whether NA values should be stripped
#'   before the computation proceeds.
#'
#' @seealso \code{\link[stats]{density}} used internally.
#'
#' @family statistics returning a subset of data
#'
#' @return A plot layer instance. Using as output \code{data} the input
#'   \code{data} after value substitution based on a 1D the filtering criterion.
#'
#' @examples
#'
#' random_string <-
#'   function(len = 6) {
#'     paste(sample(letters, len, replace = TRUE), collapse = "")
#'   }
#'
#' # Make random data.
#' set.seed(1005)
#' d <- tibble::tibble(
#'   x = rnorm(100),
#'   y = rnorm(100),
#'   group = rep(c("A", "B"), c(50, 50)),
#'   lab = replicate(100, { random_string() })
#' )
#'
#' # using defaults
#' ggplot(data = d, aes(x, y, label = lab)) +
#'   geom_point() +
#'   stat_dens1d_labels()
#'
#' ggrepel.installed <- requireNamespace("ggrepel", quietly = TRUE)
#' if (ggrepel.installed) {
#'   library(ggrepel)
#'
#' # using defaults
#'   ggplot(data = d, aes(x, y, label = lab)) +
#'     geom_point() +
#'     stat_dens1d_labels(geom = "text_repel")
#'
#' # if no mapping to label is found, it is set row names
#'   ggplot(data = d, aes(x, y)) +
#'     geom_point() +
#'     stat_dens1d_labels(geom = "text_repel")
#'
#'   ggplot(data = d, aes(x, y)) +
#'     geom_point() +
#'     stat_dens1d_labels(geom = "text_repel", pool.along = "none")
#'
#'   ggplot(data = d, aes(x, y)) +
#'     geom_point() +
#'     stat_dens1d_labels(geom = "text_repel",
#'                        keep.number = c(0, 10), pool.along = "none")
#'
#'   ggplot(data = d, aes(x, y)) +
#'     geom_point() +
#'     stat_dens1d_labels(geom = "text_repel",
#'                        keep.fraction = c(0, 0.2), pool.along = "none")
#'
#' # using defaults, along y-axis
#'   ggplot(data = d, aes(x, y, label = lab)) +
#'     geom_point() +
#'     stat_dens1d_labels(orientation = "y", geom = "text_repel")
#'
#' # example labelling with coordiantes
#'   ggplot(data = d, aes(x, y, label = sprintf("x = %.2f\ny = %.2f", x, y))) +
#'     geom_point() +
#'     stat_dens1d_filter(colour = "red") +
#'     stat_dens1d_labels(geom = "text_repel", colour = "red", size = 3)
#'
#'   ggplot(data = d, aes(x, y, label = lab, colour = group)) +
#'     geom_point() +
#'     stat_dens1d_labels(geom = "text_repel")
#'
#'   ggplot(data = d, aes(x, y, label = lab, colour = group)) +
#'     geom_point() +
#'     stat_dens1d_labels(geom = "text_repel", label.fill = NA)
#'
#' # we keep labels starting with "a" across the whole plot, but all in sparse
#' # regions. To achieve this we pass as argument to label.fill a fucntion
#' # instead of a character string.
#'   label.fun <- function(x) {ifelse(grepl("^a", x), x, "")}
#'   ggplot(data = d, aes(x, y, label = lab, colour = group)) +
#'     geom_point() +
#'     stat_dens1d_labels(geom = "text_repel", label.fill = label.fun)
#' }
#'
#' # Using geom_debug_group() we can see that all 100 rows in \code{d} are
#' # returned.
#' gginnards.installed <- requireNamespace("gginnards", quietly = TRUE)
#' if (gginnards.installed) {
#'   library(gginnards)
#'
#'   ggplot(data = d, aes(x, y, label = lab)) +
#'     geom_point() +
#'     stat_dens1d_labels(geom = "debug_group")
#' }
#' if (gginnards.installed) {
#'   ggplot(data = d, aes(x, y, label = lab)) +
#'     geom_point() +
#'     stat_dens1d_labels(geom = "debug_group", return.density = TRUE)
#' }
#'
#' @export
#'
stat_dens1d_labels <-
  function(mapping = NULL,
           data = NULL,
           geom = "text",
           position = "identity",
           ...,
           keep.fraction = 0.10,
           keep.number = Inf,
           keep.sparse = TRUE,
           keep.these = FALSE,
           exclude.these = FALSE,
           these.target = "label",
           pool.along = c("x", "none"),
           xintercept = 0,
           invert.selection = FALSE,
           bw = "SJ",
           kernel = "gaussian",
           adjust = 1,
           n = 512,
           orientation = c("x", "y"),
           label.fill = "",
           return.density = FALSE,
           na.rm = TRUE,
           show.legend = FALSE,
           inherit.aes = TRUE) {

    pool.along <- rlang::arg_match(pool.along)
    orientation <- rlang::arg_match(orientation)

    if (length(label.fill) > 1L) {
      stop("Length for 'label.fill' is not 0 or 1: ", label.fill)
    }
    if (is.numeric(label.fill)) {
      stop("'label.fill' should not be a 'numeric' value: ", label.fill)
    }
    if (any(is.na(keep.fraction) | keep.fraction < 0 | keep.fraction > 1)) {
      stop("Out of range or missing value for 'keep.fraction': ", keep.fraction)
    }
    if (any(is.na(keep.number) | keep.number < 0)) {
      stop("Out of range or missing value for 'keep.number': ", keep.number)
    }

    ggplot2::layer(
      stat = StatDens1dLabels, data = data, mapping = mapping, geom = geom,
      position = position, show.legend = show.legend, inherit.aes = inherit.aes,
      params = list(na.rm = na.rm,
                    keep.fraction = keep.fraction,
                    keep.number = keep.number,
                    keep.sparse = keep.sparse,
                    keep.these = keep.these,
                    exclude.these = exclude.these,
                    these.target = these.target,
                    pool.along = pool.along,
                    xintercept = xintercept,
                    invert.selection = invert.selection,
                    bw = bw,
                    adjust = adjust,
                    kernel = kernel,
                    n = n,
                    orientation = orientation,
                    label.fill = label.fill,
                    return.density = return.density,
                    ...)
    )
  }

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
StatDens1dLabels <-
  ggplot2::ggproto(
    "StatDens1dLabels",
    ggplot2::Stat,
    compute_panel =
      function(data,
               scales,
               keep.fraction,
               keep.number,
               keep.sparse,
               keep.these,
               exclude.these,
               these.target = "label",
               pool.along,
               xintercept,
               invert.selection,
               bw,
               kernel,
               adjust,
               n,
               orientation,
               label.fill,
               return.density) {

        force(data)
        if (!exists("label", data) && !is.null(label.fill)) {
          data[["label"]] <- rownames(data)
        }

        keep.these <- these2logical(these = keep.these,
                                    data = data,
                                    these.target = these.target)

        exclude.these <- these2logical(these = exclude.these,
                                       data = data,
                                       these.target = these.target)

        # discard redundant splits and make list of logical vectors
        if (pool.along != "x" &&
            xintercept < max(data[[orientation]]) &&
            xintercept > min(data[[orientation]])) {

          selectors <-list(low.tail = data[[orientation]] <= xintercept,
                           high.tail = data[[orientation]] > xintercept)
          if (length(keep.fraction) != 2L) {
            keep.fraction <- rep_len(keep.fraction, length.out = 2)
          }
          if (length(keep.number) != 2L) {
            if (length(keep.number) == 1L) {
              keep.number <- keep.number %/% 2
            }
            keep.number <- rep_len(keep.number, length.out = 2)
          }
          num.rows <- sapply(selectors, sum) # selectors are logical
        } else {
          keep.fraction <- keep.fraction[[1]] # can be a vector or a list
          keep.number <- keep.number[[1]]
          num.rows <- nrow(data)
          selectors <- list(all = rep.int(TRUE, times = num.rows))
        }

        # vectorized
        too.large.frac <- num.rows * keep.fraction > keep.number
        keep.fraction[too.large.frac] <-
          keep.number[too.large.frac] / num.rows[too.large.frac]

        # density on a grid
        # data with fewer than 2 rows is as a special case as density() fails
        if (length(unique(data[[orientation]])) >= 2L) {
          dens <-
            stats::density(data[[orientation]],
                           bw = bw, kernel = kernel, adjust = adjust, n = n,
                           from = scales[[orientation]]$dimension()[1],
                           to = scales[[orientation]]$dimension()[2])

          # estimate density at each observations coordinates
          fdens <- stats::splinefun(dens$x, dens$y) # y contains estimate of density
          dens <- fdens(data[[orientation]])
        } else {
          if (nrow(data) > 1L) {
            message("Density not computed, too few distinct values in '", orientation, "'")
          }
          dens <- rep_len(1, nrow(data))
        }
        # we construct one logical vector by adding observations/label to be kept
        # we may have a list of 1 or 2 logical vectors
        keep <- logical(nrow(data))
        for (i in seq_along(selectors)) {
          if (keep.fraction[i] == 1) {
            keep[ selectors[[i]] ] <- TRUE
          } else if (keep.fraction[i] != 0 && length(selectors[[i]]) >= 2L) {
            if (keep.sparse) {
              keep[ selectors[[i]] ] <-
                dens[ selectors[[i]] ] < stats::quantile(dens[ selectors[[i]] ],
                                                         keep.fraction[i],
                                                         names = FALSE,
                                                         type = 8)
            } else {
              keep[ selectors[[i]] ] <-
                dens[ selectors[[i]] ] >= stats::quantile(dens[ selectors[[i]] ],
                                                          1 - keep.fraction[i],
                                                          names = FALSE,
                                                          type = 8)
            }
          }
        }
        keep <- (keep | keep.these) & !exclude.these

        if (invert.selection){
          keep <- !keep
        }

        if (return.density) {
          data[["keep.obs"]] <- keep
          data[["density"]] <- dens
        }

        if (is.null(label.fill)) {
          data <- data[keep, ]
        } else if (is.function(label.fill)) {
          data[["label"]][!keep] <- label.fill(data[["label"]][!keep])
        } else if (is.na(label.fill)) {
          # NA_logical_, the default NA, cannot always be assigned to character
          label.fill <- NA_character_
          data[["label"]][!keep] <- label.fill
        } else if (is.character(label.fill)) {
          data[["label"]][!keep] <- label.fill
        } else if (is.logical(label.fill)) {
          if (label.fill) {
            data[["label"]][!keep] <- ""
          } # if FALSE data is not modified
        } else {
          stop("'label.fill' is : ", mode(label.fill),
               " instead of 'character' or 'function'.")
        }
        data
      },

    required_aes = "x|y"
  )

#' Nudge labels away from a line
#'
#' \code{position_nudge_line()} is generally useful for adjusting the starting
#' position of labels or text to be repelled while preserving the original
#' position as the start of the segments. The difference compared to
#' \code{\link{position_nudge_center}} is that the nudging is away from from a
#' line or curve fitted to the data points or supplied as coefficients. While
#' \code{position_nudge_center()} is most useful for "round-shaped", vertically-
#' or horizontally elongated clouds of points, \code{position_nudge_line()} is
#' most suitable when observations follow a linear or curvilinear relationship
#' between \emph{x} and \emph{y} values. In contrast to
#' \code{\link[ggplot2]{position_nudge}}, \code{position_nudge_line()} returns
#' in `data` both the original coordinates and the nudged coordinates.
#'
#' @family position adjustments
#'
#' @param x,y Amount of vertical and horizontal distance to move. A numeric
#'   vector of length 1 or longer.
#' @param xy_relative Nudge relative to \emph{x} and \emph{y} data expanse, ignored unless
#'   \code{x} and \code{y} are both \code{NA}s.
#' @param abline a vector of length two giving the intercept and slope.
#' @param method One of \code{"spline"}, \code{"lm"} or \code{"auto"}.
#' @param formula A model formula for \code{\link{lm}} when \code{method =
#'   "lm"}. Ignored otherwise.
#' @param direction One of \code{"automatic"}, \code{"none"}, or \code{"split"}.
#' @param line_nudge A positive multiplier >= 1, increasing nudging away from
#'   the curve or line compared to nudging from points.
#' @param kept.origin One of \code{"original"} or \code{"none"}.
#'
#' @details The default amount of nudging is 3% of the spread of the data along
#'   \emph{x} and \emph{y} axes, which in most cases is good. In most cases it is best to
#'   apply nudging along a direction perpendicular to the line or curve, if this
#'   is the aim, passing an argument to only one of \code{x}, \code{y} or
#'   \code{xy_relative} will be enough. When \code{direction = "split"} nudging
#'   is away from an implicit line or curve on either side with positive
#'   nudging. The line or curve can be smooth spline or linear regression fitted
#'   on-the-fly to the data points, or a straight line defined by its
#'   coefficients passed to \code{abline}. The fitting is well defined only if
#'   the observations fall roughly on a curve or straight line that is monotonic
#'   in \code{y}. By means of \code{line_nudge} one can increment nudging away
#'   from the line or curve compared to away from the points, which is useful
#'   for example to keep labels outside of a confidence band. Direction defaults
#'   to \code{"split"} when \code{line_nudge} > 1, and otherwise to
#'   \code{"none"}.
#'
#' @note For \code{method = "lm"} only model formulas corresponding to
#'   polynomials with no missing terms are supported. If using\code{\link{poly}}
#'   in the model formula, \code{raw = TRUE} is required.
#'
#'   In practice, \code{x} and \code{y} should have the same sign for nudging to
#'   work correctly.
#'
#'   This position is most useful when labeling points conforming a cloud along
#'   an arbitrary curve or line.
#'
#' @seealso \code{\link[ggplot2]{position_nudge}},
#'  \code{\link[ggrepel]{position_nudge_repel}}.
#'
#' @return A \code{"Position"} object.
#'
#' @importFrom polynom polynomial
#'
#' @export
#'
#' @examples
#'
#' set.seed(16532)
#' df <- data.frame(
#'   x = -10:10,
#'   y = (-10:10)^2,
#'   yy = (-10:10)^2 + rnorm(21, 0, 4),
#'   yyy = (-10:10) + rnorm(21, 0, 4),
#'   l = letters[1:21]
#' )
#'
#' # Setting the nudging distance
#'
#' ggplot(df, aes(x, y, label = l)) +
#'   geom_line(linetype = "dotted") +
#'   geom_point() +
#'   geom_text(position = position_nudge_line())
#'
#' ggplot(df, aes(x, y, label = l)) +
#'   geom_line(linetype = "dotted") +
#'   geom_point() +
#'   geom_text_s(position = position_nudge_line())
#'
#' ggplot(df, aes(x, y, label = l)) +
#'   geom_line(linetype = "dotted") +
#'   geom_point() +
#'   geom_text(position = position_nudge_line(xy_relative = -0.03))
#'
#' ggplot(df, aes(x, y, label = l)) +
#'   geom_line(linetype = "dotted") +
#'   geom_point() +
#'   geom_text(position = position_nudge_line(x = 0.6, y = 3.2))
#'
#' ggplot(df, aes(x, y, label = l)) +
#'   geom_line(linetype = "dotted") +
#'   geom_point() +
#'   geom_text(position = position_nudge_line(x = -0.6, y = -4))
#'
#' # Other curves, using defaults
#'
#' ggplot(df, aes(x, -y, label = l)) +
#'   geom_line(linetype = "dotted") +
#'   geom_point() +
#'   geom_text(position = position_nudge_line())
#'
#' ggplot(subset(df, x >= 0), aes(y, sqrt(y), label = l)) +
#'   geom_line(linetype = "dotted") +
#'   geom_point() +
#'   geom_text(position = position_nudge_line())
#'
#' # Points scattered near a curve or line, we use 'direction = "split"'
#'
#' ggplot(df, aes(x)) +
#'   geom_line(aes(y = y), linetype = "dotted") +
#'   geom_point(aes(y = yy)) +
#'   geom_text(aes(y = yy, label = l),
#'             position = position_nudge_line(direction = "split"))
#'
#' ggplot(subset(df, x >= 0), aes(y, yy)) +
#'   stat_smooth(method = "lm", formula = y ~ x) +
#'   geom_point() +
#'   geom_text(aes(label = l),
#'             position = position_nudge_line(direction = "split"))
#'
#' # increasing the nudging for labels near the line
#'
#' ggplot(subset(df, x >= 0), aes(y, yy)) +
#'   stat_smooth(method = "lm", formula = y ~ x) +
#'   geom_point() +
#'   geom_text(aes(label = l),
#'               position = position_nudge_line(line_nudge = 2,
#'                                              direction = "split"))
#'
#' # fitting a linear model instead of the default spline
#'
#' ggplot(subset(df, x >= 0), aes(y, yy)) +
#'   stat_smooth(method = "lm", formula = y ~ x) +
#'   geom_point() +
#'   geom_text(aes(label = l),
#'             position = position_nudge_line(method = "lm",
#'                                            direction = "split"))
#'
#' ggplot(subset(df, x >= 0), aes(x, x^2)) +
#'   stat_smooth(method = "lm", formula = y ~ poly(x, 2, raw = TRUE)) +
#'   geom_point() +
#'   geom_text(aes(label = l),
#'             position = position_nudge_line(method = "lm",
#'                                            formula = y ~ poly(x, 2, raw = TRUE)))
#'
position_nudge_line <-
  function(x = NA_real_,
           y = NA_real_,
           xy_relative = c(0.03, 0.03),
           abline = NULL,
           method = NULL,
           formula = y ~ x,
           direction = c("automatic", "none", "split"),
           line_nudge = 1,
           kept.origin = c("original", "none")) {

    direction <- rlang::arg_match(direction)
    kept.origin <- rlang::arg_match(kept.origin)

    # set defaults
    if (!is.null(abline)) {
      method <- "abline"
    } else {
      abline <- rep(NA_real_, 2) # to ensure that a list member is created
    }

    if (is.null(method)) {
      method <- "auto" # decided later based on nrow(data)
    }

    if (method == "linear") {
      method <- "lm"
    }

    if (direction == "automatic") {
      if (line_nudge > 1) {
        direction <- "split"
      } else {
        direction <- "none"
      }
    }

    if (length(xy_relative) == 1) {
      xy_relative <- rep(xy_relative, 2)
    }

    stopifnot(length(xy_relative) == 2)

    if (lubridate::is.duration(x)) {
      x <- as.numeric(x)
    }
    if (lubridate::is.duration(y)) {
      y <- as.numeric(y)
    }

    ggplot2::ggproto(
      NULL,
      PositionNudgeLine,
      x = x,
      y = y,
      xy_relative = xy_relative,
      abline = abline,
      method = method,
      formula = formula,
      direction = direction,
      line_nudge = line_nudge,
      kept.origin = kept.origin
    )
  }

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @export
PositionNudgeLine <-
  ggplot2::ggproto(
    "PositionNudgeLine",
    Position,
    x = 0,
    y = 0,
    xy_relative = c(0.03, 0.03),
    abline = rep(NA_real_, 2),
    method = "spline",
    formula = y ~ x,
    direction = "none",
    line_nudge = 1,

    setup_params = function(self, data) {
      list(x = self$x,
           y = self$y,
           x.reorder = !is.null(self$x) && length(self$x) > 1 && length(self$x) < nrow(data),
           y.reorder = !is.null(self$y) && length(self$y) > 1 && length(self$y) < nrow(data),
           xy_relative = self$xy_relative,
           abline = self$abline,
           method = self$method,
           formula = self$formula,
           kept.origin = self$kept.origin,
           direction = self$direction,
           line_nudge = self$line_nudge
      )
    },

    compute_panel = function(data, params, scales) {

      if (length(params$x) > nrow(data)) {
        warning("Argument 'x' longer than data: some values dropped!")
      }
      if (length(params$y) > nrow(data)) {
        warning("Argument 'y' longer than data: some values dropped!")
      }
      x_orig <- data$x
      y_orig <- data$y
      # set parameter defaults that depend on the scale
      x_range <- scales$x$dimension()
      y_range <- scales$y$dimension()
      x_spread <- x_range[2] - x_range[1]
      y_spread <- y_range[2] - y_range[1]
      xy.range.ratio <- x_spread / y_spread

      if (all(is.na(params$x) & is.na(params$y))) {
        params$x <- params$xy_relative[1] * x_spread
        params$y <- params$xy_relative[2] * y_spread
      } else if (xor(all(is.na(params$x)), all(is.na(params$y)))) {
        if (is.na(params$x)) {
          params$x <- params$y * xy.range.ratio
        } else {
          params$y <- params$x / xy.range.ratio
        }
      }

      if (params$method == "auto") {
        if (nrow(data) < 5) {
          params$method <- "lm"
        } else {
          params$method <- "spline"
        }
      }

      # compute lines or curves and their derivatives
      if (params$method == "abline") {
        if (is.numeric(params$abline) && length(params$abline) == 2) {
          curve <- params$abline[1] + params$abline[2] * data$x
          # ensure same length in all cases
          sm.deriv <- rep(params$abline[2], nrow(data))
        } else {
          stop("'abline' should be a numeric vector of length 2")
        }
      } else if (params$method %in% c("lm", "spline")) {
        # we need to handle grouping by ourselves as compute_group does not work
        curve <- sm.deriv <- numeric(nrow(data))
        for (group in unique(data$group)) {
          in.grp <- data$group == group
          if (nrow(data[in.grp, ]) < 4 || params$method == "lm") {
            mf <- stats::lm(formula = params$formula, data = data[in.grp, ])
            curve[in.grp] <- stats::predict(mf)
            coef.poly <- polynom::polynomial(stats::coefficients(mf))
            deriv.poly <- stats::deriv(coef.poly)
            sm.deriv[in.grp] <- stats::predict(deriv.poly, data[in.grp, "x"])
            if (params$method != "lm") {
              message("Fitting a linear regression as n < 4")
            }
          } else if (params$method == "spline") {
            sm.spline <- stats::smooth.spline(data[in.grp, "x"], data[in.grp, "y"])
            curve[in.grp] <- stats::predict(sm.spline, x = data[in.grp, "x"], deriv = 0)$y
            sm.deriv[in.grp] <- stats::predict(sm.spline, x = data[in.grp, "x"], deriv = 1)$y
          }
        }
      } else {
        stop("Method \"", params$method, "\"not recognized")
      }

      # compute x and y nudge for each point
      # By changing the sign we ensure consistent positions in opposite slopes
      angle.rotation <- ifelse(sm.deriv > 0, -0.5 * pi, +0.5 * pi)
      # scaling is needed to compute the angle on the plot
      angle <- atan2(sm.deriv * xy.range.ratio, 1) + angle.rotation

      if (params$x.reorder) {
        x_nudge <- rep_len(params$x, nrow(data))[order(order(data$x))] *
          cos(angle) * ifelse(sm.deriv > 0, -1, +1)
      } else {
        x_nudge <- rep_len(params$x, nrow(data)) *
          cos(angle) * ifelse(sm.deriv > 0, -1, +1)
      }
      if (params$y.reorder) {
        y_nudge <- rep_len(params$y, nrow(data))[order(order(data$y))] *
          sin(angle) * ifelse(sm.deriv > 0, -1, +1)
      } else {
        y_nudge <- rep_len(params$y, nrow(data)) *
          sin(angle) * ifelse(sm.deriv > 0, -1, +1)
      }

      if (params$direction == "split") {
        # sign depends on position relative to the line or curve
        x_nudge <- ifelse(data$y >= curve, x_nudge, -x_nudge)
        y_nudge <- ifelse(data$y >= curve, y_nudge, -y_nudge)
      } else if (params$direction != "none") {
        warning("Ignoring unrecognized direction \"", params$direction, "\".")
      }

      if (params$line_nudge > 1) {
        # nudging further away from line or curve than from points
        adj_y_nudge <- y_nudge * params$line_nudge - (data$y - curve)
        adj_x_nudge <- x_nudge * adj_y_nudge / y_nudge
        y_nudge <- ifelse(sign(y_nudge) == sign(adj_y_nudge) &
                            abs(y_nudge) < abs(adj_y_nudge),
                          adj_y_nudge,
                          y_nudge)
        x_nudge <- ifelse(sign(y_nudge) == sign(adj_y_nudge) &
                            abs(x_nudge) >= abs(adj_x_nudge),
                          adj_x_nudge,
                          x_nudge)
      }
      # transform only the dimensions for which new coordinates exist
      if (any(x_nudge != 0)) {
        if (any(y_nudge != 0)) {
          data <- ggplot2::transform_position(data, function(x) x + x_nudge, function(y) y + y_nudge)
        } else {
          data <- ggplot2::transform_position(data, function(x) x + x_nudge, NULL)
        }
      } else if (any(y_nudge != 0)) {
        data <- ggplot2::transform_position(data, NULL, function(y) y + y_nudge)
      }
      if (params$kept.origin == "original") {
        data$x_orig <- x_orig
        data$y_orig <- y_orig
      }

      data
    }

  )

#' Annotations supporting NPC
#'
#' A revised version of \code{annotate()} from package 'ggplot2' adding support
#' for \code{npcx} and \code{npcy} position aesthetics, allowing use of the
#' geometries defined in the current package such as \code{geom_text_npc()}. It
#' also has a parameter \code{label} that directly accepts data frames, ggplots
#' and grobs as arguments in addition to objects of atomic classes like
#' character. When package 'ggpmisc' is loaded this definition of
#' \code{annotate()} overrides that in package 'ggplot2'.
#'
#' @param geom character Name of geom to use for annotation.
#' @param x,y,xmin,ymin,xmax,ymax,xend,yend,npcx,npcy	numeric Positioning
#'   aesthetics - you must specify at least one of these.
#' @param label character, data.frame, ggplot or grob.
#' @param ...	Other named arguments passed on to \code{layer()}. These are often
#'   aesthetics, used to set an aesthetic to a fixed value, like color = "red"
#'   or size = 3. They may also be parameters to the paired geom/stat.
#' @param na.rm	logical If \code{FALSE}, the default, missing values are removed
#'   with a warning. If TRUE, missing values are silently removed.
#'
#' @details Note that all position aesthetics are scaled (i.e., they will
#'   expand the limits of the plot so they are visible), but all other
#'   aesthetics are set. This means that layers created with this function will
#'   never affect the legend.
#'
#' @note To use the original definition of \code{annotate()} after loading
#'   package 'ggpmisc', use \code{ggplot2::annotate()}.
#'
#' @return A plot layer instance.
#'
#' @export
#'
#' @examples
#'
#' p <- ggplot(mtcars, aes(x = wt, y = mpg)) + geom_point()
#'
#' # Works as ggplot2::annotate()
#' p + annotate("text", x = 5, y = 32, label = "Some text")
#' p + annotate("label", x = c(2, 5), y = c(15, 32),
#'              label = c("A", "B"))
#' p + annotate("table", x = 5, y = 30,
#'              label = data.frame(A = 1:2, B = letters[1:2]))
#' p + annotate("plot", x = 5.5, y = 34,
#'              label = p + theme_bw(9))
#' p + annotate("rect", xmin = 3, xmax = 4.2, ymin = 12, ymax = 21, alpha = .2)
#' p + annotate("segment", x = 2.5, xend = 4, y = 15, yend = 25, colour = "blue")
#' p + annotate("pointrange", x = 3.5, y = 20, ymin = 12, ymax = 28,
#'   colour = "red", size = 1.5)
#'
#' # But ggpmisc::annotate() also works with npcx and npcy pseudo-aesthetics
#' p + annotate("label_npc", npcx = c(0.1, 0.9), npcy = c(0.1, 0.9),
#'              label = c("A", "B"))
#' p + annotate("label_npc", npcx = 0.9, npcy = c(0.1, 0.9),
#'              label = c("A", "B"))
#'
#' p + annotate("text_npc", npcx = 0.9, npcy = 0.9, label = "Some text")
#' p + annotate("text_npc", npcx = "right", npcy = "top", label = "Some text")
#'
#' p + annotate("table_npc", npcx = 0.9, npcy = 0.9,
#'              label = data.frame(A = 1:2, B = letters[1:2]))
#'
#' p + annotate("plot_npc", npcx = 1, npcy = 1,
#'              label = p + theme_bw(9))
#' p + annotate("plot_npc", npcx = c(0, 1), npcy = c(0, 1),
#'              label = list(p + theme_bw(9), p + theme_grey(9)),
#'              vp.width = 0.3, vp.height = 0.4)
#'
annotate <-
  function (geom, x = NULL, y = NULL, xmin = NULL, xmax = NULL,
            ymin = NULL, ymax = NULL, xend = NULL, yend = NULL,
            npcx = NULL, npcy = NULL, label = NULL, ...,
            na.rm = FALSE)
  {
    # convert to lists for 'ggpp' geoms
    if (inherits(label, what = c("data.frame", "gg", "grob"))) {
      label <- list(label)
    }

    position <- compact(list(x = x,
                             xmin = xmin,
                             xmax = xmax,
                             xend = xend,
                             y = y,
                             ymin = ymin,
                             ymax = ymax,
                             yend = yend,
                             npcx = npcx, # for 'ggpp' geoms
                             npcy = npcy, # for 'ggpp' geoms
                             label = label))

    aesthetics <- c(position, list(...))

    # Check that all aesthetic have compatible lengths
    lengths <- lengths(aesthetics)
    n <- unique(lengths)

    # if there is more than one unique length, ignore constants
    if (length(n) > 1L) {
      n <- setdiff(n, 1L)
    }

    # if there is still more than one unique length, we error out
    if (length(n) > 1L) {
      bad <- lengths != 1L
      details <- paste(names(aesthetics)[bad], " (", lengths[bad],
                       ")", sep = "", collapse = ", ")
      rlang::abort(glue::glue("Unequal parameter lengths: {details}"))
    }

    data <- new_data_frame(position, n = n)

    ggplot2::layer(geom = geom,
                   params = list(na.rm = na.rm, ...),
                   stat = StatIdentity,
                   position = PositionIdentity,
                   data = data,
                   mapping = aes_all(names(data)),
                   inherit.aes = FALSE,
                   show.legend = FALSE)
  }

context("annotate")
# tests copied from 'ggplot2' to ensure that pass through is working as expected

test_that("dates in segment annotation work", {
  dt <- structure(list(month = structure(c(1364774400, 1377993600),
                                         class = c("POSIXct", "POSIXt"), tzone = "UTC"), total = c(-10.3,
                                                                                                   11.7)), .Names = c("month", "total"), row.names = c(NA, -2L), class =
                    "data.frame")

  p <- ggplot(dt, aes(month, total)) +
    geom_point() +
    annotate("segment",
             x = as.POSIXct("2013-04-01"),
             xend = as.POSIXct("2013-07-01"),
             y = -10,
             yend = 10
    )

  expect_true(all(c("xend", "yend") %in% names(layer_data(p, 2))))
})


test_that("segment annotations transform with scales", {
  # Line should match data points
  df <- tibble::tibble(x = c(1, 10), y = c(10, 1))
  plot <- ggplot(df, aes(x, y)) +
    geom_point() +
    annotate("segment", x = 1, y = 10, xend = 10, yend = 1, colour = "red") +
    scale_y_reverse(NULL, breaks = NULL) +
    scale_x_continuous(NULL, breaks = NULL)

  vdiffr::expect_doppelganger("line matches points", plot)
})

context("ggpp_annotate")

p <- ggplot(mtcars, aes(x = wt, y = mpg)) + geom_point()

test_that("ggpp::annotate handles input of uneven lengths", {
  expect_error(p + annotate("text", label = "a", x = 1:3, y = 2:1))
  expect_error(p + annotate("text", label = c("a", "b"), x = 1:3, y = 3:1))
  expect_silent(p + annotate("text", label = "a", x = 1, y = 30))
  expect_silent(p + annotate("text", label = "a", x = 1:3, y = 3:1))
  expect_silent(p + annotate("text", label = c("a", "b"), x = 1:2, y = 1))
})

test_that("ggpp::annotate works with npc pseudo-aesthetics", {

vdiffr::expect_doppelganger("annotate_label_npc1",
                            p + annotate("label_npc", npcx = c(0.1, 0.9), npcy = c(0.1, 0.9),
                                         label = c("A", "B"))
                            )

vdiffr::expect_doppelganger("annotate_text_npc1",
                            p + annotate("text_npc", npcx = 0.9, npcy = 0.9, label = "Some text")
)

vdiffr::expect_doppelganger("annotate_text_npc2",
                            p + annotate("text_npc", npcx = "right", npcy = "top", label = "Some text")
)

vdiffr::expect_doppelganger("annotate_table_npc1",
                            p + annotate("table_npc", npcx = 0.9, npcy = 0.9,
                                         label = data.frame(A = 1:2, B = letters[1:2]))
)

vdiffr::expect_doppelganger("annotate_plot_npc1",
                            p + annotate("plot_npc", npcx = 1, npcy = 1,
                                         label = p + theme_bw(9))
)

vdiffr::expect_doppelganger("annotate_plot_npc2",
                            p + annotate("plot_npc", npcx = c(0, 1), npcy = c(0, 1),
                                         label = list(p + theme_bw(9), p + theme_grey(9)),
                                         vp.width = 0.3, vp.height = 0.4)
)

})

test_that("ggpp::annotate works with x and y aesthetics", {

vdiffr::expect_doppelganger("annotate_text",
                            p + annotate("text", x = c(2, 5), y = c(15, 32),
                                         label = c("A", "B"))
)

vdiffr::expect_doppelganger("annotate_label",
                            p + annotate("label", x = c(2, 5), y = c(15, 32),
                                         label = c("A", "B"))
)

vdiffr::expect_doppelganger("annotate_table",
                            p + annotate("table", x = 5, y = 30,
                                         label = data.frame(A = 1:2, B = letters[1:2]))
)

vdiffr::expect_doppelganger("annotate_table_colour",
                            p + annotate("table", x = 5, y = 30, colour = "blue",
                                         label = data.frame(A = 1:2, B = letters[1:2]))
)

vdiffr::expect_doppelganger("annotate_table_alpha",
                            p + annotate("table", x = 5, y = 30, alpha = 0.25,
                                         label = data.frame(A = 1:2, B = letters[1:2]))
)

vdiffr::expect_doppelganger("annotate_table_colour_target_rules",
                            p + annotate("table", x = 5, y = 30,
                                         colour = "red", colour.target = "table.rules",
                                         label = data.frame(A = 1:2, B = letters[1:2]))
)

vdiffr::expect_doppelganger("annotate_table_colour_default_colour",
                            p + annotate("table", x = 5, y = 30,
                                         default.colour = "blue",
                                         colour = "red", colour.target = "table.rules",
                                         label = data.frame(A = 1:2, B = letters[1:2]))
)

vdiffr::expect_doppelganger("annotate_plot",
                            p + annotate("plot", x = 5.5, y = 34,
                                         label = p + theme_bw(9))
)

})

#' Key glyphs for legends
#'
#' Each geom has an associated function that draws the key when the geom needs
#' to be displayed in a legend. These functions are called `draw_key_*()`, where
#' `*` stands for the name of the respective key glyph. The key glyphs can be
#' customized for individual geoms by providing a geom with the `key_glyph`
#' argument (see [`layer()`] or examples below.)
#'
#' @return A grid grob.
#' @param data A single row data frame containing the scaled aesthetics to
#'   display in this key
#' @param params A list of additional parameters supplied to the geom.
#' @param size Width and height of key in mm.
#' @examples
#' p <- ggplot(economics, aes(date, psavert, color = "savings rate"))
#' # key glyphs can be specified by their name
#' p + geom_line(key_glyph = "timeseries")
#'
#' # key glyphs can be specified via their drawing function
#' p + geom_line(key_glyph = draw_key_rect)
#'
#' @keywords internal
#' @name ggpp_draw_key
NULL

#' @export
#' @rdname ggpp_draw_key
draw_key_text_s <- function(data, params, size) {
  data  <- replace_null(unclass(data), label = "a", angle = 0)
  hjust <- ifelse(data$hjust %in% c("left", "middle", "right"),
                  compute_just(data$hjust %||% 0.5),
                  0.5)
  vjust <- ifelse(data$vjust %in% c("left", "middle", "right"),
                  compute_just(data$vjust %||% 0.5),
                  0.5)
  just  <- rotate_just(data$angle, hjust, vjust)

  colour2text <- any((params$colour.target %||% "text") %in% c("text", "all"))
  alpha2text <- any((params$alpha.target %||% "text") %in% c("text", "all"))
  text.colour <- ifelse(colour2text,
                        data$colour %||% params$default.colour  %||% "black",
                        params$default.colour %||% "black")
  colour.lacks.alpha <- ifelse(nrow(grDevices::col2rgb(text.colour)) == 3,
                               rep(TRUE, length(text.colour)),
                               is.na(grDevices::col2rgb(text.colour)[4, ]))
  colour.alpha <- ifelse(is.na(data$alpha),
                         ifelse(colour.lacks.alpha,
                                params$default.alpha %||% 1,
                                params$default.alpha),
                         data$alpha)
  if (alpha2text) {
    text.colour <- ggplot2::alpha(text.colour, colour.alpha)
  }

  grob  <- titleGrob(
    data$label,
    x = unit(just$hjust, "npc"), y = unit(just$vjust, "npc"),
    angle = data$angle,
    hjust = hjust,
    vjust = vjust,
    gp = gpar(
      col = text.colour,
      fontfamily = data$family   %||% "",
      fontface   = data$fontface %||% 1,
      fontsize   = (data$size %||% 3.88) * .pt
    ),
    margin = margin(0.1, 0.1, 0.1, 0.1, unit = "lines"),
    margin_x = TRUE, margin_y = TRUE
  )
  attr(grob, "width")  <- convertWidth(grobWidth(grob),   "cm", valueOnly = TRUE)
  attr(grob, "height") <- convertHeight(grobHeight(grob), "cm", valueOnly = TRUE)
  grob
}

#' @export
#' @rdname ggpp_draw_key
draw_key_label_s <- function(data, params, size) {
  data <- replace_null(unclass(data), label = "a", angle = 0)
  params$label.size <- params$label.size %||% 0.25
  hjust <- ifelse(data$hjust %in% c("left", "middle", "right"),
                  compute_just(data$hjust %||% 0.5),
                  0.5)
  vjust <- ifelse(data$vjust %in% c("left", "middle", "right"),
                  compute_just(data$vjust %||% 0.5),
                  0.5)
  just  <- rotate_just(data$angle, hjust, vjust)
  padding <- rep(params$label.padding %||% unit(0.25, "lines"), length.out = 4)
  descent <- font_descent(
    family = data$family %||% "",
    face = data$fontface %||% 1,
    size = data$size %||% 3.88
  )

  colour2text <- any((params$colour.target %||% "text") %in% c("text", "all"))
  alpha2text <- any((params$alpha.target %||% "all") %in% c("text", "all"))
  text.colour <- ifelse(colour2text,
                        data$colour %||% params$default.colour  %||% "black",
                        params$default.colour %||% "black")
  colour.lacks.alpha <- ifelse(nrow(grDevices::col2rgb(text.colour)) == 3,
                               rep(TRUE, length(text.colour)),
                               is.na(grDevices::col2rgb(text.colour)[4, ]))
  colour.alpha <- ifelse(is.na(data$alpha),
                         ifelse(colour.lacks.alpha,
                                params$default.alpha %||% 1,
                                params$default.alpha),
                         data$alpha)
  if (alpha2text) {
    text.colour <- ggplot2::alpha(text.colour, colour.alpha)
  }

  grob <- labelGrob(
    data$label,
    x = unit(just$hjust, "npc"),
    y = unit(just$vjust, "npc") + descent,
    angle = data$angle,
    just = c(hjust, vjust),
    padding = padding,
    r = params$label.r %||% unit(0.15, "lines"),
    text.gp = gpar(
      col = text.colour,
      fontfamily = data$family   %||% "",
      fontface   = data$fontface %||% 1,
      fontsize   = (data$size %||% 3.88) * .pt
    ),
    rect.gp = gpar(
      col  = if (isTRUE(all.equal(params$label.size, 0))) {
        NA
      } else {
        colour2box <- any((params$colour.target %||% "text") %in% c("box", "box.line", "all"))
        alpha2box <- any((params$alpha.target %||% "all") %in% c("box", "box.line", "all"))
        box.colour <- ifelse(colour2box,
                             data$colour %||% params$default.colour  %||% "black",
                             params$default.colour %||% "black")
        colour.lacks.alpha <- ifelse(nrow(grDevices::col2rgb(box.colour)) == 3,
                                     rep(TRUE, length(box.colour)),
                                     is.na(grDevices::col2rgb(box.colour)[4, ]))
        colour.alpha <- ifelse(is.na(data$alpha),
                               ifelse(colour.lacks.alpha,
                                      params$default.alpha %||% 1,
                                      params$default.alpha),
                               data$alpha)
        if (alpha2box) {
          ggplot2::alpha(box.colour, colour.alpha)
        } else {
          box.colour
        }

      },
      fill =
        {
          alpha2fill <- any((params$alpha.target %||% "all") %in% c("box", "box.fill", "all"))
          box.fill <- data$fill %||% params$default.fill %||% "white"
          fill.lacks.alpha <- ifelse(nrow(grDevices::col2rgb(box.fill)) == 3,
                                     rep(TRUE, length(box.fill)),
                                     is.na(grDevices::col2rgb(box.fill)[4, ]))
          fill.alpha <- ifelse(is.na(data$alpha),
                               ifelse(fill.lacks.alpha,
                                      params$default.alpha %||% 1,
                                      params$default.alpha),
                               data$alpha)
          if (alpha2fill) {
            box.fill <- alpha(box.fill, fill.alpha)
          }

          box.fill
        },
      lwd  = (data$linewidth %||% 0.25) * ggplot2::.stroke,
      lty = data$linetype %||% "solid"
    )
  )
  angle  <- deg2rad(data$angle %||% 0)
  text   <- grob$children[[2]]
  width  <- convertWidth(grobWidth(text),   "cm", valueOnly = TRUE)
  height <- convertHeight(grobHeight(text), "cm", valueOnly = TRUE)
  x <- c(0, 0, width, width)
  y <- c(0, height, height, 0)
  attr(grob, "width")  <- diff(range(x * cos(angle) - y * sin(angle)))
  attr(grob, "height") <- diff(range(x * sin(angle) + y * cos(angle)))
  grob
}

context("draw_key_label_s")

test_that("Returns expected value", {

  test.data <- data.frame(x = 1,
                          y = 5,
                          label = "a",
                          angle = 0,
                          colour = "red",
                          fill = "white",
                          alpha = 0.5,
                          vjust = 0.5,
                          hjust = 0.5)

  test.params.01 <- list(colour.target = "all",
                      alpha.target = "all",
                      default.colour = "black",
                      default.alpha = 1
                      )

  set.seed(1)
  obj.01 <- draw_key_label_s(data = test.data,
                             params = test.params.01,
                             size = 4)
  # expect_known_value(unname(obj.01),
  #                    "draw-key-label-s-01")
  expect_is(obj.01, "gTree")
  expect_length(obj.01, 5)
  expect_named(obj.01, c("name", "gp", "vp", "children", "childrenOrder"))


  test.params.02 <- list(colour.target = "text",
                      alpha.target = "none",
                      default.colour = "grey30",
                      default.alpha = 1
  )

  set.seed(1)
  obj.02 <- draw_key_label_s(data = test.data,
                             params = test.params.02,
                             size = 4)
  # expect_known_value(unname(obj.02),
  #                    "draw-key-label-s-02")
  expect_is(obj.02, "gTree")
  expect_length(obj.02, 5)
  expect_named(obj.02, c("name", "gp", "vp", "children", "childrenOrder"))

  test.params.03 <- list(colour.target = "text",
                         alpha.target = "none",
                         default.colour = "grey30",
                         default.alpha = 0
  )

  set.seed(1)
  obj.03 <- draw_key_label_s(data = test.data,
                             params = test.params.03,
                             size = 4)
  # expect_known_value(unname(obj.03),
  #                    "draw-key-label-s-03")
  expect_is(obj.03, "gTree")
  expect_length(obj.03, 5)
  expect_named(obj.03, c("name", "gp", "vp", "children", "childrenOrder"))

  test.params.04 <- list(colour.target = "box",
                         alpha.target = "none",
                         default.colour = "grey30",
                         default.alpha = 0
  )

  set.seed(1)
  obj.04 <- draw_key_label_s(data = test.data,
                             params = test.params.04,
                             size = 4)
  # expect_known_value(unname(obj.04),
  #                    "draw-key-label-s-04")
  expect_is(obj.04, "gTree")
  expect_length(obj.04, 5)
  expect_named(obj.04, c("name", "gp", "vp", "children", "childrenOrder"))

  test.params.05 <- list(colour.target = c("segment", "text"),
                         alpha.target = "none",
                         default.colour = "grey30",
                         default.alpha = 1
  )

  set.seed(1)
  obj.05 <- draw_key_label_s(data = test.data,
                             params = test.params.05,
                             size = 4)
  # expect_known_value(unname(obj.05),
  #                    "draw-key-label-s-05")
  expect_is(obj.05, "gTree")
  expect_length(obj.05, 5)
  expect_named(obj.05, c("name", "gp", "vp", "children", "childrenOrder"))



})

context("draw_key_text_s")

test_that("Returns expected value", {

  test.data <- data.frame(x = 1,
                          y = 5,
                          label = "a",
                          angle = 0,
                          colour = "red",
                          alpha = 0.5,
                          vjust = 0.5,
                          hjust = 0.5)

  test.params.01 <- list(colour.target = "all",
                         alpha.target = "all",
                         default.colour = "black",
                         default.alpha = 1
  )

  # names of grob members seem to change randomly from run to run
  # set.seed() and unname() do not seem to help!
  set.seed(1)
  obj.01 <- draw_key_text_s(data = test.data,
                            params = test.params.01,
                            size = 4)
  # expect_known_value(unname(obj.01),
  #                    "draw-key-text-s-01")
  expect_is(obj.01, "titleGrob")
  expect_length(obj.01, 7)
  expect_named(obj.01, c("widths", "heights", "name", "gp", "vp", "children", "childrenOrder"))

  test.params.02 <- list(colour.target = "text",
                      alpha.target = "none",
                      default.colour = "grey30",
                      default.alpha = 1
  )

  set.seed(1)
  obj.02 <- draw_key_text_s(data = test.data,
                            params = test.params.02,
                            size = 4)
  # expect_known_value(unname(obj.02),
  #                    "draw-key-text-s-02")
  expect_is(obj.02, "titleGrob")
  expect_length(obj.02, 7)
  expect_named(obj.02, c("widths", "heights", "name", "gp", "vp", "children", "childrenOrder"))

  test.params.03 <- list(colour.target = "text",
                         alpha.target = "none",
                         default.colour = "grey30",
                         default.alpha = 0
  )

  set.seed(1)
  obj.03 <- draw_key_text_s(data = test.data,
                            params = test.params.03,
                            size = 4)
  # expect_known_value(unname(obj.03),
  #                    "draw-key-text-s-03")
  expect_is(obj.03, "titleGrob")
  expect_length(obj.03, 7)
  expect_named(obj.03, c("widths", "heights", "name", "gp", "vp", "children", "childrenOrder"))

})

#' Combined positions dodge and nudge
#'
#' \code{position_dodgenudge()} combines into one function the action of
#' \code{\link[ggplot2]{position_dodge}} and
#' \code{\link[ggplot2]{position_nudge}} and \code{position_dodge2nudge()}
#' combines into one function the action of
#' \code{\link[ggplot2]{position_dodge2}} and
#' \code{\link[ggplot2]{position_nudge}}. They are useful when labelling plots
#' such as grouped bars, columns, etc. and when adding dodged text labels linked
#' to observations plotted without dodge. It can replace other position
#' functions as they are backwards compatible. Like all other position functions
#' in 'ggpp' and 'ggrepel' they preserve the initial position to allow drawing
#' of segments or arrow linking the original position to the displaced one.
#'
#' @details The applied dodge is identical to that by
#'   \code{\link[ggplot2]{position_dodge}} or
#'   \code{\link[ggplot2]{position_dodge2}} while nudging is similar to that by
#'   \code{\link[ggplot2]{position_nudge}}.
#'
#' There are two possible uses for these functions. First they can be used to
#' label dodged bars or boxplots. In this case, it is mandatory to use the same
#' argument to \code{width} when passing \code{position_dodge()} to
#' \code{geom_col()} and \code{position_dodgenudge()} to \code{geom_text()} or
#' \code{geom_label()} or their repulsive equivalents. Otherwise the arrows or
#' segments will fail to connect to the labels. In other words dodging is
#' computed twice. Dodge is identical to that obtained with the same arguments
#' in \code{\link[ggplot2]{position_dodge}} as \code{position_dodgenudge()} last
#' function simply calls the same code from package 'ggplot2' ahead of applying
#' nudging.
#'
#' The second use is to dodge labels to be connected to elements that have not
#' been dodged.
#'
#' When applying dodging, the return of original positions instead of the dodged
#' ones is achieved by passing \code{origin = "original"} instead of the default
#' of \code{origin = "dodged"}.
#'
#' @family position adjustments
#'
#' @param width Dodging width, when different to the width of the individual
#'   elements. This is useful when you want to align narrow geoms with wider
#'   geoms. See the examples.
#' @param preserve Should dodging preserve the total width of all elements at a
#'   position, or the width of a single element?.
#' @param reverse logical If TRUE, will reverse the default dodging order.
#' @param padding Padding between elements at the same position. Elements are
#'   shrunk by this proportion to allow space between them. Defaults to 0.1.
#' @param reverse If TRUE, will reverse the default stacking order. This is
#'   useful if you're rotating both the plot and legend.
#' @param x,y Amount of vertical and horizontal distance to move. A numeric
#'   vector of length 1, or of the same length as rows there are in `data`,
#'   with nudge values in data rows order.
#' @param direction One of \code{"none"}, \code{"split"}, \code{"split.x"} or
#'   \code{"split.y"}. A value of \code{"none"} replicates the behavior of
#'   \code{\link[ggplot2]{position_nudge}}. At the moment \code{"split"} changes
#'   the sign of the nudge at zero, which is suitable for column plots with
#'   negative slices.
#' @param kept.origin One of \code{"original"}, \code{"dodged"} or
#'   \code{"none"}.
#'
#' @seealso \code{\link[ggplot2]{position_nudge}},
#' \code{\link[ggrepel]{position_nudge_repel}}.
#'
#' @return A \code{"Position"} object.
#'
#' @export
#'
#' @author Michał Krassowski, edited by Pedro J. Aphalo.
#'
#' @source \url{https://github.com/slowkow/ggrepel/issues/161}.
#'
#' @examples
#' df <- data.frame(x1 = c(1, 2, 1, 3, -1),
#'                  x2 = c("a", "a", "b", "b", "b"),
#'                  grp = c("some long name", "other name", "some name",
#'                          "another name", "some long name"))
#'
#' # Add labels to a horizontal column plot (stacked by default)
#' ggplot(data = df, aes(x1, x2, group = grp)) +
#'   geom_col(aes(fill = grp), width = 0.8,
#'            position = position_dodge()) +
#'   geom_vline(xintercept = 0) +
#'   geom_text(
#'     aes(label = grp),
#'     position = position_dodgenudge(x = 0.09, direction = "split", width = 0.8),
#'     angle = 90, size = 3) +
#'   theme(legend.position = "none")
#'
#' ggplot(data = df, aes(x2, x1, group = grp)) +
#'   geom_col(aes(fill = grp), width = 0.75,
#'            position = position_dodge(width = 0.75)) +
#'   geom_vline(xintercept = 0) +
#'   geom_text(aes(label = grp),
#'             position = position_dodgenudge(y = 0.1,
#'                                            direction = "split",
#'                                            width = 0.75),
#'             size = 3) +
#'   theme(legend.position = "none")
#'
position_dodgenudge <-
  function(width = 1,
           preserve = c("total", "single"),
           reverse = FALSE,
           x = 0,
           y = 0,
           direction = c("none", "split", "split.x", "split.y", "center"),
           kept.origin = c("dodged", "original", "none")) {

    preserve <- rlang::arg_match(preserve)
    direction <- rlang::arg_match(direction)
    kept.origin <- rlang::arg_match(kept.origin)

    ggplot2::ggproto(NULL, PositionDodgeAndNudge,
                     x = x,
                     y = y,
                     .fun_x = switch(direction,
                                     none = function(x) {1},
                                     split = sign,
                                     split.y = function(x) {1},
                                     split.x = sign,
                                     center = sign),
                     .fun_y = switch(direction,
                                     none = function(x) {1},
                                     split = sign,
                                     split.x = function(x) {1},
                                     split.y = sign,
                                     center = sign),
                     kept.origin = kept.origin,
                     width = width,
                     preserve = preserve,
                     reverse = reverse
    )
  }

#' @rdname ggpp-ggproto
#' @format NULL
#' @usage NULL
#' @noRd
PositionDodgeAndNudge <-
  ggplot2::ggproto("PositionDodgeAndNudge", ggplot2::PositionDodge,
                   x = 0,
                   y = 0,

                   setup_params = function(self, data) {
                     c(
                       list(nudge_x = self$x,
                            nudge_y = self$y,
                            .fun_x = self$.fun_x,
                            .fun_y = self$.fun_y,
                            x.reorder = !is.null(self$x) &&
                              length(self$x) > 1 && length(self$x) < nrow(data),
                            y.reorder = !is.null(self$y) &&
                              length(self$y) > 1 && length(self$y) < nrow(data),
                            kept.origin = self$kept.origin),
                       ggplot2::ggproto_parent(ggplot2::PositionDodge, self)$setup_params(data)
                     )
                   },

                   compute_layer = function(self, data, params, layout) {

                     if (length(params$nudge_x) > nrow(data)) {
                       warning("Argument 'x' longer than data: some values dropped!")
                     }
                     if (length(params$nudge_y) > nrow(data)) {
                       warning("Argument 'y' longer than data: some values dropped!")
                     }
                     x_orig <- data$x
                     y_orig <- data$y
                     # operate on the dodged positions
                     data = ggplot2::ggproto_parent(ggplot2::PositionDodge, self)$compute_layer(data, params, layout)

                     x_dodged <- data$x
                     y_dodged <- data$y

                     if (params$x.reorder) {
                       params$nudge_x <- rep_len(params$nudge_x, nrow(data))[order(order(data$x))]
                     } else {
                       params$nudge_x <- rep_len(params$nudge_x, nrow(data))
                     }
                     if (params$y.reorder) {
                       params$nudge_y <- rep_len(params$nudge_y, nrow(data))[order(order(data$y))]
                     } else {
                       params$nudge_y <- rep_len(params$nudge_y, nrow(data))
                     }

                     # transform only the dimensions for which non-zero nudging is requested
                     if (any(params$nudge_x != 0)) {
                       if (any(params$nudge_y != 0)) {
                         data <- ggplot2::transform_position(data,
                                                             function(x) x + params$nudge_x * params$.fun_x(x),
                                                             function(y) y + params$nudge_y * params$.fun_y(y))
                       } else {
                         data <- ggplot2::transform_position(data,
                                                             function(x) x + params$nudge_x * params$.fun_x(x),
                                                             NULL)
                       }
                     } else if (any(params$nudge_y != 0)) {
                       data <- ggplot2::transform_position(data,
                                                           function(x) x,
                                                           function(y) y + params$nudge_y * params$.fun_y(y))
                     }
                     # add original position
                     if (params$kept.origin == "dodged") {
                       data$x_orig <- x_dodged
                       data$y_orig <- y_dodged
                     } else if (params$kept.origin == "original") {
                       data$x_orig <- x_orig
                       data$y_orig <- y_orig
                     }

                     data
                   },

                   compute_panel = function(self, data, params, scales) {
                     ggplot2::ggproto_parent(PositionDodge, self)$compute_panel(data, params, scales)
                   }
  )

#' @rdname position_dodgenudge
#'
#' @export
#'
position_dodge_keep <- function(width = 1,
                                preserve = c("total", "single"),
                                reverse = FALSE,
                                kept.origin = "original") {
  position_dodgenudge(width = width,
                      preserve = preserve,
                      reverse = reverse,
                      x = 0,
                      y = 0,
                      direction = "none",
                      kept.origin = kept.origin)
}

context("position_dodge_keep")

test_that("test expected default arguments", {
  position <- position_dodge_keep()
  expect_identical(position$width, 1)
  expect_identical(position$preserve, "total")
  expect_identical(position$x, 0)
  expect_identical(position$y, 0)
  expect_identical(position$kept.origin, "original")
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
  skip_if(condition = utils::packageVersion("ggplot2") < "4.0.0",
          message = "'ggplot2' < 4.0.0")
  expect_false(position$reverse)
  expect_identical(position$orientation, "x")
  expect_identical(position$required_aes, character(0))
  expect_type(position$aesthetics, "closure")
  expect_type(position$use_defaults, "closure")
})

test_that("test expected user arguments", {
  position <-
    position_dodge_keep(preserve = "single", width = 0.5, kept.origin = "dodged")
  expect_identical(position$width, 0.5)
  expect_identical(position$preserve, "single")
  expect_identical(position$x, 0)
  expect_identical(position$y, 0)
  expect_identical(position$kept.origin, "dodged")
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
  skip_if(condition = utils::packageVersion("ggplot2") < "4.0.0",
          message = "'ggplot2' < 4.0.0")
  expect_false(position$reverse)
  expect_identical(position$orientation, "x")
  expect_identical(position$required_aes, character(0))
  expect_type(position$aesthetics, "closure")
  expect_type(position$use_defaults, "closure")
})

test_that("test expected user argument 'reverse'", {
  position <- position_dodge_keep(reverse = TRUE, width = 0.5)
  expect_identical(position$width, 0.5)
  expect_identical(position$preserve, "total")
  expect_identical(position$x, 0)
  expect_identical(position$y, 0)
  expect_identical(position$kept.origin, "original")
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
  skip_if(condition = utils::packageVersion("ggplot2") < "4.0.0",
          message = "'ggplot2' < 4.0.0")
  expect_true(position$reverse)
  expect_identical(position$orientation, "x")
  expect_identical(position$required_aes, character(0))
  expect_type(position$aesthetics, "closure")
  expect_type(position$use_defaults, "closure")
})

test_that("test expected user argument 'orientation'", {
  skip("Not yet implemented")
  position <- position_dodge_keep(orientation = "y", width = 0.5)
  expect_identical(position$width, 0.5)
  expect_identical(position$preserve, "total")
  expect_identical(position$x, 0)
  expect_identical(position$y, 0)
  expect_identical(position$kept.origin, "original")
  expect_type(position$compute_panel, "closure")
  expect_type(position$compute_layer, "closure")
  expect_type(position$setup_data, "closure")
  expect_type(position$setup_params, "closure")
  skip_if(condition = utils::packageVersion("ggplot2") < "4.0.0",
          message = "'ggplot2' < 4.0.0")
  expect_true(position$reverse)
  expect_identical(position$orientation, "y")
  expect_identical(position$required_aes, character(0))
  expect_type(position$aesthetics, "closure")
  expect_type(position$use_defaults, "closure")
})

#' Chose between dark and light color
#'
#' Chose between a pair of contrasting dark and light colors based on a weighted
#' mean of RGB channels of a color. This function implements a simple approach
#' to the choice for a color of a plot element to ensure it is visible against a
#' background color.
#'
#' @param colors A vector of color definitions or color names in the background.
#' @param threshold numeric A value of luminance in [0..1] indicating the switch
#'   point between dark and light background.
#' @param dark.color,light.color A color definition or color name to return as
#'   dark and light colors to contrast light and dark backgrounds respectively.
#'
#' @details The switch between dark and light color is based on a quick and
#'   dirty approximation of the luminance of colors computed from RGB values.
#'   This easily computed approximation seems to work well enough. The default
#'   threshold chosen for a switch between black and white may need to be
#'   adjusted for other pairs of colors. Graphic devices can differ in the color
#'   spaces they support, but this is unlikely to affect the choice between
#'   black and white or other pairs of colors with large differences in
#'   luminance.
#'
#' @note The current implementation of \code{dark_or_light()} ignores
#'   \code{alpha}, the transparency component, of all its arguments.
#'
#' @export
#'
#' @examples
#'
#' dark_or_light("yellow")
#' dark_or_light("darkblue")
#' dark_or_light(c("darkblue", "yellow", "red"))
#' dark_or_light("#FFFFFF")
#' dark_or_light("#FFFFFF", dark.color = "darkblue", light.color = "lightgrey")
#' dark_or_light("#000000", dark.color = "darkblue", light.color = "lightgrey")
#'
dark_or_light <- function(colors,
                          threshold = 0.45,
                          dark.color = "black",
                          light.color = "white")
{
  if (!length(colors))
    return(character())
  stopifnot("'threshold' should have length = 1" = length(threshold) == 1L,
            "'threshold' should be within 0..1" = threshold >= 0 && threshold <= 1)
  threshold <- trunc(threshold * 255)
  # approximate luminance in 0..255
  lum <- sapply(colors, function(x) {
    y <- grDevices::col2rgb(x)
    sum(y * c(1.5, 2.5, 1))/5},
    USE.NAMES = FALSE)
  out <- rep_len(dark.color, length.out = length(colors))
  out[lum <= threshold] <- light.color
  out
}

context("dark_or_light")

test_that("character(0) on NULL input", {
  res <- dark_or_light(NULL)
  expect_equal(res, character(0))
})

test_that("character(0) on character(0) input", {
  res <- dark_or_light(character(0))
  expect_equal(res, character(0))
})

test_that("color: yellow, switch to black, default threshold", {
  res <- dark_or_light("yellow")
  expect_equal(res, "black")
})

test_that("color: darkblue, switch to white, default threshold", {
  res <- dark_or_light("darkblue")
  expect_equal(res, "white")
})

test_that("color: #FFFFFF, switch to black, default threshold", {
  res <- dark_or_light("#FFFFFF")
  expect_equal(res, "black")
})

test_that("color: #000000, switch to white, default threshold", {
  res <- dark_or_light("#000000")
  expect_equal(res, "white")
})

test_that("color: #000000, switch to light.color specified, default threshold", {
  res <- dark_or_light("#000000", dark.color = "darkblue", light.color = "lightgrey")
  expect_equal(res, "lightgrey")
})

test_that("color: #000000, switch to dark.color specified, default threshold", {
  res <- dark_or_light("#FFFFFF", dark.color = "darkblue", light.color = "lightgrey")
  expect_equal(res, "darkblue")
})

test_that("color: yellow, threshold 0.8, switch to white", {
  res <- dark_or_light("yellow", threshold = 0.8)
  expect_equal(res, "white")
})

test_that("threshold: off range triggers error", {
  expect_no_error(dark_or_light("yellow", threshold = 0))
  expect_no_error(dark_or_light("yellow", threshold = 1))
  expect_error(dark_or_light("yellow", threshold = -0.1))
  expect_error(dark_or_light("yellow", threshold = 1.1))
})

test_that("threshold: length != 1 triggers error", {
  expect_error(dark_or_light("yellow", threshold = numeric()))
  expect_error(dark_or_light("yellow", threshold = NULL))
  expect_error(dark_or_light("yellow", threshold = c(0.45, 0.45)))
})

context("RNAseq_data")

test_that("volcano_example.df test size and names", {
expect_identical(nrow(volcano_example.df), 1218L)
expect_identical(ncol(volcano_example.df), 6L)
expect_named(volcano_example.df,
             c("tag", "gene", "outcome", "logFC", "PValue", "genotype"))
})

test_that("quadrant_example.df test size and names", {
  expect_identical(nrow(quadrant_example.df), 1015L)
  expect_identical(ncol(quadrant_example.df), 7L)
  expect_named(quadrant_example.df,
               c("tag", "gene", "outcome.x", "outcome.y",
                 "logFC.x", "logFC.y", "genotype"))
})

context("geom_y_margin_point")

test_that("geom_y_margin_point, missing yintercept", {
  p <- ggplot(mtcars, aes(wt, mpg)) +
    geom_point()
  p1 <- p + geom_y_margin_point()

  result <- expect_error(layer_data(p1))
})

test_that("geom_y_margin_point, yintercept", {
  p <- ggplot(mtcars, aes(wt, mpg)) +
    geom_point()
  p1 <- p + geom_y_margin_point(yintercept = c(18, 28, 15))

  expect_true(p1$layers[[2]]$data$yintercept[1] == 18)
  expect_true(p1$layers[[2]]$data$yintercept[2] == 28)
  expect_true(p1$layers[[2]]$data$yintercept[3] == 15)
})

test_that("geom_y_margin_point, yintercept dataframe", {
  p <- ggplot(mtcars, aes(wt, mpg)) +
    geom_point()
  p1 <- p + geom_y_margin_point(data = data.frame(x = c(2.5, 4.5)),
                                mapping = aes(yintercept = x))

  expect_true(p1$layers[[2]]$data$x[1] == 2.5)
  expect_true(p1$layers[[2]]$data$x[2] == 4.5)
})

context("weather_data")

test_that("test size and sums", {
expect_identical(nrow(weather_18_june_2019.df), 1440L)
expect_identical(ncol(weather_18_june_2019.df), 18L)
expect_named(weather_18_june_2019.df,
             c("time", "PAR_umol", "PAR_diff_fr", "global_watt", "day_of_year",
               "month_of_year", "month_name", "calendar_year", "solar_time",
               "sun_elevation", "sun_azimuth", "was_sunny", "wind_speed",
               "wind_direction", "air_temp_C", "air_RH", "air_DP", "air_pressure" ))
})

context("birch_data")

test_that("birch.df test size and names", {
expect_identical(nrow(birch.df), 350L)
expect_identical(ncol(birch.df), 8L)
expect_named(birch.df,
             c("Container", "Density", "block", "height", "diameter",
               "dwstem", "dwroot", "healthy"))
})

test_that("birch_dw.df test size and names", {
  expect_identical(nrow(birch_dw.df), 700L)
  expect_identical(ncol(birch_dw.df), 5L)
  expect_named(birch_dw.df,
               c("Container", "Density", "block", "dry.weight", "Part"))
})

---
title: "Extensions to the Grammar of Graphics"
subtitle: "Package 'ggpp' `r packageVersion('ggpp')` "
author: "Pedro J. Aphalo"
date: "`r packageDate('ggpp')`"
output: 
  rmarkdown::html_vignette:
    toc: yes
vignette: >
  %\VignetteIndexEntry{Extensions to the Grammar of Graphics}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  markdown: 
    wrap: 72
---

```{r, include=FALSE, echo=FALSE}
library(knitr)
opts_chunk$set(fig.align = 'center', 
               fig.show = 'hold', fig.width = 7, fig.height = 4)
options(warnPartialMatchArgs = FALSE,
        tibble.print.max = 4,
        tibble.print.min = 4,
        dplyr.summarise.inform = FALSE)
```

## Preliminaries

We load all the packages used in the examples, if installed, and set
flags to skip examples if they are not available.

```{r, message=FALSE}
library(ggpp)
library(tibble)
library(dplyr)
eval_magick <- requireNamespace("magick", quietly = TRUE)
```

We also set an uncluttered theme as default for 'ggplot2'.

```{r, message=FALSE}
old_theme <- theme_set(theme_bw() + theme(panel.grid = element_blank()))
```

# Introduction to the extensions

## Data labels, annotations and insets

*Data labels* add textual information directly related to individual
data points (shown as glyphs). Text position in this case is dependent
on the scales used to represent data points. Text is usually displaced
so that it does not occlude the glyph representing the data point and
when the link to the data point is unclear, this link is signaled with a
line segment or arrow. Data labels are distinct from annotations but
instead contribute to the representation of data on a plot or map.

*References* are lines, shading or marks used to help the reading of a
plot. These elements are used to highlight specific values on an axis or
a region in a plot. They, like data labels, are positioned relative to
the scales used for data. The position of data labels and lines, glyphs
or shading used as reference cannot be altered by the designer of a
plot, as the position conveys information. I will use the term *data
labels* irrespective if the "labels" are textual or graphical, like
icons and small plots and simple tables linked to data points or map
features.

According to Koponen and Hildén (2019), in a statistical chart 
"annotations can be used to draw reader attention to relevant detail".
These authors use as an example a text box in a plot to highlight a data
point that is off-scale and has been "squeezed" to a position
immediately outside the plotting area.

*Annotations* differ from data labels, in that their position is
decoupled from their meaning. Insets can be thought as larger, but still
self-contained annotations. In most cases the reading of inset tables
and plots depends only weakly on the plot or map in which they are
included.

In the case of annotations and insets the designer of a data
visualization has the freedom to locate them anywhere, as long as they do
not occlude features used to describe data. I will use the term
*annotation* irrespective if the "labels" are textual or graphical.
Insets are similar to annotations, but the term *inset* is used when an
annotation's graphical or textual element is complex and occupies more
space within the plotting area. Insets can be moved from within the main
plotting area to being adjacent to it, e.g., as a smaller panel, without
any loss of meaning .

That the position of annotations and insets is independent of the
plotted data cannot be expressed using the grammar of graphics (GG) as
implemented in package 'ggplot2'.

## Annotations in 'ggplot2'

The plotting of data using package 'ggplot2' is described by Wickham
(2016) and several other books. Chapter 7 in Aphalo (2020) not only
describes the grammar used by 'ggplot2' but also several extensions to
it, including those provided by packages 'ggpp' and 'ggpmisc' at the time of
writing.

In *The Layered Grammar of Graphics* (Wickham 2000) as implemented in
package 'ggplot2' (Wickham 2016) annotations are "second class"
features. As layers they behave differently than data layers: Only
constant values can be mapped to aesthetics and do no support faceting
into panels. In essence `annotate()` disconnects the resulting plot
elements from the data source and faceting, but not from the scales used
to graphically display the data values.

From the data visualization perspective the main practical and
conceptual difference between data labels and annotations is in the
scales used to position them within the plotting area. Instead of using
an `annotate()` function that deviates from the grammar of graphics to
implement annotations, we can retain the use of the grammar of
graphics for annotations but add support for native plot coordinates
(npc). Support of annotations done in this way allows "annotation
layers" to behave almost identically to "data layers", and use the same
syntax. The *x* and *y* position aesthetics used for data can be
supplemented with pseudo-aesthetics without any translation relative to
the native plotting coordinates of the plotting area or viewport. Doing
so, allows the graphic design flexibility conceptually inherent to
annotations within a user-friendly syntax.

## Support of native plot coordinates 

Based on this insight, a new approach to adding annotations and insets
was implemented in package 'ggpmisc' (>= 0.3.1) through two new *x* and
*y* pseudo-aesthetics, *npcx* and *npcy*, and corresponding dumb scales
and various geometries that make use of them. These scales and
geometries were are currently in package 'ggpp' and still used in package
'ggpmisc'.

Native plot coordinates have range 0..1. However, the "npc" geoms from 'ggpp'
also recognize certain positions by name with the same character strings as used
for text justification in 'ggplot2'. The default justification is
`"inward"`, as this protects from clipping at the edges the plotting
area irrespective of the arguments passed to the `witdth` and `height`
parameters of R's graphic devices.

```{r}
p1 <- ggplot(mtcars, aes(factor(cyl), mpg)) +
  geom_point() +
  geom_text_npc(data = data.frame(x = c("left", "left"),
                                  y = c("top", "bottom"),
                                  label = c("Most\nefficient",
                                            "Least\nefficient")),
                mapping = aes(npcx = x, npcy = y, label = label),
                size = 3)
p1

```

The advantage of this approach becomes apparent when the limits of the
*y* scale are unknown or vary. When a script or user defined plotting
function sets the scale limits based on the input data, in the absence
of the extensions proposed here, setting annotations consistently within
the plotting area becomes laborious. The example below, shows how the
annotations remain at the desired position when the *y* scale limits are
expanded.

```{r}
p1 + expand_limits(y = 0)
```

To support the existing syntax for annotations using the new geometries,
function `ggplot2::annotate()` is overridden when package 'ggpp' is
loaded. The new definition adds support for the new pseudo-aesthetics
*npcx* and *npcy* retaining its original 'ggplot2' behaviour in all
other respects.

```{r}
ggplot(mtcars, aes(factor(cyl), mpg)) +
  geom_point() +
  annotate(geom = "text_npc",
           npcx = c("left", "left"),
           npcy = c("top", "bottom"),
           label = c("Most\nefficient",
                     "Least\nefficient"),
           size = 3)
```

Inset plots can be added with the same syntax using `geom_plot_npc()`.
They can be thought also as being annotations. Here we use `annotate()`
but `geom_plot()` can be also used directly, in which case the inset
plots can be different for each panel.

```{r}
p2 <- ggplot(mtcars, aes(factor(cyl), mpg, colour = factor(cyl))) +
  stat_boxplot() +
  labs(y = NULL) +
  theme_bw(9) + 
  theme(legend.position = "none",
        panel.grid = element_blank())

ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
  geom_point() +
  annotate("plot_npc", npcx = "left", npcy = "bottom", label = p2) +
  expand_limits(y = 0, x = 0)

```

A simple example with facets, labelling of panels in a traditional way
as required by some book and journal styles. In this case panel tags are
added within the plotting area at a consistent "npc" location with free
scale limits in panels.

```{r}
ggplot(mtcars, aes(wt, mpg)) +
  geom_point() +
  geom_text_npc(data = data.frame(cyl = levels(factor(mtcars$cyl)),
                                  label = LETTERS[seq_along(levels(factor(mtcars$cyl)))],
                                  x = 0.90,
                                  y = 0.95),
                mapping = aes(npcx = x, npcy = y, label = label),
                size = 4) +
  facet_wrap(~factor(cyl), scales = "free") +
  theme(strip.background = element_blank(),
        strip.text = element_blank())
```

This approach was first implemented in 'ggpmisc' version 0.3.1 released
in April 2919. The implementation now in package 'ggpp' can be
considered stable. However, this implementation is to an extent
dependent on undocumented behaviour of 'ggplot2' functions, which means
that future updates to 'ggplot2' could break this functionality.

## When to use annotations and when insets

Package 'ggpmisc' adds support for various plot annotations and
reference guides based on model fits and other statistics. It also adds
support for some data labels related to data features. While developing
these statistics in 'ggpmisc', it became clear that expanding the
grammar of graphics's support for annotations would simplify the new
code considerably and also more cleanly separate the computations on
data from the positioning of annotations. These extensions to the
grammar are now in 'ggpp'.

**Table**. Geometries useful for data labels and annotations. Currently
implemented *ordinary* geometries and their *npc* versions. The
rightmost column shows the expected class of the objects mapped to the
`label` aesthetic. When using `annotate()` to add a single plot, table
or grob as an inset, enclosing them in a list is allowed, but not a
requirement. All those geoms whose names are highlighted in *`italics`*
support plotting of connecting segments or arrows when a the coordinates
have been modified by a *position* function.

+-------------------------------+---------------------+----------------------+
|  Data labels                  | Annotations  (npc)  | `label` aes.         |
+:==============================+:====================+:=====================+
| `ggplot2::geom_text()`        | `geom_text_npc()`   | `character`          |
+-------------------------------+---------------------+----------------------+
| `ggplot2::geom_label()`       | `geom_label_npc()`  | `character`          |
+-------------------------------+---------------------+----------------------+
| `ggrepel::geom_text_repel()`  |                     | `character`          |
+-------------------------------+---------------------+----------------------+
| `ggrepel::geom_label_repel()` |                     | `character`          |
+-------------------------------+---------------------+----------------------+
| `geom_text_s()`               |                     | `character`          |
+-------------------------------+---------------------+----------------------+
| `geom_label_s()`              |                     | `character`          |
+-------------------------------+---------------------+----------------------+
| `geom_text_pairwise()`        |                     | `character`          |
+-------------------------------+---------------------+----------------------+
| `geom_label_pairwise()`       |                     | `character`          |
+-------------------------------+---------------------+----------------------+
| `geom_point_s()`              |                     | _not supported_      |
+-------------------------------+---------------------+----------------------+
| `geom_plot()`                 | `geom_plot_npc()`   | `list(<ggplot>)`     |
+-------------------------------+---------------------+----------------------+
| `geom_table()`                | `geom_table_npc()`  | `list(<data.frame>)` |
+-------------------------------+---------------------+----------------------+
| `geom_grob()`                 | `geom_grob_npc()`   | `list(<grob>)`       |
+-------------------------------+---------------------+----------------------+
| `geom_img()`  *(planned)*     | `geom_img_npc()`    | `list(<raster>)`     |
+===============================+=====================+======================+

: **Table**. Geometries useful for data labels and annotations. Currently
implemented *ordinary* geometries and their *npc* versions. The
rightmost column shows the expected class of the objects mapped to the
`label` aesthetic. When using `annotate()` to add a single plot, table
or grob as an inset, enclosing them in a list is allowed, but not a
requirement.

## Highlighting data features

Marginal annotations with symbols and vertical and horizontal lines can be used
to highlight summary values or events/conditions. The *geoms*
`geom_quadrant_lines()`, `geom_vhlines()`, `geom_x_margin_arrow()`,
`geom_x_margin_grob()`, `geom_x_margin_point()`, `geom_y_margin_arrow()`,
`geom_y_margin_grob()` and `geom_y_margin_point()` make such labelling easy.
These work similarly to geoms `geom_hline()` and `geom_vline()` from 'ggplot2'
and plays a similar role. Of course, they can also be used to add annotations
not derived from the observations.


+-------------------------------+---------------------+----------------------+
|  Marginal marks               | Location aesthetic  | `label` aes.         |
+===============================+=====================+======================+
| `geom_x_margin_arrow()`       | `xintercept`        | _not supported_      |
+-------------------------------+---------------------+----------------------+
| `geom_y_margin_arrow()`       | `yintercept`        | _not supported_      |
+-------------------------------+---------------------+----------------------+
| `geom_x_margin_grob()`        | `xintercept`        | `list(<grob>)`       |
+-------------------------------+---------------------+----------------------+
| `geom_y_margin_grob()`        | `yintercept`        | `list(<grob>)`       |
+-------------------------------+---------------------+----------------------+
| `geom_x_margin_point()`       | `xintercept`        | _not supported_      |
+-------------------------------+---------------------+----------------------+
| `geom_y_margin_point()`       | `yintercept`        | _not supported_      |
+-------------------------------+---------------------+----------------------+
| `geom_quadrant_lines()`       |                     | _not supported_      |
+-------------------------------+---------------------+----------------------+
| `geom_vhlines()`              | `xintercept`        | _not supported_      |
|                               | `yintercept`        |                      |
+===============================+=====================+======================+

: **Table**. Geometries useful for data labels and annotations, lines and marginal marks. The middle columns shows the aesthetic that determines the position of the lines or marks on the edge of the plotting area. The rightmost column shows the expected class of the objects mapped to the `label` aesthetic.


When adding an informative element to a plot, assess whether it is an
*annotation* or a *data label*. To decide on the best approach, consider
if the location of the element is more "naturally" expressed in the
original data units or as position relative to the edges or centre of
the plotting area. In the second case, prefer the "npc" geoms as you are
dealing with *annotations*, otherwise, use the ordinary geometries as
you are dealing with data labels or data points.

## Summary statistics

With some types of data it is common to compute a summary per plot quadrant,
rather than based on a prior grouping. The *stat* `stat_quadrant_counts()` can
de used to annotate plots with the number of observations in each quadrant,
selected quadrants, or quadrants pooled along _x_ and/or _y_. Similarly,
`stat_group_counts()` and `stat_panel_counts()` can be used to annotate plots
with the number of observations per panel, or per group within each panel.

In many cases, such as rug plots or other representations on plot margins, _x_
and _y_ summaries can be computed in separate plot layers using the statistics
from 'ggplot2'. In contrast, in other cases, such as when showing the _x_ and
_y_ position of the means in the plotting area itself, it is necessary to
compute summaries for both _x_ and _y_ in the same plot layer. Function
`stat_summary_xy()` is similar to `stat_summary()` from 'ggplot2' but useful
when both *x* and *y* are continuous variables. It allows the use of the same,
or different functions to summarize the data over *x* and *y*. In contrast, its
simpler counterpart, `stat_centroid()`, always applies the same function over *x*
and *y*. 

An additional variation is statistic `stat_apply_group`, which can apply
functions to _x_ and _y_ values in parallel, which is most useful for
accumulating or differencing consecutive data values.

## Positioning data labels

Nudging is normally used to displace text labels so that data labels do not
overlap the points or lines representing the data being labelled. Two
enhanced versions of `position_nudge()` are provided,
`position_nudge_center()` and `position_nudge_line()`. These functions
make it possible to apply nudging that varies automatically according to
the relative position of points with respect to arbitrary points or
lines, or with respect to a polynomial or smoothing spline fitted
on-the-fly to the the observations.

A limitation of 'ggplot2' is that only one position function can be used
in a layer. This makes it difficult to add data labels to stacked or
dodged bars or columns, or the labelling of points that have been
jittered. The solution provided by 'ggpp' are wrappers on these position
functions from 'ggplot2'. These wrappers add support for nudging.

The position functions defined in 'ggpp', instead of deleting the original
positions, rename the variables in `data` containing these coordinates. This
makes it possible for geometries to draw connecting segments or arrows between
old and new positions. For consistency, positions equivalent to all those
defined in 'ggplot2' are defined in 'ggpp'. These functions are backwards
compatible with those from 'ggplot2' and can be used in place of them. The
geometries from 'ggrepel' (>= 0.9.2) and many of those from 'ggpp' can make use
of the original coordinates kept in data.

| Position (origin kept)                      | Position (origin discarded)       |
|---------------------------------------------|-----------------------------------|
| `position_nudge_keep()`                     | `ggplot2::position_nudge()`       |
| `ggrepel::postion_nudge_repel()`            | `ggplot2::position_nudge()`       |
| `position_nudge_centre()`                   |                                   |
| `position_nudge_line()`                     |                                   |
| `position_jitter_keep()`                    | `ggplot2::position_jitter()`      |
| `position_dodge_keep()`                     | `ggplot2::position_dodge()`       |
| `position_dodge2_keep()`                    | `ggplot2::position_dodge2()`      |
| `position_stack_keep()`                     | `ggplot2::position_stack()`       |
| `position_stack_minmax()`                   |                                   |
| `position_fill_keep()`                      | `ggplot2::position_fill()`        |
| `position_jitternudge()`                    |                                   |
| `position_dodgenudge()`                     |                                   |
| `position_dodge2nudge()`                    |                                   |
| `position_stacknudge()`                     |                                   |
| `position_fillnudge()`                      |                                   |
| `position_jitterdodge_keep()` *(planned)*   | `ggplot2::position_jitterdodge()` |
| `position_jitterdodgenudge()` *(planned)*   |                                   |

: Position functions

There is a storage-use cost in saving the original position, by making the
`data` member stored in the ggplot object bigger. This may need to be taken into
consideration when `data` has many rows. In such cases, when a counterpart
position function that does not keep the original coordinates exists, it is wise
not to use the new "keep" position functions in combination with geometries that
cannot make use of the additional information stored by them. Alternatively, and
when using those new position functions with no counterpart in 'ggplot2' the
keeping of the original position can be disabled by passing `kept.position =
"none"` as an additional argument when they are called.

## Local density of observations

With stats from 'ggplot2' it is common to plot estimates of empirical density
curves or empirical density surfaces. There are other uses for density
estimates: 1) highlighting dense or sparse regions in clouds of observations
plotted as points, and 2) using density to decide which observations to label
and which not to label. _If the highlighting or labeling is to be interpreted as
meaningful as a description of properties of the population then the fitted
curve of surface should not be too flexible, it should not track local anomalies
in the distribution of observations. In contrast if the use of the density
estimate is only used as a way of avoiding overlaps among labels with no intent
of describing properties of the data, much more flexible curves and surfaces are
most effective.

Functions `stat_dens1d_filter()` and `stat_dens2d_filter()`, and
`stat_dens1d_filter_g()` and `stat_dens2d_filter_g()` can be used to 
filter/subset data based on local density estimates. 

Functions `stat_dens1d_labels()` and `stat_dens2d_labels()` instead of removing
rows from data, replace or edit the values mapped to the `label` aesthetic
before they are passed to the geometry. This is especially useful in combination with
the repulsive geometries from package 'ggrepel' because by retaining all
observations in data these geometries can repulse labels away from
observations with and without data labels attached.

Although useful for diverse types of plots, these statistics are most useful
when creating volcano and quadrant plots. A case where the observations near the
edges of a cloud of observations are of special interest.

The density estimates used for the "filtering" of data rows or labels can be of
interest both when selecting the parameters for the fit of the empirical
density function or possibly even for mapping to an aesthetic. These are 
optionally returned.

One final twist is that even if the density estimate is in all cases obtained
from the whole of `data`, the selection criteria can be applied separately,
and even be different, for each quadrant or tail. For example, we can
highlight or label the same fraction of observations, or alternatively the same
number of observations in each quadrant or tail even if the overall density
of observations is different between tails or among quadrants.

## ggplot methods

That `ggplot()` is defined as a generic method in 'ggplot2' makes it possible to
define specializations. We provide two `ggplot()` methods for time series stored
in objects of classes `ts` and `xts`. These methods automatically convert these
objects into tibbles and set by default the aesthetic mappings for `x` and `y`
automatically. A companion function `try_tibble()` is also exported.

## List of references

Aphalo, Pedro J. (2020) *Learn R: As a Language.* The R Series. Boca
Raton and London: Chapman and Hall/CRC Press. ISBN: 978-0-367-18253-3.
350 pp.

Koponen, J; Hildén, J. (2019) *Data visualization handbook*. Aalto ARTS
books, Espoo. ISBN 978-952-60-7449-8.

Wickham H. (2010) A Layered Grammar of Graphics. *Journal of
Computational and Graphical Statistics* **19:** 3--28.

Wickham H. (2016) *ggplot2: Elegant Graphics for Data Analysis*.
Springer International Publishing. ISBN 978-3-319-24275-0.

# Examples

Some of the examples below are evaluated only if packages 'ggrepel', and
'magick' are available.

## ggplot methods

`ggplot()` methods for classes `"ts"` and `"xts"` automate plotting of
time series data, as *x* and *y* aesthetics are mapped to time and the
variable of the time series, respectively. For plotting time series data
stored in objects of other classes, see the conversion functions
`try_tibble()` and `try_data_frame()` in the last section of this
vignette.

By default, `time` is of class `Date` or of class `POSIXct` depending on
how time is stored in the time series object. On-the-fly coversion of
`time` to `numeric` is also possible (shown in the second example
below).

```{r}
class(lynx)
ggplot(lynx) + geom_line()
```

The class of variable `time`, mapped to the *x* aesthetic, affects the
scale used by default as well as the formatting of values when converted
to character strings or printed. Here we force its conversion to
`numeric`.

```{r}
ggplot(lynx, as.numeric = FALSE) + geom_line()
```

## Geometries

Three of the geometries described below allow the addition of plot
layers containing insets. Insets can be plots, tables, bitmaps, or grid
objects. Insets can be also added as annotations. Using for `data` a
tibble with a list column containing data frames or tibbles allows like
any other geom, the use of grouping, multiple insets per panel, faceting
with different tables per panel, and different number of insets in each
panel, i.e., individual tables added to a plot with `geom_table` behave
similarly to individual `character` values added with `geom_text`.

Other geometries also described in this section support the use of
native plot coordinates for positioning elements in the plotting area.
Obviously these geometries are not meant to be used to display data, but
instead they make it possible to add annotations to plots consistently
across data sets even when using scales with varying limits.

| Geometry                                   | Main use                        | Aesthetics                                                                                                | Segment |
|-------------|-------------|---------------------------------|-------------|
| `geom_text_s()`                            | data labels                     | x, y, label, size, family, font face, colour, alpha, group, angle, vjust, hjust                           | yes     |
| `geom_label_s()`                           | data labels                     | x, y, label, size, family, font face, colour, fill, alpha, linewidth, linetype, group, vjust, hjust       | yes     |
| `geom_text_pairwise()`                     | data labels                     | x, xmin, xmax, y, label, size, family, font face, colour, alpha, group, angle, vjust, hjust                     | horiz.     |
| `geom_label_pairwise()`                    | data labels                     | x, xmin, xmax, y, label, size, family, font face, colour, fill, alpha, linewidth, linetype, group, vjust, hjust | horiz.     |
| `geom_text_npc()`                          | annotations                     | npcx, npcy, label, size, family, font face, colour, alpha, group, angle, vjust, hjust                     | no      |
| `geom_label_npc()`                         | annotations                     | npcx, npcy, label, size, family, font face, colour, fill, alpha, linewidth, linetype, group, vjust, hjust | no      |
| `geom_point_s()`                           | data labels                     | x, y, size, colour, fill, alpha, shape, stroke, group                                                     | yes     |
| `geom_table()`                             | data labels                     | x, y, label, size, family, font face, colour, alpha, group, angle, vjust, hjust                           | yes     |
| `geom_table_npc()`                         | annotations                     | npcx, npcy, label, size, family, font face, colour, alpha, group, angle, vjust, hjust                     | no      |
| `geom_plot()` , `geom_grob()`              | data labels                     | x, y, label, group, angle, vjust, hjust                                                                   | yes     |
| `geom_plot_npc()` , `geom_grob_npc()`      | annotations                     | npcx, npcy, label, group, vjust, hjust                                                                    | no      |
| `geom_margin_arrow()`                      | data labels, scale labels, data | xintercept, yintercept, label, size, family, font face, colour, alpha, group, vjust, hjust                | no      |
| `geom_margin_point()`                      | data labels, scale labels, data | xintercept, yintercept, label, size, family, font face, colour, alpha, group, vjust, hjust                | no      |
| `geom_margin_grob()`                       | data labels, scale labels, data | xintercept, yintercept, label, size, family, font face, colour, alpha, group, vjust, hjust                | no      |
| `geom_quadrant_lines()` , `geom_vhlines()` | data labels, scale labels, data | xintercept, yintercept, label, size, family, font face, colour, alpha, group, vjust, hjust                | no                                                                                         |         |
: Geometries defined in package 'ggpp'

### geom_table() and stat_fmt_table()

The *geometry* `geom_table()` plots a data frame or tibble, nested in a
tibble passed as data argument, using *aesthetics* `x` and `y` for
positioning, and `label` for the list of data frames containing the data
for the tables. The tables are created as 'grid' grobs and added as
usual to the ggplot object. In contrast to "standard" geoms, this geom
by default does not inherit the globally mapped aesthetics. Tables are
always added at their native sizes, which can be altered by changing the
`size` of the text in them.

Passing a data frame or a tibble results in the same plot, but there is a
difference in how a `list` is added as a column to them, as shown below.

```{r}
tb <- mpg %>%
  group_by(cyl) %>%
  summarise(hwy = median(hwy), cty = median(cty))

# using R's data frames we need to call I() to add the list as is
data.df <- data.frame(x = 7, y = 44, tb = I(list(tb)))

ggplot(mpg, aes(displ, hwy, colour = factor(cyl))) +
  geom_table(data = data.df, aes(x, y, label = tb)) +
  geom_point() 

# using 'tibble' the list is added as is by default
data.tb <- tibble(x = 7, y = 44, tb = list(tb))

ggplot(mpg, aes(displ, hwy, colour = factor(cyl))) +
  geom_table(data = data.tb, aes(x, y, label = tb)) +
  geom_point() 
```

In plots with a single panel it can be easier to use `annotate()` to add
inset tables, here creating the same plot as above. When using `annotate()`
single data frames, ggplots or grobs do not need to be wrapped in a list, 
although lists are also supported.

```{r}
tb <- mpg %>%
  group_by(cyl) %>%
  summarise(hwy = median(hwy), cty = median(cty))

ggplot(mpg, aes(displ, hwy, colour = factor(cyl))) +
  annotate("table", x = 7, y = 44, label = tb) +
  geom_point() 
```

Table themes are supported through parameter `table.theme` and if
variables or constants are mapped to the `colour`, `fill`, `size`, `alpha` or
`family` aesthetics they override the corresponding default theme
settings. The display of `rownames` and `colnames` can be enabled or
disable through parameter `table.rownames` and `table.colnames` and the
horizontal justification of text in the core of the table through
parameter `table.hjust`.

Parameter `table.theme` accepts as arguments `NULL` for use of the
the default as set at time of rendering, a `ttheme` constructor function such 
as those defined in package 'gridExtra', or the variations on them defined in
package 'ggpp'. The active default can be set with function `ttheme_set()`.

```{r}
ggplot(mpg, aes(displ, hwy, colour = factor(cyl))) +
  geom_table(data = data.tb, aes(x, y, label = tb),
             table.theme = ttheme_gtsimple,
             table.hjust = 0, colour = "darkred", fill = "#FFFFBB") +
  geom_point() 
```

Transparency can be set with the `alpha`aesthetic or in the table theme.

```{r}
ggplot(mpg, aes(displ, hwy, colour = factor(cyl))) +
  geom_table(data = data.tb, aes(x, y, label = tb),
             table.theme = ttheme_gtdefault,
             table.hjust = 0, 
             colour = "darkred", fill = "#FFFFBB",
             alpha = 0.7) +
  geom_point()
```


Using `stat_fmt_tb()` we can rename columns and rows of the tibble,
reorder them and/or select a subset of columns or rows as shown below.
To provide a complete example we also replace the names of the scales
for *x*, *y* and *color* aesthetics. Here we pass a `character` vector
with the original names of the columns in full, but partial matching is
tried when needed. It is also possible to use a `numeric` vector of
positional indexes.

```{r}
ggplot(mpg, aes(displ, hwy, colour = factor(cyl))) +
  geom_table(data = data.tb, aes(x, y, label = tb),
             table.theme = ttheme_gtlight,
             size = 3, colour = "darkblue",
             stat = "fmt_tb", 
             tb.vars = c(Cylinders = "cyl", MPG = "hwy"), # rename
             tb.rows = 4:1) + # change order
  labs(x = "Engine displacement (l)", y = "Fuel use efficiency (MPG)",
       colour = "Engine cylinders\n(number)") +
  geom_point() +
  theme_bw()
```

Parsed text, using plot math syntax is supported in the table, with
fall-back to plain text in case of parsing errors, on a cell by cell
basis (see `tableGrob()` in package 'gridExtra' for details, as this
function is used to build the table). Here we plot the MPG for city
traffic and we can see that the plotting area expands to include the
coordinates at which the table is anchored. Justification is by default
set to `"inward"` which ensures that the table is fully within the
plotting region.

```{r}
tb.pm <- tibble(Parameter = c("frac(beta[1], a^2)", "frac(beta[2], a^3)"),
                Value = c("10^2.4", "10^3.532"))
data.tb <- tibble(x = 7, y = 44, tb = list(tb.pm))
ggplot(mpg, aes(displ, cty)) +
  geom_point() +
  geom_table(data = data.tb, aes(x, y, label = tb), parse = TRUE) +
  theme_bw()
```

As implemented, there is no limitation to the number of insets, and
faceting is respected. If the base plot shows a map, multiple small
tables could be superimposed on different countries or regions. The size
of the table inset is given by the *size* aesthetic, so like for
`geom_text()` it is independent of the `ggplot2::theme()` and controlled
by the `table.theme` directly or indirectly through the *size*
aesthetic.

Please see section **Normalised Parent Coordinates** below for a
description of `geom_table_npc()`.

### geom_plot()

The `geom_plot()` *geometry* plots ggplot objects, nested in a tibble
passed as data argument, using *aesthetics* `x` and `y` for positioning,
and `label` for the ggplot object containing the definition of the plot
to be nested. With this approach in plots with facets the insets can be
different in each panel. It is also possible to inset more than one plot
in a single call simply by creating a tibble with multiple rows.

Behind the scenes, one Grob is created for each plot to be inset. The
conversion is done with `ggplotGrob()` and the Grobs added to the main
ggplot object.

As an example we produce a plot where the inset plot is a zoomed-in
detail from the main plot. In this case the main and inset plots start
as the same plot. In most cases the size of text and other elements in
the inset should be smaller than in the main plot. Here we override the
default theme setting the `base_size` from its default of 11 pt to 8 pt.

```{r}
p <- ggplot(mpg, aes(displ, hwy, colour = factor(cyl))) +
  geom_point() 

data.tb <- 
  tibble(x = 7, y = 44, 
         plot = list(p + 
                       coord_cartesian(xlim = c(4.9, 6.2), 
                                       ylim = c(13, 21)) +
                       labs(x = NULL, y = NULL) +
                       theme_bw(8) +
                       scale_colour_discrete(guide = "none")))

ggplot(mpg, aes(displ, hwy, colour = factor(cyl))) +
  geom_plot(data = data.tb, aes(x, y, label = plot)) +
  annotate(geom = "rect", 
           xmin = 4.9, xmax = 6.2, ymin = 13, ymax = 21,
           linetype = "dotted", fill = NA, colour = "black") +
  geom_point() 
```

In general, the inset plot can be any ggplot object, allowing the
creation of very different combinations of main plot and inset plots.
Here we use the inset to show summaries as in the previous example of an
inset table.

```{r}
p <- ggplot(mpg, aes(factor(cyl), hwy, fill = factor(cyl))) +
  stat_summary(geom = "col", fun = mean, width = 2/3) +
  labs(x = "Number of cylinders", y = NULL, title = "Means") +
  scale_fill_discrete(guide = "none")

data.tb <- tibble(x = 7, y = 44, 
                  plot = list(p +
                                theme_bw(8)))

ggplot(mpg, aes(displ, hwy, colour = factor(cyl))) +
  geom_plot(data = data.tb, aes(x, y, label = plot)) +
  geom_point() +
  labs(x = "Engine displacement (l)", y = "Fuel use efficiency (MPG)",
       colour = "Engine cylinders\n(number)") +
  theme_bw()
```

The same plot as above can be created using `annotate()`, but be aware
that when using facets, 'ggplot2' annotations are identical in all
panels.

```{r}
p <- ggplot(mpg, aes(factor(cyl), hwy, fill = factor(cyl))) +
  stat_summary(geom = "col", fun = mean, width = 2/3) +
  labs(x = "Number of cylinders", y = NULL, title = "Means") +
  scale_fill_discrete(guide = "none")

ggplot(mpg, aes(displ, hwy, colour = factor(cyl))) +
  annotate("plot", x = 7, y = 44, label = p + theme_bw(8)) +
  geom_point() +
  labs(x = "Engine displacement (l)", y = "Fuel use efficiency (MPG)",
       colour = "Engine cylinders\n(number)") +
  theme_bw()
```

As implemented in 'ggpp' geometries, there is no limitation to the number of
insets, and faceting is respected. If the base plot shows a map or a bitmap,
multiple small plots could be superimposed on different countries or regions.
The size of the insets is controlled by the `vp.width` and `vp.height`
aesthetics as a fraction of the main plot's plotting region. Consequently, the
insets are scaled together with the main plot. A possible unintuitive but useful
feature, is that the theme is linked to each plot.

Please see section **Normalised Parent Coordinates** below for a
description of `geom_plot_npc()`.

### geom_grob()

The `geom_grob()` *geometry* plots grobs (graphical objects as created
with package 'grid'), nested in a tibble passed as data argument, using
*aesthetics* `x` and `y` for positioning, and `label` for the Grob
object. While `geom_table()` and `geom_plot()` take as values mapped to
the `label` aesthetics tibbles or data frames, and ggplots,
respectively, and convert them into Grobs, `geom_grob()` expects Grobs
ready to be rendered. This means that any Grob created using 'grid' or
its extensions can be added as a data label to a ggplot.

```{r, eval=eval_magick}
file.name <- 
  system.file("extdata", "Isoquercitin.png", 
              package = "ggpp", mustWork = TRUE)
Isoquercitin <- magick::image_read(file.name)
grobs.tb <- tibble(x = c(0, 10, 20, 40), y = c(4, 5, 6, 9),
                   width = c(0.05, 0.05, 0.01, 1),
                   height =  c(0.05, 0.05, 0.01, 0.3),
                   grob = list(grid::circleGrob(), 
                               grid::rectGrob(), 
                               grid::textGrob("I am a Grob"),
                               grid::rasterGrob(image = Isoquercitin)))

ggplot() +
  geom_grob(data = grobs.tb, 
            aes(x, y, label = grob, vp.width = width, vp.height = height),
            hjust = 0.7, vjust = 0.55) +
  scale_y_continuous(expand = expansion(mult = 0.3, add = 0)) +
  scale_x_continuous(expand = expansion(mult = 0.2, add = 0)) +
  theme_bw(12)
```

As shown above for inset tables and inset plots, it is also possible to
use `annotate()` with Grobs. The next example insets a single `Grob`.
Here we reuse the bitmap `Isoquercitin` read in the previous example.
The Grob is contained in a viewport. Here setting `width = 1` ("npc"
units) when creating the `Grob` from the bitmap ensures that the bitmap
fills the width of the viewport (to ensure that the inset is not
distorted, set only one of `width` or `height`). The argument to
`vp.width` or `vp.height`, also in "npc" units, determines the size of
the `Grob` relative to the size of the plotting area.

```{r, eval=eval_magick}
ggplot() +
  annotate("grob", x = 1, y = 3, vp.width = 0.5,
           label = grid::rasterGrob(image = Isoquercitin, width = 1)) +
  theme_bw(12)
```

`geom_grob()` is designed thinking that its main use will be for graphical
annotations, although one could use it for infographics with multiple
copies of each grob, this would go against the grammar of graphics. In
this implementation grobs cannot be mapped to an aesthetic through a
scale.

As implemented, there is no limitation to the number of insets and
faceting is respected. If the base plot shows a map or a bitmap,
multiple simple grobs (e.g. national flags) could be superimposed on
different countries. The size of the insets is controlled by the
`vp.width` and `vp.height` aesthetics as a fraction of the main plot's
plotting region. Consequently, the insets are scaled together with the
main plot.

Please see section **Normalised Parent Coordinates** below for a
description of `geom_grob_npc()`.

### geom_vhlines()

This is a convenience geometry that adds both vertical and horizontal
guide lines on the same plot layer, using the same syntax as
`geom_hline()` and `geom_vline()` from package 'ggplot2'.

```{r}
ggplot(mpg, aes(displ, hwy, colour = factor(cyl))) +
  geom_vhlines(xintercept = c(2.75, 4), yintercept = 27, linetype = "dashed") +
  geom_point() +
  labs(x = "Engine displacement (l)", y = "Fuel use efficiency (MPG)",
       colour = "Engine cylinders\n(number)") 

```

### geom_text_s(), geom_label_s(), and geom_point_s()

With package 'ggrepel' we have become used to easily add data labels linked
to the observations or points by a line segment or arrow. However,
sometimes a plot has few text labels and repulsion is unnecessary. In such cases
nudging or dodging may be enough, or even preferable so as to achieve consistent
positions irrespective of plot scaling or graphic device used for rendering.

We can achieve this with `geom_text_s()` (formerly named `geom_text_linked()`)
or `geom_label_s()`, together with a suitable position function that retains the
original *x* and *y* coordinates of the point. We provide several different
position functions described in a later section, all of which retain the initial
position in `data$x_orig` and `data_$y_orig`. Function `position_nudge_keep()`
is used by default providing the same default behaviour as `geom_text()` and
`geom_label()` from 'ggplot2'. However, when nudging using the `nudeg_x` and/or
`nudeg_y` parameters, or passing as argument to `position` a position function
that keeps the original position, by default linking segments are drawn between
the original and nudged or displaced positions. Passing `add.segments = FALSE`
disables the drawing of segments. The equivalent geometry for points connected
by segments to their original positions is `geom_point_s()`.

These geometries do not use like those in package 'ggrepel' special aesthetics 
to set the properties of segments. These geometries allow the user to select to
which graphic elements to apply the mapped aesthetics. In addition 
`geom_label_s()` adds support for `linewidth` and `linetype` aesthetics for
the box.

```{r}
my.cars <- mtcars[c(TRUE, FALSE, FALSE, FALSE), ]
my.cars$name <- rownames(my.cars)
my.cars <- my.cars[order(my.cars$wt), ]
ggplot(my.cars, aes(wt, mpg, label = name)) +
  geom_point() +
  geom_text_s(aes(colour = factor(cyl)),
              vjust = 0.5,
              angle = 90,
              nudge_y = -1.5, 
              arrow = arrow(length = grid::unit(1.5, "mm"))) +
  scale_colour_discrete(l = 40) +
  expand_limits(y = 0)
```

```{r}
my.cars <- mtcars[c(TRUE, FALSE, FALSE, FALSE), ]
my.cars$name <- rownames(my.cars)
my.cars <- my.cars[order(my.cars$wt), ]
ggplot(my.cars, aes(wt, mpg, label = name)) +
  geom_point() +
  geom_label_s(aes(colour = factor(cyl)),
               size = 2.5,
               linewidth = 0.5,
               colour.target = c("box", "segment"),
               nudge_x = 0.2,
               arrow = arrow(length = grid::unit(1.5, "mm"))) +
  scale_colour_discrete(l = 40) +
  expand_limits(x = 6.5)
```

### geom_text_pairwise() and geom_label_pairwise()

When labels show the results of pairwise comparisons, it is natural to use a 
segment to connect the members of the pairs compared to a textual label.

```{r}
# With a factor mapped to x, highlight pairs

my.cars <- mtcars
my.cars$name <- rownames(my.cars)
p1 <- ggplot(my.cars, aes(factor(cyl), mpg)) +
       geom_boxplot(width = 0.33)
```

```{r}
my.pairs <-
  data.frame(A = 1:2, B = 2:3, bar.height = c(12, 30),
             p.value = c(0.01, 0.05678))
```

```{r}
p1 +
  geom_text_pairwise(data = my.pairs,
                     aes(xmin = A, xmax = B,
                         y = bar.height,
                         label = p.value),
                     parse = TRUE)

```
```{r}
p1 +
  geom_label_pairwise(data = my.pairs,
                      aes(xmin = A, xmax = B,
                          y = bar.height,
                          label = sprintf("italic(P)~`=`~%.2f", p.value)),
                      colour = "red", size = 2.75,
                      arrow = grid::arrow(angle = 30,
                                          length = unit(1.5, "mm"),
                                          ends = "both"),
                      parse = TRUE)
```

```{r}
p1 +
  geom_text_pairwise(data = my.pairs,
                     aes(xmin = A, xmax = B,
                         y = bar.height,
                         label = sprintf("italic(P)~`=`~%.2f", p.value)),
                     colour = "red", colour.target = "segment",
                     arrow = grid::arrow(angle = 90,
                                         length = unit(1, "mm"),
                                         ends = "both"),
                     parse = TRUE)
```

```{r}
p1 +
  geom_text_pairwise(data = my.pairs,
                     aes(xmin = A, xmax = B,
                         y = bar.height,
                         label = sprintf("italic(P)~`=`~%.2f", p.value)),
                     colour = "red", colour.target = "text",
                     arrow = grid::arrow(angle = 90,
                                         length = unit(1, "mm"),
                                         ends = "both"),
                     parse = TRUE)
```

Another situation where horizontal or vertical segments can be useful if to label ranges of values of a continuous variable.

```{r}
# with a numeric vector mapped to x, indicate range

p2 <-
  ggplot(my.cars, aes(disp, mpg)) +
    geom_point()

my.ranges <-
  data.frame(A = c(50, 400),
             B = c(200, 500),
             bar.height = 5,
             text = c("small", "large"))
```

```{r}
p2 +
  geom_text_pairwise(data = my.ranges,
                     aes(xmin = A, xmax = B,
                     y = bar.height, label = text))
```

```{r}
p2 +
  geom_label_pairwise(data = my.ranges,
                      aes(xmin = A, xmax = B,
                          y = bar.height, label = text))
```

### Normalised Parent Coordinates

R's 'grid' package defines several units that can be used to describe
the locations of plot elements. In 'ggplot2' the *x* and *y* aesthetics
are directly mapped to `"native"` or data units. For consistent location
of **annotations** with respect to the plotting area we need to rely on
`"npc"` which are expressed relative to the size of the grid viewport.
The plotting area in a ggplot is implemented as a 'grid' viewport and
support for `"npc"` coordinates is relatively easy to implement.

To support `"npc"` positions we have implemented *scales* for two new
(pseudo) *aesthetics*, `npcx` and `npcy`. These are very simple
continuous scales which do not support any transformation or changes to
their limits, both of which would be meaningless for `"npc"` units.
Variables mapped to these aesthetics can be either numerical with values
in the range zero to one or character. A limited set of strings are
recognised and converted to `"npc"` units: `"bottom"`, `"center"`,
`"top"`, `"left"`, `"middle"`, `"right"` (`"centre"` is a synonym for
`"center"`).

To make these scales useful we need also to define *geometries* that use
these new aesthetics. Package 'ggpp' currently provides
`geom_text_npc()`, `geom_label_npc()`, `geom_table_npc()`,
`geom_plot_npc()` and `geom_grob_npc()`.

As is the case for `geom_text()` and `geom_label()` from package
'ggplot2', `"bottom"`, `"center"`, `"top"`, `"left"`, `"middle"`,
`"right"`, plus `"inward"` and `"outward"` can be used, as well as
numeric values, to control the justification. Justification defaults to
`"inward"` in the geometries described here.

While the usual *x* and *y* aesthetics are used whenever the positions
of plot elements represent data values, these new scales and geometries
are useful only for annotations, i.e., in those cases when we want plot
elements at specific positions within the plotting area irrespective of
the ranges of the data mapped to the *x* and *y* aesthetics. When
writing scripts or functions that may be applied to different data sets
these new *aesthetics* help by keeping the code concise and reusable.
These *geometries* are used by default by several of the statistics
described in later sections and those defined in package 'ggpmisc'.

As an example let's imagine that we want to add the structure of a
metabolite to a plot. Its position has nothing to do with the data
mapped to *x* and *y*, so it is conceptually better to use `"npc"`
coordinates. The big practical advantage is that this also allows to
keep this part of the plot definition independent of the data being
plotted, giving a major advantage in the case of plots with facets with
free scale limits. This example can be easily adapted to
`geom_plot_npc()` where a list of ggplots is mapped to `label`, and to
`geom_table_npc()` where a list of data frame is mapped to `label`.

We produce the example plot by first constructing a tibble to contain
the grob and the coordinate data, and then map these variables to
aesthetics using `aes()`. In the example the tibble has a single row,
but this is not a requirement. In this respect these geoms behave as
normal geoms, with facets also supported.

```{r, eval=eval_magick}
file.name <- 
  system.file("extdata", "Robinin.png", 
              package = "ggpp", mustWork = TRUE)
Robinin <- magick::image_read(file.name)

set.seed(123456)
data.tb <- tibble(x = 1:20, y = (1:20) + rnorm(20, 0, 10))

flavo.tb <- tibble(x = 0.02,
                   y = 0.95,
                   width = 1/2,
                   height = 1/4,
                   grob = list(grid::rasterGrob(image = Robinin)))

ggplot(data.tb, aes(x, y)) +
  geom_grob_npc(data = flavo.tb, 
                aes(label = grob, npcx = x, npcy = y, 
                    vp.width = width, vp.height = height)) +
  geom_point() +
  expand_limits(y = 55, x = 0)
```

Alternatively, we can pass constant values to `geom_grob_npc()` to
obtain the same plot. This approach can be handy in simple cases.

```{r, eval=eval_magick}
ggplot(data.tb, aes(x, y)) +
  geom_grob_npc(label = list(grid::rasterGrob(image = Robinin, width = 1)), 
                npcx = 0.02, npcy = 0.95,
                vp.width = 1/2, vp.height = 1/4) +
  geom_point() +
  expand_limits(y = 55, x = 0)
```

We can also use `annotate()` if the annotation should be the same for
all panels, or if we have a single figure panel. In this case there is
no need to wrap a single `grob` in a list.

```{r, eval=eval_magick}
ggplot(data.tb, aes(x, y)) +
  annotate("grob_npc", label = grid::rasterGrob(image = Robinin, width = 1), 
                npcx = 0.02, npcy = 0.95, vp.width = 1/2, vp.height = 1/4) +
  geom_point() +
  expand_limits(y = 55, x = 0)

```

Two additional geometries are based on existing 'ggplot2' geometries.
They are based on `geom_text()` and `geom_label()`. We give an example
using `geom_text_npc()` to produce a "classic" labelling for facets
matching the style of `theme_classic()` and traditional scientific
journals' design.

```{r}
corner_letters.tb <- tibble(label = LETTERS[1:4],
                            x = "right", 
                            y = "top",
                            cyl = c(4,5,6,8))
ggplot(mpg, aes(displ, hwy)) +
  geom_point() +
  facet_wrap(~cyl, scales = "free") +
  geom_text_npc(data = corner_letters.tb,
                aes(npcx = x, npcy = y, label = label)) +
  theme_classic() +
  theme(strip.background = element_blank(),
        strip.text.x = element_blank())
```

### Marginal markings

'ggplot2' provides `geom_rug()`, `geom_vline()` and `geom_hline()`. Rug
plots are intended to be used to represent distributions along the
margins of plot. `geom_vline()` and `geom_hline()`are normally used to
separate regions in a plot or to highlight important values along the
*x* or *y* axis. When creating plots it is sometimes useful to put small
marks along the axes, just inside the plotting area, similar to those in
a rug plot, but like `geom_vline()` and `geom_hline()` in their purpose.

Three *geometries* provide such markers: `geom_margin_point()`,
`geom_margin_arrow()`, and `geom_margin_grob()`. They behave similarly
to `geom_vline()` and `geom_hline()` and their positions are determined
also by the `xintercept` and `yintercept` aesthetics.

In the example below we indicate the group medians along the *x* axis
with filled triangles.

```{r}
data.tb <- mpg %>%
  group_by(cyl) %>%
  summarise(hwy = median(hwy), displ = median(displ))
ggplot(mpg, aes(displ, hwy, colour = factor(cyl))) +
  geom_x_margin_point(data = data.tb,
                      aes(xintercept = displ, fill = factor(cyl))) +
  expand_limits(y = 10) +
  geom_point() 
```

## Statistics

| Statistic                 | Main use                   | Usual geometries    | Most used with   | Compute function |
|---------------------------|----------------------------|---------------------|------------------|-------------|
| `stat_fmt_tb()`           | formatting and selection   | `geom_table()`    | tables as data labels                | group |
| `stat_fmt_tb()`           | formatting and selection   | `geom_table_npc()`    | tables as annotations                | group |
| `stat_dens2d_filter()`    | local 2D density filtering | `geom_text_s()`, `geom_label_s()`, `geom_text()`, `geom_label()`  | text as data labels  | panel |
| `stat_dens2d_label()`     | local 2D density filtering | `geom_text_repel()`, `geom_label_repel()`  | text as data labels  | panel |
| `stat_dens1d_filter()`    | local 1D density filtering | `geom_text_s()`, `geom_label_s()`, `geom_text()`, `geom_label()`  | text as data labels  | panel |
| `stat_dens1d_label()`     | local 1D density filtering | `geom_text_repel()`, `geom_label_repel()`  | text as data labels  | panel |
| `stat_dens2d_filter_g()`  | local 2D density filtering | `geom_text_s()`, `geom_label_s()`, `geom_text()`, `geom_label()`  | text as data labels  | group |
| `stat_dens2d_label_g()`   | local 2D density filtering | `geom_text_repel()`, `geom_label_repel()`  | text as data labels  | group |
| `stat_dens1d_filter_g()`  | local 1D density filtering | `geom_text_s()`, `geom_label_s()`, `geom_text()`, `geom_label()`  | text as data labels  | group |
| `stat_dens1d_label_g()`   | local 1D density filtering | `geom_text_repel()`, `geom_label_repel()`  | data labels  | group |
| `stat_panel_counts()`     | number of observations     | `geom_text()`, `geom_label()` | text as annotation | panel |
| `stat_group_counts()`     | number of observations     | `geom_text()`, `geom_label()` | text as annotation | panel |
| `stat_quadrant_counts()`  | number of observations     | `geom_text()`, `geom_label()` | text as annotation | panel |
| `stat_apply_panel()`      | cummulative summaries      | `geom_point()`, `geom_line()`, etc. | scatter and line plots | panel |
| `stat_apply_group()`      | cummulative summaries      | `geom_point()`, `geom_line()`, etc. | scatter and line plots | group |
| `stat_centroid()`         | joint x and y summaries    | `geom_point()`, `geom_rug()`, `geom_margin_arrow()`, etc. | data summary | group |
| `stat_summary_xy()`       | joint x and y summaries    | `geom_point()`, `geom_rug()`, `geom_margin_arrow()`, etc. | data summary | group |
| `stat_functions()`        | compute y from x range     | `geom_line()`, `geom_point()`, etc. | draw function curves | group |
: Statistics defined in package 'ggpp'.

### stat_centroid()

It can be useful to mark the centroid of a group of observations with a
point or with a label. By default `stat_centroid()` applies function
`mean_se()` to both `x` and `y` by group. If the value mapped to an
aesthetics across rows within each group is unique this value is copied
to the returned `data`.

```{r}
ggplot(mpg, aes(displ, hwy, colour = factor(cyl))) +
  geom_point(alpha = 0.33) +
  stat_centroid(shape = "cross", size = 4)
```

Other functions can be passed to this statistic as long as they return a
single value that can be mapped to the `x` and `y` aesthetics (numeric,
time or a factor).

```{r}
ggplot(mpg, aes(displ, hwy, colour = factor(cyl))) +
  geom_point(alpha = 0.33) +
  stat_centroid(shape = "cross", size = 4, .fun = median)
```

The very similar `stat_summary_xy()` accepts different functions for `x`
and `y`.

### stat_group_counts()

```{r}
ggplot(mpg, aes(displ, hwy, colour = factor(cyl))) +
  geom_point(alpha = 0.25) +
  stat_panel_counts()
```

```{r}
ggplot(mpg, aes(displ, hwy, colour = factor(cyl))) +
  geom_point(alpha = 0.2) +
  stat_group_counts(hstep = 0.09, vstep = 0, label.x = "left", label.y = "bottom")
```

Labels with counts of observations can be also added when aesthetics `x` or `y`
are factors, as in ggplots factors create a grouping. If the labels are to be
aligned with the factor levels, we need to use a "normal" geometry such as
`geom_text()` instead of the default `geom_text_npc()` and set the corresponding
label positions to `"factor"`.

```{r}
ggplot(mpg,
       aes(factor(cyl), hwy)) +
  stat_boxplot() +
  stat_group_counts(geom = "text",
                    label.y = 10,
                    label.x = "factor") +
  stat_panel_counts()
```

### stat_quadrant_counts() and geom_quadrant_lines()

This statistic automates the annotation of plots with number of observations
(counts, fractions, or percents), either by quadrant, by pairs of quadrants or
the four quadrants taken together (whole plotting area). Its companion geometry,
`geom_quadrant_lines()` is used in the examples to highlight the quadrants.

We generate some artificial data.

```{r}
set.seed(4321)
# generate artificial data
x <- -99:100
y <- x + rnorm(length(x), mean = 0, sd = abs(x))
my.data <- data.frame(x, 
                      y, 
                      group = c("A", "B"))
```

Using defaults except for color.

```{r}
ggplot(my.data, aes(x, y)) +
  geom_quadrant_lines(colour = "red") +
  stat_quadrant_counts(colour = "red") +
  geom_point() +
  expand_limits(y = c(-250, 250))
```

Number in quadrant over total observations.

```{r}
ggplot(my.data, aes(x, y)) +
  geom_quadrant_lines(colour = "red") +
  stat_quadrant_counts(aes(label = after_stat(fr.label)), colour = "red") +
  geom_point() +
  expand_limits(y = c(-250, 250))
```

Decimal fractions.


```{r}
ggplot(my.data, aes(x, y)) +
  geom_quadrant_lines(colour = "red") +
  stat_quadrant_counts(aes(label = after_stat(dec.label)), colour = "red") +
  geom_point() +
  expand_limits(y = c(-250, 250))
```

Percent of observations.

```{r}
ggplot(my.data, aes(x, y)) +
  geom_quadrant_lines(colour = "red") +
  stat_quadrant_counts(aes(label = after_stat(pc.label)), colour = "red") +
  geom_point() +
  expand_limits(y = c(-250, 250))
```

User-constructed labels.


```{r}
ggplot(my.data, aes(x, y)) +
  geom_quadrant_lines(colour = "red") +
  stat_quadrant_counts(aes(label = sprintf("%i genes", after_stat(count))), colour = "red") +
  geom_point() +
  expand_limits(y = c(-250, 250))
```
Pooling quadrants along the *x*-axis. (`pool.along = "y"` pools along
*y*.)

```{r}
ggplot(my.data, aes(x, y)) +
  geom_quadrant_lines(colour = "red", pool.along = "x") +
  stat_quadrant_counts(colour = "red", pool.along = "x") +
  geom_point()
```

Manual positioning of the text annotations and pooling of all four
quadrants, and overriding the default formatting for the label.

```{r}
ggplot(my.data, aes(x, y)) +
  geom_point() +
  stat_quadrant_counts(quadrants = 0L, label.x = "left", 
                       aes(label = sprintf("%i observations", after_stat(count))))
```

Annotation of only specific quadrants.

```{r}
ggplot(my.data, aes(x, y)) +
  geom_quadrant_lines(colour = "red") +
  stat_quadrant_counts(colour = "red", quadrants = c(1:4)) +
  scale_y_continuous(expand = expansion(mult = 0.12)) + # add space
  geom_point()
```

```{r}
ggplot(my.data, aes(x, y)) +
  geom_quadrant_lines(colour = "red") +
  stat_quadrant_counts(colour = "red", quadrants = c(2, 4)) +
  geom_point()
```

Using facets, even with free scale limits, the labels are placed
consistently. This achieved by the default use of `geom_text_npc()` or
as shown below by use of \`geom_label_npc(). We expand the *y* limits to
ensure that no observations are occluded by the labels.

```{r}
ggplot(my.data, aes(x, y, colour = group)) +
  geom_quadrant_lines() +
  stat_quadrant_counts(geom = "label_npc") +
  geom_point() +
  expand_limits(y = c(-260, 260)) +
  facet_wrap(~group)
```

### stat_apply_group

This statistic applies functions to `x` and `y` data. The function(s)
supplied as argument are expected to return a numeric vector. Both
functions should return vectors of the same length. When possible it is
usually better to apply functions through mappings using `aes()` when
they are independent of grouping, or to use a transformation for the
scale. However, when grouping is important, these statistics make it
possible to avoid pre-computation of the data for a layer.

Here we plot the difference in trunk circumference between dates for
each tree. (As the vector returned by `diff()` is one element shorter
than its input, we delete the first value of `x` in `.fun.x`.).

```{r}
ggplot(Orange, aes(age, circumference, colour = Tree)) +
  stat_apply_group(.fun.x = function(x) {x[-1L]},
                   .fun.y = diff)
```

### stat_dens2d_labels() and stat_dens2d_filter()

These stats had their origin in an enhancement suggestion for 'ggrepel'
from Hadley Wickham and discussion with Kamil Slowikowski (ggrepel's
author) and others. In fact the code is based on code Kamil gave during
the discussion, but simplified and taking a few additional ideas from
`ggplot::stat_dens2d`.

---

**Warning!** Which observations are selected by the algorithm used,
based on `MASS:kde2d`, depends strongly on the values of parameters `h`
and `n`. You may need to alter the defaults by passing explicit
arguments to these stats. Beware, though, that what are good values, may
depend on individual data sets even if they include the same number of
observations. 

If we want to highlight the outermost observations at the edges of a empirical
density distribution, values of `n` similar to those in the examples in the
documentation of `MASS::kde2d` and `ggplot2::stat_dens2d` are suitable.

In contrast, if the aim is to prevent overlaps among labels, by preventing
observation labeling in regions with _local_ high density we need to fit a
much more flexible surface based on a very dense grid of estimates. In this
case, for the selection of observations to work cleanly, much larger values of `n`
than in the examples in the documentation of `MASS::kde2d` and
`ggplot2::stat_dens2d` are needed in most cases.

Functions `stat_dens2d_labels()` and `stat_dens2d_filter()` will return the
density estimates if `return.density = TRUE` is passed when called. These
density estimates can be plotted to visualize the density function being used
to filter or reset the labels.

---

Some random data with random labels.

```{r}
random_string <- function(len = 3) {
paste(sample(letters, len, replace = TRUE), collapse = "")
}

# Make random data.
set.seed(1001)
d <- tibble::tibble(
  x = rnorm(100),
  y = rnorm(100),
  group = rep(c("A", "B"), c(50, 50)),
  lab = replicate(100, { random_string() })
)
```

The stat `stat_dens2d_filter` *filters* observations, in other words
passes to the geom a subset of the data received as input. The default
argument for `geom` is `"point"`.

Using defaults except for the color aesthetic. Highlight 1/4 of observations
from lowest density areas of the whole plot panel (using the default of
`pool.along = "xy"`) using overplotting of selected black points with red
points. We also highlight the boundaries of the quadrants.

```{r}
ggplot(data = d, aes(x, y)) +
  geom_quadrant_lines(linetype = "dashed") +
  geom_point() +
  stat_dens2d_filter(keep.fraction = 1/4, 
                     colour = "red")
```
Highlight 1/4 of observations from lowest density areas in each quadrant of
the plot. We also highlight the boundaries of the quadrants. Compare the plot
above to that below.

```{r}
ggplot(data = d, aes(x, y)) +
  geom_quadrant_lines(linetype = "dashed") +
  geom_point() +
  stat_dens2d_filter(keep.fraction = 1/4, 
                     colour = "red", 
                     pool.along = "none")
```
Highlight a given fraction of observations in each quadrant, by setting `pool.along = "none"`,
and passing a vector of length four to `keep.fraction`.

Keep at most 20 observations.

```{r}
ggplot(data = d, aes(x, y)) +
  geom_quadrant_lines(linetype = "dashed") +
  geom_point() +
  stat_dens2d_filter(keep.fraction = c(1/2, 1/4, 0, 1/2),
                     pool.along = "none",
                     colour = "red")
```

Keep always 20 observations by setting `keep.fraction = 1`.

```{r}
ggplot(data = d, aes(x, y)) +
  geom_quadrant_lines(linetype = "dashed") +
  geom_point() +
  stat_dens2d_filter(keep.fraction = 1, 
                     keep.number = 20, 
                     colour = "red")
```

Keep always 20 observations, 5 per quadrant, by setting `pool.along = "none"`.

```{r}
ggplot(data = d, aes(x, y)) +
  geom_quadrant_lines(linetype = "dashed") +
  geom_point() +
  stat_dens2d_filter(keep.fraction = 1, 
                     keep.number = 20, 
                     pool.along = "none", 
                     colour = "red")
```

Keep a given number of observations in each quadrant, by setting `pool.along = "none"`,
and passing a vector of length four to `keep.number`.

```{r}
ggplot(data = d, aes(x, y)) +
  geom_quadrant_lines(linetype = "dashed") +
  geom_point() +
  stat_dens2d_filter(keep.fraction = 1, 
                     keep.number = c(1, 2, 3, 0), 
                     pool.along = "none", 
                     colour = "red")
```
A more elaborate example, 50% of observations, with color representing the
local empirical density, only for them.

```{r}
ggplot(data = d, aes(x, y)) +
  geom_quadrant_lines(linetype = "dashed") +
  geom_point(size = 3, colour = "grey50") +
  stat_dens2d_filter(keep.fraction = 1/2, 
                     return.density = TRUE,
                     aes(color = after_stat(density)),
                     size = 2,
                     show.legend = TRUE) +
  scale_color_viridis_c(direction = -1, option = "magma", begin = 0.5)
```

Highlighting 1/4 of the observations from the panel, using
over-plotting with a 'hollow' shape. (We also shift one group with
respect to the other in `data`, to show that grouping is ignored by
`stat_dens2d_filter()`. 

```{r}
ggplot(data = d, aes(x + rep(c(-2,2), rep(50,2)), 
                     y, 
                     colour = group)) +
   geom_point(size = 1) +
   stat_dens2d_filter(shape = 1, size = 3,
                      keep.fraction = 0.25)
```

Highlight 1/4 of observations from lowest density areas of the plot,
with density assessed separately for each group. In this
example grouping is based on the color aesthetic.

```{r}
ggplot(data = d, aes(x + rep(c(-2,2), rep(50,2)), 
                     y, 
                     colour = group)) +
   geom_point(size = 1) +
   stat_dens2d_filter_g(shape = 1,
                        size = 3,
                        keep.fraction = 0.25)
```

The stat `stat_dens2d_labels` *replaces* the values of the label
(aesthetic) variable in `data` based on density of observations along
the *x* or *y* axis in the plot panel. The replacement is given by the
argument passed to `label.fill`, which can be a character string or a
function accepting a character string as argument and returning also a
character string.

The default value for `geom` is `"text"`. The default value of
`label.fill` is `""` which results in empty labels, while using `NA` as
fill label results in observations being omitted. Using `NA` as
`label.fill` is similar to using `stat_dens2d_filter` as long as the
geom used requires a `label` aesthetic.

Label 1/10 of observations from lowest density areas of the plot panels.

```{r}
ggplot(data = d, aes(x, y, label = lab, colour = group)) +
  geom_point() +
  stat_dens2d_labels(keep.fraction = 1/5,
                     position = position_nudge_center(x = 0.05, 
                                                      y = 0.05,
                                                      center_x = 0,
                                                      center_y = 0),
                     vjust = "outward", hjust = "outward") +
  scale_x_continuous(expand = expansion(c(0.1, 0.1)))
```

Using the geoms from package 'ggrepel' avoids clashes among labels and clashes
of labels with data points. Please, see vignette _Combining repulsion and
nudging_ for more examples.

### stat_dens1d_labels() and stat_dens1d_filter()

These stats are similar to `stat_dens2d_labels()` and
`stat_dens2d_filter()` but compute the density in a single dimension,
either the *x* or *y* aesthetics. A similar warning as for 2D densities 
applies to 1D density estimation.

**Warning!** Which observations are selected by the algorithm used,
based on `stats::density`, depends strongly on the values of parameters
`bw`, `adjust` and `kernel`. You may need to alter the defaults by
passing explicit arguments. Beware that what are good values, may depend
on individual data sets even if they include the same number of
observations. For the selection of observations to work cleanly, the
argument for `n` must large enough to generate a dense grid or the
bandwidth may need to be increased by passing a number \> 1 as argument.
Increasing the bandwidth makes the empirical density function smoother,
and the selection of points less dependent on immediate neighbours.

We use the same data as in the previous sections.

```{r}
random_string <- function(len = 6) {
paste(sample(letters, len, replace = TRUE), collapse = "")
}

# Make random data.
set.seed(1001)
d <- tibble::tibble(
  x = rnorm(100),
  y = rnorm(100),
  group = rep(c("A", "B"), c(50, 50)),
  lab = replicate(100, { random_string() })
)
```

The stat `stat_dens1d_filter` *filters* observations, in other words
passes to the geom a subset of the data received as input. The default
value for `geom` is `"point"` and the default `orientation` is `"x"`.

Using defaults except for the color aesthetic, we highlight 1/4 of
observations from lowest density region along the *x* axis of the plot
panel.

```{r}
ggplot(data = d, aes(x, y)) +
  geom_point() +
  stat_dens1d_filter(keep.fraction = 0.25,
                     colour = "red")
```

We repeat the example above, we highlight 1/4 of observations, but now
from lowest density region along the *y* axis of the plot panel.

```{r}
ggplot(data = d, aes(x, y)) +
  geom_point() +
  stat_dens1d_filter(keep.fraction = 0.25,
                     colour = "red",
                     orientation = "y")
```

In other respects than `orientation` and the parameters passed
internally to `stats::density()` the examples given earlier for
`stat_dens2d_filter()` also apply.

The stat `stat_dens1d_labels` *replaces* the values of the label
(aesthetic) variable in `data` based on density of observations along
the *x* or *y* axis in the plot panel. The replacement is given by the
argument passed to `label.fill`, which can be a character string or a
function accepting a character string as argument and returning also a
character string.

The default value for `geom` is `"text"`. The default value of
`label.fill` is `""` which results in empty labels, while using `NA` as
fill label results in observations being omitted. Using `NA` as
`label.fill` is similar to using `stat_dens2d_filter` as long as the
geom used requires a `label` aesthetic.

In other respects than `orientation` and the parameters passed
internally to `stats::density()` the examples given earlier for
`stat_dens2d_labels()` also apply.

## Simple position functions

While _jitter_ and _dodge_ are normally use to displace the observations being
plotted to avoid overlaps and _stack_ used to construct columns or areas that
represent sums and contributions to a sum from different groups of observations,
_nudge_ is almost always used to displace the position of data labels relative
to observations they are attached to, to avoid overlaps.

Nudging shifts deterministically the coordinates giving an *x* and/or
*y* position and also expands the limits of the corresponding scales to
match. By default in 'ggplot2' geometries and position functions no
nudging is applied.

Function `position_nudge()` from package 'ggplot2' simply applies the
nudge, or *x* and/or *y* shifts based directly on the values passed to
its parameters `x` and `y`. Passing arguments to the `nudge_x` and/or
`nudge_y` parameters of a geometry has the same effect as these values
are passed to `position_nudge()`. Geometries also have a `position`
parameter to which we can pass an expression based on a *position
function* which opens the door to more elaborate approaches to nudging.

A new variation on simple nudge is provided by function `position_nudge_to()`,
which accepts the desired nudged final coordinates directly instead of as a
shift.

We can do better than simply shifting all data to the same extent and
direction or to a fixed position. For example by nudging away from a
focal point, a line or a curve. In `position_nudge_center()` and
`position_nudge_line()` described below, this reference alters only the
direction (angle) along which nudge is applied but not the extent of the
shift. Advanced nudging works very well, but only for some patterns of
observations and may require manual adjustment of positions, repulsion
is more generally applicable but like jitter is aleatory. Combining
nudging and repulsion we can make repulsion more predictable with little
loss of its applicability.

These functions can be used with any geometry but if segments joining
the labels to the points are desired, layer functions `geom_text_s()`,
`geom_label_s()`, etc., from 'ggpp' or `geom_text_repel()` or
`geom_label_repel()` feom 'ggrepel' have to be used. 

Several geometries defined in package 'ggpp' even if not supporting repulsion,
can plot connecting segments. Please see the documentation for the geometries
for the details. Drawing of segments or arrows is possible if both the nudged
and original *x* and *y* coordinates are stored in `data`. Support of this
feature across packages is possible by coordinated development of 'ggpp' and
'ggrepel' and agreement on a naming convention for storing the original
position in position functions as those described next.

| Position                  | Main use       | Displacement                    | Most used with              |
|-----------------|-----------------|---------------------|------------------|
| `position_nudge_keep()`   | nudge          | x, y (fixed distance)           | data labels                 |
| `position_jitter_keep()`  | jitter         | x, y (random)                   | dot plots                   |
| `position_stack_keep()`   | stack          | vertical (absolute)             | column and bar plots        |
| `position_stack_minmax()` | stack          | vertical (absolute)             | error bars                  |
| `position_fill_keep()`    | fill           | vertical (relative, fractional) | column plots                |
| `position_dodge_keep()`   | dodge          | sideways (absolute)             | column and bar plots        |
| `position_dosge2_keep()`  | dodge2         | sideways (absolute)             | box plots                   |
| `position_nudge_to()`     | nudge          | x, y (fixed position)           | data labels                 |
| `position_nudge_center()` | nudge          | x, y (away or towards target)   | data labels                 |
| `position_nudge_line()`   | nudge          | x, y (away or towards target)   | data labels                 |

: Simple position functions defined in package 'ggpp'


### position_nudge_keep() and friends

Function `position_nudge_keep()` is like `ggplot2::position_nudge()` but
keeps (stores) the original *x* and *y* coordinates;
`position_nudge_keep()` can be used interchangeably with
`ggplot2::position_nudge()` with other geometries. Although possibly
only occasionally needed, `position_jitter_keep()` ,
`position_dodge_keep()` and `position_stack_keep()` are also made
available for completeness.

```{r}
set.seed(84532)
df <- data.frame(
  x = rnorm(20),
  y = rnorm(20, 2, 2),
  l = paste("label:", letters[1:20])
)

```

When used together with `geom_text_s()` or `geom_label_s()` segments
between a nudged label and the original position (here
indicated by a point) are drawn.

```{r}
ggplot(df, aes(x, y, label = l)) +
  geom_point() +
  geom_text_s(position = position_nudge_keep(x = 0.1)) +
  expand_limits(x = 2.5)
```

We can supress the drawing of segments.

```{r}
ggplot(df, aes(x, y, label = l)) +
  geom_point() +
  geom_text_s(position = position_nudge_keep(x = 0.1), add.segments = FALSE) +
  expand_limits(x = 2.5)
```

`position_nudge_keep()` and all other position functions described below
save the original positions in `data` in columns `x_orig` and `y_orig`
and the shifted positions in columns `x` and `y`. Because of this, they
can be used together with any 'ggplot2' geometry, even though these
geometries will ignore the stored original positions.

```{r}
ggplot(df, aes(x, y, label = l)) +
  geom_point() +
  geom_text(position = position_nudge(x = 0.3))
```

### position_nudge_to()

Function `position_nudge_to()` nudges to a given position instead of
using the same shift for each observation. Can be used to align labels
for points that are not themselves aligned. By left-justifying the label
text we ensure the alignment of labels differing in width.

```{r}
ggplot(df, aes(x, y, label = ifelse(x < 1, "", l) )) +
  geom_point() +
  geom_text_s(position = position_nudge_to(y = 2.3),
              colour = "red",
              arrow = arrow(length = unit(0.015, "npc")),
              angle = 90) +
  expand_limits(x = 3)
```
This function also supports distributing the positions in the available space,
with action `"spread".

```{r}
ggplot(df, aes(x, y, label = l)) +
  geom_point() +
  geom_text_s(data = function(x) {x[x$x > 0, ]},
              position = position_nudge_to(x = 2.3, y.action = "spread"),
              colour = "red",
              arrow = arrow(length = unit(0.015, "npc")),
              vjust = 0.5) +
  expand_limits(x = 2.7)
```

The space used can be expanded.

```{r}
ggplot(df, aes(x, y, label = l)) +
  geom_point() +
  geom_text_s(data = function(x) {x[x$x > 0, ]},
              position = position_nudge_to(x = 2.3, 
                                           y.action = "spread",
                                           y.expansion = 0.1),
              colour = "red",
              arrow = arrow(length = unit(0.015, "npc")),
              vjust = 0.5) +
  expand_limits(x = 2.7)
```

The range can be also set manually.

```{r}
ggplot(df, aes(x, y, label = l)) +
  geom_point() +
  geom_text_s(data = function(x) {x[x$x > 0, ]},
              position = position_nudge_to(x = 2.3, 
                                           y = c(-1.5, 5),
                                           y.action = "spread"),
              colour = "red",
              arrow = arrow(length = unit(0.015, "npc")),
              vjust = 0.5) +
  expand_limits(x = 2.7)
```

### position_nudge_center()

Function `position_nudge_center()` can nudge radially away from a focal
point if both `x` and `y` are passed as arguments, or towards opposite
sides of a boundary vertical or horizontal *virtual* line if only one of
`x` or `y` is passed an argument. By default, the "center" is the
centroid computed using `mean()`, but other functions or numeric values
can be passed to override it. When data are sparse, such nudging may be
effective in avoiding label overlaps, and achieving a visually pleasing
positioning.

In all cases nudging shifts the coordinates giving an *x* and/or *y*
position and also expands the limits of the corresponding scales to
include the nudged coordinate values.

```{r}
ggplot(df, aes(x, y, label = l)) +
  geom_point() +
  geom_text_s(position = 
                    position_nudge_center(x = -0.1, center_x = 0))
```

By default, split is away or towards the `mean()`. Here we allow
repulsion to separate the labels (compare with previous plot).

```{r}
ggplot(df, aes(x, y, label = l)) +
  geom_point() +
  geom_text_s(position = position_nudge_center(x = 0.1, direction = "split")) +
  expand_limits(x = c(-3, 3))
```

We set a different split point as a constant value.

```{r}
ggplot(df, aes(x, y, label = l)) +
  geom_point() +
  geom_text_s(position = 
                position_nudge_center(x = 0.1, 
                                      center_x = 1, 
                                      direction = "split")) +
  expand_limits(x = c(-3, 3))
```

We set a different split point as the value computed by a function
function, by name.

```{r}
ggplot(df, aes(x, y, label = l)) +
  geom_point() +
  geom_text_s(position = position_nudge_center(x = 0.1, 
                                               center_x = median, 
                                               direction = "split")) +
  expand_limits(x = c(-3, 3))
```

We set a different split point as the value computed by an anonymous
function. Here we split on the first quartile along *x*.

```{r}
ggplot(df, aes(x, y, label = l)) +
  geom_point() +
  geom_text_s(position = 
                    position_nudge_center(x = 0.1,
                                          center_x = function(x) {
                                            quantile(x, 
                                                     probs = 1/4, 
                                                     names = FALSE)
                                          },
                                          direction = "split")) +
  expand_limits(x = c(-3, 3))
```

### position_nudge_line()

Function `position_nudge_line()` nudges away from a line, which can be a
user supplied straight line as well as a smooth spline or a polynomial
fitted to the observations themselves. The nudging is away and
perpendicular to the local slope of the straight or curved line. It
relies on the same assumptions as linear regression, assuming that *x*
values are not subject to error. This in most cases prevents labels from
overlaping a curve fitted to the data, even if not exactly based on the
same model fit. When observations are sparse, this may be enough to
obtain a nice arrangement of data labels.

```{r}
set.seed(16532)
df <- data.frame(
  x = -10:10,
  y = (-10:10)^2,
  yy = (-10:10)^2 + rnorm(21, 0, 4),
  yyy = (-10:10) + rnorm(21, 0, 4),
  l = letters[1:21]
)
```

The first, simple example shows that `position_nudge_line()` has shifted
the direction of the nudging based on the alignment of the observations
along a line. One could, of course, have in this case passed suitable
values as arguments to *x* and *y* using `position_nudge()` from package
'ggplot2'. However, `position_nudge_line()` will work without change
irrespective of the slope or intercept along which the observations
fall.

```{r}
ggplot(df, aes(x, 2 * x, label = l)) +
  geom_point() +
  geom_abline(intercept = 0, slope = 2, linetype = "dotted") +
  geom_text(position = position_nudge_line(x = -0.5, y = -0.8))
```

With observations with high variation in *y*, a linear model fit may
need to be used. In this case fitted twice, once in `stat_smooth()` and
once in `position_nudge_line()`.

```{r}
ggplot(subset(df, x >= 0), aes(x, yyy)) +
  geom_point() +
  stat_smooth(method = "lm", formula = y ~ x) +
  geom_text(aes(label = l),
            vjust = "center", hjust = "center",
            position = position_nudge_line(x = 0, y = 1.2,
                                           method = "lm",
                                           direction = "split"))
```

With lower variation in *y*, we can pass to `line_nudge` a multiplier to
keep labels outside of the confidence band.

```{r}
ggplot(subset(df, x >= 0), aes(y, yy)) +
  geom_point() +
  stat_smooth(method = "lm", formula = y ~ x) +
  geom_text(aes(label = l),
            position = position_nudge_line(method = "lm",
                                           x = 3, y = 3, 
                                           line_nudge = 2.5,
                                           direction = "split"))
```

If we want the nudging based on an arbitrary straight line not computed
from `data`, we can pass the intercept and slope in a numeric vector of
length two as an argument to parameter `abline`.

```{r}
ggplot(subset(df, x >= 0), aes(y, yy)) +
  geom_point() +
  geom_abline(intercept = 0, slope = 1, linetype = "dotted") +
  geom_text(aes(label = l),
            position = position_nudge_line(abline = c(0, 1),
                                           x = 3, y = 3, 
                                           direction = "split"))
```

With observations that follow exactly a simple curve the defaults work
well to automate the nudging of individual data labels away from the
implicit curve. Positive values as arguments to `x` and `y` correspond
to above and inside the curve. One could, of course, pass also in this
case suitable values as arguments to *x* and *y* using
`position_nudge()` from package 'ggplot2', but these arguments would
need to be vectors with different *nudge* values for each observation.

```{r}
ggplot(df, aes(x, y, label = l)) +
  geom_point() +
  geom_line(linetype = "dotted") +
  geom_text(position = position_nudge_line(x = 0.6, y = 6))
```

Negative values passed as arguments to `x` and `y` correspond to labels
below and outside the curve.

```{r}
ggplot(df, aes(x, y, label = l)) +
  geom_point() +
  geom_line(linetype = "dotted") +
  geom_text(position = position_nudge_line(x = -0.6, y = -6))
```

When the observations include random variation along *y*, it is
important that the smoother used for the line added to a plot and that
passed to `position_nudge_line()` are similar. By default
`stat_smooth()` uses `"loess"` and `position_nudge_line()` with method
`"spline"`, `smooth.sline()`, which are a good enough match.

```{r}
ggplot(df, aes(x, yy)) +
  geom_point() +
  stat_smooth(method = "loess", formula = y ~ x) +
  geom_text(aes(label = l),
            position = position_nudge_line(x = 0.6, 
                                           y = 6,
                                           direction = "split"))
```

We can use other geometries.

```{r}
ggplot(df, aes(x, yy)) +
  geom_point() +
  stat_smooth(method = "loess", formula = y ~ x) +
  geom_label_s(aes(label = l),
                   position = position_nudge_line(x = 0.4, 
                                                  y = 4,
                                                  direction = "split")) +
  expand_limits(y = -12)
```

When fitting a polynomial, `"lm"` should be the argument passed to
`method` and a model formula preferably based on `poly()`, setting
`raw = TRUE`, as argument to `formula`. *Currently no other methods are
implemented in* `position_nudge_line()`.

```{r}
ggplot(df, aes(x, yy)) +
  geom_point() +
  stat_smooth(method = "lm", formula = y ~ poly(x, 2, raw = TRUE)) +
  geom_text_s(aes(label = l),
              position = position_nudge_line(method = "lm",
                                             x = 0.5, 
                                             y = 5,
                                             formula = y ~ poly(x, 2, raw = TRUE),
                                             direction = "split"))
```

## Combined position functions

As mentioned above, while _jitter_ and _dodge_ are normally use to displace the
observations being plotted to avoid overlaps and _stack_ used to construct
columns or areas that represent sums and contributions to a sum from different
groups of observations, _nudge_ is almost always used to displace the position
of data labels relative to observations they are attached to, to avoid overlaps.
Because _nudge_ plays in most cases a different role than _jitter_, _dodge_,
and _stack_, rather frequently it is useful to combine _nudge_ with the other
position displacements. Additional examples for this use case are given in the
`vignette("Combining repulsion and nudging")`.

Nudging text or labels in a dodged or stacked bar or column plot has not been
easy in 'ggplot2'. This is a rather frequent situation. However, within
'ggplot2' and the Grammar of Graphics this imposed some limitations on the
design of plots and the need to use workarounds to modify the data before
plotting.

Package 'ggpp' implements
*position* functions that implement the functionality of more than one of the
position functions available in 'ggplot2'. The examples below demonstrate their
use.

| Position                  | Main use       | Displacement                    | Most used with              |
|-----------------|-----------------|---------------------|------------------|
| `position_stacknudge()`   | stack + nudge  | combined, see above             | data labels in column plots |
| `position_fillnudge()`    | fill + nudge   | combined, see above             | data labels in column plots |
| `position_jitternudge()`  | jitter + nudge | combined, see above             | data labels in dot plots    |
| `position_dodgenudge()`   | dodge + nudge  | combined, see above             | data labels in column plots |
| `position_dodge2nudge()`  | dodge2 + nudge | combined, see above             | data labels in box plots    |

: Combined position functions defined in package 'ggpp'

**In all cases when labelling points or columns, to achieve the correct
alignment, the arguments passed to the equivalent position functions used
for observations and text must be the same, except for `x` and `y` which
determine amount of nudging to be added to the other dispalcements.**

All the position functions described in this section are compatible with
most geometries from `ggpplot2` and extensions. However, only those from
packages 'ggpp' and 'ggrepel' (> 0.9.1) plot segments and arrows linking
original and displaced positions using the saved original positions as
"kept", or saved, by these position functions.

```{r}
 df <- data.frame(x1 = c(1, 2, 1, 3, -1),
                  x2 = c("a", "a", "b", "b", "b"),
                  grp = c("A", "B", "C", "D", "E"))
```

### position_dodgenudge()

When labelling dodged columns, we can in addition apply nudging. Here we add a
label just outside the tip of each horizontal bar. With `direction = "split"`
we ensure that the text is correctly located in the _negative_ columns.

```{r}
ggplot(data = df, aes(x1, x2, group = grp)) +
   geom_col(aes(fill = grp), width = 0.8,
            position = position_dodge()) +
   geom_vline(xintercept = 0) +
   geom_text(
     aes(label = grp),
     position = position_dodgenudge(x = 0.09, direction = "split", width = 0.8)) +
   theme(legend.position = "none")
```

```{r}
 ggplot(data = df, aes(x2, x1, group = grp)) +
   geom_col(aes(fill = grp), width = 0.75,
            position = position_dodge(width = 0.75)) +
   geom_hline(yintercept = 0) +
   geom_text(aes(label = grp),
             position = position_dodgenudge(y = 0.1,
                                            direction = "split",
                                            width = 0.75),
             size = 3) +
   theme(legend.position = "none")
```

### position_stacknudge()

With horizontally stacked bar we add labels above each bar, at their horizontal
center with `vjust = 0.5`. (As with text, the justification is relative, with
vertical always meaning along the direction representing continuous data values
or counts.)

```{r}
 ggplot(data = df, aes(x1, x2, group = grp)) +
   geom_col(aes(fill = grp), width = 0.5) +
   geom_vline(xintercept = 0) +
   geom_text(
     aes(label = grp),
     position = position_stacknudge(vjust = 0.5, y = 0.33)) +
   theme(legend.position = "none")
```

### position_fillnudge()

Here we nudge the labels down from the top of each stacked bar with `vjust = 1`.

```{r}
 ggplot(data = subset(df, x1 >= 0), aes(x2, x1, group = grp)) +
   geom_col(aes(fill = grp), width=0.5, position = position_fill()) +
   geom_vline(xintercept = 0) +
   geom_text(
     aes(label = grp),
     position = position_fillnudge(vjust = 1, y = -0.05)) +
   theme(legend.position = "none")

```

### position_stack_minmax()

Functions `position_stack()` and `position_stack_nudge()` from package 'ggplot2'
displace the position of a variable mapped to only one of `y`, `ymax` or `ymin`.
So they cannot be used with geometries `geom_linerange()`, `geom_pointrange()`
or `geom_errorbar()`. Position function `position_stack_minmax()` can be used
with these geometries as it displaces `y`, `ymin` and `ymax` together by the
same distance. Like `position_stack_nudge()` it also supports nudging. Thus,
this new position function makes it possible to add error bars to each member of
a stacked colum plot. As nudging is also supported, it is possible, as shown
below, to also include an error bar for the total quantity represented by the
stacked bars.

```{r}
ggplot(birch_dw.df,
       aes(y = dry.weight * 1e-3, x = Density, fill = Part)) +
  stat_summary(geom = "col", fun = mean,
               position = "stack", alpha = 0.7, width = 0.67) +
  # error bars for each stack bar
  stat_summary(geom = "linerange", fun.data = mean_cl_normal,
               position = position_stack_minmax(x = -0.1)) +
  # error bar for the total
  stat_summary(data = birch.df, aes(y = (dwstem + dwroot) * 1e-3, fill = NULL),
               geom = "linerange", linewidth = 0.75,
               position = position_nudge(x = 0.1), fun.data = mean_cl_normal) +
  labs(y = "Seedling dry mass (g)") +
  scale_fill_grey(start = 0.7, end = 0.3) +
  facet_wrap(facets = vars(Container))
```

### position_jitternudge()

When combining jitter and nudge, we can nudge away from the jittered positions
or from the original positions. In this example, we nudge away from the jittered
positions with `nudge.from = "jittered"` and the direction of the nudge is on
either side of the original position with `direction = "split"`, with nudging
depending on the direction the jitter has moved the point.

```{r}
 jitter <- position_jitter(width = 0.2, height = 2, seed = 123)

 jitter_nudge <- position_jitternudge(width = 0.2, height = 2,
                                      seed = 123, x = 0.1,
                                      direction = "split",
                                      nudge.from = "jittered")
 ggplot(mpg[1:20, ],
        aes(cyl, hwy, label = drv)) +
   geom_point(position = jitter) +
   geom_text_s(position = jitter_nudge)

```

In this second example, the nudging is away from the original _x_-position while using the
jittered _y_-position, which gives vertically aligned labels.

```{r}
jitter <- position_jitter(width = 0.2, height = 2, seed = 123)

jitter_nudge <- position_jitternudge(width = 0.2, height = 2,
                                      seed = 123, x = 0.35,
                                      direction = "split",
                                      nudge.from = "original.x")
 ggplot(mpg[1:20, ],
        aes(cyl, hwy, label = drv)) +
   geom_point(position = jitter) +
   geom_text_s(position = jitter_nudge)
```

Jitter along the _y_-axis or the observed data can cause overlaps of data
labels. To avoid this, it is a good approach to use the repulsive geometries
`geom_text_repel()` or `geon_label_repel()` in combination with the position
functions from package 'ggrepel'.

### position_jitter_keep()

With no nudging but keeping the original positions we show how points have moved
with jitter.

```{r}
 ggplot(mpg[1:20, ],
        aes(cyl, hwy, label = drv)) +
   geom_point() +
   geom_point_s(position =
                position_jitter_keep(width = 0.3, height = 2, seed = 123),
                color = "red")
```

## Appendix A: More on density filtering

This appendix includes additional examples of the use of the local-density-based
filtering. These additional examples show the behaviour of these statistics with
with pseudo-random artificial data based on different theoretical probability
density functions. Here we use the `_filter` variants of these statistics. The
companion `_labels` variants are meant to be used together with the repulsive
geometries from package 'ggrepel'.

To start, we define a function to simplify the generation of random data sets
based on different probability distributions.

```{r}
make_data_tbl <- function(nrow = 100, rfun = rnorm, ...) {
  if (nrow %% 2) {
    nrow <- nrow + 1
  }
  
  set.seed(1001)
  
  tibble::tibble(
    x = rfun(nrow, ...),
    y = rfun(nrow, ...),
    group = rep(c("A", "B"), c(nrow / 2, nrow / 2))
  )
}
```

In all the examples in this section of the vignette we use colours to
demonstrate which data points are selected, but any other suitable aesthetic and
discrete scale can be used instead. By overriding the default `keep.sparse =
TRUE` with `keep.sparse = FALSE` we keep 1/3 of the observations in the denser
region of the plot. Although here we first plot all data points and later
overplot the selected ones this is not necessary.

```{r}
ggplot(data = make_data_tbl(300), aes(x, y)) +
  geom_point() +
  stat_dens2d_filter(colour = "red", 
                     keep.sparse = FALSE, 
                     keep.fraction = 1/3)
```

Here we highlight the observations split into three groups, each containing 1/3
of the observations in `data`. Each group, corresponding to a different local
density of observations.

```{r}
ggplot(data = make_data_tbl(300), aes(x, y)) +
  geom_point() +
  stat_dens2d_filter(colour = "red", 
                     keep.sparse = FALSE, 
                     keep.fraction = 1/3)+
  stat_dens2d_filter(colour = "blue", 
                     keep.fraction = 1/3)
```

The algorithm seems to work well also with other distributions, in this
example we use the uniform distribution.

```{r}
ggplot(data = make_data_tbl(300, rfun = runif), aes(x, y)) +
  geom_point() +
  stat_dens2d_filter(colour = "red", keep.fraction = 1/2)
```

One example with the gamma distribution, which is asymmetric.

```{r}
ggplot(data = make_data_tbl(300, rfun = rgamma, shape = 2), 
       aes(x, y)) +
  geom_point() +
  stat_dens2d_filter(colour = "red", keep.fraction = 1/3)
```

## Appendix B: try_tibble()

### Time series

Several different formats for storing time series data are used in R. Here we
use in the examples objects of class `ts` but several other classes are
supported as `try.xts()` from package 'xts' is used internally. Package
'ggplot2' supports plotting with variables containing date-times represented as
objects of classes `POSIXct` or `Date`. Functions `try_data_frame()` and
`try_tibble()` are used in 'ggpp' internally in the definition of `ggplot()`
method for time series data. Non-the-less, `try_data_frame()` and `try_tibble()`
are exported. The two functions differ only in the class of the returned object.
The time variable in the returned data frame or tibble by default belongs to
class `Date` or to class `POSIXct` depending on the time step of the time series
object. In the examples below we use `try_tibble()` but they work unchanged
with `try_data_frame()`.

In the first example we convert a quarterly time series into a tibble.

```{r}
print(austres)
class(austres)
austres.df <- try_tibble(austres)
class(austres.df)
head(austres.df, 4)
```

The next chunk demonstrates that when passing `as.numeric = TRUE` to override the default, numeric times are expressed as years with decimal fractions in the returned data frame.

```{r}
austres.df <- try_tibble(austres, as.numeric = TRUE)
head(austres.df, 4)
```

This additional example is for a series of yearly values.

```{r}
class(lynx)
lynx.df <- try_tibble(lynx)
class(lynx.df)
head(lynx.df, 3)
```

In some cases there can be small rounding errors or the time steps can be shorter
that the default `time.resolution = "month"`. We should in those cases pass
an argument to `time.resolution` to override the default rounding. Rounding is
done with `lubridate::round_date()` and the argument from `time.resolution`
passed to its parameter `unit`, except if it is `NULL` or `NA`, in which case
no rounding is applied.

```{r}
lynx.df <- try_tibble(lynx, time.resolution = NULL)
head(lynx.df, 3)
```

In addition we can convert the POSIXct values into numeric values in
calendar years plus a decimal fraction.

```{r}
lynx_n.df <- try_tibble(lynx, time.resolution = "year", as.numeric = TRUE)
lapply(lynx_n.df, "class")
head(lynx_n.df, 3)
```

### Other classes

Functions `try_data_frame()` and `try_tibble()` attempt to handle gracefully
objects that are not time series.

```{r}
try_tibble(1:5)
```

```{r}
try_tibble(letters[1:5])
```

```{r}
try_tibble(factor(letters[1:5]))
```

```{r}
try_tibble(list(x = rep(1,5), y = 1:5))
```

```{r}
try_tibble(data.frame(x = rep(1,5), y = 1:5))
```

```{r}
try_tibble(matrix(1:10, ncol = 2))
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weather-data.R
\docType{data}
\name{weather_18_june_2019.df}
\alias{weather_18_june_2019.df}
\title{Weather data}
\format{
A tibble with 18 columns and 1440 rows.
}
\usage{
weather_18_june_2019.df
}
\description{
A data set containing weather data measured in Viikki, Helsinki,
Finland. Values for all variables are means of 12 readings at 5 seconds
intervals. Sun angles were computed with R package 'photobiology'.
}
\details{
The variables are as follows:
\itemize{
  \item time (yyyy-mm-dd hh:mm:ss)
  \item PAR_umol (umol m-2 s-1)
  \item PAR_diff_fr (/1)
  \item global_watt (W m-2)
  \item day_of_year
  \item month_of_year
  \item month_name
  \item calendar_year
  \item solar_time (h)
  \item sun_elevation (degrees above horizon)
  \item sun_azimuth (degrees)
  \item was_sunny (T/F)
  \item wind_speed (m s-1)
  \item wind_direction (degrees)
  \item air_temperature_C (C)
  \item air_RH (%)
  \item air_DP (C)
  \item air_pressure
}
}
\examples{
names(weather_18_june_2019.df)
head(weather_18_june_2019.df)
nrow(weather_18_june_2019.df)

}
\references{
P. J. Aphalo, unpublished data.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat-apply.R
\name{stat_apply_group}
\alias{stat_apply_group}
\alias{stat_summary_xy}
\alias{stat_centroid}
\title{Apply a function to x or y values}
\usage{
stat_apply_group(
  mapping = NULL,
  data = NULL,
  geom = "line",
  .fun.x = NULL,
  .fun.x.args = list(),
  .fun.y = NULL,
  .fun.y.args = list(),
  position = "identity",
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = TRUE,
  ...
)

stat_summary_xy(
  mapping = NULL,
  data = NULL,
  geom = "point",
  .fun.x = NULL,
  .fun.x.args = list(),
  .fun.y = NULL,
  .fun.y.args = list(),
  position = "identity",
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = TRUE,
  ...
)

stat_centroid(
  mapping = NULL,
  data = NULL,
  geom = "point",
  .fun = NULL,
  .fun.args = list(),
  position = "identity",
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = TRUE,
  ...
)
}
\arguments{
\item{mapping}{The aesthetic mapping, usually constructed with
\code{\link[ggplot2]{aes}}. Only needs to be set at the layer level if you
are overriding the plot defaults.}

\item{data}{A layer specific dataset - only needed if you want to override
the plot defaults.}

\item{geom}{The geometric object to use display the data}

\item{.fun.x, .fun.y, .fun}{function to be applied or the name of the function
to be applied as a character string.}

\item{.fun.x.args, .fun.y.args, .fun.args}{additional arguments to be passed to
the function as a named list.}

\item{position}{The position adjustment to use for overlapping points on this
layer}

\item{na.rm}{a logical value indicating whether NA values should be stripped
before the computation proceeds.}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
never includes, and \code{TRUE} always includes.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics, rather
than combining with them. This is most useful for helper functions that
define both data and aesthetics and shouldn't inherit behaviour from the
default plot specification, e.g. \code{\link[ggplot2]{borders}}.}

\item{...}{other arguments passed on to \code{\link[ggplot2]{layer}}. This
can include aesthetics whose values you want to set, not map. See
\code{\link[ggplot2]{layer}} for more details.}
}
\value{
A data frame with the same variables as the data input, with either a
  single or multiple rows, with the values of \code{x} and \code{y} variables
  replaced by the values returned by the applied functions, or possibly
  filled with \code{NA} if no function was supplied or available by default.
  If the applied function returns a named vector, the names are copied into
  columns \code{x.names} and/or \code{y.names}. If the summary function
  applied returns a one row data frame, it will be column bound keeping
  the column names, but overwritting columns x and/or y with y from the
  summary data frame. In the names returned by \code{.fun.x} the letter
  "y" is replaced by "x". These allows the use of the same functions as in
  \code{ggplot2::stat_summary()}.

  \describe{
  \item{x}{x-value as returned by \code{.fun.x}, with names removed}
  \item{y}{y-value as returned by \code{.fun.y}, with names removed}
  \item{x.names}{if the x-value returned by \code{.fun.x} is named, these names}
  \item{y.names}{if the y-value returned by \code{.fun.y} is named, these names}
  \item{xmin, xmax}{values returned by \code{.fun.x} under these names, if present}
  \item{ymin, ymax}{values returned by \code{.fun.y} under these names, if present}
  \item{<other>}{additional values as returned by \code{.fun.y} under other names}
  }
}
\description{
\code{stat_summary_xy()} and \code{stat_centroid()} are similar to
\code{ggplot2::stat_summary()} but summarize both \code{x} and \code{y}
values in the same plot layer. Differently to \code{stat_summary()} no
grouping based on data \code{values} is done; the grouping respected is that
already present based on mappings to aesthetics. This makes it possible to
highlight the actual location of the centroid with \code{geom_point()},
\code{geom_text()}, and similar geometries. Instead, if we use
\code{geom_rug()} they are only a convenience avoiding the need to add two
separate layers and flipping one of them using \code{orientation = "y"}.
}
\details{
\code{stat_apply_group} applies functions to data.
When possible it is preferable to use transformations through scales or
summary functions such as \code{ggplot2::stat_summary()},
\code{stat_summary_xy()} or \code{stat_centroid()}. There are some
computations that are not scale transformations but are not usual summaries
either, as the number of data values does not decrease all the way to one row
per group. A typical case for a summary is the computation of quantiles. For
transformations are cumulative ones, e.g., using \code{cumsum()},
\code{runmed()} and similar functions. Obviously, it is always possible to
apply such functions to the data before plotting and passing them to a single
layer function. However, it can be useful to apply such functions on-the-fly
to ensure that grouping is consistent between computations and aesthetics.
One particularity of these statistics is that they can apply simultaneously
different functions to \code{x} values and to \code{y} values when needed. In
contrast to these statistics, \code{\link[ggplot2]{geom_smooth}} applies a
function that takes both \code{x} and \code{y} values as arguments.

These four statistics are similar. They differ on whether they
  return a single or multiple rows of data per group.
}
\note{
The applied function(s) must accept as first argument a vector that
  matches the variables mapped to \code{x} or \code{y} aesthetics. For
  \code{stat_summary_xy()} and \code{stat_centroid()} the function(s) to be
  applied is(are) expected to return a vector of length 1 or a data frame
  with only one row, as \code{mean_se()}, \code{mean_cl_normal()}
  \code{mean_cl_boot()}, \code{mean_sdl()} and \code{median_hilow()} from
  'ggplot2' do.

  For \code{stat_apply_group} the vectors returned by the
  the functions applied to \code{x} and \code{y} must be of exactly the same
  length. When only one of \code{.fun.x} or \code{.fun.y} are passed a
  function as argument, the other variable in the returned data is filled
  with \code{NA_real_}. If other values are desired, they can be set by means
  of a user-defined function.
}
\examples{
set.seed(123456)
my.df <- data.frame(X = rep(1:20,2),
                    Y = runif(40),
                    category = rep(c("A","B"), each = 20))

# make sure rows are ordered for X as we will use functions that rely on this
my.df <- my.df[order(my.df[["X"]]), ]

# Centroid
ggplot(my.df, aes(x = X, y = Y, colour = category)) +
  stat_centroid(shape = "cross", size = 6) +
  geom_point()

ggplot(my.df, aes(x = X, y = Y, colour = category)) +
  stat_centroid(geom = "rug", linewidth = 1.5, .fun = median) +
  geom_point()

ggplot(my.df, aes(x = X, y = Y, colour = category)) +
  stat_centroid(geom = "text", aes(label = category)) +
  geom_point()

ggplot(my.df, aes(x = X, y = Y, colour = category)) +
  stat_summary_xy(geom = "pointrange",
                  .fun.x = mean, .fun.y = mean_se) +
  geom_point()

# quantiles
ggplot(my.df, aes(x = X, y = Y, colour = category)) +
  geom_point() +
  stat_apply_group(geom = "rug", .fun.y = quantile, .fun.x = quantile)

ggplot(my.df, aes(x = X, y = Y)) +
  geom_point() +
  stat_apply_group(geom = "rug", sides = "lr", color = "darkred",
                   .fun.y = quantile) +
  stat_apply_group(geom = "text", hjust = "right", color = "darkred",
                   .fun.y = quantile,
                   .fun.x = function(x) {rep(22, 5)}, # set x to 22
                   mapping = aes(label = after_stat(y.names))) +
                   expand_limits(x = 21)

my.probs <- c(0.25, 0.5, 0.75)
ggplot(my.df, aes(x = X, y = Y, colour = category)) +
  geom_point() +
  stat_apply_group(geom = "hline",
                   aes(yintercept = after_stat(y)),
                   .fun.y = quantile,
                   .fun.y.args = list(probs = my.probs))

# cummulative summaries
ggplot(my.df, aes(x = X, y = Y, colour = category)) +
  stat_apply_group(.fun.x = function(x) {x},
                   .fun.y = cummax)

ggplot(my.df, aes(x = X, y = Y, colour = category)) +
  stat_apply_group(.fun.x = cumsum, .fun.y = cumsum)

# diff returns a shorter vector by 1 for each group
ggplot(my.df, aes(x = X, y = Y, colour = category)) +
  stat_apply_group(.fun.x = function(x) {x[-1L]},
                   .fun.y = diff, na.rm = TRUE)

# Running summaries
ggplot(my.df, aes(x = X, y = Y, colour = category)) +
  geom_point() +
  stat_apply_group(.fun.x = function(x) {x},
                   .fun.y = runmed, .fun.y.args = list(k = 5))

# Rescaling per group
ggplot(my.df, aes(x = X, y = Y, colour = category)) +
  stat_apply_group(.fun.x = function(x) {x},
                   .fun.y = function(x) {(x - min(x)) / (max(x) - min(x))})

# inspecting the returned data
if (requireNamespace("gginnards", quietly = TRUE)) {
  library(gginnards)

  ggplot(my.df, aes(x = X, y = Y, colour = category)) +
    stat_centroid(.fun = mean_se, geom = "debug_group")

  ggplot(my.df, aes(x = X, y = Y, colour = category)) +
    stat_summary_xy(.fun.y = mean_se, geom = "debug_group")

  ggplot(my.df, aes(x = X, y = Y, colour = category)) +
    stat_apply_group(.fun.y = cumsum, geom = "debug_group")

  ggplot(my.df, aes(x = X, y = Y, colour = category)) +
    geom_point() +
    stat_apply_group(geom = "debug_group",
                    .fun.x = quantile,
                    .fun.x.args = list(probs = my.probs),
                    .fun.y = quantile,
                   .fun.y.args = list(probs = my.probs))
}

}
\references{
Answers to question "R ggplot on-the-fly calculation by grouping variable" at
\url{https://stackoverflow.com/questions/51412522}.
}
\concept{summary stats}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggpp-legend-draw.R
\name{ggpp_draw_key}
\alias{ggpp_draw_key}
\alias{draw_key_text_s}
\alias{draw_key_label_s}
\title{Key glyphs for legends}
\usage{
draw_key_text_s(data, params, size)

draw_key_label_s(data, params, size)
}
\arguments{
\item{data}{A single row data frame containing the scaled aesthetics to
display in this key}

\item{params}{A list of additional parameters supplied to the geom.}

\item{size}{Width and height of key in mm.}
}
\value{
A grid grob.
}
\description{
Each geom has an associated function that draws the key when the geom needs
to be displayed in a legend. These functions are called `draw_key_*()`, where
`*` stands for the name of the respective key glyph. The key glyphs can be
customized for individual geoms by providing a geom with the `key_glyph`
argument (see [`layer()`] or examples below.)
}
\examples{
p <- ggplot(economics, aes(date, psavert, color = "savings rate"))
# key glyphs can be specified by their name
p + geom_line(key_glyph = "timeseries")

# key glyphs can be specified via their drawing function
p + geom_line(key_glyph = draw_key_rect)

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-quadrant-lines.R
\name{geom_quadrant_lines}
\alias{geom_quadrant_lines}
\alias{geom_vhlines}
\title{Reference lines: horizontal plus vertical, and quadrants}
\usage{
geom_quadrant_lines(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  pool.along = c("none", "x", "y", "xy"),
  xintercept = 0,
  yintercept = 0,
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = FALSE,
  ...
)

geom_vhlines(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  xintercept = NULL,
  yintercept = NULL,
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = FALSE,
  ...
)
}
\arguments{
\item{mapping}{The aesthetic mapping, usually constructed with
\code{\link[ggplot2]{aes}}. Only needs to be set at the layer level if you
are overriding the plot defaults.}

\item{data}{A layer specific data set - only needed if you want to override
the plot defaults.}

\item{stat}{The statistic object to use display the data}

\item{position}{The position adjustment to use for overlapping points on this
layer}

\item{pool.along}{character, one of \code{"none"}, \code{"x"}, \code{"y"}, or
\code{"xy"} indicating whether to plot or not lines separating quadrants.}

\item{xintercept, yintercept}{numeric vectors the coordinates of the origin of the
quadrants.}

\item{na.rm}{a logical indicating whether NA values should be stripped before
the computation proceeds.}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
never includes, and \code{TRUE} always includes.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics, rather
than combining with them. This is most useful for helper functions that
define both data and aesthetics and should not inherit behaviour from the
default plot specification, e.g. \code{\link[ggplot2]{borders}}.}

\item{...}{other arguments passed on to \code{\link[ggplot2]{layer}}. This
can include aesthetics whose values you want to set, not map. See
\code{\link[ggplot2]{layer}} for more details.}
}
\value{
A plot layer instance.
}
\description{
\code{geom_vhlines()} adds in a single layer both vertical and horizontal
guide lines. Can be thought of as a convenience function that helps with
producing consistent vertical and horizontal guide lines. It behaves like
\code{geom_vline()} and \code{geom_hline()}.
\code{geom_quadrant_lines()} displays the boundaries of four quadrants
with an arbitrary origin. The quadrants are specified in the same way as
in \code{stat_quadrant_counts()} and is intended to be used to add guide
lines consistent with the counts by quadrant computed by this stat.
}
\details{
While \code{geom_vhlines()} does not provide defaults for the
  intercepts and accepts vectors of length > 1, \code{geom_quadrant_lines()}
  sets by default the intercepts to zero producing the natural quadrants and
  only accepts vectors of length one per panel. That is \code{geom_vhlines()}
  can be used to plot a grid while \code{geom_quadrant_lines()} plots at
  most one vertical and one horizontal line. In the case of
  \code{geom_quadrant_lines()} the pooling along axes can be specified in the
  same way as in \code{\link{stat_quadrant_counts}()}.
}
\examples{
# generate artificial data
set.seed(4321)
x <- 1:100
y <- rnorm(length(x), mean = 10)
my.data <- data.frame(x, y)

ggplot(my.data, aes(x, y)) +
  geom_quadrant_lines() +
  geom_point()

ggplot(my.data, aes(x, y)) +
  geom_quadrant_lines(linetype = "dotted") +
  geom_point()

ggplot(my.data, aes(x, y)) +
  geom_quadrant_lines(xintercept = 50,
                      yintercept = 10,
                      colour = "blue") +
  geom_point()

ggplot(my.data, aes(x, y)) +
  geom_quadrant_lines(xintercept = 50,
                      pool.along = "y",
                      colour = "blue") +
  geom_point()

ggplot(my.data, aes(x, y)) +
  geom_vhlines(xintercept = c(25, 50, 75),
               yintercept = 10 ,
               linetype = "dotted",
               colour = "red") +
  geom_point() +
  theme_bw()

ggplot(my.data, aes(x, y)) +
  geom_vhlines(xintercept = c(25, 50, 75),
               yintercept = c(10, 8),
               linetype = "dotted",
               colour = "red") +
  geom_point() +
  theme_bw()

}
\seealso{
\code{\link[ggplot2]{geom_abline}}, the topic where
  \code{geom_vline()} and \code{geom_hline()} are described.

Other Functions for quadrant and volcano plots:
\code{\link[=stat_panel_counts]{stat_panel_counts()}},
\code{\link[=stat_quadrant_counts]{stat_quadrant_counts()}}
}
\concept{Functions for quadrant and volcano plots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat-dens2d-labels.r
\name{these2logical}
\alias{these2logical}
\title{Convert keep.these argument into logical vector}
\usage{
these2logical(these, data, these.target = "label")
}
\arguments{
\item{these}{character vector, integer vector, logical vector or
function that takes the variable mapped to the \code{label} aesthetic as
first argument and returns a character vector or a logical vector. These
rows from \code{data} are selected irrespective of the local density.}

\item{data}{data.frame The plot layer's data set.}

\item{these.target}{character The name of the variable (or aesthetic) in
which to select the target passed through \code{these}.}
}
\description{
Convert keep.these argument into logical vector
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-margin-arrow.r
\name{geom_x_margin_arrow}
\alias{geom_x_margin_arrow}
\alias{geom_y_margin_arrow}
\title{Reference arrows on the margins}
\usage{
geom_x_margin_arrow(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  ...,
  xintercept,
  sides = "b",
  arrow.length = 0.03,
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = FALSE
)

geom_y_margin_arrow(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  ...,
  yintercept,
  sides = "l",
  arrow.length = 0.03,
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = FALSE
)
}
\arguments{
\item{mapping}{The aesthetic mapping, usually constructed with
\code{\link[ggplot2]{aes}}. Only needs to be set at the layer level if you
are overriding the plot defaults.}

\item{data}{A layer specific dataset - only needed if you want to override
the plot defaults.}

\item{stat}{The statistical transformation to use on the data for this layer,
as a string.}

\item{position}{Position adjustment, either as a string, or the result of a
call to a position adjustment function.}

\item{...}{other arguments passed on to \code{\link[ggplot2]{layer}}. This
can include aesthetics whose values you want to set, not map. See
\code{\link[ggplot2]{layer}} for more details.}

\item{xintercept, yintercept}{numeric Parameters that control the position of
the marginal points. If these are set, data, mapping and show.legend are
overridden.}

\item{sides}{A string that controls which sides of the plot the rugs appear
on. It can be set to a string containing any combination of \code{"trbl"},
for top, right, bottom, and left.}

\item{arrow.length}{numeric value expressed in npc units for the length of the
arows inwards from the edge of the plotting area.}

\item{na.rm}{If \code{FALSE} (the default), removes missing values with a
warning.  If \code{TRUE} silently removes missing values.}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
never includes, and \code{TRUE} always includes.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics, rather
than combining with them. This is most useful for helper functions that
define both data and aesthetics and shouldn't inherit behaviour from the
default plot specification, e.g., \code{\link[ggplot2]{borders}}.}
}
\value{
A plot layer instance.
}
\description{
Small arrows on plot margins can supplement a 2d display with annotations.
Arrows can be used to highlight specific values along a margin. The geometries
\code{geom_x_margin_arrow()} and \code{geom_y_margin_arrow()} behave
similarly \code{geom_vline()} and \code{geom_hline()} and share their "double
personality" as both annotations and geometries.
}
\examples{
p <- ggplot(mtcars, aes(wt, mpg)) +
  geom_point()
p
p + geom_x_margin_arrow(xintercept = 3.5)
p + geom_y_margin_arrow(yintercept = c(18, 28, 15))
p + geom_x_margin_arrow(data = data.frame(x = c(2.5, 4.5)),
                        mapping = aes(xintercept = x))
p + geom_x_margin_arrow(data = data.frame(x = c(2.5, 4.5)),
                        mapping = aes(xintercept = x),
                        sides="tb")

}
\seealso{
Other Geometries for marginal annotations in ggplots:
\code{\link[=geom_x_margin_grob]{geom_x_margin_grob()}},
\code{\link[=geom_x_margin_point]{geom_x_margin_point()}}
}
\concept{Geometries for marginal annotations in ggplots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/position-nudge-line.R
\name{position_nudge_line}
\alias{position_nudge_line}
\title{Nudge labels away from a line}
\usage{
position_nudge_line(
  x = NA_real_,
  y = NA_real_,
  xy_relative = c(0.03, 0.03),
  abline = NULL,
  method = NULL,
  formula = y ~ x,
  direction = c("automatic", "none", "split"),
  line_nudge = 1,
  kept.origin = c("original", "none")
)
}
\arguments{
\item{x, y}{Amount of vertical and horizontal distance to move. A numeric
vector of length 1 or longer.}

\item{xy_relative}{Nudge relative to \emph{x} and \emph{y} data expanse, ignored unless
\code{x} and \code{y} are both \code{NA}s.}

\item{abline}{a vector of length two giving the intercept and slope.}

\item{method}{One of \code{"spline"}, \code{"lm"} or \code{"auto"}.}

\item{formula}{A model formula for \code{\link{lm}} when \code{method =
"lm"}. Ignored otherwise.}

\item{direction}{One of \code{"automatic"}, \code{"none"}, or \code{"split"}.}

\item{line_nudge}{A positive multiplier >= 1, increasing nudging away from
the curve or line compared to nudging from points.}

\item{kept.origin}{One of \code{"original"} or \code{"none"}.}
}
\value{
A \code{"Position"} object.
}
\description{
\code{position_nudge_line()} is generally useful for adjusting the starting
position of labels or text to be repelled while preserving the original
position as the start of the segments. The difference compared to
\code{\link{position_nudge_center}} is that the nudging is away from from a
line or curve fitted to the data points or supplied as coefficients. While
\code{position_nudge_center()} is most useful for "round-shaped", vertically-
or horizontally elongated clouds of points, \code{position_nudge_line()} is
most suitable when observations follow a linear or curvilinear relationship
between \emph{x} and \emph{y} values. In contrast to
\code{\link[ggplot2]{position_nudge}}, \code{position_nudge_line()} returns
in `data` both the original coordinates and the nudged coordinates.
}
\details{
The default amount of nudging is 3% of the spread of the data along
  \emph{x} and \emph{y} axes, which in most cases is good. In most cases it is best to
  apply nudging along a direction perpendicular to the line or curve, if this
  is the aim, passing an argument to only one of \code{x}, \code{y} or
  \code{xy_relative} will be enough. When \code{direction = "split"} nudging
  is away from an implicit line or curve on either side with positive
  nudging. The line or curve can be smooth spline or linear regression fitted
  on-the-fly to the data points, or a straight line defined by its
  coefficients passed to \code{abline}. The fitting is well defined only if
  the observations fall roughly on a curve or straight line that is monotonic
  in \code{y}. By means of \code{line_nudge} one can increment nudging away
  from the line or curve compared to away from the points, which is useful
  for example to keep labels outside of a confidence band. Direction defaults
  to \code{"split"} when \code{line_nudge} > 1, and otherwise to
  \code{"none"}.
}
\note{
For \code{method = "lm"} only model formulas corresponding to
  polynomials with no missing terms are supported. If using\code{\link{poly}}
  in the model formula, \code{raw = TRUE} is required.

  In practice, \code{x} and \code{y} should have the same sign for nudging to
  work correctly.

  This position is most useful when labeling points conforming a cloud along
  an arbitrary curve or line.
}
\examples{

set.seed(16532)
df <- data.frame(
  x = -10:10,
  y = (-10:10)^2,
  yy = (-10:10)^2 + rnorm(21, 0, 4),
  yyy = (-10:10) + rnorm(21, 0, 4),
  l = letters[1:21]
)

# Setting the nudging distance

ggplot(df, aes(x, y, label = l)) +
  geom_line(linetype = "dotted") +
  geom_point() +
  geom_text(position = position_nudge_line())

ggplot(df, aes(x, y, label = l)) +
  geom_line(linetype = "dotted") +
  geom_point() +
  geom_text_s(position = position_nudge_line())

ggplot(df, aes(x, y, label = l)) +
  geom_line(linetype = "dotted") +
  geom_point() +
  geom_text(position = position_nudge_line(xy_relative = -0.03))

ggplot(df, aes(x, y, label = l)) +
  geom_line(linetype = "dotted") +
  geom_point() +
  geom_text(position = position_nudge_line(x = 0.6, y = 3.2))

ggplot(df, aes(x, y, label = l)) +
  geom_line(linetype = "dotted") +
  geom_point() +
  geom_text(position = position_nudge_line(x = -0.6, y = -4))

# Other curves, using defaults

ggplot(df, aes(x, -y, label = l)) +
  geom_line(linetype = "dotted") +
  geom_point() +
  geom_text(position = position_nudge_line())

ggplot(subset(df, x >= 0), aes(y, sqrt(y), label = l)) +
  geom_line(linetype = "dotted") +
  geom_point() +
  geom_text(position = position_nudge_line())

# Points scattered near a curve or line, we use 'direction = "split"'

ggplot(df, aes(x)) +
  geom_line(aes(y = y), linetype = "dotted") +
  geom_point(aes(y = yy)) +
  geom_text(aes(y = yy, label = l),
            position = position_nudge_line(direction = "split"))

ggplot(subset(df, x >= 0), aes(y, yy)) +
  stat_smooth(method = "lm", formula = y ~ x) +
  geom_point() +
  geom_text(aes(label = l),
            position = position_nudge_line(direction = "split"))

# increasing the nudging for labels near the line

ggplot(subset(df, x >= 0), aes(y, yy)) +
  stat_smooth(method = "lm", formula = y ~ x) +
  geom_point() +
  geom_text(aes(label = l),
              position = position_nudge_line(line_nudge = 2,
                                             direction = "split"))

# fitting a linear model instead of the default spline

ggplot(subset(df, x >= 0), aes(y, yy)) +
  stat_smooth(method = "lm", formula = y ~ x) +
  geom_point() +
  geom_text(aes(label = l),
            position = position_nudge_line(method = "lm",
                                           direction = "split"))

ggplot(subset(df, x >= 0), aes(x, x^2)) +
  stat_smooth(method = "lm", formula = y ~ poly(x, 2, raw = TRUE)) +
  geom_point() +
  geom_text(aes(label = l),
            position = position_nudge_line(method = "lm",
                                           formula = y ~ poly(x, 2, raw = TRUE)))

}
\seealso{
\code{\link[ggplot2]{position_nudge}},
 \code{\link[ggrepel]{position_nudge_repel}}.

Other position adjustments:
\code{\link[=position_dodgenudge]{position_dodgenudge()}},
\code{\link[=position_dodgenudge_to]{position_dodgenudge_to()}},
\code{\link[=position_jitternudge]{position_jitternudge()}},
\code{\link[=position_nudge_center]{position_nudge_center()}},
\code{\link[=position_nudge_keep]{position_nudge_keep()}},
\code{\link[=position_nudge_to]{position_nudge_to()}},
\code{\link[=position_stacknudge]{position_stacknudge()}},
\code{\link[=position_stacknudge_to]{position_stacknudge_to()}}
}
\concept{position adjustments}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-table.R
\name{ttheme_set}
\alias{ttheme_set}
\alias{set_ttheme}
\title{Set default table theme}
\usage{
ttheme_set(table.theme = NULL)

set_ttheme(table.theme = NULL)
}
\arguments{
\item{table.theme}{NULL, list or function A gridExtra ttheme defintion, or
a constructor for a ttheme or NULL for default.}
}
\value{
A named list with the previous value of the option.
}
\description{
Set R option to the theme to use as current default. This function is
implemented differently but is used in the same way as
\code{ggplot2::theme_set()} but affects the default table-theme instead
of the plot theme.
}
\note{
The ttheme is set when a plot object is constructed, and consequently
the option setting does not affect rendering of ready built plot objects.
}
\examples{
library(dplyr)
library(tibble)

mtcars |>
  group_by(cyl) |>
  summarize(wt = mean(wt), mpg = mean(mpg)) |>
  ungroup() |>
  mutate(wt = sprintf("\%.2f", wt),
         mpg = sprintf("\%.1f", mpg)) -> tb

df <- tibble(x = 5.45, y = 34, tb = list(tb))

# Same as the default theme constructor
ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
  geom_point() +
  geom_table(data = df, aes(x = x, y = y, label = tb))

# set a new default
old_ttheme <- ttheme_set(ttheme_gtstripes)

ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
  geom_point() +
  geom_table(data = df, aes(x = x, y = y, label = tb))

# restore previous setting
ttheme_set(old_ttheme)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrap-labels.R
\name{wrap_labels}
\alias{wrap_labels}
\title{Wrap character strings in a vector}
\usage{
wrap_labels(x, width = 20, indent = 0, new.line = "\\n")
}
\arguments{
\item{x}{character vector, or an object which can be converted to a character
vector by \code{as.character}.}

\item{width}{a positive integer giving the target column for wrapping lines
in the output.}

\item{indent}{a positive or negative integer giving the indentation of the
first line in a member character string.}

\item{new.line}{character sting; use \code{"<br>"} for HTML encoded strings.}
}
\value{
A character vector of the same length as \code{x}, with new line
  characters inserted to wrap text lines longer than \code{width}. Names in
  \code{x} are preserved in the returned value, no names are added if none
  are present in \code{x}.
}
\description{
Wrap the members of a character vector to a given maximum width by inserting
new line characters at word boundaries.
}
\details{
Function \code{wrap_labels()} is a wrapper on \code{link{strwrap}}
that returns a vector of character strings instead of a list of vectors. In
addition to wrapping, indentation is supported. Wrapping is always at white
space, so \code{width = 0} wraps word by word.

Because the returned value is a character vector of the same length as the
input, this function can be used within a call to \code{aes()} when mapping a
character vector to the \code{label} aesthetic, as long as the character
strings will not be parsed into R expressions. It can be also used to wrap
the strings in a variable stored in a data frame.
}
\examples{
my.text <- c(A = "This is the first string",
             B = "This is the second string, which is longer")

wrap_labels(my.text, width = 20)
wrap_labels(unname(my.text), width = 20)

cat(wrap_labels(my.text, width = 20), sep = "\n--\n")
cat(wrap_labels(my.text, width = 20, indent = 2), sep = "\n--\n")
cat(wrap_labels(my.text, width = 20, indent = -2), sep = "\n--\n")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/try-data-frame.R
\name{try_data_frame}
\alias{try_data_frame}
\alias{try_tibble}
\title{Convert an R object into a tibble}
\usage{
try_data_frame(
  x,
  time.resolution = "month",
  as.numeric = FALSE,
  col.names = NULL
)

try_tibble(x, time.resolution = "month", as.numeric = FALSE, col.names = NULL)
}
\arguments{
\item{x}{An R object}

\item{time.resolution}{character The time unit to which the returned time
values will be rounded.}

\item{as.numeric}{logical If TRUE convert time to numeric, expressed as
fractional calendar years.}

\item{col.names}{character vector}
}
\value{
A \code{tibble::tibble} object, derived from \code{data.frame}.
}
\description{
This functions tries to convert any R object into a data.frame object.
If \code{x} is already a data.frame, it is returned as is. If it is
a list or a vector it is converted by means of \code{as.data.frame()}.
If of any other type, a conversion into an object of class \code{xts} is
attempted by means of \code{try.xts()} and if successful the \code{xts}
object is converted into a data frame with a variable \code{time}
containing times as \code{POSIXct} and the remaining data columns with
the time series data. In this conversion row names are stripped.
}
\note{
This function can be used to easily convert time series data into a
  format that can be easily plotted with package \code{ggplot2}.
  \code{try_tibble} is another name for \code{try_data_frame} which tracks
  the separation and re-naming of \code{data_frame} into
  \code{tibble::tibble} in the imported packages.
}
\section{Warning!}{
 The time zone was set to "UTC" by try.xts() in the test
  cases I used. Setting TZ to "UTC" can cause some trouble as several
  frequently used functions have as default the local or system TZ and will
  apply a conversion before printing or plotting time data, which in addition
  is affected by summer/winter time transitions. This should be taken into
  account as even for yearly data when conversion is to POSIXct a day (1st of
  January) will be set, but then shifted some hours if printed on a TZ
  different from "UTC". I recommend reading the documentation of package
  \code{\link[lubridate]{lubridate-package}} where the irregularities of time
  data and the difficulties they cause are very well described. In many cases
  when working with time series with yearly observations it is best to work
  with numeric values for years.
}

\examples{
class(lynx)
try_tibble(lynx)
try_tibble(lynx, as.numeric = TRUE)
try_tibble(lynx, "year")
class(austres)
try_tibble(austres)
try_tibble(austres, as.numeric = TRUE)
try_tibble(austres, "quarter")
class(cars)
try_tibble(cars)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/position-stack-nudge.R
\name{position_stacknudge}
\alias{position_stacknudge}
\alias{position_fillnudge}
\alias{position_stack_keep}
\alias{position_fill_keep}
\alias{position_stack_minmax}
\title{Combined positions stack and nudge}
\source{
\url{https://github.com/slowkow/ggrepel/issues/161}.
}
\usage{
position_stacknudge(
  vjust = 1,
  reverse = FALSE,
  x = 0,
  y = 0,
  direction = c("none", "split", "split.x", "split.y"),
  kept.origin = c("stacked", "original", "none")
)

position_fillnudge(
  vjust = 1,
  reverse = FALSE,
  x = 0,
  y = 0,
  direction = c("none", "split", "split.x", "split.y"),
  kept.origin = c("stacked", "original", "none")
)

position_stack_keep(vjust = 1, reverse = FALSE, kept.origin = "original")

position_fill_keep(vjust = 1, reverse = FALSE, kept.origin = "original")

position_stack_minmax(
  vjust = 1,
  reverse = FALSE,
  x = 0,
  y = 0,
  direction = c("none", "split", "split.x", "split.y"),
  kept.origin = c("stacked", "original", "none")
)
}
\arguments{
\item{vjust}{Vertical adjustment for geoms that have a position (like points
or lines), not a dimension (like bars or areas). Set to 0 to align with the
bottom, 0.5 for the middle, and 1 (the default) for the top.}

\item{reverse}{If TRUE, will reverse the default stacking order. This is
useful if you're rotating both the plot and legend.}

\item{x, y}{Amount of vertical and horizontal distance to move. A numeric
vector of length 1, or of the same length as rows there are in \code{data},
with nudge values in data rows order.}

\item{direction}{One of \code{"none"}, \code{"split"}, \code{"split.x"} or
\code{"split.y"}. A value of \code{"none"} replicates the behavior of
\code{\link[ggplot2]{position_nudge}}. At the moment \code{"split"} changes
the sign of the nudge at zero, which is suitable for column plots with
negative slices.}

\item{kept.origin}{One of \code{"original"}, \code{"stacked"} or
\code{"none"}.}
}
\value{
A \code{"Position"} object.
}
\description{
\code{position_stacknudge()} is useful when labelling plots such as stacked
bars, stacked columns, stacked lines, etc. In contrast to
\code{\link[ggplot2]{position_nudge}}, \code{position_stacknudge()} returns
in \code{data} both the original coordinates and the nudged coordinates.
}
\details{
\code{position_fillnudge()} is useful when labelling plots such as filled
bars, filled columns, filled lines, etc. In contrast to
\code{\link[ggplot2]{position_nudge}}, \code{position_fillnudge()} returns
in \code{data} both the original coordinates and the nudged coordinates.

The wrapper \code{position_nudge_keep()} has the same signature and
behaviour as \code{\link[ggplot2]{position_nudge}} nad provides an easier to
remember name when the need is only to have access to both the original and
nudged coordinates.

These position functions are backwards compatible with
\code{\link[ggplot2]{position_nudge}} but extends it by adding support for
stacking and for geometries that make use of the original position to draw
connecting segments or arrows.

The wrapper \code{position_stack_keep()} has the same signature and
behaviour as \code{\link[ggplot2]{position_stack}} and provides an easier to
remember name when the need is only to have access to both the original and
nudged coordinates.

The wrapper \code{position_fill_keep()} has the same signature and
behaviour as \code{\link[ggplot2]{position_fill}} and provides an easier to
remember name when the need is only to have access to both the original and
nudged coordinates.

The wrapper \code{position_stack_minmax()} has the same signature and
behaviour as \code{\link{position_stacknudge}} but stacks y, ymin and
ymax in parallel, making it possible to stack summaries with error bars,
works correctly with \code{geom_pointrange()}, \code{geom_linerange()} and
\code{geom_errorbar()}.
}
\examples{

df <- data.frame(x1 = c("a", "a", "b", "b", "b"),
                 x2 = c(1, 2, 1, 3, -1),
                 grp = c("some long name", "other name", "some name",
                         "another name", "some long name"))

# Add labels to a horizontal column plot (stacked by default)
ggplot(data = df, aes(x1, x2, group = grp)) +
  geom_col(aes(fill = grp), width=0.5) +
  geom_vline(xintercept = 0) +
  geom_text(
    aes(label = grp),
    position = position_stacknudge(vjust = 0.5, y = 0.3)) +
  theme(legend.position = "none")

# Add labels to a vertical column plot (stacked by default)
ggplot(data = df, aes(x2, x1, group = grp)) +
  geom_col(aes(fill = grp), width=0.5) +
  geom_vline(xintercept = 0) +
  geom_text(
    aes(label = grp),
    position = position_stacknudge(vjust = 0.5, x = -0.3),
    angle = 90) +
  theme(legend.position = "none")

# Add labels to a vertical column plot (stacked by default)
ggplot(data = subset(df, x1 >= 0), aes(x1, x2, group = grp)) +
  geom_col(aes(fill = grp), width=0.5, position = position_fill()) +
  geom_vline(xintercept = 0) +
  geom_text(
    aes(label = grp),
    position = position_fillnudge(vjust = 0.5, x = -0.3),
    angle = 90) +
  theme(legend.position = "none")

# Add label at a fixed distance from the top of each column slice
ggplot(data = df, aes(x1, x2, group = grp)) +
  geom_col(aes(fill = grp), width=0.5) +
  geom_vline(xintercept = 0) +
  geom_text(
    aes(label = grp),
    position = position_stacknudge(vjust = 1, y = -0.2)) +
  theme(legend.position = "none")

# Use geom_text_s(), geom_text_repel() or geom_label_repel() to link
# label to labelled segment or object with an arrow
ggplot(data = df, aes(x2, x1, group = grp)) +
  geom_col(aes(fill = grp), width=0.5) +
  geom_vline(xintercept = 0) +
  geom_text_s(
    aes(label = grp),
    position = position_stacknudge(vjust = 0.5, y = 0.35),
    vjust = "bottom") +
  theme(legend.position = "none")

ggplot(birch_dw.df,
       aes(y = dry.weight * 1e-3, x = Density, fill = Part)) +
   stat_summary(geom = "col", fun = mean,
                position = "stack", alpha = 0.7, width = 0.67) +
   stat_summary(geom = "linerange", fun.data = mean_cl_normal,
                position = position_stack_minmax()) +
  labs(y = "Seedling dry mass (g)") +
  scale_fill_grey(start = 0.7, end = 0.3) +
  facet_wrap(facets = vars(Container))

}
\seealso{
\code{\link[ggplot2]{position_nudge}},
  \code{\link[ggplot2]{position_stack}},
  \code{\link[ggrepel]{position_nudge_repel}}.

Other position adjustments:
\code{\link[=position_dodgenudge]{position_dodgenudge()}},
\code{\link[=position_dodgenudge_to]{position_dodgenudge_to()}},
\code{\link[=position_jitternudge]{position_jitternudge()}},
\code{\link[=position_nudge_center]{position_nudge_center()}},
\code{\link[=position_nudge_keep]{position_nudge_keep()}},
\code{\link[=position_nudge_line]{position_nudge_line()}},
\code{\link[=position_nudge_to]{position_nudge_to()}},
\code{\link[=position_stacknudge_to]{position_stacknudge_to()}}
}
\author{
Michał Krassowski, edited by Pedro J. Aphalo.
}
\concept{position adjustments}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/position-dodge-nudge-to.R,
%   R/position-dodge2nudge-to.R
\name{position_dodgenudge_to}
\alias{position_dodgenudge_to}
\alias{position_dodge2nudge_to}
\title{Dodge plus nudge labels to new positions}
\usage{
position_dodgenudge_to(
  width = 1,
  preserve = c("total", "single"),
  reverse = FALSE,
  x = NULL,
  y = NULL,
  x.action = c("none", "spread"),
  y.action = c("none", "spread"),
  x.distance = "equal",
  y.distance = "equal",
  x.expansion = 0,
  y.expansion = 0,
  kept.origin = c("dodged", "original", "none")
)

position_dodge2nudge_to(
  width = 1,
  preserve = c("total", "single"),
  padding = 0.1,
  reverse = FALSE,
  x = NULL,
  y = NULL,
  x.action = c("none", "spread"),
  y.action = c("none", "spread"),
  x.distance = "equal",
  y.distance = "equal",
  x.expansion = 0,
  y.expansion = 0,
  kept.origin = c("dodged", "original", "none")
)
}
\arguments{
\item{width}{Dodging width, when different to the width of the individual
elements. This is useful when you want to align narrow geoms with wider
geoms. See the examples.}

\item{preserve}{Should dodging preserve the total width of all elements at a
position, or the width of a single element?.}

\item{reverse}{If TRUE, will reverse the default dodging order.}

\item{x, y}{Coordinates of the destination position. A vector of mode
\code{numeric}, that is extended if needed, to the same length as rows
there are in \code{data}. The default, \code{NULL}, leaves the original
coordinates unchanged after dodging.}

\item{x.action, y.action}{character string, one of \code{"none"}, or
\code{"spread"}. With \code{"spread"} distributing the positions
within the range of argument \code{x} or \code{y}, if non-null, or the
range the variable mapped to \emph{x} or \code{y}, otherwise.}

\item{x.distance, y.distance}{character or numeric Currently only \code{"equal"} is
implemented.}

\item{x.expansion, y.expansion}{numeric vectors of length 1 or 2, as a
fraction of width of the range.}

\item{kept.origin}{One of \code{"original"}, \code{"dodged"} or
\code{"none"}.}

\item{padding}{Padding between elements at the same position. Elements are
shrunk by this proportion to allow space between them. Defaults to 0.1.}
}
\value{
A \code{"Position"} object.
}
\description{
Functions \code{position_dodgenudge_to()} and
\code{position_dodge2nudge_to()} are meant to complement
\code{position_dodge()} and \code{position_dodge2()} from 'ggplot2', adding
as a second action that of \code{position_nudge_to()}. These positions are
generally useful for adjusting the position of labels or text. As with other
position functions in this package, the original positions are preserved to
allow the text or labels to be linked back to their original position with a
segment or arrow.
}
\details{
These positions apply sequentially two actions, in the order they
  appear in their names. The applied dodge is similar to that by
  \code{\link[ggplot2]{position_dodge}} and
  \code{\link[ggplot2]{position_dodge2}} while nudging is different to that
  by \code{\link[ggplot2]{position_nudge}} and equal to that applied by
  \code{\link{position_nudge_to}}.

  The dodged and nudged to \code{x} and/or \code{y} values replace the
  original ones in \code{data}, while the original or the dodged coordinates
  are returned in \code{x_orig} and \code{y_orig}. Nudge values supported are
  those of \emph{mode} numeric, thus including dates and times when they
  match the mapped data.

  If the length of \code{x} and/or \code{y} is more than one but less than
  rows are present in the data, the vector is both recycled and reordered so
  that the nudges are applied sequentially based on the data values. If their
  length matches the number of rows in data, they are assumed to be already
  in data order.

  The intended use is to label dodged bars, boxplots or points with labels
  aligned. In this case, it is mandatory to use the same argument to
  \code{width} when passing \code{position_dodge()} to \code{geom_col()} and
  \code{position_dodgenudge_to()} to \code{geom_text()}, \code{geom_label()},
  \code{geom_text_s()}, \code{geom_label_s()} or their repulsive equivalents
  from package 'ggrepel'. Otherwise the arrows or segments will fail to
  connect to the labels.

  When applying dodging, the return of original positions instead of the
  dodged ones is achieved by passing \code{origin = "original"} instead of
  the default of \code{origin = "dodged"}.
}
\note{
Irrespective of the action, the ordering of rows in \code{data} is
  preserved.
}
\examples{

df <- data.frame(
  x = c(1,3,2,5,4,2.5),
  y = c(2, 3, 2.5, 1.8, 2.8, 1.5),
  grp = c("A", "A", "A", "B", "B", "B"),
  grp.inner = c("a", "b", "c", "a", "b", "c"),
  label = c("abc","cd","d","c","bcd","a")
)

# default is no nudging
ggplot(df, aes(grp, y, label = label, fill = label)) +
  geom_col(position = position_dodge(width = 0.92)) +
  geom_text(position = position_dodgenudge_to(width = 0.92),
            vjust = -0.2) +
  theme(legend.position = "none")

ggplot(df, aes(grp, y, label = label, fill = grp.inner)) +
  geom_col(position = position_dodge(width = 0.92)) +
  geom_text(position = position_dodgenudge_to(width = 0.92),
            vjust = -0.2)

ggplot(df, aes(grp, y, label = label, fill = label)) +
  geom_col(position = position_dodge2(width = 0.92)) +
  geom_text(position = position_dodge2nudge_to(width = 0.92),
            vjust = -0.2) +
  theme(legend.position = "none")

ggplot(df, aes(grp, y, label = label, fill = grp.inner)) +
  geom_col(position = position_dodge2(width = 0.92)) +
  geom_text(position = position_dodge2nudge_to(width = 0.92),
            vjust = -0.2)

# nudging all labels to a given y value
ggplot(df, aes(grp, y, label = label, fill = grp.inner)) +
  geom_col(position = position_dodge(width = 0.92)) +
  geom_text(position = position_dodgenudge_to(width = 0.92, y = 0.8))

ggplot(df, aes(grp, y, label = label, fill = grp.inner)) +
  geom_col(position = position_dodge2(width = 0.92)) +
  geom_text(position = position_dodge2nudge_to(width = 0.92, y = 0.8))

ggplot(df, aes(grp, y, label = label, fill = grp.inner)) +
  geom_col(position = position_dodge(width = 0.92)) +
  geom_text(position = position_dodgenudge_to(width = 0.92, y = 0.8))

ggplot(df[-1, ], aes(grp, y, label = label, fill = grp.inner)) +
  geom_col(position = position_dodge(width = 0.92)) +
  geom_text(position = position_dodgenudge_to(width = 0.92, y = 0.8))

ggplot(df[-1, ], aes(grp, y, label = label, fill = grp.inner)) +
  geom_col(position = position_dodge(width = 0.92, preserve = "total")) +
  geom_text(position = position_dodgenudge_to(width = 0.92, y = 0.8,
                                              preserve = "total"))

ggplot(df[-1, ], aes(grp, y, label = label, fill = grp.inner)) +
  geom_col(position = position_dodge(width = 0.92, preserve = "single")) +
  geom_text(position = position_dodgenudge_to(width = 0.92, y = 0.8,
                                              preserve = "single"))

ggplot(df[-1, ], aes(grp, y, label = label, fill = grp.inner)) +
  geom_col(position = position_dodge2(width = 0.92, preserve = "total")) +
  geom_text(position = position_dodge2nudge_to(width = 0.92, y = 0.8,
                                              preserve = "total"))

ggplot(df[-1, ], aes(grp, y, label = label, fill = grp.inner)) +
  geom_col(position = position_dodge2(width = 0.92, preserve = "single")) +
  geom_text(position = position_dodge2nudge_to(width = 0.92, y = 0.8,
                                              preserve = "single"))

}
\seealso{
\code{\link{position_nudge_to}},
  \code{\link[ggplot2]{position_dodge}},
  \code{\link[ggplot2]{position_dodge2}}.

Other position adjustments:
\code{\link[=position_dodgenudge]{position_dodgenudge()}},
\code{\link[=position_jitternudge]{position_jitternudge()}},
\code{\link[=position_nudge_center]{position_nudge_center()}},
\code{\link[=position_nudge_keep]{position_nudge_keep()}},
\code{\link[=position_nudge_line]{position_nudge_line()}},
\code{\link[=position_nudge_to]{position_nudge_to()}},
\code{\link[=position_stacknudge]{position_stacknudge()}},
\code{\link[=position_stacknudge_to]{position_stacknudge_to()}}
}
\concept{position adjustments}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-table.R
\name{ttheme_gtdefault}
\alias{ttheme_gtdefault}
\alias{ttheme_gtminimal}
\alias{ttheme_gtbw}
\alias{ttheme_gtplain}
\alias{ttheme_gtdark}
\alias{ttheme_gtlight}
\alias{ttheme_gtsimple}
\alias{ttheme_gtstripes}
\title{Table themes}
\usage{
ttheme_gtdefault(
  base_size = 10,
  base_colour = "black",
  base_family = "",
  parse = FALSE,
  padding = grid::unit(c(0.8, 0.6), "char"),
  text.alpha = NA,
  rules.alpha = NA,
  canvas.alpha = NA,
  ...
)

ttheme_gtminimal(
  base_size = 10,
  base_colour = "black",
  base_family = "",
  parse = FALSE,
  padding = grid::unit(c(0.5, 0.4), "char"),
  text.alpha = NA,
  rules.alpha = NA,
  canvas.alpha = NA,
  ...
)

ttheme_gtbw(
  base_size = 10,
  base_colour = "black",
  base_family = "",
  parse = FALSE,
  padding = grid::unit(c(1, 0.6), "char"),
  text.alpha = NA,
  rules.alpha = NA,
  canvas.alpha = NA,
  ...
)

ttheme_gtplain(
  base_size = 10,
  base_colour = "black",
  base_family = "",
  parse = FALSE,
  padding = grid::unit(c(0.8, 0.6), "char"),
  text.alpha = NA,
  rules.alpha = NA,
  canvas.alpha = NA,
  ...
)

ttheme_gtdark(
  base_size = 10,
  base_colour = "grey90",
  base_family = "",
  parse = FALSE,
  padding = grid::unit(c(0.8, 0.6), "char"),
  text.alpha = NA,
  rules.alpha = NA,
  canvas.alpha = NA,
  ...
)

ttheme_gtlight(
  base_size = 10,
  base_colour = "grey10",
  base_family = "",
  parse = FALSE,
  padding = grid::unit(c(0.8, 0.6), "char"),
  text.alpha = NA,
  rules.alpha = NA,
  canvas.alpha = NA,
  ...
)

ttheme_gtsimple(
  base_size = 10,
  base_colour = "grey10",
  base_family = "",
  parse = FALSE,
  padding = grid::unit(c(0.5, 0.4), "char"),
  text.alpha = NA,
  rules.alpha = NA,
  canvas.alpha = NA,
  ...
)

ttheme_gtstripes(
  base_size = 10,
  base_colour = "grey10",
  base_family = "",
  parse = FALSE,
  padding = grid::unit(c(0.8, 0.6), "char"),
  text.alpha = NA,
  rules.alpha = NA,
  canvas.alpha = NA,
  ...
)
}
\arguments{
\item{base_size}{numeric, default font size of text in table.}

\item{base_colour}{default font colour for text in table.}

\item{base_family}{default font family for text in table.}

\item{parse}{logical, behaviour for parsing text as plotmath.}

\item{padding}{length-2 unit vector specifying the horizontal and vertical
padding of text within each cell.}

\item{text.alpha, canvas.alpha, rules.alpha}{numeric in [0..1] Transparency
applied to table \code{base_colour}, to table body background
\code{fill} and rules \code{colour}, respectively.}

\item{...}{further arguments to control the gtable.}
}
\value{
A \code{list} object that can be used as \code{ttheme} in the
  construction of tables with functions from package 'gridExtra'.
}
\description{
Additional theme constructors for use with \code{\link{geom_table}}.
}
\details{
These wrapper functions are table theme (\code{ttheme}) constructors
  making it easier to change the style of tables created with
  \code{\link[gridExtra]{tableGrob}}. When passed as argument to
  \code{\link{geom_table}} the table theme's \code{base_colour},
  \code{base_family}, \code{base_colour} and \code{base_size} function as
  defaults for the text in the body of the table. They are overriden if the
  corresponding text-related aesthetics are mapped or set to a constant
  through the usual 'ggplot2' mechanisms. On the other hand the properties
  of the background fill, rules and column and row headings can be set only
  through the theme. The \code{ttheme} constructors defined in
  'ggpp' have formal parameters for \code{alpha} transparency of the
  text, background and rules. Transparency is useful as plot insets can
  accidentally overlap observations hiding them from view depending on the
  stacking order of plot layers.

  These theme constructors are wrappers on the constructors
  \code{gridExtra::ttheme_default()} and \code{gridExtra::ttheme_minimal()}.
  They can also be used directly with \code{\link[gridExtra]{grid.table}} if
  desired.
}
\examples{
library(dplyr)
library(tibble)

mtcars |>
  group_by(cyl) |>
  summarize(wt = mean(wt), mpg = mean(mpg)) |>
  ungroup() |>
  mutate(wt = sprintf("\%.2f", wt),
         mpg = sprintf("\%.1f", mpg)) -> tb

df <- tibble(x = 5.45, y = 34, tb = list(tb))

# Same as the default theme constructor
ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
  geom_point() +
  geom_table(data = df, aes(x = x, y = y, label = tb),
             table.theme = ttheme_gtdefault) +
  theme_classic()

# Minimal theme constructor
ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
  geom_point() +
  geom_table(data = df, aes(x = x, y = y, label = tb),
             table.theme = ttheme_gtminimal) +
  theme_classic()

# A theme with white background
ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
  geom_point() +
  geom_table(data = df, aes(x = x, y = y, label = tb),
             table.theme = ttheme_gtbw) +
  theme_bw()

# Base colour of theme overridden by aesthetic constant
ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
  geom_point() +
  geom_table(data = df, aes(x = x, y = y, label = tb),
             table.theme = ttheme_gtbw, colour = "darkblue") +
  theme_bw()

# A theme with dark background
ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
  geom_point() +
  geom_table(data = df, aes(x = x, y = y, label = tb),
             table.theme = ttheme_gtdark) +
  theme_dark()

# Base colour of theme overridden by aesthetic constant
ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
  geom_point() +
  geom_table(data = df, aes(x = x, y = y, label = tb),
             table.theme = ttheme_gtdark, colour = "yellow") +
  theme_dark()

# A theme with light background
ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
  geom_point() +
  geom_table(data = df, aes(x = x, y = y, label = tb),
             table.theme = ttheme_gtlight)

# Base colour of theme overridden by aesthetic constant
ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
  geom_point() +
  geom_table(data = df, aes(x = x, y = y, label = tb),
             table.theme = ttheme_gtlight, colour = "darkred")

# Base colour of theme overridden by aesthetic constant
ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
  geom_point() +
  geom_table(data = df, aes(x = x, y = y, label = tb),
             table.theme = ttheme_gtsimple)

# Default colour of theme superceded by aesthetic constant
ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
  geom_point() +
  geom_table(data = df, aes(x = x, y = y, label = tb),
             table.theme = ttheme_gtstripes) +
  theme_dark()

# Transparency of table background fill and grid lines colour
ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
  geom_point() +
  geom_table(data = df, aes(x = 1.5, y = y, label = tb),
             table.theme = ttheme_gtplain(canvas.alpha = 0.5,
                                          rules.alpha = 0.2)) +
  theme_classic()

# Transparency of table background fill and grid lines colour
# and table text base colour: black with 50\% transparency
ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
  geom_point() +
  geom_table(data = df, aes(x = 1.5, y = y, label = tb),
             table.theme = ttheme_gtplain(text.alpha = 0.5)) +
  theme_classic()

}
\concept{geometries for adding insets to ggplots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-label-npc.r, R/geom-text-npc.r
\name{geom_label_npc}
\alias{geom_label_npc}
\alias{geom_text_npc}
\title{Text with Normalised Parent Coordinates}
\usage{
geom_label_npc(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  ...,
  parse = FALSE,
  nudge_x = 0,
  nudge_y = 0,
  label.padding = grid::unit(0.25, "lines"),
  label.r = grid::unit(0.15, "lines"),
  label.size = 0.25,
  size.unit = "mm",
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = FALSE
)

geom_text_npc(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  ...,
  parse = FALSE,
  nudge_x = 0,
  nudge_y = 0,
  check_overlap = FALSE,
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = FALSE
)
}
\arguments{
\item{mapping}{The aesthetic mapping, usually constructed with
\code{\link[ggplot2]{aes}}. Only needs
to be set at the layer level if you are overriding the plot defaults.}

\item{data}{A layer specific data set - only needed if you want to override
the plot defaults.}

\item{stat}{The statistical transformation to use on the data for this layer,
as a string.}

\item{position}{Position adjustment, either as a string, or the result of a
call to a position adjustment function.}

\item{...}{other arguments passed on to \code{\link[ggplot2]{layer}}. This
can include aesthetics whose values you want to set, not map. See
\code{\link[ggplot2]{layer}} for more details.}

\item{parse}{If TRUE, the labels will be parsed into expressions and
displayed as described in \code{?plotmath}.}

\item{nudge_x, nudge_y}{Horizontal and vertical adjustment to nudge labels by.
Useful for offsetting text from points, particularly on discrete scales.}

\item{label.padding}{Amount of padding around label. Defaults to 0.25 lines.}

\item{label.r}{Radius of rounded corners. Defaults to 0.15 lines.}

\item{label.size}{Size of label border, in mm.}

\item{size.unit}{How the `size` aesthetic is interpreted: as millimetres
(`"mm"`, default), points (`"pt"`), centimetres (`"cm"`), inches (`"in"`),
or picas (`"pc"`).}

\item{na.rm}{If \code{FALSE} (the default), removes missing values with a
warning.  If \code{TRUE} silently removes missing values.}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
never includes, and \code{TRUE} always includes.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics, rather
than combining with them. This is most useful for helper functions that
define both data and aesthetics and shouldn't inherit behaviour from the
default plot specification, e.g. \code{\link[ggplot2]{borders}}.}

\item{check_overlap}{If `TRUE`, text that overlaps previous text in the same
layer will not be plotted.}
}
\description{
\code{geom_text_npc()} adds text directly to the plot.
\code{geom_label_npc()} draws a rectangle behind the text, making it easier
to read. The difference is that \code{x} and \code{y} mappings are expected
to be given in \code{npc} graphic units, using pseudo-aesthetics. Their
intended use is to add annotations to a plot.
}
\details{
These geoms are identical to 'ggplot2'
  \code{\link[ggplot2]{geom_text}} and \code{\link[ggplot2]{geom_label}}
  except that they interpret \code{npcx} and \code{npcy} positions in
  \code{npc} units. They translate \code{npcx} and \code{npcy} coordinates
  using a pseudo-aesthetic with a fixed scale, the translation is done
  separately for each plot panel. All aesthetics other than \emph{x} and
  \emph{y} and grouping work as in normal geoms. These include
  \code{linetype} and \code{angle} in \code{geom_label_npc()}.
}
\section{Alignment}{
 With textual positions and groups a shift is added to
  successive labels to avoid overlaps. The shift is based on grouping,
  however unused levels are not dropped. In plots with faceting, if not
  all groups appear in each panel, there will be blank spaces in between
  labels. To solve this pass numeric values for the npc coordinates of each
  label instead of character strings.

 You can modify text alignment with the \code{vjust} and
  \code{hjust} aesthetics. These can either be a number between 0
  (right/bottom) and 1 (top/left) or a character (\code{"left"},
  \code{"middle"}, \code{"right"}, \code{"bottom"}, \code{"center"},
  \code{"top"}). In addition, you can use special alignments for
  justification including \code{"position"}, \code{"inward"} and
  \code{"outward"}. Inward always aligns text towards the center of the
  plotting area, and outward aligns it away from the center of the plotting
  area. If tagged with \code{_mean} or \code{_median} (e.g.,
  \code{"outward_mean"}) the mean or median of the data in the panel along
  the corresponding axis is used as center. If the characters following the
  underscore represent a number (e.g., \code{"outward_10.5"}) the reference
  point will be this value in data units. Position justification is computed
  based on the direction of the displacement of the position of the label so
  that each individual text or label is justified outwards from its original
  position. The default justification is \code{"position"}.

  If no position displacement is applied, or a position function defined in
  'ggplot2' is used, these geometries behave similarly to the corresponding
  ones from package 'ggplot2' with a default justification of \code{0.5} and
  no segment drawn.
}

\section{Plot boundaries and clipping}{
 Note that when you change the scale
  limits for \emph{x} and/or \emph{y} of a plot, text labels stay the same
  size, as determined by the \code{size} aesthetic, given in millimetres. The
  actual size as seen in the plotted output is decided during the rendering
  of the plot to a graphics device. Limits are expanded only to include the
  anchor point of the labels because the "width" and "height" of a text
  element are 0 (as seen by ggplot2). Text labels do have height and width,
  but in grid units, not data units.
}

\examples{

df <- data.frame(
  x = c(0, 0, 1, 1, 0.5),
  x.chr = c("left", "left", "right", "right", "center"),
  y = c(0, 1, 0, 1, 0.5),
  y.chr = c("bottom", "top", "bottom", "top", "middle"),
  text = c("bottom-left", "top-left", "bottom-right", "top-right", "center-middle")
)

ggplot(df) +
  geom_text_npc(aes(npcx = x, npcy = y, label = text))

ggplot(df) +
  geom_text_npc(aes(npcx = x.chr, npcy = y.chr, label = text))

ggplot(df) +
  geom_text_npc(aes(npcx = x.chr, npcy = y.chr, label = text),
                angle = 90)

ggplot(data = mtcars, mapping = aes(wt, mpg)) +
  geom_point() +
  geom_text_npc(data = df, aes(npcx = x, npcy = y, label = text))

ggplot(data = mtcars, mapping = aes(wt, mpg)) +
  geom_point() +
  geom_text_npc(data = df, aes(npcx = x, npcy = y, label = text)) +
  expand_limits(y = 40, x = 6)

ggplot(data = mtcars) +
  geom_point(mapping = aes(wt, mpg)) +
  geom_label_npc(data = df, aes(npcx = x, npcy = y, label = text))

ggplot(data = mtcars) +
  geom_point(mapping = aes(wt, mpg)) +
  geom_label_npc(data = df, aes(npcx = x.chr, npcy = y.chr, label = text),
                 angle = 90) # ignored by ggplot2 < 3.5.0

}
\seealso{
\code{\link[ggplot2]{geom_text}} and
   \code{\link[ggplot2]{geom_label}} for additional details.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-label-pairwise.r, R/geom-text-pairwise.R
\name{geom_label_pairwise}
\alias{geom_label_pairwise}
\alias{geom_text_pairwise}
\title{Label pairwise comparisons}
\usage{
geom_label_pairwise(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  ...,
  orientation = NA,
  parse = FALSE,
  nudge_x = 0,
  nudge_y = 0,
  default.colour = NULL,
  default.color = default.colour,
  colour.target = "all",
  color.target = colour.target,
  default.alpha = NA,
  alpha.target = "segment",
  label.padding = grid::unit(0.25, "lines"),
  label.r = grid::unit(0.15, "lines"),
  segment.linewidth = 0.5,
  arrow = NULL,
  size.unit = "mm",
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = FALSE
)

geom_text_pairwise(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  ...,
  orientation = NA,
  parse = FALSE,
  nudge_x = 0,
  nudge_y = 0,
  default.colour = "black",
  default.color = default.colour,
  colour.target = "all",
  color.target = colour.target,
  default.alpha = NA,
  alpha.target = "all",
  segment.linewidth = 0.5,
  arrow = NULL,
  check_overlap = FALSE,
  size.unit = "mm",
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = FALSE
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by
\code{\link[ggplot2]{aes}}. With \code{inherit.aes = FALSE}
(the default) it is not combined with the default mapping at the top level of
the plot. You always need to supply a \code{mapping} unless you set
\code{inherit.aes = TRUE}.}

\item{data}{A data frame. If specified, overrides the default data frame
defined at the top level of the plot.}

\item{stat}{The statistical transformation to use on the data for this layer,
as a string.}

\item{position}{Position adjustment, either as a string, or the result of a
call to a position adjustment function.}

\item{...}{other arguments passed on to \code{\link[ggplot2]{layer}}. There
  are three types of arguments you can use here:

  \itemize{ \item Aesthetics: to set an aesthetic to a fixed value, like
  \code{colour = "red"} or \code{size = 3}. \item Other arguments to the
  layer, for example you override the default \code{stat} associated with the
  layer. \item Other arguments passed on to the stat. }}

\item{orientation}{character The orientation of the layer. The default
(\code{NA}) automatically determines the orientation from the aesthetic
mapping. In the rare event that this fails it can be given explicitly by
setting orientation to either "x" or "y".}

\item{parse}{If \code{TRUE}, the labels will be parsed into expressions and
displayed as described in \code{?plotmath}.}

\item{nudge_x, nudge_y}{Horizontal and vertical adjustments to nudge the
starting position of each text label. The units for \code{nudge_x} and
\code{nudge_y} are the same as for the data units on the x-axis and y-axis.}

\item{default.colour, default.color}{A colour definition to use for elements
not targeted by the colour aesthetic.}

\item{colour.target, color.target}{A vector of character strings; \code{"all"},
\code{"text"}, \code{"segment"}, \code{"box"}, \code{"box.line"}, and
\code{"box.fill"} or \code{"none"}.}

\item{default.alpha}{numeric in [0..1] A transparency value to use for
elements not targeted by the alpha aesthetic.}

\item{alpha.target}{A vector of character strings; \code{"all"},
\code{"text"}, \code{"segment"}, \code{"box"}, \code{"box.line"}, and
\code{"box.fill"} or \code{"none"}.}

\item{label.padding}{Amount of padding around label. Defaults to 0.25 lines.}

\item{label.r}{Radius of rounded corners. Defaults to 0.15 lines.}

\item{segment.linewidth}{numeric Width of the segments or arrows in mm.}

\item{arrow}{specification for arrow heads, as created by
\code{\link[grid]{arrow}}}

\item{size.unit}{How the `size` aesthetic is interpreted: as millimetres
(`"mm"`, default), points (`"pt"`), centimetres (`"cm"`), inches (`"in"`),
or picas (`"pc"`).}

\item{na.rm}{If \code{FALSE} (the default), removes missing values with a
warning.  If \code{TRUE} silently removes missing values.}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA} includes a legend if any aesthetics are mapped.
\code{FALSE}, the default, never includes it, and \code{TRUE} always includes it.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics, rather
than combining them.}

\item{check_overlap}{If \code{TRUE}, text that overlaps previous text in the
same layer will not be plotted. \code{check_overlap} takes place at draw
time and in the order of the data, thus its action depends of the size at
which the plot is drawn.}
}
\value{
A plot layer instance.
}
\description{
Add a plot layer with a text label and a segment connecting two
  values along the \code{x} aesthetic. These are usually two levels of a
  factor mapped to the \code{x} aesthetic when used to report significance or
  highlighting pairwise comparisons.
}
\details{
Geometries \code{geom_text_pairwise()} and
  \code{geom_label_pairwise()} have an interface similar to that of
  \code{\link[ggplot2]{geom_text}} and \code{\link[ggplot2]{geom_label}}, but
  add a segment connecting two values along \code{x}. In the most
  frequent use case they add a segment connecting pairs of levels from a
  grouping factor mapped to the \emph{x} or \emph{y} aesthetic. They can
  also be used to label ranges of values.

  The segment extends from \code{xmin} to \code{xmax}, and the text label is
  located at \code{x} with a default that positions the label at the centre
  of the bar. The ends of the bar can be terminated with arrow heads given
  by parameter \code{arrow}, with a default of a plain segment without
  arrow tips. The text label is located slightly above the segment by the
  default value of \code{vjust} in \code{geom_text_pairwise()} and on top
  of the segment in \code{geom_label_pairwise()}.

  Layer functions \code{geom_text_pairwise()} and
  \code{geom_label_pairwise()} use by default
  \code{\link[ggplot2]{position_nudge}}. Nudging affects both text label and
  bar, and its default of no displacement will very rarely need to be
  changed.

  Differently to \code{geom_text_repel()} and \code{geom_label_repel()},
  \code{geom_text_pairwise()} and \code{geom_label_pairwise()} do not make
  use of additional aesthetics for the segments or boxes, but instead allow
  the choice of which elements are targeted by the usual 'ggplot2' aesthetics
  and which are rendered using a default constant value. In the grammar of
  graphics using the same aesthetic with multiple meanings is not allowed,
  thus, the approach used in package 'ggpp' attempts to enforce this.
}
\section{Under development!}{
 This geometry is still under development
  and its user interface subject to change.
}

\section{Plot boundaries and clipping}{
 Note that when you change the scale
  limits for \emph{x} and/or \emph{y} of a plot, text labels stay the same
  size, as determined by the \code{size} aesthetic, given in millimetres. The
  actual size as seen in the plotted output is decided during the rendering
  of the plot to a graphics device. Limits are expanded only to include the
  anchor point of the labels because the "width" and "height" of a text
  element are 0 (as seen by ggplot2). Text labels do have height and width,
  but in grid units, not data units. Either function
  \code{\link[ggplot2]{expand_limits}} or the scale expansion can be used to
  ensure text labels remain within the plotting area.
}

\section{Alignment}{
 You can modify text alignment with the \code{vjust} and
  \code{hjust} aesthetics. These can either be a number between 0
  (right/bottom) and 1 (top/left) or a character (\code{"left"},
  \code{"middle"}, \code{"right"}, \code{"bottom"}, \code{"center"},
  \code{"top"}). Values outside the range 0..1 displace the text label so
  that the anchor point is outside the text label. In addition, you can use
  special alignments for justification including \code{"position"},
  \code{"inward"} and \code{"outward"}. Inward always aligns text towards the
  center of the plotting area, and outward aligns it away from the center of
  the plotting area. If tagged with \code{_mean} or \code{_median} (e.g.,
  \code{"outward_mean"}) the mean or median of the data in the panel along
  the corresponding axis is used as center. If the characters following the
  underscore represent a number (e.g., \code{"outward_10.5"}) the reference
  point will be this value in data units. Position justification is computed
  based on the direction of the displacement of the position of the label so
  that each individual text or label is justified outwards from its original
  position. The default justification is \code{"identity"}.
}

\section{Aesthetics}{
 Layer functions \code{geom_text_pairwise()} and
  \code{geom_label_pairwise()} require aesthetics \code{xmin}, \code{xmax},
  \code{x}, \code{y} and
  \code{label} and support aesthetics: \code{alpha}, \code{colour},
  \code{group}, \code{size} (of text), \code{family}, \code{fontface},
  \code{linewidth}, \code{linetype}, \code{hjust} and \code{vjust}. In addition,
  \code{geom_text_pairwise} supports \code{angle} and \code{geom_label_pairwise} supports
  \code{fill}. See
  \code{\link[ggplot2]{aes_colour_fill_alpha}},
  \code{\link[ggplot2]{aes_linetype_size_shape}},
  \code{\link[ggplot2]{aes_position}}, and
  \code{\link[ggplot2]{aes_group_order}}.

  In 'ggplot2' \code{linewidth} when applied to the border of the box drawn
  by \code{geom_label()} is given in points rather than in mm because of a
  historical error in the code. In other geometries such as
  \code{geom_segment()} \code{linewidth} is given in mm. As in
  \code{geom_label_pairwise()} it is important to remain consistent among
  different \code{linewidth} specifications, mm are used both for the box
  border and linking segment. To imitate the behaviour of \code{geom_label()}
  a correction factor of 0.75 (more exactly 1 pt = 0.7528 mm) can be used for
  the line width of the border of the box.
}

\examples{

my.cars <- mtcars
my.cars$name <- rownames(my.cars)
p1 <- ggplot(my.cars, aes(x = factor(cyl), y = mpg)) +
       geom_boxplot(width = 0.33)

# With a factor mapped to x, highlight pairs

my.pairs <-
  data.frame(A = 1:2, B = 2:3, bar.height = c(12, 30),
             p.value = c(0.01, 0.05678))
p1 +
  geom_text_pairwise(data = my.pairs,
                     aes(xmin = A, xmax = B,
                         y = bar.height,
                         label = p.value),
                     parse = TRUE)

p1 +
  geom_text_pairwise(data = my.pairs,
                     aes(xmin = A, xmax = B,
                         y = bar.height,
                         label = sprintf("italic(P)~`=`~\%.2f", p.value)),
                     arrow = grid::arrow(angle = 90,
                                         length = unit(1, "mm"),
                                         ends = "both"),
                     parse = TRUE)

p1 +
  geom_text_pairwise(data = my.pairs,
                     aes(xmin = A, xmax = B,
                         y = bar.height,
                         label = sprintf("italic(P)~`=`~\%.2f", p.value)),
                     colour = "red",
                     arrow = grid::arrow(angle = 90,
                                         length = unit(1, "mm"),
                                         ends = "both"),
                     parse = TRUE)

p1 +
  geom_label_pairwise(data = my.pairs,
                      aes(xmin = A, xmax = B,
                          y = bar.height,
                          label = sprintf("italic(P)~`=`~\%.2f", p.value)),
                      colour = "red", size = 2.75,
                      arrow = grid::arrow(angle = 30,
                                          length = unit(1.5, "mm"),
                                          ends = "both"),
                      parse = TRUE)

p1 +
  geom_text_pairwise(data = my.pairs,
                     aes(xmin = A, xmax = B,
                         y = bar.height,
                         label = sprintf("italic(P)~`=`~\%.2f", p.value)),
                     colour = "red", colour.target = "segment",
                     arrow = grid::arrow(angle = 90,
                                         length = unit(1, "mm"),
                                         ends = "both"),
                     parse = TRUE)

p1 +
  geom_text_pairwise(data = my.pairs,
                     aes(xmin = A, xmax = B,
                         y = bar.height,
                         label = sprintf("italic(P)~`=`~\%.2f", p.value)),
                     colour = "red", colour.target = "text",
                     arrow = grid::arrow(angle = 90,
                                         length = unit(1, "mm"),
                                         ends = "both"),
                     parse = TRUE)

# with a numeric vector mapped to x, indicate range

p2 <-
  ggplot(my.cars, aes(disp, mpg)) +
    geom_point()

my.ranges <-
  data.frame(A = c(50, 400),
             B = c(200, 500),
             bar.height = 5,
             text = c("small", "large"))

p2 +
  geom_text_pairwise(data = my.ranges,
                     aes(xmin = A, xmax = B,
                     y = bar.height, label = text))

p2 +
  geom_text_pairwise(data = my.ranges,
                     aes(xmin = A, xmax = B,
                         y = bar.height, label = text),
                     angle = 90, hjust = -0.1)

p2 +
  geom_label_pairwise(data = my.ranges,
                      aes(xmin = A, xmax = B,
                          y = bar.height, label = text),
                     angle = 90, hjust = -0.1)

p2 +
  geom_label_pairwise(data = my.ranges,
                      aes(xmin = A, xmax = B,
                          y = bar.height, label = text))

p2 +
  geom_text_pairwise(data = my.ranges,
                     aes(xmin = A, xmax = B,
                         y = bar.height, label = text),
                     arrow = grid::arrow(ends = "both", length = unit(2, "mm")))

# implicit orientation = "y"
p3 <- ggplot(my.cars, aes(x = mpg, y = factor(cyl))) +
       geom_boxplot(width = 0.33)

p3 +
  geom_text_pairwise(data = my.pairs,
                     aes(ymin = A, ymax = B,
                         x = bar.height,
                         label = p.value))

p3 +
  geom_label_pairwise(data = my.pairs,
                      aes(ymin = A, ymax = B,
                          x = bar.height,
                          label = p.value))

}
\seealso{
\code{\link{geom_text_s}}, \code{\link{geom_label_s}},
  \code{\link[ggplot2]{geom_text}}, \code{\link[ggplot2]{geom_label}} and
  other documentation of package 'ggplot2'.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat-dens1d-labels.r
\name{stat_dens1d_labels}
\alias{stat_dens1d_labels}
\title{Replace labels in data based on 1D density}
\usage{
stat_dens1d_labels(
  mapping = NULL,
  data = NULL,
  geom = "text",
  position = "identity",
  ...,
  keep.fraction = 0.1,
  keep.number = Inf,
  keep.sparse = TRUE,
  keep.these = FALSE,
  exclude.these = FALSE,
  these.target = "label",
  pool.along = c("x", "none"),
  xintercept = 0,
  invert.selection = FALSE,
  bw = "SJ",
  kernel = "gaussian",
  adjust = 1,
  n = 512,
  orientation = c("x", "y"),
  label.fill = "",
  return.density = FALSE,
  na.rm = TRUE,
  show.legend = FALSE,
  inherit.aes = TRUE
)
}
\arguments{
\item{mapping}{The aesthetic mapping, usually constructed with
\code{\link[ggplot2]{aes}} or \code{\link[ggplot2]{aes_}}. Only needs to be
set at the layer level if you are overriding the plot defaults.}

\item{data}{A layer specific dataset - only needed if you want to override
the plot defaults.}

\item{geom}{The geometric object to use display the data.}

\item{position}{The position adjustment to use for overlapping points on this
layer}

\item{...}{other arguments passed on to \code{\link[ggplot2]{layer}}. This
can include aesthetics whose values you want to set, not map. See
\code{\link[ggplot2]{layer}} for more details.}

\item{keep.fraction}{numeric vector of length 1 or 2 [0..1]. The fraction of
the observations (or rows) in \code{data} to be retained.}

\item{keep.number}{integer vector of length 1 or 2. Set the maximum number of
observations to retain, effective only if obeying \code{keep.fraction}
would result in a larger number.}

\item{keep.sparse}{logical If \code{TRUE}, the default, observations from the
more sparse regions are retained, if \code{FALSE} those from the densest
regions.}

\item{keep.these, exclude.these}{character vector, integer vector, logical
vector or function that takes one or more variables in data selected by
\code{these.target}. Negative integers behave as in R's extraction methods.
The rows from \code{data} indicated by \code{keep.these} and
\code{exclude.these} are kept or excluded irrespective of the local
density.}

\item{these.target}{character, numeric or logical selecting one or more
column(s) of \code{data}. If \code{TRUE} the whole \code{data} object is
passed.}

\item{pool.along}{character, one of \code{"none"} or \code{"x"},
indicating if selection should be done pooling the observations along the
\emph{x} aesthetic, or separately on either side of \code{xintercept}.}

\item{xintercept}{numeric The split point for the data filtering.}

\item{invert.selection}{logical If \code{TRUE}, the complement of the
selected rows are returned.}

\item{bw}{numeric or character The smoothing bandwidth to be used. If
numeric, the standard deviation of the smoothing kernel. If character, a
rule to choose the bandwidth, as listed in \code{\link[stats]{bw.nrd}}.}

\item{kernel}{character See \code{\link{density}} for details.}

\item{adjust}{numeric A multiplicative bandwidth adjustment. This makes it
possible to adjust the bandwidth while still using the a bandwidth
estimator through an argument passed to \code{bw}. The larger the value
passed to \code{adjust} the stronger the smoothing, hence decreasing
sensitivity to local changes in density.}

\item{n}{numeric Number of equally spaced points at which the density is to
be estimated for applying the cut point. See \code{\link{density}} for
details.}

\item{orientation}{character The aesthetic along which density is computed.
Given explicitly by setting orientation to either "x" or "y".}

\item{label.fill}{character vector of length 1 or a function.}

\item{return.density}{logical vector of lenght 1. If \code{TRUE} add columns
\code{"density"} and \code{"keep.obs"} to the returned data frame.}

\item{na.rm}{a logical value indicating whether NA values should be stripped
before the computation proceeds.}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
never includes, and \code{TRUE} always includes.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics, rather
than combining with them. This is most useful for helper functions that
define both data and aesthetics and shouldn't inherit behaviour from the
default plot specification, e.g. \code{\link[ggplot2]{borders}}.}
}
\value{
A plot layer instance. Using as output \code{data} the input
  \code{data} after value substitution based on a 1D the filtering criterion.
}
\description{
\code{stat_dens1d_labels()} Sets values mapped to the
  \code{label} aesthetic to \code{""} or a user provided character string
  based on the local density in regions of a plot panel. Its main use is
  together with repulsive geoms from package \code{\link[ggrepel]{ggrepel}}
  to restrict labeling to the low density tails of a distribution. By default
  the data are handled all together, but it is also possible to control
  labeling separately in each tail.

  If there is no mapping to \code{label} in \code{data}, the mapping is set
  to \code{rownames(data)}, with a message.
}
\details{
\code{stat_dens1d_labels()} is designed to work together with
  geometries from package 'ggrepel'. To avoid text labels being plotted over
  unlabelled points the corresponding rows in data need to be retained but
  labels replaced with the empty character string, \code{""}. Function
  \code{\link{stat_dens1d_filter}} cannot be used with the repulsive geoms
  from 'ggrepel' because it drops the observations.

  \code{stat_dens1d_labels()} can be useful also in other situations, as the
  substitution character string can be set by the user by passing an argument
  to \code{label.fill}. If this argument is \code{NULL} the unselected rows
  are filtered out.

  The local density of observations along \emph{x} or \emph{y} is computed
  with function \code{\link[stats]{density}} and used to select observations,
  passing to the geom all the rows in its \code{data} input but with with the
  text of labels replaced in those "not kept". The default is to select
  observations in sparse regions of the plot, but the selection can be
  inverted so that only observations in the densest regions are returned.
  Specific observations can be protected from having the label replaced by
  passing a suitable argument to \code{keep.these}. Logical and integer
  vectors function as indexes to rows in \code{data}, while a character
  vector is compared to values in the variable mapped to the \code{label}
  aesthetic. A function passed as argument to keep.these will receive as
  argument the values in the variable mapped to \code{label} and should
  return a character, logical or numeric vector as described above.

  How many labels are retained intact in addition to those in
  \code{keep.these} is controlled with arguments passed to \code{keep.number}
  and \code{keep.fraction}. \code{keep.number} sets the maximum number of
  observations selected, whenever \code{keep.fraction} results in fewer
  observations selected, it is obeyed. If \code{xintercept} is a finite value
  within the \emph{x} range of the data and \code{pool.along} is passed
  \code{"none"} the data are split into two groups and \code{keep.number} and
  \code{keep.fraction} are applied separately to each tail with density still
  computed jointly from all observations. If the length of \code{keep.number}
  and \code{keep.fraction} is one, half this value is used each tail, if
  their length is two, the first value is use for the left tail and the
  second value for the right tail (or if using \code{orientation = "y"} the
  lower and upper tails, respectively).

  Computation of density and of the default bandwidth require at least
  two observations with different values. If data do not fulfill this
  condition, they are kept only if \code{keep.fraction = 1}. This is correct
  behavior for a single observation, but can be surprising in the case of
  multiple observations.

  Parameters \code{keep.these} and \code{exclude.these} make it possible to
  force inclusion or exclusion of labels after the density is computed.
  In case of conflict, \code{exclude.these} overrides \code{keep.these}.
}
\note{
Which points are kept and which not depends on how dense and flexible
  is the density curve estimate. This depends on the values passed as
  arguments to parameters \code{n}, \code{bw} and \code{kernel}. It is
  also important to be aware that both \code{geom_text()} and
  \code{geom_text_repel()} can avoid overplotting by discarding labels at
  the plot rendering stage, i.e., what is plotted may differ from what is
  returned by this statistic.
}
\examples{

random_string <-
  function(len = 6) {
    paste(sample(letters, len, replace = TRUE), collapse = "")
  }

# Make random data.
set.seed(1005)
d <- tibble::tibble(
  x = rnorm(100),
  y = rnorm(100),
  group = rep(c("A", "B"), c(50, 50)),
  lab = replicate(100, { random_string() })
)

# using defaults
ggplot(data = d, aes(x, y, label = lab)) +
  geom_point() +
  stat_dens1d_labels()

ggrepel.installed <- requireNamespace("ggrepel", quietly = TRUE)
if (ggrepel.installed) {
  library(ggrepel)

# using defaults
  ggplot(data = d, aes(x, y, label = lab)) +
    geom_point() +
    stat_dens1d_labels(geom = "text_repel")

# if no mapping to label is found, it is set row names
  ggplot(data = d, aes(x, y)) +
    geom_point() +
    stat_dens1d_labels(geom = "text_repel")

  ggplot(data = d, aes(x, y)) +
    geom_point() +
    stat_dens1d_labels(geom = "text_repel", pool.along = "none")

  ggplot(data = d, aes(x, y)) +
    geom_point() +
    stat_dens1d_labels(geom = "text_repel",
                       keep.number = c(0, 10), pool.along = "none")

  ggplot(data = d, aes(x, y)) +
    geom_point() +
    stat_dens1d_labels(geom = "text_repel",
                       keep.fraction = c(0, 0.2), pool.along = "none")

# using defaults, along y-axis
  ggplot(data = d, aes(x, y, label = lab)) +
    geom_point() +
    stat_dens1d_labels(orientation = "y", geom = "text_repel")

# example labelling with coordiantes
  ggplot(data = d, aes(x, y, label = sprintf("x = \%.2f\ny = \%.2f", x, y))) +
    geom_point() +
    stat_dens1d_filter(colour = "red") +
    stat_dens1d_labels(geom = "text_repel", colour = "red", size = 3)

  ggplot(data = d, aes(x, y, label = lab, colour = group)) +
    geom_point() +
    stat_dens1d_labels(geom = "text_repel")

  ggplot(data = d, aes(x, y, label = lab, colour = group)) +
    geom_point() +
    stat_dens1d_labels(geom = "text_repel", label.fill = NA)

# we keep labels starting with "a" across the whole plot, but all in sparse
# regions. To achieve this we pass as argument to label.fill a fucntion
# instead of a character string.
  label.fun <- function(x) {ifelse(grepl("^a", x), x, "")}
  ggplot(data = d, aes(x, y, label = lab, colour = group)) +
    geom_point() +
    stat_dens1d_labels(geom = "text_repel", label.fill = label.fun)
}

# Using geom_debug_group() we can see that all 100 rows in \code{d} are
# returned.
gginnards.installed <- requireNamespace("gginnards", quietly = TRUE)
if (gginnards.installed) {
  library(gginnards)

  ggplot(data = d, aes(x, y, label = lab)) +
    geom_point() +
    stat_dens1d_labels(geom = "debug_group")
}
if (gginnards.installed) {
  ggplot(data = d, aes(x, y, label = lab)) +
    geom_point() +
    stat_dens1d_labels(geom = "debug_group", return.density = TRUE)
}

}
\seealso{
\code{\link[stats]{density}} used internally.

Other statistics returning a subset of data:
\code{\link[=stat_dens1d_filter]{stat_dens1d_filter()}},
\code{\link[=stat_dens2d_filter]{stat_dens2d_filter()}},
\code{\link[=stat_dens2d_labels]{stat_dens2d_labels()}}
}
\concept{statistics returning a subset of data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/position-dodge-nudge.R,
%   R/position-dodge2-nudge.R
\name{position_dodgenudge}
\alias{position_dodgenudge}
\alias{position_dodge_keep}
\alias{position_dodge2nudge}
\alias{position_dodge2_keep}
\title{Combined positions dodge and nudge}
\source{
\url{https://github.com/slowkow/ggrepel/issues/161}.
}
\usage{
position_dodgenudge(
  width = 1,
  preserve = c("total", "single"),
  reverse = FALSE,
  x = 0,
  y = 0,
  direction = c("none", "split", "split.x", "split.y", "center"),
  kept.origin = c("dodged", "original", "none")
)

position_dodge_keep(
  width = 1,
  preserve = c("total", "single"),
  reverse = FALSE,
  kept.origin = "original"
)

position_dodge2nudge(
  width = 1,
  preserve = c("total", "single"),
  padding = 0.1,
  reverse = FALSE,
  x = 0,
  y = 0,
  direction = c("none", "split", "split.x", "split.y", "center"),
  kept.origin = c("dodged", "original", "none")
)

position_dodge2_keep(
  width = 1,
  preserve = c("total", "single"),
  kept.origin = "original"
)
}
\arguments{
\item{width}{Dodging width, when different to the width of the individual
elements. This is useful when you want to align narrow geoms with wider
geoms. See the examples.}

\item{preserve}{Should dodging preserve the total width of all elements at a
position, or the width of a single element?.}

\item{reverse}{If TRUE, will reverse the default stacking order. This is
useful if you're rotating both the plot and legend.}

\item{x, y}{Amount of vertical and horizontal distance to move. A numeric
vector of length 1, or of the same length as rows there are in `data`,
with nudge values in data rows order.}

\item{direction}{One of \code{"none"}, \code{"split"}, \code{"split.x"} or
\code{"split.y"}. A value of \code{"none"} replicates the behavior of
\code{\link[ggplot2]{position_nudge}}. At the moment \code{"split"} changes
the sign of the nudge at zero, which is suitable for column plots with
negative slices.}

\item{kept.origin}{One of \code{"original"}, \code{"dodged"} or
\code{"none"}.}

\item{padding}{Padding between elements at the same position. Elements are
shrunk by this proportion to allow space between them. Defaults to 0.1.}
}
\value{
A \code{"Position"} object.
}
\description{
\code{position_dodgenudge()} combines into one function the action of
\code{\link[ggplot2]{position_dodge}} and
\code{\link[ggplot2]{position_nudge}} and \code{position_dodge2nudge()}
combines into one function the action of
\code{\link[ggplot2]{position_dodge2}} and
\code{\link[ggplot2]{position_nudge}}. They are useful when labelling plots
such as grouped bars, columns, etc. and when adding dodged text labels linked
to observations plotted without dodge. It can replace other position
functions as they are backwards compatible. Like all other position functions
in 'ggpp' and 'ggrepel' they preserve the initial position to allow drawing
of segments or arrow linking the original position to the displaced one.
}
\details{
The applied dodge is identical to that by
  \code{\link[ggplot2]{position_dodge}} or
  \code{\link[ggplot2]{position_dodge2}} while nudging is similar to that by
  \code{\link[ggplot2]{position_nudge}}.

There are two possible uses for these functions. First they can be used to
label dodged bars or boxplots. In this case, it is mandatory to use the same
argument to \code{width} when passing \code{position_dodge()} to
\code{geom_col()} and \code{position_dodgenudge()} to \code{geom_text()} or
\code{geom_label()} or their repulsive equivalents. Otherwise the arrows or
segments will fail to connect to the labels. In other words dodging is
computed twice. Dodge is identical to that obtained with the same arguments
in \code{\link[ggplot2]{position_dodge}} as \code{position_dodgenudge()} last
function simply calls the same code from package 'ggplot2' ahead of applying
nudging.

The second use is to dodge labels to be connected to elements that have not
been dodged.

When applying dodging, the return of original positions instead of the dodged
ones is achieved by passing \code{origin = "original"} instead of the default
of \code{origin = "dodged"}.
}
\examples{
df <- data.frame(x1 = c(1, 2, 1, 3, -1),
                 x2 = c("a", "a", "b", "b", "b"),
                 grp = c("some long name", "other name", "some name",
                         "another name", "some long name"))

# Add labels to a horizontal column plot (stacked by default)
ggplot(data = df, aes(x1, x2, group = grp)) +
  geom_col(aes(fill = grp), width = 0.8,
           position = position_dodge()) +
  geom_vline(xintercept = 0) +
  geom_text(
    aes(label = grp),
    position = position_dodgenudge(x = 0.09, direction = "split", width = 0.8),
    angle = 90, size = 3) +
  theme(legend.position = "none")

ggplot(data = df, aes(x2, x1, group = grp)) +
  geom_col(aes(fill = grp), width = 0.75,
           position = position_dodge(width = 0.75)) +
  geom_vline(xintercept = 0) +
  geom_text(aes(label = grp),
            position = position_dodgenudge(y = 0.1,
                                           direction = "split",
                                           width = 0.75),
            size = 3) +
  theme(legend.position = "none")

}
\seealso{
\code{\link[ggplot2]{position_nudge}},
\code{\link[ggrepel]{position_nudge_repel}}.

Other position adjustments:
\code{\link[=position_dodgenudge_to]{position_dodgenudge_to()}},
\code{\link[=position_jitternudge]{position_jitternudge()}},
\code{\link[=position_nudge_center]{position_nudge_center()}},
\code{\link[=position_nudge_keep]{position_nudge_keep()}},
\code{\link[=position_nudge_line]{position_nudge_line()}},
\code{\link[=position_nudge_to]{position_nudge_to()}},
\code{\link[=position_stacknudge]{position_stacknudge()}},
\code{\link[=position_stacknudge_to]{position_stacknudge_to()}}
}
\author{
Michał Krassowski, edited by Pedro J. Aphalo.
}
\concept{position adjustments}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/position-nudge-center.R
\name{position_nudge_keep}
\alias{position_nudge_keep}
\title{Nudge points a fixed distance}
\usage{
position_nudge_keep(
  x = 0,
  y = 0,
  obey_grouping = NULL,
  kept.origin = c("original", "none")
)
}
\arguments{
\item{x, y}{Amount of vertical and horizontal distance to move. A numeric
vector of length 1, or of the same length as rows there are in \code{data},
with nudge values in data rows order.}

\item{obey_grouping}{A logical flag indicating whether to obey or not
groupings of the observations. By default, grouping is obeyed when both of
the variables mapped to \emph{x} and \emph{y} are continuous numeric and
ignored otherwise.}

\item{kept.origin}{One of \code{"original"} or \code{"none"}.}
}
\value{
A \code{"Position"} object.
}
\description{
The function \code{position_nudge_keep()} has an additional parameters
compared to \code{\link[ggplot2]{position_nudge}}, \code{obey_grouping} and
by default the same behaviour when the values passed as arguments to \code{x}
and \code{y} have length one.
}
\details{
When \code{x} or \code{y} have length > 1, they are treated
  specially. If the lengths is the same as there are rows in data, the nudges
  are applied in the order of the rows in data. When they are shorter, they
  are recycled and applied to the data values after ordering. This makes it
  possible to have alternating mudging right and left or up and down. If
  \code{obey_grouping = TRUE} is passed in the call, the alternation will
  take place within groups.

  As other position functions from package 'ggpp', \code{position_nudge_keep()}
  by default renames and keeps the original positions of the observations in
  \code{data} making it possible to draw connecting segments or conencting
  arrows.
}
\note{
Irrespective of the action, the ordering of rows in \code{data} is
  preserved.
}
\examples{
df <- data.frame(
  x = c(1,3,2,5,4,2.5),
  y = c("abc","cd","d","c","bcd","a")
)

# Plain nudging, same as with ggplot2::position_nudge()

ggplot(df, aes(x, y, label = y)) +
  geom_point() +
  geom_text_s(hjust = "left", vjust = "bottom",
              position = position_nudge_keep(x = 0.2, y = 0.2))

# alternating nudging
ggplot(df, aes(x, y, label = y)) +
  geom_point() +
  geom_text_s(position = position_nudge_keep(x = c(0.2, -0.2)))

# direct nudging
ggplot(df, aes(x, y, label = y)) +
  geom_point() +
  geom_text_s(position = position_nudge_keep(x = rep_len(c(0.2, -0.2), 6)))

}
\seealso{
Other position adjustments:
\code{\link[=position_dodgenudge]{position_dodgenudge()}},
\code{\link[=position_dodgenudge_to]{position_dodgenudge_to()}},
\code{\link[=position_jitternudge]{position_jitternudge()}},
\code{\link[=position_nudge_center]{position_nudge_center()}},
\code{\link[=position_nudge_line]{position_nudge_line()}},
\code{\link[=position_nudge_to]{position_nudge_to()}},
\code{\link[=position_stacknudge]{position_stacknudge()}},
\code{\link[=position_stacknudge_to]{position_stacknudge_to()}}
}
\concept{position adjustments}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/try-data-frame.R
\name{ggplot}
\alias{ggplot}
\alias{ggplot.ts}
\alias{ggplot.xts}
\title{Create a new ggplot plot from time series data}
\usage{
\method{ggplot}{ts}(
  data,
  mapping = NULL,
  ...,
  time.resolution = "day",
  as.numeric = TRUE,
  environment = parent.frame()
)

\method{ggplot}{xts}(
  data,
  mapping = NULL,
  ...,
  time.resolution = "day",
  as.numeric = TRUE,
  environment = parent.frame()
)
}
\arguments{
\item{data}{Default spectrum dataset to use for plot. If not a spectrum, the
methods used will be those defined in package \code{ggplot2}. See \code{\link[ggplot2]{ggplot}}.
If not specified,
must be suppled in each layer added to the plot.}

\item{mapping}{Default list of aesthetic mappings to use for plot.
If not specified, in the case of spectral objects, a default mapping will
be used.}

\item{...}{Other arguments passed on to methods. Not currently used.}

\item{time.resolution}{character The time unit to which the returned time
values will be rounded.}

\item{as.numeric}{logical If TRUE convert time to numeric, expressed as
fractional calendar years.}

\item{environment}{If an variable defined in the aesthetic mapping is not
found in the data, ggplot will look for it in this environment. It defaults
to using the environment in which \code{ggplot()} is called.}
}
\value{
A \code{"ggplot"} object.
}
\description{
\code{ggplot()} initializes a ggplot object. It can be used to
declare the input spectral object for a graphic and to optionally specify the
set of plot aesthetics intended to be common throughout all
subsequent layers unless specifically overridden.
}
\details{
\code{ggplot()} is typically used to construct a plot
incrementally, using the + operator to add layers to the
existing ggplot object. This is advantageous in that the
code is explicit about which layers are added and the order
in which they are added. For complex graphics with multiple
layers, initialization with \code{ggplot} is recommended.

There are three common ways to invoke \code{ggplot}:
\itemize{
   \item \code{ggplot(ts, aes(x, y, <other aesthetics>))}
   \item \code{ggplot(ts)}
  }
The first method is recommended if all layers use the same
data and the same set of aesthetics, although this method
can also be used to add a layer using data from another
data frame. See the first example below. The second
method specifies the default spectrum object to use for the plot, and the
units to be used for y in the plot,
but no aesthetics are defined up front. This is useful when
one data frame is used predominantly as layers are added,
but the aesthetics may vary from one layer to another. The
third method specifies the default spectrum object to use for the plot,
but no aesthetics are defined up front. This is useful when
one spectrum is used predominantly as layers are added,
but the aesthetics may vary from one layer to another.
}
\note{
Current implementation does not merge default mapping with user
supplied mapping. If user supplies a mapping, it is used as is.
To add to the default mapping, aes() can be used by itself to compose
the ggplot.
}
\examples{
ggplot(lynx) + geom_line()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat-dens2d-filter.r
\name{stat_dens2d_filter}
\alias{stat_dens2d_filter}
\alias{stat_dens2d_filter_g}
\title{Filter observations by local 2D density}
\usage{
stat_dens2d_filter(
  mapping = NULL,
  data = NULL,
  geom = "point",
  position = "identity",
  ...,
  keep.fraction = 0.1,
  keep.number = Inf,
  keep.sparse = TRUE,
  keep.these = FALSE,
  exclude.these = FALSE,
  these.target = "label",
  pool.along = c("xy", "x", "y", "none"),
  xintercept = 0,
  yintercept = 0,
  invert.selection = FALSE,
  na.rm = TRUE,
  show.legend = FALSE,
  inherit.aes = TRUE,
  h = NULL,
  n = NULL,
  return.density = FALSE
)

stat_dens2d_filter_g(
  mapping = NULL,
  data = NULL,
  geom = "point",
  position = "identity",
  ...,
  keep.fraction = 0.1,
  keep.number = Inf,
  keep.sparse = TRUE,
  keep.these = FALSE,
  exclude.these = FALSE,
  these.target = "label",
  pool.along = c("xy", "x", "y", "none"),
  xintercept = 0,
  yintercept = 0,
  invert.selection = FALSE,
  na.rm = TRUE,
  show.legend = FALSE,
  inherit.aes = TRUE,
  h = NULL,
  n = NULL,
  return.density = FALSE
)
}
\arguments{
\item{mapping}{The aesthetic mapping, usually constructed with
\code{\link[ggplot2]{aes}} or \code{\link[ggplot2]{aes_}}. Only needs
to be set at the layer level if you are overriding the plot defaults.}

\item{data}{A layer specific dataset - only needed if you want to override
the plot defaults.}

\item{geom}{The geometric object to use display the data.}

\item{position}{The position adjustment to use for overlapping points on this
layer}

\item{...}{other arguments passed on to \code{\link[ggplot2]{layer}}. This
can include aesthetics whose values you want to set, not map. See
\code{\link[ggplot2]{layer}} for more details.}

\item{keep.fraction}{numeric [0..1]. The fraction of the observations (or
rows) in \code{data} to be retained.}

\item{keep.number}{integer Set the maximum number of observations to retain,
effective only if obeying \code{keep.fraction} would result in a larger
number.}

\item{keep.sparse}{logical If \code{TRUE}, the default, observations from the
more sparse regions are retained, if \code{FALSE} those from the densest
regions.}

\item{keep.these, exclude.these}{character vector, integer vector, logical
vector or function that takes one or more variables in data selected by
\code{these.target}. Negative integers behave as in R's extraction methods.
The rows from \code{data} indicated by \code{keep.these} and
\code{exclude.these} are kept or excluded irrespective of the local
density.}

\item{these.target}{character, numeric or logical selecting one or more
column(s) of \code{data}. If \code{TRUE} the whole \code{data} object is
passed.}

\item{pool.along}{character, one of \code{"none"}, \code{"x"}, \code{"y"}, or
\code{"xy"} indicating if selection should be done pooling the observations
along the \emph{x}, \code{y}, both axes or none based on quadrants given by
\code{xintercept} and \code{yintercept}.}

\item{xintercept, yintercept}{numeric The center point of the quadrants.}

\item{invert.selection}{logical If \code{TRUE}, the complement of the
selected rows are returned.}

\item{na.rm}{a logical value indicating whether NA values should be stripped
before the computation proceeds.}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
never includes, and \code{TRUE} always includes.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics, rather
than combining with them. This is most useful for helper functions that
define both data and aesthetics and shouldn't inherit behaviour from the
default plot specification, e.g. \code{\link[ggplot2]{borders}}.}

\item{h}{vector of bandwidths for x and y directions. Defaults to normal
reference bandwidth (see bandwidth.nrd). A scalar value will be taken to
apply to both directions.}

\item{n}{Number of grid points in each direction. Can be scalar or a
length-2 integer vector}

\item{return.density}{logical vector of lenght 1. If \code{TRUE} add columns
\code{"density"} and \code{"keep.obs"} to the returned data frame.}
}
\value{
A plot layer instance. Using as output \code{data} a subset of the
  rows in input \code{data} retained based on a 2D-density-based filtering
  criterion.
}
\description{
\code{stat_dens2d_filter} Filters-out/filters-in observations in
  regions of a plot panel with high density of observations, based on the
  values mapped to both \code{x} and \code{y} aesthetics.
  \code{stat_dens2d_filter_g} does the filtering by group instead of by
  panel. This second stat is useful for highlighting observations, while the
  first one tends to be most useful when the aim is to prevent clashes among
  text labels. If there is no mapping to \code{label} in \code{data}, the
  mapping is silently set to \code{rownames(data)}.
}
\details{
The local density of observations in 2D (\emph{x} and \emph{y}) is
  computed with function \code{\link[MASS]{kde2d}} and used to select
  observations, passing to the geom a subset of the rows in its \code{data}
  input. The default is to select observations in sparse regions of the plot,
  but the selection can be inverted so that only observations in the densest
  regions are returned. Specific observations can be protected from being
  deselected and "kept" by passing a suitable argument to \code{keep.these}.
  Logical and integer vectors work as indexes to rows in \code{data}, while a
  character vector values are compared to the character values mapped to the
  \code{label} aesthetic. A function passed as argument to keep.these will
  receive as argument the values in the variable mapped to \code{label} and
  should return a character, logical or numeric vector as described above. If
  no variable has been mapped to \code{label}, row names are used in its
  place.

  How many rows are retained in addition to those in \code{keep.these} is
  controlled with arguments passed to \code{keep.number} and
  \code{keep.fraction}. \code{keep.number} sets the maximum number of
  observations selected, whenever \code{keep.fraction} results in fewer
  observations selected, it is obeyed.

  Computation of density and of the default bandwidth require at least
  two observations with different values. If data do not fulfill this
  condition, they are kept only if \code{keep.fraction = 1}. This is correct
  behavior for a single observation, but can be surprising in the case of
  multiple observations.

  Parameters \code{keep.these} and \code{exclude.these} make it possible to
  force inclusion or exclusion of observations after the density is computed.
  In case of conflict, \code{exclude.these} overrides \code{keep.these}.
}
\note{
Which points are kept and which not depends on how dense a grid is used
  and how flexible the density surface estimate is. This depends on the
  values passed as arguments to parameters \code{n}, \code{bw} and
  \code{kernel}. It is also important to be aware that both
  \code{geom_text()} and \code{geom_text_repel()} can avoid overplotting by
  discarding labels at the plot rendering stage, i.e., what is plotted may
  differ from what is returned by this statistic.
}
\examples{

random_string <-
  function(len = 6) {
    paste(sample(letters, len, replace = TRUE), collapse = "")
  }

# Make random data.
set.seed(1001)
d <- tibble::tibble(
  x = rnorm(100),
  y = rnorm(100),
  group = rep(c("A", "B"), c(50, 50)),
  lab = replicate(100, { random_string() })
)

# filter (and here highlight) 1/10 observations in sparsest regions
ggplot(data = d, aes(x, y)) +
  geom_point() +
  stat_dens2d_filter(colour = "red")

# filter observations not in the sparsest regions
ggplot(data = d, aes(x, y)) +
  geom_point() +
  stat_dens2d_filter(colour = "blue", invert.selection = TRUE)

# filter observations in dense regions of the plot
ggplot(data = d, aes(x, y)) +
  geom_point() +
  stat_dens2d_filter(colour = "blue", keep.sparse = FALSE)

# filter 1/2 the observations
ggplot(data = d, aes(x, y)) +
  geom_point() +
  stat_dens2d_filter(colour = "red", keep.fraction = 0.5)

# filter 1/2 the observations but cap their number to maximum 12 observations
ggplot(data = d, aes(x, y)) +
  geom_point() +
  stat_dens2d_filter(colour = "red",
                     keep.fraction = 0.5,
                     keep.number = 12)

# density filtering done jointly across groups
ggplot(data = d, aes(x, y, colour = group)) +
  geom_point() +
  stat_dens2d_filter(shape = 1, size = 3, keep.fraction = 1/4)

# density filtering done independently for each group
ggplot(data = d, aes(x, y, colour = group)) +
  geom_point() +
  stat_dens2d_filter_g(shape = 1, size = 3, keep.fraction = 1/4)

# density filtering done jointly across groups by overriding grouping
ggplot(data = d, aes(x, y, colour = group)) +
  geom_point() +
  stat_dens2d_filter_g(colour = "black",
                       shape = 1, size = 3, keep.fraction = 1/4)

# label observations
ggplot(data = d, aes(x, y, label = lab, colour = group)) +
  geom_point() +
  stat_dens2d_filter(geom = "text")

ggplot(data = d, aes(x, y, label = lab, colour = group)) +
  geom_point() +
  stat_dens2d_filter(geom = "text",
                     keep.these = function(x) {grepl("^u", x)})

ggplot(data = d, aes(x, y, label = lab, colour = group)) +
  geom_point() +
  stat_dens2d_filter(geom = "text",
                     keep.these = function(x) {grepl("^u", x)})

ggplot(data = d, aes(x, y, label = lab, colour = group)) +
  geom_point() +
  stat_dens2d_filter(geom = "text",
                     keep.these = 1:30)

# looking under the hood with gginnards::geom_debug_group()
gginnards.installed <- requireNamespace("gginnards", quietly = TRUE)
if (gginnards.installed) {
  library(gginnards)

  ggplot(data = d, aes(x, y, label = lab, colour = group)) +
    stat_dens2d_filter(geom = "debug_group")
}
if (gginnards.installed) {
  ggplot(data = d, aes(x, y, label = lab, colour = group)) +
    geom_point() +
    stat_dens2d_filter(geom = "debug_group", return.density = TRUE)
}

}
\seealso{
\code{\link{stat_dens2d_labels}} and \code{\link[MASS]{kde2d}} used
  internally. Parameters \code{n}, \code{h} in these statistics correspond to
  the parameters with the same name in this imported function. Limits are set
  to the limits of the plot scales.

Other statistics returning a subset of data:
\code{\link[=stat_dens1d_filter]{stat_dens1d_filter()}},
\code{\link[=stat_dens1d_labels]{stat_dens1d_labels()}},
\code{\link[=stat_dens2d_labels]{stat_dens2d_labels()}}
}
\concept{statistics returning a subset of data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat-quadrant-counts.R
\name{stat_quadrant_counts}
\alias{stat_quadrant_counts}
\title{Number of observations in quadrants}
\usage{
stat_quadrant_counts(
  mapping = NULL,
  data = NULL,
  geom = "text_npc",
  position = "identity",
  quadrants = NULL,
  pool.along = c("none", "x", "y", "xy"),
  xintercept = 0,
  yintercept = 0,
  label.x = NULL,
  label.y = NULL,
  digits = 2,
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = TRUE,
  ...
)
}
\arguments{
\item{mapping}{The aesthetic mapping, usually constructed with
\code{\link[ggplot2]{aes}} or \code{\link[ggplot2]{aes_}}. Only needs to be
set at the layer level if you are overriding the plot defaults.}

\item{data}{A layer specific dataset - only needed if you want to override
the plot defaults.}

\item{geom}{The geometric object to use display the data}

\item{position}{The position adjustment to use on this layer}

\item{quadrants}{integer vector indicating which quadrants are of interest,
with a \code{OL} indicating the whole plot.}

\item{pool.along}{character, one of \code{"none"}, \code{"x"} or \code{"y"},
indicating which quadrants to pool to calculate counts by pair of
quadrants.}

\item{xintercept, yintercept}{numeric the coordinates of the origin of the
quadrants.}

\item{label.x, label.y}{\code{numeric} Coordinates (in npc units) to be used
for absolute positioning of the labels.}

\item{digits}{integer Number of digits for fraction and percent labels.}

\item{na.rm}{a logical indicating whether \code{NA} values should be stripped
before the computation proceeds.}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
never includes, and \code{TRUE} always includes.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics, rather
than combining with them. This is most useful for helper functions that
define both data and aesthetics and should not inherit behaviour from the
default plot specification, e.g., \code{\link[ggplot2]{borders}}.}

\item{...}{other arguments passed on to \code{\link[ggplot2]{layer}}. This
can include aesthetics whose values you want to set, not map. See
\code{\link[ggplot2]{layer}} for more details.}
}
\value{
A plot layer instance. Using as output \code{data} the counts of
  observations per plot quadrant.
}
\description{
\code{stat_quadrant_counts()} counts the number of observations in each
quadrant of a plot panel. By default it adds a text label to the far corner
of each quadrant. It can also be used to obtain the total number of
observations in each of two pairs of quadrants or in the whole panel.
Grouping is ignored, so en every case a single count is computed for each
quadrant in a plot panel.
}
\details{
This statistic can be used to automatically count observations in
  each of the four quadrants of a plot, and by default add these counts as
  text labels. Values exactly equal to \code{xintercept} or
  \code{yintercept} are counted together with those larger than the
  intercepts. An argument value of zero, passed to formal parameter
  \code{quadrants} is interpreted as a request for the count of all
  observations in each plot panel.

  The default origin of quadrants is at \code{xintercept = 0},
  \code{yintercept = 0}. Also by default, counts are computed for all
  quadrants within the \emph{x} and \emph{y} scale limits, but ignoring any
  marginal scale expansion. The default positions of the labels is in the
  farthest corner or edge of each quadrant using npc coordinates.
  Consequently, when using facets even with free limits for \emph{x} and
  \emph{y} axes, the location of the labels is consistent across panels. This
  is achieved by use of \code{geom = "text_npc"} or \code{geom =
  "label_npc"}. To pass the positions in native data units, pass \code{geom =
  "text"} explicitly as argument.
}
\section{Computed variables}{
 Data frame with one to four rows, one for each
  quadrant for which counts are counted in \code{data}. \describe{
  \item{quadrant}{integer, one of 0:4} \item{x}{x value of label position in
  data units} \item{y}{y value of label position in data units} \item{npcx}{x
  value of label position in npc units} \item{npcy}{y value of label position
  in npc units} \item{count}{number of  observations in the quadrant(s)}
  \item{total}{number of onservations in data}
  \item{count.label}{number of observations as character}
  \item{pc.label}{percent of observations as character}
  \item{fr.label}{fraction of observations as character} }.

  As shown in one example below \code{\link[gginnards]{geom_debug}} can be
  used to print the computed values returned by any statistic. The output
  shown includes also values mapped to aesthetics, like \code{label} in the
  example.
}

\examples{

# generate artificial data
set.seed(4321)
x <- -50:50
y <- rnorm(length(x), mean = 0)
my.data <- data.frame(x, y)

# using automatically generated text labels, default origin at (0, 0)

ggplot(my.data, aes(x, y)) +
  geom_point() +
  geom_quadrant_lines() +
  stat_quadrant_counts()

ggplot(my.data, aes(x, y)) +
  geom_point() +
  geom_quadrant_lines() +
  stat_quadrant_counts(aes(label = after_stat(pc.label)))

ggplot(my.data, aes(x, y)) +
  geom_point() +
  geom_quadrant_lines() +
  stat_quadrant_counts(aes(label = after_stat(fr.label)))

ggplot(my.data, aes(x, y)) +
  geom_point() +
  geom_quadrant_lines() +
  stat_quadrant_counts(aes(label = after_stat(dec.label)))

ggplot(my.data, aes(x, y)) +
 geom_point() +
  geom_quadrant_lines() +
  stat_quadrant_counts(aes(label = sprintf("\%i observations", after_stat(count)))) +
  scale_y_continuous(expand = expansion(c(0.05, 0.15))) # reserve space

# user specified origin

ggplot(my.data, aes(x, y)) +
  geom_quadrant_lines(colour = "blue", xintercept = 10, yintercept = -1) +
  stat_quadrant_counts(colour = "blue", xintercept = 10, yintercept = -1) +
  geom_point() +
  scale_y_continuous(expand = expansion(mult = 0.15))

ggplot(my.data, aes(x, y)) +
  geom_quadrant_lines(colour = "blue", xintercept = 10, yintercept = -1) +
  stat_quadrant_counts(aes(label = after_stat(pc.label)),
                       colour = "blue", xintercept = 10, yintercept = -1) +
  geom_point() +
  scale_y_continuous(expand = expansion(mult = 0.15))

# more digits in labels

ggplot(my.data, aes(x, y)) +
  geom_quadrant_lines(colour = "blue", xintercept = 10, yintercept = -1) +
  stat_quadrant_counts(aes(label = after_stat(pc.label)), digits = 3,
                       colour = "blue", xintercept = 10, yintercept = -1) +
  geom_point() +
  scale_y_continuous(expand = expansion(mult = 0.15))

ggplot(my.data, aes(x, y)) +
  geom_quadrant_lines(colour = "blue", xintercept = 10, yintercept = -1) +
  stat_quadrant_counts(aes(label = after_stat(fr.label)),
                       colour = "blue", xintercept = 10, yintercept = -1) +
  geom_point() +
  scale_y_continuous(expand = expansion(mult = 0.15))

# grouped quadrants

ggplot(my.data, aes(x, y)) +
  geom_quadrant_lines(colour = "blue",
                      pool.along = "x") +
  stat_quadrant_counts(colour = "blue", label.x = "right",
                       pool.along = "x") +
  geom_point() +
  scale_y_continuous(expand = expansion(mult = 0.15))

# whole panel

ggplot(my.data, aes(x, y)) +
  geom_point() +
  stat_quadrant_counts(quadrants = 0, label.x = "left", label.y = "bottom") +
  scale_y_continuous(expand = expansion(mult = c(0.15, 0.05)))

# use a different geometry

ggplot(my.data, aes(x, y)) +
  geom_point() +
  stat_quadrant_counts(geom = "text") # use geom_text()

# Numeric values can be used to build labels with alternative formats
# Here with sprintf(), but paste() and format() also work.

ggplot(my.data, aes(x, y)) +
  geom_quadrant_lines(colour = "blue") +
  stat_quadrant_counts(aes(label = sprintf("\%i of \%i genes",
                       after_stat(count), after_stat(total))),
                       colour = "blue") +
  geom_point() +
  scale_y_continuous(expand = expansion(mult = 0.15))

# We use geom_debug_group() to see the computed values

gginnards.installed <- requireNamespace("gginnards", quietly = TRUE)
if (gginnards.installed) {
  library(gginnards)

  ggplot(my.data, aes(x, y)) +
    geom_point() +
    stat_quadrant_counts(geom = "debug_group")
}

if (gginnards.installed) {
  ggplot(my.data, aes(x, y)) +
    geom_point() +
    stat_quadrant_counts(geom = "debug_group", xintercept = 50)
}

}
\seealso{
Other Functions for quadrant and volcano plots:
\code{\link[=geom_quadrant_lines]{geom_quadrant_lines()}},
\code{\link[=stat_panel_counts]{stat_panel_counts()}}
}
\concept{Functions for quadrant and volcano plots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-margin-grob.r
\name{geom_x_margin_grob}
\alias{geom_x_margin_grob}
\alias{geom_y_margin_grob}
\title{Add Grobs on the margins}
\usage{
geom_x_margin_grob(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  ...,
  xintercept,
  sides = "b",
  grob.shift = 0,
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = FALSE
)

geom_y_margin_grob(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  ...,
  yintercept,
  sides = "l",
  grob.shift = 0,
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = FALSE
)
}
\arguments{
\item{mapping}{The aesthetic mapping, usually constructed with
\code{\link[ggplot2]{aes}}. Only needs to be set at the layer level if you
are overriding the plot defaults.}

\item{data}{A layer specific dataset - only needed if you want to override
the plot defaults.}

\item{stat}{The statistical transformation to use on the data for this layer,
as a string.}

\item{position}{Position adjustment, either as a string, or the result of a
call to a position adjustment function.}

\item{...}{other arguments passed on to \code{\link[ggplot2]{layer}}. This
can include aesthetics whose values you want to set, not map. See
\code{\link[ggplot2]{layer}} for more details.}

\item{xintercept, yintercept}{numeric Parameters that control the position of
the marginal points. If these are set, data, mapping and show.legend are
overridden.}

\item{sides}{A character string of length one that controls on which side of
the plot the grob annotations appear on. It can be set to a string
containing one of \code{"t"}, \code{"r"}, \code{"b"} or \code{"l"}, for
top, right, bottom, and left.}

\item{grob.shift}{numeric value expressed in npc units for the shift of the
marginal grob inwards from the edge of the plotting area.}

\item{na.rm}{If \code{FALSE} (the default), removes missing values with a
warning.  If \code{TRUE} silently removes missing values.}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
never includes, and \code{TRUE} always includes.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics, rather
than combining with them. This is most useful for helper functions that
define both data and aesthetics and shouldn't inherit behaviour from the
default plot specification, e.g. \code{\link[ggplot2]{borders}}.}
}
\value{
A plot layer instance.
}
\description{
Margin grobs can supplement a 2d display with annotations. Margin grobs such
as icons or symbols can highlight individual values along a margin. The
geometries \code{geom_x_margin_grob()} and \code{geom_y_margin_grob()} behave
similarly \code{geom_vline()} and \code{geom_hline()} and share their "double
personality" as both annotations and geometries.
}
\section{Alignment}{
 You can modify text alignment with the \code{vjust} and
  \code{hjust} aesthetics. These can either be a number between 0
  (right/bottom) and 1 (top/left) or a character (\code{"left"},
  \code{"middle"}, \code{"right"}, \code{"bottom"}, \code{"center"},
  \code{"top"}). In addition, you can use special alignments for
  justification including \code{"position"}, \code{"inward"} and
  \code{"outward"}. Inward always aligns text towards the center of the
  plotting area, and outward aligns it away from the center of the plotting
  area. If tagged with \code{_mean} or \code{_median} (e.g.,
  \code{"outward_mean"}) the mean or median of the data in the panel along
  the corresponding axis is used as center. If the characters following the
  underscore represent a number (e.g., \code{"outward_10.5"}) the reference
  point will be this value in data units. Position justification is computed
  based on the direction of the displacement of the position of the label so
  that each individual text or label is justified outwards from its original
  position. The default justification is \code{"position"}.

  If no position displacement is applied, or a position function defined in
  'ggplot2' is used, these geometries behave similarly to the corresponding
  ones from package 'ggplot2' with a default justification of \code{0.5} and
  no segment drawn.
}

\section{Position functions}{
 Many layer functions from package 'ggpp' are
  designed to work seamlessly with position functions that keep, rather than
  discard, the original \code{x} and \code{y} positions in \code{data} when
  computing a new displaced position. See \code{\link{position_nudge_keep}},
  \code{\link{position_dodge_keep}}, \code{\link{position_jitter_keep}},
  \code{\link{position_nudge_center}}, \code{\link{position_nudge_line}},
  \code{\link{position_nudge_to}}, \code{\link{position_dodgenudge}},
  \code{\link{position_jitternudge}}, and \code{\link{position_stacknudge}}
  for examples and details of their use.
}

\examples{
# We can add icons to the margin of a plot to signal events




}
\seealso{
\code{\link[grid]{grid-package}}, \code{\link[ggplot2]{geom_rug}},
  and other documentation of package 'ggplot2'.

Other Geometries for marginal annotations in ggplots:
\code{\link[=geom_x_margin_arrow]{geom_x_margin_arrow()}},
\code{\link[=geom_x_margin_point]{geom_x_margin_point()}}
}
\concept{Geometries for marginal annotations in ggplots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-plot.R
\name{geom_plot}
\alias{geom_plot}
\alias{geom_plot_npc}
\title{Inset plots}
\usage{
geom_plot(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  ...,
  nudge_x = 0,
  nudge_y = 0,
  default.colour = NULL,
  default.color = default.colour,
  colour.target = "box",
  color.target = colour.target,
  default.alpha = 1,
  alpha.target = "all",
  add.segments = TRUE,
  box.padding = 0.25,
  point.padding = 1e-06,
  segment.linewidth = 0.5,
  min.segment.length = 0,
  arrow = NULL,
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = FALSE
)

geom_plot_npc(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  ...,
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = FALSE
)
}
\arguments{
\item{mapping}{The aesthetic mapping, usually constructed with
\code{\link[ggplot2]{aes}}. Only needs to be set at the layer level if you
are overriding the plot defaults.}

\item{data}{A layer specific dataset - only needed if you want to override
the plot defaults.}

\item{stat}{The statistical transformation to use on the data for this layer,
as a string.}

\item{position}{Position adjustment, either as a string, or the result of a
call to a position adjustment function.}

\item{...}{other arguments passed on to \code{\link[ggplot2]{layer}}. This
can include aesthetics whose values you want to set, not map. See
\code{\link[ggplot2]{layer}} for more details.}

\item{nudge_x, nudge_y}{Horizontal and vertical adjustments to nudge the
starting position of each "label". The units for \code{nudge_x} and
\code{nudge_y} are the same as for the data units on the x-axis and y-axis.}

\item{default.colour, default.color}{A colour definition to use for elements not targeted by
the colour aesthetic.}

\item{colour.target, color.target}{A vector of character strings; \code{"all"},
\code{"text"}, \code{"box"} and \code{"segment"}.}

\item{default.alpha}{numeric in [0..1] A transparency value to use for
elements not targeted by the alpha aesthetic.}

\item{alpha.target}{A vector of character strings; \code{"all"},
\code{"text"}, \code{"segment"}, \code{"box"}, \code{"box.line"}, and
\code{"box.fill"}.}

\item{add.segments}{logical Display connecting segments or arrows between
original positions and displaced ones if both are available.}

\item{box.padding, point.padding}{numeric By how much each end of the segments
should shortened in mm.}

\item{segment.linewidth}{numeric Width of the segments or arrows in mm.}

\item{min.segment.length}{numeric Segments shorter that the minimum length
are not rendered, in mm.}

\item{arrow}{specification for arrow heads, as created by
\code{\link[grid]{arrow}}}

\item{na.rm}{If \code{FALSE} (the default), removes missing values with a
warning.  If \code{TRUE} silently removes missing values.}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
never includes, and \code{TRUE} always includes.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics, rather
than combining with them. This is most useful for helper functions that
define both data and aesthetics and shouldn't inherit behaviour from the
default plot specification, e.g. \code{\link[ggplot2]{borders}}.}
}
\value{
A plot layer instance.
}
\description{
\code{geom_plot} and \code{geom_plot_npc} add ggplot objects as insets to the
base ggplot, using syntax similar to that of
\code{\link[ggplot2]{geom_label}} and \code{\link{geom_text_s}}.
In most respects they behave as any other ggplot geometry: they add a layer
containing one or more grobs and grouping and faceting works as usual. The
most common use of \code{geom_plot} is to add data labels that are themselves
ggplots rather than text. \code{\link{geom_plot_npc}} is used to add ggplots
as annotations to plots, but contrary to layer function \code{annotate()},
\code{\link{geom_plot_npc}} is data driven and respects grouping and facets,
thus plot insets can differ among panels.
}
\details{
You can modify the size of inset plots with the \code{vp.width} and
  \code{vp.height} aesthetics. These can take a number between 0 (smallest
  possible inset) and 1 (whole plotting area width or height). The default
  value for for both of these aesthetics is 1/5. Thus, in contrast to
  \code{\link[ggplot2]{geom_text}} and \code{\link{geom_text_s}} the size of
  the insets remains the same relative to the size of the plotting area
  irrespective of how the plot is rendered. The aspect ratio of insets is
  preserved and size is adjusted until the whole inset fits within the
  viewport.

  \code{geom_plot} and \code{geom_plot_npc} expect a list of ggplot objects
  (\code{"gg"} class) to be mapped to the \code{label} aesthetic. These geoms
  work with tibbles or data frames as \code{data} as they both support
  \code{list} objects as member variables.

  The \code{x} and \code{y} aesthetics determine the position of the whole
  inset plot, similarly to that of a text label, justification is interpreted
  as indicating the position of the plot with respect to its x and y
  coordinates in the data, and \code{angle} is used to rotate the plot as a
  whole.

  Of these two geoms only \code{\link{geom_plot}} supports the plotting of
  segments, as \code{\link{geom_plot_npc}} uses a coordinate system that is
  unrelated to data units and data.In the case of \code{geom_plot_npc()},
  \code{npcx} and \code{npcy} aesthetics determine the position of the inset
  plot.
}
\note{
The insets are stored nested within the main ggplot object and
  contain their own copy of the data, and are rendered as grid grobs as normal
  ggplots at the time the main ggplot is rendered. They can have different
  themes.

  Use \code{\link{annotate}} as redefined in 'ggpp' when adding insets
  as annotations (automatically available unless 'ggpp' is not attached).
  \code{\link[ggplot2]{annotate}} cannot be used with the \code{npcx} and
  \code{npcy} pseudo-aesthetics.
}
\section{Alignment}{
 You can modify text alignment with the \code{vjust} and
  \code{hjust} aesthetics. These can either be a number between 0
  (right/bottom) and 1 (top/left) or a character (\code{"left"},
  \code{"middle"}, \code{"right"}, \code{"bottom"}, \code{"center"},
  \code{"top"}). In addition, you can use special alignments for
  justification including \code{"position"}, \code{"inward"} and
  \code{"outward"}. Inward always aligns text towards the center of the
  plotting area, and outward aligns it away from the center of the plotting
  area. If tagged with \code{_mean} or \code{_median} (e.g.,
  \code{"outward_mean"}) the mean or median of the data in the panel along
  the corresponding axis is used as center. If the characters following the
  underscore represent a number (e.g., \code{"outward_10.5"}) the reference
  point will be this value in data units. Position justification is computed
  based on the direction of the displacement of the position of the label so
  that each individual text or label is justified outwards from its original
  position. The default justification is \code{"position"}.

  If no position displacement is applied, or a position function defined in
  'ggplot2' is used, these geometries behave similarly to the corresponding
  ones from package 'ggplot2' with a default justification of \code{0.5} and
  no segment drawn.
}

\section{Position functions}{
 Many layer functions from package 'ggpp' are
  designed to work seamlessly with position functions that keep, rather than
  discard, the original \code{x} and \code{y} positions in \code{data} when
  computing a new displaced position. See \code{\link{position_nudge_keep}},
  \code{\link{position_dodge_keep}}, \code{\link{position_jitter_keep}},
  \code{\link{position_nudge_center}}, \code{\link{position_nudge_line}},
  \code{\link{position_nudge_to}}, \code{\link{position_dodgenudge}},
  \code{\link{position_jitternudge}}, and \code{\link{position_stacknudge}}
  for examples and details of their use.
}

\section{Plot boundaries and clipping}{
 The "width" and "height" of an inset
  as for a text element are 0, so stacking and dodging inset plots will not
  work by default, and axis limits are not automatically expanded to include
  all inset plots. Obviously, insets do have height and width, but they are
  physical units, not data units. The amount of space they occupy on the main
  plot is not constant in data units of the base plot: when you modify scale
  limits, inset plots stay the same size relative to the physical size of the
  base plot.
}

\examples{
# inset plot with enlarged detail from a region of the main plot
library(tibble)
p <-
  ggplot(data = mtcars, mapping = aes(wt, mpg)) +
  geom_point()

p.crop <- p +
          coord_cartesian(xlim = c(3, 4),
          ylim = c(13, 16)) +
          labs(x = NULL, y = NULL) +
          theme_bw(10)

df <- data.frame(x = 0.01,
                 y = 0.01,
                 plot = I(list(p.crop)))
p +
  expand_limits(x = 0, y = 0) +
  geom_plot_npc(data = df,
                aes(npcx = x, npcy = y, label = plot))

p +
  expand_limits(x = 0, y = 0) +
  geom_plot_npc(data = df,
                aes(npcx = x, npcy = y, label = plot,
                vp.width = 1/2, vp.height = 1/4))

p +
  expand_limits(x = 0, y = 0) +
  geom_plot_npc(data = df,
                aes(npcx = x, npcy = y, label = plot),
                vp.width = 1/4, vp.height = 1/4)

p +
  geom_plot(data = df,
            aes(x = x + 3, y = y + 20, label = plot),
            nudge_x = -1, nudge_y = - 7,
            hjust = 0.5, vjust = 0.5,
            arrow = arrow(length = unit(0.5, "lines")),
            colour = "red",
            vp.width = 1/5, vp.height = 1/5)

}
\references{
The idea of implementing a \code{geom_custom()} for grobs has
  been discussed as an issue at
  \url{https://github.com/tidyverse/ggplot2/issues/1399}.
}
\seealso{
Other geometries adding layers with insets:
\code{\link[=geom_table]{geom_table()}}
}
\concept{geometries adding layers with insets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-grob.R
\name{geom_grob}
\alias{geom_grob}
\alias{geom_grob_npc}
\title{Inset graphical objects}
\usage{
geom_grob(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  ...,
  nudge_x = 0,
  nudge_y = 0,
  default.colour = NULL,
  default.color = default.colour,
  colour.target = "segment",
  color.target = colour.target,
  default.alpha = 1,
  alpha.target = "segment",
  add.segments = TRUE,
  box.padding = 0.25,
  point.padding = 1e-06,
  segment.linewidth = 0.5,
  min.segment.length = 0,
  arrow = NULL,
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = FALSE
)

geom_grob_npc(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  ...,
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = FALSE
)
}
\arguments{
\item{mapping}{The aesthetic mapping, usually constructed with
\code{\link[ggplot2]{aes}}. Only needs to be set at the layer level if you
are overriding the plot defaults.}

\item{data}{A layer specific dataset - only needed if you want to override
the plot defaults.}

\item{stat}{The statistical transformation to use on the data for this layer,
as a string.}

\item{position}{Position adjustment, either as a string, or the result of a
call to a position adjustment function.}

\item{...}{other arguments passed on to \code{\link[ggplot2]{layer}}. This
can include aesthetics whose values you want to set, not map. See
\code{\link[ggplot2]{layer}} for more details.}

\item{nudge_x, nudge_y}{Horizontal and vertical adjustments to nudge the
starting position of each "label". The units for \code{nudge_x} and
\code{nudge_y} are the same as for the data units on the x-axis and y-axis.}

\item{default.colour, default.color}{A colour definition to use for elements
not targeted by the colour aesthetic.}

\item{colour.target, color.target}{A vector of character strings; \code{"all"},
\code{"text"}, \code{"box"} and \code{"segment"} or \code{"none"}.}

\item{default.alpha}{numeric in [0..1] A transparency value to use for
elements not targeted by the alpha aesthetic.}

\item{alpha.target}{A vector of character strings; \code{"all"},
\code{"text"}, \code{"segment"}, \code{"box"}, \code{"box.line"}, and
\code{"box.fill"} or \code{"none"}.}

\item{add.segments}{logical Display connecting segments or arrows between
original positions and displaced ones if both are available.}

\item{box.padding, point.padding}{numeric By how much each end of the segments
should shortened in mm.}

\item{segment.linewidth}{numeric Width of the segments or arrows in mm.}

\item{min.segment.length}{numeric Segments shorter that the minimum length
are not rendered, in mm.}

\item{arrow}{specification for arrow heads, as created by
\code{\link[grid]{arrow}}}

\item{na.rm}{If \code{FALSE} (the default), removes missing values with a
warning.  If \code{TRUE} silently removes missing values.}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
never includes, and \code{TRUE} always includes.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics, rather
than combining with them. This is most useful for helper functions that
define both data and aesthetics and shouldn't inherit behaviour from the
default plot specification, e.g. \code{\link[ggplot2]{borders}}.}
}
\value{
A plot layer instance.
}
\description{
\code{geom_grob} and \code{geom_grob_npc} add Grobs as insets to the ggplot
using syntax similar to that of \code{\link[ggplot2]{geom_text}},
\code{\link{geom_text_s}} and \code{\link{geom_text_npc}}.
In most respects they behave as any other ggplot geometry: they add a layer
containing one or more grobs and grouping and faceting works as usual. The
most common use of \code{geom_grob} is to add data labels that are graphical
objects rather than text. \code{\link{geom_grob_npc}} is used to add grobs
as annotations to plots, but contrary to layer function \code{annotate()},
\code{\link{geom_grob_npc}} is data driven and respects grouping and facets,
thus plot insets can differ among panels. Of these two geoms only
\code{\link{geom_grob}} supports the plotting of segments, as
\code{\link{geom_grob_npc}} uses a coordinate system that is unrelated
to data units and data.
}
\details{
You can modify the size of insets with the \code{vp.width} and
  \code{vp.height} aesthetics. These can take a number between 0 (smallest
  possible inset) and 1 (whole plotting area width or height). The default
  value for for both of these aesthetics is 1/5. Thus, in contrast to
  \code{\link[ggplot2]{geom_text}}, \code{\link[ggplot2]{geom_label}},
  \code{\link{geom_text_s}} and \code{\link{geom_label_s}} the size of the
  insets remains the same relative to the size of the plotting area
  irrespective of the size the plot is rendered at. The aspect ratio of
  insets is preserved and size is adjusted until the whole inset fits within
  the viewport.

  By default \code{geom_grob} uses \code{\link{position_nudge_center}} and
  justification \code{"position"}, while \code{geom_grob_npc} uses
  \code{\link[ggplot2]{position_nudge}} and justification \code{"inward"}. In
  contrast to \code{\link[ggplot2]{position_nudge}},
  \code{\link{position_nudge_center}} and all other position functions
  defined in packages 'ggpp' keep the original coordinates thus allowing the
  plotting of connecting segments and arrows.

  \code{geom_grob} and \code{geom_grob_npc} expect a list of graphic objects
  ("grob") to be mapped to the \code{label} aesthetic. These geoms work with
  tibbles or data frames as \code{data} as they both support \code{list}
  objects as member variables.

  The \code{x} and \code{y} aesthetics determine the position of the whole
  inset grob, similarly to that of a text label, justification is interpreted
  as indicating the position of the grob with respect to its \emph{x} and
  \emph{y} coordinates in the data, and \code{angle} is used to rotate the
  grob as a whole.
}
\note{
The insets are stored nested within the main ggplot object and
  contain their own copy of the data, and are rendered as grid grobs as normal
  ggplots at the time the main ggplot is rendered. They can have different
  themes.

  Use \code{\link{annotate}} as redefined in 'ggpp' when adding insets
  as annotations (automatically available unless 'ggpp' is not attached).
  \code{\link[ggplot2]{annotate}} cannot be used with the \code{npcx} and
  \code{npcy} pseudo-aesthetics.
}
\section{Plot boundaries and clipping}{
 The "width" and "height" of an inset
  as for a text element are 0, so stacking and dodging inset plots will not
  work by default, and axis limits are not automatically expanded to include
  all inset plots. Obviously, insets do have height and width, but they are
  physical units, not data units. The amount of space they occupy on the main
  plot is not constant in data units of the base plot: when you modify scale
  limits, inset plots stay the same size relative to the physical size of the
  base plot.
}

\section{Alignment}{
 You can modify text alignment with the \code{vjust} and
  \code{hjust} aesthetics. These can either be a number between 0
  (right/bottom) and 1 (top/left) or a character (\code{"left"},
  \code{"middle"}, \code{"right"}, \code{"bottom"}, \code{"center"},
  \code{"top"}). In addition, you can use special alignments for
  justification including \code{"position"}, \code{"inward"} and
  \code{"outward"}. Inward always aligns text towards the center of the
  plotting area, and outward aligns it away from the center of the plotting
  area. If tagged with \code{_mean} or \code{_median} (e.g.,
  \code{"outward_mean"}) the mean or median of the data in the panel along
  the corresponding axis is used as center. If the characters following the
  underscore represent a number (e.g., \code{"outward_10.5"}) the reference
  point will be this value in data units. Position justification is computed
  based on the direction of the displacement of the position of the label so
  that each individual text or label is justified outwards from its original
  position. The default justification is \code{"position"}.

  If no position displacement is applied, or a position function defined in
  'ggplot2' is used, these geometries behave similarly to the corresponding
  ones from package 'ggplot2' with a default justification of \code{0.5} and
  no segment drawn.
}

\section{Position functions}{
 Many layer functions from package 'ggpp' are
  designed to work seamlessly with position functions that keep, rather than
  discard, the original \code{x} and \code{y} positions in \code{data} when
  computing a new displaced position. See \code{\link{position_nudge_keep}},
  \code{\link{position_dodge_keep}}, \code{\link{position_jitter_keep}},
  \code{\link{position_nudge_center}}, \code{\link{position_nudge_line}},
  \code{\link{position_nudge_to}}, \code{\link{position_dodgenudge}},
  \code{\link{position_jitternudge}}, and \code{\link{position_stacknudge}}
  for examples and details of their use.
}

\examples{
library(tibble)
df <- tibble(x = 2, y = 15, grob = list(grid::circleGrob(r = 0.2)))

# without nudging no segments are drawn
ggplot(data = mtcars,
       aes(wt, mpg)) +
  geom_point(aes(colour = factor(cyl))) +
  geom_grob(data = df,
            aes(x, y, label = grob))

# with nudging segments are drawn
ggplot(data = mtcars,
       aes(wt, mpg)) +
  geom_point(aes(colour = factor(cyl))) +
  geom_grob(data = df,
            aes(x, y, label = grob),
            nudge_x = 0.5,
            colour = "red",
            hjust = 0.5,
            vjust = 0.5)

ggplot(data = mtcars,
       aes(wt, mpg)) +
  geom_point(aes(colour = factor(cyl))) +
  geom_grob(data = df,
            aes(x, y, label = grob),
            nudge_x = 0.5,
            colour = "red",
            colour.target = "none",
            hjust = 0.5,
            vjust = 0.5)

# with nudging plotting of segments can be disabled
ggplot(data = mtcars,
       aes(wt, mpg)) +
  geom_point(aes(colour = factor(cyl))) +
  geom_grob(data = df,
            aes(x, y, label = grob),
            add.segments = FALSE,
            nudge_x = 0.5,
            hjust = 0.5,
            vjust = 0.5)

}
\references{
The idea of implementing a \code{geom_custom()} for grobs has
  been discussed as an issue at
  \url{https://github.com/tidyverse/ggplot2/issues/1399}.
}
\seealso{
\code{\link[grid]{grid-package}}, \code{\link[ggplot2]{geom_text}},
  and other documentation of package 'ggplot2'.
}
\concept{geometries adding layers with insets.}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat-dens1d-filter.r
\name{stat_dens1d_filter}
\alias{stat_dens1d_filter}
\alias{stat_dens1d_filter_g}
\title{Filter observations by local 1D density}
\usage{
stat_dens1d_filter(
  mapping = NULL,
  data = NULL,
  geom = "point",
  position = "identity",
  ...,
  keep.fraction = 0.1,
  keep.number = Inf,
  keep.sparse = TRUE,
  keep.these = FALSE,
  exclude.these = FALSE,
  these.target = "label",
  pool.along = c("x", "none"),
  xintercept = 0,
  invert.selection = FALSE,
  bw = "SJ",
  kernel = "gaussian",
  adjust = 1,
  n = 512,
  return.density = FALSE,
  orientation = c("x", "y"),
  na.rm = TRUE,
  show.legend = FALSE,
  inherit.aes = TRUE
)

stat_dens1d_filter_g(
  mapping = NULL,
  data = NULL,
  geom = "point",
  position = "identity",
  keep.fraction = 0.1,
  keep.number = Inf,
  keep.sparse = TRUE,
  keep.these = FALSE,
  exclude.these = FALSE,
  these.target = "label",
  pool.along = c("x", "none"),
  xintercept = 0,
  invert.selection = FALSE,
  na.rm = TRUE,
  show.legend = FALSE,
  inherit.aes = TRUE,
  bw = "SJ",
  adjust = 1,
  kernel = "gaussian",
  n = 512,
  return.density = FALSE,
  orientation = c("x", "y"),
  ...
)
}
\arguments{
\item{mapping}{The aesthetic mapping, usually constructed with
\code{\link[ggplot2]{aes}} or \code{\link[ggplot2]{aes_}}. Only needs to be
set at the layer level if you are overriding the plot defaults.}

\item{data}{A layer specific dataset - only needed if you want to override
the plot defaults.}

\item{geom}{The geometric object to use display the data.}

\item{position}{The position adjustment to use for overlapping points on this
layer}

\item{...}{other arguments passed on to \code{\link[ggplot2]{layer}}. This
can include aesthetics whose values you want to set, not map. See
\code{\link[ggplot2]{layer}} for more details.}

\item{keep.fraction}{numeric vector of length 1 or 2 [0..1]. The fraction of
the observations (or rows) in \code{data} to be retained.}

\item{keep.number}{integer vector of length 1 or 2. Set the maximum number of
observations to retain, effective only if obeying \code{keep.fraction}
would result in a larger number.}

\item{keep.sparse}{logical If \code{TRUE}, the default, observations from the
more sparse regions are retained, if \code{FALSE} those from the densest
regions.}

\item{keep.these, exclude.these}{character vector, integer vector, logical
vector or function that takes one or more variables in data selected by
\code{these.target}. Negative integers behave as in R's extraction methods.
The rows from \code{data} indicated by \code{keep.these} and
\code{exclude.these} are kept or excluded irrespective of the local
density.}

\item{these.target}{character, numeric or logical selecting one or more
column(s) of \code{data}. If \code{TRUE} the whole \code{data} object is
passed.}

\item{pool.along}{character, one of \code{"none"} or \code{"x"},
indicating if selection should be done pooling the observations along the
\emph{x} aesthetic, or separately on either side of \code{xintercept}.}

\item{xintercept}{numeric The split point for the data filtering. If
\code{NA} the data are not split.}

\item{invert.selection}{logical If \code{TRUE}, the complement of the
selected rows are returned.}

\item{bw}{numeric or character The smoothing bandwidth to be used. If
numeric, the standard deviation of the smoothing kernel. If character, a
rule to choose the bandwidth, as listed in \code{\link[stats]{bw.nrd}}.}

\item{kernel}{character See \code{\link{density}} for details.}

\item{adjust}{numeric A multiplicative bandwidth adjustment. This makes it
possible to adjust the bandwidth while still using the a bandwidth
estimator through an argument passed to \code{bw}. The larger the value
passed to \code{adjust} the stronger the smoothing, hence decreasing
sensitivity to local changes in density.}

\item{n}{numeric Number of equally spaced points at which the density is to
be estimated for applying the cut point. See \code{\link{density}} for
details.}

\item{return.density}{logical vector of lenght 1. If \code{TRUE} add columns
\code{"density"} and \code{"keep.obs"} to the returned data frame.}

\item{orientation}{character The aesthetic along which density is computed.
Given explicitly by setting orientation to either \code{"x"} or \code{"y"}.}

\item{na.rm}{a logical value indicating whether \code{NA} values should be
stripped before the computation proceeds.}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
never includes, and \code{TRUE} always includes.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics, rather
than combining with them. This is most useful for helper functions that
define both data and aesthetics and shouldn't inherit behaviour from the
default plot specification, e.g. \code{\link[ggplot2]{borders}}.}
}
\value{
A plot layer instance. Using as output \code{data} a subset of the
  rows in input \code{data} retained based on a 1D filtering criterion.
}
\description{
\code{stat_dens1d_filter} Filters-out/filters-in observations in
  regions of a plot panel with high density of observations, based on the
  values mapped to one of \code{x} and \code{y} aesthetics.
  \code{stat_dens1d_filter_g} does the same filtering by group instead of by
  panel. This second stat is useful for highlighting observations, while the
  first one tends to be most useful when the aim is to prevent clashes among
  text labels. By default the data are handled all together, but it is also
  possible to control labeling separately in each tail.
}
\details{
The 1D density of observations of \emph{x} or \emph{y} is computed
  with function \code{\link[stats]{density}} and used to select observations,
  passing to the geom a subset of the rows in its \code{data} input. The
  default is to select observations in sparse regions of the plot, but the
  selection can be inverted so that only observations in the densest regions
  are returned. Specific observations can be protected from being deselected
  and "kept" by passing a suitable argument to \code{keep.these}. Logical and
  integer vectors work as indexes to rows in \code{data}, while a values in a
  character vector are compared to the character values mapped to the
  \code{label} aesthetic. A function passed as argument to keep.these will
  receive as argument the values in the variable mapped to \code{label} and
  should return a character, logical or numeric vector as described above. If
  no variable has been mapped to \code{label}, row names are used in its
  place.

  How many rows are retained in addition to those in \code{keep.these} is
  controlled with arguments passed to \code{keep.number} and
  \code{keep.fraction}. \code{keep.number} sets the maximum number of
  observations selected, whenever \code{keep.fraction} results in fewer
  observations selected, it is obeyed. If `xintercept` is a finite value
  within the \emph{x} range of the data and \code{pool.along}
  is passed \code{"none"} the data as are split into two groups
  and \code{keep.number} and \code{keep.fraction} are applied separately to
  each tail with density still computed jointly from all observations. If the
  length of \code{keep.number} and \code{keep.fraction} is one, this value
  is used for both tails, if their length is two, the first value is use
  for the left tail and the second value for the right tail.

  Computation of density and of the default bandwidth require at least
  two observations with different values. If data do not fulfill this
  condition, they are kept only if \code{keep.fraction = 1}. This is correct
  behavior for a single observation, but can be surprising in the case of
  multiple observations.

  Parameters \code{keep.these} and \code{exclude.these} make it possible to
  force inclusion or exclusion of observations after the density is computed.
  In case of conflict, \code{exclude.these} overrides \code{keep.these}.
}
\note{
Which points are kept and which not depends on how dense and flexible
  is the density curve estimate. This depends on the values passed as
  arguments to parameters \code{n}, \code{bw} and \code{kernel}. It is
  also important to be aware that both \code{geom_text()} and
  \code{geom_text_repel()} can avoid over plotting by discarding labels at
  the plot rendering stage, i.e., what is plotted may differ from what is
  returned by this statistic.
}
\examples{

random_string <-
  function(len = 6) {
    paste(sample(letters, len, replace = TRUE), collapse = "")
  }

# Make random data.
set.seed(1001)
d <- tibble::tibble(
  x = rnorm(100),
  y = rnorm(100),
  group = rep(c("A", "B"), c(50, 50)),
  lab = replicate(100, { random_string() })
)
d$xg <- d$x
d$xg[51:100] <- d$xg[51:100] + 1

# highlight the 1/10 of observations in sparsest regions of the plot
ggplot(data = d, aes(x, y)) +
  geom_point() +
  geom_rug(sides = "b") +
  stat_dens1d_filter(colour = "red") +
  stat_dens1d_filter(geom = "rug", colour = "red", sides = "b")

# highlight the 1/4 of observations in densest regions of the plot
ggplot(data = d, aes(x, y)) +
  geom_point() +
  geom_rug(sides = "b") +
  stat_dens1d_filter(colour = "blue",
                     keep.fraction = 1/4, keep.sparse = FALSE) +
  stat_dens1d_filter(geom = "rug", colour = "blue",
                     keep.fraction = 1/4, keep.sparse = FALSE,
                     sides = "b")

# switching axes
ggplot(data = d, aes(x, y)) +
  geom_point() +
  geom_rug(sides = "l") +
  stat_dens1d_filter(colour = "red", orientation = "y") +
  stat_dens1d_filter(geom = "rug", colour = "red", orientation = "y",
                     sides = "l")

# highlight 1/10 plus 1/10 observations in high and low density regions
ggplot(data = d, aes(x, y)) +
  geom_point() +
  geom_rug(sides = "b") +
  stat_dens1d_filter(colour = "red") +
  stat_dens1d_filter(geom = "rug", colour = "red", sides = "b") +
  stat_dens1d_filter(colour = "blue", keep.sparse = FALSE) +
  stat_dens1d_filter(geom = "rug",
                     colour = "blue", keep.sparse = FALSE, sides = "b")

# selecting the 1/10 observations in sparsest regions and their complement
ggplot(data = d, aes(x, y)) +
  stat_dens1d_filter(colour = "red") +
  stat_dens1d_filter(geom = "rug", colour = "red", sides = "b") +
  stat_dens1d_filter(colour = "blue", invert.selection = TRUE) +
  stat_dens1d_filter(geom = "rug",
                     colour = "blue", invert.selection = TRUE, sides = "b")

# density filtering done jointly across groups
ggplot(data = d, aes(xg, y, colour = group)) +
  geom_point() +
  geom_rug(sides = "b", colour = "black") +
  stat_dens1d_filter(shape = 1, size = 3, keep.fraction = 1/4, adjust = 2)

# density filtering done independently for each group
ggplot(data = d, aes(xg, y, colour = group)) +
  geom_point() +
  geom_rug(sides = "b") +
  stat_dens1d_filter_g(shape = 1, size = 3, keep.fraction = 1/4, adjust = 2)

# density filtering done jointly across groups by overriding grouping
ggplot(data = d, aes(xg, y, colour = group)) +
  geom_point() +
  geom_rug(sides = "b") +
  stat_dens1d_filter_g(colour = "black",
                       shape = 1, size = 3, keep.fraction = 1/4, adjust = 2)

# label observations
ggplot(data = d, aes(x, y, label = lab, colour = group)) +
  geom_point() +
  stat_dens1d_filter(geom = "text", hjust = "outward")

# looking under the hood with gginnards::geom_debug_group()
gginnards.installed <- requireNamespace("gginnards", quietly = TRUE)
if (gginnards.installed) {
  library(gginnards)

  ggplot(data = d, aes(x, y, label = lab, colour = group)) +
    stat_dens1d_filter(geom = "debug_group")
}

if (gginnards.installed) {
  ggplot(data = d, aes(x, y, label = lab, colour = group)) +
    stat_dens1d_filter(geom = "debug_group", return.density = TRUE)
}
}
\seealso{
\code{\link[stats]{density}} used internally.

Other statistics returning a subset of data:
\code{\link[=stat_dens1d_labels]{stat_dens1d_labels()}},
\code{\link[=stat_dens2d_filter]{stat_dens2d_filter()}},
\code{\link[=stat_dens2d_labels]{stat_dens2d_labels()}}
}
\concept{statistics returning a subset of data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggpp.R
\docType{package}
\name{ggpp-package}
\alias{ggpp}
\alias{ggpp-package}
\title{ggpp: Grammar Extensions to 'ggplot2'}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Extensions to 'ggplot2' respecting the grammar of graphics paradigm. Geometries: geom_table(), geom_plot() and geom_grob() add insets to plots using native data coordinates, while geom_table_npc(), geom_plot_npc() and geom_grob_npc() do the same using "npc" coordinates through new aesthetics "npcx" and "npcy". Statistics: select observations based on 2D density. Positions: radial nudging away from a center point and nudging away from a line or curve; combined stacking and nudging; combined dodging and nudging.
}
\details{
Package 'ggpp' provides functions that extend the grammar of
graphics as implemented in 'ggplot2'. It attempts to stay true to the
original grammar and to respect the naming conventions used in 'ggplot2'.

Extensions provided:
\itemize{
\item Geoms adding support for plot, table and grob insets within the
gramamr. Geoms using a parallel pseudo-scale based on native plot coordinates
(npc) to allow annotations consistent with the grammar and so supporting
facets and grouping. Geoms for annotations on the edges of the plotting
area. Geom for easily drawing lines separating the quadrants of a plot.
\item Stats for filtering-out/filtering-in observations in regions of a
panel or group where the density of observations is high. Statistics
simultaneously computing summaries, optionally using different functions,
along x and y. Stat computing quadrant counts.
\item Position functions implementing multi-directional nudging based on the
data.
\item Scales. Pseudo-scales supporting npc coordinates for x and y.
\item Specializations of the \code{ggplot()} generic accepting time series
objects of classes \code{ts} and \code{xts} as data argument.
}
}
\section{Acknowledgements}{
 We thank Kamil Slowikowski not only for
  contributing ideas and code examples to this package but also for adding
  new features to his package 'ggrepel' that allow new use cases for
  \code{stat_dens2d_labels()}, \code{position_nudge_center()},
  \code{position_nudge_line()} and \code{position_nudge_to()} from this
  package. This package includes code copied and/or modified from
  that in package 'ggplot2'.
}

\references{
Package 'ggplot2' documentation is available at
\url{https://ggplot2.tidyverse.org/}\cr Package 'ggplot2' source code at
\url{https://github.com/tidyverse/ggplot2}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://docs.r4photobiology.info/ggpp/}
  \item \url{https://github.com/aphalo/ggpp}
  \item Report bugs at \url{https://github.com/aphalo/ggpp/issues}
}

}
\author{
\strong{Maintainer}: Pedro J. Aphalo \email{pedro.aphalo@helsinki.fi} (\href{https://orcid.org/0000-0003-3385-972X}{ORCID})

Authors:
\itemize{
  \item Pedro J. Aphalo \email{pedro.aphalo@helsinki.fi} (\href{https://orcid.org/0000-0003-3385-972X}{ORCID})
}

Other contributors:
\itemize{
  \item Kamil Slowikowski (\href{https://orcid.org/0000-0002-2843-6370}{ORCID}) [contributor]
  \item Michał Krassowski (\href{https://orcid.org/0000-0002-9638-7785}{ORCID}) [contributor]
  \item Daniel Sabanés Bové (\href{https://orcid.org/0000-0002-0176-9239}{ORCID}) [contributor]
  \item Stella Banjo [contributor]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/example-data.R
\docType{data}
\name{ivy.df}
\alias{ivy.df}
\title{Ivy photosynthesis light response}
\format{
A \code{data.frame} object with 36 rows and 6 variables.
}
\usage{
ivy.df
}
\description{
A dataset containing photosynthesis measurements on four ivy plants.
}
\details{
For each plant a light response curve of photosynthesis was measured
using a custom-built system and software that allowed controlling the
concentrations of water vapour and carbon dioxide at the surface of the
leaves, i.e., inside the air boundary layer.
}
\examples{
colnames(ivy.df)
head(ivy.df)

}
\references{
Aphalo, P. J. (1991) Interactions in Stomatal Function. PhD
thesis, University of Edinburgh. \url{http://hdl.handle.net/1842/14758}.
}
\seealso{
Other Plant growth and morphology data:
\code{\link{birch.df}}
}
\concept{Plant growth and morphology data}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/example-data.R
\docType{data}
\name{quadrant_example.df}
\alias{quadrant_example.df}
\title{Gene expression data}
\format{
A \code{data.frame} object with 6088 rows and 6 variables
}
\usage{
quadrant_example.df
}
\description{
A dataset containing reshaped and simplified output from an
  analysis of data from RNAseq done with package edgeR. Original data from
  gene expression in the plant species \emph{Arabidopsis thaliana}.
}
\examples{
colnames(quadrant_example.df)
head(quadrant_example.df)

}
\references{
Rai, Neha; O'Hara, Andrew; Farkas, Daniel; Safronov, Omid;
Ratanasopa, Khuanpiroon; Wang, Fang; Lindfors, Anders V.; Jenkins,
Gareth I.; Lehto, Tarja; Salojärvi, Jarkko; Brosché, Mikael; Strid. Åke;
 Aphalo, Pedro José; Morales, Luis Orlando (2020) The photoreceptor UVR8
 mediates the perception of both UV-B and UV-A wavelengths up to 350 nm of
 sunlight with responsivity moderated by cryptochromes. \emph{Plant, Cell &
 Environment}, 43:1513-1527.
}
\seealso{
Other Transcriptomics data:
\code{\link{volcano_example.df}}
}
\concept{Transcriptomics data}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/position-nudge-center.R
\name{position_nudge_center}
\alias{position_nudge_center}
\alias{position_nudge_centre}
\title{Nudge labels away from a central point}
\usage{
position_nudge_center(
  x = 0,
  y = 0,
  center_x = NULL,
  center_y = NULL,
  direction = NULL,
  obey_grouping = NULL,
  kept.origin = c("original", "none")
)

position_nudge_centre(
  x = 0,
  y = 0,
  center_x = NULL,
  center_y = NULL,
  direction = NULL,
  obey_grouping = NULL,
  kept.origin = c("original", "none")
)
}
\arguments{
\item{x, y}{Amount of vertical and horizontal distance to move. A numeric
vector, that is recycled if shorter than the number of rows in \code{data}.}

\item{center_x, center_y}{The coordinates of the virtual origin out from which
nudging radiates or splits in opposite directions. A numeric vector of
length 1 or of the same length as rows there are in \code{data}, or a
function returning either of these vectors computed from the variables in
data mapped to \code{x} or \code{y}, respectively.}

\item{direction}{One of \code{"none"}, \code{"radial"}, or \code{"split"}. A
value of \code{"none"} replicates the behavior of
\code{\link[ggplot2]{position_nudge}}. Which of these three values is the
default depends on the values passed to the other parameters.}

\item{obey_grouping}{A logical flag indicating whether to obey or not
groupings of the observations. By default, grouping is obeyed when both of
the variables mapped to \emph{x} and \emph{y} are continuous numeric and
ignored otherwise.}

\item{kept.origin}{One of \code{"original"} or \code{"none"}.}
}
\value{
A \code{"Position"} object.
}
\description{
\code{position_nudge_center()} is generally useful for adjusting the position of
labels or text, both on a discrete or continuous scale. In contrast to
\code{\link[ggplot2]{position_nudge}}, \code{position_nudge_center()} returns in \code{data} both
the original coordinates and the nudged coordinates.
}
\details{
This position function is backwards compatible with \code{\link[ggplot2]{position_nudge}}
but extends it by adding support for nudging that varies across the plotting
region, either in opposite directions or radially from a virtual \emph{center
point}.

Positive values as arguments to \code{x} and \code{y} are added to
  the original position along either axis. If no arguments are passed to
  \code{center_x}, \code{center_y} or \code{direction}, the nudging is
  applied as is, as is the case if \code{direction = "none"}. If
  non-\code{NULL} arguments are passed to both \code{center_x} and
  \code{center_y}, \code{direction = "radial"} is assumed. In this case, if
  \code{x} and/or \code{y} positive nudging is applied radially outwards from
  the center, while if negative, inwards towards the center. When a
  non-\code{NULL} argument is passed only to one of \code{center_x} or
  \code{center_y}, \code{direction = "split"} is assumed. In this case when
  the initial location of the point is to the left of \code{center_x},
  \code{-x} is used instead of \code{x} for nudging, and when the initial
  location of the point is to the below of \code{center_y}, \code{-y} is used
  instead of \code{y} for nudging. If non-\code{NULL} arguments are passed to
  both \code{center_x} and \code{center_y}, and \code{direction} is passed
  \code{"split"} as argument, then the split as described above is applied to
  both to \emph{x} and \emph{y} coordinates.
}
\note{
Some situations are handled as special cases. When \code{direction =
  "split"} or \code{direction = "radial"}, observations at exactly the _center_
  are nudged using \code{x} and \code{y} unchanged. When\code{direction = "split"},
  and
  both \code{center_x} and \code{center_y} have been supplied, segments are drawn at
  eight different possible angles. When segments are exactly horizontal or
  vertical they would be shorter than when drawn at the other four angles, in
  which case \code{x} or \code{y} are adjusted to ensure these segments are of the same
  lengths as those at other angles.

  This position is most useful when labelling points forming a cloud or
  grouped along vertical or horizontal lines or "divides".
}
\examples{
df <- data.frame(
  x = c(1,3,2,5,4,2.5),
  y = c("abc","cd","d","c","bcd","a")
)

# Plain nudging, same as with ggplot2::position_nudge()

ggplot(df, aes(x, y, label = y)) +
  geom_point() +
  geom_text_s(hjust = "left", vjust = "bottom",
              position = position_nudge(x = 0.2, y = 0.2))

ggplot(df, aes(x, y, label = y)) +
  geom_point() +
  geom_text_s(add.segments = FALSE,
              position = position_nudge_center(x = 0.2, y = 0.2)
  )

# "split" nudging

ggplot(df, aes(x, y)) +
  geom_point() +
  geom_text_s(aes(label = y),
              add.segments = FALSE,
              position = position_nudge_center(x = 0.2,
                                               y = 0.2,
                                               direction = "split"))

ggplot(df, aes(x, y)) +
  geom_point() +
  geom_text_s(aes(label = y),
              position = position_nudge_center(x = 0.4,
                                               direction = "split"))

ggplot(df, aes(x, y)) +
  geom_point() +
  geom_text_s(aes(label = y),
              position = position_nudge_center(y = 0.2,
                                               direction = "split"))

ggplot(df, aes(x, y)) +
  geom_point() +
  geom_text_s(aes(label = y),
              position = position_nudge_center(x = 0.2,
                                               y = 0.3,
                                               center_y = 2,
                                               center_x = 1.5,
                                               direction = "split"))

ggplot(df, aes(x, y)) +
  geom_point() +
  geom_text_s(aes(label = y),
              position = position_nudge_center(x = 0.06,
                                               y = 0.08,
                                               center_y = 2))

ggplot(df, aes(x, y)) +
  geom_point() +
  geom_text_s(aes(label = y),
              position = position_nudge_center(x = 0.1,
                                               center_x = 2.51))

ggplot(df, aes(x, y)) +
  geom_point() +
  geom_text_s(aes(label = y),
              position = position_nudge_center(x = 0.06,
                                               y = 0.08,
                                               center_x = median,
                                               center_y = median,
                                               direction = "split"))

# "Radial" nudging

ggplot(df, aes(x, y)) +
  geom_point() +
  geom_text_s(aes(label = y),
              position = position_nudge_center(x = 0.1,
                                               y = 0.2,
                                               direction = "radial"))

ggplot(df, aes(x, y)) +
  geom_point() +
  geom_text_s(aes(label = y),
              position = position_nudge_center(x = -0.1,
                                               y = -0.1,
                                               direction = "radial"))

df <- data.frame(
  x = -10:10,
  z = (-10:10)^2,
  y = letters[1:21],
  group = rep(c("a", "b"), rep(c(11, 10)))
)

ggplot(df, aes(x, z)) +
  geom_point() +
  geom_line() +
  geom_text_s(aes(label = y),
              position = position_nudge_center(x = 0.9,
                                               y = 2.7,
                                               center_x = mean,
                                               center_y = max))

ggplot(df, aes(x, z, color = group)) +
  geom_point() +
  geom_line(color = "black", linetype = "dotted") +
  geom_text_s(aes(label = y),
              position = position_nudge_center(x = -1.2,
                                               y = -3,
                                               center_x = 0,
                                               center_y = "above_max"))

ggplot(df, aes(x, z, color = group)) +
  geom_point() +
  geom_line(color = "black", linetype = "dotted") +
  geom_text(aes(label = y),
            vjust = "inward", hjust = "inward",
            position = position_nudge_center(x = -0.9,
                                             y = -2.7,
                                             center_x = mean,
                                             center_y = max,
                                             obey_grouping = FALSE))

}
\seealso{
[ggplot2::position_nudge()], [ggrepel::position_nudge_repel()].

Other position adjustments:
\code{\link[=position_dodgenudge]{position_dodgenudge()}},
\code{\link[=position_dodgenudge_to]{position_dodgenudge_to()}},
\code{\link[=position_jitternudge]{position_jitternudge()}},
\code{\link[=position_nudge_keep]{position_nudge_keep()}},
\code{\link[=position_nudge_line]{position_nudge_line()}},
\code{\link[=position_nudge_to]{position_nudge_to()}},
\code{\link[=position_stacknudge]{position_stacknudge()}},
\code{\link[=position_stacknudge_to]{position_stacknudge_to()}}
}
\concept{position adjustments}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/position-jitter-nudge.R
\name{position_jitternudge}
\alias{position_jitternudge}
\alias{position_jitter_keep}
\title{Combined positions jitter and nudge}
\source{
\url{https://github.com/slowkow/ggrepel/issues/161}.
}
\usage{
position_jitternudge(
  width = NULL,
  height = NULL,
  seed = NA,
  x = 0,
  y = 0,
  direction = c("as.is", "alternate", "split"),
  nudge.from = c("original", "original.x", "original.y", "jittered", "jittered.y",
    "jittered.x"),
  kept.origin = c("jittered", "original", "none")
)

position_jitter_keep(
  width = NULL,
  height = NULL,
  seed = NA,
  kept.origin = "original"
)
}
\arguments{
\item{width, height}{Amount of vertical and horizontal jitter. The jitter is
added in both positive and negative directions, so the total spread is
twice the value specified here. If omitted, defaults to 40% of the
resolution of the data: this means the jitter values will occupy 80% of the
implied bins. Categorical data is aligned on the integers, so a width or
height of 0.5 will spread the data so it's not possible to see the
distinction between the categories.}

\item{seed}{A random seed to make the jitter reproducible. Useful if you need
to apply the same jitter twice, e.g., for a point and a corresponding
label. The random seed is reset after jittering. If \code{NA} (the default
value), the seed is initialised with a random value; this makes sure that
two subsequent calls start with a different seed. Use \code{NULL} to use
the current random seed and also avoid resetting (the behaviour of ggplot
2.2.1 and earlier).}

\item{x, y}{Amount of vertical and horizontal distance to move. A numeric
vector of length 1, or of the same length as rows there are in \code{data},
with nudge values in data rows order.}

\item{direction}{One of \code{"as.is"}, \code{"alternate"}, \code{"split"},
\code{"split.x"} or \code{"split.y"}. A value of \code{"none"} replicates
the behavior of \code{\link[ggplot2]{position_nudge}}. \code{"split"}
changes the sign of the nudge depending on the direction of the random
jitter applied to each individual observation, which is suitable for
nudging labels outward of the jittered data.}

\item{nudge.from}{One of \code{"original"}, \code{"jittered"},
\code{"original.y"} (or \code{"jittered.x"}), \code{"original.x"} (or
\code{"jittered.y"}). A value of \code{"original"} applies the nudge before
jittering the observations, while \code{"jittered"} applies the nudging
after jittering.}

\item{kept.origin}{One of \code{"original"}, \code{"jittered"} or
\code{"none"}.}
}
\value{
A \code{"Position"} object. The layer function within it returns a
  data frame, with the jittered + nudged values in columns \code{x} and
  \code{y} and by default the jittered values with no nudging as
  \code{x_orig} and \code{y_orig}. With \code{nudge.from = "original"} the
  original values with no jitter and no nudge applied are returned as
  \code{x_orig} and \code{y_orig}.
}
\description{
\code{position_jitternudge()} combines into one function the action of
\code{\link[ggplot2]{position_jitter}} and
\code{\link[ggplot2]{position_nudge}}. It is useful when labels to jittered
plots and when adding jitter to text labels linked to points plotted without
jitter. It can replace other position functions as it is backwards
compatible. Like all other position functions in 'ggpp' and 'ggrepel' it
preserves the initial position to allow drawing of segments or arrow linking
the original position to the displaced one.
}
\details{
Jitter with \code{position_jitternudge()} is identical to that with
  \code{\link[ggplot2]{position_jitter}} while nudging is enhanced compared
  to \code{\link[ggplot2]{position_nudge}} by taking into use cases specific
  to the combination of jitter and nudge.

  There are two posible uses for this function. First it can be used to label
  jittered points in a plot. In this case, it is mandatory to use the same
  arguments to \code{width}, \code{height} and \code{seed} when passing
  \code{position_jitter()} to \code{geom_point()} and
  \code{position_jitternudge()} to \code{geom_text()} or to
  \code{geom_label()} or their repulsive equivalents. Otherwise the arrows or
  segments will fail to connect to the labels. In other words jittering is
  computed twice. Jitter should be identical with the same arguments as
  \code{position_jitternudge()} as this last function calls the same code
  imported from package 'ggplot2'.

  The second use is to jitter labels to be connected to points that have not
  been jittered. The return of original positions instead of the jittered
  ones is achieved by passing \code{origin = "original"} to override the
  default \code{origin = "jittered"}.
}
\note{
When \code{direction = "split"} is used together with no jitter, the
  split to left and right, or up and down is done at random.
}
\examples{

jitter <- position_jitter(width = 0.2, height = 2, seed = 123)

jitter_nudge <- position_jitternudge(width = 0.2, height = 2,
                                     seed = 123, x = 0.1,
                                     direction = "split",
                                     nudge.from = "jittered")
ggplot(mpg[1:20, ],
       aes(cyl, hwy, label = drv)) +
  geom_point(position = jitter) +
  geom_text_s(position = jitter_nudge)

jitter_nudge <- position_jitternudge(width = 0.2, height = 2,
                                     seed = 123, x = 0.35,
                                     direction = "split",
                                     nudge.from = "original.x")
ggplot(mpg[1:20, ],
       aes(cyl, hwy, label = drv)) +
  geom_point(position = jitter) +
  geom_text_s(position = jitter_nudge)

jitter <- position_jitter(width = 0, height = 2, seed = 123)

jitter_nudge <- position_jitternudge(width = 0, height = 2,
                                     seed = 123, x = 0.4,
                                     direction = "split",
                                     nudge.from = "original.x")
ggplot(mpg[1:20, ],
       aes(cyl, hwy, label = drv)) +
  geom_point(position = jitter) +
  geom_text_s(position = jitter_nudge)

jitter_nudge <- position_jitternudge(width = 0, height = 2,
                                     seed = 123, x = 0.4,
                                     direction = "alternate",
                                     nudge.from = "original.x")
ggplot(mpg[1:20, ],
       aes(cyl, hwy, label = drv)) +
  geom_point(position = jitter) +
  geom_text_s(position = jitter_nudge)

# No nudge, show how points have moved with jitter

ggplot(mpg[1:20, ],
       aes(cyl, hwy, label = drv)) +
  geom_point() +
  geom_point_s(position =
               position_jitter_keep(width = 0.3, height = 2, seed = 123),
               color = "red",
               arrow = grid::arrow(length = unit(0.4, "lines")))

}
\seealso{
\code{\link[ggplot2]{position_jitter}},
\code{\link[ggplot2]{position_nudge}},
 \code{\link[ggrepel]{position_nudge_repel}}.

Other position adjustments:
\code{\link[=position_dodgenudge]{position_dodgenudge()}},
\code{\link[=position_dodgenudge_to]{position_dodgenudge_to()}},
\code{\link[=position_nudge_center]{position_nudge_center()}},
\code{\link[=position_nudge_keep]{position_nudge_keep()}},
\code{\link[=position_nudge_line]{position_nudge_line()}},
\code{\link[=position_nudge_to]{position_nudge_to()}},
\code{\link[=position_stacknudge]{position_stacknudge()}},
\code{\link[=position_stacknudge_to]{position_stacknudge_to()}}
}
\author{
Michał Krassowski, edited by Pedro J. Aphalo.
}
\concept{position adjustments}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/example-data.R
\docType{data}
\name{birch.df}
\alias{birch.df}
\alias{birch_dw.df}
\title{Birch seedlings' size}
\format{
A \code{data.frame} object with 350 rows and 8 variables.

A \code{data.frame} object with 700 rows and 5 variables.
}
\usage{
birch.df

birch_dw.df
}
\description{
A dataset containing the measurements on 350 birch seedlings.
}
\details{
The data are for seedlings grown in trays with cells or containers
of two different volumes. For each of these types of trays, all cells, 1/2 of
the cells or 1/4 of the cells contained seedlings. Root-collar diameter (mm),
height (cm), dry mass (mg) of stems and roots. Measurements done at the end
of the first growing season, after leaf fall.
}
\examples{
colnames(birch.df)
head(birch.df)

colnames(birch_dw.df)
head(birch_dw.df)

}
\references{
Aphalo, P. J. and Rikala, R. (2003) Field performance of
silver-birch planting-stock grown at different spacing and in containers of
different volume. \emph{New Forests}, 25:93-108. \doi{10.1023/A:1022618810937}.
}
\seealso{
Other Plant growth and morphology data:
\code{\link{ivy.df}}
}
\concept{Plant growth and morphology data}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-point-linked.r
\name{geom_point_s}
\alias{geom_point_s}
\title{Points linked by a segment}
\usage{
geom_point_s(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  ...,
  nudge_x = 0,
  nudge_y = 0,
  move.point = TRUE,
  arrow = grid::arrow(length = unit(1/3, "lines"), ends = "first"),
  default.colour = NULL,
  default.color = default.colour,
  colour.target = "point",
  color.target = colour.target,
  default.alpha = NA,
  alpha.target = "all",
  add.segments = TRUE,
  box.padding = 0.25,
  point.padding = 1e-06,
  segment.linewidth = 0.5,
  min.segment.length = 0,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by
\code{\link[ggplot2]{aes}}. If specified and \code{inherit.aes = TRUE} (the
default), is combined with the default mapping at the top level of the
plot. You only need to supply \code{mapping} if there isn't a mapping
defined for the plot.}

\item{data}{A data frame. If specified, overrides the default data frame
defined at the top level of the plot.}

\item{stat}{The statistical transformation to use on the data for this layer,
as a string.}

\item{position}{Position adjustment, either as a string, or the result of a
call to a position adjustment function.}

\item{...}{other arguments passed on to \code{\link[ggplot2]{layer}}. There
are three types of arguments you can use here:
\itemize{ \item Aesthetics: to set an aesthetic to a fixed value, like
\code{colour = "red"} or \code{size = 3}. \item Other arguments to the
layer, for example you override the default \code{stat} associated with the
layer. \item Other arguments passed on to the stat. }}

\item{nudge_x, nudge_y}{Horizontal and vertical adjustments to nudge the
starting position of each text label. The units for \code{nudge_x} and
\code{nudge_y} are the same as for the data units on the x-axis and y-axis.}

\item{move.point}{logical If \code{TRUE} the point is drawn at the nudged
position while if \code{FALSE} the point is drawn at the original position.}

\item{arrow}{specification for arrow heads, as created by
\code{\link[grid]{arrow}}}

\item{default.colour, default.color}{A colour definition to use for elements not targeted by
the colour aesthetic.}

\item{colour.target, color.target}{A character string, one of \code{"all"},
\code{"point"} and \code{"segment"} or \code{"none"}.}

\item{default.alpha}{numeric in [0..1] A transparency value to use for
elements not targeted by the alpha aesthetic.}

\item{alpha.target}{A character string, one of \code{"all"},
\code{"segment"}, \code{"point"}, or \code{"none"}.}

\item{add.segments}{logical Display connecting segments or arrows between
original positions and displaced ones if both are available.}

\item{box.padding, point.padding}{numeric By how much each end of the segments
should shortened in mm.}

\item{segment.linewidth}{numeric Width of the segments or arrows in mm.}

\item{min.segment.length}{numeric Segments shorter that the minimum length
are not rendered, in mm.}

\item{na.rm}{If \code{FALSE} (the default), removes missing values with a
warning.  If \code{TRUE} silently removes missing values.}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
never includes, and \code{TRUE} always includes.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics, rather
than combining with them. This is most useful for helper functions that
define both data and aesthetics and shouldn't inherit behaviour from the
default plot specification, e.g. \code{\link[ggplot2]{borders}}.}
}
\value{
A plot layer instance.
}
\description{
The geometry \code{"geom_point_s"} provides a super set of the capabilities of
geom \code{\link[ggplot2]{geom_point}} from package 'ggplot2' by allowing
plotting of arrows or segments joining the original position of displaced observations
to their current position rendered as points or graphic symbols. The most
common use is to demonstrate the action of different position functions. It
can be also used to highlight observations.
}
\details{
The plotting of segments is similar in idea to that implemented in
  \code{\link[ggrepel]{geom_text_repel}} and relies on position functions
  that rename instead of only replacing the original \code{x} and \code{y}
  coordinates from the \code{data} object.

  By default this geom uses \code{\link{position_nudge_center}} which is backwards
  compatible with \code{\link[ggplot2]{position_nudge}} but provides additional control
  on the direction of the nudging.
}
\note{
The insets are stored nested within the main ggplot object and
  contain their own copy of the data, and are rendered as grid grobs as normal
  ggplots at the time the main ggplot is rendered. They can have different
  themes.

  Use \code{\link{annotate}} as redefined in 'ggpp' when adding insets
  as annotations (automatically available unless 'ggpp' is not attached).
  \code{\link[ggplot2]{annotate}} cannot be used with the \code{npcx} and
  \code{npcy} pseudo-aesthetics.
}
\section{Position functions}{
 Many layer functions from package 'ggpp' are
  designed to work seamlessly with position functions that keep, rather than
  discard, the original \code{x} and \code{y} positions in \code{data} when
  computing a new displaced position. See \code{\link{position_nudge_keep}},
  \code{\link{position_dodge_keep}}, \code{\link{position_jitter_keep}},
  \code{\link{position_nudge_center}}, \code{\link{position_nudge_line}},
  \code{\link{position_nudge_to}}, \code{\link{position_dodgenudge}},
  \code{\link{position_jitternudge}}, and \code{\link{position_stacknudge}}
  for examples and details of their use.
}

\examples{

# Same output as with geom_point()
ggplot(mpg[1:20, ],
       aes(cyl, hwy)) +
  geom_point_s(colour = "blue")

# with segment drawn after nudging
ggplot(mpg[1:20, ],
       aes(cyl, hwy, label = drv)) +
  geom_point_s(position = position_nudge_keep(x = 0.2),
               colour = "red") +
  geom_point_s(colour = "blue") +
  expand_limits(x = c(3.5, 8.5))

ggplot(mpg[1:20, ],
       aes(cyl, hwy, label = drv)) +
  geom_point_s(position = position_nudge_keep(x = 0.2),
               colour = "blue",
               move.point = FALSE) +
  expand_limits(x = c(3.5, 8.5))

# with segment drawn after nudging
ggplot(mpg[1:20, ],
       aes(cyl, hwy, label = drv)) +
  geom_point_s(position = position_nudge_keep(x = 0.2),
               colour = "red",
               colour.target = "all") +
  geom_point_s(colour = "blue")

ggplot(mpg[1:20, ],
       aes(cyl, hwy, label = drv)) +
  geom_point_s(position = position_nudge_keep(x = 0.2),
               colour = "red",
               colour.target = "segment") +
  geom_point_s(colour = "blue")

ggplot(mpg[1:20, ],
       aes(cyl, hwy, label = drv)) +
  geom_point_s(position = position_nudge_keep(x = 0.2),
               colour = "red",
               colour.target = "point") +
  geom_point_s(colour = "blue")

ggplot(mpg[1:50, ],
       aes(cyl, hwy, label = drv)) +
  geom_point_s(position = position_jitternudge(width = 0.66, height = 2,
                                               seed = 456,
                                               nudge.from = "jittered",
                                               kept.origin = "original"),
               colour = "red",
               alpha = 0.3, alpha.target = "segment",
               arrow = grid::arrow(length = grid::unit(0.4, "lines"),
                                   ends = "first")) +
  geom_point_s(colour = "blue")

}
\seealso{
\code{\link[ggplot2]{geom_point}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-label-linked.r, R/geom-text-linked.r
\name{geom_label_s}
\alias{geom_label_s}
\alias{geom_text_s}
\title{Linked Text}
\usage{
geom_label_s(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  ...,
  parse = FALSE,
  nudge_x = 0,
  nudge_y = 0,
  default.colour = NULL,
  default.color = default.colour,
  colour.target = c("text", "box"),
  color.target = colour.target,
  default.alpha = NA,
  alpha.target = "all",
  label.padding = grid::unit(0.25, "lines"),
  label.r = grid::unit(0.15, "lines"),
  segment.linewidth = 0.5,
  add.segments = TRUE,
  box.padding = 1e-06,
  point.padding = 1e-06,
  min.segment.length = 0,
  arrow = NULL,
  size.unit = "mm",
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)

geom_text_s(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  ...,
  parse = FALSE,
  nudge_x = 0,
  nudge_y = 0,
  default.colour = NULL,
  default.color = default.colour,
  colour.target = "text",
  color.target = colour.target,
  default.alpha = NA,
  alpha.target = "all",
  add.segments = TRUE,
  box.padding = 0.25,
  point.padding = 1e-06,
  segment.linewidth = 0.5,
  min.segment.length = 0,
  arrow = NULL,
  check_overlap = FALSE,
  size.unit = "mm",
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by
\code{\link[ggplot2]{aes}}. If specified and with \code{inherit.aes = TRUE}
(the default), it is combined with the default mapping at the top level of
the plot. You only need to supply \code{mapping} if there isn't a mapping
defined for the plot.}

\item{data}{A data frame. If specified, overrides the default data frame
defined at the top level of the plot.}

\item{stat}{The statistical transformation to use on the data for this layer,
as a string.}

\item{position}{Position adjustment, either as a string, or the result of a
call to a position adjustment function.}

\item{...}{other arguments passed on to \code{\link[ggplot2]{layer}}. There
  are three types of arguments you can use here:

  \itemize{ \item Aesthetics: to set an aesthetic to a fixed value, like
  \code{colour = "red"} or \code{size = 3}. \item Other arguments to the
  layer, for example you override the default \code{stat} associated with the
  layer. \item Other arguments passed on to the stat. }}

\item{parse}{If \code{TRUE}, the labels will be parsed into expressions and
displayed as described in \code{?plotmath}.}

\item{nudge_x, nudge_y}{Horizontal and vertical adjustments to nudge the
starting position of each text label. The units for \code{nudge_x} and
\code{nudge_y} are the same as for the data units on the x-axis and y-axis.}

\item{default.colour, default.color}{A colour definition to use for elements
not targeted by the colour aesthetic.}

\item{colour.target, color.target}{A vector of character strings; \code{"all"},
\code{"text"}, \code{"segment"}, \code{"box"}, \code{"box.line"}, and
\code{"box.fill"} or \code{"none"}.}

\item{default.alpha}{numeric in [0..1] A transparency value to use for
elements not targeted by the alpha aesthetic.}

\item{alpha.target}{A vector of character strings; \code{"all"},
\code{"text"}, \code{"segment"}, \code{"box"}, \code{"box.line"}, and
\code{"box.fill"} or \code{"none"}.}

\item{label.padding}{Amount of padding around label. Defaults to 0.25 lines.}

\item{label.r}{Radius of rounded corners. Defaults to 0.15 lines.}

\item{segment.linewidth}{numeric Width of the segments or arrows in mm.}

\item{add.segments}{logical Display connecting segments or arrows between
original positions and displaced ones if both are available.}

\item{box.padding, point.padding}{numeric By how much each end of the segments
should shortened in mm.}

\item{min.segment.length}{numeric Segments shorter that the minimum length
are not rendered, in mm.}

\item{arrow}{specification for arrow heads, as created by
\code{\link[grid]{arrow}}}

\item{size.unit}{How the `size` aesthetic is interpreted: as millimetres
(`"mm"`, default), points (`"pt"`), centimetres (`"cm"`), inches (`"in"`),
or picas (`"pc"`).}

\item{na.rm}{If \code{FALSE} (the default), removes missing values with a
warning.  If \code{TRUE} silently removes missing values.}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA}, the default, includes a legend if any aesthetics are mapped.
\code{FALSE} never includes it, and \code{TRUE} always includes it.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics, rather
than combining them. This is most useful for helper functions that define
both data and aesthetics and shouldn't inherit behaviour from the default
plot specification, e.g., \code{\link[ggplot2]{borders}}.}

\item{check_overlap}{If \code{TRUE}, text that overlaps previous text in the
same layer will not be plotted. \code{check_overlap} takes place at draw
time and in the order of the data, thus its action depends of the size at
which the plot is drawn.}
}
\value{
A plot layer instance.
}
\description{
Linked text geometries are most useful for adding data labels to
  plots. `geom_text_s()` and `geom_label_s()` add text to the plot and for
  nudged positions link the original location to the nudged text with a
  segment or arrow.
}
\details{
Geometries \code{geom_text_s()} and \code{geom_label_s()} have an
  interface similar to that of \code{\link[ggplot2]{geom_text}} and
  \code{\link[ggplot2]{geom_label}}, but support additional features.
  Similarly to \code{geom_text_repel()} and \code{geom_label_repel()} when
  used together with position functions defined in package 'ggpp' they draw a
  segment linking the label at a displaced position to the original position,
  usually a point corresponding to an observation to which the label refers.
  Another difference is that they allow control of to which graphical
  elements the mappings to colour and alpha aesthetics are applied.
  Differently to \code{geom_label()}, \code{geom_label_s()} obeys aesthetic
  mappings to \code{linewidth} and \code{linetype} applied to the line at the
  edge of the label box. These features are reflected in the plot key, except
  for the segment, assumed not to be used to display information
  only in coordination with other graphic elements.

  In \code{geom_label_s()} the default \code{fill} is similar to
  \code{"white"} but with its \code{alpha} component set to 0.75. This
  differs from \code{"white"} used in \code{geom_label()}: the default fill
  is semitransparent with the intention that accidental occlusion of
  observations is obvious irrespective of the order in which layers are added
  to the plot.

  Layer functions \code{geom_text_s()} and \code{geom_label_s()} use by
  default \code{\link{position_nudge_keep}} which is backwards compatible
  with \code{\link[ggplot2]{position_nudge}}. In contrast to
  \code{\link[ggplot2]{position_nudge}}, \code{\link{position_nudge_keep}}
  and all other position functions defined in packages 'ggpp' and 'ggrepel'
  keep the original coordinates, thus allowing the plotting of connecting
  segments and arrows.

  Differently to \code{geom_text_repel()} and \code{geom_label_repel()},
  \code{geom_text_s()} and \code{geom_label_s()} do not make use of
  additional aesthetics for the segments or boxes, but instead allow the
  choice of which elements are targeted by the aesthetics and which are
  rendered in a default colour. In the grammar of graphics using the same
  aesthetic with multiple meanings is not allowed, thus, the approach used in
  the geometry layer functions from package 'ggpp' attempts to enforce this.
}
\section{Plot boundaries and clipping}{
 Note that when you change the scale
  limits for \emph{x} and/or \emph{y} of a plot, text labels stay the same
  size, as determined by the \code{size} aesthetic, given in millimetres. The
  actual size as seen in the plotted output is decided during the rendering
  of the plot to a graphics device. Limits are expanded only to include the
  anchor point of the labels because the "width" and "height" of a text
  element are 0 (as seen by ggplot2). Text labels do have height and width,
  but in grid units, not data units.
}

\section{Alignment}{
 You can modify text alignment with the \code{vjust} and
  \code{hjust} aesthetics. These can either be a number between 0
  (right/bottom) and 1 (top/left) or a character (\code{"left"},
  \code{"middle"}, \code{"right"}, \code{"bottom"}, \code{"center"},
  \code{"top"}). In addition, you can use special alignments for
  justification including \code{"position"}, \code{"inward"} and
  \code{"outward"}. Inward always aligns text towards the center of the
  plotting area, and outward aligns it away from the center of the plotting
  area. If tagged with \code{_mean} or \code{_median} (e.g.,
  \code{"outward_mean"}) the mean or median of the data in the panel along
  the corresponding axis is used as center. If the characters following the
  underscore represent a number (e.g., \code{"outward_10.5"}) the reference
  point will be this value in data units. Position justification is computed
  based on the direction of the displacement of the position of the label so
  that each individual text or label is justified outwards from its original
  position. The default justification is \code{"position"}.

  If no position displacement is applied, or a position function defined in
  'ggplot2' is used, these geometries behave similarly to the corresponding
  ones from package 'ggplot2' with a default justification of \code{0.5} and
  no segment drawn.
}

\section{Differences from earlier versions}{

  The user interface is for the most part stable starting from 'ggpp' (==
  0.5.7). In 'ggpp' (== 0.5.0) support for aesthetics related to segments was
  removed, and replaced by parameters and a new mechanism for targeting the
  usual \code{colour} and \code{alpha} aesthetics to text, border, and
  segment.
}

\section{Aesthetics}{
 Layer functions \code{geom_text_s()} and
  \code{geom_label_s()} require aesthetics \code{x}, \code{y} and
  \code{label} and support aesthetics: \code{alpha}, \code{colour},
  \code{group}, \code{size} (of text), \code{family}, \code{fontface},
  \code{lineheight}, \code{hjust} and \code{vjust}. In addition,
  \code{geom_text_s} supports \code{angle} and \code{geom_label_s} supports
  \code{fill}, \code{linewidth} and \code{linetype}. See
  \code{\link[ggplot2]{aes_colour_fill_alpha}},
  \code{\link[ggplot2]{aes_linetype_size_shape}},
  \code{\link[ggplot2]{aes_position}}, and
  \code{\link[ggplot2]{aes_group_order}}.

  In 'ggplot2' \code{linewidth} when applied to the border of the box drawn
  by \code{geom_label()} is given in points rather than in mm because of a
  historical error in the code. In other geometries such as
  \code{geom_segment()} \code{linewidth} is given in mm. As in
  \code{geom_label_s()} it is important to remain consistent among
  different \code{linewidth} specifications, mm are used both for the box
  border and linking segment. To imitate the behaviour of \code{geom_label()}
  a correction factor of 0.75 (more exactly 1 pt = 0.7528 mm) can be used for
  the line width of the border of the box.
}

\section{Position functions}{
 Many layer functions from package 'ggpp' are
  designed to work seamlessly with position functions that keep, rather than
  discard, the original \code{x} and \code{y} positions in \code{data} when
  computing a new displaced position. See \code{\link{position_nudge_keep}},
  \code{\link{position_dodge_keep}}, \code{\link{position_jitter_keep}},
  \code{\link{position_nudge_center}}, \code{\link{position_nudge_line}},
  \code{\link{position_nudge_to}}, \code{\link{position_dodgenudge}},
  \code{\link{position_jitternudge}}, and \code{\link{position_stacknudge}}
  for examples and details of their use.
}

\examples{

my.cars <- mtcars[c(TRUE, FALSE, FALSE, FALSE), ]
my.cars$name <- rownames(my.cars)

# no nudging
ggplot(my.cars, aes(wt, mpg, label = name)) +
  geom_text_s() +
  expand_limits(x = c(2, 6))

# base plot
p <- ggplot(my.cars, aes(wt, mpg, label = name)) +
       geom_point()

# Using nudging
p +
  geom_text_s(nudge_x = 0.12) +
  expand_limits(x = 6.2)
p +
  geom_text_s(nudge_x = -0.12) +
  expand_limits(x = 1.5)
p +
  geom_text_s(nudge_x = 0.12,
              arrow = arrow(length = grid::unit(1.5, "mm")),
              point.padding = 0.4) +
  expand_limits(x = 6.2)
p +
  geom_text_s(nudge_y = 0.1, nudge_x = 0.07) +
  expand_limits(x = 6.2)
p +
  geom_text_s(nudge_y = 1, angle = 90) +
  expand_limits(y = 30)
p +
  geom_text_s(angle = 90, nudge_y = 1,
              arrow = arrow(length = grid::unit(1.5, "mm")),
              colour.target = "segment", colour = "red") +
  expand_limits(y = 30)
p +
  geom_text_s(aes(colour = factor(cyl)),
              angle = 90, nudge_y = 1,
              arrow = arrow(length = grid::unit(1.5, "mm")),
              alpha.target = "segment", alpha = 0.3) +
  expand_limits(y = 30)

p +
  geom_label_s(nudge_x = 0.12) +
  expand_limits(x = 6.2)
p +
  geom_label_s(nudge_x = 0.12, linetype = "dotted", linewidth = 0.3) +
  expand_limits(x = 6.2)
p +
  geom_label_s(aes(colour = factor(cyl)),
               nudge_x = 0.12,
               colour.target = "box",
               linewidth = 0.5,
               label.r = unit(0, "lines")) +
  expand_limits(x = 6.2)
p +
  geom_label_s(nudge_x = 0.12, linewidth = 0) +
  expand_limits(x = 6.2)

# No segments
p +
  geom_label_s(nudge_x = 0.05, segment.linewidth = 0) +
  expand_limits(x = 6.2)

# Nudging away from arbitrary point
p +
  geom_label_s(hjust = "outward_1", nudge_x = 0.12) +
  expand_limits(x = 6.2)
p +
  geom_label_s(hjust = "inward_3", nudge_y = 0.4)

p +
  geom_label_s(nudge_y = 1, angle = 90) +
  expand_limits(y = 30)

# Add aesthetic mappings and adjust arrows
p +
  geom_text_s(aes(colour = factor(cyl)),
              angle = 90,
              nudge_y = 1,
              arrow = arrow(angle = 20,
                            length = grid::unit(1.5, "mm"),
                            ends = "first",
                            type = "closed")) +
  scale_colour_discrete(l = 40) + # luminance, make colours darker
  expand_limits(y = 27)

p +
  geom_text_s(aes(colour = factor(cyl)),
              angle = 90,
              nudge_y = 1,
              arrow = arrow(angle = 20,
                            length = grid::unit(1.5, "mm"),
                            ends = "first",
                            type = "closed")) +
  scale_colour_discrete(l = 40) + # luminance, make colours darker
  expand_limits(y = 27)

p +
  geom_label_s(aes(colour = factor(cyl)),
              colour.target = c("box", "text"),
              nudge_x = 0.3,
              arrow = arrow(angle = 20,
                            length = grid::unit(1/3, "lines"))) +
  scale_colour_discrete(l = 40) + # luminance, make colours darker
  expand_limits(x = 7)

p +
  geom_label_s(aes(colour = factor(cyl)),
              nudge_x = 0.3,
              colour.target = c("box", "segment"),
              linewidth = 0.5,
              arrow = arrow(angle = 20,
                            length = grid::unit(1/3, "lines"))) +
  scale_colour_discrete(l = 40) + # luminance, make colours darker
  expand_limits(x = 7)

p +
  geom_label_s(aes(colour = factor(cyl), fill = factor(cyl)),
              nudge_x = 0.3,
              alpha.target = "box",
              alpha = 0.1,
              linewidth = 0.5,
              arrow = arrow(angle = 20,
                            length = grid::unit(1/3, "lines"))) +
  scale_colour_discrete(l = 40) + # luminance, make colours darker
  expand_limits(x = 7)#' # Scale height of text, rather than sqrt(height)

p +
  geom_text_s(aes(size = wt), nudge_x = -0.1) +
  scale_radius(range = c(3,6)) + # override scale_area()
    expand_limits(x = c(1.8, 5.5))

}
\seealso{
\code{\link[ggplot2]{geom_text}}, \code{\link[ggplot2]{geom_label}}
  and other documentation of package 'ggplot2'.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annotate.r
\name{annotate}
\alias{annotate}
\title{Annotations supporting NPC}
\usage{
annotate(
  geom,
  x = NULL,
  y = NULL,
  xmin = NULL,
  xmax = NULL,
  ymin = NULL,
  ymax = NULL,
  xend = NULL,
  yend = NULL,
  npcx = NULL,
  npcy = NULL,
  label = NULL,
  ...,
  na.rm = FALSE
)
}
\arguments{
\item{geom}{character Name of geom to use for annotation.}

\item{x, y, xmin, ymin, xmax, ymax, xend, yend, npcx, npcy}{numeric Positioning
aesthetics - you must specify at least one of these.}

\item{label}{character, data.frame, ggplot or grob.}

\item{...}{Other named arguments passed on to \code{layer()}. These are often
aesthetics, used to set an aesthetic to a fixed value, like color = "red"
or size = 3. They may also be parameters to the paired geom/stat.}

\item{na.rm}{logical If \code{FALSE}, the default, missing values are removed
with a warning. If TRUE, missing values are silently removed.}
}
\value{
A plot layer instance.
}
\description{
A revised version of \code{annotate()} from package 'ggplot2' adding support
for \code{npcx} and \code{npcy} position aesthetics, allowing use of the
geometries defined in the current package such as \code{geom_text_npc()}. It
also has a parameter \code{label} that directly accepts data frames, ggplots
and grobs as arguments in addition to objects of atomic classes like
character. When package 'ggpmisc' is loaded this definition of
\code{annotate()} overrides that in package 'ggplot2'.
}
\details{
Note that all position aesthetics are scaled (i.e., they will
  expand the limits of the plot so they are visible), but all other
  aesthetics are set. This means that layers created with this function will
  never affect the legend.
}
\note{
To use the original definition of \code{annotate()} after loading
  package 'ggpmisc', use \code{ggplot2::annotate()}.
}
\examples{

p <- ggplot(mtcars, aes(x = wt, y = mpg)) + geom_point()

# Works as ggplot2::annotate()
p + annotate("text", x = 5, y = 32, label = "Some text")
p + annotate("label", x = c(2, 5), y = c(15, 32),
             label = c("A", "B"))
p + annotate("table", x = 5, y = 30,
             label = data.frame(A = 1:2, B = letters[1:2]))
p + annotate("plot", x = 5.5, y = 34,
             label = p + theme_bw(9))
p + annotate("rect", xmin = 3, xmax = 4.2, ymin = 12, ymax = 21, alpha = .2)
p + annotate("segment", x = 2.5, xend = 4, y = 15, yend = 25, colour = "blue")
p + annotate("pointrange", x = 3.5, y = 20, ymin = 12, ymax = 28,
  colour = "red", size = 1.5)

# But ggpmisc::annotate() also works with npcx and npcy pseudo-aesthetics
p + annotate("label_npc", npcx = c(0.1, 0.9), npcy = c(0.1, 0.9),
             label = c("A", "B"))
p + annotate("label_npc", npcx = 0.9, npcy = c(0.1, 0.9),
             label = c("A", "B"))

p + annotate("text_npc", npcx = 0.9, npcy = 0.9, label = "Some text")
p + annotate("text_npc", npcx = "right", npcy = "top", label = "Some text")

p + annotate("table_npc", npcx = 0.9, npcy = 0.9,
             label = data.frame(A = 1:2, B = letters[1:2]))

p + annotate("plot_npc", npcx = 1, npcy = 1,
             label = p + theme_bw(9))
p + annotate("plot_npc", npcx = c(0, 1), npcy = c(0, 1),
             label = list(p + theme_bw(9), p + theme_grey(9)),
             vp.width = 0.3, vp.height = 0.4)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-table.R
\name{geom_table}
\alias{geom_table}
\alias{geom_table_npc}
\title{Inset tables}
\usage{
geom_table(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  ...,
  nudge_x = 0,
  nudge_y = 0,
  default.colour = NA,
  default.color = default.colour,
  colour.target = "table.text",
  color.target = colour.target,
  default.alpha = 1,
  alpha.target = "all",
  fontsize.scaling = 0.825,
  add.segments = TRUE,
  box.padding = 0.25,
  point.padding = 1e-06,
  segment.linewidth = 0.5,
  min.segment.length = 0,
  arrow = NULL,
  table.theme = NULL,
  table.rownames = FALSE,
  table.colnames = TRUE,
  table.hjust = 0.5,
  parse = FALSE,
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = FALSE
)

geom_table_npc(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  ...,
  default.colour = NA,
  default.color = default.colour,
  colour.target = "table.text",
  color.target = colour.target,
  default.alpha = 1,
  alpha.target = "all",
  fontsize.scaling = 0.825,
  table.theme = NULL,
  table.rownames = FALSE,
  table.colnames = TRUE,
  table.hjust = 0.5,
  parse = FALSE,
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = FALSE
)
}
\arguments{
\item{mapping}{The aesthetic mapping, usually constructed with
\code{\link[ggplot2]{aes}}. Only needs to be set at the layer level if you
are overriding the plot defaults.}

\item{data}{A layer specific dataset - only needed if you want to override
the plot defaults.}

\item{stat}{The statistical transformation to use on the data for this layer,
as a string.}

\item{position}{Position adjustment, either as a string, or the result of a
call to a position adjustment function.}

\item{...}{other arguments passed on to \code{\link[ggplot2]{layer}}. This
can include aesthetics whose values you want to set, not map. See
\code{\link[ggplot2]{layer}} for more details.}

\item{nudge_x, nudge_y}{Horizontal and vertical adjustments to nudge the
starting position of each "label". The units for \code{nudge_x} and
\code{nudge_y} are the same as for the data units on the x-axis and y-axis.}

\item{default.colour, default.color}{A colour definition to use for elements
not targeted by the colour aesthetic. If \code{NA} the colours in the table
theme are not modified.}

\item{colour.target, color.target}{A vector of character strings with one or
more of \code{"all"}, \code{"table"}, \code{"table.text"},
\code{"table.rules"}, \code{"segment"} or \code{"none"}.}

\item{default.alpha}{numeric in [0..1] A transparency value to use for
elements not targeted by the alpha aesthetic. If \code{NA} the alpha
channel of the colour definitions is not modified.}

\item{alpha.target}{A vector of character strings with one or
more of \code{"all"}, \code{"table"}, \code{"table.text"},
\code{"table.rules"} and \code{"table.canvas"}, \code{"segment"} or
\code{"none"}.}

\item{fontsize.scaling}{A scaling factor to apply to the \emph{size}
aesthetic retrieved from the theme or mapped, applied to table text.}

\item{add.segments}{logical Display connecting segments or arrows between
original positions and displaced ones if both are available.}

\item{box.padding, point.padding}{numeric By how much each end of the segments
should shortened in mm.}

\item{segment.linewidth}{numeric Width of the segments or arrows in mm.}

\item{min.segment.length}{numeric Segments shorter that the minimum length
are not rendered, in mm.}

\item{arrow}{specification for arrow heads, as created by
\code{\link[grid]{arrow}}}

\item{table.theme}{NULL, list or function A gridExtra ttheme defintion, or
a constructor for a ttheme or NULL for default. If \code{nULL} the theme
is retrieved from R option \code{} at the time the plot is rendered.}

\item{table.rownames, table.colnames}{logical flag to enable or disable
printing of row names and column names.}

\item{table.hjust}{numeric Horizontal justification for the core and column
headings of the table.}

\item{parse}{If TRUE, the labels will be parsed into expressions and
displayed as described in \code{?plotmath}.}

\item{na.rm}{If \code{FALSE} (the default), removes missing values with a
warning.  If \code{TRUE} silently removes missing values.}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
never includes, and \code{TRUE} always includes.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics, rather
than combining with them. This is most useful for helper functions that
define both data and aesthetics and shouldn't inherit behaviour from the
default plot specification, e.g. \code{\link[ggplot2]{borders}}.}
}
\value{
A plot layer instance.
}
\description{
\code{geom_table} and \code{geom_table_npc} add data frames as table insets
to the base ggplot, using syntax similar to that of
\code{\link[ggplot2]{geom_text}} and \code{\link{geom_text_s}}. In most
respects they behave as any other ggplot geometry: they add a layer
containing one or more grobs and grouping and faceting works as usual. The
most common use of \code{geom_table} is to add data labels that are whole
tables rather than text. \code{\link{geom_table_npc}} is used to add tables
as annotations to plots, but contrary to layer function \code{annotate},
\code{\link{geom_table_npc}} is data driven and respects grouping and facets,
thus plot insets can differ among panels.
}
\details{
\code{geom_table} and \code{geom_table_npc} expect a list of data frames
  (\code{"data.frame"} class or derived) to be mapped to the \code{label}
  aesthetic. These geoms work with tibbles or data frames as \code{data} as
  they both support \code{list} objects as member variables.

  A table is built with function \code{gridExtra::gtable} for each
  data frame in the list, and formatted according to a \code{ttheme} (table
  theme) list object or \code{ttheme} constructor function passed as argument
  to parameter \code{table.theme}. If the value passed as argument to
  \code{table.theme} is \code{NULL} the table theme used is that set as
  default through R option \code{ggpmisc.ttheme.default} at the time the plot
  is rendered or the \code{\link{ttheme_gtdefault}} constructor function if
  not set.

  If the argument passed to \code{table.theme} or set through R option
  \code{ggpmisc.ttheme.default} is a constructor function (passing its name
  without parenthesis), the values mapped to \code{size}, \code{colour},
  \code{fill}, \code{alpha}, and \code{family} aesthetics will the passed to
  this theme constructor for each individual table. In contrast, if a ready
  constructed \code{ttheme} stored as a list object is passed as argument
  (e.g., by calling the constructor, using constructor name followed by
  parenthesis), it is used as is, with mappings to aesthetics \code{colour},
  \code{fill}, \code{alpha}, and \code{family} ignored if present. By default
  the constructor \code{ttheme_gtdefault} is used and \code{colour} and
  \code{fill}, are mapped to \code{NA}. Mapping these aesthetics to \code{NA}
  triggers the use the values set in the \code{ttheme}. As the table is built
  with function \code{gridExtra::gtable()}, for details, please, consult
  \code{\link[gridExtra]{tableGrob}} and \code{\link{ttheme_gtdefault}}.

  The character strings in the data frame can be parsed into R expressions so
  the inset tables can include maths. With \code{parse = TRUE} parsing is
  attempted on each table cell, but failure triggers fall-back to rendering
  without parsing, on a cell by cell basis. Thus, a table can contain a
  mixture cells and/or headings that require parsing or not (see the
  documentation in \link[gridExtra]{gridExtra-package} for details).

  The \code{x} and \code{y} aesthetics determine the position of the whole
  inset table, similarly to that of a text label, justification is
  interpreted as indicating the position of the inset table with respect to
  its \emph{horizontal} and \emph{vertical} axes (rows and columns in the
  data frame), and \code{angle} is used to rotate the inset table as a whole.
  The default for the \code{colour} and \code{fill} aesthetics is to retrieve
  them from the table theme, while \code{family} and \code{size} are
  retrieved from the \code{geom} element of the ggplot theme.

  Of these two geoms only \code{\link{geom_table}} supports the plotting of
  connecting segments or arrows when its position has been modified by a
  \code{position} function. This is because \code{\link{geom_table_npc}} uses
  a coordinate system that is unrelated to data units, scales or data in
  other plot layers. In the case of \code{geom_table_npc}, \code{npcx} and
  \code{npcy} pseudo-aesthetics determine the position of the inset table. By
  default, \code{hjust} and \code{vjust} are set to \code{"inward"} to avoid
  clipping at the edge of the plot canvas.
}
\note{
Complex tables with annotations or different colouring of rows or cells
  can be constructed with functions in package 'gridExtra' or in any other
  way as long as they can be saved as grid graphical objects and then added
  to a ggplot as a new layer with \code{\link{geom_grob}}.
}
\section{Alignment}{
 You can modify text alignment with the \code{vjust} and
  \code{hjust} aesthetics. These can either be a number between 0
  (right/bottom) and 1 (top/left) or a character (\code{"left"},
  \code{"middle"}, \code{"right"}, \code{"bottom"}, \code{"center"},
  \code{"top"}). In addition, you can use special alignments for
  justification including \code{"position"}, \code{"inward"} and
  \code{"outward"}. Inward always aligns text towards the center of the
  plotting area, and outward aligns it away from the center of the plotting
  area. If tagged with \code{_mean} or \code{_median} (e.g.,
  \code{"outward_mean"}) the mean or median of the data in the panel along
  the corresponding axis is used as center. If the characters following the
  underscore represent a number (e.g., \code{"outward_10.5"}) the reference
  point will be this value in data units. Position justification is computed
  based on the direction of the displacement of the position of the label so
  that each individual text or label is justified outwards from its original
  position. The default justification is \code{"position"}.

  If no position displacement is applied, or a position function defined in
  'ggplot2' is used, these geometries behave similarly to the corresponding
  ones from package 'ggplot2' with a default justification of \code{0.5} and
  no segment drawn.
}

\section{Position functions}{
 Many layer functions from package 'ggpp' are
  designed to work seamlessly with position functions that keep, rather than
  discard, the original \code{x} and \code{y} positions in \code{data} when
  computing a new displaced position. See \code{\link{position_nudge_keep}},
  \code{\link{position_dodge_keep}}, \code{\link{position_jitter_keep}},
  \code{\link{position_nudge_center}}, \code{\link{position_nudge_line}},
  \code{\link{position_nudge_to}}, \code{\link{position_dodgenudge}},
  \code{\link{position_jitternudge}}, and \code{\link{position_stacknudge}}
  for examples and details of their use.
}

\section{Plot boundaries and clipping}{
 The "width" and "height" of an inset
  as for a text element are 0, so stacking and dodging inset plots will not
  work by default, and axis limits are not automatically expanded to include
  all inset plots. Obviously, insets do have height and width, but they are
  physical units, not data units. The amount of space they occupy on the main
  plot is not constant in data units of the base plot: when you modify scale
  limits, inset plots stay the same size relative to the physical size of the
  base plot.
}

\examples{
library(dplyr)
library(tibble)

theme_set(theme_bw())

mtcars |>
  group_by(cyl) |>
  summarize(wt = mean(wt), mpg = mean(mpg)) |>
  ungroup() |>
  mutate(wt = sprintf("\%.2f", wt),
         mpg = sprintf("\%.1f", mpg)) -> tb

df <- data.frame(x = 5.45, y = 34, tb = I(list(tb)))

# using defaults
ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
  geom_point() +
  geom_table(data = df,
             aes(x = x, y = y, label = tb))

# settings aesthetics to constants
ggplot(mtcars,
       aes(wt, mpg, colour = factor(cyl))) +
  geom_point() +
  geom_table(data = df,
             aes(x = x, y = y, label = tb),
             color = "red",
             fill = "#FFCCCC",
             family = "serif", size = 5,
             angle = 90, vjust = 0)

# passing a theme constructor as argument
ggplot(mtcars,
       aes(wt, mpg, colour = factor(cyl))) +
  geom_point() +
  geom_table(data = df,
             aes(x = x, y = y, label = tb),
             table.theme = ttheme_gtstripes) +
  theme_classic()

# transparency
ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
  geom_point() +
  geom_table(data = df,
             aes(x = x, y = y, label = tb),
             alpha = 0.5) +
  theme_bw()

ggplot(mtcars, aes(wt, mpg, colour = factor(cyl))) +
  geom_point() +
  geom_table(data = df,
             aes(x = x, y = y, label = tb),
             alpha = 0.5, alpha.target = "table.canvas")

df2 <- tibble(x = 5.45,
              y = c(34, 29, 24),
              x1 = c(2.29, 3.12, 4.00),
              y1 = c(26.6, 19.7, 15.1),
              cyl = c(4, 6, 8),
              tb = list(tb[1, 1:3], tb[2, 1:3], tb[3, 1:3]))

# mapped aesthetics
ggplot(mtcars,
       aes(wt, mpg, color = factor(cyl))) +
  geom_point() +
  geom_table(data = df2,
             inherit.aes = TRUE,
             mapping = aes(x = x, y = y, label = tb))

ggplot(mtcars,
       aes(wt, mpg, color = factor(cyl))) +
  geom_point() +
  geom_table(data = df2,
             inherit.aes = TRUE,
             colour.target = "table.rules",
             mapping = aes(x = x, y = y, label = tb))

# nudging and segments
ggplot(mtcars,
       aes(wt, mpg, color = factor(cyl))) +
  geom_point(show.legend = FALSE) +
  geom_table(data = df2,
             inherit.aes = TRUE,
             mapping = aes(x = x1, y = y1, label = tb),
             nudge_x = 0.7, nudge_y = 3,
             vjust = 0.5, hjust = 0.5,
             arrow = arrow(length = unit(0.5, "lines"))) +
  theme_classic()

ggplot(mtcars,
       aes(wt, mpg, color = factor(cyl))) +
  geom_point(show.legend = FALSE) +
  geom_table(data = df2,
             inherit.aes = TRUE,
             mapping = aes(x = x1, y = y1, label = tb),
             nudge_x = 0.7, nudge_y = 3,
             vjust = 0.5, hjust = 0.5,
             arrow = arrow(length = unit(0.5, "lines")),
             colour.target = c("table.rules", "segment")) +
  theme_classic()

# Using native plot coordinates instead of data coordinates
dfnpc <- tibble(x = 0.95, y = 0.95, tb = list(tb))

ggplot(mtcars,
       aes(wt, mpg, colour = factor(cyl))) +
  geom_point() +
  geom_table_npc(data = dfnpc,
                 aes(npcx = x, npcy = y, label = tb))

}
\references{
This geometry is inspired on answers to two questions in
  Stackoverflow. In contrast to these earlier examples, the current geom
  obeys the grammar of graphics, and attempts to be consistent with the
  behaviour of 'ggplot2' geometries.
  \url{https://stackoverflow.com/questions/12318120/adding-table-within-the-plotting-region-of-a-ggplot-in-r}
  \url{https://stackoverflow.com/questions/25554548/adding-sub-tables-on-each-panel-of-a-facet-ggplot-in-r?}
}
\seealso{
Formatting of tables \code{stat_fmt_table},
  \code{\link{ttheme_gtdefault}}, \code{\link{ttheme_set}},
  \code{\link[gridExtra]{tableGrob}}.

Other geometries adding layers with insets:
\code{\link[=geom_plot]{geom_plot()}}
}
\concept{geometries adding layers with insets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat-format-table.R
\name{stat_fmt_tb}
\alias{stat_fmt_tb}
\title{Select and slice a tibble nested in \code{data}}
\usage{
stat_fmt_tb(
  mapping = NULL,
  data = NULL,
  geom = "table",
  tb.vars = NULL,
  tb.rows = NULL,
  tb.funs = list(),
  digits = 3,
  position = "identity",
  table.theme = NULL,
  table.rownames = FALSE,
  table.colnames = TRUE,
  table.hjust = 0.5,
  parse = FALSE,
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = TRUE,
  ...
)
}
\arguments{
\item{mapping}{The aesthetic mapping, usually constructed with
\code{\link[ggplot2]{aes}} or \code{\link[ggplot2]{aes_}}. Only needs to be
set at the layer level if you are overriding the plot defaults.}

\item{data}{A layer specific dataset - only needed if you want to override
the plot defaults.}

\item{geom}{The geometric object to use display the data}

\item{tb.vars, tb.rows}{character or numeric vectors, optionally named, used
to select and/or rename the columns or rows in the table
returned.}

\item{tb.funs}{named list of functions to be applied to \code{data} columns.}

\item{digits}{integer indicating the number of significant digits to be
retained in data. Use \code{digits = Inf} to skip.}

\item{position}{The position adjustment to use for overlapping points on this
layer}

\item{table.theme}{NULL, list or function A 'gridExtra' \code{ttheme}
definition, or a constructor for a \code{ttheme} or \code{NULL} for
default.}

\item{table.rownames, table.colnames}{logical flag to enable or disabling
printing of row names and column names.}

\item{table.hjust}{numeric Horizontal justification for the core and column
headings of the table.}

\item{parse}{If \code{TRUE}, the labels will be parsed into expressions and
displayed as described in \code{?plotmath}.}

\item{na.rm}{a logical indicating whether \code{NA} values should be stripped
before the computation proceeds.}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA}, the default, includes if any aesthetics are mapped. \code{FALSE}
never includes, and \code{TRUE} always includes.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics, rather
than combining with them. This is most useful for helper functions that
define both data and aesthetics and shouldn't inherit behaviour from the
default plot specification, e.g. \code{\link[ggplot2]{borders}}.}

\item{...}{other arguments passed on to \code{\link[ggplot2]{layer}}. This
can include aesthetics whose values you want to set, not map. See
\code{\link[ggplot2]{layer}} for more details.}
}
\value{
A plot layer instance. Using as output \code{data} a copy of the
  input \code{data} in which the data frames mapped to \code{label} have been
  modified.
}
\description{
\code{stat_fmt_tb} selects, reorders and/or renames columns and
  or rows of a tibble nested in \code{data}. It can also apply user supplied
  functions to data columns. This stat is intended to be used to pre-process
  \code{tibble} objects mapped to the \code{label} aesthetic before adding
  them to a plot with \code{geom_table}.
}
\details{
One or more functions to be applied can be passed in a named list to
  parameter `tb.funs`. Functions are matched by name to columns, after column
  selection and renaming have been applied.
}
\section{Computed variables}{
 The output of sequentially applying
  \code{\link[dplyr]{slice}} with \code{tb.rows} as argument and
  \code{\link[dplyr]{select}} with \code{tb.vars} to a list variable
  list mapped to \code{label} and containing a single tibble per row
  in \code{data}.
}

\examples{
my.df <-
  tibble::tibble(
    x = c(1, 2),
    y = c(0, 4),
    group = c("A", "B"),
    tbs = list(a = tibble::tibble(Xa = 1:6, Y = rep(c("x", "y"), 3)),
               b = tibble::tibble(Xb = 1:3, Y = "x"))
  )

ggplot(my.df, aes(x, y, label = tbs)) +
  stat_fmt_tb() +
  expand_limits(x = c(0,3), y = c(-2, 6))

# Hide column names, diplay row names
ggplot(my.df, aes(x, y, label = tbs)) +
  stat_fmt_tb(table.colnames = FALSE,
              table.rownames = TRUE) +
  expand_limits(x = c(0,3), y = c(-2, 6))

# Use a theme for the table
ggplot(my.df, aes(x, y, label = tbs)) +
  stat_fmt_tb(table.theme = ttheme_gtlight) +
  expand_limits(x = c(0,3), y = c(-2, 6))

# selection and renaming by column position
ggplot(my.df, aes(x, y, label = tbs)) +
  stat_fmt_tb(tb.vars = c(value = 1, group = 2),
               tb.rows = 1:3) +
  expand_limits(x = c(0,3), y = c(-2, 6))

# apply functions to columns
ggplot(my.df, aes(x, y, label = tbs)) +
  stat_fmt_tb(tb.vars = c(value = 1, group = 2),
               tb.rows = 1:3,
               tb.funs = list(group = function(x) {sprintf("italic(\%s)", x)},
                              value = function(x) {ifelse(x > 2, "bold(zz)", x)}),
               parse = TRUE) +
  expand_limits(x = c(0,3), y = c(-2, 6))

# selection, reordering and renaming by column position
ggplot(my.df, aes(x, y, label = tbs)) +
  stat_fmt_tb(tb.vars = c(group = 2, value = 1),
              tb.rows = 1:3) +
  expand_limits(x = c(0,3), y = c(-2, 6))

# selection and renaming, using partial matching to column name
ggplot(my.df, aes(x, y, label = tbs)) +
  stat_fmt_tb(tb.vars = c(value = "X", group = "Y"),
              tb.rows = 1:3) +
  expand_limits(x = c(0,3), y = c(-2, 6))

}
\seealso{
See \code{\link{geom_table}} for details on how tables respond
  to mapped aesthetics and table themes. For details on predefined table
  themes see \code{\link{ttheme_gtdefault}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat-functions.R
\name{stat_functions}
\alias{stat_functions}
\title{Draw functions as curves}
\usage{
stat_functions(
  mapping = NULL,
  data = NULL,
  n = 101,
  geom = "line",
  position = "identity",
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE,
  ...
)
}
\arguments{
\item{mapping}{The aesthetic mapping, usually constructed with
\code{\link[ggplot2]{aes}} or \code{\link[ggplot2]{aes_}}. Only needs to be
set at the layer level if you are overriding the plot defaults.}

\item{data}{A layer specific dataset. Useful if the function curve is to be
overlaid on other layers.}

\item{n}{integer Number of points to interpolate along the x axis.}

\item{geom}{The geometric object to use display the data}

\item{position}{The position adjustment to use on this layer}

\item{na.rm}{a logical indicating whether \code{NA} values should be stripped
before the computation proceeds.}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA}, the default, includes it if any aesthetics are mapped.
\code{FALSE} never includes, and \code{TRUE} always includes.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics, rather
than combining with them.}

\item{...}{other arguments passed on to \code{\link[ggplot2]{layer}}. This
can include aesthetics whose values you want to set, not map. See
\code{\link[ggplot2]{layer}} for more details.}
}
\value{
A plot layer instance.
}
\description{
\code{stat_functions()} computes values from functions and returns new data
containing numeric vectors for \code{x} and \code{y}. As function definitions
are passed through \code{data} this statistic follows the grammar of graphics in
its behaviour.
}
\details{
This statistic can be used to plot values computed by functions. As
  it follows the grammar of graphics, grouping and facets are supported. In
  this it differs from \code{\link[ggplot2]{geom_function}} which behaves
  like a plot annotation.

  Aesthetics \code{xmin} and \code{xmax} should be mapped to numeric values
  defining the range of the vector to be created and passed as argument to
  the function to compute the \code{y} values, and returned as \code{x} in
  data. \code{n} is the length of this \code{x} vector.
}
\section{Computed variables}{
 Data frame with \code{n} rows or a multiple of
  this, one for each
  row in \code{data}. \describe{
  \item{x}{numeric vector}
  \item{y}{numeric vactor}
  \item{idx}{integer vector, with values corresponding to rows in the input
      \code{data}, i.e., for each function}
  }
  As shown in one example below \code{\link[gginnards]{geom_debug}} can be
  used to print the computed values returned by any statistic. The output
  shown includes also values mapped to aesthetics.
}

\examples{

# one function

df1 <- data.frame(min = 0, max = pi, fun = I(list(sin)))

ggplot(df1, aes(xmin = min, xmax = max, y = fun)) +
  stat_functions()

ggplot(df1, aes(xmin = min, xmax = max, y = fun)) +
  stat_functions(geom = "point", n = 20)

# two functions

df2 <- data.frame(min = -pi, max = pi,
                  fun = I(list(sin, cos)), name = c("sin", "cos"))

# each function must be in a separate group for correct plotting of lines

ggplot(df2, aes(xmin = min, xmax = max, y = fun, group = after_stat(idx))) +
  stat_functions()

ggplot(df2, aes(xmin = min, xmax = max, y = fun, colour = name)) +
  stat_functions()

ggplot(df2, aes(xmin = min, xmax = max, y = fun)) +
  stat_functions() +
  facet_grid(~ name)

# two curves with same function

df3 <- data.frame(min = c(-pi, 0),
                  max = c(0,pi),
                  fun = I(list(sin, sin)),
                  name = c("negative", "positive"))

ggplot(df3, aes(xmin = min, xmax = max, y = fun, colour = name)) +
  stat_functions()

# We use geom_debug_group() to see the computed values

gginnards.installed <- requireNamespace("gginnards", quietly = TRUE)
if (gginnards.installed) {
  library(gginnards)

  ggplot(df1, aes(xmin = min, xmax = max, y = fun)) +
    stat_functions(geom = "debug_group")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dark-or-light.R
\name{dark_or_light}
\alias{dark_or_light}
\title{Chose between dark and light color}
\usage{
dark_or_light(
  colors,
  threshold = 0.45,
  dark.color = "black",
  light.color = "white"
)
}
\arguments{
\item{colors}{A vector of color definitions or color names in the background.}

\item{threshold}{numeric A value of luminance in [0..1] indicating the switch
point between dark and light background.}

\item{dark.color, light.color}{A color definition or color name to return as
dark and light colors to contrast light and dark backgrounds respectively.}
}
\description{
Chose between a pair of contrasting dark and light colors based on a weighted
mean of RGB channels of a color. This function implements a simple approach
to the choice for a color of a plot element to ensure it is visible against a
background color.
}
\details{
The switch between dark and light color is based on a quick and
  dirty approximation of the luminance of colors computed from RGB values.
  This easily computed approximation seems to work well enough. The default
  threshold chosen for a switch between black and white may need to be
  adjusted for other pairs of colors. Graphic devices can differ in the color
  spaces they support, but this is unlikely to affect the choice between
  black and white or other pairs of colors with large differences in
  luminance.
}
\note{
The current implementation of \code{dark_or_light()} ignores
  \code{alpha}, the transparency component, of all its arguments.
}
\examples{

dark_or_light("yellow")
dark_or_light("darkblue")
dark_or_light(c("darkblue", "yellow", "red"))
dark_or_light("#FFFFFF")
dark_or_light("#FFFFFF", dark.color = "darkblue", light.color = "lightgrey")
dark_or_light("#000000", dark.color = "darkblue", light.color = "lightgrey")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/position-dodge-nudge-to.R
\name{position_nudge_to}
\alias{position_nudge_to}
\title{Nudge labels to new positions}
\usage{
position_nudge_to(
  x = NULL,
  y = NULL,
  x.action = c("none", "spread"),
  y.action = c("none", "spread"),
  x.distance = "equal",
  y.distance = "equal",
  x.expansion = 0,
  y.expansion = 0,
  kept.origin = c("original", "none")
)
}
\arguments{
\item{x, y}{Coordinates of the destination position. A vector of mode
\code{numeric}, that is extended if needed, to the same length as rows
there are in \code{data}. The default, \code{NULL}, leaves the original
coordinates unchanged.}

\item{x.action, y.action}{character string, one of \code{"none"}, or
\code{"spread"}. With \code{"spread"} distributing the positions
within the range of argument \code{x} or \code{y}, if non-null. Otherwise,
using the range the variable mapped to \emph{x} or \code{y}.}

\item{x.distance, y.distance}{character or numeric Currently only
\code{"equal"} is implemented, indicating equal spacing between the
spread positions.}

\item{x.expansion, y.expansion}{numeric vectors of length 1 or 2, as a
fraction of width of the range used to spread positions.}

\item{kept.origin}{One of \code{"original"} or \code{"none"}.}
}
\value{
A \code{"Position"} object.
}
\description{
\code{position_nudge_to()} differs from \code{\link[ggplot2]{position_nudge}}
in that the coordinates of the new position are given directly, rather than
as a displacement from the original location. It optionally sets an even
spacing among positions within a range. As with other position functions in
this package, the original positions are preserved to allow the text or
labels to be linked back to their original position with a segment or arrow.
}
\details{
The nudged to
  \code{x} and/or \code{y} values replace the original ones in
  \code{data}, while the original coordinates are returned in
  \code{x_orig} and \code{y_orig}. Nudge values supported are those of
  \emph{mode} numeric, thus including dates and times when they match the
  mapped data.

  If the length of \code{x} and/or \code{y} is more than one but less than
  the rows present in the \code{data}, the vector is both recycled and
  reordered so that the nudges are applied sequentially based on the data
  values. If their length matches the number of rows in \code{data}, they are
  assumed to be already in \code{data} order.

  Irrespective of the action, the ordering of rows in \code{data} is
  preserved.
}
\note{
The current implementation DOES NOT support flipping geoms with the
\code{orientation} argument or implicitly by the mapping. It DOES NOT
apply scale transformations when spreading the positions.
}
\examples{
# The examples below exemplify the features of position_nudge_to().
# Please see the vignette for examples of use cases.

df <- data.frame(
  x = c(1,3,2,5,4,2.5),
  y = c(2, 3, 2.5, 1.8, 2.8, 1.5),
  grp = c("A", "A", "A", "B", "B", "B"),
  grp.inner = c("a", "b", "c", "a", "b", "c"),
  label = c("abc","cd","d","c","bcd","a")
)

# default is no nudging
ggplot(df, aes(label, y, label = y)) +
  geom_col() +
  geom_text(position = position_nudge_to(),
            vjust = -0.2)

# a single y (or x) value nudges all observations to this data value
ggplot(df, aes(label, y, label = y)) +
  geom_col() +
  geom_label(position = position_nudge_to(y = 1))

ggplot(df, aes(x, y, label = label)) +
  geom_point() +
  geom_text(position = position_nudge_to(y = 3.2))

ggplot(df, aes(x, y, label = label)) +
  geom_point() +
  geom_line() +
  geom_text(position = position_nudge_to(y = 0.1))

# with a suitable geom, segments or arrows can be added
ggplot(df, aes(x, y, label = label)) +
  geom_point() +
  geom_text_s(position = position_nudge_to(y = 2.25))

# alternating in y value order because y has fewer values than rows in data
ggplot(df, aes(x, y, label = label)) +
  geom_point() +
  geom_text_s(position = position_nudge_to(y = c(3, 0.1)))

# in data row order with as many nudge y values as rows in data
ggplot(df, aes(x, y, label = label)) +
  geom_point() +
  geom_text_s(position = position_nudge_to(y = c(1.8, 2.3, 1.3, 2.8, 3, 0.1)))

# spread the values at equal distance within the available space
ggplot(df, aes(x, y, label = label)) +
  geom_point() +
  geom_text_s(position =
                position_nudge_to(y = 4, x.action = "spread"))

ggplot(df, aes(x, y, label = label)) +
  geom_point() +
  geom_text_s(position =
                position_nudge_to(y = 4, x.action = "spread")) +
scale_x_log10()

# spread the values at equal distance within the expanded available space
ggplot(df, aes(x, y, label = label)) +
  geom_point() +
  geom_text_s(position =
                position_nudge_to(y = 4, x.action = "spread", x.expansion = 0.1))

# spread the values at equal distance within the range given by x
ggplot(df, aes(x, y, label = label)) +
  geom_point() +
  geom_text_s(position =
                position_nudge_to(y = 4, x = c(1.5, 4), x.action = "spread"))

# currently if scale transformations are used, the x and/or y arguments must
# be transformed. WARNING: This will change in the near future!!

ggplot(df, aes(x, y, label = label)) +
  geom_point() +
  geom_text_s(position =
                position_nudge_to(y = 4, x = log10(c(1.5, 4)), x.action = "spread")) +
scale_x_log10()

ggplot(df, aes(x, y, label = label)) +
  geom_point() +
  geom_text_s(position =
                position_nudge_to(y = log10(4), x.action = "spread")) +
scale_y_log10()

}
\seealso{
\code{\link[ggplot2]{position_nudge}},
  \code{\link[ggrepel]{position_nudge_repel}}.

Other position adjustments:
\code{\link[=position_dodgenudge]{position_dodgenudge()}},
\code{\link[=position_dodgenudge_to]{position_dodgenudge_to()}},
\code{\link[=position_jitternudge]{position_jitternudge()}},
\code{\link[=position_nudge_center]{position_nudge_center()}},
\code{\link[=position_nudge_keep]{position_nudge_keep()}},
\code{\link[=position_nudge_line]{position_nudge_line()}},
\code{\link[=position_stacknudge]{position_stacknudge()}},
\code{\link[=position_stacknudge_to]{position_stacknudge_to()}}
}
\concept{position adjustments}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-grob.R, R/geom-label-linked.r,
%   R/geom-label-npc.r, R/geom-label-pairwise.r, R/geom-margin-arrow.r,
%   R/geom-margin-grob.r, R/geom-margin-point.r, R/geom-plot.R,
%   R/geom-point-linked.r, R/geom-quadrant-lines.R, R/geom-table.R,
%   R/geom-text-linked.r, R/geom-text-npc.r, R/geom-text-pairwise.R,
%   R/position-dodge-nudge-to.R, R/position-dodge2nudge-to.R,
%   R/position-nudge-center.R, R/position-nudge-line.R,
%   R/position-stack-nudge.R, R/position-stacknudge-to.R, R/stat-apply.R,
%   R/stat-dens1d-filter.r, R/stat-dens1d-labels.r, R/stat-dens2d-filter.r,
%   R/stat-dens2d-labels.r, R/stat-format-table.R, R/stat-functions.R,
%   R/stat-panel-counts.R, R/stat-quadrant-counts.R
\name{GeomGrob}
\alias{GeomGrob}
\alias{GeomGrobNpc}
\alias{GeomLabelS}
\alias{GeomLabelNpc}
\alias{GeomLabelPairwise}
\alias{GeomXMarginArrow}
\alias{GeomYMarginArrow}
\alias{GeomXMarginGrob}
\alias{GeomYMarginGrob}
\alias{GeomXMarginPoint}
\alias{GeomYMarginPoint}
\alias{GeomPlot}
\alias{GeomPlotNpc}
\alias{GeomPointS}
\alias{GeomQuadrantLines}
\alias{GeomVHLines}
\alias{GeomTable}
\alias{GeomTableNpc}
\alias{GeomTextS}
\alias{GeomTextNpc}
\alias{GeomTextPairwise}
\alias{PositionDodgeNudgeTo}
\alias{PositionDodge2AndNudgeTo}
\alias{PositionNudgeCenter}
\alias{PositionNudgeLine}
\alias{PositionFillAndNudge}
\alias{PositionStackNudgeTo}
\alias{PositionFillNudgeTo}
\alias{Stats}
\alias{StatApplyGroup}
\alias{StatDens1dFilter}
\alias{StatDens1dFilterG}
\alias{StatDens1dLabels}
\alias{StatDens2dFilter}
\alias{StatDens2dFilterG}
\alias{StatDens2dLabels}
\alias{StatFmtTb}
\alias{StatFunctions}
\alias{StatPanelCounts}
\alias{StatGroupCounts}
\alias{StatQuadrantCounts}
\title{\code{Stat*} Objects}
\description{
All \code{stat_*} functions (like \code{stat_bin}) return a layer that
contains a \code{Stat*} object (like \code{StatBin}). The \code{Stat*}
object is responsible for rendering the data in the plot.
}
\details{
Each of the \code{Stat*} objects is a \code{\link[ggplot2]{ggproto}} object, descended
from the top-level \code{Stat}, and each implements various methods and
fields. To create a new type of Stat object, you typically will want to
implement one or more of the following:
}
\seealso{
\code{\link[ggplot2]{ggplot2-ggproto}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/position-stacknudge-to.R
\name{position_stacknudge_to}
\alias{position_stacknudge_to}
\alias{position_fillnudge_to}
\title{Stack plus nudge labels to new positions}
\usage{
position_stacknudge_to(
  vjust = 1,
  reverse = FALSE,
  x = NULL,
  y = NULL,
  x.action = c("none", "spread"),
  y.action = c("none", "spread"),
  x.distance = "equal",
  y.distance = "equal",
  x.expansion = 0,
  y.expansion = 0,
  kept.origin = c("stacked", "original", "none")
)

position_fillnudge_to(
  vjust = 1,
  reverse = FALSE,
  x = NULL,
  y = NULL,
  x.action = c("none", "spread"),
  y.action = c("none", "spread"),
  x.distance = "equal",
  y.distance = "equal",
  x.expansion = 0,
  y.expansion = 0,
  kept.origin = c("stacked", "original", "none")
)
}
\arguments{
\item{vjust}{Vertical adjustment for geoms that have a position (like points
or lines), not a dimension (like bars or areas). Set to 0 to align with the
bottom, 0.5 for the middle, and 1 (the default) for the top.}

\item{reverse}{If TRUE, will reverse the default stacking order. This is
useful if you're rotating both the plot and legend.}

\item{x, y}{Coordinates of the destination position. A vector of mode
\code{numeric}, that is extended if needed, to the same length as rows
there are in \code{data}. The default, \code{NULL}, leaves the original
coordinates unchanged after dodging.}

\item{x.action, y.action}{character string, one of \code{"none"}, or
\code{"spread"}. With \code{"spread"} distributing the positions
within the range of argument \code{x} or \code{y}, if non-null, or the
range the variable mapped to \emph{x} or \code{y}, otherwise.}

\item{x.distance, y.distance}{character or numeric Currently only
\code{"equal"} is implemented.}

\item{x.expansion, y.expansion}{numeric vectors of length 1 or 2, as a
fraction of width of the range.}

\item{kept.origin}{One of \code{"original"}, \code{"stacked"} or
\code{"none"}.}
}
\value{
A \code{"Position"} object.
}
\description{
Functions \code{position_stacknudge_to()} and \code{position_fillnudge_to()}
are meant to complement \code{position_stack()} and \code{position_fill()}
from 'ggplot2', adding as a second action that of \code{position_nudge_to()}.
These positions are generally useful for adjusting the position of labels or
text. As with other position functions in this package, the original
positions are preserved to allow the text or labels to be linked back to
their original position with a segment or arrow.
}
\details{
These positions apply sequentially two actions, in the order they
  appear in their names. The applied stacking is similar to that by
  \code{\link[ggplot2]{position_stack}} and
  \code{\link[ggplot2]{position_fill}} while nudging is different to that
  by \code{\link[ggplot2]{position_nudge}} and equal to that applied by
  \code{\link{position_nudge_to}}.

  The nudged to \code{x} and/or \code{y} values replace the original
  ones in \code{data}, while the original or the stacked coordinates are
  returned in \code{x_orig} and \code{y_orig}. Nudge values supported are
  those of \emph{mode} numeric, thus including dates and times when they
  match the mapped data.

  If the length of \code{x} and/or \code{y} is more than one but less than
  rows are present in the data, the vector is both recycled and reordered so
  that the nudges are applied sequentially based on the data values. If their
  length matches the number of rows in data, they are assumed to be already
  in data order.

  When applying stacking, the return of original positions instead of the
  stacked ones is achieved by passing \code{origin = "original"} instead of
  the default of \code{origin = "stacked"}.
}
\note{
Irrespective of the action, the ordering of rows in \code{data} is
  preserved.
}
\examples{

cols.df <- data.frame(x1 = c(1, 2, 1, 3),
                       x2 = c("a", "a", "b", "b"),
                       grp = c("A", "B", "C", "D"))

 ggplot(data = cols.df, aes(x2, x1, group = grp)) +
 geom_col(aes(fill = grp), width=0.5, position = "stack") +
   geom_vline(xintercept = 0) +
   geom_text_s(
     aes(label = x1),
     position = position_stacknudge_to(x = c(1.4, 1.4, 1.6, 1.6))) +
   theme(legend.position = "none")

 ggplot(data = cols.df, aes(x2, x1, group = grp)) +
 geom_col(aes(fill = grp), width=0.5, position = "fill") +
   geom_vline(xintercept = 0) +
   geom_text_s(
     aes(label = grp),
     position = position_fillnudge_to(x = c(1.4, 1.4, 1.6, 1.6))) +
   theme(legend.position = "none")

 ggplot(data = cols.df, aes(x2, x1, group = grp)) +
 geom_col(aes(fill = grp), width=0.5, position = "fill") +
   geom_vline(xintercept = 0) +
   geom_text_s(
     aes(label = x1),
     position = position_fillnudge_to(vjust = 0.5,
                                      x = c(1.4, 1.4, 1.6, 1.6))) +
   theme(legend.position = "none")

 ggplot(data = cols.df, aes(x2, x1, group = grp)) +
 geom_col(aes(fill = grp), width=0.5, position = "fill") +
   geom_vline(xintercept = 0) +
   geom_text_s(
     aes(label = x1), vjust = 0.5,
     position = position_fillnudge_to(vjust = 0.5,
                                      x = c(1.4, 1.4, 1.6, 1.6),
                                      y = c(0.85, 0.35, 0.85, 0.35))) +
   theme(legend.position = "none")

}
\seealso{
\code{\link{position_nudge_to}},
  \code{\link[ggplot2]{position_stack}}.

Other position adjustments:
\code{\link[=position_dodgenudge]{position_dodgenudge()}},
\code{\link[=position_dodgenudge_to]{position_dodgenudge_to()}},
\code{\link[=position_jitternudge]{position_jitternudge()}},
\code{\link[=position_nudge_center]{position_nudge_center()}},
\code{\link[=position_nudge_keep]{position_nudge_keep()}},
\code{\link[=position_nudge_line]{position_nudge_line()}},
\code{\link[=position_nudge_to]{position_nudge_to()}},
\code{\link[=position_stacknudge]{position_stacknudge()}}
}
\concept{position adjustments}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat-panel-counts.R
\name{stat_panel_counts}
\alias{stat_panel_counts}
\alias{stat_group_counts}
\title{Number of observations in a plot panel}
\usage{
stat_panel_counts(
  mapping = NULL,
  data = NULL,
  geom = "text_npc",
  position = "identity",
  label.x = "right",
  label.y = "top",
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = TRUE,
  ...
)

stat_group_counts(
  mapping = NULL,
  data = NULL,
  geom = "text_npc",
  position = "identity",
  label.x = "right",
  label.y = "top",
  hstep = 0,
  vstep = NULL,
  digits = 2,
  na.rm = FALSE,
  show.legend = FALSE,
  inherit.aes = TRUE,
  ...
)
}
\arguments{
\item{mapping}{The aesthetic mapping, usually constructed with
\code{\link[ggplot2]{aes}} or \code{\link[ggplot2]{aes_}}. Only needs to be
set at the layer level if you are overriding the plot defaults.}

\item{data}{A layer specific dataset. Rarely used, as you will not want to
override the plot defaults.}

\item{geom}{The geometric object to use display the data}

\item{position}{The position adjustment to use on this layer}

\item{label.x, label.y}{\code{numeric} Coordinates (in npc units) to be used
for absolute positioning of the labels.}

\item{na.rm}{a logical indicating whether \code{NA} values should be stripped
before the computation proceeds.}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA}, the default, includes it if any aesthetics are mapped.
\code{FALSE} never includes, and \code{TRUE} always includes.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics, rather
than combining with them. This is most useful for helper functions that
define both data and aesthetics and should not inherit behaviour from the
default plot specification, e.g., \code{\link[ggplot2]{borders}}.}

\item{...}{other arguments passed on to \code{\link[ggplot2]{layer}}. This
can include aesthetics whose values you want to set, not map. See
\code{\link[ggplot2]{layer}} for more details.}

\item{hstep, vstep}{numeric in npc units, the horizontal and vertical step
used between labels for different groups.}

\item{digits}{integer Number of digits for fraction and percent labels.}
}
\value{
A plot layer instance. Using as output \code{data} the counts of
  observations in each plot panel or per group in each plot panel.
}
\description{
\code{stat_panel_counts()} counts the number of observations in each panel.
\code{stat_group_counts()} counts the number of observations in each group.
By default they add one or more text labels to the top right corner of each
panel. Grouping is ignored by \code{stat_panel_counts()}. If no grouping
exists, the two statistics behave similarly.
}
\details{
These statistics can be used to automatically count observations in
  each panel of a plot, and by default add these counts as text labels. These
  statistics, unlike \code{stat_quadrant_counts()} requires only one of
  \emph{x} or \emph{y} aesthetics and can be used together with statistics
  that have the same requirement, like \code{stat_density()}.

  The default position of the label is in the top right corner. When using
  facets even with free limits for \emph{x} and \emph{y} axes, the location
  of the labels is consistent across panels. This is achieved by use of
  \code{geom = "text_npc"} or \code{geom = "label_npc"}. To pass the
  positions in native data units to \code{label.x} and \code{label.y}, pass
  also explicitly \code{geom = "text"}, \code{geom = "label"} or some other
  geometry that use the \emph{x} and/or \emph{y} aesthetics. A vector with
  the same length as the number of panels in the figure can be used if
  needed.
}
\note{
If a factor is mapped to \code{x} or to \code{y} aesthetics each level
  of the factor constitutes a group, in this case the default positioning and
  geom using NPC pseudo aesthetics will have to be overriden by passing
  \code{geom = "text"} and data coordinates used. The default for factors
  may change in the future.
}
\section{Computed variables}{
 Data frame with one or more rows, one for each
  group of observations for which counts are counted in \code{data}. \describe{
  \item{x,npcx}{x value of label position in data- or npc units, respectively}
  \item{y,npcy}{y value of label position in data- or npc units, respectively}
  \item{count}{number of  observations as an integer}
  \item{count.label}{number of observations as character}}

  As shown in one example below \code{\link[gginnards]{geom_debug}} can be
  used to print the computed values returned by any statistic. The output
  shown includes also values mapped to aesthetics, like \code{label} in the
  example. \code{x} and \code{y} are included in the output only if mapped.
}

\examples{

# generate artificial data with numeric x and y
set.seed(67821)
x <- 1:100
y <- rnorm(length(x), mean = 10)
group <- factor(rep(c("A", "B"), times = 50))
my.data <- data.frame(x, y, group)

# using automatically generated text labels

ggplot(my.data, aes(x, y)) +
  geom_point() +
  stat_panel_counts()

ggplot(my.data, aes(x, y, colour = group)) +
  geom_point() +
  stat_panel_counts()

ggplot(my.data, aes(x, y, colour = group)) +
  geom_point() +
  stat_group_counts()

ggplot(my.data, aes(x, y, colour = group)) +
  geom_point() +
  stat_group_counts(label.x = "left", hstep = 0.06, vstep = 0)

ggplot(my.data, aes(x, y, colour = group)) +
  geom_point() +
  stat_group_counts(aes(label = after_stat(pc.label)))

ggplot(my.data, aes(x, y, colour = group)) +
  geom_point() +
  stat_group_counts(aes(label = after_stat(pc.label)), digits = 3)

ggplot(my.data, aes(x, y, colour = group)) +
  geom_point() +
  stat_group_counts(aes(label = after_stat(fr.label)))

ggplot(my.data, aes(x, y, colour = group)) +
  geom_point() +
  stat_group_counts(aes(label = after_stat(dec.label)))

# one of x or y can be a factor
# label.x or label.y along the factor can be set to "factor" together
# with the use of geom_text()

ggplot(mpg,
       aes(factor(cyl), hwy)) +
  stat_boxplot() +
  stat_group_counts(geom = "text",
                    label.y = 10,
                    label.x = "factor") +
  stat_panel_counts()

# Numeric values can be used to build labels with alternative formats
# Here with sprintf(), but paste() and format() also work.

ggplot(my.data, aes(x, y)) +
 geom_point() +
 stat_panel_counts(aes(label = sprintf("\%i observations",
                                       after_stat(count)))) +
 scale_y_continuous(expand = expansion(mult = c(0.05, 0.12)))

ggplot(mpg,
       aes(factor(cyl), hwy)) +
  stat_boxplot() +
  stat_group_counts(geom = "text",
                    aes(label = sprintf("(\%i)", after_stat(count))),
                    label.y = 10,
                    label.x = "factor")

ggplot(mpg,
       aes(factor(cyl), hwy)) +
  stat_boxplot() +
  stat_group_counts(aes(label = sprintf("n[\%i]~`=`~\%i",
                                        after_stat(x), after_stat(count))),
                    parse = TRUE,
                    geom = "text",
                    label.y = 10,
                    label.x = "factor") +
  stat_panel_counts(aes(label = sprintf("sum(n[i])~`=`~\%i",
                                        after_stat(count))),
                    parse = TRUE)

# label position

ggplot(my.data, aes(y)) +
  stat_panel_counts(label.x = "left") +
  stat_density(alpha = 0.5)

ggplot(my.data, aes(y, colour = group)) +
  stat_group_counts(label.y = "top") +
  stat_density(aes(fill = group), alpha = 0.3)

# The numeric value can be used as a label as is

ggplot(mpg,
       aes(factor(cyl), hwy)) +
  stat_boxplot() +
  stat_group_counts(geom = "text",
                    aes(label = after_stat(count)),
                    label.x = "factor",
                    label.y = 10) +
  annotate(geom = "text", x = 0.55, y = 10, label = "n[i]~`=`", parse = TRUE)

# We use geom_debug_group() to see the computed values

gginnards.installed <- requireNamespace("gginnards", quietly = TRUE)
if (gginnards.installed) {
  library(gginnards)

  ggplot(my.data, aes(x, y)) +
    geom_point() +
    stat_panel_counts(geom = "debug_group")
}

if (gginnards.installed) {
  ggplot(my.data, aes(x, y, colour = group)) +
    geom_point() +
    stat_group_counts(geom = "debug_group")
}

}
\seealso{
Other Functions for quadrant and volcano plots:
\code{\link[=geom_quadrant_lines]{geom_quadrant_lines()}},
\code{\link[=stat_quadrant_counts]{stat_quadrant_counts()}}
}
\concept{Functions for quadrant and volcano plots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scale-continuous-npc.r
\name{scale_continuous_npc}
\alias{scale_continuous_npc}
\alias{scale_npcx_continuous}
\alias{scale_npcy_continuous}
\title{Position scales for continuous data (npcx & npcy)}
\usage{
scale_npcx_continuous(...)

scale_npcy_continuous(...)
}
\arguments{
\item{...}{Other arguments passed on to \code{continuous_scale()}}
}
\value{
A \code{"Scale"} object.
}
\description{
\code{scale_npcx_continuous()} and \code{scale_npcy_continuous()} are
scales for continuous npcx and npcy aesthetics expressed in "npc" units.
There are no variants. Obviously limits are always the full range of "npc"
units and transformations meaningless. These scales are used by the newly
defined aesthetics \code{npcx} and \code{npcy}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{check_default_colour}
\alias{check_default_colour}
\title{Check default.colour argument}
\usage{
check_default_colour(default.colour, theme = NULL)
}
\arguments{
\item{default.colour}{The value to be checked}

\item{theme}{The theme from where to fetch a replacement value when
\code{default.colour} is \code{NULL}.}
}
\value{
A colour definition or NA
}
\description{
if possible use ink from theme as default colour
}
\details{
The replacement value is fetched from the theme only if
  'ggplot2' >= 4.0.0 and otherwise \code{"black"} is used. \code{NA}
  values are passed through as they can be used to disable rendering
  of grob elements.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute-npc.r
\name{compute_npcx}
\alias{compute_npcx}
\alias{compute_npcy}
\alias{as_npcx}
\alias{as_npcy}
\alias{compute_npc}
\alias{as_npc}
\title{Compute NPC coordinates}
\usage{
compute_npcx(x, group = 1L, h.step = 0.1, margin.npc = 0.05, each.len = 1)

compute_npcy(y, group = 1L, v.step = 0.1, margin.npc = 0.05, each.len = 1)

as_npcx(x, ...)

as_npcy(y, ...)

compute_npc(a, margin.npc = 0.05)

as_npc(a, margin.npc = 0.05)
}
\arguments{
\item{x}{numeric or if character, one of "right", "left", "maximum",
"minimum", "centre", "center" or "middle".}

\item{group}{integer vector, ggplot's group id. Used to shift coordinates to
avoid overlaps.}

\item{h.step, v.step}{numeric [0..1] The step size for shifting coordinates
in npc units. Usually << 1.}

\item{margin.npc}{numeric [0..1] The margin added towards the nearest
plotting area edge when converting character coordinates into npc. Usually
<< 1.}

\item{each.len}{integer The number of steps per group.}

\item{y}{numeric or if character, one of "top", "bottom", "maximum",
"minimum", "centre", "center" or "middle".}

\item{...}{named arguments passed to \code{compute_npcx()} or
\code{compute_npcy()}.}

\item{a}{numeric or if character, one of "right", "left", "top", "maximum",
"minimum", "bottom", "centre", "center" or "middle".}
}
\value{
A numeric vector with values in the range [0..1] representing
  npc coordinates.
}
\description{
Translate and/or compute NPC (Normalised Parent Coordinates) for use with
aesthetics \code{x} and \code{y}.
}
\details{
Functions \code{compute_npcx} and \code{compute_npcy} convert
  character-encoded positions to npc units and shift positions to avoid
  overlaps when grouping is active. If numeric, they validate the npc values.
  Function \code{compute_npcx} does the translation either for both \code{x}
  and \code{y} aesthetics, but does not implement a shift for grpups.
  Functions \code{as_npcx()},  \code{as_npcy()} and \code{as_npc()} are
  wrappers on these functions that return the value as objects of class
  \code{"AsIs"} so that in 'ggplot2' >= 3.5.0 they can be used with any layer
  function.

  These functions use NPC (Normalised Parent Coordinates) instead of data
  coordinates. They translate named positions into numeric values in [0..1]
  and they can also shift the position according to the group, e.g., for each
  increase in the group number displace the position inwards or outwards, by
  a user-supplied distance. They make it possible to set automatically set
  default positions for grouped text labels.

  Out of bounds numeric values are constrained to [0..1]. Unrecognized
  character values are silently converted into \code{NA_integer_}.
}
\note{
The \emph{as_npc()} functions make it easier the use of NPC coordinates
  with 'ggplot2' >= 3.5.0. The _compute_ functions are used by several layer
  functions in packages 'ggpp' and 'ggpmisc', are compatible with 'ggplot2'
  <= 3.4.4 and can be useful to developers of other 'ggplot2' extensions.
}
\examples{
compute_npcx("right")
compute_npcx(c("left", "right"))
compute_npcx(c("minimum", "maximum"))
compute_npcx(c("left", "right"), margin.npc = 0)
compute_npcy("bottom")
compute_npcy("bottom", group = 1L:3L)
compute_npcy("bottom", group = 1L:3L, v.step = 0.2)
compute_npcy("bottom", group = 2L)
compute_npcx(0.5)
compute_npcx(1)
compute_npcx(-2)

as_npc("right")
class(as_npc("right"))
class(compute_npcx("right"))

}
